/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/core/include/numpy/ufuncobject.h",
            "pyseobnr/eob/utils/eob_parameters.h"
        ],
        "extra_compile_args": [
            "-O3"
        ],
        "include_dirs": [
            "pyseobnr/eob/utils",
            "./pyseobnr/eob/utils",
            "/local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/core/include",
            "pyseobnr/eob/hamiltonian"
        ],
        "name": "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD",
        "sources": [
            "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx"
        ]
    },
    "module_name": "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyseobnr__eob__hamiltonian__Ham_AvgS2precess_simple_cython_PA_AD
#define __PYX_HAVE_API__pyseobnr__eob__hamiltonian__Ham_AvgS2precess_simple_cython_PA_AD
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include "eob_parameters.h"
#include <math.h>
#include "pythread.h"
#include <stdlib.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "stringsource",
  "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx",
  "__init__.pxd",
  "type.pxd",
  "pyseobnr/eob/utils/containers.pxd",
};
/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_incr_aligned(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && CYTHON_COMPILING_IN_NOGIL
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd)
    #define __pyx_atomic_incr_aligned(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_aligned(value) _InterlockedExchangeAdd(value, -1)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_PhysicalParams;
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs;
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_Dynamics;
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_FluxParams;
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_EOBParams;
struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C;
struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "pyseobnr/eob/utils/containers.pxd":8
 * 
 * 
 * cdef class PhysicalParams:             # <<<<<<<<<<<<<<
 *     cdef public double m_1,m_2,nu,M,X_1,X_2,delta,chi_1,chi_2,chi_S,chi_A,a,ap,am
 *     cdef public double chi1_L, chi2_L,a1,a2, H_val
 */
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_PhysicalParams {
  PyObject_HEAD
  struct __pyx_vtabstruct_8pyseobnr_3eob_5utils_10containers_PhysicalParams *__pyx_vtab;
  double m_1;
  double m_2;
  double nu;
  double M;
  double X_1;
  double X_2;
  double delta;
  double chi_1;
  double chi_2;
  double chi_S;
  double chi_A;
  double a;
  double ap;
  double am;
  double chi1_L;
  double chi2_L;
  double a1;
  double a2;
  double H_val;
  __Pyx_memviewslice chi1_v;
  __Pyx_memviewslice chi2_v;
  __Pyx_memviewslice lN;
  double omega;
  double omega_circ;
};


/* "pyseobnr/eob/utils/containers.pxd":23
 *     cdef public double flagNLOSO,flagNLOSO2,flagNLOSO3,flagNLOSS,flagNLOSS2,flagS3
 * '''
 * cdef class CalibCoeffs:             # <<<<<<<<<<<<<<
 *     cdef public dict dc
 * 
 */
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs {
  PyObject_HEAD
  PyObject *dc;
};


/* "pyseobnr/eob/utils/containers.pxd":26
 *     cdef public dict dc
 * 
 * cdef class Dynamics:             # <<<<<<<<<<<<<<
 *     cdef public double[:] p_circ
 * 
 */
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_Dynamics {
  PyObject_HEAD
  __Pyx_memviewslice p_circ;
};


/* "pyseobnr/eob/utils/containers.pxd":29
 *     cdef public double[:] p_circ
 * 
 * cdef class FluxParams:             # <<<<<<<<<<<<<<
 *     # Coefficients entering the amplitude residual
 *     cdef public double[:,:,:] rho_coeffs
 */
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_FluxParams {
  PyObject_HEAD
  __Pyx_memviewslice rho_coeffs;
  __Pyx_memviewslice rho_coeffs_log;
  __Pyx_memviewslice f_coeffs;
  __Pyx_memviewslice f_coeffs_vh;
  __Pyx_memviewslice delta_coeffs;
  __Pyx_memviewslice delta_coeffs_vh;
  __Pyx_memviewslice prefixes;
  __Pyx_memviewslice prefixes_abs;
  __Pyx_memviewslice Tlm;
  __Pyx_memviewslice rholm;
  __Pyx_memviewslice deltalm;
  __Pyx_memviewslice nqc_coeffs;
  __Pyx_memviewslice extra_coeffs;
  __Pyx_memviewslice extra_coeffs_log;
  PyObject *special_modes;
  int rho_initialized;
  int extra_PN_terms;
};


/* "pyseobnr/eob/utils/containers.pxd":67
 *     cdef public bint extra_PN_terms
 * 
 * cdef class EOBParams:             # <<<<<<<<<<<<<<
 *     cdef public PhysicalParams p_params
 *     cdef public CalibCoeffs c_coeffs
 */
struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_EOBParams {
  PyObject_HEAD
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_PhysicalParams *p_params;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *c_coeffs;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_FluxParams *flux_params;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_Dynamics *dynamics;
  PyObject *mode_array;
  int aligned;
};


/* "pyseobnr/eob/hamiltonian/Hamiltonian_v5PHM_C.pxd":4
 * cimport cython
 * from pyseobnr.eob.utils.containers cimport EOBParams
 * cdef class Hamiltonian_v5PHM_C:             # <<<<<<<<<<<<<<
 *     cdef public EOBParams EOBpars
 *     cpdef _call(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL_1,double chiL_2)
 */
struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C {
  PyObject_HEAD
  struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *__pyx_vtab;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_EOBParams *EOBpars;
};


/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":17
 * 
 * 
 * cdef class Ham_AvgS2precess_simple_cython_PA_AD(Hamiltonian_v5PHM_C):             # <<<<<<<<<<<<<<
 *     cdef public CalibCoeffs calibration_coeffs
 *     def __cinit__(self,EOBParams params):
 */
struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD {
  struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C __pyx_base;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *calibration_coeffs;
};


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":280
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "pyseobnr/eob/utils/containers.pxd":8
 * 
 * 
 * cdef class PhysicalParams:             # <<<<<<<<<<<<<<
 *     cdef public double m_1,m_2,nu,M,X_1,X_2,delta,chi_1,chi_2,chi_S,chi_A,a,ap,am
 *     cdef public double chi1_L, chi2_L,a1,a2, H_val
 */

struct __pyx_vtabstruct_8pyseobnr_3eob_5utils_10containers_PhysicalParams {
  PyObject *(*update_spins)(struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_PhysicalParams *, double, double, int __pyx_skip_dispatch);
  PyObject *(*_compute_derived_quants)(struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_PhysicalParams *);
};
static struct __pyx_vtabstruct_8pyseobnr_3eob_5utils_10containers_PhysicalParams *__pyx_vtabptr_8pyseobnr_3eob_5utils_10containers_PhysicalParams;


/* "pyseobnr/eob/hamiltonian/Hamiltonian_v5PHM_C.pxd":4
 * cimport cython
 * from pyseobnr.eob.utils.containers cimport EOBParams
 * cdef class Hamiltonian_v5PHM_C:             # <<<<<<<<<<<<<<
 *     cdef public EOBParams EOBpars
 *     cpdef _call(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL_1,double chiL_2)
 */

struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C {
  PyObject *(*_call)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch);
  PyObject *(*grad)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch);
  PyObject *(*hessian)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch);
  double (*csi)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch);
  PyObject *(*dynamics)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch);
  double (*omega)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch);
  PyObject *(*aux_derivs)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *__pyx_vtabptr_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C;


/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":17
 * 
 * 
 * cdef class Ham_AvgS2precess_simple_cython_PA_AD(Hamiltonian_v5PHM_C):             # <<<<<<<<<<<<<<
 *     cdef public CalibCoeffs calibration_coeffs
 *     def __cinit__(self,EOBParams params):
 */

struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD {
  struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C __pyx_base;
  double (*xi)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double);
  PyObject *(*auxderivs)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_vtabptr_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD;


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* Profile.proto */
#ifndef CYTHON_PROFILE
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
  #define CYTHON_PROFILE 0
#else
  #define CYTHON_PROFILE 1
#endif
#endif
#ifndef CYTHON_TRACE_NOGIL
  #define CYTHON_TRACE_NOGIL 0
#else
  #if CYTHON_TRACE_NOGIL && !defined(CYTHON_TRACE)
    #define CYTHON_TRACE 1
  #endif
#endif
#ifndef CYTHON_TRACE
  #define CYTHON_TRACE 0
#endif
#if CYTHON_TRACE
  #undef CYTHON_PROFILE_REUSE_FRAME
#endif
#ifndef CYTHON_PROFILE_REUSE_FRAME
  #define CYTHON_PROFILE_REUSE_FRAME 0
#endif
#if CYTHON_PROFILE || CYTHON_TRACE
  #include "compile.h"
  #include "frameobject.h"
  #include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #if CYTHON_PROFILE_REUSE_FRAME
    #define CYTHON_FRAME_MODIFIER static
    #define CYTHON_FRAME_DEL(frame)
  #else
    #define CYTHON_FRAME_MODIFIER
    #define CYTHON_FRAME_DEL(frame) Py_CLEAR(frame)
  #endif
  #define __Pyx_TraceDeclarations\
      static PyCodeObject *__pyx_frame_code = NULL;\
      CYTHON_FRAME_MODIFIER PyFrameObject *__pyx_frame = NULL;\
      int __Pyx_use_tracing = 0;
  #define __Pyx_TraceFrameInit(codeobj)\
      if (codeobj) __pyx_frame_code = (PyCodeObject*) codeobj;
#if PY_VERSION_HEX >= 0x030b00a2
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->cframe->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_EnterTracing(tstate) PyThreadState_EnterTracing(tstate)
  #define __Pyx_LeaveTracing(tstate) PyThreadState_LeaveTracing(tstate)
#elif PY_VERSION_HEX >= 0x030a00b1
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->cframe->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_EnterTracing(tstate)\
      do { tstate->tracing++; tstate->cframe->use_tracing = 0; } while (0)
  #define __Pyx_LeaveTracing(tstate)\
      do {\
          tstate->tracing--;\
          tstate->cframe->use_tracing = ((CYTHON_TRACE && tstate->c_tracefunc != NULL)\
                                 || tstate->c_profilefunc != NULL);\
      } while (0)
#else
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_EnterTracing(tstate)\
      do { tstate->tracing++; tstate->use_tracing = 0; } while (0)
  #define __Pyx_LeaveTracing(tstate)\
      do {\
          tstate->tracing--;\
          tstate->use_tracing = ((CYTHON_TRACE && tstate->c_tracefunc != NULL)\
                                         || tstate->c_profilefunc != NULL);\
      } while (0)
#endif
  #ifdef WITH_THREAD
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  if (nogil) {\
      if (CYTHON_TRACE_NOGIL) {\
          PyThreadState *tstate;\
          PyGILState_STATE state = PyGILState_Ensure();\
          tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 1, 1)) {\
              __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          }\
          PyGILState_Release(state);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  } else {\
      PyThreadState* tstate = PyThreadState_GET();\
      if (__Pyx_IsTracing(tstate, 1, 1)) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #else
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  {   PyThreadState* tstate = PyThreadState_GET();\
      if (__Pyx_IsTracing(tstate, 1, 1)) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #endif
  #define __Pyx_TraceException()\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 1)) {\
          __Pyx_EnterTracing(tstate);\
          PyObject *exc_info = __Pyx_GetExceptionTuple(tstate);\
          if (exc_info) {\
              if (CYTHON_TRACE && tstate->c_tracefunc)\
                  tstate->c_tracefunc(\
                      tstate->c_traceobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              tstate->c_profilefunc(\
                  tstate->c_profileobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              Py_DECREF(exc_info);\
          }\
          __Pyx_LeaveTracing(tstate);\
      }\
  }
  static void __Pyx_call_return_trace_func(PyThreadState *tstate, PyFrameObject *frame, PyObject *result) {
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      __Pyx_EnterTracing(tstate);
      if (CYTHON_TRACE && tstate->c_tracefunc)
          tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_RETURN, result);
      if (tstate->c_profilefunc)
          tstate->c_profilefunc(tstate->c_profileobj, frame, PyTrace_RETURN, result);
      CYTHON_FRAME_DEL(frame);
      __Pyx_LeaveTracing(tstate);
      __Pyx_ErrRestoreInState(tstate, type, value, traceback);
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0)) {\
                  __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0)) {\
              __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
          }\
      }\
  }
  #else
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 0)) {\
          __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
      }\
  }
  #endif
  static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno);
  static int __Pyx_TraceSetupAndCall(PyCodeObject** code, PyFrameObject** frame, PyThreadState* tstate, const char *funcname, const char *srcfile, int firstlineno);
#else
  #define __Pyx_TraceDeclarations
  #define __Pyx_TraceFrameInit(codeobj)
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)   if ((1)); else goto_error;
  #define __Pyx_TraceException()
  #define __Pyx_TraceReturn(result, nogil)
#endif
#if CYTHON_TRACE
  static int __Pyx_call_line_trace_func(PyThreadState *tstate, PyFrameObject *frame, int lineno) {
      int ret;
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      __Pyx_PyFrame_SetLineNumber(frame, lineno);
      __Pyx_EnterTracing(tstate);
      ret = tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_LINE, NULL);
      __Pyx_LeaveTracing(tstate);
      if (likely(!ret)) {
          __Pyx_ErrRestoreInState(tstate, type, value, traceback);
      } else {
          Py_XDECREF(type);
          Py_XDECREF(value);
          Py_XDECREF(traceback);
      }
      return ret;
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              int ret = 0;\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
                  ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              }\
              PyGILState_Release(state);\
              if (unlikely(ret)) goto_error;\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
              int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              if (unlikely(ret)) goto_error;\
          }\
      }\
  }
  #else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
          int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
          if (unlikely(ret)) goto_error;\
      }\
  }
  #endif
#else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)   if ((1)); else goto_error;
#endif

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *, int writable_flag);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD__call(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, CYTHON_UNUSED double __pyx_v_m_1, CYTHON_UNUSED double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_grad(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_hessian(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_xi(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, CYTHON_UNUSED double __pyx_v_m_1, CYTHON_UNUSED double __pyx_v_m_2, CYTHON_UNUSED double __pyx_v_chi_1, CYTHON_UNUSED double __pyx_v_chi_2, CYTHON_UNUSED double __pyx_v_chiL1, CYTHON_UNUSED double __pyx_v_chiL2); /* proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_dynamics(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_omega(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_auxderivs(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, CYTHON_UNUSED double __pyx_v_m_1, CYTHON_UNUSED double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'cython.view' */

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;

/* Module declarations from 'pyseobnr.eob.utils.containers' */
static PyTypeObject *__pyx_ptype_8pyseobnr_3eob_5utils_10containers_PhysicalParams = 0;
static PyTypeObject *__pyx_ptype_8pyseobnr_3eob_5utils_10containers_CalibCoeffs = 0;
static PyTypeObject *__pyx_ptype_8pyseobnr_3eob_5utils_10containers_Dynamics = 0;
static PyTypeObject *__pyx_ptype_8pyseobnr_3eob_5utils_10containers_FluxParams = 0;
static PyTypeObject *__pyx_ptype_8pyseobnr_3eob_5utils_10containers_EOBParams = 0;

/* Module declarations from 'pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C' */
static PyTypeObject *__pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C = 0;

/* Module declarations from 'libc.math' */

/* Module declarations from 'pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD' */
static PyTypeObject *__pyx_ptype_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD"
extern int __pyx_module_is_main_pyseobnr__eob__hamiltonian__Ham_AvgS2precess_simple_cython_PA_AD;
int __pyx_module_is_main_pyseobnr__eob__hamiltonian__Ham_AvgS2precess_simple_cython_PA_AD = 0;

/* Implementation of 'pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_IndexError;
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_q[] = "q";
static const char __pyx_k_a6[] = "a6";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_dSO[] = "dSO";
static const char __pyx_k_m_1[] = "m_1";
static const char __pyx_k_m_2[] = "m_2";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_call[] = "_call";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_grad[] = "grad";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_chiL1[] = "chiL1";
static const char __pyx_k_chiL2[] = "chiL2";
static const char __pyx_k_chi_1[] = "chi_1";
static const char __pyx_k_chi_2[] = "chi_2";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_omega[] = "omega";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_chi1_v[] = "chi1_v";
static const char __pyx_k_chi2_v[] = "chi2_v";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_hessian[] = "hessian";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_reshape[] = "reshape";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_dynamics[] = "dynamics";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_auxderivs[] = "auxderivs";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Ham_AvgS2precess_simple_cython_P[] = "Ham_AvgS2precess_simple_cython_PA_AD";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_n_s_Ham_AvgS2precess_simple_cython_P;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_n_u_a6;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_auxderivs;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_call;
static PyObject *__pyx_n_s_chi1_v;
static PyObject *__pyx_n_s_chi2_v;
static PyObject *__pyx_n_s_chiL1;
static PyObject *__pyx_n_s_chiL2;
static PyObject *__pyx_n_s_chi_1;
static PyObject *__pyx_n_s_chi_2;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_u_dSO;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_dynamics;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_grad;
static PyObject *__pyx_n_s_hessian;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_m_1;
static PyObject *__pyx_n_s_m_2;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_omega;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_params;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_q;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_reshape;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static int __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD___cinit__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_EOBParams *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_2__call__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_4_call(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_6grad(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_8hessian(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_10dynamics(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_12omega(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_14auxderivs(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs___get__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self); /* proto */
static int __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_2__set__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_4__del__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_112105877;
static PyObject *__pyx_int_136983863;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__20;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_codeobj__24;
/* Late includes */

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":19
 * cdef class Ham_AvgS2precess_simple_cython_PA_AD(Hamiltonian_v5PHM_C):
 *     cdef public CalibCoeffs calibration_coeffs
 *     def __cinit__(self,EOBParams params):             # <<<<<<<<<<<<<<
 *         self.EOBpars = params
 * 
 */

/* Python wrapper */
static int __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_EOBParams *__pyx_v_params = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_params,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 19, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_params = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_EOBParams *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 19, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_params), __pyx_ptype_8pyseobnr_3eob_5utils_10containers_EOBParams, 1, "params", 0))) __PYX_ERR(1, 19, __pyx_L1_error)
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD___cinit__(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), __pyx_v_params);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD___cinit__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_EOBParams *__pyx_v_params) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_TraceCall("__cinit__", __pyx_f[1], 19, 0, __PYX_ERR(1, 19, __pyx_L1_error));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":20
 *     cdef public CalibCoeffs calibration_coeffs
 *     def __cinit__(self,EOBParams params):
 *         self.EOBpars = params             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 */
  __Pyx_TraceLine(20,0,__PYX_ERR(1, 20, __pyx_L1_error))
  __Pyx_INCREF(((PyObject *)__pyx_v_params));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_params));
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.EOBpars);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.EOBpars));
  __pyx_v_self->__pyx_base.EOBpars = __pyx_v_params;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":19
 * cdef class Ham_AvgS2precess_simple_cython_PA_AD(Hamiltonian_v5PHM_C):
 *     cdef public CalibCoeffs calibration_coeffs
 *     def __cinit__(self,EOBParams params):             # <<<<<<<<<<<<<<
 *         self.EOBpars = params
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":22
 *         self.EOBpars = params
 * 
 *     def __call__(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 *         return self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_3__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_3__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_q = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi1_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi2_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_m_1;
  double __pyx_v_m_2;
  double __pyx_v_chi_1;
  double __pyx_v_chi_2;
  double __pyx_v_chiL1;
  double __pyx_v_chiL2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_q,&__pyx_n_s_p,&__pyx_n_s_chi1_v,&__pyx_n_s_chi2_v,&__pyx_n_s_m_1,&__pyx_n_s_m_2,&__pyx_n_s_chi_1,&__pyx_n_s_chi_2,&__pyx_n_s_chiL1,&__pyx_n_s_chiL2,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_q)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 1); __PYX_ERR(1, 22, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi1_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 2); __PYX_ERR(1, 22, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi2_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 3); __PYX_ERR(1, 22, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 4); __PYX_ERR(1, 22, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 5); __PYX_ERR(1, 22, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 6); __PYX_ERR(1, 22, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 7); __PYX_ERR(1, 22, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 8); __PYX_ERR(1, 22, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, 9); __PYX_ERR(1, 22, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__call__") < 0)) __PYX_ERR(1, 22, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_q = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_q.memview)) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_p.memview)) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_chi1_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi1_v.memview)) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_chi2_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi2_v.memview)) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_m_1 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_m_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_m_2 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_m_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_chi_1 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_chi_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_chi_2 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_chi_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_chiL1 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_chiL1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 22, __pyx_L3_error)
    __pyx_v_chiL2 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_chiL2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 22, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 22, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_2__call__(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_2__call__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);
  __Pyx_TraceCall("__call__", __pyx_f[1], 22, 0, __PYX_ERR(1, 22, __pyx_L1_error));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":23
 * 
 *     def __call__(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 *         return self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _call(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 */
  __Pyx_TraceLine(23,0,__PYX_ERR(1, 23, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._call(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":22
 *         self.EOBpars = params
 * 
 *     def __call__(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 *         return self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_q, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi1_v, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi2_v, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":25
 *         return self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)
 * 
 *     cpdef _call(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 *         """
 *         Evaluate the Hamiltonian as well as several potentials.
 */

static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_5_call(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD__call(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, CYTHON_UNUSED double __pyx_v_m_1, CYTHON_UNUSED double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch) {
  double __pyx_v_r;
  CYTHON_UNUSED double __pyx_v_phi;
  double __pyx_v_prst;
  double __pyx_v_L;
  double __pyx_v_chix1;
  double __pyx_v_chiy1;
  double __pyx_v_chiz1;
  double __pyx_v_chix2;
  double __pyx_v_chiy2;
  double __pyx_v_chiz2;
  CYTHON_UNUSED double __pyx_v_pphi;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *__pyx_v_c_coeffs = 0;
  double __pyx_v_a6;
  double __pyx_v_dSO;
  double __pyx_v_M;
  double __pyx_v_nu;
  double __pyx_v_X_1;
  double __pyx_v_X_2;
  double __pyx_v_delta;
  double __pyx_v_Dbpm;
  double __pyx_v_Apm;
  double __pyx_v_t2;
  double __pyx_v_t1;
  double __pyx_v_ap2;
  double __pyx_v_xi;
  double __pyx_v_apam;
  double __pyx_v_am2;
  double __pyx_v_QSalign2;
  double __pyx_v_Qpm;
  double __pyx_v_Qq;
  double __pyx_v_Bnpa;
  double __pyx_v_BnpSalign2;
  double __pyx_v_Bnp;
  double __pyx_v_amz;
  double __pyx_v_apz;
  double __pyx_v_napnam;
  double __pyx_v_amz2;
  double __pyx_v_nam2;
  double __pyx_v_apz2;
  double __pyx_v_nap2;
  double __pyx_v_BpSprec2;
  double __pyx_v_Bp;
  double __pyx_v_ASprec2;
  double __pyx_v_ASalign2;
  double __pyx_v_A;
  double __pyx_v_ap;
  double __pyx_v_lap;
  double __pyx_v_Heven;
  double __pyx_v_am;
  double __pyx_v_lam;
  double __pyx_v_Ga3;
  double __pyx_v_SOcalib;
  double __pyx_v_gam;
  double __pyx_v_gap;
  double __pyx_v_Hodd;
  CYTHON_UNUSED double __pyx_v_pr;
  double __pyx_v_Heff;
  double __pyx_v_H;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  double __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call", 0);
  __Pyx_TraceCall("_call", __pyx_f[1], 25, 0, __PYX_ERR(1, 25, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 25, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_5_call)) {
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 25, __pyx_L1_error) }
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_q, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 25, __pyx_L1_error) }
        __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_p, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 25, __pyx_L1_error) }
        __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_chi1_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 25, __pyx_L1_error) }
        __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_chi2_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_m_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyFloat_FromDouble(__pyx_v_m_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_chi_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyFloat_FromDouble(__pyx_v_chi_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_chiL1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = PyFloat_FromDouble(__pyx_v_chiL2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_13 = __pyx_t_1; __pyx_t_14 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(10+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_15, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_15, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_15, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_15, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_15, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_15, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_16, 8+__pyx_t_15, __pyx_t_11);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_16, 9+__pyx_t_15, __pyx_t_12);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_12 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":49
 *         # Coordinate definitions
 * 
 *         cdef double r = q[0]             # <<<<<<<<<<<<<<
 *         cdef double phi = q[1]
 * 
 */
  __Pyx_TraceLine(49,0,__PYX_ERR(1, 49, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_r = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":50
 * 
 *         cdef double r = q[0]
 *         cdef double phi = q[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double prst = p[0]
 */
  __Pyx_TraceLine(50,0,__PYX_ERR(1, 50, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_phi = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":52
 *         cdef double phi = q[1]
 * 
 *         cdef double prst = p[0]             # <<<<<<<<<<<<<<
 *         cdef double L = p[1]
 * 
 */
  __Pyx_TraceLine(52,0,__PYX_ERR(1, 52, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_prst = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":53
 * 
 *         cdef double prst = p[0]
 *         cdef double L = p[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix1 = chi1_v[0]
 */
  __Pyx_TraceLine(53,0,__PYX_ERR(1, 53, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_L = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":55
 *         cdef double L = p[1]
 * 
 *         cdef double chix1 = chi1_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]
 */
  __Pyx_TraceLine(55,0,__PYX_ERR(1, 55, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":56
 * 
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz1 = chi1_v[2]
 * 
 */
  __Pyx_TraceLine(56,0,__PYX_ERR(1, 56, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":57
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix2 = chi2_v[0]
 */
  __Pyx_TraceLine(57,0,__PYX_ERR(1, 57, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":59
 *         cdef double chiz1 = chi1_v[2]
 * 
 *         cdef double chix2 = chi2_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]
 */
  __Pyx_TraceLine(59,0,__PYX_ERR(1, 59, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":60
 * 
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz2 = chi2_v[2]
 * 
 */
  __Pyx_TraceLine(60,0,__PYX_ERR(1, 60, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":61
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double pphi = L
 */
  __Pyx_TraceLine(61,0,__PYX_ERR(1, 61, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":63
 *         cdef double chiz2 = chi2_v[2]
 * 
 *         cdef double pphi = L             # <<<<<<<<<<<<<<
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 */
  __Pyx_TraceLine(63,0,__PYX_ERR(1, 63, __pyx_L1_error))
  __pyx_v_pphi = __pyx_v_L;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":65
 *         cdef double pphi = L
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs             # <<<<<<<<<<<<<<
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']
 */
  __Pyx_TraceLine(65,0,__PYX_ERR(1, 65, __pyx_L1_error))
  __pyx_t_1 = ((PyObject *)__pyx_v_self->calibration_coeffs);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_c_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":66
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']             # <<<<<<<<<<<<<<
 *         cdef double dSO = c_coeffs['dSO']
 * 
 */
  __Pyx_TraceLine(66,0,__PYX_ERR(1, 66, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_a6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_a6 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":67
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(67,0,__PYX_ERR(1, 67, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_dSO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dSO = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":71
 * 
 *         # Extra quantities used in the Hamiltonian
 *         cdef double M = self.EOBpars.p_params.M             # <<<<<<<<<<<<<<
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 */
  __Pyx_TraceLine(71,0,__PYX_ERR(1, 71, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->M;
  __pyx_v_M = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":72
 *         # Extra quantities used in the Hamiltonian
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu             # <<<<<<<<<<<<<<
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 */
  __Pyx_TraceLine(72,0,__PYX_ERR(1, 72, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->nu;
  __pyx_v_nu = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":73
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1             # <<<<<<<<<<<<<<
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 */
  __Pyx_TraceLine(73,0,__PYX_ERR(1, 73, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_1;
  __pyx_v_X_1 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":74
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2             # <<<<<<<<<<<<<<
 *         cdef double delta = self.EOBpars.p_params.delta
 * 
 */
  __Pyx_TraceLine(74,0,__PYX_ERR(1, 74, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_2;
  __pyx_v_X_2 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":75
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(75,0,__PYX_ERR(1, 75, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->delta;
  __pyx_v_delta = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":79
 * 
 *         # Actual Hamiltonian expressions
 *         cdef double Dbpm = r*(6730497718123.02*nu**3 + 133772083200.0*nu**2*r**2 + 1822680546449.21*nu**2*r + 80059249540278.2*nu**2 + 2589101062873.81*nu*r**2 + 10611661054566.2*nu*r - 12049908701745.2*nu + 5107745331375.71*r**2 - 326837426.241486*r*(14700.0*nu + 42911.0) - 39476764256925.6*r - (-5041721180160.0*nu**2 - 25392914995744.3*nu - 879923036160.0*r**2 - 283115520.0*r*(14700.0*nu + 42911.0) + 104186110149937.0)*log(r) + 5787938193408.0*log(r)**2 + 275059053208689.0)/(55296.0*nu*(14515200.0*nu**3 - 42636451.6032331*nu**2 - 2510664218.28128*nu + 1002013764.01019) - 967680.0*r**3*(-138240.0*nu**2 - 2675575.66847905*nu - 5278341.3229329) - 9216.0*r**2*(-197773496.793534*nu**2 - 630116198.873299*nu + 5805304367.87913) + r*(5927865218923.02*nu**3 + 43133561885859.3*nu**2 + 43393301259014.8*nu + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0) + 5787938193408.0*r*log(r)**2 + (-1698693120.0*nu*(11592.0*nu + 69847.0) + 879923036160.0*r**3 + 283115520.0*r**2*(14700.0*nu + 42911.0) + 49152.0*r*(102574080.0*nu**2 + 409207698.136075*nu - 2119671837.36038))*log(r))             # <<<<<<<<<<<<<<
 * 
 *         cdef double Apm = 7680.0*r**4*(-5416406.59541186*nu**2 + 28.0*nu*(1920.0*a6 + 733955.307463037) + 2048.0*nu*(756.0*nu + 336.0*r + 407.0)*log(r) - 7.0*r*(-185763.092693281*nu**2 + 938918.400156317*nu - 245760.0) - 3440640.0)/(241555486248.807*nu**4 + 1120.0*nu**3*(-17833256.898555*r**2 - 163683964.822551*r - 1188987459.03162) + 7.0*nu**2*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 1426660551.8844*r**5 - 3089250703.76879*r**4 - 6178501407.53758*r**3 + 2064783811.32587*r**2 + 122635399361.987*r + 276057889687.011) + 67645734912.0*nu**2*log(r)**2 + 53760.0*nu*(7680.0*a6*(r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 128.0*r*(-6852.34813868015*r**4 + 4264.6962773603*r**3 + 8529.39255472061*r**2 + 13218.7851094412*r - 33722.4297811176) + 113485.217444961*r*(-r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 148.04406601634*r*(349.0*r**4 + 1926.0*r**3 + 3852.0*r**2 + 7704.0*r + 36400.0)) + 32768.0*nu*(-1882456.23663972*nu**2 - 38842241.4769507*nu + 161280.0*r**5 + 480.0*r**4*(756.0*nu + 1079.0) + 960.0*r**3*(756.0*nu + 1079.0) + 1920.0*r**2*(588.0*nu + 1079.0) + 240.0*r*(-3024.0*nu**2 - 7466.27061066206*nu + 17264.0) + 13447680.0)*log(r) + 13212057600.0*r**5)
 */
  __Pyx_TraceLine(79,0,__PYX_ERR(1, 79, __pyx_L1_error))
  __pyx_v_Dbpm = ((__pyx_v_r * (((((((((((((6730497718123.02 * pow(__pyx_v_nu, 3.0)) + ((133772083200.0 * pow(__pyx_v_nu, 2.0)) * pow(__pyx_v_r, 2.0))) + ((1822680546449.21 * pow(__pyx_v_nu, 2.0)) * __pyx_v_r)) + (80059249540278.2 * pow(__pyx_v_nu, 2.0))) + ((2589101062873.81 * __pyx_v_nu) * pow(__pyx_v_r, 2.0))) + ((10611661054566.2 * __pyx_v_nu) * __pyx_v_r)) - (12049908701745.2 * __pyx_v_nu)) + (5107745331375.71 * pow(__pyx_v_r, 2.0))) - ((326837426.241486 * __pyx_v_r) * ((14700.0 * __pyx_v_nu) + 42911.0))) - (39476764256925.6 * __pyx_v_r)) - ((((((-5041721180160.0 * pow(__pyx_v_nu, 2.0)) - (25392914995744.3 * __pyx_v_nu)) - (879923036160.0 * pow(__pyx_v_r, 2.0))) - ((283115520.0 * __pyx_v_r) * ((14700.0 * __pyx_v_nu) + 42911.0))) + 104186110149937.0) * log(__pyx_v_r))) + (5787938193408.0 * pow(log(__pyx_v_r), 2.0))) + 275059053208689.0)) / (((((((55296.0 * __pyx_v_nu) * ((((14515200.0 * pow(__pyx_v_nu, 3.0)) - (42636451.6032331 * pow(__pyx_v_nu, 2.0))) - (2510664218.28128 * __pyx_v_nu)) + 1002013764.01019)) - ((967680.0 * pow(__pyx_v_r, 3.0)) * (((-138240.0 * pow(__pyx_v_nu, 2.0)) - (2675575.66847905 * __pyx_v_nu)) - 5278341.3229329))) - ((9216.0 * pow(__pyx_v_r, 2.0)) * (((-197773496.793534 * pow(__pyx_v_nu, 2.0)) - (630116198.873299 * __pyx_v_nu)) + 5805304367.87913))) + (__pyx_v_r * (((((5927865218923.02 * pow(__pyx_v_nu, 3.0)) + (43133561885859.3 * pow(__pyx_v_nu, 2.0))) + (43393301259014.8 * __pyx_v_nu)) + (86618264430493.3 * pow((1.0 - (0.496948781616935 * __pyx_v_nu)), 2.0))) + 188440788778196.0))) + ((5787938193408.0 * __pyx_v_r) * pow(log(__pyx_v_r), 2.0))) + ((((((-1698693120.0 * __pyx_v_nu) * ((11592.0 * __pyx_v_nu) + 69847.0)) + (879923036160.0 * pow(__pyx_v_r, 3.0))) + ((283115520.0 * pow(__pyx_v_r, 2.0)) * ((14700.0 * __pyx_v_nu) + 42911.0))) + ((49152.0 * __pyx_v_r) * (((102574080.0 * pow(__pyx_v_nu, 2.0)) + (409207698.136075 * __pyx_v_nu)) - 2119671837.36038))) * log(__pyx_v_r))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":81
 *         cdef double Dbpm = r*(6730497718123.02*nu**3 + 133772083200.0*nu**2*r**2 + 1822680546449.21*nu**2*r + 80059249540278.2*nu**2 + 2589101062873.81*nu*r**2 + 10611661054566.2*nu*r - 12049908701745.2*nu + 5107745331375.71*r**2 - 326837426.241486*r*(14700.0*nu + 42911.0) - 39476764256925.6*r - (-5041721180160.0*nu**2 - 25392914995744.3*nu - 879923036160.0*r**2 - 283115520.0*r*(14700.0*nu + 42911.0) + 104186110149937.0)*log(r) + 5787938193408.0*log(r)**2 + 275059053208689.0)/(55296.0*nu*(14515200.0*nu**3 - 42636451.6032331*nu**2 - 2510664218.28128*nu + 1002013764.01019) - 967680.0*r**3*(-138240.0*nu**2 - 2675575.66847905*nu - 5278341.3229329) - 9216.0*r**2*(-197773496.793534*nu**2 - 630116198.873299*nu + 5805304367.87913) + r*(5927865218923.02*nu**3 + 43133561885859.3*nu**2 + 43393301259014.8*nu + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0) + 5787938193408.0*r*log(r)**2 + (-1698693120.0*nu*(11592.0*nu + 69847.0) + 879923036160.0*r**3 + 283115520.0*r**2*(14700.0*nu + 42911.0) + 49152.0*r*(102574080.0*nu**2 + 409207698.136075*nu - 2119671837.36038))*log(r))
 * 
 *         cdef double Apm = 7680.0*r**4*(-5416406.59541186*nu**2 + 28.0*nu*(1920.0*a6 + 733955.307463037) + 2048.0*nu*(756.0*nu + 336.0*r + 407.0)*log(r) - 7.0*r*(-185763.092693281*nu**2 + 938918.400156317*nu - 245760.0) - 3440640.0)/(241555486248.807*nu**4 + 1120.0*nu**3*(-17833256.898555*r**2 - 163683964.822551*r - 1188987459.03162) + 7.0*nu**2*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 1426660551.8844*r**5 - 3089250703.76879*r**4 - 6178501407.53758*r**3 + 2064783811.32587*r**2 + 122635399361.987*r + 276057889687.011) + 67645734912.0*nu**2*log(r)**2 + 53760.0*nu*(7680.0*a6*(r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 128.0*r*(-6852.34813868015*r**4 + 4264.6962773603*r**3 + 8529.39255472061*r**2 + 13218.7851094412*r - 33722.4297811176) + 113485.217444961*r*(-r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 148.04406601634*r*(349.0*r**4 + 1926.0*r**3 + 3852.0*r**2 + 7704.0*r + 36400.0)) + 32768.0*nu*(-1882456.23663972*nu**2 - 38842241.4769507*nu + 161280.0*r**5 + 480.0*r**4*(756.0*nu + 1079.0) + 960.0*r**3*(756.0*nu + 1079.0) + 1920.0*r**2*(588.0*nu + 1079.0) + 240.0*r*(-3024.0*nu**2 - 7466.27061066206*nu + 17264.0) + 13447680.0)*log(r) + 13212057600.0*r**5)             # <<<<<<<<<<<<<<
 * 
 *         cdef double t2 = chix2**2 + chiy2**2 + chiz2**2
 */
  __Pyx_TraceLine(81,0,__PYX_ERR(1, 81, __pyx_L1_error))
  __pyx_v_Apm = (((7680.0 * pow(__pyx_v_r, 4.0)) * (((((-5416406.59541186 * pow(__pyx_v_nu, 2.0)) + ((28.0 * __pyx_v_nu) * ((1920.0 * __pyx_v_a6) + 733955.307463037))) + (((2048.0 * __pyx_v_nu) * (((756.0 * __pyx_v_nu) + (336.0 * __pyx_v_r)) + 407.0)) * log(__pyx_v_r))) - ((7.0 * __pyx_v_r) * (((-185763.092693281 * pow(__pyx_v_nu, 2.0)) + (938918.400156317 * __pyx_v_nu)) - 245760.0))) - 3440640.0)) / (((((((241555486248.807 * pow(__pyx_v_nu, 4.0)) + ((1120.0 * pow(__pyx_v_nu, 3.0)) * (((-17833256.898555 * pow(__pyx_v_r, 2.0)) - (163683964.822551 * __pyx_v_r)) - 1188987459.03162))) + ((7.0 * pow(__pyx_v_nu, 2.0)) * (((((((((-39321600.0 * __pyx_v_a6) * ((3.0 * __pyx_v_r) + 59.0)) + (745857848.115604 * __pyx_v_a6)) + (1426660551.8844 * pow(__pyx_v_r, 5.0))) - (3089250703.76879 * pow(__pyx_v_r, 4.0))) - (6178501407.53758 * pow(__pyx_v_r, 3.0))) + (2064783811.32587 * pow(__pyx_v_r, 2.0))) + (122635399361.987 * __pyx_v_r)) + 276057889687.011))) + ((67645734912.0 * pow(__pyx_v_nu, 2.0)) * pow(log(__pyx_v_r), 2.0))) + ((53760.0 * __pyx_v_nu) * (((((7680.0 * __pyx_v_a6) * ((((pow(__pyx_v_r, 4.0) + (2.0 * pow(__pyx_v_r, 3.0))) + (4.0 * pow(__pyx_v_r, 2.0))) + (8.0 * __pyx_v_r)) + 16.0)) + ((128.0 * __pyx_v_r) * (((((-6852.34813868015 * pow(__pyx_v_r, 4.0)) + (4264.6962773603 * pow(__pyx_v_r, 3.0))) + (8529.39255472061 * pow(__pyx_v_r, 2.0))) + (13218.7851094412 * __pyx_v_r)) - 33722.4297811176))) + ((113485.217444961 * __pyx_v_r) * (((((-pow(__pyx_v_r, 4.0)) + (2.0 * pow(__pyx_v_r, 3.0))) + (4.0 * pow(__pyx_v_r, 2.0))) + (8.0 * __pyx_v_r)) + 16.0))) + ((148.04406601634 * __pyx_v_r) * (((((349.0 * pow(__pyx_v_r, 4.0)) + (1926.0 * pow(__pyx_v_r, 3.0))) + (3852.0 * pow(__pyx_v_r, 2.0))) + (7704.0 * __pyx_v_r)) + 36400.0))))) + (((32768.0 * __pyx_v_nu) * ((((((((-1882456.23663972 * pow(__pyx_v_nu, 2.0)) - (38842241.4769507 * __pyx_v_nu)) + (161280.0 * pow(__pyx_v_r, 5.0))) + ((480.0 * pow(__pyx_v_r, 4.0)) * ((756.0 * __pyx_v_nu) + 1079.0))) + ((960.0 * pow(__pyx_v_r, 3.0)) * ((756.0 * __pyx_v_nu) + 1079.0))) + ((1920.0 * pow(__pyx_v_r, 2.0)) * ((588.0 * __pyx_v_nu) + 1079.0))) + ((240.0 * __pyx_v_r) * (((-3024.0 * pow(__pyx_v_nu, 2.0)) - (7466.27061066206 * __pyx_v_nu)) + 17264.0))) + 13447680.0)) * log(__pyx_v_r))) + (13212057600.0 * pow(__pyx_v_r, 5.0))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":83
 *         cdef double Apm = 7680.0*r**4*(-5416406.59541186*nu**2 + 28.0*nu*(1920.0*a6 + 733955.307463037) + 2048.0*nu*(756.0*nu + 336.0*r + 407.0)*log(r) - 7.0*r*(-185763.092693281*nu**2 + 938918.400156317*nu - 245760.0) - 3440640.0)/(241555486248.807*nu**4 + 1120.0*nu**3*(-17833256.898555*r**2 - 163683964.822551*r - 1188987459.03162) + 7.0*nu**2*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 1426660551.8844*r**5 - 3089250703.76879*r**4 - 6178501407.53758*r**3 + 2064783811.32587*r**2 + 122635399361.987*r + 276057889687.011) + 67645734912.0*nu**2*log(r)**2 + 53760.0*nu*(7680.0*a6*(r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 128.0*r*(-6852.34813868015*r**4 + 4264.6962773603*r**3 + 8529.39255472061*r**2 + 13218.7851094412*r - 33722.4297811176) + 113485.217444961*r*(-r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 148.04406601634*r*(349.0*r**4 + 1926.0*r**3 + 3852.0*r**2 + 7704.0*r + 36400.0)) + 32768.0*nu*(-1882456.23663972*nu**2 - 38842241.4769507*nu + 161280.0*r**5 + 480.0*r**4*(756.0*nu + 1079.0) + 960.0*r**3*(756.0*nu + 1079.0) + 1920.0*r**2*(588.0*nu + 1079.0) + 240.0*r*(-3024.0*nu**2 - 7466.27061066206*nu + 17264.0) + 13447680.0)*log(r) + 13212057600.0*r**5)
 * 
 *         cdef double t2 = chix2**2 + chiy2**2 + chiz2**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double t1 = chix1**2 + chiy1**2 + chiz1**2
 */
  __Pyx_TraceLine(83,0,__PYX_ERR(1, 83, __pyx_L1_error))
  __pyx_v_t2 = ((pow(__pyx_v_chix2, 2.0) + pow(__pyx_v_chiy2, 2.0)) + pow(__pyx_v_chiz2, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":85
 *         cdef double t2 = chix2**2 + chiy2**2 + chiz2**2
 * 
 *         cdef double t1 = chix1**2 + chiy1**2 + chiz1**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double ap2 = X_1**2*t1 + X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2
 */
  __Pyx_TraceLine(85,0,__PYX_ERR(1, 85, __pyx_L1_error))
  __pyx_v_t1 = ((pow(__pyx_v_chix1, 2.0) + pow(__pyx_v_chiy1, 2.0)) + pow(__pyx_v_chiz1, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":87
 *         cdef double t1 = chix1**2 + chiy1**2 + chiz1**2
 * 
 *         cdef double ap2 = X_1**2*t1 + X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2             # <<<<<<<<<<<<<<
 * 
 *         cdef double xi = Dbpm**0.5*r**2*(Apm + ap2/r**2)/(ap2 + r**2)
 */
  __Pyx_TraceLine(87,0,__PYX_ERR(1, 87, __pyx_L1_error))
  __pyx_v_ap2 = (((pow(__pyx_v_X_1, 2.0) * __pyx_v_t1) + ((__pyx_v_X_1 * __pyx_v_X_2) * ((((2.0 * __pyx_v_chix1) * __pyx_v_chix2) + ((2.0 * __pyx_v_chiy1) * __pyx_v_chiy2)) + ((2.0 * __pyx_v_chiz1) * __pyx_v_chiz2)))) + (pow(__pyx_v_X_2, 2.0) * __pyx_v_t2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":89
 *         cdef double ap2 = X_1**2*t1 + X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2
 * 
 *         cdef double xi = Dbpm**0.5*r**2*(Apm + ap2/r**2)/(ap2 + r**2)             # <<<<<<<<<<<<<<
 * 
 *         cdef double apam = X_1**2*t1 - X_2**2*t2
 */
  __Pyx_TraceLine(89,0,__PYX_ERR(1, 89, __pyx_L1_error))
  __pyx_v_xi = (((pow(__pyx_v_Dbpm, 0.5) * pow(__pyx_v_r, 2.0)) * (__pyx_v_Apm + (__pyx_v_ap2 / pow(__pyx_v_r, 2.0)))) / (__pyx_v_ap2 + pow(__pyx_v_r, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":91
 *         cdef double xi = Dbpm**0.5*r**2*(Apm + ap2/r**2)/(ap2 + r**2)
 * 
 *         cdef double apam = X_1**2*t1 - X_2**2*t2             # <<<<<<<<<<<<<<
 * 
 *         cdef double am2 = X_1**2*t1 - X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2
 */
  __Pyx_TraceLine(91,0,__PYX_ERR(1, 91, __pyx_L1_error))
  __pyx_v_apam = ((pow(__pyx_v_X_1, 2.0) * __pyx_v_t1) - (pow(__pyx_v_X_2, 2.0) * __pyx_v_t2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":93
 *         cdef double apam = X_1**2*t1 - X_2**2*t2
 * 
 *         cdef double am2 = X_1**2*t1 - X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2             # <<<<<<<<<<<<<<
 * 
 *         cdef double QSalign2 = prst**4*(-0.46875*am2*(4.0*nu**2 - 5.0*nu + 1.0) - 0.15625*ap2*(32.0*nu**2 - 33.0*nu - 5.0) + 0.3125*apam*delta*(18.0*nu - 1.0))/(r**3*xi**4)
 */
  __Pyx_TraceLine(93,0,__PYX_ERR(1, 93, __pyx_L1_error))
  __pyx_v_am2 = (((pow(__pyx_v_X_1, 2.0) * __pyx_v_t1) - ((__pyx_v_X_1 * __pyx_v_X_2) * ((((2.0 * __pyx_v_chix1) * __pyx_v_chix2) + ((2.0 * __pyx_v_chiy1) * __pyx_v_chiy2)) + ((2.0 * __pyx_v_chiz1) * __pyx_v_chiz2)))) + (pow(__pyx_v_X_2, 2.0) * __pyx_v_t2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":95
 *         cdef double am2 = X_1**2*t1 - X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2
 * 
 *         cdef double QSalign2 = prst**4*(-0.46875*am2*(4.0*nu**2 - 5.0*nu + 1.0) - 0.15625*ap2*(32.0*nu**2 - 33.0*nu - 5.0) + 0.3125*apam*delta*(18.0*nu - 1.0))/(r**3*xi**4)             # <<<<<<<<<<<<<<
 * 
 *         cdef double Qpm = 1.48275342024365*nu*prst**8/r**2.5 + 0.121954868780449*nu*prst**8/r - 11.3175085791863*nu*prst**6/r**3.5 + 147.443752990146*nu*prst**4/r**4.5 + prst**8*(-6.0*nu**4 + 3.42857142857143*nu**3 + 3.33842023648322*nu**2 + 1.38977750996128*nu)/r**2 + prst**6*(6.0*nu**3 - 5.4*nu**2 - 2.78300763695006*nu)/r**2 + prst**6*(-14.0*nu**4 + 188.0*nu**3 - 89.5298327361234*nu**2 - 33.9782122170436*nu)/r**3 + prst**4*(-6.0*nu**2 + 8.0*nu)/r**2 + prst**4*(10.0*nu**3 - 131.0*nu**2 + 92.7110442849544*nu)/r**3 + prst**4*(602.318540416564*nu**3 + nu**2*(118.4*log(r) - 1796.13660498019) + nu*(452.542166996693 - 51.6952380952381*log(r)))/r**4
 */
  __Pyx_TraceLine(95,0,__PYX_ERR(1, 95, __pyx_L1_error))
  __pyx_v_QSalign2 = ((pow(__pyx_v_prst, 4.0) * ((((-0.46875 * __pyx_v_am2) * (((4.0 * pow(__pyx_v_nu, 2.0)) - (5.0 * __pyx_v_nu)) + 1.0)) - ((0.15625 * __pyx_v_ap2) * (((32.0 * pow(__pyx_v_nu, 2.0)) - (33.0 * __pyx_v_nu)) - 5.0))) + (((0.3125 * __pyx_v_apam) * __pyx_v_delta) * ((18.0 * __pyx_v_nu) - 1.0)))) / (pow(__pyx_v_r, 3.0) * pow(__pyx_v_xi, 4.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":97
 *         cdef double QSalign2 = prst**4*(-0.46875*am2*(4.0*nu**2 - 5.0*nu + 1.0) - 0.15625*ap2*(32.0*nu**2 - 33.0*nu - 5.0) + 0.3125*apam*delta*(18.0*nu - 1.0))/(r**3*xi**4)
 * 
 *         cdef double Qpm = 1.48275342024365*nu*prst**8/r**2.5 + 0.121954868780449*nu*prst**8/r - 11.3175085791863*nu*prst**6/r**3.5 + 147.443752990146*nu*prst**4/r**4.5 + prst**8*(-6.0*nu**4 + 3.42857142857143*nu**3 + 3.33842023648322*nu**2 + 1.38977750996128*nu)/r**2 + prst**6*(6.0*nu**3 - 5.4*nu**2 - 2.78300763695006*nu)/r**2 + prst**6*(-14.0*nu**4 + 188.0*nu**3 - 89.5298327361234*nu**2 - 33.9782122170436*nu)/r**3 + prst**4*(-6.0*nu**2 + 8.0*nu)/r**2 + prst**4*(10.0*nu**3 - 131.0*nu**2 + 92.7110442849544*nu)/r**3 + prst**4*(602.318540416564*nu**3 + nu**2*(118.4*log(r) - 1796.13660498019) + nu*(452.542166996693 - 51.6952380952381*log(r)))/r**4             # <<<<<<<<<<<<<<
 * 
 *         cdef double Qq = QSalign2 + Qpm
 */
  __Pyx_TraceLine(97,0,__PYX_ERR(1, 97, __pyx_L1_error))
  __pyx_v_Qpm = ((((((((((((1.48275342024365 * __pyx_v_nu) * pow(__pyx_v_prst, 8.0)) / pow(__pyx_v_r, 2.5)) + (((0.121954868780449 * __pyx_v_nu) * pow(__pyx_v_prst, 8.0)) / __pyx_v_r)) - (((11.3175085791863 * __pyx_v_nu) * pow(__pyx_v_prst, 6.0)) / pow(__pyx_v_r, 3.5))) + (((147.443752990146 * __pyx_v_nu) * pow(__pyx_v_prst, 4.0)) / pow(__pyx_v_r, 4.5))) + ((pow(__pyx_v_prst, 8.0) * ((((-6.0 * pow(__pyx_v_nu, 4.0)) + (3.42857142857143 * pow(__pyx_v_nu, 3.0))) + (3.33842023648322 * pow(__pyx_v_nu, 2.0))) + (1.38977750996128 * __pyx_v_nu))) / pow(__pyx_v_r, 2.0))) + ((pow(__pyx_v_prst, 6.0) * (((6.0 * pow(__pyx_v_nu, 3.0)) - (5.4 * pow(__pyx_v_nu, 2.0))) - (2.78300763695006 * __pyx_v_nu))) / pow(__pyx_v_r, 2.0))) + ((pow(__pyx_v_prst, 6.0) * ((((-14.0 * pow(__pyx_v_nu, 4.0)) + (188.0 * pow(__pyx_v_nu, 3.0))) - (89.5298327361234 * pow(__pyx_v_nu, 2.0))) - (33.9782122170436 * __pyx_v_nu))) / pow(__pyx_v_r, 3.0))) + ((pow(__pyx_v_prst, 4.0) * ((-6.0 * pow(__pyx_v_nu, 2.0)) + (8.0 * __pyx_v_nu))) / pow(__pyx_v_r, 2.0))) + ((pow(__pyx_v_prst, 4.0) * (((10.0 * pow(__pyx_v_nu, 3.0)) - (131.0 * pow(__pyx_v_nu, 2.0))) + (92.7110442849544 * __pyx_v_nu))) / pow(__pyx_v_r, 3.0))) + ((pow(__pyx_v_prst, 4.0) * (((602.318540416564 * pow(__pyx_v_nu, 3.0)) + (pow(__pyx_v_nu, 2.0) * ((118.4 * log(__pyx_v_r)) - 1796.13660498019))) + (__pyx_v_nu * (452.542166996693 - (51.6952380952381 * log(__pyx_v_r)))))) / pow(__pyx_v_r, 4.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":99
 *         cdef double Qpm = 1.48275342024365*nu*prst**8/r**2.5 + 0.121954868780449*nu*prst**8/r - 11.3175085791863*nu*prst**6/r**3.5 + 147.443752990146*nu*prst**4/r**4.5 + prst**8*(-6.0*nu**4 + 3.42857142857143*nu**3 + 3.33842023648322*nu**2 + 1.38977750996128*nu)/r**2 + prst**6*(6.0*nu**3 - 5.4*nu**2 - 2.78300763695006*nu)/r**2 + prst**6*(-14.0*nu**4 + 188.0*nu**3 - 89.5298327361234*nu**2 - 33.9782122170436*nu)/r**3 + prst**4*(-6.0*nu**2 + 8.0*nu)/r**2 + prst**4*(10.0*nu**3 - 131.0*nu**2 + 92.7110442849544*nu)/r**3 + prst**4*(602.318540416564*nu**3 + nu**2*(118.4*log(r) - 1796.13660498019) + nu*(452.542166996693 - 51.6952380952381*log(r)))/r**4
 * 
 *         cdef double Qq = QSalign2 + Qpm             # <<<<<<<<<<<<<<
 * 
 *         cdef double Bnpa = -r*(r + 2.0)/(ap2*r*(r + 2.0) + r**4)
 */
  __Pyx_TraceLine(99,0,__PYX_ERR(1, 99, __pyx_L1_error))
  __pyx_v_Qq = (__pyx_v_QSalign2 + __pyx_v_Qpm);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":101
 *         cdef double Qq = QSalign2 + Qpm
 * 
 *         cdef double Bnpa = -r*(r + 2.0)/(ap2*r*(r + 2.0) + r**4)             # <<<<<<<<<<<<<<
 * 
 *         cdef double BnpSalign2 = (0.1875*am2*(4.0*nu - 1.0) + ap2*(3.0*nu + 2.8125) - 2.625*apam*delta)/r**3 + (0.015625*am2*(4.0*nu**2 + 115.0*nu - 37.0) + 0.015625*ap2*(-1171.0*nu - 861.0) + 0.03125*apam*delta*(26.0*nu + 449.0))/r**4
 */
  __Pyx_TraceLine(101,0,__PYX_ERR(1, 101, __pyx_L1_error))
  __pyx_v_Bnpa = (((-__pyx_v_r) * (__pyx_v_r + 2.0)) / (((__pyx_v_ap2 * __pyx_v_r) * (__pyx_v_r + 2.0)) + pow(__pyx_v_r, 4.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":103
 *         cdef double Bnpa = -r*(r + 2.0)/(ap2*r*(r + 2.0) + r**4)
 * 
 *         cdef double BnpSalign2 = (0.1875*am2*(4.0*nu - 1.0) + ap2*(3.0*nu + 2.8125) - 2.625*apam*delta)/r**3 + (0.015625*am2*(4.0*nu**2 + 115.0*nu - 37.0) + 0.015625*ap2*(-1171.0*nu - 861.0) + 0.03125*apam*delta*(26.0*nu + 449.0))/r**4             # <<<<<<<<<<<<<<
 * 
 *         cdef double Bnp = Apm*Dbpm + BnpSalign2 + ap2/r**2 - 1.0
 */
  __Pyx_TraceLine(103,0,__PYX_ERR(1, 103, __pyx_L1_error))
  __pyx_v_BnpSalign2 = ((((((0.1875 * __pyx_v_am2) * ((4.0 * __pyx_v_nu) - 1.0)) + (__pyx_v_ap2 * ((3.0 * __pyx_v_nu) + 2.8125))) - ((2.625 * __pyx_v_apam) * __pyx_v_delta)) / pow(__pyx_v_r, 3.0)) + (((((0.015625 * __pyx_v_am2) * (((4.0 * pow(__pyx_v_nu, 2.0)) + (115.0 * __pyx_v_nu)) - 37.0)) + ((0.015625 * __pyx_v_ap2) * ((-1171.0 * __pyx_v_nu) - 861.0))) + (((0.03125 * __pyx_v_apam) * __pyx_v_delta) * ((26.0 * __pyx_v_nu) + 449.0))) / pow(__pyx_v_r, 4.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":105
 *         cdef double BnpSalign2 = (0.1875*am2*(4.0*nu - 1.0) + ap2*(3.0*nu + 2.8125) - 2.625*apam*delta)/r**3 + (0.015625*am2*(4.0*nu**2 + 115.0*nu - 37.0) + 0.015625*ap2*(-1171.0*nu - 861.0) + 0.03125*apam*delta*(26.0*nu + 449.0))/r**4
 * 
 *         cdef double Bnp = Apm*Dbpm + BnpSalign2 + ap2/r**2 - 1.0             # <<<<<<<<<<<<<<
 * 
 *         cdef double amz = chi_1*X_1 - chi_2*X_2
 */
  __Pyx_TraceLine(105,0,__PYX_ERR(1, 105, __pyx_L1_error))
  __pyx_v_Bnp = ((((__pyx_v_Apm * __pyx_v_Dbpm) + __pyx_v_BnpSalign2) + (__pyx_v_ap2 / pow(__pyx_v_r, 2.0))) - 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":107
 *         cdef double Bnp = Apm*Dbpm + BnpSalign2 + ap2/r**2 - 1.0
 * 
 *         cdef double amz = chi_1*X_1 - chi_2*X_2             # <<<<<<<<<<<<<<
 * 
 *         cdef double apz = chi_1*X_1 + chi_2*X_2
 */
  __Pyx_TraceLine(107,0,__PYX_ERR(1, 107, __pyx_L1_error))
  __pyx_v_amz = ((__pyx_v_chi_1 * __pyx_v_X_1) - (__pyx_v_chi_2 * __pyx_v_X_2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":109
 *         cdef double amz = chi_1*X_1 - chi_2*X_2
 * 
 *         cdef double apz = chi_1*X_1 + chi_2*X_2             # <<<<<<<<<<<<<<
 * 
 *         cdef double napnam = -0.5*amz*apz + 0.5*apam
 */
  __Pyx_TraceLine(109,0,__PYX_ERR(1, 109, __pyx_L1_error))
  __pyx_v_apz = ((__pyx_v_chi_1 * __pyx_v_X_1) + (__pyx_v_chi_2 * __pyx_v_X_2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":111
 *         cdef double apz = chi_1*X_1 + chi_2*X_2
 * 
 *         cdef double napnam = -0.5*amz*apz + 0.5*apam             # <<<<<<<<<<<<<<
 * 
 *         cdef double amz2 = amz**2
 */
  __Pyx_TraceLine(111,0,__PYX_ERR(1, 111, __pyx_L1_error))
  __pyx_v_napnam = (((-0.5 * __pyx_v_amz) * __pyx_v_apz) + (0.5 * __pyx_v_apam));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":113
 *         cdef double napnam = -0.5*amz*apz + 0.5*apam
 * 
 *         cdef double amz2 = amz**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double nam2 = 0.5*am2 - 0.5*amz2
 */
  __Pyx_TraceLine(113,0,__PYX_ERR(1, 113, __pyx_L1_error))
  __pyx_v_amz2 = pow(__pyx_v_amz, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":115
 *         cdef double amz2 = amz**2
 * 
 *         cdef double nam2 = 0.5*am2 - 0.5*amz2             # <<<<<<<<<<<<<<
 * 
 *         cdef double apz2 = apz**2
 */
  __Pyx_TraceLine(115,0,__PYX_ERR(1, 115, __pyx_L1_error))
  __pyx_v_nam2 = ((0.5 * __pyx_v_am2) - (0.5 * __pyx_v_amz2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":117
 *         cdef double nam2 = 0.5*am2 - 0.5*amz2
 * 
 *         cdef double apz2 = apz**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double nap2 = 0.5*ap2 - 0.5*apz2
 */
  __Pyx_TraceLine(117,0,__PYX_ERR(1, 117, __pyx_L1_error))
  __pyx_v_apz2 = pow(__pyx_v_apz, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":119
 *         cdef double apz2 = apz**2
 * 
 *         cdef double nap2 = 0.5*ap2 - 0.5*apz2             # <<<<<<<<<<<<<<
 * 
 *         cdef double BpSprec2 = -nap2/r**2 + (nam2*(0.1875 - 0.75*nu) + nap2*(-1.75*nu - 0.9375) + napnam*(0.75 - 1.5*X_2))/r**3 + (-0.125*delta*napnam*(98.0*nu + 43.0) + 0.015625*nam2*(152.0*nu**2 - 1090.0*nu + 219.0) + 0.00520833333333333*nap2*(264.0*nu**2 - 1610.0*nu + 375.0))/r**4
 */
  __Pyx_TraceLine(119,0,__PYX_ERR(1, 119, __pyx_L1_error))
  __pyx_v_nap2 = ((0.5 * __pyx_v_ap2) - (0.5 * __pyx_v_apz2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":121
 *         cdef double nap2 = 0.5*ap2 - 0.5*apz2
 * 
 *         cdef double BpSprec2 = -nap2/r**2 + (nam2*(0.1875 - 0.75*nu) + nap2*(-1.75*nu - 0.9375) + napnam*(0.75 - 1.5*X_2))/r**3 + (-0.125*delta*napnam*(98.0*nu + 43.0) + 0.015625*nam2*(152.0*nu**2 - 1090.0*nu + 219.0) + 0.00520833333333333*nap2*(264.0*nu**2 - 1610.0*nu + 375.0))/r**4             # <<<<<<<<<<<<<<
 * 
 *         cdef double Bp = BpSprec2 + 1.0
 */
  __Pyx_TraceLine(121,0,__PYX_ERR(1, 121, __pyx_L1_error))
  __pyx_v_BpSprec2 = ((((-__pyx_v_nap2) / pow(__pyx_v_r, 2.0)) + ((((__pyx_v_nam2 * (0.1875 - (0.75 * __pyx_v_nu))) + (__pyx_v_nap2 * ((-1.75 * __pyx_v_nu) - 0.9375))) + (__pyx_v_napnam * (0.75 - (1.5 * __pyx_v_X_2)))) / pow(__pyx_v_r, 3.0))) + ((((((-0.125 * __pyx_v_delta) * __pyx_v_napnam) * ((98.0 * __pyx_v_nu) + 43.0)) + ((0.015625 * __pyx_v_nam2) * (((152.0 * pow(__pyx_v_nu, 2.0)) - (1090.0 * __pyx_v_nu)) + 219.0))) + ((0.00520833333333333 * __pyx_v_nap2) * (((264.0 * pow(__pyx_v_nu, 2.0)) - (1610.0 * __pyx_v_nu)) + 375.0))) / pow(__pyx_v_r, 4.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":123
 *         cdef double BpSprec2 = -nap2/r**2 + (nam2*(0.1875 - 0.75*nu) + nap2*(-1.75*nu - 0.9375) + napnam*(0.75 - 1.5*X_2))/r**3 + (-0.125*delta*napnam*(98.0*nu + 43.0) + 0.015625*nam2*(152.0*nu**2 - 1090.0*nu + 219.0) + 0.00520833333333333*nap2*(264.0*nu**2 - 1610.0*nu + 375.0))/r**4
 * 
 *         cdef double Bp = BpSprec2 + 1.0             # <<<<<<<<<<<<<<
 * 
 *         cdef double ASprec2 = 2.0*nap2/r**3 + (4.125*delta*napnam + 0.125*nam2*(-4.0*nu - 3.0) + 0.25*nap2*(7.0*nu - 31.0))/r**4 + (0.25*delta*napnam*(68.0*nu - 1.0) + 0.015625*nam2*(-328.0*nu**2 + 1166.0*nu - 171.0) + 0.00520833333333333*nap2*(-264.0*nu**2 + 2870.0*nu + 561.0))/r**5
 */
  __Pyx_TraceLine(123,0,__PYX_ERR(1, 123, __pyx_L1_error))
  __pyx_v_Bp = (__pyx_v_BpSprec2 + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":125
 *         cdef double Bp = BpSprec2 + 1.0
 * 
 *         cdef double ASprec2 = 2.0*nap2/r**3 + (4.125*delta*napnam + 0.125*nam2*(-4.0*nu - 3.0) + 0.25*nap2*(7.0*nu - 31.0))/r**4 + (0.25*delta*napnam*(68.0*nu - 1.0) + 0.015625*nam2*(-328.0*nu**2 + 1166.0*nu - 171.0) + 0.00520833333333333*nap2*(-264.0*nu**2 + 2870.0*nu + 561.0))/r**5             # <<<<<<<<<<<<<<
 * 
 *         cdef double ASalign2 = (0.125*am2*(4.0*nu + 1.0) + 1.125*ap2 - 1.25*apam*delta)/r**4 + (0.046875*am2*(28.0*nu**2 - 27.0*nu - 3.0) - 0.390625*ap2*(7.0*nu + 9.0) - 1.21875*apam*delta*(2.0*nu - 3.0))/r**5
 */
  __Pyx_TraceLine(125,0,__PYX_ERR(1, 125, __pyx_L1_error))
  __pyx_v_ASprec2 = ((((2.0 * __pyx_v_nap2) / pow(__pyx_v_r, 3.0)) + (((((4.125 * __pyx_v_delta) * __pyx_v_napnam) + ((0.125 * __pyx_v_nam2) * ((-4.0 * __pyx_v_nu) - 3.0))) + ((0.25 * __pyx_v_nap2) * ((7.0 * __pyx_v_nu) - 31.0))) / pow(__pyx_v_r, 4.0))) + ((((((0.25 * __pyx_v_delta) * __pyx_v_napnam) * ((68.0 * __pyx_v_nu) - 1.0)) + ((0.015625 * __pyx_v_nam2) * (((-328.0 * pow(__pyx_v_nu, 2.0)) + (1166.0 * __pyx_v_nu)) - 171.0))) + ((0.00520833333333333 * __pyx_v_nap2) * (((-264.0 * pow(__pyx_v_nu, 2.0)) + (2870.0 * __pyx_v_nu)) + 561.0))) / pow(__pyx_v_r, 5.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":127
 *         cdef double ASprec2 = 2.0*nap2/r**3 + (4.125*delta*napnam + 0.125*nam2*(-4.0*nu - 3.0) + 0.25*nap2*(7.0*nu - 31.0))/r**4 + (0.25*delta*napnam*(68.0*nu - 1.0) + 0.015625*nam2*(-328.0*nu**2 + 1166.0*nu - 171.0) + 0.00520833333333333*nap2*(-264.0*nu**2 + 2870.0*nu + 561.0))/r**5
 * 
 *         cdef double ASalign2 = (0.125*am2*(4.0*nu + 1.0) + 1.125*ap2 - 1.25*apam*delta)/r**4 + (0.046875*am2*(28.0*nu**2 - 27.0*nu - 3.0) - 0.390625*ap2*(7.0*nu + 9.0) - 1.21875*apam*delta*(2.0*nu - 3.0))/r**5             # <<<<<<<<<<<<<<
 * 
 *         cdef double A = (ASalign2 + ASprec2 + Apm + ap2/r**2)/(ap2*(1.0 + 2.0/r)/r**2 + 1.0)
 */
  __Pyx_TraceLine(127,0,__PYX_ERR(1, 127, __pyx_L1_error))
  __pyx_v_ASalign2 = ((((((0.125 * __pyx_v_am2) * ((4.0 * __pyx_v_nu) + 1.0)) + (1.125 * __pyx_v_ap2)) - ((1.25 * __pyx_v_apam) * __pyx_v_delta)) / pow(__pyx_v_r, 4.0)) + (((((0.046875 * __pyx_v_am2) * (((28.0 * pow(__pyx_v_nu, 2.0)) - (27.0 * __pyx_v_nu)) - 3.0)) - ((0.390625 * __pyx_v_ap2) * ((7.0 * __pyx_v_nu) + 9.0))) - (((1.21875 * __pyx_v_apam) * __pyx_v_delta) * ((2.0 * __pyx_v_nu) - 3.0))) / pow(__pyx_v_r, 5.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":129
 *         cdef double ASalign2 = (0.125*am2*(4.0*nu + 1.0) + 1.125*ap2 - 1.25*apam*delta)/r**4 + (0.046875*am2*(28.0*nu**2 - 27.0*nu - 3.0) - 0.390625*ap2*(7.0*nu + 9.0) - 1.21875*apam*delta*(2.0*nu - 3.0))/r**5
 * 
 *         cdef double A = (ASalign2 + ASprec2 + Apm + ap2/r**2)/(ap2*(1.0 + 2.0/r)/r**2 + 1.0)             # <<<<<<<<<<<<<<
 * 
 *         cdef double ap = X_1*chiL1 + X_2*chiL2
 */
  __Pyx_TraceLine(129,0,__PYX_ERR(1, 129, __pyx_L1_error))
  __pyx_v_A = ((((__pyx_v_ASalign2 + __pyx_v_ASprec2) + __pyx_v_Apm) + (__pyx_v_ap2 / pow(__pyx_v_r, 2.0))) / (((__pyx_v_ap2 * (1.0 + (2.0 / __pyx_v_r))) / pow(__pyx_v_r, 2.0)) + 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":131
 *         cdef double A = (ASalign2 + ASprec2 + Apm + ap2/r**2)/(ap2*(1.0 + 2.0/r)/r**2 + 1.0)
 * 
 *         cdef double ap = X_1*chiL1 + X_2*chiL2             # <<<<<<<<<<<<<<
 * 
 *         cdef double lap = ap
 */
  __Pyx_TraceLine(131,0,__PYX_ERR(1, 131, __pyx_L1_error))
  __pyx_v_ap = ((__pyx_v_X_1 * __pyx_v_chiL1) + (__pyx_v_X_2 * __pyx_v_chiL2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":133
 *         cdef double ap = X_1*chiL1 + X_2*chiL2
 * 
 *         cdef double lap = ap             # <<<<<<<<<<<<<<
 * 
 *         cdef double Heven = (A*(Bnpa*L**2*lap**2/r**2 + Bp*L**2/r**2 + Qq + prst**2*(Bnp + 1.0)/xi**2 + 1.0))**0.5
 */
  __Pyx_TraceLine(133,0,__PYX_ERR(1, 133, __pyx_L1_error))
  __pyx_v_lap = __pyx_v_ap;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":135
 *         cdef double lap = ap
 * 
 *         cdef double Heven = (A*(Bnpa*L**2*lap**2/r**2 + Bp*L**2/r**2 + Qq + prst**2*(Bnp + 1.0)/xi**2 + 1.0))**0.5             # <<<<<<<<<<<<<<
 * 
 *         cdef double am = X_1*chiL1 - X_2*chiL2
 */
  __Pyx_TraceLine(135,0,__PYX_ERR(1, 135, __pyx_L1_error))
  __pyx_v_Heven = pow((__pyx_v_A * (((((((__pyx_v_Bnpa * pow(__pyx_v_L, 2.0)) * pow(__pyx_v_lap, 2.0)) / pow(__pyx_v_r, 2.0)) + ((__pyx_v_Bp * pow(__pyx_v_L, 2.0)) / pow(__pyx_v_r, 2.0))) + __pyx_v_Qq) + ((pow(__pyx_v_prst, 2.0) * (__pyx_v_Bnp + 1.0)) / pow(__pyx_v_xi, 2.0))) + 1.0)), 0.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":137
 *         cdef double Heven = (A*(Bnpa*L**2*lap**2/r**2 + Bp*L**2/r**2 + Qq + prst**2*(Bnp + 1.0)/xi**2 + 1.0))**0.5
 * 
 *         cdef double am = X_1*chiL1 - X_2*chiL2             # <<<<<<<<<<<<<<
 * 
 *         cdef double lam = am
 */
  __Pyx_TraceLine(137,0,__PYX_ERR(1, 137, __pyx_L1_error))
  __pyx_v_am = ((__pyx_v_X_1 * __pyx_v_chiL1) - (__pyx_v_X_2 * __pyx_v_chiL2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":139
 *         cdef double am = X_1*chiL1 - X_2*chiL2
 * 
 *         cdef double lam = am             # <<<<<<<<<<<<<<
 * 
 *         cdef double Ga3 = L*lam*(-0.25*L**2*delta*nap2/r**3 + (0.0416666666666667*ap2*delta + nap2*(0.416666666666667 - 0.833333333333333*X_2))/r**2) + L*lap*(L**2*(-0.25*nap2 + napnam*(0.5 - X_2))/r**3 + (-0.25*ap2 + 0.208333333333333*apam*delta - 1.66666666666667*delta*napnam - 0.75*nap2)/r**2)
 */
  __Pyx_TraceLine(139,0,__PYX_ERR(1, 139, __pyx_L1_error))
  __pyx_v_lam = __pyx_v_am;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":141
 *         cdef double lam = am
 * 
 *         cdef double Ga3 = L*lam*(-0.25*L**2*delta*nap2/r**3 + (0.0416666666666667*ap2*delta + nap2*(0.416666666666667 - 0.833333333333333*X_2))/r**2) + L*lap*(L**2*(-0.25*nap2 + napnam*(0.5 - X_2))/r**3 + (-0.25*ap2 + 0.208333333333333*apam*delta - 1.66666666666667*delta*napnam - 0.75*nap2)/r**2)             # <<<<<<<<<<<<<<
 * 
 *         cdef double SOcalib = L*nu*dSO*lap/r**3
 */
  __Pyx_TraceLine(141,0,__PYX_ERR(1, 141, __pyx_L1_error))
  __pyx_v_Ga3 = (((__pyx_v_L * __pyx_v_lam) * (((((-0.25 * pow(__pyx_v_L, 2.0)) * __pyx_v_delta) * __pyx_v_nap2) / pow(__pyx_v_r, 3.0)) + ((((0.0416666666666667 * __pyx_v_ap2) * __pyx_v_delta) + (__pyx_v_nap2 * (0.416666666666667 - (0.833333333333333 * __pyx_v_X_2)))) / pow(__pyx_v_r, 2.0)))) + ((__pyx_v_L * __pyx_v_lap) * (((pow(__pyx_v_L, 2.0) * ((-0.25 * __pyx_v_nap2) + (__pyx_v_napnam * (0.5 - __pyx_v_X_2)))) / pow(__pyx_v_r, 3.0)) + (((((-0.25 * __pyx_v_ap2) + ((0.208333333333333 * __pyx_v_apam) * __pyx_v_delta)) - ((1.66666666666667 * __pyx_v_delta) * __pyx_v_napnam)) - (0.75 * __pyx_v_nap2)) / pow(__pyx_v_r, 2.0)))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":143
 *         cdef double Ga3 = L*lam*(-0.25*L**2*delta*nap2/r**3 + (0.0416666666666667*ap2*delta + nap2*(0.416666666666667 - 0.833333333333333*X_2))/r**2) + L*lap*(L**2*(-0.25*nap2 + napnam*(0.5 - X_2))/r**3 + (-0.25*ap2 + 0.208333333333333*apam*delta - 1.66666666666667*delta*napnam - 0.75*nap2)/r**2)
 * 
 *         cdef double SOcalib = L*nu*dSO*lap/r**3             # <<<<<<<<<<<<<<
 * 
 *         cdef double gam = L**4*(0.29296875*nu**2 - 0.3515625*nu - 0.41015625)/r**4 + L**2*(0.46875 - 0.28125*nu)/r**2 + L**2*(-0.798177083333333*nu**2 - 0.2734375*nu - 0.23046875)/r**3 + 0.25 + (0.34375*nu + 0.09375)/r + (0.536458333333333*nu**2 - 0.03125*nu + 0.078125)/r**2
 */
  __Pyx_TraceLine(143,0,__PYX_ERR(1, 143, __pyx_L1_error))
  __pyx_v_SOcalib = ((((__pyx_v_L * __pyx_v_nu) * __pyx_v_dSO) * __pyx_v_lap) / pow(__pyx_v_r, 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":145
 *         cdef double SOcalib = L*nu*dSO*lap/r**3
 * 
 *         cdef double gam = L**4*(0.29296875*nu**2 - 0.3515625*nu - 0.41015625)/r**4 + L**2*(0.46875 - 0.28125*nu)/r**2 + L**2*(-0.798177083333333*nu**2 - 0.2734375*nu - 0.23046875)/r**3 + 0.25 + (0.34375*nu + 0.09375)/r + (0.536458333333333*nu**2 - 0.03125*nu + 0.078125)/r**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double gap = L**4*(1.34765625*nu**2 + 0.5859375*nu + 0.41015625)/r**4 + L**2*(-1.40625*nu - 0.46875)/r**2 + L**2*(-2.07161458333333*nu**2 - 2.0859375*nu + 0.23046875)/r**3 + 1.75 + (0.71875*nu - 0.09375)/r + (0.567708333333333*nu**2 - 5.53125*nu - 0.078125)/r**2
 */
  __Pyx_TraceLine(145,0,__PYX_ERR(1, 145, __pyx_L1_error))
  __pyx_v_gam = (((((((pow(__pyx_v_L, 4.0) * (((0.29296875 * pow(__pyx_v_nu, 2.0)) - (0.3515625 * __pyx_v_nu)) - 0.41015625)) / pow(__pyx_v_r, 4.0)) + ((pow(__pyx_v_L, 2.0) * (0.46875 - (0.28125 * __pyx_v_nu))) / pow(__pyx_v_r, 2.0))) + ((pow(__pyx_v_L, 2.0) * (((-0.798177083333333 * pow(__pyx_v_nu, 2.0)) - (0.2734375 * __pyx_v_nu)) - 0.23046875)) / pow(__pyx_v_r, 3.0))) + 0.25) + (((0.34375 * __pyx_v_nu) + 0.09375) / __pyx_v_r)) + ((((0.536458333333333 * pow(__pyx_v_nu, 2.0)) - (0.03125 * __pyx_v_nu)) + 0.078125) / pow(__pyx_v_r, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":147
 *         cdef double gam = L**4*(0.29296875*nu**2 - 0.3515625*nu - 0.41015625)/r**4 + L**2*(0.46875 - 0.28125*nu)/r**2 + L**2*(-0.798177083333333*nu**2 - 0.2734375*nu - 0.23046875)/r**3 + 0.25 + (0.34375*nu + 0.09375)/r + (0.536458333333333*nu**2 - 0.03125*nu + 0.078125)/r**2
 * 
 *         cdef double gap = L**4*(1.34765625*nu**2 + 0.5859375*nu + 0.41015625)/r**4 + L**2*(-1.40625*nu - 0.46875)/r**2 + L**2*(-2.07161458333333*nu**2 - 2.0859375*nu + 0.23046875)/r**3 + 1.75 + (0.71875*nu - 0.09375)/r + (0.567708333333333*nu**2 - 5.53125*nu - 0.078125)/r**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double Hodd = (Ga3 + L*delta*gam*lam + L*gap*lap + SOcalib)/(ap2*(r + 2.0) + r**3)
 */
  __Pyx_TraceLine(147,0,__PYX_ERR(1, 147, __pyx_L1_error))
  __pyx_v_gap = (((((((pow(__pyx_v_L, 4.0) * (((1.34765625 * pow(__pyx_v_nu, 2.0)) + (0.5859375 * __pyx_v_nu)) + 0.41015625)) / pow(__pyx_v_r, 4.0)) + ((pow(__pyx_v_L, 2.0) * ((-1.40625 * __pyx_v_nu) - 0.46875)) / pow(__pyx_v_r, 2.0))) + ((pow(__pyx_v_L, 2.0) * (((-2.07161458333333 * pow(__pyx_v_nu, 2.0)) - (2.0859375 * __pyx_v_nu)) + 0.23046875)) / pow(__pyx_v_r, 3.0))) + 1.75) + (((0.71875 * __pyx_v_nu) - 0.09375) / __pyx_v_r)) + ((((0.567708333333333 * pow(__pyx_v_nu, 2.0)) - (5.53125 * __pyx_v_nu)) - 0.078125) / pow(__pyx_v_r, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":149
 *         cdef double gap = L**4*(1.34765625*nu**2 + 0.5859375*nu + 0.41015625)/r**4 + L**2*(-1.40625*nu - 0.46875)/r**2 + L**2*(-2.07161458333333*nu**2 - 2.0859375*nu + 0.23046875)/r**3 + 1.75 + (0.71875*nu - 0.09375)/r + (0.567708333333333*nu**2 - 5.53125*nu - 0.078125)/r**2
 * 
 *         cdef double Hodd = (Ga3 + L*delta*gam*lam + L*gap*lap + SOcalib)/(ap2*(r + 2.0) + r**3)             # <<<<<<<<<<<<<<
 * 
 *         cdef double pr = prst/xi
 */
  __Pyx_TraceLine(149,0,__PYX_ERR(1, 149, __pyx_L1_error))
  __pyx_v_Hodd = ((((__pyx_v_Ga3 + (((__pyx_v_L * __pyx_v_delta) * __pyx_v_gam) * __pyx_v_lam)) + ((__pyx_v_L * __pyx_v_gap) * __pyx_v_lap)) + __pyx_v_SOcalib) / ((__pyx_v_ap2 * (__pyx_v_r + 2.0)) + pow(__pyx_v_r, 3.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":151
 *         cdef double Hodd = (Ga3 + L*delta*gam*lam + L*gap*lap + SOcalib)/(ap2*(r + 2.0) + r**3)
 * 
 *         cdef double pr = prst/xi             # <<<<<<<<<<<<<<
 * 
 *         cdef double Heff = Heven + Hodd
 */
  __Pyx_TraceLine(151,0,__PYX_ERR(1, 151, __pyx_L1_error))
  __pyx_v_pr = (__pyx_v_prst / __pyx_v_xi);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":153
 *         cdef double pr = prst/xi
 * 
 *         cdef double Heff = Heven + Hodd             # <<<<<<<<<<<<<<
 * 
 *         # Evaluate H_real/nu
 */
  __Pyx_TraceLine(153,0,__PYX_ERR(1, 153, __pyx_L1_error))
  __pyx_v_Heff = (__pyx_v_Heven + __pyx_v_Hodd);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":156
 * 
 *         # Evaluate H_real/nu
 *         cdef double H = M * sqrt(1+2*nu*(Heff-1)) / nu             # <<<<<<<<<<<<<<
 *         return H,xi, A, Bnp, Bnpa, Qq, Heven, Hodd, Bp
 * 
 */
  __Pyx_TraceLine(156,0,__PYX_ERR(1, 156, __pyx_L1_error))
  __pyx_v_H = ((__pyx_v_M * sqrt((1.0 + ((2.0 * __pyx_v_nu) * (__pyx_v_Heff - 1.0))))) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":157
 *         # Evaluate H_real/nu
 *         cdef double H = M * sqrt(1+2*nu*(Heff-1)) / nu
 *         return H,xi, A, Bnp, Bnpa, Qq, Heven, Hodd, Bp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(157,0,__PYX_ERR(1, 157, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_H); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_xi); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_A); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_16 = PyFloat_FromDouble(__pyx_v_Bnp); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_12 = PyFloat_FromDouble(__pyx_v_Bnpa); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_11 = PyFloat_FromDouble(__pyx_v_Qq); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_Heven); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_Hodd); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_Bp); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = PyTuple_New(9); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_7, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_7, 8, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_13 = 0;
  __pyx_t_16 = 0;
  __pyx_t_12 = 0;
  __pyx_t_11 = 0;
  __pyx_t_10 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":25
 *         return self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)
 * 
 *     cpdef _call(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 *         """
 *         Evaluate the Hamiltonian as well as several potentials.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD._call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c_coeffs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_5_call(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_4_call[] = "\n        Evaluate the Hamiltonian as well as several potentials.\n\n        Args:\n          q (double[:]): Canonical positions (r,phi).\n          p (double[:]): Canonical momenta  (prstar,pphi).\n          chi1_v (double[:]): Dimensionless spin vector of the primary.\n          chi2_v (double[:]): Dimensionless spin vector of the secondary.\n          m_1 (double): Primary mass component.\n          m_2 (double): Secondary mass component.\n          chi_1 (double): Projection of chi1_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chi_2 (double): Projection of chi2_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chiL1 (double): Projection of chi1_v onto the orbital angular momentum unit vector (l).\n          chiL2 (double): Projection of chi2_v onto the orbital angular momentum unit vector (l).\n\n        Returns:\n           (tuple)  H,xi, A, Bnp, Bnpa, Qq, Heven, Hodd, Bp\n\n        ";
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_5_call(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_q = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi1_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi2_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_m_1;
  double __pyx_v_m_2;
  double __pyx_v_chi_1;
  double __pyx_v_chi_2;
  double __pyx_v_chiL1;
  double __pyx_v_chiL2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_call (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_q,&__pyx_n_s_p,&__pyx_n_s_chi1_v,&__pyx_n_s_chi2_v,&__pyx_n_s_m_1,&__pyx_n_s_m_2,&__pyx_n_s_chi_1,&__pyx_n_s_chi_2,&__pyx_n_s_chiL1,&__pyx_n_s_chiL2,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_q)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 1); __PYX_ERR(1, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi1_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 2); __PYX_ERR(1, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi2_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 3); __PYX_ERR(1, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 4); __PYX_ERR(1, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 5); __PYX_ERR(1, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 6); __PYX_ERR(1, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 7); __PYX_ERR(1, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 8); __PYX_ERR(1, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, 9); __PYX_ERR(1, 25, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_call") < 0)) __PYX_ERR(1, 25, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_q = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_q.memview)) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_p.memview)) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_chi1_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi1_v.memview)) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_chi2_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi2_v.memview)) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_m_1 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_m_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_m_2 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_m_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_chi_1 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_chi_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_chi_2 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_chi_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_chiL1 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_chiL1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L3_error)
    __pyx_v_chiL2 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_chiL2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_call", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 25, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD._call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_4_call(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_4_call(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call", 0);
  __Pyx_TraceCall("_call (wrapper)", __pyx_f[1], 25, 0, __PYX_ERR(1, 25, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 25, __pyx_L1_error) }
  if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 25, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 25, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 25, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD__call(__pyx_v_self, __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD._call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_q, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi1_v, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi2_v, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":160
 * 
 * 
 *     cpdef grad(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 *         """
 *         Compute the gradient of the Hamiltonian in polar coordinates.
 */

static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_7grad(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_grad(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch) {
  double __pyx_v_r;
  CYTHON_UNUSED double __pyx_v_phi;
  double __pyx_v_prst;
  double __pyx_v_L;
  double __pyx_v_chix1;
  double __pyx_v_chiy1;
  double __pyx_v_chiz1;
  double __pyx_v_chix2;
  double __pyx_v_chiy2;
  double __pyx_v_chiz2;
  double __pyx_v_pphi;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *__pyx_v_c_coeffs = 0;
  double __pyx_v_a6;
  double __pyx_v_dSO;
  double __pyx_v_M;
  double __pyx_v_nu;
  double __pyx_v_X_1;
  double __pyx_v_X_2;
  double __pyx_v_delta;
  double __pyx_v_x0;
  double __pyx_v_x1;
  double __pyx_v_x2;
  double __pyx_v_x3;
  double __pyx_v_x4;
  double __pyx_v_x5;
  double __pyx_v_x6;
  double __pyx_v_x7;
  double __pyx_v_x8;
  double __pyx_v_x9;
  double __pyx_v_x10;
  double __pyx_v_x11;
  double __pyx_v_x12;
  double __pyx_v_x13;
  double __pyx_v_x14;
  double __pyx_v_x15;
  double __pyx_v_x16;
  double __pyx_v_x17;
  double __pyx_v_x18;
  double __pyx_v_x19;
  double __pyx_v_x20;
  double __pyx_v_x21;
  double __pyx_v_x22;
  double __pyx_v_x23;
  double __pyx_v_x24;
  double __pyx_v_x25;
  double __pyx_v_x26;
  double __pyx_v_x27;
  double __pyx_v_x28;
  double __pyx_v_x29;
  double __pyx_v_x30;
  double __pyx_v_x31;
  double __pyx_v_x32;
  double __pyx_v_x33;
  double __pyx_v_x34;
  double __pyx_v_x35;
  double __pyx_v_x36;
  double __pyx_v_x37;
  double __pyx_v_x38;
  double __pyx_v_x39;
  double __pyx_v_x40;
  double __pyx_v_x41;
  double __pyx_v_x42;
  double __pyx_v_x43;
  double __pyx_v_x44;
  double __pyx_v_x45;
  double __pyx_v_x46;
  double __pyx_v_x47;
  double __pyx_v_x48;
  double __pyx_v_x49;
  double __pyx_v_x50;
  double __pyx_v_x51;
  double __pyx_v_x52;
  double __pyx_v_x53;
  double __pyx_v_x54;
  double __pyx_v_x55;
  double __pyx_v_x56;
  double __pyx_v_x57;
  double __pyx_v_x58;
  double __pyx_v_x59;
  double __pyx_v_x60;
  double __pyx_v_x61;
  double __pyx_v_x62;
  double __pyx_v_x63;
  double __pyx_v_x64;
  double __pyx_v_x65;
  double __pyx_v_x66;
  double __pyx_v_x67;
  double __pyx_v_x68;
  double __pyx_v_x69;
  double __pyx_v_x70;
  double __pyx_v_x71;
  double __pyx_v_x72;
  double __pyx_v_x73;
  double __pyx_v_x74;
  double __pyx_v_x75;
  double __pyx_v_x76;
  double __pyx_v_x77;
  double __pyx_v_x78;
  double __pyx_v_x79;
  double __pyx_v_x80;
  double __pyx_v_x81;
  double __pyx_v_x82;
  double __pyx_v_x83;
  double __pyx_v_x84;
  double __pyx_v_x85;
  double __pyx_v_x86;
  double __pyx_v_x87;
  double __pyx_v_x88;
  double __pyx_v_x89;
  double __pyx_v_x90;
  double __pyx_v_x91;
  double __pyx_v_x92;
  double __pyx_v_x93;
  double __pyx_v_x94;
  double __pyx_v_x95;
  double __pyx_v_x96;
  double __pyx_v_x97;
  double __pyx_v_x98;
  double __pyx_v_x99;
  double __pyx_v_x100;
  double __pyx_v_x101;
  double __pyx_v_x102;
  double __pyx_v_x103;
  double __pyx_v_x104;
  double __pyx_v_x105;
  double __pyx_v_x106;
  double __pyx_v_x107;
  double __pyx_v_x108;
  double __pyx_v_x109;
  double __pyx_v_x110;
  double __pyx_v_x111;
  double __pyx_v_x112;
  double __pyx_v_x113;
  double __pyx_v_x114;
  double __pyx_v_x115;
  double __pyx_v_x116;
  double __pyx_v_x117;
  double __pyx_v_x118;
  double __pyx_v_x119;
  double __pyx_v_x120;
  double __pyx_v_x121;
  double __pyx_v_x122;
  double __pyx_v_x123;
  double __pyx_v_x124;
  double __pyx_v_x125;
  double __pyx_v_x126;
  double __pyx_v_x127;
  double __pyx_v_x128;
  double __pyx_v_x129;
  double __pyx_v_x130;
  double __pyx_v_x131;
  double __pyx_v_x132;
  double __pyx_v_x133;
  double __pyx_v_x134;
  double __pyx_v_x135;
  double __pyx_v_x136;
  double __pyx_v_x137;
  double __pyx_v_x138;
  double __pyx_v_x139;
  double __pyx_v_x140;
  double __pyx_v_x141;
  double __pyx_v_x142;
  double __pyx_v_x143;
  double __pyx_v_x144;
  double __pyx_v_x145;
  double __pyx_v_x146;
  double __pyx_v_x147;
  double __pyx_v_x148;
  double __pyx_v_x149;
  double __pyx_v_x150;
  double __pyx_v_x151;
  double __pyx_v_x152;
  double __pyx_v_x153;
  double __pyx_v_x154;
  double __pyx_v_x155;
  double __pyx_v_x156;
  double __pyx_v_x157;
  double __pyx_v_x158;
  double __pyx_v_x159;
  double __pyx_v_x160;
  double __pyx_v_x161;
  double __pyx_v_x162;
  double __pyx_v_x163;
  double __pyx_v_x164;
  double __pyx_v_x165;
  double __pyx_v_x166;
  double __pyx_v_x167;
  double __pyx_v_x168;
  double __pyx_v_x169;
  double __pyx_v_x170;
  double __pyx_v_x171;
  double __pyx_v_x172;
  double __pyx_v_x173;
  double __pyx_v_x174;
  double __pyx_v_x175;
  double __pyx_v_x176;
  double __pyx_v_x177;
  double __pyx_v_x178;
  double __pyx_v_x179;
  double __pyx_v_x180;
  double __pyx_v_x181;
  double __pyx_v_x182;
  double __pyx_v_x183;
  double __pyx_v_x184;
  double __pyx_v_x185;
  double __pyx_v_x186;
  double __pyx_v_x187;
  double __pyx_v_x188;
  double __pyx_v_x189;
  double __pyx_v_x190;
  double __pyx_v_x191;
  double __pyx_v_x192;
  double __pyx_v_x193;
  double __pyx_v_x194;
  double __pyx_v_x195;
  double __pyx_v_x196;
  double __pyx_v_x197;
  double __pyx_v_x198;
  double __pyx_v_x199;
  double __pyx_v_x200;
  double __pyx_v_x201;
  double __pyx_v_x202;
  double __pyx_v_x203;
  double __pyx_v_x204;
  double __pyx_v_x205;
  double __pyx_v_x206;
  double __pyx_v_x207;
  double __pyx_v_x208;
  double __pyx_v_x209;
  double __pyx_v_x210;
  double __pyx_v_x211;
  double __pyx_v_x212;
  double __pyx_v_x213;
  double __pyx_v_x214;
  double __pyx_v_x215;
  double __pyx_v_x216;
  double __pyx_v_x217;
  double __pyx_v_x218;
  double __pyx_v_x219;
  double __pyx_v_x220;
  double __pyx_v_x221;
  double __pyx_v_x222;
  double __pyx_v_x223;
  double __pyx_v_x224;
  double __pyx_v_x225;
  double __pyx_v_x226;
  double __pyx_v_x227;
  double __pyx_v_x228;
  double __pyx_v_x229;
  double __pyx_v_x230;
  double __pyx_v_x231;
  double __pyx_v_x232;
  double __pyx_v_x233;
  double __pyx_v_x234;
  double __pyx_v_H;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  double __pyx_v_dHeffdr;
  double __pyx_v_dHeffdphi;
  double __pyx_v_dHeffdpr;
  double __pyx_v_dHeffdpphi;
  double __pyx_v_dHdr;
  double __pyx_v_dHdphi;
  double __pyx_v_dHdpr;
  double __pyx_v_dHdpphi;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  double __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("grad", 0);
  __Pyx_TraceCall("grad", __pyx_f[1], 160, 0, __PYX_ERR(1, 160, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_grad); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_7grad)) {
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 160, __pyx_L1_error) }
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_q, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 160, __pyx_L1_error) }
        __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_p, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 160, __pyx_L1_error) }
        __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_chi1_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 160, __pyx_L1_error) }
        __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_chi2_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_m_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyFloat_FromDouble(__pyx_v_m_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_chi_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyFloat_FromDouble(__pyx_v_chi_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_chiL1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = PyFloat_FromDouble(__pyx_v_chiL2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_13 = __pyx_t_1; __pyx_t_14 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 160, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 160, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(10+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 160, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_15, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_15, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_15, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_15, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_15, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_15, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_16, 8+__pyx_t_15, __pyx_t_11);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_16, 9+__pyx_t_15, __pyx_t_12);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_12 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 160, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":183
 *         # Coordinate definitions
 * 
 *         cdef double r = q[0]             # <<<<<<<<<<<<<<
 *         cdef double phi = q[1]
 * 
 */
  __Pyx_TraceLine(183,0,__PYX_ERR(1, 183, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_r = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":184
 * 
 *         cdef double r = q[0]
 *         cdef double phi = q[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double prst = p[0]
 */
  __Pyx_TraceLine(184,0,__PYX_ERR(1, 184, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_phi = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":186
 *         cdef double phi = q[1]
 * 
 *         cdef double prst = p[0]             # <<<<<<<<<<<<<<
 *         cdef double L = p[1]
 * 
 */
  __Pyx_TraceLine(186,0,__PYX_ERR(1, 186, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_prst = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":187
 * 
 *         cdef double prst = p[0]
 *         cdef double L = p[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix1 = chi1_v[0]
 */
  __Pyx_TraceLine(187,0,__PYX_ERR(1, 187, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_L = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":189
 *         cdef double L = p[1]
 * 
 *         cdef double chix1 = chi1_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]
 */
  __Pyx_TraceLine(189,0,__PYX_ERR(1, 189, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":190
 * 
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz1 = chi1_v[2]
 * 
 */
  __Pyx_TraceLine(190,0,__PYX_ERR(1, 190, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":191
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix2 = chi2_v[0]
 */
  __Pyx_TraceLine(191,0,__PYX_ERR(1, 191, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":193
 *         cdef double chiz1 = chi1_v[2]
 * 
 *         cdef double chix2 = chi2_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]
 */
  __Pyx_TraceLine(193,0,__PYX_ERR(1, 193, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":194
 * 
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz2 = chi2_v[2]
 * 
 */
  __Pyx_TraceLine(194,0,__PYX_ERR(1, 194, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":195
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double pphi = L
 */
  __Pyx_TraceLine(195,0,__PYX_ERR(1, 195, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":197
 *         cdef double chiz2 = chi2_v[2]
 * 
 *         cdef double pphi = L             # <<<<<<<<<<<<<<
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 */
  __Pyx_TraceLine(197,0,__PYX_ERR(1, 197, __pyx_L1_error))
  __pyx_v_pphi = __pyx_v_L;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":199
 *         cdef double pphi = L
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs             # <<<<<<<<<<<<<<
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']
 */
  __Pyx_TraceLine(199,0,__PYX_ERR(1, 199, __pyx_L1_error))
  __pyx_t_1 = ((PyObject *)__pyx_v_self->calibration_coeffs);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_c_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":200
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']             # <<<<<<<<<<<<<<
 *         cdef double dSO = c_coeffs['dSO']
 * 
 */
  __Pyx_TraceLine(200,0,__PYX_ERR(1, 200, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_a6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_a6 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":201
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(201,0,__PYX_ERR(1, 201, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_dSO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dSO = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":205
 * 
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M             # <<<<<<<<<<<<<<
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 */
  __Pyx_TraceLine(205,0,__PYX_ERR(1, 205, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->M;
  __pyx_v_M = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":206
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu             # <<<<<<<<<<<<<<
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 */
  __Pyx_TraceLine(206,0,__PYX_ERR(1, 206, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->nu;
  __pyx_v_nu = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":207
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1             # <<<<<<<<<<<<<<
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 */
  __Pyx_TraceLine(207,0,__PYX_ERR(1, 207, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_1;
  __pyx_v_X_1 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":208
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2             # <<<<<<<<<<<<<<
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3
 */
  __Pyx_TraceLine(208,0,__PYX_ERR(1, 208, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_2;
  __pyx_v_X_2 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":209
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta             # <<<<<<<<<<<<<<
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0
 */
  __Pyx_TraceLine(209,0,__PYX_ERR(1, 209, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->delta;
  __pyx_v_delta = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":210
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3             # <<<<<<<<<<<<<<
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 */
  __Pyx_TraceLine(210,0,__PYX_ERR(1, 210, __pyx_L1_error))
  __pyx_v_x0 = pow(__pyx_v_r, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":211
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0             # <<<<<<<<<<<<<<
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2
 */
  __Pyx_TraceLine(211,0,__PYX_ERR(1, 211, __pyx_L1_error))
  __pyx_v_x1 = (__pyx_v_r + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":212
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)             # <<<<<<<<<<<<<<
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2
 */
  __Pyx_TraceLine(212,0,__PYX_ERR(1, 212, __pyx_L1_error))
  __pyx_v_x2 = (__pyx_v_X_1 * ((((2.0 * __pyx_v_chix1) * __pyx_v_chix2) + ((2.0 * __pyx_v_chiy1) * __pyx_v_chiy2)) + ((2.0 * __pyx_v_chiz1) * __pyx_v_chiz2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":213
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2             # <<<<<<<<<<<<<<
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 */
  __Pyx_TraceLine(213,0,__PYX_ERR(1, 213, __pyx_L1_error))
  __pyx_v_x3 = (__pyx_v_X_2 * __pyx_v_x2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":214
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2             # <<<<<<<<<<<<<<
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5
 */
  __Pyx_TraceLine(214,0,__PYX_ERR(1, 214, __pyx_L1_error))
  __pyx_v_x4 = pow(__pyx_v_X_1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":215
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2             # <<<<<<<<<<<<<<
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 */
  __Pyx_TraceLine(215,0,__PYX_ERR(1, 215, __pyx_L1_error))
  __pyx_v_x5 = ((pow(__pyx_v_chix1, 2.0) + pow(__pyx_v_chiy1, 2.0)) + pow(__pyx_v_chiz1, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":216
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5             # <<<<<<<<<<<<<<
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7
 */
  __Pyx_TraceLine(216,0,__PYX_ERR(1, 216, __pyx_L1_error))
  __pyx_v_x6 = (__pyx_v_x4 * __pyx_v_x5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":217
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)             # <<<<<<<<<<<<<<
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8
 */
  __Pyx_TraceLine(217,0,__PYX_ERR(1, 217, __pyx_L1_error))
  __pyx_v_x7 = (pow(__pyx_v_X_2, 2.0) * ((pow(__pyx_v_chix2, 2.0) + pow(__pyx_v_chiy2, 2.0)) + pow(__pyx_v_chiz2, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":218
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7             # <<<<<<<<<<<<<<
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9
 */
  __Pyx_TraceLine(218,0,__PYX_ERR(1, 218, __pyx_L1_error))
  __pyx_v_x8 = (__pyx_v_x6 + __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":219
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8             # <<<<<<<<<<<<<<
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10
 */
  __Pyx_TraceLine(219,0,__PYX_ERR(1, 219, __pyx_L1_error))
  __pyx_v_x9 = (__pyx_v_x3 + __pyx_v_x8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":220
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9             # <<<<<<<<<<<<<<
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)
 */
  __Pyx_TraceLine(220,0,__PYX_ERR(1, 220, __pyx_L1_error))
  __pyx_v_x10 = (__pyx_v_x1 * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":221
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10             # <<<<<<<<<<<<<<
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4
 */
  __Pyx_TraceLine(221,0,__PYX_ERR(1, 221, __pyx_L1_error))
  __pyx_v_x11 = (__pyx_v_x0 + __pyx_v_x10);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":222
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)
 */
  __Pyx_TraceLine(222,0,__PYX_ERR(1, 222, __pyx_L1_error))
  __pyx_v_x12 = pow(__pyx_v_x11, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":223
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4             # <<<<<<<<<<<<<<
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu
 */
  __Pyx_TraceLine(223,0,__PYX_ERR(1, 223, __pyx_L1_error))
  __pyx_v_x13 = pow(__pyx_v_r, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":224
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1
 */
  __Pyx_TraceLine(224,0,__PYX_ERR(1, 224, __pyx_L1_error))
  __pyx_v_x14 = pow(__pyx_v_x13, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":225
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2
 */
  __Pyx_TraceLine(225,0,__PYX_ERR(1, 225, __pyx_L1_error))
  __pyx_v_x15 = (3.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":226
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1             # <<<<<<<<<<<<<<
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17
 */
  __Pyx_TraceLine(226,0,__PYX_ERR(1, 226, __pyx_L1_error))
  __pyx_v_x16 = (__pyx_v_X_1 * __pyx_v_chiL1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":227
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2             # <<<<<<<<<<<<<<
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18
 */
  __Pyx_TraceLine(227,0,__PYX_ERR(1, 227, __pyx_L1_error))
  __pyx_v_x17 = (__pyx_v_X_2 * __pyx_v_chiL2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":228
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17             # <<<<<<<<<<<<<<
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2
 */
  __Pyx_TraceLine(228,0,__PYX_ERR(1, 228, __pyx_L1_error))
  __pyx_v_x18 = (__pyx_v_x16 + __pyx_v_x17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":229
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18             # <<<<<<<<<<<<<<
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)
 */
  __Pyx_TraceLine(229,0,__PYX_ERR(1, 229, __pyx_L1_error))
  __pyx_v_x19 = (__pyx_v_pphi * __pyx_v_x18);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":230
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2             # <<<<<<<<<<<<<<
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375
 */
  __Pyx_TraceLine(230,0,__PYX_ERR(1, 230, __pyx_L1_error))
  __pyx_v_x20 = pow(__pyx_v_r, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":231
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875
 */
  __Pyx_TraceLine(231,0,__PYX_ERR(1, 231, __pyx_L1_error))
  __pyx_v_x21 = pow(__pyx_v_x20, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":232
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375             # <<<<<<<<<<<<<<
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)
 */
  __Pyx_TraceLine(232,0,__PYX_ERR(1, 232, __pyx_L1_error))
  __pyx_v_x22 = ((0.71875 * __pyx_v_nu) - 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":233
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875             # <<<<<<<<<<<<<<
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2
 */
  __Pyx_TraceLine(233,0,__PYX_ERR(1, 233, __pyx_L1_error))
  __pyx_v_x23 = ((-1.40625 * __pyx_v_nu) - 0.46875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":234
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25
 */
  __Pyx_TraceLine(234,0,__PYX_ERR(1, 234, __pyx_L1_error))
  __pyx_v_x24 = pow(__pyx_v_x0, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":235
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2             # <<<<<<<<<<<<<<
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26
 */
  __Pyx_TraceLine(235,0,__PYX_ERR(1, 235, __pyx_L1_error))
  __pyx_v_x25 = pow(__pyx_v_pphi, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":236
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25             # <<<<<<<<<<<<<<
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2
 */
  __Pyx_TraceLine(236,0,__PYX_ERR(1, 236, __pyx_L1_error))
  __pyx_v_x26 = (__pyx_v_x24 * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":237
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26             # <<<<<<<<<<<<<<
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 */
  __Pyx_TraceLine(237,0,__PYX_ERR(1, 237, __pyx_L1_error))
  __pyx_v_x27 = (2.0 * __pyx_v_x26);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":238
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2             # <<<<<<<<<<<<<<
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14
 */
  __Pyx_TraceLine(238,0,__PYX_ERR(1, 238, __pyx_L1_error))
  __pyx_v_x28 = pow(__pyx_v_nu, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":239
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30
 */
  __Pyx_TraceLine(239,0,__PYX_ERR(1, 239, __pyx_L1_error))
  __pyx_v_x29 = (((-2.0859375 * __pyx_v_nu) - (2.07161458333333 * __pyx_v_x28)) + 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":240
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14             # <<<<<<<<<<<<<<
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 */
  __Pyx_TraceLine(240,0,__PYX_ERR(1, 240, __pyx_L1_error))
  __pyx_v_x30 = (3.0 * __pyx_v_x14);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":241
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30             # <<<<<<<<<<<<<<
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4
 */
  __Pyx_TraceLine(241,0,__PYX_ERR(1, 241, __pyx_L1_error))
  __pyx_v_x31 = (__pyx_v_x25 * __pyx_v_x30);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":242
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5
 */
  __Pyx_TraceLine(242,0,__PYX_ERR(1, 242, __pyx_L1_error))
  __pyx_v_x32 = (((0.5859375 * __pyx_v_nu) + (1.34765625 * __pyx_v_x28)) + 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":243
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4             # <<<<<<<<<<<<<<
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)
 */
  __Pyx_TraceLine(243,0,__PYX_ERR(1, 243, __pyx_L1_error))
  __pyx_v_x33 = pow(__pyx_v_pphi, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":244
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5             # <<<<<<<<<<<<<<
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35
 */
  __Pyx_TraceLine(244,0,__PYX_ERR(1, 244, __pyx_L1_error))
  __pyx_v_x34 = pow(__pyx_v_r, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":245
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36
 */
  __Pyx_TraceLine(245,0,__PYX_ERR(1, 245, __pyx_L1_error))
  __pyx_v_x35 = pow(__pyx_v_x34, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":246
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35             # <<<<<<<<<<<<<<
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375
 */
  __Pyx_TraceLine(246,0,__PYX_ERR(1, 246, __pyx_L1_error))
  __pyx_v_x36 = (4.0 * __pyx_v_x35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":247
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36             # <<<<<<<<<<<<<<
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu
 */
  __Pyx_TraceLine(247,0,__PYX_ERR(1, 247, __pyx_L1_error))
  __pyx_v_x37 = (__pyx_v_x33 * __pyx_v_x36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":248
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375             # <<<<<<<<<<<<<<
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 */
  __Pyx_TraceLine(248,0,__PYX_ERR(1, 248, __pyx_L1_error))
  __pyx_v_x38 = ((0.34375 * __pyx_v_nu) + 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":249
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu             # <<<<<<<<<<<<<<
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 */
  __Pyx_TraceLine(249,0,__PYX_ERR(1, 249, __pyx_L1_error))
  __pyx_v_x39 = (0.46875 - (0.28125 * __pyx_v_nu));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":250
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17
 */
  __Pyx_TraceLine(250,0,__PYX_ERR(1, 250, __pyx_L1_error))
  __pyx_v_x40 = (((-0.2734375 * __pyx_v_nu) - (0.798177083333333 * __pyx_v_x28)) - 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":251
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42
 */
  __Pyx_TraceLine(251,0,__PYX_ERR(1, 251, __pyx_L1_error))
  __pyx_v_x41 = (((-0.3515625 * __pyx_v_nu) + (0.29296875 * __pyx_v_x28)) - 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":252
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17             # <<<<<<<<<<<<<<
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43
 */
  __Pyx_TraceLine(252,0,__PYX_ERR(1, 252, __pyx_L1_error))
  __pyx_v_x42 = (__pyx_v_x16 - __pyx_v_x17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":253
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42             # <<<<<<<<<<<<<<
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1
 */
  __Pyx_TraceLine(253,0,__PYX_ERR(1, 253, __pyx_L1_error))
  __pyx_v_x43 = (__pyx_v_pphi * __pyx_v_x42);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":254
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43             # <<<<<<<<<<<<<<
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2
 */
  __Pyx_TraceLine(254,0,__PYX_ERR(1, 254, __pyx_L1_error))
  __pyx_v_x44 = (__pyx_v_delta * __pyx_v_x43);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":255
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1             # <<<<<<<<<<<<<<
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46
 */
  __Pyx_TraceLine(255,0,__PYX_ERR(1, 255, __pyx_L1_error))
  __pyx_v_x45 = (__pyx_v_chi_1 * __pyx_v_X_1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":256
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2             # <<<<<<<<<<<<<<
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2
 */
  __Pyx_TraceLine(256,0,__PYX_ERR(1, 256, __pyx_L1_error))
  __pyx_v_x46 = (__pyx_v_chi_2 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":257
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46             # <<<<<<<<<<<<<<
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9
 */
  __Pyx_TraceLine(257,0,__PYX_ERR(1, 257, __pyx_L1_error))
  __pyx_v_x47 = (__pyx_v_x45 + __pyx_v_x46);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":258
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2             # <<<<<<<<<<<<<<
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25
 */
  __Pyx_TraceLine(258,0,__PYX_ERR(1, 258, __pyx_L1_error))
  __pyx_v_x48 = pow(__pyx_v_x47, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":259
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9             # <<<<<<<<<<<<<<
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9
 */
  __Pyx_TraceLine(259,0,__PYX_ERR(1, 259, __pyx_L1_error))
  __pyx_v_x49 = ((-__pyx_v_x48) + __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":260
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25             # <<<<<<<<<<<<<<
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 */
  __Pyx_TraceLine(260,0,__PYX_ERR(1, 260, __pyx_L1_error))
  __pyx_v_x50 = (__pyx_v_delta * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":261
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9             # <<<<<<<<<<<<<<
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46
 */
  __Pyx_TraceLine(261,0,__PYX_ERR(1, 261, __pyx_L1_error))
  __pyx_v_x51 = (__pyx_v_delta * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":262
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)             # <<<<<<<<<<<<<<
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 */
  __Pyx_TraceLine(262,0,__PYX_ERR(1, 262, __pyx_L1_error))
  __pyx_v_x52 = (__pyx_v_x49 * (0.416666666666667 - (0.833333333333333 * __pyx_v_X_2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":263
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46             # <<<<<<<<<<<<<<
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 */
  __Pyx_TraceLine(263,0,__PYX_ERR(1, 263, __pyx_L1_error))
  __pyx_v_x53 = (__pyx_v_x45 - __pyx_v_x46);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":264
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7             # <<<<<<<<<<<<<<
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7
 */
  __Pyx_TraceLine(264,0,__PYX_ERR(1, 264, __pyx_L1_error))
  __pyx_v_x54 = (((__pyx_v_x4 * __pyx_v_x5) - (__pyx_v_x47 * __pyx_v_x53)) - __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":265
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7             # <<<<<<<<<<<<<<
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56
 */
  __Pyx_TraceLine(265,0,__PYX_ERR(1, 265, __pyx_L1_error))
  __pyx_v_x55 = (((((-0.125 * __pyx_v_x3) + (0.125 * __pyx_v_x48)) + ((0.5 * __pyx_v_x54) * (0.5 - __pyx_v_X_2))) - (0.125 * __pyx_v_x6)) - (0.125 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":266
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7             # <<<<<<<<<<<<<<
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54
 */
  __Pyx_TraceLine(266,0,__PYX_ERR(1, 266, __pyx_L1_error))
  __pyx_v_x56 = (__pyx_v_x6 - __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":267
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56             # <<<<<<<<<<<<<<
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24
 */
  __Pyx_TraceLine(267,0,__PYX_ERR(1, 267, __pyx_L1_error))
  __pyx_v_x57 = (__pyx_v_delta * __pyx_v_x56);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":268
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54             # <<<<<<<<<<<<<<
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)
 */
  __Pyx_TraceLine(268,0,__PYX_ERR(1, 268, __pyx_L1_error))
  __pyx_v_x58 = (__pyx_v_delta * __pyx_v_x54);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":269
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24             # <<<<<<<<<<<<<<
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25
 */
  __Pyx_TraceLine(269,0,__PYX_ERR(1, 269, __pyx_L1_error))
  __pyx_v_x59 = ((__pyx_v_nu * __pyx_v_dSO) * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":270
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33
 */
  __Pyx_TraceLine(270,0,__PYX_ERR(1, 270, __pyx_L1_error))
  __pyx_v_x60 = pow(__pyx_v_r, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":271
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25             # <<<<<<<<<<<<<<
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 */
  __Pyx_TraceLine(271,0,__PYX_ERR(1, 271, __pyx_L1_error))
  __pyx_v_x61 = (__pyx_v_x21 * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":272
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33             # <<<<<<<<<<<<<<
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 */
  __Pyx_TraceLine(272,0,__PYX_ERR(1, 272, __pyx_L1_error))
  __pyx_v_x62 = (__pyx_v_x14 * __pyx_v_x33);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":273
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)             # <<<<<<<<<<<<<<
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49
 */
  __Pyx_TraceLine(273,0,__PYX_ERR(1, 273, __pyx_L1_error))
  __pyx_v_x63 = (__pyx_v_x18 * ((((((__pyx_v_x21 * (((-5.53125 * __pyx_v_nu) + (0.567708333333333 * __pyx_v_x28)) - 0.078125)) + (__pyx_v_x22 * __pyx_v_x60)) + (__pyx_v_x23 * __pyx_v_x61)) + (__pyx_v_x26 * __pyx_v_x29)) + (__pyx_v_x32 * __pyx_v_x62)) + 1.75));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":274
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)             # <<<<<<<<<<<<<<
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65
 */
  __Pyx_TraceLine(274,0,__PYX_ERR(1, 274, __pyx_L1_error))
  __pyx_v_x64 = (__pyx_v_delta * ((((((__pyx_v_x21 * (((-0.03125 * __pyx_v_nu) + (0.536458333333333 * __pyx_v_x28)) + 0.078125)) + (__pyx_v_x26 * __pyx_v_x40)) + (__pyx_v_x38 * __pyx_v_x60)) + (__pyx_v_x39 * __pyx_v_x61)) + (__pyx_v_x41 * __pyx_v_x62)) + 0.25));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":275
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49             # <<<<<<<<<<<<<<
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 */
  __Pyx_TraceLine(275,0,__PYX_ERR(1, 275, __pyx_L1_error))
  __pyx_v_x65 = (__pyx_v_x24 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":276
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65             # <<<<<<<<<<<<<<
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55
 */
  __Pyx_TraceLine(276,0,__PYX_ERR(1, 276, __pyx_L1_error))
  __pyx_v_x66 = (__pyx_v_x50 * __pyx_v_x65);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":277
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)             # <<<<<<<<<<<<<<
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 */
  __Pyx_TraceLine(277,0,__PYX_ERR(1, 277, __pyx_L1_error))
  __pyx_v_x67 = (__pyx_v_x42 * ((__pyx_v_x21 * ((0.0416666666666667 * __pyx_v_x51) + (0.5 * __pyx_v_x52))) - (0.125 * __pyx_v_x66)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":278
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55             # <<<<<<<<<<<<<<
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4
 */
  __Pyx_TraceLine(278,0,__PYX_ERR(1, 278, __pyx_L1_error))
  __pyx_v_x68 = (__pyx_v_x26 * __pyx_v_x55);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":279
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)             # <<<<<<<<<<<<<<
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)
 */
  __Pyx_TraceLine(279,0,__PYX_ERR(1, 279, __pyx_L1_error))
  __pyx_v_x69 = (__pyx_v_x18 * ((__pyx_v_x21 * (((((((0.208333333333333 * __pyx_v_delta) * __pyx_v_x56) - (0.625 * __pyx_v_x3)) + (0.375 * __pyx_v_x48)) - (0.833333333333333 * __pyx_v_x58)) - (0.625 * __pyx_v_x6)) - (0.625 * __pyx_v_x7))) + __pyx_v_x68));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":280
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4             # <<<<<<<<<<<<<<
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71
 */
  __Pyx_TraceLine(280,0,__PYX_ERR(1, 280, __pyx_L1_error))
  __pyx_v_x70 = pow(__pyx_v_prst, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":281
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)             # <<<<<<<<<<<<<<
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6
 */
  __Pyx_TraceLine(281,0,__PYX_ERR(1, 281, __pyx_L1_error))
  __pyx_v_x71 = pow(__pyx_v_r, -4.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":282
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71             # <<<<<<<<<<<<<<
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)
 */
  __Pyx_TraceLine(282,0,__PYX_ERR(1, 282, __pyx_L1_error))
  __pyx_v_x72 = (__pyx_v_nu * __pyx_v_x71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":283
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6             # <<<<<<<<<<<<<<
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74
 */
  __Pyx_TraceLine(283,0,__PYX_ERR(1, 283, __pyx_L1_error))
  __pyx_v_x73 = pow(__pyx_v_prst, 6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":284
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)             # <<<<<<<<<<<<<<
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)
 */
  __Pyx_TraceLine(284,0,__PYX_ERR(1, 284, __pyx_L1_error))
  __pyx_v_x74 = pow(__pyx_v_r, -3.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":285
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74             # <<<<<<<<<<<<<<
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8
 */
  __Pyx_TraceLine(285,0,__PYX_ERR(1, 285, __pyx_L1_error))
  __pyx_v_x75 = (__pyx_v_nu * __pyx_v_x74);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":286
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)             # <<<<<<<<<<<<<<
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77
 */
  __Pyx_TraceLine(286,0,__PYX_ERR(1, 286, __pyx_L1_error))
  __pyx_v_x76 = pow(__pyx_v_r, -2.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":287
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8             # <<<<<<<<<<<<<<
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60
 */
  __Pyx_TraceLine(287,0,__PYX_ERR(1, 287, __pyx_L1_error))
  __pyx_v_x77 = pow(__pyx_v_prst, 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":288
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77             # <<<<<<<<<<<<<<
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77
 */
  __Pyx_TraceLine(288,0,__PYX_ERR(1, 288, __pyx_L1_error))
  __pyx_v_x78 = (__pyx_v_nu * __pyx_v_x77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":289
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60             # <<<<<<<<<<<<<<
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28
 */
  __Pyx_TraceLine(289,0,__PYX_ERR(1, 289, __pyx_L1_error))
  __pyx_v_x79 = (__pyx_v_nu * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":290
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77             # <<<<<<<<<<<<<<
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81
 */
  __Pyx_TraceLine(290,0,__PYX_ERR(1, 290, __pyx_L1_error))
  __pyx_v_x80 = (0.121954868780449 * __pyx_v_x77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":291
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3
 */
  __Pyx_TraceLine(291,0,__PYX_ERR(1, 291, __pyx_L1_error))
  __pyx_v_x81 = ((8.0 * __pyx_v_nu) - (6.0 * __pyx_v_x28));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":292
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81             # <<<<<<<<<<<<<<
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 */
  __Pyx_TraceLine(292,0,__PYX_ERR(1, 292, __pyx_L1_error))
  __pyx_v_x82 = (__pyx_v_x21 * __pyx_v_x81);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":293
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3             # <<<<<<<<<<<<<<
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70
 */
  __Pyx_TraceLine(293,0,__PYX_ERR(1, 293, __pyx_L1_error))
  __pyx_v_x83 = pow(__pyx_v_nu, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":294
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83             # <<<<<<<<<<<<<<
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 */
  __Pyx_TraceLine(294,0,__PYX_ERR(1, 294, __pyx_L1_error))
  __pyx_v_x84 = (((92.7110442849544 * __pyx_v_nu) - (131.0 * __pyx_v_x28)) + (10.0 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":295
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70             # <<<<<<<<<<<<<<
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86
 */
  __Pyx_TraceLine(295,0,__PYX_ERR(1, 295, __pyx_L1_error))
  __pyx_v_x85 = (__pyx_v_x24 * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":296
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83             # <<<<<<<<<<<<<<
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4
 */
  __Pyx_TraceLine(296,0,__PYX_ERR(1, 296, __pyx_L1_error))
  __pyx_v_x86 = (((-2.78300763695006 * __pyx_v_nu) - (5.4 * __pyx_v_x28)) + (6.0 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":297
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86             # <<<<<<<<<<<<<<
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 */
  __Pyx_TraceLine(297,0,__PYX_ERR(1, 297, __pyx_L1_error))
  __pyx_v_x87 = (__pyx_v_x21 * __pyx_v_x86);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":298
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4             # <<<<<<<<<<<<<<
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73
 */
  __Pyx_TraceLine(298,0,__PYX_ERR(1, 298, __pyx_L1_error))
  __pyx_v_x88 = pow(__pyx_v_nu, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":299
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88             # <<<<<<<<<<<<<<
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 */
  __Pyx_TraceLine(299,0,__PYX_ERR(1, 299, __pyx_L1_error))
  __pyx_v_x89 = ((((-33.9782122170436 * __pyx_v_nu) - (89.5298327361234 * __pyx_v_x28)) + (188.0 * __pyx_v_x83)) - (14.0 * __pyx_v_x88));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":300
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73             # <<<<<<<<<<<<<<
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91
 */
  __Pyx_TraceLine(300,0,__PYX_ERR(1, 300, __pyx_L1_error))
  __pyx_v_x90 = (__pyx_v_x24 * __pyx_v_x73);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":301
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88             # <<<<<<<<<<<<<<
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)
 */
  __Pyx_TraceLine(301,0,__PYX_ERR(1, 301, __pyx_L1_error))
  __pyx_v_x91 = ((((1.38977750996128 * __pyx_v_nu) + (3.33842023648322 * __pyx_v_x28)) + (3.42857142857143 * __pyx_v_x83)) - (6.0 * __pyx_v_x88));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":302
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91             # <<<<<<<<<<<<<<
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 */
  __Pyx_TraceLine(302,0,__PYX_ERR(1, 302, __pyx_L1_error))
  __pyx_v_x92 = (__pyx_v_x77 * __pyx_v_x91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":303
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)             # <<<<<<<<<<<<<<
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70
 */
  __Pyx_TraceLine(303,0,__PYX_ERR(1, 303, __pyx_L1_error))
  __pyx_v_x93 = log(__pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":304
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83             # <<<<<<<<<<<<<<
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13
 */
  __Pyx_TraceLine(304,0,__PYX_ERR(1, 304, __pyx_L1_error))
  __pyx_v_x94 = (((__pyx_v_nu * (452.542166996693 - (51.6952380952381 * __pyx_v_x93))) + (__pyx_v_x28 * ((118.4 * __pyx_v_x93) - 1796.13660498019))) + (602.318540416564 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":305
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70             # <<<<<<<<<<<<<<
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)
 */
  __Pyx_TraceLine(305,0,__PYX_ERR(1, 305, __pyx_L1_error))
  __pyx_v_x95 = (__pyx_v_x14 * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":306
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13             # <<<<<<<<<<<<<<
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2
 */
  __Pyx_TraceLine(306,0,__PYX_ERR(1, 306, __pyx_L1_error))
  __pyx_v_x96 = ((__pyx_v_r * __pyx_v_x10) + __pyx_v_x13);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":307
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98
 */
  __Pyx_TraceLine(307,0,__PYX_ERR(1, 307, __pyx_L1_error))
  __pyx_v_x97 = pow(__pyx_v_x96, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":308
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2             # <<<<<<<<<<<<<<
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99
 */
  __Pyx_TraceLine(308,0,__PYX_ERR(1, 308, __pyx_L1_error))
  __pyx_v_x98 = pow(__pyx_v_x18, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":309
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98             # <<<<<<<<<<<<<<
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49
 */
  __Pyx_TraceLine(309,0,__PYX_ERR(1, 309, __pyx_L1_error))
  __pyx_v_x99 = (__pyx_v_x97 * __pyx_v_x98);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":310
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99             # <<<<<<<<<<<<<<
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54
 */
  __Pyx_TraceLine(310,0,__PYX_ERR(1, 310, __pyx_L1_error))
  __pyx_v_x100 = ((__pyx_v_x25 * __pyx_v_x60) * __pyx_v_x99);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":311
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49             # <<<<<<<<<<<<<<
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8
 */
  __Pyx_TraceLine(311,0,__PYX_ERR(1, 311, __pyx_L1_error))
  __pyx_v_x101 = (0.5 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":312
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54             # <<<<<<<<<<<<<<
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2
 */
  __Pyx_TraceLine(312,0,__PYX_ERR(1, 312, __pyx_L1_error))
  __pyx_v_x102 = (0.5 * __pyx_v_x54);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":313
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8             # <<<<<<<<<<<<<<
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104
 */
  __Pyx_TraceLine(313,0,__PYX_ERR(1, 313, __pyx_L1_error))
  __pyx_v_x103 = ((-__pyx_v_x3) + __pyx_v_x8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":314
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2             # <<<<<<<<<<<<<<
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 */
  __Pyx_TraceLine(314,0,__PYX_ERR(1, 314, __pyx_L1_error))
  __pyx_v_x104 = (__pyx_v_x103 - pow(__pyx_v_x53, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":315
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104             # <<<<<<<<<<<<<<
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28
 */
  __Pyx_TraceLine(315,0,__PYX_ERR(1, 315, __pyx_L1_error))
  __pyx_v_x105 = (0.5 * __pyx_v_x104);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":316
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)             # <<<<<<<<<<<<<<
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 */
  __Pyx_TraceLine(316,0,__PYX_ERR(1, 316, __pyx_L1_error))
  __pyx_v_x106 = (__pyx_v_x58 * ((98.0 * __pyx_v_nu) + 43.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":317
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49
 */
  __Pyx_TraceLine(317,0,__PYX_ERR(1, 317, __pyx_L1_error))
  __pyx_v_x107 = (264.0 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":318
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0             # <<<<<<<<<<<<<<
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 */
  __Pyx_TraceLine(318,0,__PYX_ERR(1, 318, __pyx_L1_error))
  __pyx_v_x108 = (((-1610.0 * __pyx_v_nu) + __pyx_v_x107) + 375.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":319
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49             # <<<<<<<<<<<<<<
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104
 */
  __Pyx_TraceLine(319,0,__PYX_ERR(1, 319, __pyx_L1_error))
  __pyx_v_x109 = (0.00260416666666667 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":320
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0             # <<<<<<<<<<<<<<
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 */
  __Pyx_TraceLine(320,0,__PYX_ERR(1, 320, __pyx_L1_error))
  __pyx_v_x110 = (((-1090.0 * __pyx_v_nu) + (152.0 * __pyx_v_x28)) + 219.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":321
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104             # <<<<<<<<<<<<<<
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)
 */
  __Pyx_TraceLine(321,0,__PYX_ERR(1, 321, __pyx_L1_error))
  __pyx_v_x111 = (0.0078125 * __pyx_v_x104);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":322
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9
 */
  __Pyx_TraceLine(322,0,__PYX_ERR(1, 322, __pyx_L1_error))
  __pyx_v_x112 = (((((-__pyx_v_x101) * __pyx_v_x21) + (__pyx_v_x14 * (((-0.0625 * __pyx_v_x106) + (__pyx_v_x108 * __pyx_v_x109)) + (__pyx_v_x110 * __pyx_v_x111)))) + (__pyx_v_x24 * (((__pyx_v_x101 * ((-1.75 * __pyx_v_nu) - 0.9375)) + (__pyx_v_x102 * (0.75 - (1.5 * __pyx_v_X_2)))) + (__pyx_v_x105 * (0.1875 - (0.75 * __pyx_v_nu)))))) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":323
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)             # <<<<<<<<<<<<<<
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4
 */
  __Pyx_TraceLine(323,0,__PYX_ERR(1, 323, __pyx_L1_error))
  __pyx_v_x113 = pow(__pyx_v_r, -13.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":324
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9             # <<<<<<<<<<<<<<
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9
 */
  __Pyx_TraceLine(324,0,__PYX_ERR(1, 324, __pyx_L1_error))
  __pyx_v_x114 = (__pyx_v_x20 + __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":325
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4             # <<<<<<<<<<<<<<
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu
 */
  __Pyx_TraceLine(325,0,__PYX_ERR(1, 325, __pyx_L1_error))
  __pyx_v_x115 = pow(__pyx_v_x114, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":326
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9             # <<<<<<<<<<<<<<
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0
 */
  __Pyx_TraceLine(326,0,__PYX_ERR(1, 326, __pyx_L1_error))
  __pyx_v_x116 = (__pyx_v_x21 * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":327
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 */
  __Pyx_TraceLine(327,0,__PYX_ERR(1, 327, __pyx_L1_error))
  __pyx_v_x117 = (756.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":328
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0             # <<<<<<<<<<<<<<
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2
 */
  __Pyx_TraceLine(328,0,__PYX_ERR(1, 328, __pyx_L1_error))
  __pyx_v_x118 = (((336.0 * __pyx_v_r) + __pyx_v_x117) + 407.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":329
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0             # <<<<<<<<<<<<<<
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28
 */
  __Pyx_TraceLine(329,0,__PYX_ERR(1, 329, __pyx_L1_error))
  __pyx_v_x119 = (((((((2048.0 * __pyx_v_nu) * __pyx_v_x118) * __pyx_v_x93) + ((28.0 * __pyx_v_nu) * ((1920.0 * __pyx_v_a6) + 733955.307463037))) - ((7.0 * __pyx_v_r) * (((938918.400156317 * __pyx_v_nu) - (185763.092693281 * __pyx_v_x28)) - 245760.0))) - (5416406.59541186 * __pyx_v_x28)) - 3440640.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":330
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2             # <<<<<<<<<<<<<<
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 */
  __Pyx_TraceLine(330,0,__PYX_ERR(1, 330, __pyx_L1_error))
  __pyx_v_x120 = pow(__pyx_v_x93, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":331
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28             # <<<<<<<<<<<<<<
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 */
  __Pyx_TraceLine(331,0,__PYX_ERR(1, 331, __pyx_L1_error))
  __pyx_v_x121 = (__pyx_v_x120 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":332
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)             # <<<<<<<<<<<<<<
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0
 */
  __Pyx_TraceLine(332,0,__PYX_ERR(1, 332, __pyx_L1_error))
  __pyx_v_x122 = (__pyx_v_x83 * (((-163683964.822551 * __pyx_v_r) - (17833256.898555 * __pyx_v_x20)) - 1188987459.03162));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":333
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)             # <<<<<<<<<<<<<<
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0
 */
  __Pyx_TraceLine(333,0,__PYX_ERR(1, 333, __pyx_L1_error))
  __pyx_v_x123 = (__pyx_v_x28 * (((((((((-39321600.0 * __pyx_v_a6) * ((3.0 * __pyx_v_r) + 59.0)) + (745857848.115604 * __pyx_v_a6)) + (122635399361.987 * __pyx_v_r)) - (6178501407.53758 * __pyx_v_x0)) - (3089250703.76879 * __pyx_v_x13)) + (2064783811.32587 * __pyx_v_x20)) + (1426660551.8844 * __pyx_v_x34)) + 276057889687.011));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":334
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125
 */
  __Pyx_TraceLine(334,0,__PYX_ERR(1, 334, __pyx_L1_error))
  __pyx_v_x124 = ((588.0 * __pyx_v_nu) + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":335
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 */
  __Pyx_TraceLine(335,0,__PYX_ERR(1, 335, __pyx_L1_error))
  __pyx_v_x125 = (__pyx_v_x117 + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":336
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125             # <<<<<<<<<<<<<<
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93
 */
  __Pyx_TraceLine(336,0,__PYX_ERR(1, 336, __pyx_L1_error))
  __pyx_v_x126 = (__pyx_v_x0 * __pyx_v_x125);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":337
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0             # <<<<<<<<<<<<<<
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128
 */
  __Pyx_TraceLine(337,0,__PYX_ERR(1, 337, __pyx_L1_error))
  __pyx_v_x127 = ((((((((-38842241.4769507 * __pyx_v_nu) + ((240.0 * __pyx_v_r) * (((-7466.27061066206 * __pyx_v_nu) - (3024.0 * __pyx_v_x28)) + 17264.0))) + ((1920.0 * __pyx_v_x124) * __pyx_v_x20)) + ((480.0 * __pyx_v_x125) * __pyx_v_x13)) + (960.0 * __pyx_v_x126)) - (1882456.23663972 * __pyx_v_x28)) + (161280.0 * __pyx_v_x34)) + 13447680.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":338
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93             # <<<<<<<<<<<<<<
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r
 */
  __Pyx_TraceLine(338,0,__PYX_ERR(1, 338, __pyx_L1_error))
  __pyx_v_x128 = (__pyx_v_nu * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":339
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128             # <<<<<<<<<<<<<<
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 */
  __Pyx_TraceLine(339,0,__PYX_ERR(1, 339, __pyx_L1_error))
  __pyx_v_x129 = (__pyx_v_x127 * __pyx_v_x128);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":340
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r             # <<<<<<<<<<<<<<
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6
 */
  __Pyx_TraceLine(340,0,__PYX_ERR(1, 340, __pyx_L1_error))
  __pyx_v_x130 = (8.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":341
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0             # <<<<<<<<<<<<<<
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r
 */
  __Pyx_TraceLine(341,0,__PYX_ERR(1, 341, __pyx_L1_error))
  __pyx_v_x131 = ((((2.0 * __pyx_v_x0) + __pyx_v_x130) + (4.0 * __pyx_v_x20)) + 16.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":342
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6             # <<<<<<<<<<<<<<
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r
 */
  __Pyx_TraceLine(342,0,__PYX_ERR(1, 342, __pyx_L1_error))
  __pyx_v_x132 = (7680.0 * __pyx_v_a6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":343
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r             # <<<<<<<<<<<<<<
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r
 */
  __Pyx_TraceLine(343,0,__PYX_ERR(1, 343, __pyx_L1_error))
  __pyx_v_x133 = (128.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":344
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r             # <<<<<<<<<<<<<<
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r
 */
  __Pyx_TraceLine(344,0,__PYX_ERR(1, 344, __pyx_L1_error))
  __pyx_v_x134 = (7704.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":345
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r             # <<<<<<<<<<<<<<
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 */
  __Pyx_TraceLine(345,0,__PYX_ERR(1, 345, __pyx_L1_error))
  __pyx_v_x135 = (148.04406601634 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":346
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r             # <<<<<<<<<<<<<<
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 */
  __Pyx_TraceLine(346,0,__PYX_ERR(1, 346, __pyx_L1_error))
  __pyx_v_x136 = (113485.217444961 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":347
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))             # <<<<<<<<<<<<<<
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138
 */
  __Pyx_TraceLine(347,0,__PYX_ERR(1, 347, __pyx_L1_error))
  __pyx_v_x137 = (__pyx_v_nu * ((((__pyx_v_x132 * (__pyx_v_x13 + __pyx_v_x131)) + (__pyx_v_x133 * (((((13218.7851094412 * __pyx_v_r) + (4264.6962773603 * __pyx_v_x0)) - (6852.34813868015 * __pyx_v_x13)) + (8529.39255472061 * __pyx_v_x20)) - 33722.4297811176))) + (__pyx_v_x135 * (((((1926.0 * __pyx_v_x0) + (349.0 * __pyx_v_x13)) + __pyx_v_x134) + (3852.0 * __pyx_v_x20)) + 36400.0))) + (__pyx_v_x136 * ((-__pyx_v_x13) + __pyx_v_x131))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":348
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139
 */
  __Pyx_TraceLine(348,0,__PYX_ERR(1, 348, __pyx_L1_error))
  __pyx_v_x138 = pow((((((((67645734912.0 * __pyx_v_x121) + (1120.0 * __pyx_v_x122)) + (7.0 * __pyx_v_x123)) + (32768.0 * __pyx_v_x129)) + (53760.0 * __pyx_v_x137)) + (13212057600.0 * __pyx_v_x34)) + (241555486248.807 * __pyx_v_x88)), -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":349
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138             # <<<<<<<<<<<<<<
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 */
  __Pyx_TraceLine(349,0,__PYX_ERR(1, 349, __pyx_L1_error))
  __pyx_v_x139 = (__pyx_v_x13 * __pyx_v_x138);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":350
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139             # <<<<<<<<<<<<<<
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)
 */
  __Pyx_TraceLine(350,0,__PYX_ERR(1, 350, __pyx_L1_error))
  __pyx_v_x140 = (__pyx_v_x119 * __pyx_v_x139);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":351
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140             # <<<<<<<<<<<<<<
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120
 */
  __Pyx_TraceLine(351,0,__PYX_ERR(1, 351, __pyx_L1_error))
  __pyx_v_x141 = ((0.000130208333333333 * __pyx_v_x116) + __pyx_v_x140);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":352
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)             # <<<<<<<<<<<<<<
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 */
  __Pyx_TraceLine(352,0,__PYX_ERR(1, 352, __pyx_L1_error))
  __pyx_v_x142 = pow(__pyx_v_x141, -4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":353
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120             # <<<<<<<<<<<<<<
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20
 */
  __Pyx_TraceLine(353,0,__PYX_ERR(1, 353, __pyx_L1_error))
  __pyx_v_x143 = (__pyx_v_r * __pyx_v_x120);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":354
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913             # <<<<<<<<<<<<<<
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 */
  __Pyx_TraceLine(354,0,__PYX_ERR(1, 354, __pyx_L1_error))
  __pyx_v_x144 = (((-630116198.873299 * __pyx_v_nu) - (197773496.793534 * __pyx_v_x28)) + 5805304367.87913);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":355
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20             # <<<<<<<<<<<<<<
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146
 */
  __Pyx_TraceLine(355,0,__PYX_ERR(1, 355, __pyx_L1_error))
  __pyx_v_x145 = (__pyx_v_x144 * __pyx_v_x20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":356
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329             # <<<<<<<<<<<<<<
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 */
  __Pyx_TraceLine(356,0,__PYX_ERR(1, 356, __pyx_L1_error))
  __pyx_v_x146 = (((-2675575.66847905 * __pyx_v_nu) - (138240.0 * __pyx_v_x28)) - 5278341.3229329);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":357
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146             # <<<<<<<<<<<<<<
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 */
  __Pyx_TraceLine(357,0,__PYX_ERR(1, 357, __pyx_L1_error))
  __pyx_v_x147 = (__pyx_v_x0 * __pyx_v_x146);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":358
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)             # <<<<<<<<<<<<<<
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149
 */
  __Pyx_TraceLine(358,0,__PYX_ERR(1, 358, __pyx_L1_error))
  __pyx_v_x148 = (__pyx_v_nu * ((((-2510664218.28128 * __pyx_v_nu) - (42636451.6032331 * __pyx_v_x28)) + (14515200.0 * __pyx_v_x83)) + 1002013764.01019));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":359
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0             # <<<<<<<<<<<<<<
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0
 */
  __Pyx_TraceLine(359,0,__PYX_ERR(1, 359, __pyx_L1_error))
  __pyx_v_x149 = (((((43393301259014.8 * __pyx_v_nu) + (43133561885859.3 * __pyx_v_x28)) + (5927865218923.02 * __pyx_v_x83)) + (86618264430493.3 * pow((1.0 - (0.496948781616935 * __pyx_v_nu)), 2.0))) + 188440788778196.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":360
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149             # <<<<<<<<<<<<<<
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151
 */
  __Pyx_TraceLine(360,0,__PYX_ERR(1, 360, __pyx_L1_error))
  __pyx_v_x150 = (__pyx_v_r * __pyx_v_x149);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":361
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0             # <<<<<<<<<<<<<<
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 */
  __Pyx_TraceLine(361,0,__PYX_ERR(1, 361, __pyx_L1_error))
  __pyx_v_x151 = ((14700.0 * __pyx_v_nu) + 42911.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":362
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151             # <<<<<<<<<<<<<<
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93
 */
  __Pyx_TraceLine(362,0,__PYX_ERR(1, 362, __pyx_L1_error))
  __pyx_v_x152 = (283115520.0 * __pyx_v_x151);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":363
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20             # <<<<<<<<<<<<<<
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 */
  __Pyx_TraceLine(363,0,__PYX_ERR(1, 363, __pyx_L1_error))
  __pyx_v_x153 = (((((-1698693120.0 * __pyx_v_nu) * ((11592.0 * __pyx_v_nu) + 69847.0)) + ((49152.0 * __pyx_v_r) * (((409207698.136075 * __pyx_v_nu) + (102574080.0 * __pyx_v_x28)) - 2119671837.36038))) + (879923036160.0 * __pyx_v_x0)) + (__pyx_v_x152 * __pyx_v_x20));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":364
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93             # <<<<<<<<<<<<<<
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r
 */
  __Pyx_TraceLine(364,0,__PYX_ERR(1, 364, __pyx_L1_error))
  __pyx_v_x154 = (__pyx_v_x153 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":365
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2             # <<<<<<<<<<<<<<
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20
 */
  __Pyx_TraceLine(365,0,__PYX_ERR(1, 365, __pyx_L1_error))
  __pyx_v_x155 = pow((((((__pyx_v_x143 - (1.59227685093395e-9 * __pyx_v_x145)) - (1.67189069348064e-7 * __pyx_v_x147)) + (9.55366110560367e-9 * __pyx_v_x148)) + (1.72773095804465e-13 * __pyx_v_x150)) + (1.72773095804465e-13 * __pyx_v_x154)), 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":366
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r             # <<<<<<<<<<<<<<
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28
 */
  __Pyx_TraceLine(366,0,__PYX_ERR(1, 366, __pyx_L1_error))
  __pyx_v_x156 = (__pyx_v_nu * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":367
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20             # <<<<<<<<<<<<<<
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151
 */
  __Pyx_TraceLine(367,0,__PYX_ERR(1, 367, __pyx_L1_error))
  __pyx_v_x157 = (__pyx_v_nu * __pyx_v_x20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":368
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28             # <<<<<<<<<<<<<<
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28
 */
  __Pyx_TraceLine(368,0,__PYX_ERR(1, 368, __pyx_L1_error))
  __pyx_v_x158 = (__pyx_v_r * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":369
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151             # <<<<<<<<<<<<<<
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 */
  __Pyx_TraceLine(369,0,__PYX_ERR(1, 369, __pyx_L1_error))
  __pyx_v_x159 = (__pyx_v_r * __pyx_v_x151);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":370
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28             # <<<<<<<<<<<<<<
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 */
  __Pyx_TraceLine(370,0,__PYX_ERR(1, 370, __pyx_L1_error))
  __pyx_v_x160 = (__pyx_v_x20 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":371
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0             # <<<<<<<<<<<<<<
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93
 */
  __Pyx_TraceLine(371,0,__PYX_ERR(1, 371, __pyx_L1_error))
  __pyx_v_x161 = ((5041721180160.0 * __pyx_v_x28) - 104186110149937.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":372
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20             # <<<<<<<<<<<<<<
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 */
  __Pyx_TraceLine(372,0,__PYX_ERR(1, 372, __pyx_L1_error))
  __pyx_v_x162 = ((((-25392914995744.3 * __pyx_v_nu) - (__pyx_v_r * __pyx_v_x152)) - __pyx_v_x161) - (879923036160.0 * __pyx_v_x20));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":373
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93             # <<<<<<<<<<<<<<
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)
 */
  __Pyx_TraceLine(373,0,__PYX_ERR(1, 373, __pyx_L1_error))
  __pyx_v_x163 = (__pyx_v_x162 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":374
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1             # <<<<<<<<<<<<<<
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28
 */
  __Pyx_TraceLine(374,0,__PYX_ERR(1, 374, __pyx_L1_error))
  __pyx_v_x164 = (((((((((((((-0.0438084424460039 * __pyx_v_nu) - (0.143521050466841 * __pyx_v_r)) + (0.0210425293255724 * __pyx_v_x120)) + (0.0385795738434214 * __pyx_v_x156)) + (0.00941289164152486 * __pyx_v_x157)) + (0.00662650629087394 * __pyx_v_x158)) - (1.18824456940711e-6 * __pyx_v_x159)) + (0.000486339502879429 * __pyx_v_x160)) - (3.63558293513537e-15 * __pyx_v_x163)) + (0.0185696317637669 * __pyx_v_x20)) + (0.291062041428379 * __pyx_v_x28)) + (0.0244692826489756 * __pyx_v_x83)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":375
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)             # <<<<<<<<<<<<<<
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 */
  __Pyx_TraceLine(375,0,__PYX_ERR(1, 375, __pyx_L1_error))
  __pyx_v_x165 = pow(__pyx_v_x164, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":376
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 */
  __Pyx_TraceLine(376,0,__PYX_ERR(1, 376, __pyx_L1_error))
  __pyx_v_x166 = (4.0 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":377
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)             # <<<<<<<<<<<<<<
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2
 */
  __Pyx_TraceLine(377,0,__PYX_ERR(1, 377, __pyx_L1_error))
  __pyx_v_x167 = (((((0.3125 * __pyx_v_delta) * __pyx_v_x56) * ((18.0 * __pyx_v_nu) - 1.0)) - ((0.03125 * (((-33.0 * __pyx_v_nu) + (32.0 * __pyx_v_x28)) - 5.0)) * (((5.0 * __pyx_v_x3) + (5.0 * __pyx_v_x6)) + (5.0 * __pyx_v_x7)))) - ((0.03125 * (((-5.0 * __pyx_v_nu) + __pyx_v_x166) + 1.0)) * (((-15.0 * __pyx_v_x3) + (15.0 * __pyx_v_x6)) + (15.0 * __pyx_v_x7))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":378
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70             # <<<<<<<<<<<<<<
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2
 */
  __Pyx_TraceLine(378,0,__PYX_ERR(1, 378, __pyx_L1_error))
  __pyx_v_x168 = (((((__pyx_v_x115 * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":379
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2             # <<<<<<<<<<<<<<
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)
 */
  __Pyx_TraceLine(379,0,__PYX_ERR(1, 379, __pyx_L1_error))
  __pyx_v_x169 = pow(__pyx_v_x114, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":380
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2             # <<<<<<<<<<<<<<
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28
 */
  __Pyx_TraceLine(380,0,__PYX_ERR(1, 380, __pyx_L1_error))
  __pyx_v_x170 = pow(__pyx_v_prst, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":381
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)             # <<<<<<<<<<<<<<
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120
 */
  __Pyx_TraceLine(381,0,__PYX_ERR(1, 381, __pyx_L1_error))
  __pyx_v_x171 = pow(__pyx_v_x141, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":382
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28             # <<<<<<<<<<<<<<
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 */
  __Pyx_TraceLine(382,0,__PYX_ERR(1, 382, __pyx_L1_error))
  __pyx_v_x172 = (1822680546449.21 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":383
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120             # <<<<<<<<<<<<<<
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)
 */
  __Pyx_TraceLine(383,0,__PYX_ERR(1, 383, __pyx_L1_error))
  __pyx_v_x173 = (5787938193408.0 * __pyx_v_x120);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":384
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0             # <<<<<<<<<<<<<<
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu
 */
  __Pyx_TraceLine(384,0,__PYX_ERR(1, 384, __pyx_L1_error))
  __pyx_v_x174 = (((((((((((((-12049908701745.2 * __pyx_v_nu) + (__pyx_v_r * __pyx_v_x172)) - (39476764256925.6 * __pyx_v_r)) + (10611661054566.2 * __pyx_v_x156)) + (2589101062873.81 * __pyx_v_x157)) - (326837426.241486 * __pyx_v_x159)) + (133772083200.0 * __pyx_v_x160)) - __pyx_v_x163) + __pyx_v_x173) + (5107745331375.71 * __pyx_v_x20)) + (80059249540278.2 * __pyx_v_x28)) + (6730497718123.02 * __pyx_v_x83)) + 275059053208689.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":385
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3
 */
  __Pyx_TraceLine(385,0,__PYX_ERR(1, 385, __pyx_L1_error))
  __pyx_v_x175 = pow(__pyx_v_x174, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":386
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 */
  __Pyx_TraceLine(386,0,__PYX_ERR(1, 386, __pyx_L1_error))
  __pyx_v_x176 = (4.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":387
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3             # <<<<<<<<<<<<<<
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178
 */
  __Pyx_TraceLine(387,0,__PYX_ERR(1, 387, __pyx_L1_error))
  __pyx_v_x177 = (3.0 * __pyx_v_x3);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":388
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7             # <<<<<<<<<<<<<<
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179
 */
  __Pyx_TraceLine(388,0,__PYX_ERR(1, 388, __pyx_L1_error))
  __pyx_v_x178 = ((3.0 * __pyx_v_x6) + (3.0 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":389
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178             # <<<<<<<<<<<<<<
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 */
  __Pyx_TraceLine(389,0,__PYX_ERR(1, 389, __pyx_L1_error))
  __pyx_v_x179 = ((-__pyx_v_x177) + __pyx_v_x178);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":390
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179             # <<<<<<<<<<<<<<
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 */
  __Pyx_TraceLine(390,0,__PYX_ERR(1, 390, __pyx_L1_error))
  __pyx_v_x180 = (0.0625 * __pyx_v_x179);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":391
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)             # <<<<<<<<<<<<<<
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 */
  __Pyx_TraceLine(391,0,__PYX_ERR(1, 391, __pyx_L1_error))
  __pyx_v_x181 = (__pyx_v_x57 * ((26.0 * __pyx_v_nu) + 449.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":392
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)             # <<<<<<<<<<<<<<
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 */
  __Pyx_TraceLine(392,0,__PYX_ERR(1, 392, __pyx_L1_error))
  __pyx_v_x182 = (__pyx_v_x9 * ((-1171.0 * __pyx_v_nu) - 861.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":393
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)             # <<<<<<<<<<<<<<
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)
 */
  __Pyx_TraceLine(393,0,__PYX_ERR(1, 393, __pyx_L1_error))
  __pyx_v_x183 = (__pyx_v_x103 * (((115.0 * __pyx_v_nu) + __pyx_v_x166) - 37.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":394
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154             # <<<<<<<<<<<<<<
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185
 */
  __Pyx_TraceLine(394,0,__PYX_ERR(1, 394, __pyx_L1_error))
  __pyx_v_x184 = ((((((5787938193408.0 * __pyx_v_x143) - (9216.0 * __pyx_v_x145)) - (967680.0 * __pyx_v_x147)) + (55296.0 * __pyx_v_x148)) + __pyx_v_x150) + __pyx_v_x154);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":395
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34
 */
  __Pyx_TraceLine(395,0,__PYX_ERR(1, 395, __pyx_L1_error))
  __pyx_v_x185 = pow(__pyx_v_x184, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":396
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185             # <<<<<<<<<<<<<<
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138
 */
  __Pyx_TraceLine(396,0,__PYX_ERR(1, 396, __pyx_L1_error))
  __pyx_v_x186 = (__pyx_v_x174 * __pyx_v_x185);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":397
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34             # <<<<<<<<<<<<<<
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188
 */
  __Pyx_TraceLine(397,0,__PYX_ERR(1, 397, __pyx_L1_error))
  __pyx_v_x187 = (__pyx_v_x186 * __pyx_v_x34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":398
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138             # <<<<<<<<<<<<<<
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 */
  __Pyx_TraceLine(398,0,__PYX_ERR(1, 398, __pyx_L1_error))
  __pyx_v_x188 = (__pyx_v_x119 * __pyx_v_x138);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":399
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188             # <<<<<<<<<<<<<<
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 */
  __Pyx_TraceLine(399,0,__PYX_ERR(1, 399, __pyx_L1_error))
  __pyx_v_x189 = (7680.0 * __pyx_v_x188);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":400
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))             # <<<<<<<<<<<<<<
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 */
  __Pyx_TraceLine(400,0,__PYX_ERR(1, 400, __pyx_L1_error))
  __pyx_v_x190 = (((__pyx_v_x116 + (__pyx_v_x14 * (((0.03125 * __pyx_v_x181) + (0.015625 * __pyx_v_x182)) + (0.015625 * __pyx_v_x183)))) + (__pyx_v_x187 * __pyx_v_x189)) + (__pyx_v_x24 * (((__pyx_v_x180 * (__pyx_v_x176 - 1.0)) - (2.625 * __pyx_v_x57)) + (__pyx_v_x9 * (__pyx_v_x15 + 2.8125)))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":401
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190             # <<<<<<<<<<<<<<
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60
 */
  __Pyx_TraceLine(401,0,__PYX_ERR(1, 401, __pyx_L1_error))
  __pyx_v_x191 = (((((__pyx_v_x169 * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x190);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":402
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 */
  __Pyx_TraceLine(402,0,__PYX_ERR(1, 402, __pyx_L1_error))
  __pyx_v_x192 = ((((((((((((((((-__pyx_v_x1) * __pyx_v_x100) + (__pyx_v_x112 * __pyx_v_x61)) + ((1.27277314139085e-19 * __pyx_v_x113) * __pyx_v_x168)) + ((1.69542100694444e-8 * __pyx_v_x191) * __pyx_v_x35)) + (__pyx_v_x21 * __pyx_v_x92)) + ((147.443752990146 * __pyx_v_x70) * __pyx_v_x72)) + (__pyx_v_x70 * __pyx_v_x82)) - ((11.3175085791863 * __pyx_v_x73) * __pyx_v_x75)) + (__pyx_v_x73 * __pyx_v_x87)) + ((1.48275342024365 * __pyx_v_x76) * __pyx_v_x78)) + (__pyx_v_x79 * __pyx_v_x80)) + (__pyx_v_x84 * __pyx_v_x85)) + (__pyx_v_x89 * __pyx_v_x90)) + (__pyx_v_x94 * __pyx_v_x95)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":403
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60             # <<<<<<<<<<<<<<
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)
 */
  __Pyx_TraceLine(403,0,__PYX_ERR(1, 403, __pyx_L1_error))
  __pyx_v_x193 = (2.0 * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":404
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)
 */
  __Pyx_TraceLine(404,0,__PYX_ERR(1, 404, __pyx_L1_error))
  __pyx_v_x194 = ((__pyx_v_x116 * (__pyx_v_x193 + 1.0)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":405
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 */
  __Pyx_TraceLine(405,0,__PYX_ERR(1, 405, __pyx_L1_error))
  __pyx_v_x195 = pow(__pyx_v_x194, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":406
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)             # <<<<<<<<<<<<<<
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu
 */
  __Pyx_TraceLine(406,0,__PYX_ERR(1, 406, __pyx_L1_error))
  __pyx_v_x196 = (__pyx_v_x103 * (__pyx_v_x176 + 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":407
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 */
  __Pyx_TraceLine(407,0,__PYX_ERR(1, 407, __pyx_L1_error))
  __pyx_v_x197 = (__pyx_v_x57 * ((2.0 * __pyx_v_nu) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":408
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)
 */
  __Pyx_TraceLine(408,0,__PYX_ERR(1, 408, __pyx_L1_error))
  __pyx_v_x198 = (7.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":409
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7             # <<<<<<<<<<<<<<
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)
 */
  __Pyx_TraceLine(409,0,__PYX_ERR(1, 409, __pyx_L1_error))
  __pyx_v_x199 = (((0.390625 * __pyx_v_x3) + (0.390625 * __pyx_v_x6)) + (0.390625 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":410
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)             # <<<<<<<<<<<<<<
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 */
  __Pyx_TraceLine(410,0,__PYX_ERR(1, 410, __pyx_L1_error))
  __pyx_v_x200 = (__pyx_v_x49 * (__pyx_v_x198 - 31.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":411
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 */
  __Pyx_TraceLine(411,0,__PYX_ERR(1, 411, __pyx_L1_error))
  __pyx_v_x201 = (__pyx_v_x104 * ((-__pyx_v_x176) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":412
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)             # <<<<<<<<<<<<<<
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203
 */
  __Pyx_TraceLine(412,0,__PYX_ERR(1, 412, __pyx_L1_error))
  __pyx_v_x202 = (__pyx_v_x58 * ((68.0 * __pyx_v_nu) - 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":413
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65             # <<<<<<<<<<<<<<
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)
 */
  __Pyx_TraceLine(413,0,__PYX_ERR(1, 413, __pyx_L1_error))
  __pyx_v_x203 = ((((((__pyx_v_x116 + (__pyx_v_x14 * (((0.125 * __pyx_v_x200) + (0.0625 * __pyx_v_x201)) + (2.0625 * __pyx_v_x58)))) + (__pyx_v_x14 * (((((0.125 * __pyx_v_x196) + (1.125 * __pyx_v_x3)) - (1.25 * __pyx_v_x57)) + (1.125 * __pyx_v_x6)) + (1.125 * __pyx_v_x7)))) + (7680.0 * __pyx_v_x140)) + (__pyx_v_x35 * (((__pyx_v_x109 * (((2870.0 * __pyx_v_nu) - __pyx_v_x107) + 561.0)) + (__pyx_v_x111 * (((1166.0 * __pyx_v_nu) - (328.0 * __pyx_v_x28)) - 171.0))) + (0.125 * __pyx_v_x202)))) + (__pyx_v_x35 * ((((0.015625 * __pyx_v_x179) * (((-27.0 * __pyx_v_nu) + (28.0 * __pyx_v_x28)) - 3.0)) - (1.21875 * __pyx_v_x197)) - (__pyx_v_x199 * (__pyx_v_x198 + 9.0))))) + __pyx_v_x65);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":414
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203             # <<<<<<<<<<<<<<
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 */
  __Pyx_TraceLine(414,0,__PYX_ERR(1, 414, __pyx_L1_error))
  __pyx_v_x204 = (__pyx_v_x195 * __pyx_v_x203);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":415
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)             # <<<<<<<<<<<<<<
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0
 */
  __Pyx_TraceLine(415,0,__PYX_ERR(1, 415, __pyx_L1_error))
  __pyx_v_x205 = pow((__pyx_v_x192 * __pyx_v_x204), -0.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":416
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7             # <<<<<<<<<<<<<<
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2
 */
  __Pyx_TraceLine(416,0,__PYX_ERR(1, 416, __pyx_L1_error))
  __pyx_v_x206 = (((-2.0 * __pyx_v_x3) - (2.0 * __pyx_v_x6)) - (2.0 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":417
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0             # <<<<<<<<<<<<<<
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)
 */
  __Pyx_TraceLine(417,0,__PYX_ERR(1, 417, __pyx_L1_error))
  __pyx_v_x207 = ((4.0 * __pyx_v_x60) + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":418
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2             # <<<<<<<<<<<<<<
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 */
  __Pyx_TraceLine(418,0,__PYX_ERR(1, 418, __pyx_L1_error))
  __pyx_v_x208 = (4.5 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":419
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)             # <<<<<<<<<<<<<<
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60
 */
  __Pyx_TraceLine(419,0,__PYX_ERR(1, 419, __pyx_L1_error))
  __pyx_v_x209 = pow(__pyx_v_r, -6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":420
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0             # <<<<<<<<<<<<<<
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0
 */
  __Pyx_TraceLine(420,0,__PYX_ERR(1, 420, __pyx_L1_error))
  __pyx_v_x210 = (((((-6572428.80109422 * __pyx_v_nu) + ((2048.0 * __pyx_v_x118) * __pyx_v_x79)) + (688128.0 * __pyx_v_x128)) + (1300341.64885296 * __pyx_v_x28)) + 1720320.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":421
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60             # <<<<<<<<<<<<<<
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 */
  __Pyx_TraceLine(421,0,__PYX_ERR(1, 421, __pyx_L1_error))
  __pyx_v_x211 = (__pyx_v_x28 * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":422
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0             # <<<<<<<<<<<<<<
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 */
  __Pyx_TraceLine(422,0,__PYX_ERR(1, 422, __pyx_L1_error))
  __pyx_v_x212 = (4.0 * __pyx_v_x0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":423
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0             # <<<<<<<<<<<<<<
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214
 */
  __Pyx_TraceLine(423,0,__PYX_ERR(1, 423, __pyx_L1_error))
  __pyx_v_x213 = ((__pyx_v_x130 + (6.0 * __pyx_v_x20)) + 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":424
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2             # <<<<<<<<<<<<<<
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24
 */
  __Pyx_TraceLine(424,0,__PYX_ERR(1, 424, __pyx_L1_error))
  __pyx_v_x214 = (((1.31621673590926e-19 * __pyx_v_x119) * ((((((((53760.0 * __pyx_v_nu) * (((((((((3740417.71815805 * __pyx_v_r) + (1057984.42953951 * __pyx_v_x0)) - (938918.400156317 * __pyx_v_x13)) + (__pyx_v_x132 * (__pyx_v_x212 + __pyx_v_x213))) + (__pyx_v_x133 * ((((17058.7851094412 * __pyx_v_r) - (27409.3925547206 * __pyx_v_x0)) + (12794.0888320809 * __pyx_v_x20)) + 13218.7851094412))) + (__pyx_v_x135 * ((((1396.0 * __pyx_v_x0) + __pyx_v_x134) + (5778.0 * __pyx_v_x20)) + 7704.0))) + (__pyx_v_x136 * ((-__pyx_v_x212) + __pyx_v_x213))) + (2115968.85907902 * __pyx_v_x20)) + 2888096.47013111)) + ((32768.0 * __pyx_v_x127) * __pyx_v_x79)) + ((32768.0 * __pyx_v_x128) * (((((((-1791904.9465589 * __pyx_v_nu) + ((3840.0 * __pyx_v_r) * __pyx_v_x124)) + ((2880.0 * __pyx_v_x125) * __pyx_v_x20)) + (1920.0 * __pyx_v_x126)) + (806400.0 * __pyx_v_x13)) - (725760.0 * __pyx_v_x28)) + 4143360.0))) + (66060288000.0 * __pyx_v_x13)) + ((135291469824.0 * __pyx_v_x211) * __pyx_v_x93)) + ((7.0 * __pyx_v_x28) * ((((((-117964800.0 * __pyx_v_a6) + (4129567622.65173 * __pyx_v_r)) - (12357002815.0752 * __pyx_v_x0)) + (7133302759.42198 * __pyx_v_x13)) - (18535504222.6128 * __pyx_v_x20)) + 122635399361.987))) + ((7.0 * __pyx_v_x83) * ((-5706642207.53758 * __pyx_v_r) - 26189434371.6082)))) / pow((((((((0.28004222119933 * __pyx_v_x121) + (4.63661586574928e-9 * __pyx_v_x122)) + (2.8978849160933e-11 * __pyx_v_x123)) + (1.35654132757922e-7 * __pyx_v_x129)) + (2.22557561555966e-7 * __pyx_v_x137)) + (0.0546957463279941 * __pyx_v_x34)) + __pyx_v_x88), 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":425
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214             # <<<<<<<<<<<<<<
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93
 */
  __Pyx_TraceLine(425,0,__PYX_ERR(1, 425, __pyx_L1_error))
  __pyx_v_x215 = (__pyx_v_x13 * __pyx_v_x214);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":426
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24             # <<<<<<<<<<<<<<
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 */
  __Pyx_TraceLine(426,0,__PYX_ERR(1, 426, __pyx_L1_error))
  __pyx_v_x216 = (2.0 * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":427
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93             # <<<<<<<<<<<<<<
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70
 */
  __Pyx_TraceLine(427,0,__PYX_ERR(1, 427, __pyx_L1_error))
  __pyx_v_x217 = (11575876386816.0 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":428
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1             # <<<<<<<<<<<<<<
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 */
  __Pyx_TraceLine(428,0,__PYX_ERR(1, 428, __pyx_L1_error))
  __pyx_v_x218 = (((((((((5807150888816.34 * __pyx_v_nu) + (10215490662751.4 * __pyx_v_r)) + (5178202125747.62 * __pyx_v_x156)) + (267544166400.0 * __pyx_v_x158)) - (__pyx_v_x162 * __pyx_v_x60)) + __pyx_v_x172) + (__pyx_v_x217 * __pyx_v_x60)) + (__pyx_v_x93 * (((4161798144000.0 * __pyx_v_nu) + (1759846072320.0 * __pyx_v_r)) + 12148770078720.0))) - 53501685054374.1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":429
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70             # <<<<<<<<<<<<<<
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190
 */
  __Pyx_TraceLine(429,0,__PYX_ERR(1, 429, __pyx_L1_error))
  __pyx_v_x219 = ((((__pyx_v_x113 * __pyx_v_x115) * __pyx_v_x155) * __pyx_v_x167) * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":430
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)             # <<<<<<<<<<<<<<
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24
 */
  __Pyx_TraceLine(430,0,__PYX_ERR(1, 430, __pyx_L1_error))
  __pyx_v_x220 = ((((((((-18432.0 * __pyx_v_r) * __pyx_v_x144) - ((2903040.0 * __pyx_v_x146) * __pyx_v_x20)) + __pyx_v_x149) + (__pyx_v_x153 * __pyx_v_x60)) + __pyx_v_x173) + __pyx_v_x217) + (__pyx_v_x93 * ((((20113376778784.3 * __pyx_v_nu) + (566231040.0 * __pyx_v_x159)) + __pyx_v_x161) + (2639769108480.0 * __pyx_v_x20))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":431
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190             # <<<<<<<<<<<<<<
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210
 */
  __Pyx_TraceLine(431,0,__PYX_ERR(1, 431, __pyx_L1_error))
  __pyx_v_x221 = (__pyx_v_x170 * __pyx_v_x190);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":432
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24             # <<<<<<<<<<<<<<
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 */
  __Pyx_TraceLine(432,0,__PYX_ERR(1, 432, __pyx_L1_error))
  __pyx_v_x222 = (__pyx_v_x206 * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":433
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210             # <<<<<<<<<<<<<<
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35
 */
  __Pyx_TraceLine(433,0,__PYX_ERR(1, 433, __pyx_L1_error))
  __pyx_v_x223 = (7680.0 * __pyx_v_x210);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":434
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222             # <<<<<<<<<<<<<<
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3
 */
  __Pyx_TraceLine(434,0,__PYX_ERR(1, 434, __pyx_L1_error))
  __pyx_v_x224 = (((((30720.0 * __pyx_v_x0) * __pyx_v_x188) + (__pyx_v_x139 * __pyx_v_x223)) - __pyx_v_x215) + __pyx_v_x222);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":435
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35             # <<<<<<<<<<<<<<
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5
 */
  __Pyx_TraceLine(435,0,__PYX_ERR(1, 435, __pyx_L1_error))
  __pyx_v_x225 = (__pyx_v_x184 * __pyx_v_x35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":436
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3             # <<<<<<<<<<<<<<
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7
 */
  __Pyx_TraceLine(436,0,__PYX_ERR(1, 436, __pyx_L1_error))
  __pyx_v_x226 = pow(__pyx_v_prst, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":437
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5             # <<<<<<<<<<<<<<
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226
 */
  __Pyx_TraceLine(437,0,__PYX_ERR(1, 437, __pyx_L1_error))
  __pyx_v_x227 = pow(__pyx_v_prst, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":438
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7             # <<<<<<<<<<<<<<
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227
 */
  __Pyx_TraceLine(438,0,__PYX_ERR(1, 438, __pyx_L1_error))
  __pyx_v_x228 = pow(__pyx_v_prst, 7.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":439
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226             # <<<<<<<<<<<<<<
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 */
  __Pyx_TraceLine(439,0,__PYX_ERR(1, 439, __pyx_L1_error))
  __pyx_v_x229 = (4.0 * __pyx_v_x226);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":440
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227             # <<<<<<<<<<<<<<
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21
 */
  __Pyx_TraceLine(440,0,__PYX_ERR(1, 440, __pyx_L1_error))
  __pyx_v_x230 = (6.0 * __pyx_v_x227);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":441
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)             # <<<<<<<<<<<<<<
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216
 */
  __Pyx_TraceLine(441,0,__PYX_ERR(1, 441, __pyx_L1_error))
  __pyx_v_x231 = ((__pyx_v_x203 * __pyx_v_x205) / ((__pyx_v_x116 * __pyx_v_x207) + 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":442
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21             # <<<<<<<<<<<<<<
 *         cdef double x233 = pphi*x216
 *         cdef double x234 = 4.0*pphi**3*x14
 */
  __Pyx_TraceLine(442,0,__PYX_ERR(1, 442, __pyx_L1_error))
  __pyx_v_x232 = ((2.0 * __pyx_v_pphi) * __pyx_v_x21);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":443
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216             # <<<<<<<<<<<<<<
 *         cdef double x234 = 4.0*pphi**3*x14
 * 
 */
  __Pyx_TraceLine(443,0,__PYX_ERR(1, 443, __pyx_L1_error))
  __pyx_v_x233 = (__pyx_v_pphi * __pyx_v_x216);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":444
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216
 *         cdef double x234 = 4.0*pphi**3*x14             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(444,0,__PYX_ERR(1, 444, __pyx_L1_error))
  __pyx_v_x234 = ((4.0 * pow(__pyx_v_pphi, 3.0)) * __pyx_v_x14);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":449
 *         # Evaluate Hamiltonian
 *         cdef double H
 *         H,_,_,_,_,_,_,_,_ = self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)             # <<<<<<<<<<<<<<
 * 
 *         # Heff Jacobian expressions
 */
  __Pyx_TraceLine(449,0,__PYX_ERR(1, 449, __pyx_L1_error))
  __pyx_t_1 = ((struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._call(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 9)) {
      if (size > 9) __Pyx_RaiseTooManyValuesError(9);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 449, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_16 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_12 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_11 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 8); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_16 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_12 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_11 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_10 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 8); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_7);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[9] = {&__pyx_t_2,&__pyx_t_13,&__pyx_t_16,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_7};
      for (i=0; i < 9; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(1, 449, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[9] = {&__pyx_t_2,&__pyx_t_13,&__pyx_t_16,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_7};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_19 = Py_TYPE(__pyx_t_6)->tp_iternext;
    for (index=0; index < 9; index++) {
      PyObject* item = __pyx_t_19(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_6), 9) < 0) __PYX_ERR(1, 449, __pyx_L1_error)
    __pyx_t_19 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_19 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(1, 449, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 449, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_H = __pyx_t_18;
  __pyx_v__ = __pyx_t_13;
  __pyx_t_13 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_16);
  __pyx_t_16 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_12);
  __pyx_t_12 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_11);
  __pyx_t_11 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_8);
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":452
 * 
 *         # Heff Jacobian expressions
 *         cdef double dHeffdr = x12*(-dSO*x14*x15*x19 + x19*(x24*(1.25*x3 - 0.75*x48 - 0.416666666666667*x57 + 1.66666666666667*x58 + 1.25*x6 + 1.25*x7) - x31*x55) + x19*(-x21*x22 - x23*x27 - x24*(-11.0625*nu + 1.13541666666667*x28 - 0.15625) - x29*x31 - x32*x37) + x43*(0.375*x14*x49*x50 - x24*(0.0833333333333333*x51 + x52)) + x44*(-x21*x38 - x24*(-0.0625*nu + 1.07291666666667*x28 + 0.15625) - x27*x39 - x31*x40 - x37*x41)) + 0.5*x205*(x192*x195*(30720.0*x0*x119*x138 + 7680.0*x13*x138*x210 - x14*(x177 + x178 - 3.0*x48) + x206*x24 - x209*(x109*(14350.0*nu - 1320.0*x28 + 2805.0) + x111*(5830.0*nu - 1640.0*x28 - 855.0) + 0.625*x202) - x209*(0.015625*x179*(-135.0*nu + 140.0*x28 - 15.0) - 6.09375*x197 - x199*(35.0*nu + 45.0)) - x215 - x35*(0.5*x200 + 0.25*x201 + 8.25*x58) - x35*(0.5*x196 + x2*x208 - 5.0*x57 + 4.5*x6 + 4.5*x7)) - x192*x203*(x14*x206 - x207*x24*x9)/x194**2 + x204*(-663.496888455656*nu*r**(-5.5)*x70 - nu*x21*x80 + 39.6112800271521*nu*x71*x73 + x1*x21*x25*x97*x98 + x1*x25*x60*x98*(r*x9 + x10 + x212)/x96**2 - x100 - x112*x27 + 7.59859378406358e-45*x113*x115*x142*x165*x167*x184*x220*x70 + 6.78168402777778e-8*x114*x14*x170*x171*x175*x184*x190 + x14*x70*(118.4*x211 - 51.6952380952381*x79) - 9.25454462627843e-34*x142*x218*x219/x164**3 - 2.24091649004576e-37*x165*x169*x171*x184*x218*x221*x35 + 1.69542100694444e-8*x169*x170*x171*x175*x184*x35*(x138*x187*x223 - x14*(x180*(12.0*nu - 3.0) - 7.875*x57 + x9*(9.0*nu + 8.4375)) + 38400.0*x140*x186 + x185*x189*x218*x34 - x187*x214 + x222 - x35*(0.125*x181 + 0.0625*x182 + 0.0625*x183) - 2.29252167428035e-22*x174*x188*x220*x34/x155) + 1.69542100694444e-8*x169*x170*x171*x175*x190*x220*x35 - 8.47710503472222e-8*x191*x209 + x21*x25*(-x14*(x101*(-5.25*nu - 2.8125) + x102*(2.25 - x208) + x105*(0.5625 - 2.25*nu)) + x24*x49 - x35*(0.03125*x104*x110 - 0.25*x106 + 0.0104166666666667*x108*x49)) - x216*x92 - x30*x73*x89 - x36*x70*x94 - 3.70688355060912*x74*x78 - 2.0*x81*x85 - 3.0*x84*x95 - 2.0*x86*x90 - 4.41515887225116e-12*x169*x175*x221*x224*x225/x141**3 - 6.62902677807736e-23*x165*x219*x224/x141**5 + 1.01821851311268e-18*x114**3*x142*x155*x165*x167*x70/r**12 - 1.65460508380811e-18*x168/r**14)) - (3.0*x20 + x9)*(pphi*x63 + pphi*x67 + pphi*x69 + x19*x59 + x43*x64)/x11**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double dHeffdphi = 0
 */
  __Pyx_TraceLine(452,0,__PYX_ERR(1, 452, __pyx_L1_error))
  __pyx_v_dHeffdr = (((__pyx_v_x12 * ((((((((-__pyx_v_dSO) * __pyx_v_x14) * __pyx_v_x15) * __pyx_v_x19) + (__pyx_v_x19 * ((__pyx_v_x24 * ((((((1.25 * __pyx_v_x3) - (0.75 * __pyx_v_x48)) - (0.416666666666667 * __pyx_v_x57)) + (1.66666666666667 * __pyx_v_x58)) + (1.25 * __pyx_v_x6)) + (1.25 * __pyx_v_x7))) - (__pyx_v_x31 * __pyx_v_x55)))) + (__pyx_v_x19 * ((((((-__pyx_v_x21) * __pyx_v_x22) - (__pyx_v_x23 * __pyx_v_x27)) - (__pyx_v_x24 * (((-11.0625 * __pyx_v_nu) + (1.13541666666667 * __pyx_v_x28)) - 0.15625))) - (__pyx_v_x29 * __pyx_v_x31)) - (__pyx_v_x32 * __pyx_v_x37)))) + (__pyx_v_x43 * ((((0.375 * __pyx_v_x14) * __pyx_v_x49) * __pyx_v_x50) - (__pyx_v_x24 * ((0.0833333333333333 * __pyx_v_x51) + __pyx_v_x52))))) + (__pyx_v_x44 * ((((((-__pyx_v_x21) * __pyx_v_x38) - (__pyx_v_x24 * (((-0.0625 * __pyx_v_nu) + (1.07291666666667 * __pyx_v_x28)) + 0.15625))) - (__pyx_v_x27 * __pyx_v_x39)) - (__pyx_v_x31 * __pyx_v_x40)) - (__pyx_v_x37 * __pyx_v_x41))))) + ((0.5 * __pyx_v_x205) * ((((__pyx_v_x192 * __pyx_v_x195) * (((((((((((30720.0 * __pyx_v_x0) * __pyx_v_x119) * __pyx_v_x138) + (((7680.0 * __pyx_v_x13) * __pyx_v_x138) * __pyx_v_x210)) - (__pyx_v_x14 * ((__pyx_v_x177 + __pyx_v_x178) - (3.0 * __pyx_v_x48)))) + (__pyx_v_x206 * __pyx_v_x24)) - (__pyx_v_x209 * (((__pyx_v_x109 * (((14350.0 * __pyx_v_nu) - (1320.0 * __pyx_v_x28)) + 2805.0)) + (__pyx_v_x111 * (((5830.0 * __pyx_v_nu) - (1640.0 * __pyx_v_x28)) - 855.0))) + (0.625 * __pyx_v_x202)))) - (__pyx_v_x209 * ((((0.015625 * __pyx_v_x179) * (((-135.0 * __pyx_v_nu) + (140.0 * __pyx_v_x28)) - 15.0)) - (6.09375 * __pyx_v_x197)) - (__pyx_v_x199 * ((35.0 * __pyx_v_nu) + 45.0))))) - __pyx_v_x215) - (__pyx_v_x35 * (((0.5 * __pyx_v_x200) + (0.25 * __pyx_v_x201)) + (8.25 * __pyx_v_x58)))) - (__pyx_v_x35 * (((((0.5 * __pyx_v_x196) + (__pyx_v_x2 * __pyx_v_x208)) - (5.0 * __pyx_v_x57)) + (4.5 * __pyx_v_x6)) + (4.5 * __pyx_v_x7))))) - (((__pyx_v_x192 * __pyx_v_x203) * ((__pyx_v_x14 * __pyx_v_x206) - ((__pyx_v_x207 * __pyx_v_x24) * __pyx_v_x9))) / pow(__pyx_v_x194, 2.0))) + (__pyx_v_x204 * (((((((((((((((((((((((((((((-663.496888455656 * __pyx_v_nu) * pow(__pyx_v_r, -5.5)) * __pyx_v_x70) - ((__pyx_v_nu * __pyx_v_x21) * __pyx_v_x80)) + (((39.6112800271521 * __pyx_v_nu) * __pyx_v_x71) * __pyx_v_x73)) + ((((__pyx_v_x1 * __pyx_v_x21) * __pyx_v_x25) * __pyx_v_x97) * __pyx_v_x98)) + (((((__pyx_v_x1 * __pyx_v_x25) * __pyx_v_x60) * __pyx_v_x98) * (((__pyx_v_r * __pyx_v_x9) + __pyx_v_x10) + __pyx_v_x212)) / pow(__pyx_v_x96, 2.0))) - __pyx_v_x100) - (__pyx_v_x112 * __pyx_v_x27)) + ((((((((7.59859378406358e-45 * __pyx_v_x113) * __pyx_v_x115) * __pyx_v_x142) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x184) * __pyx_v_x220) * __pyx_v_x70)) + (((((((6.78168402777778e-8 * __pyx_v_x114) * __pyx_v_x14) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x190)) + ((__pyx_v_x14 * __pyx_v_x70) * ((118.4 * __pyx_v_x211) - (51.6952380952381 * __pyx_v_x79)))) - ((((9.25454462627843e-34 * __pyx_v_x142) * __pyx_v_x218) * __pyx_v_x219) / pow(__pyx_v_x164, 3.0))) - (((((((2.24091649004576e-37 * __pyx_v_x165) * __pyx_v_x169) * __pyx_v_x171) * __pyx_v_x184) * __pyx_v_x218) * __pyx_v_x221) * __pyx_v_x35)) + (((((((1.69542100694444e-8 * __pyx_v_x169) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x35) * (((((((((__pyx_v_x138 * __pyx_v_x187) * __pyx_v_x223) - (__pyx_v_x14 * (((__pyx_v_x180 * ((12.0 * __pyx_v_nu) - 3.0)) - (7.875 * __pyx_v_x57)) + (__pyx_v_x9 * ((9.0 * __pyx_v_nu) + 8.4375))))) + ((38400.0 * __pyx_v_x140) * __pyx_v_x186)) + (((__pyx_v_x185 * __pyx_v_x189) * __pyx_v_x218) * __pyx_v_x34)) - (__pyx_v_x187 * __pyx_v_x214)) + __pyx_v_x222) - (__pyx_v_x35 * (((0.125 * __pyx_v_x181) + (0.0625 * __pyx_v_x182)) + (0.0625 * __pyx_v_x183)))) - (((((2.29252167428035e-22 * __pyx_v_x174) * __pyx_v_x188) * __pyx_v_x220) * __pyx_v_x34) / __pyx_v_x155)))) + (((((((1.69542100694444e-8 * __pyx_v_x169) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x190) * __pyx_v_x220) * __pyx_v_x35)) - ((8.47710503472222e-8 * __pyx_v_x191) * __pyx_v_x209)) + ((__pyx_v_x21 * __pyx_v_x25) * ((((-__pyx_v_x14) * (((__pyx_v_x101 * ((-5.25 * __pyx_v_nu) - 2.8125)) + (__pyx_v_x102 * (2.25 - __pyx_v_x208))) + (__pyx_v_x105 * (0.5625 - (2.25 * __pyx_v_nu))))) + (__pyx_v_x24 * __pyx_v_x49)) - (__pyx_v_x35 * ((((0.03125 * __pyx_v_x104) * __pyx_v_x110) - (0.25 * __pyx_v_x106)) + ((0.0104166666666667 * __pyx_v_x108) * __pyx_v_x49)))))) - (__pyx_v_x216 * __pyx_v_x92)) - ((__pyx_v_x30 * __pyx_v_x73) * __pyx_v_x89)) - ((__pyx_v_x36 * __pyx_v_x70) * __pyx_v_x94)) - ((3.70688355060912 * __pyx_v_x74) * __pyx_v_x78)) - ((2.0 * __pyx_v_x81) * __pyx_v_x85)) - ((3.0 * __pyx_v_x84) * __pyx_v_x95)) - ((2.0 * __pyx_v_x86) * __pyx_v_x90)) - ((((((4.41515887225116e-12 * __pyx_v_x169) * __pyx_v_x175) * __pyx_v_x221) * __pyx_v_x224) * __pyx_v_x225) / pow(__pyx_v_x141, 3.0))) - ((((6.62902677807736e-23 * __pyx_v_x165) * __pyx_v_x219) * __pyx_v_x224) / pow(__pyx_v_x141, 5.0))) + (((((((1.01821851311268e-18 * pow(__pyx_v_x114, 3.0)) * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x70) / pow(__pyx_v_r, 12.0))) - ((1.65460508380811e-18 * __pyx_v_x168) / pow(__pyx_v_r, 14.0))))))) - ((((3.0 * __pyx_v_x20) + __pyx_v_x9) * (((((__pyx_v_pphi * __pyx_v_x63) + (__pyx_v_pphi * __pyx_v_x67)) + (__pyx_v_pphi * __pyx_v_x69)) + (__pyx_v_x19 * __pyx_v_x59)) + (__pyx_v_x43 * __pyx_v_x64))) / pow(__pyx_v_x11, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":454
 *         cdef double dHeffdr = x12*(-dSO*x14*x15*x19 + x19*(x24*(1.25*x3 - 0.75*x48 - 0.416666666666667*x57 + 1.66666666666667*x58 + 1.25*x6 + 1.25*x7) - x31*x55) + x19*(-x21*x22 - x23*x27 - x24*(-11.0625*nu + 1.13541666666667*x28 - 0.15625) - x29*x31 - x32*x37) + x43*(0.375*x14*x49*x50 - x24*(0.0833333333333333*x51 + x52)) + x44*(-x21*x38 - x24*(-0.0625*nu + 1.07291666666667*x28 + 0.15625) - x27*x39 - x31*x40 - x37*x41)) + 0.5*x205*(x192*x195*(30720.0*x0*x119*x138 + 7680.0*x13*x138*x210 - x14*(x177 + x178 - 3.0*x48) + x206*x24 - x209*(x109*(14350.0*nu - 1320.0*x28 + 2805.0) + x111*(5830.0*nu - 1640.0*x28 - 855.0) + 0.625*x202) - x209*(0.015625*x179*(-135.0*nu + 140.0*x28 - 15.0) - 6.09375*x197 - x199*(35.0*nu + 45.0)) - x215 - x35*(0.5*x200 + 0.25*x201 + 8.25*x58) - x35*(0.5*x196 + x2*x208 - 5.0*x57 + 4.5*x6 + 4.5*x7)) - x192*x203*(x14*x206 - x207*x24*x9)/x194**2 + x204*(-663.496888455656*nu*r**(-5.5)*x70 - nu*x21*x80 + 39.6112800271521*nu*x71*x73 + x1*x21*x25*x97*x98 + x1*x25*x60*x98*(r*x9 + x10 + x212)/x96**2 - x100 - x112*x27 + 7.59859378406358e-45*x113*x115*x142*x165*x167*x184*x220*x70 + 6.78168402777778e-8*x114*x14*x170*x171*x175*x184*x190 + x14*x70*(118.4*x211 - 51.6952380952381*x79) - 9.25454462627843e-34*x142*x218*x219/x164**3 - 2.24091649004576e-37*x165*x169*x171*x184*x218*x221*x35 + 1.69542100694444e-8*x169*x170*x171*x175*x184*x35*(x138*x187*x223 - x14*(x180*(12.0*nu - 3.0) - 7.875*x57 + x9*(9.0*nu + 8.4375)) + 38400.0*x140*x186 + x185*x189*x218*x34 - x187*x214 + x222 - x35*(0.125*x181 + 0.0625*x182 + 0.0625*x183) - 2.29252167428035e-22*x174*x188*x220*x34/x155) + 1.69542100694444e-8*x169*x170*x171*x175*x190*x220*x35 - 8.47710503472222e-8*x191*x209 + x21*x25*(-x14*(x101*(-5.25*nu - 2.8125) + x102*(2.25 - x208) + x105*(0.5625 - 2.25*nu)) + x24*x49 - x35*(0.03125*x104*x110 - 0.25*x106 + 0.0104166666666667*x108*x49)) - x216*x92 - x30*x73*x89 - x36*x70*x94 - 3.70688355060912*x74*x78 - 2.0*x81*x85 - 3.0*x84*x95 - 2.0*x86*x90 - 4.41515887225116e-12*x169*x175*x221*x224*x225/x141**3 - 6.62902677807736e-23*x165*x219*x224/x141**5 + 1.01821851311268e-18*x114**3*x142*x155*x165*x167*x70/r**12 - 1.65460508380811e-18*x168/r**14)) - (3.0*x20 + x9)*(pphi*x63 + pphi*x67 + pphi*x69 + x19*x59 + x43*x64)/x11**2
 * 
 *         cdef double dHeffdphi = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)
 */
  __Pyx_TraceLine(454,0,__PYX_ERR(1, 454, __pyx_L1_error))
  __pyx_v_dHeffdphi = 0.0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":456
 *         cdef double dHeffdphi = 0
 * 
 *         cdef double dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)             # <<<<<<<<<<<<<<
 * 
 *         cdef double dHeffdpphi = x12*(x18*x59 + 2.0*x18*x68 + x19*(x23*x232 + x233*x29 + x234*x32) + x42*x64 - 0.25*x42*x66 + x44*(x232*x39 + x233*x40 + x234*x41) + x63 + x67 + x69) + x231*(-pphi*x1*x193*x99 + 2.0*pphi*x112*x21)
 */
  __Pyx_TraceLine(456,0,__PYX_ERR(1, 456, __pyx_L1_error))
  __pyx_v_dHeffdpr = (__pyx_v_x231 * ((((((((((((((11.8620273619492 * __pyx_v_nu) * __pyx_v_x228) * __pyx_v_x76) + ((((((3.39084201388889e-8 * __pyx_v_prst) * __pyx_v_x169) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x190) * __pyx_v_x225)) + (((((((5.09109256556341e-19 * __pyx_v_x113) * __pyx_v_x115) * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x226)) + ((__pyx_v_x14 * __pyx_v_x229) * __pyx_v_x94)) + (((8.0 * __pyx_v_x21) * __pyx_v_x228) * __pyx_v_x91)) + ((589.775011960583 * __pyx_v_x226) * __pyx_v_x72)) - ((67.9050514751178 * __pyx_v_x227) * __pyx_v_x75)) + ((0.975638950243592 * __pyx_v_x228) * __pyx_v_x79)) + ((__pyx_v_x229 * __pyx_v_x24) * __pyx_v_x84)) + (__pyx_v_x229 * __pyx_v_x82)) + ((__pyx_v_x230 * __pyx_v_x24) * __pyx_v_x89)) + (__pyx_v_x230 * __pyx_v_x87)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":458
 *         cdef double dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)
 * 
 *         cdef double dHeffdpphi = x12*(x18*x59 + 2.0*x18*x68 + x19*(x23*x232 + x233*x29 + x234*x32) + x42*x64 - 0.25*x42*x66 + x44*(x232*x39 + x233*x40 + x234*x41) + x63 + x67 + x69) + x231*(-pphi*x1*x193*x99 + 2.0*pphi*x112*x21)             # <<<<<<<<<<<<<<
 * 
 *         # Compute H Jacobian
 */
  __Pyx_TraceLine(458,0,__PYX_ERR(1, 458, __pyx_L1_error))
  __pyx_v_dHeffdpphi = ((__pyx_v_x12 * (((((((((__pyx_v_x18 * __pyx_v_x59) + ((2.0 * __pyx_v_x18) * __pyx_v_x68)) + (__pyx_v_x19 * (((__pyx_v_x23 * __pyx_v_x232) + (__pyx_v_x233 * __pyx_v_x29)) + (__pyx_v_x234 * __pyx_v_x32)))) + (__pyx_v_x42 * __pyx_v_x64)) - ((0.25 * __pyx_v_x42) * __pyx_v_x66)) + (__pyx_v_x44 * (((__pyx_v_x232 * __pyx_v_x39) + (__pyx_v_x233 * __pyx_v_x40)) + (__pyx_v_x234 * __pyx_v_x41)))) + __pyx_v_x63) + __pyx_v_x67) + __pyx_v_x69)) + (__pyx_v_x231 * (((((-__pyx_v_pphi) * __pyx_v_x1) * __pyx_v_x193) * __pyx_v_x99) + (((2.0 * __pyx_v_pphi) * __pyx_v_x112) * __pyx_v_x21))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":461
 * 
 *         # Compute H Jacobian
 *         cdef double dHdr = M * M * dHeffdr / (nu*H)             # <<<<<<<<<<<<<<
 *         cdef double dHdphi = M * M * dHeffdphi / (nu*H)
 *         cdef double dHdpr = M * M * dHeffdpr / (nu*H)
 */
  __Pyx_TraceLine(461,0,__PYX_ERR(1, 461, __pyx_L1_error))
  __pyx_v_dHdr = (((__pyx_v_M * __pyx_v_M) * __pyx_v_dHeffdr) / (__pyx_v_nu * __pyx_v_H));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":462
 *         # Compute H Jacobian
 *         cdef double dHdr = M * M * dHeffdr / (nu*H)
 *         cdef double dHdphi = M * M * dHeffdphi / (nu*H)             # <<<<<<<<<<<<<<
 *         cdef double dHdpr = M * M * dHeffdpr / (nu*H)
 *         cdef double dHdpphi = M * M * dHeffdpphi / (nu*H)
 */
  __Pyx_TraceLine(462,0,__PYX_ERR(1, 462, __pyx_L1_error))
  __pyx_v_dHdphi = (((__pyx_v_M * __pyx_v_M) * __pyx_v_dHeffdphi) / (__pyx_v_nu * __pyx_v_H));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":463
 *         cdef double dHdr = M * M * dHeffdr / (nu*H)
 *         cdef double dHdphi = M * M * dHeffdphi / (nu*H)
 *         cdef double dHdpr = M * M * dHeffdpr / (nu*H)             # <<<<<<<<<<<<<<
 *         cdef double dHdpphi = M * M * dHeffdpphi / (nu*H)
 * 
 */
  __Pyx_TraceLine(463,0,__PYX_ERR(1, 463, __pyx_L1_error))
  __pyx_v_dHdpr = (((__pyx_v_M * __pyx_v_M) * __pyx_v_dHeffdpr) / (__pyx_v_nu * __pyx_v_H));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":464
 *         cdef double dHdphi = M * M * dHeffdphi / (nu*H)
 *         cdef double dHdpr = M * M * dHeffdpr / (nu*H)
 *         cdef double dHdpphi = M * M * dHeffdpphi / (nu*H)             # <<<<<<<<<<<<<<
 * 
 *         return dHdr, dHdphi, dHdpr, dHdpphi
 */
  __Pyx_TraceLine(464,0,__PYX_ERR(1, 464, __pyx_L1_error))
  __pyx_v_dHdpphi = (((__pyx_v_M * __pyx_v_M) * __pyx_v_dHeffdpphi) / (__pyx_v_nu * __pyx_v_H));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":466
 *         cdef double dHdpphi = M * M * dHeffdpphi / (nu*H)
 * 
 *         return dHdr, dHdphi, dHdpr, dHdpphi             # <<<<<<<<<<<<<<
 * 
 *     cpdef hessian(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 */
  __Pyx_TraceLine(466,0,__PYX_ERR(1, 466, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dHdr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_dHdphi); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_dHdpr); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_dHdpphi); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyTuple_New(4); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":160
 * 
 * 
 *     cpdef grad(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 *         """
 *         Compute the gradient of the Hamiltonian in polar coordinates.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.grad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c_coeffs);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_7grad(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_6grad[] = "\n        Compute the gradient of the Hamiltonian in polar coordinates.\n\n        Args:\n          q (double[:]): Canonical positions (r,phi).\n          p (double[:]): Canonical momenta  (prstar,pphi).\n          chi1_v (double[:]): Dimensionless spin vector of the primary.\n          chi2_v (double[:]): Dimensionless spin vector of the secondary.\n          m_1 (double): Primary mass component.\n          m_2 (double): Secondary mass component.\n          chi_1 (double): Projection of chi1_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chi_2 (double): Projection of chi2_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chiL1 (double): Projection of chi1_v onto the orbital angular momentum unit vector (l).\n          chiL2 (double): Projection of chi2_v onto the orbital angular momentum unit vector (l).\n\n        Returns:\n           (tuple) dHdr, dHdphi, dHdpr, dHdpphi\n\n        ";
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_7grad(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_q = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi1_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi2_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_m_1;
  double __pyx_v_m_2;
  double __pyx_v_chi_1;
  double __pyx_v_chi_2;
  double __pyx_v_chiL1;
  double __pyx_v_chiL2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("grad (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_q,&__pyx_n_s_p,&__pyx_n_s_chi1_v,&__pyx_n_s_chi2_v,&__pyx_n_s_m_1,&__pyx_n_s_m_2,&__pyx_n_s_chi_1,&__pyx_n_s_chi_2,&__pyx_n_s_chiL1,&__pyx_n_s_chiL2,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_q)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 1); __PYX_ERR(1, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi1_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 2); __PYX_ERR(1, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi2_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 3); __PYX_ERR(1, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 4); __PYX_ERR(1, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 5); __PYX_ERR(1, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 6); __PYX_ERR(1, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 7); __PYX_ERR(1, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 8); __PYX_ERR(1, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, 9); __PYX_ERR(1, 160, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "grad") < 0)) __PYX_ERR(1, 160, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_q = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_q.memview)) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_p.memview)) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_chi1_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi1_v.memview)) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_chi2_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi2_v.memview)) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_m_1 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_m_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_m_2 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_m_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_chi_1 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_chi_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_chi_2 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_chi_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_chiL1 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_chiL1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 160, __pyx_L3_error)
    __pyx_v_chiL2 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_chiL2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 160, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("grad", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 160, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.grad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_6grad(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_6grad(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("grad", 0);
  __Pyx_TraceCall("grad (wrapper)", __pyx_f[1], 160, 0, __PYX_ERR(1, 160, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 160, __pyx_L1_error) }
  if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 160, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 160, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 160, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_grad(__pyx_v_self, __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.grad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_q, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi1_v, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi2_v, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":468
 *         return dHdr, dHdphi, dHdpr, dHdpphi
 * 
 *     cpdef hessian(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_9hessian(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_hessian(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch) {
  double __pyx_v_r;
  CYTHON_UNUSED double __pyx_v_phi;
  double __pyx_v_prst;
  double __pyx_v_L;
  double __pyx_v_chix1;
  double __pyx_v_chiy1;
  double __pyx_v_chiz1;
  double __pyx_v_chix2;
  double __pyx_v_chiy2;
  double __pyx_v_chiz2;
  double __pyx_v_pphi;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *__pyx_v_c_coeffs = 0;
  double __pyx_v_a6;
  double __pyx_v_dSO;
  double __pyx_v_M;
  double __pyx_v_nu;
  double __pyx_v_X_1;
  double __pyx_v_X_2;
  double __pyx_v_delta;
  double __pyx_v_x0;
  double __pyx_v_x1;
  double __pyx_v_x2;
  double __pyx_v_x3;
  double __pyx_v_x4;
  double __pyx_v_x5;
  double __pyx_v_x6;
  double __pyx_v_x7;
  double __pyx_v_x8;
  double __pyx_v_x9;
  double __pyx_v_x10;
  double __pyx_v_x11;
  double __pyx_v_x12;
  double __pyx_v_x13;
  double __pyx_v_x14;
  double __pyx_v_x15;
  double __pyx_v_x16;
  double __pyx_v_x17;
  double __pyx_v_x18;
  double __pyx_v_x19;
  double __pyx_v_x20;
  double __pyx_v_x21;
  double __pyx_v_x22;
  double __pyx_v_x23;
  double __pyx_v_x24;
  double __pyx_v_x25;
  double __pyx_v_x26;
  double __pyx_v_x27;
  double __pyx_v_x28;
  double __pyx_v_x29;
  double __pyx_v_x30;
  double __pyx_v_x31;
  double __pyx_v_x32;
  double __pyx_v_x33;
  double __pyx_v_x34;
  double __pyx_v_x35;
  double __pyx_v_x36;
  double __pyx_v_x37;
  double __pyx_v_x38;
  double __pyx_v_x39;
  double __pyx_v_x40;
  double __pyx_v_x41;
  double __pyx_v_x42;
  double __pyx_v_x43;
  double __pyx_v_x44;
  double __pyx_v_x45;
  double __pyx_v_x46;
  double __pyx_v_x47;
  double __pyx_v_x48;
  double __pyx_v_x49;
  double __pyx_v_x50;
  double __pyx_v_x51;
  double __pyx_v_x52;
  double __pyx_v_x53;
  double __pyx_v_x54;
  double __pyx_v_x55;
  double __pyx_v_x56;
  double __pyx_v_x57;
  double __pyx_v_x58;
  double __pyx_v_x59;
  double __pyx_v_x60;
  double __pyx_v_x61;
  double __pyx_v_x62;
  double __pyx_v_x63;
  double __pyx_v_x64;
  double __pyx_v_x65;
  double __pyx_v_x66;
  double __pyx_v_x67;
  double __pyx_v_x68;
  double __pyx_v_x69;
  double __pyx_v_x70;
  double __pyx_v_x71;
  double __pyx_v_x72;
  double __pyx_v_x73;
  double __pyx_v_x74;
  double __pyx_v_x75;
  double __pyx_v_x76;
  double __pyx_v_x77;
  double __pyx_v_x78;
  double __pyx_v_x79;
  double __pyx_v_x80;
  double __pyx_v_x81;
  double __pyx_v_x82;
  double __pyx_v_x83;
  double __pyx_v_x84;
  double __pyx_v_x85;
  double __pyx_v_x86;
  double __pyx_v_x87;
  double __pyx_v_x88;
  double __pyx_v_x89;
  double __pyx_v_x90;
  double __pyx_v_x91;
  double __pyx_v_x92;
  double __pyx_v_x93;
  double __pyx_v_x94;
  double __pyx_v_x95;
  double __pyx_v_x96;
  double __pyx_v_x97;
  double __pyx_v_x98;
  double __pyx_v_x99;
  double __pyx_v_x100;
  double __pyx_v_x101;
  double __pyx_v_x102;
  double __pyx_v_x103;
  double __pyx_v_x104;
  double __pyx_v_x105;
  double __pyx_v_x106;
  double __pyx_v_x107;
  double __pyx_v_x108;
  double __pyx_v_x109;
  double __pyx_v_x110;
  double __pyx_v_x111;
  double __pyx_v_x112;
  double __pyx_v_x113;
  double __pyx_v_x114;
  double __pyx_v_x115;
  double __pyx_v_x116;
  double __pyx_v_x117;
  double __pyx_v_x118;
  double __pyx_v_x119;
  double __pyx_v_x120;
  double __pyx_v_x121;
  double __pyx_v_x122;
  double __pyx_v_x123;
  double __pyx_v_x124;
  double __pyx_v_x125;
  double __pyx_v_x126;
  double __pyx_v_x127;
  double __pyx_v_x128;
  double __pyx_v_x129;
  double __pyx_v_x130;
  double __pyx_v_x131;
  double __pyx_v_x132;
  double __pyx_v_x133;
  double __pyx_v_x134;
  double __pyx_v_x135;
  double __pyx_v_x136;
  double __pyx_v_x137;
  double __pyx_v_x138;
  double __pyx_v_x139;
  double __pyx_v_x140;
  double __pyx_v_x141;
  double __pyx_v_x142;
  double __pyx_v_x143;
  double __pyx_v_x144;
  double __pyx_v_x145;
  double __pyx_v_x146;
  double __pyx_v_x147;
  double __pyx_v_x148;
  double __pyx_v_x149;
  double __pyx_v_x150;
  double __pyx_v_x151;
  double __pyx_v_x152;
  double __pyx_v_x153;
  double __pyx_v_x154;
  double __pyx_v_x155;
  double __pyx_v_x156;
  double __pyx_v_x157;
  double __pyx_v_x158;
  double __pyx_v_x159;
  double __pyx_v_x160;
  double __pyx_v_x161;
  double __pyx_v_x162;
  double __pyx_v_x163;
  double __pyx_v_x164;
  double __pyx_v_x165;
  double __pyx_v_x166;
  double __pyx_v_x167;
  double __pyx_v_x168;
  double __pyx_v_x169;
  double __pyx_v_x170;
  double __pyx_v_x171;
  double __pyx_v_x172;
  double __pyx_v_x173;
  double __pyx_v_x174;
  double __pyx_v_x175;
  double __pyx_v_x176;
  double __pyx_v_x177;
  double __pyx_v_x178;
  double __pyx_v_x179;
  double __pyx_v_x180;
  double __pyx_v_x181;
  double __pyx_v_x182;
  double __pyx_v_x183;
  double __pyx_v_x184;
  double __pyx_v_x185;
  double __pyx_v_x186;
  double __pyx_v_x187;
  double __pyx_v_x188;
  double __pyx_v_x189;
  double __pyx_v_x190;
  double __pyx_v_x191;
  double __pyx_v_x192;
  double __pyx_v_x193;
  double __pyx_v_x194;
  double __pyx_v_x195;
  double __pyx_v_x196;
  double __pyx_v_x197;
  double __pyx_v_x198;
  double __pyx_v_x199;
  double __pyx_v_x200;
  double __pyx_v_x201;
  double __pyx_v_x202;
  double __pyx_v_x203;
  double __pyx_v_x204;
  double __pyx_v_x205;
  double __pyx_v_x206;
  double __pyx_v_x207;
  double __pyx_v_x208;
  double __pyx_v_x209;
  double __pyx_v_x210;
  double __pyx_v_x211;
  double __pyx_v_x212;
  double __pyx_v_x213;
  double __pyx_v_x214;
  double __pyx_v_x215;
  double __pyx_v_x216;
  double __pyx_v_x217;
  double __pyx_v_x218;
  double __pyx_v_x219;
  double __pyx_v_x220;
  double __pyx_v_x221;
  double __pyx_v_x222;
  double __pyx_v_x223;
  double __pyx_v_x224;
  double __pyx_v_x225;
  double __pyx_v_x226;
  double __pyx_v_x227;
  double __pyx_v_x228;
  double __pyx_v_x229;
  double __pyx_v_x230;
  double __pyx_v_x231;
  double __pyx_v_x232;
  double __pyx_v_x233;
  double __pyx_v_x234;
  double __pyx_v_y0;
  double __pyx_v_y1;
  double __pyx_v_y2;
  double __pyx_v_y3;
  double __pyx_v_y4;
  double __pyx_v_y5;
  double __pyx_v_y6;
  double __pyx_v_y7;
  double __pyx_v_y8;
  double __pyx_v_y9;
  double __pyx_v_y10;
  double __pyx_v_y11;
  double __pyx_v_y12;
  double __pyx_v_y13;
  double __pyx_v_y14;
  double __pyx_v_y15;
  double __pyx_v_y16;
  double __pyx_v_y17;
  double __pyx_v_y18;
  double __pyx_v_y19;
  double __pyx_v_y20;
  double __pyx_v_y21;
  double __pyx_v_y22;
  double __pyx_v_y23;
  double __pyx_v_y24;
  double __pyx_v_y25;
  double __pyx_v_y26;
  double __pyx_v_y27;
  double __pyx_v_y28;
  double __pyx_v_y29;
  double __pyx_v_y30;
  double __pyx_v_y31;
  double __pyx_v_y32;
  double __pyx_v_y33;
  double __pyx_v_y34;
  double __pyx_v_y35;
  double __pyx_v_y36;
  double __pyx_v_y37;
  double __pyx_v_y38;
  double __pyx_v_y39;
  double __pyx_v_y40;
  double __pyx_v_y41;
  double __pyx_v_y42;
  double __pyx_v_y43;
  double __pyx_v_y44;
  double __pyx_v_y45;
  double __pyx_v_y46;
  double __pyx_v_y47;
  double __pyx_v_y48;
  double __pyx_v_y49;
  double __pyx_v_y50;
  double __pyx_v_y51;
  double __pyx_v_y52;
  double __pyx_v_y53;
  double __pyx_v_y54;
  double __pyx_v_y55;
  double __pyx_v_y56;
  double __pyx_v_y57;
  double __pyx_v_y58;
  double __pyx_v_y59;
  double __pyx_v_y60;
  double __pyx_v_y61;
  double __pyx_v_y62;
  double __pyx_v_y63;
  double __pyx_v_y64;
  double __pyx_v_y65;
  double __pyx_v_y66;
  double __pyx_v_y67;
  double __pyx_v_y68;
  double __pyx_v_y69;
  double __pyx_v_y70;
  double __pyx_v_y71;
  double __pyx_v_y72;
  double __pyx_v_y73;
  double __pyx_v_y74;
  double __pyx_v_y75;
  double __pyx_v_y76;
  double __pyx_v_y77;
  double __pyx_v_y78;
  double __pyx_v_y79;
  double __pyx_v_y80;
  double __pyx_v_y81;
  double __pyx_v_y82;
  double __pyx_v_y83;
  double __pyx_v_y84;
  double __pyx_v_y85;
  double __pyx_v_y86;
  double __pyx_v_y87;
  double __pyx_v_y88;
  double __pyx_v_y89;
  double __pyx_v_y90;
  double __pyx_v_y91;
  double __pyx_v_y92;
  double __pyx_v_y93;
  double __pyx_v_y94;
  double __pyx_v_y95;
  double __pyx_v_y96;
  double __pyx_v_y97;
  double __pyx_v_y98;
  double __pyx_v_y99;
  double __pyx_v_y100;
  double __pyx_v_y101;
  double __pyx_v_y102;
  double __pyx_v_y103;
  double __pyx_v_y104;
  double __pyx_v_y105;
  double __pyx_v_y106;
  double __pyx_v_y107;
  double __pyx_v_y108;
  double __pyx_v_y109;
  double __pyx_v_y110;
  double __pyx_v_y111;
  double __pyx_v_y112;
  double __pyx_v_y113;
  double __pyx_v_y114;
  double __pyx_v_y115;
  double __pyx_v_y116;
  double __pyx_v_y117;
  double __pyx_v_y118;
  double __pyx_v_y119;
  double __pyx_v_y120;
  double __pyx_v_y121;
  double __pyx_v_y122;
  double __pyx_v_y123;
  double __pyx_v_y124;
  double __pyx_v_y125;
  double __pyx_v_y126;
  double __pyx_v_y127;
  double __pyx_v_y128;
  double __pyx_v_y129;
  double __pyx_v_y130;
  double __pyx_v_y131;
  double __pyx_v_y132;
  double __pyx_v_y133;
  double __pyx_v_y134;
  double __pyx_v_y135;
  double __pyx_v_y136;
  double __pyx_v_y137;
  double __pyx_v_y138;
  double __pyx_v_y139;
  double __pyx_v_y140;
  double __pyx_v_y141;
  double __pyx_v_y142;
  double __pyx_v_y143;
  double __pyx_v_y144;
  double __pyx_v_y145;
  double __pyx_v_y146;
  double __pyx_v_y147;
  double __pyx_v_y148;
  double __pyx_v_y149;
  double __pyx_v_y150;
  double __pyx_v_y151;
  double __pyx_v_y152;
  double __pyx_v_y153;
  double __pyx_v_y154;
  double __pyx_v_y155;
  double __pyx_v_y156;
  double __pyx_v_y157;
  double __pyx_v_y158;
  double __pyx_v_y159;
  double __pyx_v_y160;
  double __pyx_v_y161;
  double __pyx_v_y162;
  double __pyx_v_y163;
  double __pyx_v_y164;
  double __pyx_v_y165;
  double __pyx_v_y166;
  double __pyx_v_y167;
  double __pyx_v_y168;
  double __pyx_v_y169;
  double __pyx_v_y170;
  double __pyx_v_y171;
  double __pyx_v_y172;
  double __pyx_v_y173;
  double __pyx_v_y174;
  double __pyx_v_y175;
  double __pyx_v_y176;
  double __pyx_v_y177;
  double __pyx_v_y178;
  double __pyx_v_y179;
  double __pyx_v_y180;
  double __pyx_v_y181;
  double __pyx_v_y182;
  double __pyx_v_y183;
  double __pyx_v_y184;
  double __pyx_v_y185;
  double __pyx_v_y186;
  double __pyx_v_y187;
  double __pyx_v_y188;
  double __pyx_v_y189;
  double __pyx_v_y190;
  double __pyx_v_y191;
  double __pyx_v_y192;
  double __pyx_v_y193;
  double __pyx_v_y194;
  double __pyx_v_y195;
  double __pyx_v_y196;
  double __pyx_v_y197;
  double __pyx_v_y198;
  double __pyx_v_y199;
  double __pyx_v_y200;
  double __pyx_v_y201;
  double __pyx_v_y202;
  double __pyx_v_y203;
  double __pyx_v_y204;
  double __pyx_v_y205;
  double __pyx_v_y206;
  double __pyx_v_y207;
  double __pyx_v_y208;
  double __pyx_v_y209;
  double __pyx_v_y210;
  double __pyx_v_y211;
  double __pyx_v_y212;
  double __pyx_v_y213;
  double __pyx_v_y214;
  double __pyx_v_y215;
  double __pyx_v_y216;
  double __pyx_v_y217;
  double __pyx_v_y218;
  double __pyx_v_y219;
  double __pyx_v_y220;
  double __pyx_v_y221;
  double __pyx_v_y222;
  double __pyx_v_y223;
  double __pyx_v_y224;
  double __pyx_v_y225;
  double __pyx_v_y226;
  double __pyx_v_y227;
  double __pyx_v_y228;
  double __pyx_v_y229;
  double __pyx_v_y230;
  double __pyx_v_y231;
  double __pyx_v_y232;
  double __pyx_v_y233;
  double __pyx_v_y234;
  double __pyx_v_y235;
  double __pyx_v_y236;
  double __pyx_v_y237;
  double __pyx_v_y238;
  double __pyx_v_y239;
  double __pyx_v_y240;
  double __pyx_v_y241;
  double __pyx_v_y242;
  double __pyx_v_y243;
  double __pyx_v_y244;
  double __pyx_v_y245;
  double __pyx_v_y246;
  double __pyx_v_y247;
  double __pyx_v_y248;
  double __pyx_v_y249;
  double __pyx_v_y250;
  double __pyx_v_y251;
  double __pyx_v_y252;
  double __pyx_v_y253;
  double __pyx_v_y254;
  double __pyx_v_y255;
  double __pyx_v_y256;
  double __pyx_v_y257;
  double __pyx_v_y258;
  double __pyx_v_y259;
  double __pyx_v_y260;
  double __pyx_v_y261;
  double __pyx_v_y262;
  double __pyx_v_y263;
  double __pyx_v_y264;
  double __pyx_v_y265;
  double __pyx_v_y266;
  double __pyx_v_y267;
  double __pyx_v_y268;
  double __pyx_v_y269;
  double __pyx_v_y270;
  double __pyx_v_y271;
  double __pyx_v_y272;
  double __pyx_v_y273;
  double __pyx_v_y274;
  double __pyx_v_y275;
  double __pyx_v_y276;
  double __pyx_v_y277;
  double __pyx_v_y278;
  double __pyx_v_y279;
  double __pyx_v_y280;
  double __pyx_v_y281;
  double __pyx_v_y282;
  double __pyx_v_y283;
  double __pyx_v_y284;
  double __pyx_v_y285;
  double __pyx_v_y286;
  double __pyx_v_y287;
  double __pyx_v_y288;
  double __pyx_v_y289;
  double __pyx_v_y290;
  double __pyx_v_y291;
  double __pyx_v_y292;
  double __pyx_v_y293;
  double __pyx_v_y294;
  double __pyx_v_y295;
  double __pyx_v_y296;
  double __pyx_v_y297;
  double __pyx_v_y298;
  double __pyx_v_y299;
  double __pyx_v_y300;
  double __pyx_v_y301;
  double __pyx_v_y302;
  double __pyx_v_y303;
  double __pyx_v_y304;
  double __pyx_v_y305;
  double __pyx_v_y306;
  double __pyx_v_y307;
  double __pyx_v_y308;
  double __pyx_v_y309;
  double __pyx_v_y310;
  double __pyx_v_y311;
  double __pyx_v_y312;
  double __pyx_v_y313;
  double __pyx_v_y314;
  double __pyx_v_y315;
  double __pyx_v_y316;
  double __pyx_v_y317;
  double __pyx_v_y318;
  double __pyx_v_y319;
  double __pyx_v_y320;
  double __pyx_v_y321;
  double __pyx_v_y322;
  double __pyx_v_y323;
  double __pyx_v_y324;
  double __pyx_v_y325;
  double __pyx_v_y326;
  double __pyx_v_y327;
  double __pyx_v_y328;
  double __pyx_v_y329;
  double __pyx_v_y330;
  double __pyx_v_y331;
  double __pyx_v_y332;
  double __pyx_v_y333;
  double __pyx_v_y334;
  double __pyx_v_y335;
  double __pyx_v_y336;
  double __pyx_v_y337;
  double __pyx_v_y338;
  double __pyx_v_y339;
  double __pyx_v_y340;
  double __pyx_v_y341;
  double __pyx_v_y342;
  double __pyx_v_y343;
  double __pyx_v_y344;
  double __pyx_v_y345;
  double __pyx_v_y346;
  double __pyx_v_y347;
  double __pyx_v_y348;
  double __pyx_v_y349;
  double __pyx_v_y350;
  double __pyx_v_y351;
  double __pyx_v_y352;
  double __pyx_v_y353;
  double __pyx_v_y354;
  double __pyx_v_y355;
  double __pyx_v_y356;
  double __pyx_v_y357;
  double __pyx_v_y358;
  double __pyx_v_y359;
  double __pyx_v_y360;
  double __pyx_v_y361;
  double __pyx_v_y362;
  double __pyx_v_y363;
  double __pyx_v_y364;
  double __pyx_v_y365;
  double __pyx_v_y366;
  double __pyx_v_y367;
  double __pyx_v_y368;
  double __pyx_v_y369;
  double __pyx_v_y370;
  double __pyx_v_y371;
  double __pyx_v_y372;
  double __pyx_v_y373;
  double __pyx_v_y374;
  double __pyx_v_y375;
  double __pyx_v_y376;
  double __pyx_v_y377;
  double __pyx_v_y378;
  double __pyx_v_y379;
  double __pyx_v_y380;
  double __pyx_v_y381;
  double __pyx_v_y382;
  double __pyx_v_y383;
  double __pyx_v_y384;
  double __pyx_v_y385;
  double __pyx_v_y386;
  double __pyx_v_y387;
  double __pyx_v_y388;
  double __pyx_v_y389;
  double __pyx_v_y390;
  double __pyx_v_y391;
  double __pyx_v_y392;
  double __pyx_v_y393;
  double __pyx_v_y394;
  double __pyx_v_y395;
  double __pyx_v_y396;
  double __pyx_v_y397;
  double __pyx_v_y398;
  double __pyx_v_y399;
  double __pyx_v_y400;
  double __pyx_v_y401;
  double __pyx_v_y402;
  double __pyx_v_y403;
  double __pyx_v_y404;
  double __pyx_v_y405;
  double __pyx_v_y406;
  double __pyx_v_y407;
  double __pyx_v_y408;
  double __pyx_v_y409;
  double __pyx_v_y410;
  double __pyx_v_y411;
  double __pyx_v_y412;
  double __pyx_v_y413;
  double __pyx_v_y414;
  double __pyx_v_y415;
  double __pyx_v_y416;
  double __pyx_v_y417;
  double __pyx_v_y418;
  double __pyx_v_y419;
  double __pyx_v_y420;
  double __pyx_v_y421;
  double __pyx_v_y422;
  double __pyx_v_y423;
  double __pyx_v_y424;
  double __pyx_v_y425;
  double __pyx_v_y426;
  double __pyx_v_y427;
  double __pyx_v_y428;
  double __pyx_v_y429;
  double __pyx_v_y430;
  double __pyx_v_y431;
  double __pyx_v_y432;
  double __pyx_v_y433;
  double __pyx_v_y434;
  double __pyx_v_y435;
  double __pyx_v_y436;
  double __pyx_v_y437;
  double __pyx_v_y438;
  double __pyx_v_y439;
  double __pyx_v_y440;
  double __pyx_v_y441;
  double __pyx_v_y442;
  double __pyx_v_y443;
  double __pyx_v_y444;
  double __pyx_v_y445;
  double __pyx_v_y446;
  double __pyx_v_y447;
  double __pyx_v_y448;
  double __pyx_v_y449;
  double __pyx_v_y450;
  double __pyx_v_y451;
  double __pyx_v_y452;
  double __pyx_v_y453;
  double __pyx_v_y454;
  double __pyx_v_y455;
  double __pyx_v_y456;
  double __pyx_v_y457;
  double __pyx_v_y458;
  double __pyx_v_y459;
  double __pyx_v_y460;
  double __pyx_v_y461;
  double __pyx_v_y462;
  double __pyx_v_y463;
  double __pyx_v_y464;
  double __pyx_v_y465;
  double __pyx_v_y466;
  double __pyx_v_y467;
  double __pyx_v_y468;
  double __pyx_v_y469;
  double __pyx_v_y470;
  double __pyx_v_y471;
  double __pyx_v_y472;
  double __pyx_v_y473;
  double __pyx_v_y474;
  double __pyx_v_y475;
  double __pyx_v_y476;
  double __pyx_v_H;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  double __pyx_v_dHeffdr;
  double __pyx_v_dHeffdphi;
  double __pyx_v_dHeffdpr;
  double __pyx_v_dHeffdpphi;
  double __pyx_v_d2Heffdr2;
  double __pyx_v_d2Heffdphi2;
  double __pyx_v_d2Heffdpr2;
  double __pyx_v_d2Heffdpphi2;
  double __pyx_v_d2Heffdrdphi;
  double __pyx_v_d2Heffdrdpr;
  double __pyx_v_d2Heffdrdpphi;
  double __pyx_v_d2Heffdphidpr;
  double __pyx_v_d2Heffdphidpphi;
  double __pyx_v_d2Heffdprdpphi;
  double __pyx_v_d2Hdr2;
  double __pyx_v_d2Hdphi2;
  double __pyx_v_d2Hdpr2;
  double __pyx_v_d2Hdpphi2;
  double __pyx_v_d2Hdrdphi;
  double __pyx_v_d2Hdrdpr;
  double __pyx_v_d2Hdrdpphi;
  double __pyx_v_d2Hdphidpr;
  double __pyx_v_d2Hdphidpphi;
  double __pyx_v_d2Hdprdpphi;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  double __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hessian", 0);
  __Pyx_TraceCall("hessian", __pyx_f[1], 468, 0, __PYX_ERR(1, 468, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_hessian); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_9hessian)) {
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 468, __pyx_L1_error) }
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_q, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 468, __pyx_L1_error) }
        __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_p, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 468, __pyx_L1_error) }
        __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_chi1_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 468, __pyx_L1_error) }
        __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_chi2_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_m_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyFloat_FromDouble(__pyx_v_m_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_chi_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyFloat_FromDouble(__pyx_v_chi_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_chiL1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = PyFloat_FromDouble(__pyx_v_chiL2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_13 = __pyx_t_1; __pyx_t_14 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 468, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 468, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(10+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 468, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_15, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_15, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_15, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_15, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_15, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_15, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_16, 8+__pyx_t_15, __pyx_t_11);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_16, 9+__pyx_t_15, __pyx_t_12);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_12 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 468, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":493
 *         # Coordinate definitions
 * 
 *         cdef double r = q[0]             # <<<<<<<<<<<<<<
 *         cdef double phi = q[1]
 * 
 */
  __Pyx_TraceLine(493,0,__PYX_ERR(1, 493, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_r = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":494
 * 
 *         cdef double r = q[0]
 *         cdef double phi = q[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double prst = p[0]
 */
  __Pyx_TraceLine(494,0,__PYX_ERR(1, 494, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_phi = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":496
 *         cdef double phi = q[1]
 * 
 *         cdef double prst = p[0]             # <<<<<<<<<<<<<<
 *         cdef double L = p[1]
 * 
 */
  __Pyx_TraceLine(496,0,__PYX_ERR(1, 496, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_prst = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":497
 * 
 *         cdef double prst = p[0]
 *         cdef double L = p[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix1 = chi1_v[0]
 */
  __Pyx_TraceLine(497,0,__PYX_ERR(1, 497, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_L = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":499
 *         cdef double L = p[1]
 * 
 *         cdef double chix1 = chi1_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]
 */
  __Pyx_TraceLine(499,0,__PYX_ERR(1, 499, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":500
 * 
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz1 = chi1_v[2]
 * 
 */
  __Pyx_TraceLine(500,0,__PYX_ERR(1, 500, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":501
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix2 = chi2_v[0]
 */
  __Pyx_TraceLine(501,0,__PYX_ERR(1, 501, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":503
 *         cdef double chiz1 = chi1_v[2]
 * 
 *         cdef double chix2 = chi2_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]
 */
  __Pyx_TraceLine(503,0,__PYX_ERR(1, 503, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":504
 * 
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz2 = chi2_v[2]
 * 
 */
  __Pyx_TraceLine(504,0,__PYX_ERR(1, 504, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":505
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double pphi = L
 */
  __Pyx_TraceLine(505,0,__PYX_ERR(1, 505, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":507
 *         cdef double chiz2 = chi2_v[2]
 * 
 *         cdef double pphi = L             # <<<<<<<<<<<<<<
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 */
  __Pyx_TraceLine(507,0,__PYX_ERR(1, 507, __pyx_L1_error))
  __pyx_v_pphi = __pyx_v_L;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":509
 *         cdef double pphi = L
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs             # <<<<<<<<<<<<<<
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']
 */
  __Pyx_TraceLine(509,0,__PYX_ERR(1, 509, __pyx_L1_error))
  __pyx_t_1 = ((PyObject *)__pyx_v_self->calibration_coeffs);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_c_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":510
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']             # <<<<<<<<<<<<<<
 *         cdef double dSO = c_coeffs['dSO']
 * 
 */
  __Pyx_TraceLine(510,0,__PYX_ERR(1, 510, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_a6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 510, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_a6 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":511
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(511,0,__PYX_ERR(1, 511, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_dSO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dSO = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":515
 * 
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M             # <<<<<<<<<<<<<<
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 */
  __Pyx_TraceLine(515,0,__PYX_ERR(1, 515, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->M;
  __pyx_v_M = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":516
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu             # <<<<<<<<<<<<<<
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 */
  __Pyx_TraceLine(516,0,__PYX_ERR(1, 516, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->nu;
  __pyx_v_nu = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":517
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1             # <<<<<<<<<<<<<<
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 */
  __Pyx_TraceLine(517,0,__PYX_ERR(1, 517, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_1;
  __pyx_v_X_1 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":518
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2             # <<<<<<<<<<<<<<
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3
 */
  __Pyx_TraceLine(518,0,__PYX_ERR(1, 518, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_2;
  __pyx_v_X_2 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":519
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta             # <<<<<<<<<<<<<<
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0
 */
  __Pyx_TraceLine(519,0,__PYX_ERR(1, 519, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->delta;
  __pyx_v_delta = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":520
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3             # <<<<<<<<<<<<<<
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 */
  __Pyx_TraceLine(520,0,__PYX_ERR(1, 520, __pyx_L1_error))
  __pyx_v_x0 = pow(__pyx_v_r, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":521
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0             # <<<<<<<<<<<<<<
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2
 */
  __Pyx_TraceLine(521,0,__PYX_ERR(1, 521, __pyx_L1_error))
  __pyx_v_x1 = (__pyx_v_r + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":522
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)             # <<<<<<<<<<<<<<
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2
 */
  __Pyx_TraceLine(522,0,__PYX_ERR(1, 522, __pyx_L1_error))
  __pyx_v_x2 = (__pyx_v_X_1 * ((((2.0 * __pyx_v_chix1) * __pyx_v_chix2) + ((2.0 * __pyx_v_chiy1) * __pyx_v_chiy2)) + ((2.0 * __pyx_v_chiz1) * __pyx_v_chiz2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":523
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2             # <<<<<<<<<<<<<<
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 */
  __Pyx_TraceLine(523,0,__PYX_ERR(1, 523, __pyx_L1_error))
  __pyx_v_x3 = (__pyx_v_X_2 * __pyx_v_x2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":524
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2             # <<<<<<<<<<<<<<
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5
 */
  __Pyx_TraceLine(524,0,__PYX_ERR(1, 524, __pyx_L1_error))
  __pyx_v_x4 = pow(__pyx_v_X_1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":525
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2             # <<<<<<<<<<<<<<
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 */
  __Pyx_TraceLine(525,0,__PYX_ERR(1, 525, __pyx_L1_error))
  __pyx_v_x5 = ((pow(__pyx_v_chix1, 2.0) + pow(__pyx_v_chiy1, 2.0)) + pow(__pyx_v_chiz1, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":526
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5             # <<<<<<<<<<<<<<
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7
 */
  __Pyx_TraceLine(526,0,__PYX_ERR(1, 526, __pyx_L1_error))
  __pyx_v_x6 = (__pyx_v_x4 * __pyx_v_x5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":527
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)             # <<<<<<<<<<<<<<
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8
 */
  __Pyx_TraceLine(527,0,__PYX_ERR(1, 527, __pyx_L1_error))
  __pyx_v_x7 = (pow(__pyx_v_X_2, 2.0) * ((pow(__pyx_v_chix2, 2.0) + pow(__pyx_v_chiy2, 2.0)) + pow(__pyx_v_chiz2, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":528
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7             # <<<<<<<<<<<<<<
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9
 */
  __Pyx_TraceLine(528,0,__PYX_ERR(1, 528, __pyx_L1_error))
  __pyx_v_x8 = (__pyx_v_x6 + __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":529
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8             # <<<<<<<<<<<<<<
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10
 */
  __Pyx_TraceLine(529,0,__PYX_ERR(1, 529, __pyx_L1_error))
  __pyx_v_x9 = (__pyx_v_x3 + __pyx_v_x8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":530
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9             # <<<<<<<<<<<<<<
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)
 */
  __Pyx_TraceLine(530,0,__PYX_ERR(1, 530, __pyx_L1_error))
  __pyx_v_x10 = (__pyx_v_x1 * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":531
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10             # <<<<<<<<<<<<<<
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4
 */
  __Pyx_TraceLine(531,0,__PYX_ERR(1, 531, __pyx_L1_error))
  __pyx_v_x11 = (__pyx_v_x0 + __pyx_v_x10);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":532
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)
 */
  __Pyx_TraceLine(532,0,__PYX_ERR(1, 532, __pyx_L1_error))
  __pyx_v_x12 = pow(__pyx_v_x11, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":533
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4             # <<<<<<<<<<<<<<
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu
 */
  __Pyx_TraceLine(533,0,__PYX_ERR(1, 533, __pyx_L1_error))
  __pyx_v_x13 = pow(__pyx_v_r, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":534
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1
 */
  __Pyx_TraceLine(534,0,__PYX_ERR(1, 534, __pyx_L1_error))
  __pyx_v_x14 = pow(__pyx_v_x13, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":535
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2
 */
  __Pyx_TraceLine(535,0,__PYX_ERR(1, 535, __pyx_L1_error))
  __pyx_v_x15 = (3.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":536
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1             # <<<<<<<<<<<<<<
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17
 */
  __Pyx_TraceLine(536,0,__PYX_ERR(1, 536, __pyx_L1_error))
  __pyx_v_x16 = (__pyx_v_X_1 * __pyx_v_chiL1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":537
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2             # <<<<<<<<<<<<<<
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18
 */
  __Pyx_TraceLine(537,0,__PYX_ERR(1, 537, __pyx_L1_error))
  __pyx_v_x17 = (__pyx_v_X_2 * __pyx_v_chiL2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":538
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17             # <<<<<<<<<<<<<<
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2
 */
  __Pyx_TraceLine(538,0,__PYX_ERR(1, 538, __pyx_L1_error))
  __pyx_v_x18 = (__pyx_v_x16 + __pyx_v_x17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":539
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18             # <<<<<<<<<<<<<<
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)
 */
  __Pyx_TraceLine(539,0,__PYX_ERR(1, 539, __pyx_L1_error))
  __pyx_v_x19 = (__pyx_v_pphi * __pyx_v_x18);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":540
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2             # <<<<<<<<<<<<<<
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375
 */
  __Pyx_TraceLine(540,0,__PYX_ERR(1, 540, __pyx_L1_error))
  __pyx_v_x20 = pow(__pyx_v_r, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":541
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875
 */
  __Pyx_TraceLine(541,0,__PYX_ERR(1, 541, __pyx_L1_error))
  __pyx_v_x21 = pow(__pyx_v_x20, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":542
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375             # <<<<<<<<<<<<<<
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)
 */
  __Pyx_TraceLine(542,0,__PYX_ERR(1, 542, __pyx_L1_error))
  __pyx_v_x22 = ((0.71875 * __pyx_v_nu) - 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":543
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875             # <<<<<<<<<<<<<<
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2
 */
  __Pyx_TraceLine(543,0,__PYX_ERR(1, 543, __pyx_L1_error))
  __pyx_v_x23 = ((-1.40625 * __pyx_v_nu) - 0.46875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":544
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25
 */
  __Pyx_TraceLine(544,0,__PYX_ERR(1, 544, __pyx_L1_error))
  __pyx_v_x24 = pow(__pyx_v_x0, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":545
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2             # <<<<<<<<<<<<<<
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26
 */
  __Pyx_TraceLine(545,0,__PYX_ERR(1, 545, __pyx_L1_error))
  __pyx_v_x25 = pow(__pyx_v_pphi, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":546
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25             # <<<<<<<<<<<<<<
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2
 */
  __Pyx_TraceLine(546,0,__PYX_ERR(1, 546, __pyx_L1_error))
  __pyx_v_x26 = (__pyx_v_x24 * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":547
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26             # <<<<<<<<<<<<<<
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 */
  __Pyx_TraceLine(547,0,__PYX_ERR(1, 547, __pyx_L1_error))
  __pyx_v_x27 = (2.0 * __pyx_v_x26);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":548
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2             # <<<<<<<<<<<<<<
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14
 */
  __Pyx_TraceLine(548,0,__PYX_ERR(1, 548, __pyx_L1_error))
  __pyx_v_x28 = pow(__pyx_v_nu, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":549
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30
 */
  __Pyx_TraceLine(549,0,__PYX_ERR(1, 549, __pyx_L1_error))
  __pyx_v_x29 = (((-2.0859375 * __pyx_v_nu) - (2.07161458333333 * __pyx_v_x28)) + 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":550
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14             # <<<<<<<<<<<<<<
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 */
  __Pyx_TraceLine(550,0,__PYX_ERR(1, 550, __pyx_L1_error))
  __pyx_v_x30 = (3.0 * __pyx_v_x14);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":551
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30             # <<<<<<<<<<<<<<
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4
 */
  __Pyx_TraceLine(551,0,__PYX_ERR(1, 551, __pyx_L1_error))
  __pyx_v_x31 = (__pyx_v_x25 * __pyx_v_x30);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":552
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5
 */
  __Pyx_TraceLine(552,0,__PYX_ERR(1, 552, __pyx_L1_error))
  __pyx_v_x32 = (((0.5859375 * __pyx_v_nu) + (1.34765625 * __pyx_v_x28)) + 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":553
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4             # <<<<<<<<<<<<<<
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)
 */
  __Pyx_TraceLine(553,0,__PYX_ERR(1, 553, __pyx_L1_error))
  __pyx_v_x33 = pow(__pyx_v_pphi, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":554
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5             # <<<<<<<<<<<<<<
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35
 */
  __Pyx_TraceLine(554,0,__PYX_ERR(1, 554, __pyx_L1_error))
  __pyx_v_x34 = pow(__pyx_v_r, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":555
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36
 */
  __Pyx_TraceLine(555,0,__PYX_ERR(1, 555, __pyx_L1_error))
  __pyx_v_x35 = pow(__pyx_v_x34, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":556
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35             # <<<<<<<<<<<<<<
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375
 */
  __Pyx_TraceLine(556,0,__PYX_ERR(1, 556, __pyx_L1_error))
  __pyx_v_x36 = (4.0 * __pyx_v_x35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":557
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36             # <<<<<<<<<<<<<<
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu
 */
  __Pyx_TraceLine(557,0,__PYX_ERR(1, 557, __pyx_L1_error))
  __pyx_v_x37 = (__pyx_v_x33 * __pyx_v_x36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":558
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375             # <<<<<<<<<<<<<<
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 */
  __Pyx_TraceLine(558,0,__PYX_ERR(1, 558, __pyx_L1_error))
  __pyx_v_x38 = ((0.34375 * __pyx_v_nu) + 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":559
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu             # <<<<<<<<<<<<<<
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 */
  __Pyx_TraceLine(559,0,__PYX_ERR(1, 559, __pyx_L1_error))
  __pyx_v_x39 = (0.46875 - (0.28125 * __pyx_v_nu));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":560
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17
 */
  __Pyx_TraceLine(560,0,__PYX_ERR(1, 560, __pyx_L1_error))
  __pyx_v_x40 = (((-0.2734375 * __pyx_v_nu) - (0.798177083333333 * __pyx_v_x28)) - 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":561
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42
 */
  __Pyx_TraceLine(561,0,__PYX_ERR(1, 561, __pyx_L1_error))
  __pyx_v_x41 = (((-0.3515625 * __pyx_v_nu) + (0.29296875 * __pyx_v_x28)) - 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":562
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17             # <<<<<<<<<<<<<<
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43
 */
  __Pyx_TraceLine(562,0,__PYX_ERR(1, 562, __pyx_L1_error))
  __pyx_v_x42 = (__pyx_v_x16 - __pyx_v_x17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":563
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42             # <<<<<<<<<<<<<<
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1
 */
  __Pyx_TraceLine(563,0,__PYX_ERR(1, 563, __pyx_L1_error))
  __pyx_v_x43 = (__pyx_v_pphi * __pyx_v_x42);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":564
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43             # <<<<<<<<<<<<<<
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2
 */
  __Pyx_TraceLine(564,0,__PYX_ERR(1, 564, __pyx_L1_error))
  __pyx_v_x44 = (__pyx_v_delta * __pyx_v_x43);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":565
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1             # <<<<<<<<<<<<<<
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46
 */
  __Pyx_TraceLine(565,0,__PYX_ERR(1, 565, __pyx_L1_error))
  __pyx_v_x45 = (__pyx_v_chi_1 * __pyx_v_X_1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":566
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2             # <<<<<<<<<<<<<<
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2
 */
  __Pyx_TraceLine(566,0,__PYX_ERR(1, 566, __pyx_L1_error))
  __pyx_v_x46 = (__pyx_v_chi_2 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":567
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46             # <<<<<<<<<<<<<<
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9
 */
  __Pyx_TraceLine(567,0,__PYX_ERR(1, 567, __pyx_L1_error))
  __pyx_v_x47 = (__pyx_v_x45 + __pyx_v_x46);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":568
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2             # <<<<<<<<<<<<<<
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25
 */
  __Pyx_TraceLine(568,0,__PYX_ERR(1, 568, __pyx_L1_error))
  __pyx_v_x48 = pow(__pyx_v_x47, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":569
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9             # <<<<<<<<<<<<<<
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9
 */
  __Pyx_TraceLine(569,0,__PYX_ERR(1, 569, __pyx_L1_error))
  __pyx_v_x49 = ((-__pyx_v_x48) + __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":570
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25             # <<<<<<<<<<<<<<
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 */
  __Pyx_TraceLine(570,0,__PYX_ERR(1, 570, __pyx_L1_error))
  __pyx_v_x50 = (__pyx_v_delta * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":571
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9             # <<<<<<<<<<<<<<
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46
 */
  __Pyx_TraceLine(571,0,__PYX_ERR(1, 571, __pyx_L1_error))
  __pyx_v_x51 = (__pyx_v_delta * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":572
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)             # <<<<<<<<<<<<<<
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 */
  __Pyx_TraceLine(572,0,__PYX_ERR(1, 572, __pyx_L1_error))
  __pyx_v_x52 = (__pyx_v_x49 * (0.416666666666667 - (0.833333333333333 * __pyx_v_X_2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":573
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46             # <<<<<<<<<<<<<<
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 */
  __Pyx_TraceLine(573,0,__PYX_ERR(1, 573, __pyx_L1_error))
  __pyx_v_x53 = (__pyx_v_x45 - __pyx_v_x46);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":574
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7             # <<<<<<<<<<<<<<
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7
 */
  __Pyx_TraceLine(574,0,__PYX_ERR(1, 574, __pyx_L1_error))
  __pyx_v_x54 = (((__pyx_v_x4 * __pyx_v_x5) - (__pyx_v_x47 * __pyx_v_x53)) - __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":575
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7             # <<<<<<<<<<<<<<
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56
 */
  __Pyx_TraceLine(575,0,__PYX_ERR(1, 575, __pyx_L1_error))
  __pyx_v_x55 = (((((-0.125 * __pyx_v_x3) + (0.125 * __pyx_v_x48)) + ((0.5 * __pyx_v_x54) * (0.5 - __pyx_v_X_2))) - (0.125 * __pyx_v_x6)) - (0.125 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":576
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7             # <<<<<<<<<<<<<<
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54
 */
  __Pyx_TraceLine(576,0,__PYX_ERR(1, 576, __pyx_L1_error))
  __pyx_v_x56 = (__pyx_v_x6 - __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":577
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56             # <<<<<<<<<<<<<<
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24
 */
  __Pyx_TraceLine(577,0,__PYX_ERR(1, 577, __pyx_L1_error))
  __pyx_v_x57 = (__pyx_v_delta * __pyx_v_x56);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":578
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54             # <<<<<<<<<<<<<<
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)
 */
  __Pyx_TraceLine(578,0,__PYX_ERR(1, 578, __pyx_L1_error))
  __pyx_v_x58 = (__pyx_v_delta * __pyx_v_x54);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":579
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24             # <<<<<<<<<<<<<<
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25
 */
  __Pyx_TraceLine(579,0,__PYX_ERR(1, 579, __pyx_L1_error))
  __pyx_v_x59 = ((__pyx_v_nu * __pyx_v_dSO) * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":580
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33
 */
  __Pyx_TraceLine(580,0,__PYX_ERR(1, 580, __pyx_L1_error))
  __pyx_v_x60 = pow(__pyx_v_r, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":581
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25             # <<<<<<<<<<<<<<
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 */
  __Pyx_TraceLine(581,0,__PYX_ERR(1, 581, __pyx_L1_error))
  __pyx_v_x61 = (__pyx_v_x21 * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":582
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33             # <<<<<<<<<<<<<<
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 */
  __Pyx_TraceLine(582,0,__PYX_ERR(1, 582, __pyx_L1_error))
  __pyx_v_x62 = (__pyx_v_x14 * __pyx_v_x33);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":583
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)             # <<<<<<<<<<<<<<
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49
 */
  __Pyx_TraceLine(583,0,__PYX_ERR(1, 583, __pyx_L1_error))
  __pyx_v_x63 = (__pyx_v_x18 * ((((((__pyx_v_x21 * (((-5.53125 * __pyx_v_nu) + (0.567708333333333 * __pyx_v_x28)) - 0.078125)) + (__pyx_v_x22 * __pyx_v_x60)) + (__pyx_v_x23 * __pyx_v_x61)) + (__pyx_v_x26 * __pyx_v_x29)) + (__pyx_v_x32 * __pyx_v_x62)) + 1.75));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":584
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)             # <<<<<<<<<<<<<<
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65
 */
  __Pyx_TraceLine(584,0,__PYX_ERR(1, 584, __pyx_L1_error))
  __pyx_v_x64 = (__pyx_v_delta * ((((((__pyx_v_x21 * (((-0.03125 * __pyx_v_nu) + (0.536458333333333 * __pyx_v_x28)) + 0.078125)) + (__pyx_v_x26 * __pyx_v_x40)) + (__pyx_v_x38 * __pyx_v_x60)) + (__pyx_v_x39 * __pyx_v_x61)) + (__pyx_v_x41 * __pyx_v_x62)) + 0.25));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":585
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49             # <<<<<<<<<<<<<<
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 */
  __Pyx_TraceLine(585,0,__PYX_ERR(1, 585, __pyx_L1_error))
  __pyx_v_x65 = (__pyx_v_x24 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":586
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65             # <<<<<<<<<<<<<<
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55
 */
  __Pyx_TraceLine(586,0,__PYX_ERR(1, 586, __pyx_L1_error))
  __pyx_v_x66 = (__pyx_v_x50 * __pyx_v_x65);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":587
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)             # <<<<<<<<<<<<<<
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 */
  __Pyx_TraceLine(587,0,__PYX_ERR(1, 587, __pyx_L1_error))
  __pyx_v_x67 = (__pyx_v_x42 * ((__pyx_v_x21 * ((0.0416666666666667 * __pyx_v_x51) + (0.5 * __pyx_v_x52))) - (0.125 * __pyx_v_x66)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":588
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55             # <<<<<<<<<<<<<<
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4
 */
  __Pyx_TraceLine(588,0,__PYX_ERR(1, 588, __pyx_L1_error))
  __pyx_v_x68 = (__pyx_v_x26 * __pyx_v_x55);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":589
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)             # <<<<<<<<<<<<<<
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)
 */
  __Pyx_TraceLine(589,0,__PYX_ERR(1, 589, __pyx_L1_error))
  __pyx_v_x69 = (__pyx_v_x18 * ((__pyx_v_x21 * (((((((0.208333333333333 * __pyx_v_delta) * __pyx_v_x56) - (0.625 * __pyx_v_x3)) + (0.375 * __pyx_v_x48)) - (0.833333333333333 * __pyx_v_x58)) - (0.625 * __pyx_v_x6)) - (0.625 * __pyx_v_x7))) + __pyx_v_x68));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":590
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4             # <<<<<<<<<<<<<<
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71
 */
  __Pyx_TraceLine(590,0,__PYX_ERR(1, 590, __pyx_L1_error))
  __pyx_v_x70 = pow(__pyx_v_prst, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":591
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)             # <<<<<<<<<<<<<<
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6
 */
  __Pyx_TraceLine(591,0,__PYX_ERR(1, 591, __pyx_L1_error))
  __pyx_v_x71 = pow(__pyx_v_r, -4.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":592
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71             # <<<<<<<<<<<<<<
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)
 */
  __Pyx_TraceLine(592,0,__PYX_ERR(1, 592, __pyx_L1_error))
  __pyx_v_x72 = (__pyx_v_nu * __pyx_v_x71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":593
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6             # <<<<<<<<<<<<<<
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74
 */
  __Pyx_TraceLine(593,0,__PYX_ERR(1, 593, __pyx_L1_error))
  __pyx_v_x73 = pow(__pyx_v_prst, 6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":594
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)             # <<<<<<<<<<<<<<
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)
 */
  __Pyx_TraceLine(594,0,__PYX_ERR(1, 594, __pyx_L1_error))
  __pyx_v_x74 = pow(__pyx_v_r, -3.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":595
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74             # <<<<<<<<<<<<<<
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8
 */
  __Pyx_TraceLine(595,0,__PYX_ERR(1, 595, __pyx_L1_error))
  __pyx_v_x75 = (__pyx_v_nu * __pyx_v_x74);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":596
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)             # <<<<<<<<<<<<<<
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77
 */
  __Pyx_TraceLine(596,0,__PYX_ERR(1, 596, __pyx_L1_error))
  __pyx_v_x76 = pow(__pyx_v_r, -2.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":597
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8             # <<<<<<<<<<<<<<
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60
 */
  __Pyx_TraceLine(597,0,__PYX_ERR(1, 597, __pyx_L1_error))
  __pyx_v_x77 = pow(__pyx_v_prst, 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":598
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77             # <<<<<<<<<<<<<<
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77
 */
  __Pyx_TraceLine(598,0,__PYX_ERR(1, 598, __pyx_L1_error))
  __pyx_v_x78 = (__pyx_v_nu * __pyx_v_x77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":599
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60             # <<<<<<<<<<<<<<
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28
 */
  __Pyx_TraceLine(599,0,__PYX_ERR(1, 599, __pyx_L1_error))
  __pyx_v_x79 = (__pyx_v_nu * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":600
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77             # <<<<<<<<<<<<<<
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81
 */
  __Pyx_TraceLine(600,0,__PYX_ERR(1, 600, __pyx_L1_error))
  __pyx_v_x80 = (0.121954868780449 * __pyx_v_x77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":601
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3
 */
  __Pyx_TraceLine(601,0,__PYX_ERR(1, 601, __pyx_L1_error))
  __pyx_v_x81 = ((8.0 * __pyx_v_nu) - (6.0 * __pyx_v_x28));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":602
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81             # <<<<<<<<<<<<<<
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 */
  __Pyx_TraceLine(602,0,__PYX_ERR(1, 602, __pyx_L1_error))
  __pyx_v_x82 = (__pyx_v_x21 * __pyx_v_x81);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":603
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3             # <<<<<<<<<<<<<<
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70
 */
  __Pyx_TraceLine(603,0,__PYX_ERR(1, 603, __pyx_L1_error))
  __pyx_v_x83 = pow(__pyx_v_nu, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":604
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83             # <<<<<<<<<<<<<<
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 */
  __Pyx_TraceLine(604,0,__PYX_ERR(1, 604, __pyx_L1_error))
  __pyx_v_x84 = (((92.7110442849544 * __pyx_v_nu) - (131.0 * __pyx_v_x28)) + (10.0 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":605
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70             # <<<<<<<<<<<<<<
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86
 */
  __Pyx_TraceLine(605,0,__PYX_ERR(1, 605, __pyx_L1_error))
  __pyx_v_x85 = (__pyx_v_x24 * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":606
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83             # <<<<<<<<<<<<<<
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4
 */
  __Pyx_TraceLine(606,0,__PYX_ERR(1, 606, __pyx_L1_error))
  __pyx_v_x86 = (((-2.78300763695006 * __pyx_v_nu) - (5.4 * __pyx_v_x28)) + (6.0 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":607
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86             # <<<<<<<<<<<<<<
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 */
  __Pyx_TraceLine(607,0,__PYX_ERR(1, 607, __pyx_L1_error))
  __pyx_v_x87 = (__pyx_v_x21 * __pyx_v_x86);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":608
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4             # <<<<<<<<<<<<<<
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73
 */
  __Pyx_TraceLine(608,0,__PYX_ERR(1, 608, __pyx_L1_error))
  __pyx_v_x88 = pow(__pyx_v_nu, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":609
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88             # <<<<<<<<<<<<<<
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 */
  __Pyx_TraceLine(609,0,__PYX_ERR(1, 609, __pyx_L1_error))
  __pyx_v_x89 = ((((-33.9782122170436 * __pyx_v_nu) - (89.5298327361234 * __pyx_v_x28)) + (188.0 * __pyx_v_x83)) - (14.0 * __pyx_v_x88));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":610
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73             # <<<<<<<<<<<<<<
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91
 */
  __Pyx_TraceLine(610,0,__PYX_ERR(1, 610, __pyx_L1_error))
  __pyx_v_x90 = (__pyx_v_x24 * __pyx_v_x73);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":611
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88             # <<<<<<<<<<<<<<
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)
 */
  __Pyx_TraceLine(611,0,__PYX_ERR(1, 611, __pyx_L1_error))
  __pyx_v_x91 = ((((1.38977750996128 * __pyx_v_nu) + (3.33842023648322 * __pyx_v_x28)) + (3.42857142857143 * __pyx_v_x83)) - (6.0 * __pyx_v_x88));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":612
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91             # <<<<<<<<<<<<<<
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 */
  __Pyx_TraceLine(612,0,__PYX_ERR(1, 612, __pyx_L1_error))
  __pyx_v_x92 = (__pyx_v_x77 * __pyx_v_x91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":613
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)             # <<<<<<<<<<<<<<
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70
 */
  __Pyx_TraceLine(613,0,__PYX_ERR(1, 613, __pyx_L1_error))
  __pyx_v_x93 = log(__pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":614
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83             # <<<<<<<<<<<<<<
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13
 */
  __Pyx_TraceLine(614,0,__PYX_ERR(1, 614, __pyx_L1_error))
  __pyx_v_x94 = (((__pyx_v_nu * (452.542166996693 - (51.6952380952381 * __pyx_v_x93))) + (__pyx_v_x28 * ((118.4 * __pyx_v_x93) - 1796.13660498019))) + (602.318540416564 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":615
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70             # <<<<<<<<<<<<<<
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)
 */
  __Pyx_TraceLine(615,0,__PYX_ERR(1, 615, __pyx_L1_error))
  __pyx_v_x95 = (__pyx_v_x14 * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":616
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13             # <<<<<<<<<<<<<<
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2
 */
  __Pyx_TraceLine(616,0,__PYX_ERR(1, 616, __pyx_L1_error))
  __pyx_v_x96 = ((__pyx_v_r * __pyx_v_x10) + __pyx_v_x13);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":617
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98
 */
  __Pyx_TraceLine(617,0,__PYX_ERR(1, 617, __pyx_L1_error))
  __pyx_v_x97 = pow(__pyx_v_x96, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":618
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2             # <<<<<<<<<<<<<<
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99
 */
  __Pyx_TraceLine(618,0,__PYX_ERR(1, 618, __pyx_L1_error))
  __pyx_v_x98 = pow(__pyx_v_x18, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":619
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98             # <<<<<<<<<<<<<<
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49
 */
  __Pyx_TraceLine(619,0,__PYX_ERR(1, 619, __pyx_L1_error))
  __pyx_v_x99 = (__pyx_v_x97 * __pyx_v_x98);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":620
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99             # <<<<<<<<<<<<<<
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54
 */
  __Pyx_TraceLine(620,0,__PYX_ERR(1, 620, __pyx_L1_error))
  __pyx_v_x100 = ((__pyx_v_x25 * __pyx_v_x60) * __pyx_v_x99);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":621
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49             # <<<<<<<<<<<<<<
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8
 */
  __Pyx_TraceLine(621,0,__PYX_ERR(1, 621, __pyx_L1_error))
  __pyx_v_x101 = (0.5 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":622
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54             # <<<<<<<<<<<<<<
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2
 */
  __Pyx_TraceLine(622,0,__PYX_ERR(1, 622, __pyx_L1_error))
  __pyx_v_x102 = (0.5 * __pyx_v_x54);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":623
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8             # <<<<<<<<<<<<<<
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104
 */
  __Pyx_TraceLine(623,0,__PYX_ERR(1, 623, __pyx_L1_error))
  __pyx_v_x103 = ((-__pyx_v_x3) + __pyx_v_x8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":624
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2             # <<<<<<<<<<<<<<
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 */
  __Pyx_TraceLine(624,0,__PYX_ERR(1, 624, __pyx_L1_error))
  __pyx_v_x104 = (__pyx_v_x103 - pow(__pyx_v_x53, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":625
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104             # <<<<<<<<<<<<<<
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28
 */
  __Pyx_TraceLine(625,0,__PYX_ERR(1, 625, __pyx_L1_error))
  __pyx_v_x105 = (0.5 * __pyx_v_x104);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":626
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)             # <<<<<<<<<<<<<<
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 */
  __Pyx_TraceLine(626,0,__PYX_ERR(1, 626, __pyx_L1_error))
  __pyx_v_x106 = (__pyx_v_x58 * ((98.0 * __pyx_v_nu) + 43.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":627
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49
 */
  __Pyx_TraceLine(627,0,__PYX_ERR(1, 627, __pyx_L1_error))
  __pyx_v_x107 = (264.0 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":628
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0             # <<<<<<<<<<<<<<
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 */
  __Pyx_TraceLine(628,0,__PYX_ERR(1, 628, __pyx_L1_error))
  __pyx_v_x108 = (((-1610.0 * __pyx_v_nu) + __pyx_v_x107) + 375.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":629
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49             # <<<<<<<<<<<<<<
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104
 */
  __Pyx_TraceLine(629,0,__PYX_ERR(1, 629, __pyx_L1_error))
  __pyx_v_x109 = (0.00260416666666667 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":630
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0             # <<<<<<<<<<<<<<
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 */
  __Pyx_TraceLine(630,0,__PYX_ERR(1, 630, __pyx_L1_error))
  __pyx_v_x110 = (((-1090.0 * __pyx_v_nu) + (152.0 * __pyx_v_x28)) + 219.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":631
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104             # <<<<<<<<<<<<<<
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)
 */
  __Pyx_TraceLine(631,0,__PYX_ERR(1, 631, __pyx_L1_error))
  __pyx_v_x111 = (0.0078125 * __pyx_v_x104);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":632
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9
 */
  __Pyx_TraceLine(632,0,__PYX_ERR(1, 632, __pyx_L1_error))
  __pyx_v_x112 = (((((-__pyx_v_x101) * __pyx_v_x21) + (__pyx_v_x14 * (((-0.0625 * __pyx_v_x106) + (__pyx_v_x108 * __pyx_v_x109)) + (__pyx_v_x110 * __pyx_v_x111)))) + (__pyx_v_x24 * (((__pyx_v_x101 * ((-1.75 * __pyx_v_nu) - 0.9375)) + (__pyx_v_x102 * (0.75 - (1.5 * __pyx_v_X_2)))) + (__pyx_v_x105 * (0.1875 - (0.75 * __pyx_v_nu)))))) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":633
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)             # <<<<<<<<<<<<<<
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4
 */
  __Pyx_TraceLine(633,0,__PYX_ERR(1, 633, __pyx_L1_error))
  __pyx_v_x113 = pow(__pyx_v_r, -13.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":634
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9             # <<<<<<<<<<<<<<
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9
 */
  __Pyx_TraceLine(634,0,__PYX_ERR(1, 634, __pyx_L1_error))
  __pyx_v_x114 = (__pyx_v_x20 + __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":635
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4             # <<<<<<<<<<<<<<
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu
 */
  __Pyx_TraceLine(635,0,__PYX_ERR(1, 635, __pyx_L1_error))
  __pyx_v_x115 = pow(__pyx_v_x114, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":636
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9             # <<<<<<<<<<<<<<
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0
 */
  __Pyx_TraceLine(636,0,__PYX_ERR(1, 636, __pyx_L1_error))
  __pyx_v_x116 = (__pyx_v_x21 * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":637
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 */
  __Pyx_TraceLine(637,0,__PYX_ERR(1, 637, __pyx_L1_error))
  __pyx_v_x117 = (756.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":638
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0             # <<<<<<<<<<<<<<
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2
 */
  __Pyx_TraceLine(638,0,__PYX_ERR(1, 638, __pyx_L1_error))
  __pyx_v_x118 = (((336.0 * __pyx_v_r) + __pyx_v_x117) + 407.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":639
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0             # <<<<<<<<<<<<<<
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28
 */
  __Pyx_TraceLine(639,0,__PYX_ERR(1, 639, __pyx_L1_error))
  __pyx_v_x119 = (((((((2048.0 * __pyx_v_nu) * __pyx_v_x118) * __pyx_v_x93) + ((28.0 * __pyx_v_nu) * ((1920.0 * __pyx_v_a6) + 733955.307463037))) - ((7.0 * __pyx_v_r) * (((938918.400156317 * __pyx_v_nu) - (185763.092693281 * __pyx_v_x28)) - 245760.0))) - (5416406.59541186 * __pyx_v_x28)) - 3440640.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":640
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2             # <<<<<<<<<<<<<<
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 */
  __Pyx_TraceLine(640,0,__PYX_ERR(1, 640, __pyx_L1_error))
  __pyx_v_x120 = pow(__pyx_v_x93, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":641
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28             # <<<<<<<<<<<<<<
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 */
  __Pyx_TraceLine(641,0,__PYX_ERR(1, 641, __pyx_L1_error))
  __pyx_v_x121 = (__pyx_v_x120 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":642
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)             # <<<<<<<<<<<<<<
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0
 */
  __Pyx_TraceLine(642,0,__PYX_ERR(1, 642, __pyx_L1_error))
  __pyx_v_x122 = (__pyx_v_x83 * (((-163683964.822551 * __pyx_v_r) - (17833256.898555 * __pyx_v_x20)) - 1188987459.03162));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":643
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)             # <<<<<<<<<<<<<<
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0
 */
  __Pyx_TraceLine(643,0,__PYX_ERR(1, 643, __pyx_L1_error))
  __pyx_v_x123 = (__pyx_v_x28 * (((((((((-39321600.0 * __pyx_v_a6) * ((3.0 * __pyx_v_r) + 59.0)) + (745857848.115604 * __pyx_v_a6)) + (122635399361.987 * __pyx_v_r)) - (6178501407.53758 * __pyx_v_x0)) - (3089250703.76879 * __pyx_v_x13)) + (2064783811.32587 * __pyx_v_x20)) + (1426660551.8844 * __pyx_v_x34)) + 276057889687.011));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":644
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125
 */
  __Pyx_TraceLine(644,0,__PYX_ERR(1, 644, __pyx_L1_error))
  __pyx_v_x124 = ((588.0 * __pyx_v_nu) + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":645
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 */
  __Pyx_TraceLine(645,0,__PYX_ERR(1, 645, __pyx_L1_error))
  __pyx_v_x125 = (__pyx_v_x117 + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":646
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125             # <<<<<<<<<<<<<<
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93
 */
  __Pyx_TraceLine(646,0,__PYX_ERR(1, 646, __pyx_L1_error))
  __pyx_v_x126 = (__pyx_v_x0 * __pyx_v_x125);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":647
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0             # <<<<<<<<<<<<<<
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128
 */
  __Pyx_TraceLine(647,0,__PYX_ERR(1, 647, __pyx_L1_error))
  __pyx_v_x127 = ((((((((-38842241.4769507 * __pyx_v_nu) + ((240.0 * __pyx_v_r) * (((-7466.27061066206 * __pyx_v_nu) - (3024.0 * __pyx_v_x28)) + 17264.0))) + ((1920.0 * __pyx_v_x124) * __pyx_v_x20)) + ((480.0 * __pyx_v_x125) * __pyx_v_x13)) + (960.0 * __pyx_v_x126)) - (1882456.23663972 * __pyx_v_x28)) + (161280.0 * __pyx_v_x34)) + 13447680.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":648
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93             # <<<<<<<<<<<<<<
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r
 */
  __Pyx_TraceLine(648,0,__PYX_ERR(1, 648, __pyx_L1_error))
  __pyx_v_x128 = (__pyx_v_nu * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":649
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128             # <<<<<<<<<<<<<<
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 */
  __Pyx_TraceLine(649,0,__PYX_ERR(1, 649, __pyx_L1_error))
  __pyx_v_x129 = (__pyx_v_x127 * __pyx_v_x128);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":650
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r             # <<<<<<<<<<<<<<
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6
 */
  __Pyx_TraceLine(650,0,__PYX_ERR(1, 650, __pyx_L1_error))
  __pyx_v_x130 = (8.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":651
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0             # <<<<<<<<<<<<<<
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r
 */
  __Pyx_TraceLine(651,0,__PYX_ERR(1, 651, __pyx_L1_error))
  __pyx_v_x131 = ((((2.0 * __pyx_v_x0) + __pyx_v_x130) + (4.0 * __pyx_v_x20)) + 16.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":652
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6             # <<<<<<<<<<<<<<
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r
 */
  __Pyx_TraceLine(652,0,__PYX_ERR(1, 652, __pyx_L1_error))
  __pyx_v_x132 = (7680.0 * __pyx_v_a6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":653
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r             # <<<<<<<<<<<<<<
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r
 */
  __Pyx_TraceLine(653,0,__PYX_ERR(1, 653, __pyx_L1_error))
  __pyx_v_x133 = (128.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":654
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r             # <<<<<<<<<<<<<<
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r
 */
  __Pyx_TraceLine(654,0,__PYX_ERR(1, 654, __pyx_L1_error))
  __pyx_v_x134 = (7704.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":655
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r             # <<<<<<<<<<<<<<
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 */
  __Pyx_TraceLine(655,0,__PYX_ERR(1, 655, __pyx_L1_error))
  __pyx_v_x135 = (148.04406601634 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":656
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r             # <<<<<<<<<<<<<<
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 */
  __Pyx_TraceLine(656,0,__PYX_ERR(1, 656, __pyx_L1_error))
  __pyx_v_x136 = (113485.217444961 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":657
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))             # <<<<<<<<<<<<<<
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138
 */
  __Pyx_TraceLine(657,0,__PYX_ERR(1, 657, __pyx_L1_error))
  __pyx_v_x137 = (__pyx_v_nu * ((((__pyx_v_x132 * (__pyx_v_x13 + __pyx_v_x131)) + (__pyx_v_x133 * (((((13218.7851094412 * __pyx_v_r) + (4264.6962773603 * __pyx_v_x0)) - (6852.34813868015 * __pyx_v_x13)) + (8529.39255472061 * __pyx_v_x20)) - 33722.4297811176))) + (__pyx_v_x135 * (((((1926.0 * __pyx_v_x0) + (349.0 * __pyx_v_x13)) + __pyx_v_x134) + (3852.0 * __pyx_v_x20)) + 36400.0))) + (__pyx_v_x136 * ((-__pyx_v_x13) + __pyx_v_x131))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":658
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139
 */
  __Pyx_TraceLine(658,0,__PYX_ERR(1, 658, __pyx_L1_error))
  __pyx_v_x138 = pow((((((((67645734912.0 * __pyx_v_x121) + (1120.0 * __pyx_v_x122)) + (7.0 * __pyx_v_x123)) + (32768.0 * __pyx_v_x129)) + (53760.0 * __pyx_v_x137)) + (13212057600.0 * __pyx_v_x34)) + (241555486248.807 * __pyx_v_x88)), -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":659
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138             # <<<<<<<<<<<<<<
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 */
  __Pyx_TraceLine(659,0,__PYX_ERR(1, 659, __pyx_L1_error))
  __pyx_v_x139 = (__pyx_v_x13 * __pyx_v_x138);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":660
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139             # <<<<<<<<<<<<<<
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)
 */
  __Pyx_TraceLine(660,0,__PYX_ERR(1, 660, __pyx_L1_error))
  __pyx_v_x140 = (__pyx_v_x119 * __pyx_v_x139);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":661
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140             # <<<<<<<<<<<<<<
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120
 */
  __Pyx_TraceLine(661,0,__PYX_ERR(1, 661, __pyx_L1_error))
  __pyx_v_x141 = ((0.000130208333333333 * __pyx_v_x116) + __pyx_v_x140);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":662
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)             # <<<<<<<<<<<<<<
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 */
  __Pyx_TraceLine(662,0,__PYX_ERR(1, 662, __pyx_L1_error))
  __pyx_v_x142 = pow(__pyx_v_x141, -4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":663
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120             # <<<<<<<<<<<<<<
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20
 */
  __Pyx_TraceLine(663,0,__PYX_ERR(1, 663, __pyx_L1_error))
  __pyx_v_x143 = (__pyx_v_r * __pyx_v_x120);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":664
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913             # <<<<<<<<<<<<<<
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 */
  __Pyx_TraceLine(664,0,__PYX_ERR(1, 664, __pyx_L1_error))
  __pyx_v_x144 = (((-630116198.873299 * __pyx_v_nu) - (197773496.793534 * __pyx_v_x28)) + 5805304367.87913);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":665
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20             # <<<<<<<<<<<<<<
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146
 */
  __Pyx_TraceLine(665,0,__PYX_ERR(1, 665, __pyx_L1_error))
  __pyx_v_x145 = (__pyx_v_x144 * __pyx_v_x20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":666
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329             # <<<<<<<<<<<<<<
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 */
  __Pyx_TraceLine(666,0,__PYX_ERR(1, 666, __pyx_L1_error))
  __pyx_v_x146 = (((-2675575.66847905 * __pyx_v_nu) - (138240.0 * __pyx_v_x28)) - 5278341.3229329);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":667
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146             # <<<<<<<<<<<<<<
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 */
  __Pyx_TraceLine(667,0,__PYX_ERR(1, 667, __pyx_L1_error))
  __pyx_v_x147 = (__pyx_v_x0 * __pyx_v_x146);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":668
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)             # <<<<<<<<<<<<<<
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149
 */
  __Pyx_TraceLine(668,0,__PYX_ERR(1, 668, __pyx_L1_error))
  __pyx_v_x148 = (__pyx_v_nu * ((((-2510664218.28128 * __pyx_v_nu) - (42636451.6032331 * __pyx_v_x28)) + (14515200.0 * __pyx_v_x83)) + 1002013764.01019));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":669
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0             # <<<<<<<<<<<<<<
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0
 */
  __Pyx_TraceLine(669,0,__PYX_ERR(1, 669, __pyx_L1_error))
  __pyx_v_x149 = (((((43393301259014.8 * __pyx_v_nu) + (43133561885859.3 * __pyx_v_x28)) + (5927865218923.02 * __pyx_v_x83)) + (86618264430493.3 * pow((1.0 - (0.496948781616935 * __pyx_v_nu)), 2.0))) + 188440788778196.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":670
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149             # <<<<<<<<<<<<<<
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151
 */
  __Pyx_TraceLine(670,0,__PYX_ERR(1, 670, __pyx_L1_error))
  __pyx_v_x150 = (__pyx_v_r * __pyx_v_x149);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":671
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0             # <<<<<<<<<<<<<<
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 */
  __Pyx_TraceLine(671,0,__PYX_ERR(1, 671, __pyx_L1_error))
  __pyx_v_x151 = ((14700.0 * __pyx_v_nu) + 42911.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":672
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151             # <<<<<<<<<<<<<<
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93
 */
  __Pyx_TraceLine(672,0,__PYX_ERR(1, 672, __pyx_L1_error))
  __pyx_v_x152 = (283115520.0 * __pyx_v_x151);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":673
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20             # <<<<<<<<<<<<<<
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 */
  __Pyx_TraceLine(673,0,__PYX_ERR(1, 673, __pyx_L1_error))
  __pyx_v_x153 = (((((-1698693120.0 * __pyx_v_nu) * ((11592.0 * __pyx_v_nu) + 69847.0)) + ((49152.0 * __pyx_v_r) * (((409207698.136075 * __pyx_v_nu) + (102574080.0 * __pyx_v_x28)) - 2119671837.36038))) + (879923036160.0 * __pyx_v_x0)) + (__pyx_v_x152 * __pyx_v_x20));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":674
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93             # <<<<<<<<<<<<<<
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r
 */
  __Pyx_TraceLine(674,0,__PYX_ERR(1, 674, __pyx_L1_error))
  __pyx_v_x154 = (__pyx_v_x153 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":675
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2             # <<<<<<<<<<<<<<
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20
 */
  __Pyx_TraceLine(675,0,__PYX_ERR(1, 675, __pyx_L1_error))
  __pyx_v_x155 = pow((((((__pyx_v_x143 - (1.59227685093395e-9 * __pyx_v_x145)) - (1.67189069348064e-7 * __pyx_v_x147)) + (9.55366110560367e-9 * __pyx_v_x148)) + (1.72773095804465e-13 * __pyx_v_x150)) + (1.72773095804465e-13 * __pyx_v_x154)), 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":676
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r             # <<<<<<<<<<<<<<
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28
 */
  __Pyx_TraceLine(676,0,__PYX_ERR(1, 676, __pyx_L1_error))
  __pyx_v_x156 = (__pyx_v_nu * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":677
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20             # <<<<<<<<<<<<<<
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151
 */
  __Pyx_TraceLine(677,0,__PYX_ERR(1, 677, __pyx_L1_error))
  __pyx_v_x157 = (__pyx_v_nu * __pyx_v_x20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":678
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28             # <<<<<<<<<<<<<<
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28
 */
  __Pyx_TraceLine(678,0,__PYX_ERR(1, 678, __pyx_L1_error))
  __pyx_v_x158 = (__pyx_v_r * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":679
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151             # <<<<<<<<<<<<<<
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 */
  __Pyx_TraceLine(679,0,__PYX_ERR(1, 679, __pyx_L1_error))
  __pyx_v_x159 = (__pyx_v_r * __pyx_v_x151);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":680
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28             # <<<<<<<<<<<<<<
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 */
  __Pyx_TraceLine(680,0,__PYX_ERR(1, 680, __pyx_L1_error))
  __pyx_v_x160 = (__pyx_v_x20 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":681
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0             # <<<<<<<<<<<<<<
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93
 */
  __Pyx_TraceLine(681,0,__PYX_ERR(1, 681, __pyx_L1_error))
  __pyx_v_x161 = ((5041721180160.0 * __pyx_v_x28) - 104186110149937.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":682
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20             # <<<<<<<<<<<<<<
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 */
  __Pyx_TraceLine(682,0,__PYX_ERR(1, 682, __pyx_L1_error))
  __pyx_v_x162 = ((((-25392914995744.3 * __pyx_v_nu) - (__pyx_v_r * __pyx_v_x152)) - __pyx_v_x161) - (879923036160.0 * __pyx_v_x20));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":683
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93             # <<<<<<<<<<<<<<
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)
 */
  __Pyx_TraceLine(683,0,__PYX_ERR(1, 683, __pyx_L1_error))
  __pyx_v_x163 = (__pyx_v_x162 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":684
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1             # <<<<<<<<<<<<<<
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28
 */
  __Pyx_TraceLine(684,0,__PYX_ERR(1, 684, __pyx_L1_error))
  __pyx_v_x164 = (((((((((((((-0.0438084424460039 * __pyx_v_nu) - (0.143521050466841 * __pyx_v_r)) + (0.0210425293255724 * __pyx_v_x120)) + (0.0385795738434214 * __pyx_v_x156)) + (0.00941289164152486 * __pyx_v_x157)) + (0.00662650629087394 * __pyx_v_x158)) - (1.18824456940711e-6 * __pyx_v_x159)) + (0.000486339502879429 * __pyx_v_x160)) - (3.63558293513537e-15 * __pyx_v_x163)) + (0.0185696317637669 * __pyx_v_x20)) + (0.291062041428379 * __pyx_v_x28)) + (0.0244692826489756 * __pyx_v_x83)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":685
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)             # <<<<<<<<<<<<<<
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 */
  __Pyx_TraceLine(685,0,__PYX_ERR(1, 685, __pyx_L1_error))
  __pyx_v_x165 = pow(__pyx_v_x164, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":686
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 */
  __Pyx_TraceLine(686,0,__PYX_ERR(1, 686, __pyx_L1_error))
  __pyx_v_x166 = (4.0 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":687
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)             # <<<<<<<<<<<<<<
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2
 */
  __Pyx_TraceLine(687,0,__PYX_ERR(1, 687, __pyx_L1_error))
  __pyx_v_x167 = (((((0.3125 * __pyx_v_delta) * __pyx_v_x56) * ((18.0 * __pyx_v_nu) - 1.0)) - ((0.03125 * (((-33.0 * __pyx_v_nu) + (32.0 * __pyx_v_x28)) - 5.0)) * (((5.0 * __pyx_v_x3) + (5.0 * __pyx_v_x6)) + (5.0 * __pyx_v_x7)))) - ((0.03125 * (((-5.0 * __pyx_v_nu) + __pyx_v_x166) + 1.0)) * (((-15.0 * __pyx_v_x3) + (15.0 * __pyx_v_x6)) + (15.0 * __pyx_v_x7))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":688
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70             # <<<<<<<<<<<<<<
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2
 */
  __Pyx_TraceLine(688,0,__PYX_ERR(1, 688, __pyx_L1_error))
  __pyx_v_x168 = (((((__pyx_v_x115 * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":689
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2             # <<<<<<<<<<<<<<
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)
 */
  __Pyx_TraceLine(689,0,__PYX_ERR(1, 689, __pyx_L1_error))
  __pyx_v_x169 = pow(__pyx_v_x114, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":690
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2             # <<<<<<<<<<<<<<
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28
 */
  __Pyx_TraceLine(690,0,__PYX_ERR(1, 690, __pyx_L1_error))
  __pyx_v_x170 = pow(__pyx_v_prst, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":691
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)             # <<<<<<<<<<<<<<
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120
 */
  __Pyx_TraceLine(691,0,__PYX_ERR(1, 691, __pyx_L1_error))
  __pyx_v_x171 = pow(__pyx_v_x141, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":692
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28             # <<<<<<<<<<<<<<
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 */
  __Pyx_TraceLine(692,0,__PYX_ERR(1, 692, __pyx_L1_error))
  __pyx_v_x172 = (1822680546449.21 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":693
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120             # <<<<<<<<<<<<<<
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)
 */
  __Pyx_TraceLine(693,0,__PYX_ERR(1, 693, __pyx_L1_error))
  __pyx_v_x173 = (5787938193408.0 * __pyx_v_x120);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":694
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0             # <<<<<<<<<<<<<<
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu
 */
  __Pyx_TraceLine(694,0,__PYX_ERR(1, 694, __pyx_L1_error))
  __pyx_v_x174 = (((((((((((((-12049908701745.2 * __pyx_v_nu) + (__pyx_v_r * __pyx_v_x172)) - (39476764256925.6 * __pyx_v_r)) + (10611661054566.2 * __pyx_v_x156)) + (2589101062873.81 * __pyx_v_x157)) - (326837426.241486 * __pyx_v_x159)) + (133772083200.0 * __pyx_v_x160)) - __pyx_v_x163) + __pyx_v_x173) + (5107745331375.71 * __pyx_v_x20)) + (80059249540278.2 * __pyx_v_x28)) + (6730497718123.02 * __pyx_v_x83)) + 275059053208689.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":695
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3
 */
  __Pyx_TraceLine(695,0,__PYX_ERR(1, 695, __pyx_L1_error))
  __pyx_v_x175 = pow(__pyx_v_x174, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":696
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 */
  __Pyx_TraceLine(696,0,__PYX_ERR(1, 696, __pyx_L1_error))
  __pyx_v_x176 = (4.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":697
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3             # <<<<<<<<<<<<<<
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178
 */
  __Pyx_TraceLine(697,0,__PYX_ERR(1, 697, __pyx_L1_error))
  __pyx_v_x177 = (3.0 * __pyx_v_x3);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":698
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7             # <<<<<<<<<<<<<<
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179
 */
  __Pyx_TraceLine(698,0,__PYX_ERR(1, 698, __pyx_L1_error))
  __pyx_v_x178 = ((3.0 * __pyx_v_x6) + (3.0 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":699
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178             # <<<<<<<<<<<<<<
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 */
  __Pyx_TraceLine(699,0,__PYX_ERR(1, 699, __pyx_L1_error))
  __pyx_v_x179 = ((-__pyx_v_x177) + __pyx_v_x178);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":700
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179             # <<<<<<<<<<<<<<
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 */
  __Pyx_TraceLine(700,0,__PYX_ERR(1, 700, __pyx_L1_error))
  __pyx_v_x180 = (0.0625 * __pyx_v_x179);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":701
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)             # <<<<<<<<<<<<<<
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 */
  __Pyx_TraceLine(701,0,__PYX_ERR(1, 701, __pyx_L1_error))
  __pyx_v_x181 = (__pyx_v_x57 * ((26.0 * __pyx_v_nu) + 449.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":702
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)             # <<<<<<<<<<<<<<
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 */
  __Pyx_TraceLine(702,0,__PYX_ERR(1, 702, __pyx_L1_error))
  __pyx_v_x182 = (__pyx_v_x9 * ((-1171.0 * __pyx_v_nu) - 861.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":703
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)             # <<<<<<<<<<<<<<
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)
 */
  __Pyx_TraceLine(703,0,__PYX_ERR(1, 703, __pyx_L1_error))
  __pyx_v_x183 = (__pyx_v_x103 * (((115.0 * __pyx_v_nu) + __pyx_v_x166) - 37.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":704
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154             # <<<<<<<<<<<<<<
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185
 */
  __Pyx_TraceLine(704,0,__PYX_ERR(1, 704, __pyx_L1_error))
  __pyx_v_x184 = ((((((5787938193408.0 * __pyx_v_x143) - (9216.0 * __pyx_v_x145)) - (967680.0 * __pyx_v_x147)) + (55296.0 * __pyx_v_x148)) + __pyx_v_x150) + __pyx_v_x154);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":705
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34
 */
  __Pyx_TraceLine(705,0,__PYX_ERR(1, 705, __pyx_L1_error))
  __pyx_v_x185 = pow(__pyx_v_x184, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":706
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185             # <<<<<<<<<<<<<<
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138
 */
  __Pyx_TraceLine(706,0,__PYX_ERR(1, 706, __pyx_L1_error))
  __pyx_v_x186 = (__pyx_v_x174 * __pyx_v_x185);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":707
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34             # <<<<<<<<<<<<<<
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188
 */
  __Pyx_TraceLine(707,0,__PYX_ERR(1, 707, __pyx_L1_error))
  __pyx_v_x187 = (__pyx_v_x186 * __pyx_v_x34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":708
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138             # <<<<<<<<<<<<<<
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 */
  __Pyx_TraceLine(708,0,__PYX_ERR(1, 708, __pyx_L1_error))
  __pyx_v_x188 = (__pyx_v_x119 * __pyx_v_x138);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":709
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188             # <<<<<<<<<<<<<<
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 */
  __Pyx_TraceLine(709,0,__PYX_ERR(1, 709, __pyx_L1_error))
  __pyx_v_x189 = (7680.0 * __pyx_v_x188);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":710
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))             # <<<<<<<<<<<<<<
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 */
  __Pyx_TraceLine(710,0,__PYX_ERR(1, 710, __pyx_L1_error))
  __pyx_v_x190 = (((__pyx_v_x116 + (__pyx_v_x14 * (((0.03125 * __pyx_v_x181) + (0.015625 * __pyx_v_x182)) + (0.015625 * __pyx_v_x183)))) + (__pyx_v_x187 * __pyx_v_x189)) + (__pyx_v_x24 * (((__pyx_v_x180 * (__pyx_v_x176 - 1.0)) - (2.625 * __pyx_v_x57)) + (__pyx_v_x9 * (__pyx_v_x15 + 2.8125)))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":711
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190             # <<<<<<<<<<<<<<
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60
 */
  __Pyx_TraceLine(711,0,__PYX_ERR(1, 711, __pyx_L1_error))
  __pyx_v_x191 = (((((__pyx_v_x169 * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x190);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":712
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 */
  __Pyx_TraceLine(712,0,__PYX_ERR(1, 712, __pyx_L1_error))
  __pyx_v_x192 = ((((((((((((((((-__pyx_v_x1) * __pyx_v_x100) + (__pyx_v_x112 * __pyx_v_x61)) + ((1.27277314139085e-19 * __pyx_v_x113) * __pyx_v_x168)) + ((1.69542100694444e-8 * __pyx_v_x191) * __pyx_v_x35)) + (__pyx_v_x21 * __pyx_v_x92)) + ((147.443752990146 * __pyx_v_x70) * __pyx_v_x72)) + (__pyx_v_x70 * __pyx_v_x82)) - ((11.3175085791863 * __pyx_v_x73) * __pyx_v_x75)) + (__pyx_v_x73 * __pyx_v_x87)) + ((1.48275342024365 * __pyx_v_x76) * __pyx_v_x78)) + (__pyx_v_x79 * __pyx_v_x80)) + (__pyx_v_x84 * __pyx_v_x85)) + (__pyx_v_x89 * __pyx_v_x90)) + (__pyx_v_x94 * __pyx_v_x95)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":713
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60             # <<<<<<<<<<<<<<
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)
 */
  __Pyx_TraceLine(713,0,__PYX_ERR(1, 713, __pyx_L1_error))
  __pyx_v_x193 = (2.0 * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":714
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)
 */
  __Pyx_TraceLine(714,0,__PYX_ERR(1, 714, __pyx_L1_error))
  __pyx_v_x194 = ((__pyx_v_x116 * (__pyx_v_x193 + 1.0)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":715
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 */
  __Pyx_TraceLine(715,0,__PYX_ERR(1, 715, __pyx_L1_error))
  __pyx_v_x195 = pow(__pyx_v_x194, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":716
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)             # <<<<<<<<<<<<<<
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu
 */
  __Pyx_TraceLine(716,0,__PYX_ERR(1, 716, __pyx_L1_error))
  __pyx_v_x196 = (__pyx_v_x103 * (__pyx_v_x176 + 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":717
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 */
  __Pyx_TraceLine(717,0,__PYX_ERR(1, 717, __pyx_L1_error))
  __pyx_v_x197 = (__pyx_v_x57 * ((2.0 * __pyx_v_nu) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":718
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)
 */
  __Pyx_TraceLine(718,0,__PYX_ERR(1, 718, __pyx_L1_error))
  __pyx_v_x198 = (7.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":719
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7             # <<<<<<<<<<<<<<
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)
 */
  __Pyx_TraceLine(719,0,__PYX_ERR(1, 719, __pyx_L1_error))
  __pyx_v_x199 = (((0.390625 * __pyx_v_x3) + (0.390625 * __pyx_v_x6)) + (0.390625 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":720
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)             # <<<<<<<<<<<<<<
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 */
  __Pyx_TraceLine(720,0,__PYX_ERR(1, 720, __pyx_L1_error))
  __pyx_v_x200 = (__pyx_v_x49 * (__pyx_v_x198 - 31.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":721
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 */
  __Pyx_TraceLine(721,0,__PYX_ERR(1, 721, __pyx_L1_error))
  __pyx_v_x201 = (__pyx_v_x104 * ((-__pyx_v_x176) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":722
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)             # <<<<<<<<<<<<<<
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203
 */
  __Pyx_TraceLine(722,0,__PYX_ERR(1, 722, __pyx_L1_error))
  __pyx_v_x202 = (__pyx_v_x58 * ((68.0 * __pyx_v_nu) - 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":723
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65             # <<<<<<<<<<<<<<
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)
 */
  __Pyx_TraceLine(723,0,__PYX_ERR(1, 723, __pyx_L1_error))
  __pyx_v_x203 = ((((((__pyx_v_x116 + (__pyx_v_x14 * (((0.125 * __pyx_v_x200) + (0.0625 * __pyx_v_x201)) + (2.0625 * __pyx_v_x58)))) + (__pyx_v_x14 * (((((0.125 * __pyx_v_x196) + (1.125 * __pyx_v_x3)) - (1.25 * __pyx_v_x57)) + (1.125 * __pyx_v_x6)) + (1.125 * __pyx_v_x7)))) + (7680.0 * __pyx_v_x140)) + (__pyx_v_x35 * (((__pyx_v_x109 * (((2870.0 * __pyx_v_nu) - __pyx_v_x107) + 561.0)) + (__pyx_v_x111 * (((1166.0 * __pyx_v_nu) - (328.0 * __pyx_v_x28)) - 171.0))) + (0.125 * __pyx_v_x202)))) + (__pyx_v_x35 * ((((0.015625 * __pyx_v_x179) * (((-27.0 * __pyx_v_nu) + (28.0 * __pyx_v_x28)) - 3.0)) - (1.21875 * __pyx_v_x197)) - (__pyx_v_x199 * (__pyx_v_x198 + 9.0))))) + __pyx_v_x65);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":724
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203             # <<<<<<<<<<<<<<
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 */
  __Pyx_TraceLine(724,0,__PYX_ERR(1, 724, __pyx_L1_error))
  __pyx_v_x204 = (__pyx_v_x195 * __pyx_v_x203);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":725
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)             # <<<<<<<<<<<<<<
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0
 */
  __Pyx_TraceLine(725,0,__PYX_ERR(1, 725, __pyx_L1_error))
  __pyx_v_x205 = pow((__pyx_v_x192 * __pyx_v_x204), -0.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":726
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7             # <<<<<<<<<<<<<<
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2
 */
  __Pyx_TraceLine(726,0,__PYX_ERR(1, 726, __pyx_L1_error))
  __pyx_v_x206 = (((-2.0 * __pyx_v_x3) - (2.0 * __pyx_v_x6)) - (2.0 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":727
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0             # <<<<<<<<<<<<<<
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)
 */
  __Pyx_TraceLine(727,0,__PYX_ERR(1, 727, __pyx_L1_error))
  __pyx_v_x207 = ((4.0 * __pyx_v_x60) + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":728
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2             # <<<<<<<<<<<<<<
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 */
  __Pyx_TraceLine(728,0,__PYX_ERR(1, 728, __pyx_L1_error))
  __pyx_v_x208 = (4.5 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":729
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)             # <<<<<<<<<<<<<<
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60
 */
  __Pyx_TraceLine(729,0,__PYX_ERR(1, 729, __pyx_L1_error))
  __pyx_v_x209 = pow(__pyx_v_r, -6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":730
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0             # <<<<<<<<<<<<<<
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0
 */
  __Pyx_TraceLine(730,0,__PYX_ERR(1, 730, __pyx_L1_error))
  __pyx_v_x210 = (((((-6572428.80109422 * __pyx_v_nu) + ((2048.0 * __pyx_v_x118) * __pyx_v_x79)) + (688128.0 * __pyx_v_x128)) + (1300341.64885296 * __pyx_v_x28)) + 1720320.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":731
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60             # <<<<<<<<<<<<<<
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 */
  __Pyx_TraceLine(731,0,__PYX_ERR(1, 731, __pyx_L1_error))
  __pyx_v_x211 = (__pyx_v_x28 * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":732
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0             # <<<<<<<<<<<<<<
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 */
  __Pyx_TraceLine(732,0,__PYX_ERR(1, 732, __pyx_L1_error))
  __pyx_v_x212 = (4.0 * __pyx_v_x0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":733
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0             # <<<<<<<<<<<<<<
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214
 */
  __Pyx_TraceLine(733,0,__PYX_ERR(1, 733, __pyx_L1_error))
  __pyx_v_x213 = ((__pyx_v_x130 + (6.0 * __pyx_v_x20)) + 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":734
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2             # <<<<<<<<<<<<<<
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24
 */
  __Pyx_TraceLine(734,0,__PYX_ERR(1, 734, __pyx_L1_error))
  __pyx_v_x214 = (((1.31621673590926e-19 * __pyx_v_x119) * ((((((((53760.0 * __pyx_v_nu) * (((((((((3740417.71815805 * __pyx_v_r) + (1057984.42953951 * __pyx_v_x0)) - (938918.400156317 * __pyx_v_x13)) + (__pyx_v_x132 * (__pyx_v_x212 + __pyx_v_x213))) + (__pyx_v_x133 * ((((17058.7851094412 * __pyx_v_r) - (27409.3925547206 * __pyx_v_x0)) + (12794.0888320809 * __pyx_v_x20)) + 13218.7851094412))) + (__pyx_v_x135 * ((((1396.0 * __pyx_v_x0) + __pyx_v_x134) + (5778.0 * __pyx_v_x20)) + 7704.0))) + (__pyx_v_x136 * ((-__pyx_v_x212) + __pyx_v_x213))) + (2115968.85907902 * __pyx_v_x20)) + 2888096.47013111)) + ((32768.0 * __pyx_v_x127) * __pyx_v_x79)) + ((32768.0 * __pyx_v_x128) * (((((((-1791904.9465589 * __pyx_v_nu) + ((3840.0 * __pyx_v_r) * __pyx_v_x124)) + ((2880.0 * __pyx_v_x125) * __pyx_v_x20)) + (1920.0 * __pyx_v_x126)) + (806400.0 * __pyx_v_x13)) - (725760.0 * __pyx_v_x28)) + 4143360.0))) + (66060288000.0 * __pyx_v_x13)) + ((135291469824.0 * __pyx_v_x211) * __pyx_v_x93)) + ((7.0 * __pyx_v_x28) * ((((((-117964800.0 * __pyx_v_a6) + (4129567622.65173 * __pyx_v_r)) - (12357002815.0752 * __pyx_v_x0)) + (7133302759.42198 * __pyx_v_x13)) - (18535504222.6128 * __pyx_v_x20)) + 122635399361.987))) + ((7.0 * __pyx_v_x83) * ((-5706642207.53758 * __pyx_v_r) - 26189434371.6082)))) / pow((((((((0.28004222119933 * __pyx_v_x121) + (4.63661586574928e-9 * __pyx_v_x122)) + (2.8978849160933e-11 * __pyx_v_x123)) + (1.35654132757922e-7 * __pyx_v_x129)) + (2.22557561555966e-7 * __pyx_v_x137)) + (0.0546957463279941 * __pyx_v_x34)) + __pyx_v_x88), 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":735
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214             # <<<<<<<<<<<<<<
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93
 */
  __Pyx_TraceLine(735,0,__PYX_ERR(1, 735, __pyx_L1_error))
  __pyx_v_x215 = (__pyx_v_x13 * __pyx_v_x214);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":736
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24             # <<<<<<<<<<<<<<
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 */
  __Pyx_TraceLine(736,0,__PYX_ERR(1, 736, __pyx_L1_error))
  __pyx_v_x216 = (2.0 * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":737
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93             # <<<<<<<<<<<<<<
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70
 */
  __Pyx_TraceLine(737,0,__PYX_ERR(1, 737, __pyx_L1_error))
  __pyx_v_x217 = (11575876386816.0 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":738
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1             # <<<<<<<<<<<<<<
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 */
  __Pyx_TraceLine(738,0,__PYX_ERR(1, 738, __pyx_L1_error))
  __pyx_v_x218 = (((((((((5807150888816.34 * __pyx_v_nu) + (10215490662751.4 * __pyx_v_r)) + (5178202125747.62 * __pyx_v_x156)) + (267544166400.0 * __pyx_v_x158)) - (__pyx_v_x162 * __pyx_v_x60)) + __pyx_v_x172) + (__pyx_v_x217 * __pyx_v_x60)) + (__pyx_v_x93 * (((4161798144000.0 * __pyx_v_nu) + (1759846072320.0 * __pyx_v_r)) + 12148770078720.0))) - 53501685054374.1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":739
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70             # <<<<<<<<<<<<<<
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190
 */
  __Pyx_TraceLine(739,0,__PYX_ERR(1, 739, __pyx_L1_error))
  __pyx_v_x219 = ((((__pyx_v_x113 * __pyx_v_x115) * __pyx_v_x155) * __pyx_v_x167) * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":740
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)             # <<<<<<<<<<<<<<
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24
 */
  __Pyx_TraceLine(740,0,__PYX_ERR(1, 740, __pyx_L1_error))
  __pyx_v_x220 = ((((((((-18432.0 * __pyx_v_r) * __pyx_v_x144) - ((2903040.0 * __pyx_v_x146) * __pyx_v_x20)) + __pyx_v_x149) + (__pyx_v_x153 * __pyx_v_x60)) + __pyx_v_x173) + __pyx_v_x217) + (__pyx_v_x93 * ((((20113376778784.3 * __pyx_v_nu) + (566231040.0 * __pyx_v_x159)) + __pyx_v_x161) + (2639769108480.0 * __pyx_v_x20))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":741
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190             # <<<<<<<<<<<<<<
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210
 */
  __Pyx_TraceLine(741,0,__PYX_ERR(1, 741, __pyx_L1_error))
  __pyx_v_x221 = (__pyx_v_x170 * __pyx_v_x190);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":742
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24             # <<<<<<<<<<<<<<
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 */
  __Pyx_TraceLine(742,0,__PYX_ERR(1, 742, __pyx_L1_error))
  __pyx_v_x222 = (__pyx_v_x206 * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":743
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210             # <<<<<<<<<<<<<<
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35
 */
  __Pyx_TraceLine(743,0,__PYX_ERR(1, 743, __pyx_L1_error))
  __pyx_v_x223 = (7680.0 * __pyx_v_x210);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":744
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222             # <<<<<<<<<<<<<<
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3
 */
  __Pyx_TraceLine(744,0,__PYX_ERR(1, 744, __pyx_L1_error))
  __pyx_v_x224 = (((((30720.0 * __pyx_v_x0) * __pyx_v_x188) + (__pyx_v_x139 * __pyx_v_x223)) - __pyx_v_x215) + __pyx_v_x222);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":745
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35             # <<<<<<<<<<<<<<
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5
 */
  __Pyx_TraceLine(745,0,__PYX_ERR(1, 745, __pyx_L1_error))
  __pyx_v_x225 = (__pyx_v_x184 * __pyx_v_x35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":746
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3             # <<<<<<<<<<<<<<
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7
 */
  __Pyx_TraceLine(746,0,__PYX_ERR(1, 746, __pyx_L1_error))
  __pyx_v_x226 = pow(__pyx_v_prst, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":747
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5             # <<<<<<<<<<<<<<
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226
 */
  __Pyx_TraceLine(747,0,__PYX_ERR(1, 747, __pyx_L1_error))
  __pyx_v_x227 = pow(__pyx_v_prst, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":748
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7             # <<<<<<<<<<<<<<
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227
 */
  __Pyx_TraceLine(748,0,__PYX_ERR(1, 748, __pyx_L1_error))
  __pyx_v_x228 = pow(__pyx_v_prst, 7.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":749
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226             # <<<<<<<<<<<<<<
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 */
  __Pyx_TraceLine(749,0,__PYX_ERR(1, 749, __pyx_L1_error))
  __pyx_v_x229 = (4.0 * __pyx_v_x226);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":750
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227             # <<<<<<<<<<<<<<
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21
 */
  __Pyx_TraceLine(750,0,__PYX_ERR(1, 750, __pyx_L1_error))
  __pyx_v_x230 = (6.0 * __pyx_v_x227);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":751
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)             # <<<<<<<<<<<<<<
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216
 */
  __Pyx_TraceLine(751,0,__PYX_ERR(1, 751, __pyx_L1_error))
  __pyx_v_x231 = ((__pyx_v_x203 * __pyx_v_x205) / ((__pyx_v_x116 * __pyx_v_x207) + 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":752
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21             # <<<<<<<<<<<<<<
 *         cdef double x233 = pphi*x216
 *         cdef double x234 = 4.0*pphi**3*x14
 */
  __Pyx_TraceLine(752,0,__PYX_ERR(1, 752, __pyx_L1_error))
  __pyx_v_x232 = ((2.0 * __pyx_v_pphi) * __pyx_v_x21);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":753
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216             # <<<<<<<<<<<<<<
 *         cdef double x234 = 4.0*pphi**3*x14
 *         cdef double y0 = r**3
 */
  __Pyx_TraceLine(753,0,__PYX_ERR(1, 753, __pyx_L1_error))
  __pyx_v_x233 = (__pyx_v_pphi * __pyx_v_x216);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":754
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216
 *         cdef double x234 = 4.0*pphi**3*x14             # <<<<<<<<<<<<<<
 *         cdef double y0 = r**3
 *         cdef double y1 = r + 2.0
 */
  __Pyx_TraceLine(754,0,__PYX_ERR(1, 754, __pyx_L1_error))
  __pyx_v_x234 = ((4.0 * pow(__pyx_v_pphi, 3.0)) * __pyx_v_x14);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":755
 *         cdef double x233 = pphi*x216
 *         cdef double x234 = 4.0*pphi**3*x14
 *         cdef double y0 = r**3             # <<<<<<<<<<<<<<
 *         cdef double y1 = r + 2.0
 *         cdef double y2 = chix1*chix2
 */
  __Pyx_TraceLine(755,0,__PYX_ERR(1, 755, __pyx_L1_error))
  __pyx_v_y0 = pow(__pyx_v_r, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":756
 *         cdef double x234 = 4.0*pphi**3*x14
 *         cdef double y0 = r**3
 *         cdef double y1 = r + 2.0             # <<<<<<<<<<<<<<
 *         cdef double y2 = chix1*chix2
 *         cdef double y3 = chiy1*chiy2
 */
  __Pyx_TraceLine(756,0,__PYX_ERR(1, 756, __pyx_L1_error))
  __pyx_v_y1 = (__pyx_v_r + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":757
 *         cdef double y0 = r**3
 *         cdef double y1 = r + 2.0
 *         cdef double y2 = chix1*chix2             # <<<<<<<<<<<<<<
 *         cdef double y3 = chiy1*chiy2
 *         cdef double y4 = chiz1*chiz2
 */
  __Pyx_TraceLine(757,0,__PYX_ERR(1, 757, __pyx_L1_error))
  __pyx_v_y2 = (__pyx_v_chix1 * __pyx_v_chix2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":758
 *         cdef double y1 = r + 2.0
 *         cdef double y2 = chix1*chix2
 *         cdef double y3 = chiy1*chiy2             # <<<<<<<<<<<<<<
 *         cdef double y4 = chiz1*chiz2
 *         cdef double y5 = 2.0*y2 + 2.0*y3 + 2.0*y4
 */
  __Pyx_TraceLine(758,0,__PYX_ERR(1, 758, __pyx_L1_error))
  __pyx_v_y3 = (__pyx_v_chiy1 * __pyx_v_chiy2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":759
 *         cdef double y2 = chix1*chix2
 *         cdef double y3 = chiy1*chiy2
 *         cdef double y4 = chiz1*chiz2             # <<<<<<<<<<<<<<
 *         cdef double y5 = 2.0*y2 + 2.0*y3 + 2.0*y4
 *         cdef double y6 = X_1*X_2
 */
  __Pyx_TraceLine(759,0,__PYX_ERR(1, 759, __pyx_L1_error))
  __pyx_v_y4 = (__pyx_v_chiz1 * __pyx_v_chiz2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":760
 *         cdef double y3 = chiy1*chiy2
 *         cdef double y4 = chiz1*chiz2
 *         cdef double y5 = 2.0*y2 + 2.0*y3 + 2.0*y4             # <<<<<<<<<<<<<<
 *         cdef double y6 = X_1*X_2
 *         cdef double y7 = y5*y6
 */
  __Pyx_TraceLine(760,0,__PYX_ERR(1, 760, __pyx_L1_error))
  __pyx_v_y5 = (((2.0 * __pyx_v_y2) + (2.0 * __pyx_v_y3)) + (2.0 * __pyx_v_y4));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":761
 *         cdef double y4 = chiz1*chiz2
 *         cdef double y5 = 2.0*y2 + 2.0*y3 + 2.0*y4
 *         cdef double y6 = X_1*X_2             # <<<<<<<<<<<<<<
 *         cdef double y7 = y5*y6
 *         cdef double y8 = X_1**2
 */
  __Pyx_TraceLine(761,0,__PYX_ERR(1, 761, __pyx_L1_error))
  __pyx_v_y6 = (__pyx_v_X_1 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":762
 *         cdef double y5 = 2.0*y2 + 2.0*y3 + 2.0*y4
 *         cdef double y6 = X_1*X_2
 *         cdef double y7 = y5*y6             # <<<<<<<<<<<<<<
 *         cdef double y8 = X_1**2
 *         cdef double y9 = chix1**2
 */
  __Pyx_TraceLine(762,0,__PYX_ERR(1, 762, __pyx_L1_error))
  __pyx_v_y7 = (__pyx_v_y5 * __pyx_v_y6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":763
 *         cdef double y6 = X_1*X_2
 *         cdef double y7 = y5*y6
 *         cdef double y8 = X_1**2             # <<<<<<<<<<<<<<
 *         cdef double y9 = chix1**2
 *         cdef double y10 = chiy1**2
 */
  __Pyx_TraceLine(763,0,__PYX_ERR(1, 763, __pyx_L1_error))
  __pyx_v_y8 = pow(__pyx_v_X_1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":764
 *         cdef double y7 = y5*y6
 *         cdef double y8 = X_1**2
 *         cdef double y9 = chix1**2             # <<<<<<<<<<<<<<
 *         cdef double y10 = chiy1**2
 *         cdef double y11 = chiz1**2
 */
  __Pyx_TraceLine(764,0,__PYX_ERR(1, 764, __pyx_L1_error))
  __pyx_v_y9 = pow(__pyx_v_chix1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":765
 *         cdef double y8 = X_1**2
 *         cdef double y9 = chix1**2
 *         cdef double y10 = chiy1**2             # <<<<<<<<<<<<<<
 *         cdef double y11 = chiz1**2
 *         cdef double y12 = y10 + y11 + y9
 */
  __Pyx_TraceLine(765,0,__PYX_ERR(1, 765, __pyx_L1_error))
  __pyx_v_y10 = pow(__pyx_v_chiy1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":766
 *         cdef double y9 = chix1**2
 *         cdef double y10 = chiy1**2
 *         cdef double y11 = chiz1**2             # <<<<<<<<<<<<<<
 *         cdef double y12 = y10 + y11 + y9
 *         cdef double y13 = y12*y8
 */
  __Pyx_TraceLine(766,0,__PYX_ERR(1, 766, __pyx_L1_error))
  __pyx_v_y11 = pow(__pyx_v_chiz1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":767
 *         cdef double y10 = chiy1**2
 *         cdef double y11 = chiz1**2
 *         cdef double y12 = y10 + y11 + y9             # <<<<<<<<<<<<<<
 *         cdef double y13 = y12*y8
 *         cdef double y14 = X_2**2
 */
  __Pyx_TraceLine(767,0,__PYX_ERR(1, 767, __pyx_L1_error))
  __pyx_v_y12 = ((__pyx_v_y10 + __pyx_v_y11) + __pyx_v_y9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":768
 *         cdef double y11 = chiz1**2
 *         cdef double y12 = y10 + y11 + y9
 *         cdef double y13 = y12*y8             # <<<<<<<<<<<<<<
 *         cdef double y14 = X_2**2
 *         cdef double y15 = chix2**2
 */
  __Pyx_TraceLine(768,0,__PYX_ERR(1, 768, __pyx_L1_error))
  __pyx_v_y13 = (__pyx_v_y12 * __pyx_v_y8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":769
 *         cdef double y12 = y10 + y11 + y9
 *         cdef double y13 = y12*y8
 *         cdef double y14 = X_2**2             # <<<<<<<<<<<<<<
 *         cdef double y15 = chix2**2
 *         cdef double y16 = chiy2**2
 */
  __Pyx_TraceLine(769,0,__PYX_ERR(1, 769, __pyx_L1_error))
  __pyx_v_y14 = pow(__pyx_v_X_2, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":770
 *         cdef double y13 = y12*y8
 *         cdef double y14 = X_2**2
 *         cdef double y15 = chix2**2             # <<<<<<<<<<<<<<
 *         cdef double y16 = chiy2**2
 *         cdef double y17 = chiz2**2
 */
  __Pyx_TraceLine(770,0,__PYX_ERR(1, 770, __pyx_L1_error))
  __pyx_v_y15 = pow(__pyx_v_chix2, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":771
 *         cdef double y14 = X_2**2
 *         cdef double y15 = chix2**2
 *         cdef double y16 = chiy2**2             # <<<<<<<<<<<<<<
 *         cdef double y17 = chiz2**2
 *         cdef double y18 = y14*(y15 + y16 + y17)
 */
  __Pyx_TraceLine(771,0,__PYX_ERR(1, 771, __pyx_L1_error))
  __pyx_v_y16 = pow(__pyx_v_chiy2, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":772
 *         cdef double y15 = chix2**2
 *         cdef double y16 = chiy2**2
 *         cdef double y17 = chiz2**2             # <<<<<<<<<<<<<<
 *         cdef double y18 = y14*(y15 + y16 + y17)
 *         cdef double y19 = y13 + y18
 */
  __Pyx_TraceLine(772,0,__PYX_ERR(1, 772, __pyx_L1_error))
  __pyx_v_y17 = pow(__pyx_v_chiz2, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":773
 *         cdef double y16 = chiy2**2
 *         cdef double y17 = chiz2**2
 *         cdef double y18 = y14*(y15 + y16 + y17)             # <<<<<<<<<<<<<<
 *         cdef double y19 = y13 + y18
 *         cdef double y20 = y19 + y7
 */
  __Pyx_TraceLine(773,0,__PYX_ERR(1, 773, __pyx_L1_error))
  __pyx_v_y18 = (__pyx_v_y14 * ((__pyx_v_y15 + __pyx_v_y16) + __pyx_v_y17));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":774
 *         cdef double y17 = chiz2**2
 *         cdef double y18 = y14*(y15 + y16 + y17)
 *         cdef double y19 = y13 + y18             # <<<<<<<<<<<<<<
 *         cdef double y20 = y19 + y7
 *         cdef double y21 = y1*y20
 */
  __Pyx_TraceLine(774,0,__PYX_ERR(1, 774, __pyx_L1_error))
  __pyx_v_y19 = (__pyx_v_y13 + __pyx_v_y18);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":775
 *         cdef double y18 = y14*(y15 + y16 + y17)
 *         cdef double y19 = y13 + y18
 *         cdef double y20 = y19 + y7             # <<<<<<<<<<<<<<
 *         cdef double y21 = y1*y20
 *         cdef double y22 = y0 + y21
 */
  __Pyx_TraceLine(775,0,__PYX_ERR(1, 775, __pyx_L1_error))
  __pyx_v_y20 = (__pyx_v_y19 + __pyx_v_y7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":776
 *         cdef double y19 = y13 + y18
 *         cdef double y20 = y19 + y7
 *         cdef double y21 = y1*y20             # <<<<<<<<<<<<<<
 *         cdef double y22 = y0 + y21
 *         cdef double y23 = y22**(-1)
 */
  __Pyx_TraceLine(776,0,__PYX_ERR(1, 776, __pyx_L1_error))
  __pyx_v_y21 = (__pyx_v_y1 * __pyx_v_y20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":777
 *         cdef double y20 = y19 + y7
 *         cdef double y21 = y1*y20
 *         cdef double y22 = y0 + y21             # <<<<<<<<<<<<<<
 *         cdef double y23 = y22**(-1)
 *         cdef double y24 = r**5
 */
  __Pyx_TraceLine(777,0,__PYX_ERR(1, 777, __pyx_L1_error))
  __pyx_v_y22 = (__pyx_v_y0 + __pyx_v_y21);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":778
 *         cdef double y21 = y1*y20
 *         cdef double y22 = y0 + y21
 *         cdef double y23 = y22**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y24 = r**5
 *         cdef double y25 = y24**(-1)
 */
  __Pyx_TraceLine(778,0,__PYX_ERR(1, 778, __pyx_L1_error))
  __pyx_v_y23 = pow(__pyx_v_y22, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":779
 *         cdef double y22 = y0 + y21
 *         cdef double y23 = y22**(-1)
 *         cdef double y24 = r**5             # <<<<<<<<<<<<<<
 *         cdef double y25 = y24**(-1)
 *         cdef double y26 = 12.0*nu
 */
  __Pyx_TraceLine(779,0,__PYX_ERR(1, 779, __pyx_L1_error))
  __pyx_v_y24 = pow(__pyx_v_r, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":780
 *         cdef double y23 = y22**(-1)
 *         cdef double y24 = r**5
 *         cdef double y25 = y24**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y26 = 12.0*nu
 *         cdef double y27 = X_1*chiL1
 */
  __Pyx_TraceLine(780,0,__PYX_ERR(1, 780, __pyx_L1_error))
  __pyx_v_y25 = pow(__pyx_v_y24, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":781
 *         cdef double y24 = r**5
 *         cdef double y25 = y24**(-1)
 *         cdef double y26 = 12.0*nu             # <<<<<<<<<<<<<<
 *         cdef double y27 = X_1*chiL1
 *         cdef double y28 = X_2*chiL2
 */
  __Pyx_TraceLine(781,0,__PYX_ERR(1, 781, __pyx_L1_error))
  __pyx_v_y26 = (12.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":782
 *         cdef double y25 = y24**(-1)
 *         cdef double y26 = 12.0*nu
 *         cdef double y27 = X_1*chiL1             # <<<<<<<<<<<<<<
 *         cdef double y28 = X_2*chiL2
 *         cdef double y29 = y27 + y28
 */
  __Pyx_TraceLine(782,0,__PYX_ERR(1, 782, __pyx_L1_error))
  __pyx_v_y27 = (__pyx_v_X_1 * __pyx_v_chiL1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":783
 *         cdef double y26 = 12.0*nu
 *         cdef double y27 = X_1*chiL1
 *         cdef double y28 = X_2*chiL2             # <<<<<<<<<<<<<<
 *         cdef double y29 = y27 + y28
 *         cdef double y30 = pphi*y29
 */
  __Pyx_TraceLine(783,0,__PYX_ERR(1, 783, __pyx_L1_error))
  __pyx_v_y28 = (__pyx_v_X_2 * __pyx_v_chiL2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":784
 *         cdef double y27 = X_1*chiL1
 *         cdef double y28 = X_2*chiL2
 *         cdef double y29 = y27 + y28             # <<<<<<<<<<<<<<
 *         cdef double y30 = pphi*y29
 *         cdef double y31 = dSO*y30
 */
  __Pyx_TraceLine(784,0,__PYX_ERR(1, 784, __pyx_L1_error))
  __pyx_v_y29 = (__pyx_v_y27 + __pyx_v_y28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":785
 *         cdef double y28 = X_2*chiL2
 *         cdef double y29 = y27 + y28
 *         cdef double y30 = pphi*y29             # <<<<<<<<<<<<<<
 *         cdef double y31 = dSO*y30
 *         cdef double y32 = y0**(-1)
 */
  __Pyx_TraceLine(785,0,__PYX_ERR(1, 785, __pyx_L1_error))
  __pyx_v_y30 = (__pyx_v_pphi * __pyx_v_y29);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":786
 *         cdef double y29 = y27 + y28
 *         cdef double y30 = pphi*y29
 *         cdef double y31 = dSO*y30             # <<<<<<<<<<<<<<
 *         cdef double y32 = y0**(-1)
 *         cdef double y33 = -1.40625*nu - 0.46875
 */
  __Pyx_TraceLine(786,0,__PYX_ERR(1, 786, __pyx_L1_error))
  __pyx_v_y31 = (__pyx_v_dSO * __pyx_v_y30);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":787
 *         cdef double y30 = pphi*y29
 *         cdef double y31 = dSO*y30
 *         cdef double y32 = y0**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y33 = -1.40625*nu - 0.46875
 *         cdef double y34 = pphi**2
 */
  __Pyx_TraceLine(787,0,__PYX_ERR(1, 787, __pyx_L1_error))
  __pyx_v_y32 = pow(__pyx_v_y0, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":788
 *         cdef double y31 = dSO*y30
 *         cdef double y32 = y0**(-1)
 *         cdef double y33 = -1.40625*nu - 0.46875             # <<<<<<<<<<<<<<
 *         cdef double y34 = pphi**2
 *         cdef double y35 = r**4
 */
  __Pyx_TraceLine(788,0,__PYX_ERR(1, 788, __pyx_L1_error))
  __pyx_v_y33 = ((-1.40625 * __pyx_v_nu) - 0.46875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":789
 *         cdef double y32 = y0**(-1)
 *         cdef double y33 = -1.40625*nu - 0.46875
 *         cdef double y34 = pphi**2             # <<<<<<<<<<<<<<
 *         cdef double y35 = r**4
 *         cdef double y36 = y35**(-1)
 */
  __Pyx_TraceLine(789,0,__PYX_ERR(1, 789, __pyx_L1_error))
  __pyx_v_y34 = pow(__pyx_v_pphi, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":790
 *         cdef double y33 = -1.40625*nu - 0.46875
 *         cdef double y34 = pphi**2
 *         cdef double y35 = r**4             # <<<<<<<<<<<<<<
 *         cdef double y36 = y35**(-1)
 *         cdef double y37 = 6.0*y36
 */
  __Pyx_TraceLine(790,0,__PYX_ERR(1, 790, __pyx_L1_error))
  __pyx_v_y35 = pow(__pyx_v_r, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":791
 *         cdef double y34 = pphi**2
 *         cdef double y35 = r**4
 *         cdef double y36 = y35**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y37 = 6.0*y36
 *         cdef double y38 = y34*y37
 */
  __Pyx_TraceLine(791,0,__PYX_ERR(1, 791, __pyx_L1_error))
  __pyx_v_y36 = pow(__pyx_v_y35, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":792
 *         cdef double y35 = r**4
 *         cdef double y36 = y35**(-1)
 *         cdef double y37 = 6.0*y36             # <<<<<<<<<<<<<<
 *         cdef double y38 = y34*y37
 *         cdef double y39 = nu**2
 */
  __Pyx_TraceLine(792,0,__PYX_ERR(1, 792, __pyx_L1_error))
  __pyx_v_y37 = (6.0 * __pyx_v_y36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":793
 *         cdef double y36 = y35**(-1)
 *         cdef double y37 = 6.0*y36
 *         cdef double y38 = y34*y37             # <<<<<<<<<<<<<<
 *         cdef double y39 = nu**2
 *         cdef double y40 = -2.0859375*nu - 2.07161458333333*y39 + 0.23046875
 */
  __Pyx_TraceLine(793,0,__PYX_ERR(1, 793, __pyx_L1_error))
  __pyx_v_y38 = (__pyx_v_y34 * __pyx_v_y37);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":794
 *         cdef double y37 = 6.0*y36
 *         cdef double y38 = y34*y37
 *         cdef double y39 = nu**2             # <<<<<<<<<<<<<<
 *         cdef double y40 = -2.0859375*nu - 2.07161458333333*y39 + 0.23046875
 *         cdef double y41 = 12.0*y25
 */
  __Pyx_TraceLine(794,0,__PYX_ERR(1, 794, __pyx_L1_error))
  __pyx_v_y39 = pow(__pyx_v_nu, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":795
 *         cdef double y38 = y34*y37
 *         cdef double y39 = nu**2
 *         cdef double y40 = -2.0859375*nu - 2.07161458333333*y39 + 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double y41 = 12.0*y25
 *         cdef double y42 = y34*y41
 */
  __Pyx_TraceLine(795,0,__PYX_ERR(1, 795, __pyx_L1_error))
  __pyx_v_y40 = (((-2.0859375 * __pyx_v_nu) - (2.07161458333333 * __pyx_v_y39)) + 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":796
 *         cdef double y39 = nu**2
 *         cdef double y40 = -2.0859375*nu - 2.07161458333333*y39 + 0.23046875
 *         cdef double y41 = 12.0*y25             # <<<<<<<<<<<<<<
 *         cdef double y42 = y34*y41
 *         cdef double y43 = r**(-6)
 */
  __Pyx_TraceLine(796,0,__PYX_ERR(1, 796, __pyx_L1_error))
  __pyx_v_y41 = (12.0 * __pyx_v_y25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":797
 *         cdef double y40 = -2.0859375*nu - 2.07161458333333*y39 + 0.23046875
 *         cdef double y41 = 12.0*y25
 *         cdef double y42 = y34*y41             # <<<<<<<<<<<<<<
 *         cdef double y43 = r**(-6)
 *         cdef double y44 = 20.0*y43
 */
  __Pyx_TraceLine(797,0,__PYX_ERR(1, 797, __pyx_L1_error))
  __pyx_v_y42 = (__pyx_v_y34 * __pyx_v_y41);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":798
 *         cdef double y41 = 12.0*y25
 *         cdef double y42 = y34*y41
 *         cdef double y43 = r**(-6)             # <<<<<<<<<<<<<<
 *         cdef double y44 = 20.0*y43
 *         cdef double y45 = pphi**4
 */
  __Pyx_TraceLine(798,0,__PYX_ERR(1, 798, __pyx_L1_error))
  __pyx_v_y43 = pow(__pyx_v_r, -6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":799
 *         cdef double y42 = y34*y41
 *         cdef double y43 = r**(-6)
 *         cdef double y44 = 20.0*y43             # <<<<<<<<<<<<<<
 *         cdef double y45 = pphi**4
 *         cdef double y46 = 0.5859375*nu + 1.34765625*y39 + 0.41015625
 */
  __Pyx_TraceLine(799,0,__PYX_ERR(1, 799, __pyx_L1_error))
  __pyx_v_y44 = (20.0 * __pyx_v_y43);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":800
 *         cdef double y43 = r**(-6)
 *         cdef double y44 = 20.0*y43
 *         cdef double y45 = pphi**4             # <<<<<<<<<<<<<<
 *         cdef double y46 = 0.5859375*nu + 1.34765625*y39 + 0.41015625
 *         cdef double y47 = y45*y46
 */
  __Pyx_TraceLine(800,0,__PYX_ERR(1, 800, __pyx_L1_error))
  __pyx_v_y45 = pow(__pyx_v_pphi, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":801
 *         cdef double y44 = 20.0*y43
 *         cdef double y45 = pphi**4
 *         cdef double y46 = 0.5859375*nu + 1.34765625*y39 + 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double y47 = y45*y46
 *         cdef double y48 = 0.46875 - 0.28125*nu
 */
  __Pyx_TraceLine(801,0,__PYX_ERR(1, 801, __pyx_L1_error))
  __pyx_v_y46 = (((0.5859375 * __pyx_v_nu) + (1.34765625 * __pyx_v_y39)) + 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":802
 *         cdef double y45 = pphi**4
 *         cdef double y46 = 0.5859375*nu + 1.34765625*y39 + 0.41015625
 *         cdef double y47 = y45*y46             # <<<<<<<<<<<<<<
 *         cdef double y48 = 0.46875 - 0.28125*nu
 *         cdef double y49 = -0.2734375*nu - 0.798177083333333*y39 - 0.23046875
 */
  __Pyx_TraceLine(802,0,__PYX_ERR(1, 802, __pyx_L1_error))
  __pyx_v_y47 = (__pyx_v_y45 * __pyx_v_y46);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":803
 *         cdef double y46 = 0.5859375*nu + 1.34765625*y39 + 0.41015625
 *         cdef double y47 = y45*y46
 *         cdef double y48 = 0.46875 - 0.28125*nu             # <<<<<<<<<<<<<<
 *         cdef double y49 = -0.2734375*nu - 0.798177083333333*y39 - 0.23046875
 *         cdef double y50 = -0.3515625*nu + 0.29296875*y39 - 0.41015625
 */
  __Pyx_TraceLine(803,0,__PYX_ERR(1, 803, __pyx_L1_error))
  __pyx_v_y48 = (0.46875 - (0.28125 * __pyx_v_nu));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":804
 *         cdef double y47 = y45*y46
 *         cdef double y48 = 0.46875 - 0.28125*nu
 *         cdef double y49 = -0.2734375*nu - 0.798177083333333*y39 - 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double y50 = -0.3515625*nu + 0.29296875*y39 - 0.41015625
 *         cdef double y51 = y45*y50
 */
  __Pyx_TraceLine(804,0,__PYX_ERR(1, 804, __pyx_L1_error))
  __pyx_v_y49 = (((-0.2734375 * __pyx_v_nu) - (0.798177083333333 * __pyx_v_y39)) - 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":805
 *         cdef double y48 = 0.46875 - 0.28125*nu
 *         cdef double y49 = -0.2734375*nu - 0.798177083333333*y39 - 0.23046875
 *         cdef double y50 = -0.3515625*nu + 0.29296875*y39 - 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double y51 = y45*y50
 *         cdef double y52 = y27 - y28
 */
  __Pyx_TraceLine(805,0,__PYX_ERR(1, 805, __pyx_L1_error))
  __pyx_v_y50 = (((-0.3515625 * __pyx_v_nu) + (0.29296875 * __pyx_v_y39)) - 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":806
 *         cdef double y49 = -0.2734375*nu - 0.798177083333333*y39 - 0.23046875
 *         cdef double y50 = -0.3515625*nu + 0.29296875*y39 - 0.41015625
 *         cdef double y51 = y45*y50             # <<<<<<<<<<<<<<
 *         cdef double y52 = y27 - y28
 *         cdef double y53 = pphi*y52
 */
  __Pyx_TraceLine(806,0,__PYX_ERR(1, 806, __pyx_L1_error))
  __pyx_v_y51 = (__pyx_v_y45 * __pyx_v_y50);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":807
 *         cdef double y50 = -0.3515625*nu + 0.29296875*y39 - 0.41015625
 *         cdef double y51 = y45*y50
 *         cdef double y52 = y27 - y28             # <<<<<<<<<<<<<<
 *         cdef double y53 = pphi*y52
 *         cdef double y54 = delta*y53
 */
  __Pyx_TraceLine(807,0,__PYX_ERR(1, 807, __pyx_L1_error))
  __pyx_v_y52 = (__pyx_v_y27 - __pyx_v_y28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":808
 *         cdef double y51 = y45*y50
 *         cdef double y52 = y27 - y28
 *         cdef double y53 = pphi*y52             # <<<<<<<<<<<<<<
 *         cdef double y54 = delta*y53
 *         cdef double y55 = chi_1*X_1
 */
  __Pyx_TraceLine(808,0,__PYX_ERR(1, 808, __pyx_L1_error))
  __pyx_v_y53 = (__pyx_v_pphi * __pyx_v_y52);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":809
 *         cdef double y52 = y27 - y28
 *         cdef double y53 = pphi*y52
 *         cdef double y54 = delta*y53             # <<<<<<<<<<<<<<
 *         cdef double y55 = chi_1*X_1
 *         cdef double y56 = chi_2*X_2
 */
  __Pyx_TraceLine(809,0,__PYX_ERR(1, 809, __pyx_L1_error))
  __pyx_v_y54 = (__pyx_v_delta * __pyx_v_y53);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":810
 *         cdef double y53 = pphi*y52
 *         cdef double y54 = delta*y53
 *         cdef double y55 = chi_1*X_1             # <<<<<<<<<<<<<<
 *         cdef double y56 = chi_2*X_2
 *         cdef double y57 = y55 + y56
 */
  __Pyx_TraceLine(810,0,__PYX_ERR(1, 810, __pyx_L1_error))
  __pyx_v_y55 = (__pyx_v_chi_1 * __pyx_v_X_1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":811
 *         cdef double y54 = delta*y53
 *         cdef double y55 = chi_1*X_1
 *         cdef double y56 = chi_2*X_2             # <<<<<<<<<<<<<<
 *         cdef double y57 = y55 + y56
 *         cdef double y58 = y57**2
 */
  __Pyx_TraceLine(811,0,__PYX_ERR(1, 811, __pyx_L1_error))
  __pyx_v_y56 = (__pyx_v_chi_2 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":812
 *         cdef double y55 = chi_1*X_1
 *         cdef double y56 = chi_2*X_2
 *         cdef double y57 = y55 + y56             # <<<<<<<<<<<<<<
 *         cdef double y58 = y57**2
 *         cdef double y59 = y20 - y58
 */
  __Pyx_TraceLine(812,0,__PYX_ERR(1, 812, __pyx_L1_error))
  __pyx_v_y57 = (__pyx_v_y55 + __pyx_v_y56);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":813
 *         cdef double y56 = chi_2*X_2
 *         cdef double y57 = y55 + y56
 *         cdef double y58 = y57**2             # <<<<<<<<<<<<<<
 *         cdef double y59 = y20 - y58
 *         cdef double y60 = delta*y34
 */
  __Pyx_TraceLine(813,0,__PYX_ERR(1, 813, __pyx_L1_error))
  __pyx_v_y58 = pow(__pyx_v_y57, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":814
 *         cdef double y57 = y55 + y56
 *         cdef double y58 = y57**2
 *         cdef double y59 = y20 - y58             # <<<<<<<<<<<<<<
 *         cdef double y60 = delta*y34
 *         cdef double y61 = y59*y60
 */
  __Pyx_TraceLine(814,0,__PYX_ERR(1, 814, __pyx_L1_error))
  __pyx_v_y59 = (__pyx_v_y20 - __pyx_v_y58);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":815
 *         cdef double y58 = y57**2
 *         cdef double y59 = y20 - y58
 *         cdef double y60 = delta*y34             # <<<<<<<<<<<<<<
 *         cdef double y61 = y59*y60
 *         cdef double y62 = delta*y20
 */
  __Pyx_TraceLine(815,0,__PYX_ERR(1, 815, __pyx_L1_error))
  __pyx_v_y60 = (__pyx_v_delta * __pyx_v_y34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":816
 *         cdef double y59 = y20 - y58
 *         cdef double y60 = delta*y34
 *         cdef double y61 = y59*y60             # <<<<<<<<<<<<<<
 *         cdef double y62 = delta*y20
 *         cdef double y63 = y55 - y56
 */
  __Pyx_TraceLine(816,0,__PYX_ERR(1, 816, __pyx_L1_error))
  __pyx_v_y61 = (__pyx_v_y59 * __pyx_v_y60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":817
 *         cdef double y60 = delta*y34
 *         cdef double y61 = y59*y60
 *         cdef double y62 = delta*y20             # <<<<<<<<<<<<<<
 *         cdef double y63 = y55 - y56
 *         cdef double y64 = y12*y8 - y18 - y57*y63
 */
  __Pyx_TraceLine(817,0,__PYX_ERR(1, 817, __pyx_L1_error))
  __pyx_v_y62 = (__pyx_v_delta * __pyx_v_y20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":818
 *         cdef double y61 = y59*y60
 *         cdef double y62 = delta*y20
 *         cdef double y63 = y55 - y56             # <<<<<<<<<<<<<<
 *         cdef double y64 = y12*y8 - y18 - y57*y63
 *         cdef double y65 = -0.125*y13 - 0.125*y18 + 0.125*y58 + 0.5*y64*(0.5 - X_2) - 0.125*y7
 */
  __Pyx_TraceLine(818,0,__PYX_ERR(1, 818, __pyx_L1_error))
  __pyx_v_y63 = (__pyx_v_y55 - __pyx_v_y56);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":819
 *         cdef double y62 = delta*y20
 *         cdef double y63 = y55 - y56
 *         cdef double y64 = y12*y8 - y18 - y57*y63             # <<<<<<<<<<<<<<
 *         cdef double y65 = -0.125*y13 - 0.125*y18 + 0.125*y58 + 0.5*y64*(0.5 - X_2) - 0.125*y7
 *         cdef double y66 = y34*y65
 */
  __Pyx_TraceLine(819,0,__PYX_ERR(1, 819, __pyx_L1_error))
  __pyx_v_y64 = (((__pyx_v_y12 * __pyx_v_y8) - __pyx_v_y18) - (__pyx_v_y57 * __pyx_v_y63));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":820
 *         cdef double y63 = y55 - y56
 *         cdef double y64 = y12*y8 - y18 - y57*y63
 *         cdef double y65 = -0.125*y13 - 0.125*y18 + 0.125*y58 + 0.5*y64*(0.5 - X_2) - 0.125*y7             # <<<<<<<<<<<<<<
 *         cdef double y66 = y34*y65
 *         cdef double y67 = y13 - y18
 */
  __Pyx_TraceLine(820,0,__PYX_ERR(1, 820, __pyx_L1_error))
  __pyx_v_y65 = (((((-0.125 * __pyx_v_y13) - (0.125 * __pyx_v_y18)) + (0.125 * __pyx_v_y58)) + ((0.5 * __pyx_v_y64) * (0.5 - __pyx_v_X_2))) - (0.125 * __pyx_v_y7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":821
 *         cdef double y64 = y12*y8 - y18 - y57*y63
 *         cdef double y65 = -0.125*y13 - 0.125*y18 + 0.125*y58 + 0.5*y64*(0.5 - X_2) - 0.125*y7
 *         cdef double y66 = y34*y65             # <<<<<<<<<<<<<<
 *         cdef double y67 = y13 - y18
 *         cdef double y68 = delta*y67
 */
  __Pyx_TraceLine(821,0,__PYX_ERR(1, 821, __pyx_L1_error))
  __pyx_v_y66 = (__pyx_v_y34 * __pyx_v_y65);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":822
 *         cdef double y65 = -0.125*y13 - 0.125*y18 + 0.125*y58 + 0.5*y64*(0.5 - X_2) - 0.125*y7
 *         cdef double y66 = y34*y65
 *         cdef double y67 = y13 - y18             # <<<<<<<<<<<<<<
 *         cdef double y68 = delta*y67
 *         cdef double y69 = -1.25*y68
 */
  __Pyx_TraceLine(822,0,__PYX_ERR(1, 822, __pyx_L1_error))
  __pyx_v_y67 = (__pyx_v_y13 - __pyx_v_y18);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":823
 *         cdef double y66 = y34*y65
 *         cdef double y67 = y13 - y18
 *         cdef double y68 = delta*y67             # <<<<<<<<<<<<<<
 *         cdef double y69 = -1.25*y68
 *         cdef double y70 = delta*y64
 */
  __Pyx_TraceLine(823,0,__PYX_ERR(1, 823, __pyx_L1_error))
  __pyx_v_y68 = (__pyx_v_delta * __pyx_v_y67);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":824
 *         cdef double y67 = y13 - y18
 *         cdef double y68 = delta*y67
 *         cdef double y69 = -1.25*y68             # <<<<<<<<<<<<<<
 *         cdef double y70 = delta*y64
 *         cdef double y71 = r**2
 */
  __Pyx_TraceLine(824,0,__PYX_ERR(1, 824, __pyx_L1_error))
  __pyx_v_y69 = (-1.25 * __pyx_v_y68);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":825
 *         cdef double y68 = delta*y67
 *         cdef double y69 = -1.25*y68
 *         cdef double y70 = delta*y64             # <<<<<<<<<<<<<<
 *         cdef double y71 = r**2
 *         cdef double y72 = 6.0*y71
 */
  __Pyx_TraceLine(825,0,__PYX_ERR(1, 825, __pyx_L1_error))
  __pyx_v_y70 = (__pyx_v_delta * __pyx_v_y64);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":826
 *         cdef double y69 = -1.25*y68
 *         cdef double y70 = delta*y64
 *         cdef double y71 = r**2             # <<<<<<<<<<<<<<
 *         cdef double y72 = 6.0*y71
 *         cdef double y73 = 2.0*y13 + 2.0*y18 + 2.0*y7
 */
  __Pyx_TraceLine(826,0,__PYX_ERR(1, 826, __pyx_L1_error))
  __pyx_v_y71 = pow(__pyx_v_r, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":827
 *         cdef double y70 = delta*y64
 *         cdef double y71 = r**2
 *         cdef double y72 = 6.0*y71             # <<<<<<<<<<<<<<
 *         cdef double y73 = 2.0*y13 + 2.0*y18 + 2.0*y7
 *         cdef double y74 = y22**(-2)
 */
  __Pyx_TraceLine(827,0,__PYX_ERR(1, 827, __pyx_L1_error))
  __pyx_v_y72 = (6.0 * __pyx_v_y71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":828
 *         cdef double y71 = r**2
 *         cdef double y72 = 6.0*y71
 *         cdef double y73 = 2.0*y13 + 2.0*y18 + 2.0*y7             # <<<<<<<<<<<<<<
 *         cdef double y74 = y22**(-2)
 *         cdef double y75 = 3.0*nu
 */
  __Pyx_TraceLine(828,0,__PYX_ERR(1, 828, __pyx_L1_error))
  __pyx_v_y73 = (((2.0 * __pyx_v_y13) + (2.0 * __pyx_v_y18)) + (2.0 * __pyx_v_y7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":829
 *         cdef double y72 = 6.0*y71
 *         cdef double y73 = 2.0*y13 + 2.0*y18 + 2.0*y7
 *         cdef double y74 = y22**(-2)             # <<<<<<<<<<<<<<
 *         cdef double y75 = 3.0*nu
 *         cdef double y76 = y36*y75
 */
  __Pyx_TraceLine(829,0,__PYX_ERR(1, 829, __pyx_L1_error))
  __pyx_v_y74 = pow(__pyx_v_y22, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":830
 *         cdef double y73 = 2.0*y13 + 2.0*y18 + 2.0*y7
 *         cdef double y74 = y22**(-2)
 *         cdef double y75 = 3.0*nu             # <<<<<<<<<<<<<<
 *         cdef double y76 = y36*y75
 *         cdef double y77 = y71**(-1)
 */
  __Pyx_TraceLine(830,0,__PYX_ERR(1, 830, __pyx_L1_error))
  __pyx_v_y75 = (3.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":831
 *         cdef double y74 = y22**(-2)
 *         cdef double y75 = 3.0*nu
 *         cdef double y76 = y36*y75             # <<<<<<<<<<<<<<
 *         cdef double y77 = y71**(-1)
 *         cdef double y78 = 0.71875*nu - 0.09375
 */
  __Pyx_TraceLine(831,0,__PYX_ERR(1, 831, __pyx_L1_error))
  __pyx_v_y76 = (__pyx_v_y36 * __pyx_v_y75);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":832
 *         cdef double y75 = 3.0*nu
 *         cdef double y76 = y36*y75
 *         cdef double y77 = y71**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y78 = 0.71875*nu - 0.09375
 *         cdef double y79 = y32*y34
 */
  __Pyx_TraceLine(832,0,__PYX_ERR(1, 832, __pyx_L1_error))
  __pyx_v_y77 = pow(__pyx_v_y71, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":833
 *         cdef double y76 = y36*y75
 *         cdef double y77 = y71**(-1)
 *         cdef double y78 = 0.71875*nu - 0.09375             # <<<<<<<<<<<<<<
 *         cdef double y79 = y32*y34
 *         cdef double y80 = 2.0*y79
 */
  __Pyx_TraceLine(833,0,__PYX_ERR(1, 833, __pyx_L1_error))
  __pyx_v_y78 = ((0.71875 * __pyx_v_nu) - 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":834
 *         cdef double y77 = y71**(-1)
 *         cdef double y78 = 0.71875*nu - 0.09375
 *         cdef double y79 = y32*y34             # <<<<<<<<<<<<<<
 *         cdef double y80 = 2.0*y79
 *         cdef double y81 = 3.0*y36
 */
  __Pyx_TraceLine(834,0,__PYX_ERR(1, 834, __pyx_L1_error))
  __pyx_v_y79 = (__pyx_v_y32 * __pyx_v_y34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":835
 *         cdef double y78 = 0.71875*nu - 0.09375
 *         cdef double y79 = y32*y34
 *         cdef double y80 = 2.0*y79             # <<<<<<<<<<<<<<
 *         cdef double y81 = 3.0*y36
 *         cdef double y82 = y34*y81
 */
  __Pyx_TraceLine(835,0,__PYX_ERR(1, 835, __pyx_L1_error))
  __pyx_v_y80 = (2.0 * __pyx_v_y79);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":836
 *         cdef double y79 = y32*y34
 *         cdef double y80 = 2.0*y79
 *         cdef double y81 = 3.0*y36             # <<<<<<<<<<<<<<
 *         cdef double y82 = y34*y81
 *         cdef double y83 = 4.0*y25
 */
  __Pyx_TraceLine(836,0,__PYX_ERR(1, 836, __pyx_L1_error))
  __pyx_v_y81 = (3.0 * __pyx_v_y36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":837
 *         cdef double y80 = 2.0*y79
 *         cdef double y81 = 3.0*y36
 *         cdef double y82 = y34*y81             # <<<<<<<<<<<<<<
 *         cdef double y83 = 4.0*y25
 *         cdef double y84 = y29*(-y32*(-11.0625*nu + 1.13541666666667*y39 - 0.15625) - y33*y80 - y40*y82 - y47*y83 - y77*y78)
 */
  __Pyx_TraceLine(837,0,__PYX_ERR(1, 837, __pyx_L1_error))
  __pyx_v_y82 = (__pyx_v_y34 * __pyx_v_y81);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":838
 *         cdef double y81 = 3.0*y36
 *         cdef double y82 = y34*y81
 *         cdef double y83 = 4.0*y25             # <<<<<<<<<<<<<<
 *         cdef double y84 = y29*(-y32*(-11.0625*nu + 1.13541666666667*y39 - 0.15625) - y33*y80 - y40*y82 - y47*y83 - y77*y78)
 *         cdef double y85 = 0.34375*nu + 0.09375
 */
  __Pyx_TraceLine(838,0,__PYX_ERR(1, 838, __pyx_L1_error))
  __pyx_v_y83 = (4.0 * __pyx_v_y25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":839
 *         cdef double y82 = y34*y81
 *         cdef double y83 = 4.0*y25
 *         cdef double y84 = y29*(-y32*(-11.0625*nu + 1.13541666666667*y39 - 0.15625) - y33*y80 - y40*y82 - y47*y83 - y77*y78)             # <<<<<<<<<<<<<<
 *         cdef double y85 = 0.34375*nu + 0.09375
 *         cdef double y86 = -y32*(-0.0625*nu + 1.07291666666667*y39 + 0.15625) - y48*y80 - y49*y82 - y51*y83 - y77*y85
 */
  __Pyx_TraceLine(839,0,__PYX_ERR(1, 839, __pyx_L1_error))
  __pyx_v_y84 = (__pyx_v_y29 * ((((((-__pyx_v_y32) * (((-11.0625 * __pyx_v_nu) + (1.13541666666667 * __pyx_v_y39)) - 0.15625)) - (__pyx_v_y33 * __pyx_v_y80)) - (__pyx_v_y40 * __pyx_v_y82)) - (__pyx_v_y47 * __pyx_v_y83)) - (__pyx_v_y77 * __pyx_v_y78)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":840
 *         cdef double y83 = 4.0*y25
 *         cdef double y84 = y29*(-y32*(-11.0625*nu + 1.13541666666667*y39 - 0.15625) - y33*y80 - y40*y82 - y47*y83 - y77*y78)
 *         cdef double y85 = 0.34375*nu + 0.09375             # <<<<<<<<<<<<<<
 *         cdef double y86 = -y32*(-0.0625*nu + 1.07291666666667*y39 + 0.15625) - y48*y80 - y49*y82 - y51*y83 - y77*y85
 *         cdef double y87 = y59*(0.416666666666667 - 0.833333333333333*X_2)
 */
  __Pyx_TraceLine(840,0,__PYX_ERR(1, 840, __pyx_L1_error))
  __pyx_v_y85 = ((0.34375 * __pyx_v_nu) + 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":841
 *         cdef double y84 = y29*(-y32*(-11.0625*nu + 1.13541666666667*y39 - 0.15625) - y33*y80 - y40*y82 - y47*y83 - y77*y78)
 *         cdef double y85 = 0.34375*nu + 0.09375
 *         cdef double y86 = -y32*(-0.0625*nu + 1.07291666666667*y39 + 0.15625) - y48*y80 - y49*y82 - y51*y83 - y77*y85             # <<<<<<<<<<<<<<
 *         cdef double y87 = y59*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double y88 = y52*(-y32*(0.0833333333333333*y62 + y87) + 0.375*y36*y61)
 */
  __Pyx_TraceLine(841,0,__PYX_ERR(1, 841, __pyx_L1_error))
  __pyx_v_y86 = ((((((-__pyx_v_y32) * (((-0.0625 * __pyx_v_nu) + (1.07291666666667 * __pyx_v_y39)) + 0.15625)) - (__pyx_v_y48 * __pyx_v_y80)) - (__pyx_v_y49 * __pyx_v_y82)) - (__pyx_v_y51 * __pyx_v_y83)) - (__pyx_v_y77 * __pyx_v_y85));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":842
 *         cdef double y85 = 0.34375*nu + 0.09375
 *         cdef double y86 = -y32*(-0.0625*nu + 1.07291666666667*y39 + 0.15625) - y48*y80 - y49*y82 - y51*y83 - y77*y85
 *         cdef double y87 = y59*(0.416666666666667 - 0.833333333333333*X_2)             # <<<<<<<<<<<<<<
 *         cdef double y88 = y52*(-y32*(0.0833333333333333*y62 + y87) + 0.375*y36*y61)
 *         cdef double y89 = y29*(y32*(1.25*y13 + 1.25*y18 - 0.75*y58 - 0.416666666666667*y68 + 1.25*y7 + 1.66666666666667*y70) - y66*y81)
 */
  __Pyx_TraceLine(842,0,__PYX_ERR(1, 842, __pyx_L1_error))
  __pyx_v_y87 = (__pyx_v_y59 * (0.416666666666667 - (0.833333333333333 * __pyx_v_X_2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":843
 *         cdef double y86 = -y32*(-0.0625*nu + 1.07291666666667*y39 + 0.15625) - y48*y80 - y49*y82 - y51*y83 - y77*y85
 *         cdef double y87 = y59*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double y88 = y52*(-y32*(0.0833333333333333*y62 + y87) + 0.375*y36*y61)             # <<<<<<<<<<<<<<
 *         cdef double y89 = y29*(y32*(1.25*y13 + 1.25*y18 - 0.75*y58 - 0.416666666666667*y68 + 1.25*y7 + 1.66666666666667*y70) - y66*y81)
 *         cdef double y90 = nu*y32
 */
  __Pyx_TraceLine(843,0,__PYX_ERR(1, 843, __pyx_L1_error))
  __pyx_v_y88 = (__pyx_v_y52 * (((-__pyx_v_y32) * ((0.0833333333333333 * __pyx_v_y62) + __pyx_v_y87)) + ((0.375 * __pyx_v_y36) * __pyx_v_y61)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":844
 *         cdef double y87 = y59*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double y88 = y52*(-y32*(0.0833333333333333*y62 + y87) + 0.375*y36*y61)
 *         cdef double y89 = y29*(y32*(1.25*y13 + 1.25*y18 - 0.75*y58 - 0.416666666666667*y68 + 1.25*y7 + 1.66666666666667*y70) - y66*y81)             # <<<<<<<<<<<<<<
 *         cdef double y90 = nu*y32
 *         cdef double y91 = r**(-1)
 */
  __Pyx_TraceLine(844,0,__PYX_ERR(1, 844, __pyx_L1_error))
  __pyx_v_y89 = (__pyx_v_y29 * ((__pyx_v_y32 * ((((((1.25 * __pyx_v_y13) + (1.25 * __pyx_v_y18)) - (0.75 * __pyx_v_y58)) - (0.416666666666667 * __pyx_v_y68)) + (1.25 * __pyx_v_y7)) + (1.66666666666667 * __pyx_v_y70))) - (__pyx_v_y66 * __pyx_v_y81)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":845
 *         cdef double y88 = y52*(-y32*(0.0833333333333333*y62 + y87) + 0.375*y36*y61)
 *         cdef double y89 = y29*(y32*(1.25*y13 + 1.25*y18 - 0.75*y58 - 0.416666666666667*y68 + 1.25*y7 + 1.66666666666667*y70) - y66*y81)
 *         cdef double y90 = nu*y32             # <<<<<<<<<<<<<<
 *         cdef double y91 = r**(-1)
 *         cdef double y92 = y34*y77
 */
  __Pyx_TraceLine(845,0,__PYX_ERR(1, 845, __pyx_L1_error))
  __pyx_v_y90 = (__pyx_v_nu * __pyx_v_y32);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":846
 *         cdef double y89 = y29*(y32*(1.25*y13 + 1.25*y18 - 0.75*y58 - 0.416666666666667*y68 + 1.25*y7 + 1.66666666666667*y70) - y66*y81)
 *         cdef double y90 = nu*y32
 *         cdef double y91 = r**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y92 = y34*y77
 *         cdef double y93 = y36*y45
 */
  __Pyx_TraceLine(846,0,__PYX_ERR(1, 846, __pyx_L1_error))
  __pyx_v_y91 = pow(__pyx_v_r, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":847
 *         cdef double y90 = nu*y32
 *         cdef double y91 = r**(-1)
 *         cdef double y92 = y34*y77             # <<<<<<<<<<<<<<
 *         cdef double y93 = y36*y45
 *         cdef double y94 = y29*(y33*y92 + y40*y79 + y46*y93 + y77*(-5.53125*nu + 0.567708333333333*y39 - 0.078125) + y78*y91 + 1.75)
 */
  __Pyx_TraceLine(847,0,__PYX_ERR(1, 847, __pyx_L1_error))
  __pyx_v_y92 = (__pyx_v_y34 * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":848
 *         cdef double y91 = r**(-1)
 *         cdef double y92 = y34*y77
 *         cdef double y93 = y36*y45             # <<<<<<<<<<<<<<
 *         cdef double y94 = y29*(y33*y92 + y40*y79 + y46*y93 + y77*(-5.53125*nu + 0.567708333333333*y39 - 0.078125) + y78*y91 + 1.75)
 *         cdef double y95 = y48*y92 + y49*y79 + y50*y93 + y77*(-0.03125*nu + 0.536458333333333*y39 + 0.078125) + y85*y91 + 0.25
 */
  __Pyx_TraceLine(848,0,__PYX_ERR(1, 848, __pyx_L1_error))
  __pyx_v_y93 = (__pyx_v_y36 * __pyx_v_y45);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":849
 *         cdef double y92 = y34*y77
 *         cdef double y93 = y36*y45
 *         cdef double y94 = y29*(y33*y92 + y40*y79 + y46*y93 + y77*(-5.53125*nu + 0.567708333333333*y39 - 0.078125) + y78*y91 + 1.75)             # <<<<<<<<<<<<<<
 *         cdef double y95 = y48*y92 + y49*y79 + y50*y93 + y77*(-0.03125*nu + 0.536458333333333*y39 + 0.078125) + y85*y91 + 0.25
 *         cdef double y96 = y32*y59
 */
  __Pyx_TraceLine(849,0,__PYX_ERR(1, 849, __pyx_L1_error))
  __pyx_v_y94 = (__pyx_v_y29 * ((((((__pyx_v_y33 * __pyx_v_y92) + (__pyx_v_y40 * __pyx_v_y79)) + (__pyx_v_y46 * __pyx_v_y93)) + (__pyx_v_y77 * (((-5.53125 * __pyx_v_nu) + (0.567708333333333 * __pyx_v_y39)) - 0.078125))) + (__pyx_v_y78 * __pyx_v_y91)) + 1.75));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":850
 *         cdef double y93 = y36*y45
 *         cdef double y94 = y29*(y33*y92 + y40*y79 + y46*y93 + y77*(-5.53125*nu + 0.567708333333333*y39 - 0.078125) + y78*y91 + 1.75)
 *         cdef double y95 = y48*y92 + y49*y79 + y50*y93 + y77*(-0.03125*nu + 0.536458333333333*y39 + 0.078125) + y85*y91 + 0.25             # <<<<<<<<<<<<<<
 *         cdef double y96 = y32*y59
 *         cdef double y97 = y52*(-0.125*y60*y96 + y77*(0.0416666666666667*y62 + 0.5*y87))
 */
  __Pyx_TraceLine(850,0,__PYX_ERR(1, 850, __pyx_L1_error))
  __pyx_v_y95 = ((((((__pyx_v_y48 * __pyx_v_y92) + (__pyx_v_y49 * __pyx_v_y79)) + (__pyx_v_y50 * __pyx_v_y93)) + (__pyx_v_y77 * (((-0.03125 * __pyx_v_nu) + (0.536458333333333 * __pyx_v_y39)) + 0.078125))) + (__pyx_v_y85 * __pyx_v_y91)) + 0.25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":851
 *         cdef double y94 = y29*(y33*y92 + y40*y79 + y46*y93 + y77*(-5.53125*nu + 0.567708333333333*y39 - 0.078125) + y78*y91 + 1.75)
 *         cdef double y95 = y48*y92 + y49*y79 + y50*y93 + y77*(-0.03125*nu + 0.536458333333333*y39 + 0.078125) + y85*y91 + 0.25
 *         cdef double y96 = y32*y59             # <<<<<<<<<<<<<<
 *         cdef double y97 = y52*(-0.125*y60*y96 + y77*(0.0416666666666667*y62 + 0.5*y87))
 *         cdef double y98 = y65*y79
 */
  __Pyx_TraceLine(851,0,__PYX_ERR(1, 851, __pyx_L1_error))
  __pyx_v_y96 = (__pyx_v_y32 * __pyx_v_y59);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":852
 *         cdef double y95 = y48*y92 + y49*y79 + y50*y93 + y77*(-0.03125*nu + 0.536458333333333*y39 + 0.078125) + y85*y91 + 0.25
 *         cdef double y96 = y32*y59
 *         cdef double y97 = y52*(-0.125*y60*y96 + y77*(0.0416666666666667*y62 + 0.5*y87))             # <<<<<<<<<<<<<<
 *         cdef double y98 = y65*y79
 *         cdef double y99 = y29*(y77*(0.208333333333333*delta*y67 - 0.625*y13 - 0.625*y18 + 0.375*y58 - 0.625*y7 - 0.833333333333333*y70) + y98)
 */
  __Pyx_TraceLine(852,0,__PYX_ERR(1, 852, __pyx_L1_error))
  __pyx_v_y97 = (__pyx_v_y52 * (((-0.125 * __pyx_v_y60) * __pyx_v_y96) + (__pyx_v_y77 * ((0.0416666666666667 * __pyx_v_y62) + (0.5 * __pyx_v_y87)))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":853
 *         cdef double y96 = y32*y59
 *         cdef double y97 = y52*(-0.125*y60*y96 + y77*(0.0416666666666667*y62 + 0.5*y87))
 *         cdef double y98 = y65*y79             # <<<<<<<<<<<<<<
 *         cdef double y99 = y29*(y77*(0.208333333333333*delta*y67 - 0.625*y13 - 0.625*y18 + 0.375*y58 - 0.625*y7 - 0.833333333333333*y70) + y98)
 *         cdef double y100 = prst**4
 */
  __Pyx_TraceLine(853,0,__PYX_ERR(1, 853, __pyx_L1_error))
  __pyx_v_y98 = (__pyx_v_y65 * __pyx_v_y79);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":854
 *         cdef double y97 = y52*(-0.125*y60*y96 + y77*(0.0416666666666667*y62 + 0.5*y87))
 *         cdef double y98 = y65*y79
 *         cdef double y99 = y29*(y77*(0.208333333333333*delta*y67 - 0.625*y13 - 0.625*y18 + 0.375*y58 - 0.625*y7 - 0.833333333333333*y70) + y98)             # <<<<<<<<<<<<<<
 *         cdef double y100 = prst**4
 *         cdef double y101 = r**(-4.5)
 */
  __Pyx_TraceLine(854,0,__PYX_ERR(1, 854, __pyx_L1_error))
  __pyx_v_y99 = (__pyx_v_y29 * ((__pyx_v_y77 * (((((((0.208333333333333 * __pyx_v_delta) * __pyx_v_y67) - (0.625 * __pyx_v_y13)) - (0.625 * __pyx_v_y18)) + (0.375 * __pyx_v_y58)) - (0.625 * __pyx_v_y7)) - (0.833333333333333 * __pyx_v_y70))) + __pyx_v_y98));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":855
 *         cdef double y98 = y65*y79
 *         cdef double y99 = y29*(y77*(0.208333333333333*delta*y67 - 0.625*y13 - 0.625*y18 + 0.375*y58 - 0.625*y7 - 0.833333333333333*y70) + y98)
 *         cdef double y100 = prst**4             # <<<<<<<<<<<<<<
 *         cdef double y101 = r**(-4.5)
 *         cdef double y102 = nu*y101
 */
  __Pyx_TraceLine(855,0,__PYX_ERR(1, 855, __pyx_L1_error))
  __pyx_v_y100 = pow(__pyx_v_prst, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":856
 *         cdef double y99 = y29*(y77*(0.208333333333333*delta*y67 - 0.625*y13 - 0.625*y18 + 0.375*y58 - 0.625*y7 - 0.833333333333333*y70) + y98)
 *         cdef double y100 = prst**4
 *         cdef double y101 = r**(-4.5)             # <<<<<<<<<<<<<<
 *         cdef double y102 = nu*y101
 *         cdef double y103 = r**(-3.5)
 */
  __Pyx_TraceLine(856,0,__PYX_ERR(1, 856, __pyx_L1_error))
  __pyx_v_y101 = pow(__pyx_v_r, -4.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":857
 *         cdef double y100 = prst**4
 *         cdef double y101 = r**(-4.5)
 *         cdef double y102 = nu*y101             # <<<<<<<<<<<<<<
 *         cdef double y103 = r**(-3.5)
 *         cdef double y104 = prst**6
 */
  __Pyx_TraceLine(857,0,__PYX_ERR(1, 857, __pyx_L1_error))
  __pyx_v_y102 = (__pyx_v_nu * __pyx_v_y101);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":858
 *         cdef double y101 = r**(-4.5)
 *         cdef double y102 = nu*y101
 *         cdef double y103 = r**(-3.5)             # <<<<<<<<<<<<<<
 *         cdef double y104 = prst**6
 *         cdef double y105 = nu*y104
 */
  __Pyx_TraceLine(858,0,__PYX_ERR(1, 858, __pyx_L1_error))
  __pyx_v_y103 = pow(__pyx_v_r, -3.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":859
 *         cdef double y102 = nu*y101
 *         cdef double y103 = r**(-3.5)
 *         cdef double y104 = prst**6             # <<<<<<<<<<<<<<
 *         cdef double y105 = nu*y104
 *         cdef double y106 = r**(-2.5)
 */
  __Pyx_TraceLine(859,0,__PYX_ERR(1, 859, __pyx_L1_error))
  __pyx_v_y104 = pow(__pyx_v_prst, 6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":860
 *         cdef double y103 = r**(-3.5)
 *         cdef double y104 = prst**6
 *         cdef double y105 = nu*y104             # <<<<<<<<<<<<<<
 *         cdef double y106 = r**(-2.5)
 *         cdef double y107 = prst**8
 */
  __Pyx_TraceLine(860,0,__PYX_ERR(1, 860, __pyx_L1_error))
  __pyx_v_y105 = (__pyx_v_nu * __pyx_v_y104);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":861
 *         cdef double y104 = prst**6
 *         cdef double y105 = nu*y104
 *         cdef double y106 = r**(-2.5)             # <<<<<<<<<<<<<<
 *         cdef double y107 = prst**8
 *         cdef double y108 = nu*y107
 */
  __Pyx_TraceLine(861,0,__PYX_ERR(1, 861, __pyx_L1_error))
  __pyx_v_y106 = pow(__pyx_v_r, -2.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":862
 *         cdef double y105 = nu*y104
 *         cdef double y106 = r**(-2.5)
 *         cdef double y107 = prst**8             # <<<<<<<<<<<<<<
 *         cdef double y108 = nu*y107
 *         cdef double y109 = nu*y91
 */
  __Pyx_TraceLine(862,0,__PYX_ERR(1, 862, __pyx_L1_error))
  __pyx_v_y107 = pow(__pyx_v_prst, 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":863
 *         cdef double y106 = r**(-2.5)
 *         cdef double y107 = prst**8
 *         cdef double y108 = nu*y107             # <<<<<<<<<<<<<<
 *         cdef double y109 = nu*y91
 *         cdef double y110 = 0.121954868780449*y107
 */
  __Pyx_TraceLine(863,0,__PYX_ERR(1, 863, __pyx_L1_error))
  __pyx_v_y108 = (__pyx_v_nu * __pyx_v_y107);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":864
 *         cdef double y107 = prst**8
 *         cdef double y108 = nu*y107
 *         cdef double y109 = nu*y91             # <<<<<<<<<<<<<<
 *         cdef double y110 = 0.121954868780449*y107
 *         cdef double y111 = 8.0*nu - 6.0*y39
 */
  __Pyx_TraceLine(864,0,__PYX_ERR(1, 864, __pyx_L1_error))
  __pyx_v_y109 = (__pyx_v_nu * __pyx_v_y91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":865
 *         cdef double y108 = nu*y107
 *         cdef double y109 = nu*y91
 *         cdef double y110 = 0.121954868780449*y107             # <<<<<<<<<<<<<<
 *         cdef double y111 = 8.0*nu - 6.0*y39
 *         cdef double y112 = y100*y111
 */
  __Pyx_TraceLine(865,0,__PYX_ERR(1, 865, __pyx_L1_error))
  __pyx_v_y110 = (0.121954868780449 * __pyx_v_y107);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":866
 *         cdef double y109 = nu*y91
 *         cdef double y110 = 0.121954868780449*y107
 *         cdef double y111 = 8.0*nu - 6.0*y39             # <<<<<<<<<<<<<<
 *         cdef double y112 = y100*y111
 *         cdef double y113 = nu**3
 */
  __Pyx_TraceLine(866,0,__PYX_ERR(1, 866, __pyx_L1_error))
  __pyx_v_y111 = ((8.0 * __pyx_v_nu) - (6.0 * __pyx_v_y39));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":867
 *         cdef double y110 = 0.121954868780449*y107
 *         cdef double y111 = 8.0*nu - 6.0*y39
 *         cdef double y112 = y100*y111             # <<<<<<<<<<<<<<
 *         cdef double y113 = nu**3
 *         cdef double y114 = 92.7110442849544*nu + 10.0*y113 - 131.0*y39
 */
  __Pyx_TraceLine(867,0,__PYX_ERR(1, 867, __pyx_L1_error))
  __pyx_v_y112 = (__pyx_v_y100 * __pyx_v_y111);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":868
 *         cdef double y111 = 8.0*nu - 6.0*y39
 *         cdef double y112 = y100*y111
 *         cdef double y113 = nu**3             # <<<<<<<<<<<<<<
 *         cdef double y114 = 92.7110442849544*nu + 10.0*y113 - 131.0*y39
 *         cdef double y115 = y100*y114
 */
  __Pyx_TraceLine(868,0,__PYX_ERR(1, 868, __pyx_L1_error))
  __pyx_v_y113 = pow(__pyx_v_nu, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":869
 *         cdef double y112 = y100*y111
 *         cdef double y113 = nu**3
 *         cdef double y114 = 92.7110442849544*nu + 10.0*y113 - 131.0*y39             # <<<<<<<<<<<<<<
 *         cdef double y115 = y100*y114
 *         cdef double y116 = -2.78300763695006*nu + 6.0*y113 - 5.4*y39
 */
  __Pyx_TraceLine(869,0,__PYX_ERR(1, 869, __pyx_L1_error))
  __pyx_v_y114 = (((92.7110442849544 * __pyx_v_nu) + (10.0 * __pyx_v_y113)) - (131.0 * __pyx_v_y39));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":870
 *         cdef double y113 = nu**3
 *         cdef double y114 = 92.7110442849544*nu + 10.0*y113 - 131.0*y39
 *         cdef double y115 = y100*y114             # <<<<<<<<<<<<<<
 *         cdef double y116 = -2.78300763695006*nu + 6.0*y113 - 5.4*y39
 *         cdef double y117 = y104*y116
 */
  __Pyx_TraceLine(870,0,__PYX_ERR(1, 870, __pyx_L1_error))
  __pyx_v_y115 = (__pyx_v_y100 * __pyx_v_y114);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":871
 *         cdef double y114 = 92.7110442849544*nu + 10.0*y113 - 131.0*y39
 *         cdef double y115 = y100*y114
 *         cdef double y116 = -2.78300763695006*nu + 6.0*y113 - 5.4*y39             # <<<<<<<<<<<<<<
 *         cdef double y117 = y104*y116
 *         cdef double y118 = nu**4
 */
  __Pyx_TraceLine(871,0,__PYX_ERR(1, 871, __pyx_L1_error))
  __pyx_v_y116 = (((-2.78300763695006 * __pyx_v_nu) + (6.0 * __pyx_v_y113)) - (5.4 * __pyx_v_y39));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":872
 *         cdef double y115 = y100*y114
 *         cdef double y116 = -2.78300763695006*nu + 6.0*y113 - 5.4*y39
 *         cdef double y117 = y104*y116             # <<<<<<<<<<<<<<
 *         cdef double y118 = nu**4
 *         cdef double y119 = -33.9782122170436*nu + 188.0*y113 - 14.0*y118 - 89.5298327361234*y39
 */
  __Pyx_TraceLine(872,0,__PYX_ERR(1, 872, __pyx_L1_error))
  __pyx_v_y117 = (__pyx_v_y104 * __pyx_v_y116);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":873
 *         cdef double y116 = -2.78300763695006*nu + 6.0*y113 - 5.4*y39
 *         cdef double y117 = y104*y116
 *         cdef double y118 = nu**4             # <<<<<<<<<<<<<<
 *         cdef double y119 = -33.9782122170436*nu + 188.0*y113 - 14.0*y118 - 89.5298327361234*y39
 *         cdef double y120 = y104*y119
 */
  __Pyx_TraceLine(873,0,__PYX_ERR(1, 873, __pyx_L1_error))
  __pyx_v_y118 = pow(__pyx_v_nu, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":874
 *         cdef double y117 = y104*y116
 *         cdef double y118 = nu**4
 *         cdef double y119 = -33.9782122170436*nu + 188.0*y113 - 14.0*y118 - 89.5298327361234*y39             # <<<<<<<<<<<<<<
 *         cdef double y120 = y104*y119
 *         cdef double y121 = 1.38977750996128*nu + 3.42857142857143*y113 - 6.0*y118 + 3.33842023648322*y39
 */
  __Pyx_TraceLine(874,0,__PYX_ERR(1, 874, __pyx_L1_error))
  __pyx_v_y119 = ((((-33.9782122170436 * __pyx_v_nu) + (188.0 * __pyx_v_y113)) - (14.0 * __pyx_v_y118)) - (89.5298327361234 * __pyx_v_y39));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":875
 *         cdef double y118 = nu**4
 *         cdef double y119 = -33.9782122170436*nu + 188.0*y113 - 14.0*y118 - 89.5298327361234*y39
 *         cdef double y120 = y104*y119             # <<<<<<<<<<<<<<
 *         cdef double y121 = 1.38977750996128*nu + 3.42857142857143*y113 - 6.0*y118 + 3.33842023648322*y39
 *         cdef double y122 = y107*y121
 */
  __Pyx_TraceLine(875,0,__PYX_ERR(1, 875, __pyx_L1_error))
  __pyx_v_y120 = (__pyx_v_y104 * __pyx_v_y119);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":876
 *         cdef double y119 = -33.9782122170436*nu + 188.0*y113 - 14.0*y118 - 89.5298327361234*y39
 *         cdef double y120 = y104*y119
 *         cdef double y121 = 1.38977750996128*nu + 3.42857142857143*y113 - 6.0*y118 + 3.33842023648322*y39             # <<<<<<<<<<<<<<
 *         cdef double y122 = y107*y121
 *         cdef double y123 = log(r)
 */
  __Pyx_TraceLine(876,0,__PYX_ERR(1, 876, __pyx_L1_error))
  __pyx_v_y121 = ((((1.38977750996128 * __pyx_v_nu) + (3.42857142857143 * __pyx_v_y113)) - (6.0 * __pyx_v_y118)) + (3.33842023648322 * __pyx_v_y39));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":877
 *         cdef double y120 = y104*y119
 *         cdef double y121 = 1.38977750996128*nu + 3.42857142857143*y113 - 6.0*y118 + 3.33842023648322*y39
 *         cdef double y122 = y107*y121             # <<<<<<<<<<<<<<
 *         cdef double y123 = log(r)
 *         cdef double y124 = nu*(452.542166996693 - 51.6952380952381*y123) + 602.318540416564*y113 + y39*(118.4*y123 - 1796.13660498019)
 */
  __Pyx_TraceLine(877,0,__PYX_ERR(1, 877, __pyx_L1_error))
  __pyx_v_y122 = (__pyx_v_y107 * __pyx_v_y121);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":878
 *         cdef double y121 = 1.38977750996128*nu + 3.42857142857143*y113 - 6.0*y118 + 3.33842023648322*y39
 *         cdef double y122 = y107*y121
 *         cdef double y123 = log(r)             # <<<<<<<<<<<<<<
 *         cdef double y124 = nu*(452.542166996693 - 51.6952380952381*y123) + 602.318540416564*y113 + y39*(118.4*y123 - 1796.13660498019)
 *         cdef double y125 = y100*y36
 */
  __Pyx_TraceLine(878,0,__PYX_ERR(1, 878, __pyx_L1_error))
  __pyx_v_y123 = log(__pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":879
 *         cdef double y122 = y107*y121
 *         cdef double y123 = log(r)
 *         cdef double y124 = nu*(452.542166996693 - 51.6952380952381*y123) + 602.318540416564*y113 + y39*(118.4*y123 - 1796.13660498019)             # <<<<<<<<<<<<<<
 *         cdef double y125 = y100*y36
 *         cdef double y126 = r*y21 + y35
 */
  __Pyx_TraceLine(879,0,__PYX_ERR(1, 879, __pyx_L1_error))
  __pyx_v_y124 = (((__pyx_v_nu * (452.542166996693 - (51.6952380952381 * __pyx_v_y123))) + (602.318540416564 * __pyx_v_y113)) + (__pyx_v_y39 * ((118.4 * __pyx_v_y123) - 1796.13660498019)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":880
 *         cdef double y123 = log(r)
 *         cdef double y124 = nu*(452.542166996693 - 51.6952380952381*y123) + 602.318540416564*y113 + y39*(118.4*y123 - 1796.13660498019)
 *         cdef double y125 = y100*y36             # <<<<<<<<<<<<<<
 *         cdef double y126 = r*y21 + y35
 *         cdef double y127 = y126**(-1)
 */
  __Pyx_TraceLine(880,0,__PYX_ERR(1, 880, __pyx_L1_error))
  __pyx_v_y125 = (__pyx_v_y100 * __pyx_v_y36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":881
 *         cdef double y124 = nu*(452.542166996693 - 51.6952380952381*y123) + 602.318540416564*y113 + y39*(118.4*y123 - 1796.13660498019)
 *         cdef double y125 = y100*y36
 *         cdef double y126 = r*y21 + y35             # <<<<<<<<<<<<<<
 *         cdef double y127 = y126**(-1)
 *         cdef double y128 = y29**2
 */
  __Pyx_TraceLine(881,0,__PYX_ERR(1, 881, __pyx_L1_error))
  __pyx_v_y126 = ((__pyx_v_r * __pyx_v_y21) + __pyx_v_y35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":882
 *         cdef double y125 = y100*y36
 *         cdef double y126 = r*y21 + y35
 *         cdef double y127 = y126**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y128 = y29**2
 *         cdef double y129 = y127*y128
 */
  __Pyx_TraceLine(882,0,__PYX_ERR(1, 882, __pyx_L1_error))
  __pyx_v_y127 = pow(__pyx_v_y126, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":883
 *         cdef double y126 = r*y21 + y35
 *         cdef double y127 = y126**(-1)
 *         cdef double y128 = y29**2             # <<<<<<<<<<<<<<
 *         cdef double y129 = y127*y128
 *         cdef double y130 = y129*y91
 */
  __Pyx_TraceLine(883,0,__PYX_ERR(1, 883, __pyx_L1_error))
  __pyx_v_y128 = pow(__pyx_v_y29, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":884
 *         cdef double y127 = y126**(-1)
 *         cdef double y128 = y29**2
 *         cdef double y129 = y127*y128             # <<<<<<<<<<<<<<
 *         cdef double y130 = y129*y91
 *         cdef double y131 = y130*y34
 */
  __Pyx_TraceLine(884,0,__PYX_ERR(1, 884, __pyx_L1_error))
  __pyx_v_y129 = (__pyx_v_y127 * __pyx_v_y128);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":885
 *         cdef double y128 = y29**2
 *         cdef double y129 = y127*y128
 *         cdef double y130 = y129*y91             # <<<<<<<<<<<<<<
 *         cdef double y131 = y130*y34
 *         cdef double y132 = y59*y77
 */
  __Pyx_TraceLine(885,0,__PYX_ERR(1, 885, __pyx_L1_error))
  __pyx_v_y130 = (__pyx_v_y129 * __pyx_v_y91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":886
 *         cdef double y129 = y127*y128
 *         cdef double y130 = y129*y91
 *         cdef double y131 = y130*y34             # <<<<<<<<<<<<<<
 *         cdef double y132 = y59*y77
 *         cdef double y133 = y64*(0.75 - 1.5*X_2)
 */
  __Pyx_TraceLine(886,0,__PYX_ERR(1, 886, __pyx_L1_error))
  __pyx_v_y131 = (__pyx_v_y130 * __pyx_v_y34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":887
 *         cdef double y130 = y129*y91
 *         cdef double y131 = y130*y34
 *         cdef double y132 = y59*y77             # <<<<<<<<<<<<<<
 *         cdef double y133 = y64*(0.75 - 1.5*X_2)
 *         cdef double y134 = y59*(-1.75*nu - 0.9375)
 */
  __Pyx_TraceLine(887,0,__PYX_ERR(1, 887, __pyx_L1_error))
  __pyx_v_y132 = (__pyx_v_y59 * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":888
 *         cdef double y131 = y130*y34
 *         cdef double y132 = y59*y77
 *         cdef double y133 = y64*(0.75 - 1.5*X_2)             # <<<<<<<<<<<<<<
 *         cdef double y134 = y59*(-1.75*nu - 0.9375)
 *         cdef double y135 = y19 - y7
 */
  __Pyx_TraceLine(888,0,__PYX_ERR(1, 888, __pyx_L1_error))
  __pyx_v_y133 = (__pyx_v_y64 * (0.75 - (1.5 * __pyx_v_X_2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":889
 *         cdef double y132 = y59*y77
 *         cdef double y133 = y64*(0.75 - 1.5*X_2)
 *         cdef double y134 = y59*(-1.75*nu - 0.9375)             # <<<<<<<<<<<<<<
 *         cdef double y135 = y19 - y7
 *         cdef double y136 = y135 - y63**2
 */
  __Pyx_TraceLine(889,0,__PYX_ERR(1, 889, __pyx_L1_error))
  __pyx_v_y134 = (__pyx_v_y59 * ((-1.75 * __pyx_v_nu) - 0.9375));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":890
 *         cdef double y133 = y64*(0.75 - 1.5*X_2)
 *         cdef double y134 = y59*(-1.75*nu - 0.9375)
 *         cdef double y135 = y19 - y7             # <<<<<<<<<<<<<<
 *         cdef double y136 = y135 - y63**2
 *         cdef double y137 = y136*(0.1875 - 0.75*nu)
 */
  __Pyx_TraceLine(890,0,__PYX_ERR(1, 890, __pyx_L1_error))
  __pyx_v_y135 = (__pyx_v_y19 - __pyx_v_y7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":891
 *         cdef double y134 = y59*(-1.75*nu - 0.9375)
 *         cdef double y135 = y19 - y7
 *         cdef double y136 = y135 - y63**2             # <<<<<<<<<<<<<<
 *         cdef double y137 = y136*(0.1875 - 0.75*nu)
 *         cdef double y138 = y70*(98.0*nu + 43.0)
 */
  __Pyx_TraceLine(891,0,__PYX_ERR(1, 891, __pyx_L1_error))
  __pyx_v_y136 = (__pyx_v_y135 - pow(__pyx_v_y63, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":892
 *         cdef double y135 = y19 - y7
 *         cdef double y136 = y135 - y63**2
 *         cdef double y137 = y136*(0.1875 - 0.75*nu)             # <<<<<<<<<<<<<<
 *         cdef double y138 = y70*(98.0*nu + 43.0)
 *         cdef double y139 = 264.0*y39
 */
  __Pyx_TraceLine(892,0,__PYX_ERR(1, 892, __pyx_L1_error))
  __pyx_v_y137 = (__pyx_v_y136 * (0.1875 - (0.75 * __pyx_v_nu)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":893
 *         cdef double y136 = y135 - y63**2
 *         cdef double y137 = y136*(0.1875 - 0.75*nu)
 *         cdef double y138 = y70*(98.0*nu + 43.0)             # <<<<<<<<<<<<<<
 *         cdef double y139 = 264.0*y39
 *         cdef double y140 = -1610.0*nu + y139 + 375.0
 */
  __Pyx_TraceLine(893,0,__PYX_ERR(1, 893, __pyx_L1_error))
  __pyx_v_y138 = (__pyx_v_y70 * ((98.0 * __pyx_v_nu) + 43.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":894
 *         cdef double y137 = y136*(0.1875 - 0.75*nu)
 *         cdef double y138 = y70*(98.0*nu + 43.0)
 *         cdef double y139 = 264.0*y39             # <<<<<<<<<<<<<<
 *         cdef double y140 = -1610.0*nu + y139 + 375.0
 *         cdef double y141 = 0.00260416666666667*y59
 */
  __Pyx_TraceLine(894,0,__PYX_ERR(1, 894, __pyx_L1_error))
  __pyx_v_y139 = (264.0 * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":895
 *         cdef double y138 = y70*(98.0*nu + 43.0)
 *         cdef double y139 = 264.0*y39
 *         cdef double y140 = -1610.0*nu + y139 + 375.0             # <<<<<<<<<<<<<<
 *         cdef double y141 = 0.00260416666666667*y59
 *         cdef double y142 = -1090.0*nu + 152.0*y39 + 219.0
 */
  __Pyx_TraceLine(895,0,__PYX_ERR(1, 895, __pyx_L1_error))
  __pyx_v_y140 = (((-1610.0 * __pyx_v_nu) + __pyx_v_y139) + 375.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":896
 *         cdef double y139 = 264.0*y39
 *         cdef double y140 = -1610.0*nu + y139 + 375.0
 *         cdef double y141 = 0.00260416666666667*y59             # <<<<<<<<<<<<<<
 *         cdef double y142 = -1090.0*nu + 152.0*y39 + 219.0
 *         cdef double y143 = 0.0078125*y136
 */
  __Pyx_TraceLine(896,0,__PYX_ERR(1, 896, __pyx_L1_error))
  __pyx_v_y141 = (0.00260416666666667 * __pyx_v_y59);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":897
 *         cdef double y140 = -1610.0*nu + y139 + 375.0
 *         cdef double y141 = 0.00260416666666667*y59
 *         cdef double y142 = -1090.0*nu + 152.0*y39 + 219.0             # <<<<<<<<<<<<<<
 *         cdef double y143 = 0.0078125*y136
 *         cdef double y144 = -0.5*y132 + y32*(0.5*y133 + 0.5*y134 + 0.5*y137) + y36*(-0.0625*y138 + y140*y141 + y142*y143) + 1.0
 */
  __Pyx_TraceLine(897,0,__PYX_ERR(1, 897, __pyx_L1_error))
  __pyx_v_y142 = (((-1090.0 * __pyx_v_nu) + (152.0 * __pyx_v_y39)) + 219.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":898
 *         cdef double y141 = 0.00260416666666667*y59
 *         cdef double y142 = -1090.0*nu + 152.0*y39 + 219.0
 *         cdef double y143 = 0.0078125*y136             # <<<<<<<<<<<<<<
 *         cdef double y144 = -0.5*y132 + y32*(0.5*y133 + 0.5*y134 + 0.5*y137) + y36*(-0.0625*y138 + y140*y141 + y142*y143) + 1.0
 *         cdef double y145 = y20*y77
 */
  __Pyx_TraceLine(898,0,__PYX_ERR(1, 898, __pyx_L1_error))
  __pyx_v_y143 = (0.0078125 * __pyx_v_y136);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":899
 *         cdef double y142 = -1090.0*nu + 152.0*y39 + 219.0
 *         cdef double y143 = 0.0078125*y136
 *         cdef double y144 = -0.5*y132 + y32*(0.5*y133 + 0.5*y134 + 0.5*y137) + y36*(-0.0625*y138 + y140*y141 + y142*y143) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double y145 = y20*y77
 *         cdef double y146 = y123**2
 */
  __Pyx_TraceLine(899,0,__PYX_ERR(1, 899, __pyx_L1_error))
  __pyx_v_y144 = ((((-0.5 * __pyx_v_y132) + (__pyx_v_y32 * (((0.5 * __pyx_v_y133) + (0.5 * __pyx_v_y134)) + (0.5 * __pyx_v_y137)))) + (__pyx_v_y36 * (((-0.0625 * __pyx_v_y138) + (__pyx_v_y140 * __pyx_v_y141)) + (__pyx_v_y142 * __pyx_v_y143)))) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":900
 *         cdef double y143 = 0.0078125*y136
 *         cdef double y144 = -0.5*y132 + y32*(0.5*y133 + 0.5*y134 + 0.5*y137) + y36*(-0.0625*y138 + y140*y141 + y142*y143) + 1.0
 *         cdef double y145 = y20*y77             # <<<<<<<<<<<<<<
 *         cdef double y146 = y123**2
 *         cdef double y147 = y146*y39
 */
  __Pyx_TraceLine(900,0,__PYX_ERR(1, 900, __pyx_L1_error))
  __pyx_v_y145 = (__pyx_v_y20 * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":901
 *         cdef double y144 = -0.5*y132 + y32*(0.5*y133 + 0.5*y134 + 0.5*y137) + y36*(-0.0625*y138 + y140*y141 + y142*y143) + 1.0
 *         cdef double y145 = y20*y77
 *         cdef double y146 = y123**2             # <<<<<<<<<<<<<<
 *         cdef double y147 = y146*y39
 *         cdef double y148 = y113*(-163683964.822551*r - 17833256.898555*y71 - 1188987459.03162)
 */
  __Pyx_TraceLine(901,0,__PYX_ERR(1, 901, __pyx_L1_error))
  __pyx_v_y146 = pow(__pyx_v_y123, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":902
 *         cdef double y145 = y20*y77
 *         cdef double y146 = y123**2
 *         cdef double y147 = y146*y39             # <<<<<<<<<<<<<<
 *         cdef double y148 = y113*(-163683964.822551*r - 17833256.898555*y71 - 1188987459.03162)
 *         cdef double y149 = y39*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*y0 + 1426660551.8844*y24 - 3089250703.76879*y35 + 2064783811.32587*y71 + 276057889687.011)
 */
  __Pyx_TraceLine(902,0,__PYX_ERR(1, 902, __pyx_L1_error))
  __pyx_v_y147 = (__pyx_v_y146 * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":903
 *         cdef double y146 = y123**2
 *         cdef double y147 = y146*y39
 *         cdef double y148 = y113*(-163683964.822551*r - 17833256.898555*y71 - 1188987459.03162)             # <<<<<<<<<<<<<<
 *         cdef double y149 = y39*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*y0 + 1426660551.8844*y24 - 3089250703.76879*y35 + 2064783811.32587*y71 + 276057889687.011)
 *         cdef double y150 = 588.0*nu + 1079.0
 */
  __Pyx_TraceLine(903,0,__PYX_ERR(1, 903, __pyx_L1_error))
  __pyx_v_y148 = (__pyx_v_y113 * (((-163683964.822551 * __pyx_v_r) - (17833256.898555 * __pyx_v_y71)) - 1188987459.03162));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":904
 *         cdef double y147 = y146*y39
 *         cdef double y148 = y113*(-163683964.822551*r - 17833256.898555*y71 - 1188987459.03162)
 *         cdef double y149 = y39*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*y0 + 1426660551.8844*y24 - 3089250703.76879*y35 + 2064783811.32587*y71 + 276057889687.011)             # <<<<<<<<<<<<<<
 *         cdef double y150 = 588.0*nu + 1079.0
 *         cdef double y151 = 756.0*nu
 */
  __Pyx_TraceLine(904,0,__PYX_ERR(1, 904, __pyx_L1_error))
  __pyx_v_y149 = (__pyx_v_y39 * (((((((((-39321600.0 * __pyx_v_a6) * ((3.0 * __pyx_v_r) + 59.0)) + (745857848.115604 * __pyx_v_a6)) + (122635399361.987 * __pyx_v_r)) - (6178501407.53758 * __pyx_v_y0)) + (1426660551.8844 * __pyx_v_y24)) - (3089250703.76879 * __pyx_v_y35)) + (2064783811.32587 * __pyx_v_y71)) + 276057889687.011));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":905
 *         cdef double y148 = y113*(-163683964.822551*r - 17833256.898555*y71 - 1188987459.03162)
 *         cdef double y149 = y39*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*y0 + 1426660551.8844*y24 - 3089250703.76879*y35 + 2064783811.32587*y71 + 276057889687.011)
 *         cdef double y150 = 588.0*nu + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double y151 = 756.0*nu
 *         cdef double y152 = y151 + 1079.0
 */
  __Pyx_TraceLine(905,0,__PYX_ERR(1, 905, __pyx_L1_error))
  __pyx_v_y150 = ((588.0 * __pyx_v_nu) + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":906
 *         cdef double y149 = y39*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*y0 + 1426660551.8844*y24 - 3089250703.76879*y35 + 2064783811.32587*y71 + 276057889687.011)
 *         cdef double y150 = 588.0*nu + 1079.0
 *         cdef double y151 = 756.0*nu             # <<<<<<<<<<<<<<
 *         cdef double y152 = y151 + 1079.0
 *         cdef double y153 = y0*y152
 */
  __Pyx_TraceLine(906,0,__PYX_ERR(1, 906, __pyx_L1_error))
  __pyx_v_y151 = (756.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":907
 *         cdef double y150 = 588.0*nu + 1079.0
 *         cdef double y151 = 756.0*nu
 *         cdef double y152 = y151 + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double y153 = y0*y152
 *         cdef double y154 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*y39 + 17264.0) + 1920.0*y150*y71 + 480.0*y152*y35 + 960.0*y153 + 161280.0*y24 - 1882456.23663972*y39 + 13447680.0
 */
  __Pyx_TraceLine(907,0,__PYX_ERR(1, 907, __pyx_L1_error))
  __pyx_v_y152 = (__pyx_v_y151 + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":908
 *         cdef double y151 = 756.0*nu
 *         cdef double y152 = y151 + 1079.0
 *         cdef double y153 = y0*y152             # <<<<<<<<<<<<<<
 *         cdef double y154 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*y39 + 17264.0) + 1920.0*y150*y71 + 480.0*y152*y35 + 960.0*y153 + 161280.0*y24 - 1882456.23663972*y39 + 13447680.0
 *         cdef double y155 = nu*y123
 */
  __Pyx_TraceLine(908,0,__PYX_ERR(1, 908, __pyx_L1_error))
  __pyx_v_y153 = (__pyx_v_y0 * __pyx_v_y152);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":909
 *         cdef double y152 = y151 + 1079.0
 *         cdef double y153 = y0*y152
 *         cdef double y154 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*y39 + 17264.0) + 1920.0*y150*y71 + 480.0*y152*y35 + 960.0*y153 + 161280.0*y24 - 1882456.23663972*y39 + 13447680.0             # <<<<<<<<<<<<<<
 *         cdef double y155 = nu*y123
 *         cdef double y156 = y154*y155
 */
  __Pyx_TraceLine(909,0,__PYX_ERR(1, 909, __pyx_L1_error))
  __pyx_v_y154 = ((((((((-38842241.4769507 * __pyx_v_nu) + ((240.0 * __pyx_v_r) * (((-7466.27061066206 * __pyx_v_nu) - (3024.0 * __pyx_v_y39)) + 17264.0))) + ((1920.0 * __pyx_v_y150) * __pyx_v_y71)) + ((480.0 * __pyx_v_y152) * __pyx_v_y35)) + (960.0 * __pyx_v_y153)) + (161280.0 * __pyx_v_y24)) - (1882456.23663972 * __pyx_v_y39)) + 13447680.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":910
 *         cdef double y153 = y0*y152
 *         cdef double y154 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*y39 + 17264.0) + 1920.0*y150*y71 + 480.0*y152*y35 + 960.0*y153 + 161280.0*y24 - 1882456.23663972*y39 + 13447680.0
 *         cdef double y155 = nu*y123             # <<<<<<<<<<<<<<
 *         cdef double y156 = y154*y155
 *         cdef double y157 = 8.0*r
 */
  __Pyx_TraceLine(910,0,__PYX_ERR(1, 910, __pyx_L1_error))
  __pyx_v_y155 = (__pyx_v_nu * __pyx_v_y123);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":911
 *         cdef double y154 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*y39 + 17264.0) + 1920.0*y150*y71 + 480.0*y152*y35 + 960.0*y153 + 161280.0*y24 - 1882456.23663972*y39 + 13447680.0
 *         cdef double y155 = nu*y123
 *         cdef double y156 = y154*y155             # <<<<<<<<<<<<<<
 *         cdef double y157 = 8.0*r
 *         cdef double y158 = 2.0*y0 + y157 + 4.0*y71 + 16.0
 */
  __Pyx_TraceLine(911,0,__PYX_ERR(1, 911, __pyx_L1_error))
  __pyx_v_y156 = (__pyx_v_y154 * __pyx_v_y155);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":912
 *         cdef double y155 = nu*y123
 *         cdef double y156 = y154*y155
 *         cdef double y157 = 8.0*r             # <<<<<<<<<<<<<<
 *         cdef double y158 = 2.0*y0 + y157 + 4.0*y71 + 16.0
 *         cdef double y159 = 7680.0*a6
 */
  __Pyx_TraceLine(912,0,__PYX_ERR(1, 912, __pyx_L1_error))
  __pyx_v_y157 = (8.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":913
 *         cdef double y156 = y154*y155
 *         cdef double y157 = 8.0*r
 *         cdef double y158 = 2.0*y0 + y157 + 4.0*y71 + 16.0             # <<<<<<<<<<<<<<
 *         cdef double y159 = 7680.0*a6
 *         cdef double y160 = 128.0*r
 */
  __Pyx_TraceLine(913,0,__PYX_ERR(1, 913, __pyx_L1_error))
  __pyx_v_y158 = ((((2.0 * __pyx_v_y0) + __pyx_v_y157) + (4.0 * __pyx_v_y71)) + 16.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":914
 *         cdef double y157 = 8.0*r
 *         cdef double y158 = 2.0*y0 + y157 + 4.0*y71 + 16.0
 *         cdef double y159 = 7680.0*a6             # <<<<<<<<<<<<<<
 *         cdef double y160 = 128.0*r
 *         cdef double y161 = 7704.0*r
 */
  __Pyx_TraceLine(914,0,__PYX_ERR(1, 914, __pyx_L1_error))
  __pyx_v_y159 = (7680.0 * __pyx_v_a6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":915
 *         cdef double y158 = 2.0*y0 + y157 + 4.0*y71 + 16.0
 *         cdef double y159 = 7680.0*a6
 *         cdef double y160 = 128.0*r             # <<<<<<<<<<<<<<
 *         cdef double y161 = 7704.0*r
 *         cdef double y162 = 148.04406601634*r
 */
  __Pyx_TraceLine(915,0,__PYX_ERR(1, 915, __pyx_L1_error))
  __pyx_v_y160 = (128.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":916
 *         cdef double y159 = 7680.0*a6
 *         cdef double y160 = 128.0*r
 *         cdef double y161 = 7704.0*r             # <<<<<<<<<<<<<<
 *         cdef double y162 = 148.04406601634*r
 *         cdef double y163 = 113485.217444961*r
 */
  __Pyx_TraceLine(916,0,__PYX_ERR(1, 916, __pyx_L1_error))
  __pyx_v_y161 = (7704.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":917
 *         cdef double y160 = 128.0*r
 *         cdef double y161 = 7704.0*r
 *         cdef double y162 = 148.04406601634*r             # <<<<<<<<<<<<<<
 *         cdef double y163 = 113485.217444961*r
 *         cdef double y164 = nu*(y159*(y158 + y35) + y160*(13218.7851094412*r + 4264.6962773603*y0 - 6852.34813868015*y35 + 8529.39255472061*y71 - 33722.4297811176) + y162*(1926.0*y0 + y161 + 349.0*y35 + 3852.0*y71 + 36400.0) + y163*(y158 - y35))
 */
  __Pyx_TraceLine(917,0,__PYX_ERR(1, 917, __pyx_L1_error))
  __pyx_v_y162 = (148.04406601634 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":918
 *         cdef double y161 = 7704.0*r
 *         cdef double y162 = 148.04406601634*r
 *         cdef double y163 = 113485.217444961*r             # <<<<<<<<<<<<<<
 *         cdef double y164 = nu*(y159*(y158 + y35) + y160*(13218.7851094412*r + 4264.6962773603*y0 - 6852.34813868015*y35 + 8529.39255472061*y71 - 33722.4297811176) + y162*(1926.0*y0 + y161 + 349.0*y35 + 3852.0*y71 + 36400.0) + y163*(y158 - y35))
 *         cdef double y165 = (241555486248.807*y118 + 67645734912.0*y147 + 1120.0*y148 + 7.0*y149 + 32768.0*y156 + 53760.0*y164 + 13212057600.0*y24)**(-1)
 */
  __Pyx_TraceLine(918,0,__PYX_ERR(1, 918, __pyx_L1_error))
  __pyx_v_y163 = (113485.217444961 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":919
 *         cdef double y162 = 148.04406601634*r
 *         cdef double y163 = 113485.217444961*r
 *         cdef double y164 = nu*(y159*(y158 + y35) + y160*(13218.7851094412*r + 4264.6962773603*y0 - 6852.34813868015*y35 + 8529.39255472061*y71 - 33722.4297811176) + y162*(1926.0*y0 + y161 + 349.0*y35 + 3852.0*y71 + 36400.0) + y163*(y158 - y35))             # <<<<<<<<<<<<<<
 *         cdef double y165 = (241555486248.807*y118 + 67645734912.0*y147 + 1120.0*y148 + 7.0*y149 + 32768.0*y156 + 53760.0*y164 + 13212057600.0*y24)**(-1)
 *         cdef double y166 = 336.0*r + y151 + 407.0
 */
  __Pyx_TraceLine(919,0,__PYX_ERR(1, 919, __pyx_L1_error))
  __pyx_v_y164 = (__pyx_v_nu * ((((__pyx_v_y159 * (__pyx_v_y158 + __pyx_v_y35)) + (__pyx_v_y160 * (((((13218.7851094412 * __pyx_v_r) + (4264.6962773603 * __pyx_v_y0)) - (6852.34813868015 * __pyx_v_y35)) + (8529.39255472061 * __pyx_v_y71)) - 33722.4297811176))) + (__pyx_v_y162 * (((((1926.0 * __pyx_v_y0) + __pyx_v_y161) + (349.0 * __pyx_v_y35)) + (3852.0 * __pyx_v_y71)) + 36400.0))) + (__pyx_v_y163 * (__pyx_v_y158 - __pyx_v_y35))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":920
 *         cdef double y163 = 113485.217444961*r
 *         cdef double y164 = nu*(y159*(y158 + y35) + y160*(13218.7851094412*r + 4264.6962773603*y0 - 6852.34813868015*y35 + 8529.39255472061*y71 - 33722.4297811176) + y162*(1926.0*y0 + y161 + 349.0*y35 + 3852.0*y71 + 36400.0) + y163*(y158 - y35))
 *         cdef double y165 = (241555486248.807*y118 + 67645734912.0*y147 + 1120.0*y148 + 7.0*y149 + 32768.0*y156 + 53760.0*y164 + 13212057600.0*y24)**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y166 = 336.0*r + y151 + 407.0
 *         cdef double y167 = 2048.0*nu*y123*y166 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*y39 - 245760.0) - 5416406.59541186*y39 - 3440640.0
 */
  __Pyx_TraceLine(920,0,__PYX_ERR(1, 920, __pyx_L1_error))
  __pyx_v_y165 = pow((((((((241555486248.807 * __pyx_v_y118) + (67645734912.0 * __pyx_v_y147)) + (1120.0 * __pyx_v_y148)) + (7.0 * __pyx_v_y149)) + (32768.0 * __pyx_v_y156)) + (53760.0 * __pyx_v_y164)) + (13212057600.0 * __pyx_v_y24)), -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":921
 *         cdef double y164 = nu*(y159*(y158 + y35) + y160*(13218.7851094412*r + 4264.6962773603*y0 - 6852.34813868015*y35 + 8529.39255472061*y71 - 33722.4297811176) + y162*(1926.0*y0 + y161 + 349.0*y35 + 3852.0*y71 + 36400.0) + y163*(y158 - y35))
 *         cdef double y165 = (241555486248.807*y118 + 67645734912.0*y147 + 1120.0*y148 + 7.0*y149 + 32768.0*y156 + 53760.0*y164 + 13212057600.0*y24)**(-1)
 *         cdef double y166 = 336.0*r + y151 + 407.0             # <<<<<<<<<<<<<<
 *         cdef double y167 = 2048.0*nu*y123*y166 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*y39 - 245760.0) - 5416406.59541186*y39 - 3440640.0
 *         cdef double y168 = y165*y167
 */
  __Pyx_TraceLine(921,0,__PYX_ERR(1, 921, __pyx_L1_error))
  __pyx_v_y166 = (((336.0 * __pyx_v_r) + __pyx_v_y151) + 407.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":922
 *         cdef double y165 = (241555486248.807*y118 + 67645734912.0*y147 + 1120.0*y148 + 7.0*y149 + 32768.0*y156 + 53760.0*y164 + 13212057600.0*y24)**(-1)
 *         cdef double y166 = 336.0*r + y151 + 407.0
 *         cdef double y167 = 2048.0*nu*y123*y166 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*y39 - 245760.0) - 5416406.59541186*y39 - 3440640.0             # <<<<<<<<<<<<<<
 *         cdef double y168 = y165*y167
 *         cdef double y169 = y168*y35
 */
  __Pyx_TraceLine(922,0,__PYX_ERR(1, 922, __pyx_L1_error))
  __pyx_v_y167 = (((((((2048.0 * __pyx_v_nu) * __pyx_v_y123) * __pyx_v_y166) + ((28.0 * __pyx_v_nu) * ((1920.0 * __pyx_v_a6) + 733955.307463037))) - ((7.0 * __pyx_v_r) * (((938918.400156317 * __pyx_v_nu) - (185763.092693281 * __pyx_v_y39)) - 245760.0))) - (5416406.59541186 * __pyx_v_y39)) - 3440640.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":923
 *         cdef double y166 = 336.0*r + y151 + 407.0
 *         cdef double y167 = 2048.0*nu*y123*y166 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*y39 - 245760.0) - 5416406.59541186*y39 - 3440640.0
 *         cdef double y168 = y165*y167             # <<<<<<<<<<<<<<
 *         cdef double y169 = y168*y35
 *         cdef double y170 = 0.000130208333333333*y145 + y169
 */
  __Pyx_TraceLine(923,0,__PYX_ERR(1, 923, __pyx_L1_error))
  __pyx_v_y168 = (__pyx_v_y165 * __pyx_v_y167);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":924
 *         cdef double y167 = 2048.0*nu*y123*y166 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*y39 - 245760.0) - 5416406.59541186*y39 - 3440640.0
 *         cdef double y168 = y165*y167
 *         cdef double y169 = y168*y35             # <<<<<<<<<<<<<<
 *         cdef double y170 = 0.000130208333333333*y145 + y169
 *         cdef double y171 = y170**(-4)
 */
  __Pyx_TraceLine(924,0,__PYX_ERR(1, 924, __pyx_L1_error))
  __pyx_v_y169 = (__pyx_v_y168 * __pyx_v_y35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":925
 *         cdef double y168 = y165*y167
 *         cdef double y169 = y168*y35
 *         cdef double y170 = 0.000130208333333333*y145 + y169             # <<<<<<<<<<<<<<
 *         cdef double y171 = y170**(-4)
 *         cdef double y172 = r*y146
 */
  __Pyx_TraceLine(925,0,__PYX_ERR(1, 925, __pyx_L1_error))
  __pyx_v_y170 = ((0.000130208333333333 * __pyx_v_y145) + __pyx_v_y169);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":926
 *         cdef double y169 = y168*y35
 *         cdef double y170 = 0.000130208333333333*y145 + y169
 *         cdef double y171 = y170**(-4)             # <<<<<<<<<<<<<<
 *         cdef double y172 = r*y146
 *         cdef double y173 = -630116198.873299*nu - 197773496.793534*y39 + 5805304367.87913
 */
  __Pyx_TraceLine(926,0,__PYX_ERR(1, 926, __pyx_L1_error))
  __pyx_v_y171 = pow(__pyx_v_y170, -4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":927
 *         cdef double y170 = 0.000130208333333333*y145 + y169
 *         cdef double y171 = y170**(-4)
 *         cdef double y172 = r*y146             # <<<<<<<<<<<<<<
 *         cdef double y173 = -630116198.873299*nu - 197773496.793534*y39 + 5805304367.87913
 *         cdef double y174 = y173*y71
 */
  __Pyx_TraceLine(927,0,__PYX_ERR(1, 927, __pyx_L1_error))
  __pyx_v_y172 = (__pyx_v_r * __pyx_v_y146);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":928
 *         cdef double y171 = y170**(-4)
 *         cdef double y172 = r*y146
 *         cdef double y173 = -630116198.873299*nu - 197773496.793534*y39 + 5805304367.87913             # <<<<<<<<<<<<<<
 *         cdef double y174 = y173*y71
 *         cdef double y175 = -2675575.66847905*nu - 138240.0*y39 - 5278341.3229329
 */
  __Pyx_TraceLine(928,0,__PYX_ERR(1, 928, __pyx_L1_error))
  __pyx_v_y173 = (((-630116198.873299 * __pyx_v_nu) - (197773496.793534 * __pyx_v_y39)) + 5805304367.87913);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":929
 *         cdef double y172 = r*y146
 *         cdef double y173 = -630116198.873299*nu - 197773496.793534*y39 + 5805304367.87913
 *         cdef double y174 = y173*y71             # <<<<<<<<<<<<<<
 *         cdef double y175 = -2675575.66847905*nu - 138240.0*y39 - 5278341.3229329
 *         cdef double y176 = y0*y175
 */
  __Pyx_TraceLine(929,0,__PYX_ERR(1, 929, __pyx_L1_error))
  __pyx_v_y174 = (__pyx_v_y173 * __pyx_v_y71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":930
 *         cdef double y173 = -630116198.873299*nu - 197773496.793534*y39 + 5805304367.87913
 *         cdef double y174 = y173*y71
 *         cdef double y175 = -2675575.66847905*nu - 138240.0*y39 - 5278341.3229329             # <<<<<<<<<<<<<<
 *         cdef double y176 = y0*y175
 *         cdef double y177 = nu*(-2510664218.28128*nu + 14515200.0*y113 - 42636451.6032331*y39 + 1002013764.01019)
 */
  __Pyx_TraceLine(930,0,__PYX_ERR(1, 930, __pyx_L1_error))
  __pyx_v_y175 = (((-2675575.66847905 * __pyx_v_nu) - (138240.0 * __pyx_v_y39)) - 5278341.3229329);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":931
 *         cdef double y174 = y173*y71
 *         cdef double y175 = -2675575.66847905*nu - 138240.0*y39 - 5278341.3229329
 *         cdef double y176 = y0*y175             # <<<<<<<<<<<<<<
 *         cdef double y177 = nu*(-2510664218.28128*nu + 14515200.0*y113 - 42636451.6032331*y39 + 1002013764.01019)
 *         cdef double y178 = (1 - 0.496948781616935*nu)**2
 */
  __Pyx_TraceLine(931,0,__PYX_ERR(1, 931, __pyx_L1_error))
  __pyx_v_y176 = (__pyx_v_y0 * __pyx_v_y175);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":932
 *         cdef double y175 = -2675575.66847905*nu - 138240.0*y39 - 5278341.3229329
 *         cdef double y176 = y0*y175
 *         cdef double y177 = nu*(-2510664218.28128*nu + 14515200.0*y113 - 42636451.6032331*y39 + 1002013764.01019)             # <<<<<<<<<<<<<<
 *         cdef double y178 = (1 - 0.496948781616935*nu)**2
 *         cdef double y179 = 43393301259014.8*nu + 5927865218923.02*y113 + 86618264430493.3*y178 + 43133561885859.3*y39 + 188440788778196.0
 */
  __Pyx_TraceLine(932,0,__PYX_ERR(1, 932, __pyx_L1_error))
  __pyx_v_y177 = (__pyx_v_nu * ((((-2510664218.28128 * __pyx_v_nu) + (14515200.0 * __pyx_v_y113)) - (42636451.6032331 * __pyx_v_y39)) + 1002013764.01019));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":933
 *         cdef double y176 = y0*y175
 *         cdef double y177 = nu*(-2510664218.28128*nu + 14515200.0*y113 - 42636451.6032331*y39 + 1002013764.01019)
 *         cdef double y178 = (1 - 0.496948781616935*nu)**2             # <<<<<<<<<<<<<<
 *         cdef double y179 = 43393301259014.8*nu + 5927865218923.02*y113 + 86618264430493.3*y178 + 43133561885859.3*y39 + 188440788778196.0
 *         cdef double y180 = r*y179
 */
  __Pyx_TraceLine(933,0,__PYX_ERR(1, 933, __pyx_L1_error))
  __pyx_v_y178 = pow((1.0 - (0.496948781616935 * __pyx_v_nu)), 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":934
 *         cdef double y177 = nu*(-2510664218.28128*nu + 14515200.0*y113 - 42636451.6032331*y39 + 1002013764.01019)
 *         cdef double y178 = (1 - 0.496948781616935*nu)**2
 *         cdef double y179 = 43393301259014.8*nu + 5927865218923.02*y113 + 86618264430493.3*y178 + 43133561885859.3*y39 + 188440788778196.0             # <<<<<<<<<<<<<<
 *         cdef double y180 = r*y179
 *         cdef double y181 = 14700.0*nu + 42911.0
 */
  __Pyx_TraceLine(934,0,__PYX_ERR(1, 934, __pyx_L1_error))
  __pyx_v_y179 = (((((43393301259014.8 * __pyx_v_nu) + (5927865218923.02 * __pyx_v_y113)) + (86618264430493.3 * __pyx_v_y178)) + (43133561885859.3 * __pyx_v_y39)) + 188440788778196.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":935
 *         cdef double y178 = (1 - 0.496948781616935*nu)**2
 *         cdef double y179 = 43393301259014.8*nu + 5927865218923.02*y113 + 86618264430493.3*y178 + 43133561885859.3*y39 + 188440788778196.0
 *         cdef double y180 = r*y179             # <<<<<<<<<<<<<<
 *         cdef double y181 = 14700.0*nu + 42911.0
 *         cdef double y182 = 283115520.0*y181
 */
  __Pyx_TraceLine(935,0,__PYX_ERR(1, 935, __pyx_L1_error))
  __pyx_v_y180 = (__pyx_v_r * __pyx_v_y179);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":936
 *         cdef double y179 = 43393301259014.8*nu + 5927865218923.02*y113 + 86618264430493.3*y178 + 43133561885859.3*y39 + 188440788778196.0
 *         cdef double y180 = r*y179
 *         cdef double y181 = 14700.0*nu + 42911.0             # <<<<<<<<<<<<<<
 *         cdef double y182 = 283115520.0*y181
 *         cdef double y183 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*y39 - 2119671837.36038) + 879923036160.0*y0 + y182*y71
 */
  __Pyx_TraceLine(936,0,__PYX_ERR(1, 936, __pyx_L1_error))
  __pyx_v_y181 = ((14700.0 * __pyx_v_nu) + 42911.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":937
 *         cdef double y180 = r*y179
 *         cdef double y181 = 14700.0*nu + 42911.0
 *         cdef double y182 = 283115520.0*y181             # <<<<<<<<<<<<<<
 *         cdef double y183 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*y39 - 2119671837.36038) + 879923036160.0*y0 + y182*y71
 *         cdef double y184 = y123*y183
 */
  __Pyx_TraceLine(937,0,__PYX_ERR(1, 937, __pyx_L1_error))
  __pyx_v_y182 = (283115520.0 * __pyx_v_y181);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":938
 *         cdef double y181 = 14700.0*nu + 42911.0
 *         cdef double y182 = 283115520.0*y181
 *         cdef double y183 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*y39 - 2119671837.36038) + 879923036160.0*y0 + y182*y71             # <<<<<<<<<<<<<<
 *         cdef double y184 = y123*y183
 *         cdef double y185 = y172 - 1.59227685093395e-9*y174 - 1.67189069348064e-7*y176 + 9.55366110560367e-9*y177 + 1.72773095804465e-13*y180 + 1.72773095804465e-13*y184
 */
  __Pyx_TraceLine(938,0,__PYX_ERR(1, 938, __pyx_L1_error))
  __pyx_v_y183 = (((((-1698693120.0 * __pyx_v_nu) * ((11592.0 * __pyx_v_nu) + 69847.0)) + ((49152.0 * __pyx_v_r) * (((409207698.136075 * __pyx_v_nu) + (102574080.0 * __pyx_v_y39)) - 2119671837.36038))) + (879923036160.0 * __pyx_v_y0)) + (__pyx_v_y182 * __pyx_v_y71));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":939
 *         cdef double y182 = 283115520.0*y181
 *         cdef double y183 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*y39 - 2119671837.36038) + 879923036160.0*y0 + y182*y71
 *         cdef double y184 = y123*y183             # <<<<<<<<<<<<<<
 *         cdef double y185 = y172 - 1.59227685093395e-9*y174 - 1.67189069348064e-7*y176 + 9.55366110560367e-9*y177 + 1.72773095804465e-13*y180 + 1.72773095804465e-13*y184
 *         cdef double y186 = y185**2
 */
  __Pyx_TraceLine(939,0,__PYX_ERR(1, 939, __pyx_L1_error))
  __pyx_v_y184 = (__pyx_v_y123 * __pyx_v_y183);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":940
 *         cdef double y183 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*y39 - 2119671837.36038) + 879923036160.0*y0 + y182*y71
 *         cdef double y184 = y123*y183
 *         cdef double y185 = y172 - 1.59227685093395e-9*y174 - 1.67189069348064e-7*y176 + 9.55366110560367e-9*y177 + 1.72773095804465e-13*y180 + 1.72773095804465e-13*y184             # <<<<<<<<<<<<<<
 *         cdef double y186 = y185**2
 *         cdef double y187 = nu*r
 */
  __Pyx_TraceLine(940,0,__PYX_ERR(1, 940, __pyx_L1_error))
  __pyx_v_y185 = (((((__pyx_v_y172 - (1.59227685093395e-9 * __pyx_v_y174)) - (1.67189069348064e-7 * __pyx_v_y176)) + (9.55366110560367e-9 * __pyx_v_y177)) + (1.72773095804465e-13 * __pyx_v_y180)) + (1.72773095804465e-13 * __pyx_v_y184));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":941
 *         cdef double y184 = y123*y183
 *         cdef double y185 = y172 - 1.59227685093395e-9*y174 - 1.67189069348064e-7*y176 + 9.55366110560367e-9*y177 + 1.72773095804465e-13*y180 + 1.72773095804465e-13*y184
 *         cdef double y186 = y185**2             # <<<<<<<<<<<<<<
 *         cdef double y187 = nu*r
 *         cdef double y188 = nu*y71
 */
  __Pyx_TraceLine(941,0,__PYX_ERR(1, 941, __pyx_L1_error))
  __pyx_v_y186 = pow(__pyx_v_y185, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":942
 *         cdef double y185 = y172 - 1.59227685093395e-9*y174 - 1.67189069348064e-7*y176 + 9.55366110560367e-9*y177 + 1.72773095804465e-13*y180 + 1.72773095804465e-13*y184
 *         cdef double y186 = y185**2
 *         cdef double y187 = nu*r             # <<<<<<<<<<<<<<
 *         cdef double y188 = nu*y71
 *         cdef double y189 = r*y39
 */
  __Pyx_TraceLine(942,0,__PYX_ERR(1, 942, __pyx_L1_error))
  __pyx_v_y187 = (__pyx_v_nu * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":943
 *         cdef double y186 = y185**2
 *         cdef double y187 = nu*r
 *         cdef double y188 = nu*y71             # <<<<<<<<<<<<<<
 *         cdef double y189 = r*y39
 *         cdef double y190 = r*y181
 */
  __Pyx_TraceLine(943,0,__PYX_ERR(1, 943, __pyx_L1_error))
  __pyx_v_y188 = (__pyx_v_nu * __pyx_v_y71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":944
 *         cdef double y187 = nu*r
 *         cdef double y188 = nu*y71
 *         cdef double y189 = r*y39             # <<<<<<<<<<<<<<
 *         cdef double y190 = r*y181
 *         cdef double y191 = y39*y71
 */
  __Pyx_TraceLine(944,0,__PYX_ERR(1, 944, __pyx_L1_error))
  __pyx_v_y189 = (__pyx_v_r * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":945
 *         cdef double y188 = nu*y71
 *         cdef double y189 = r*y39
 *         cdef double y190 = r*y181             # <<<<<<<<<<<<<<
 *         cdef double y191 = y39*y71
 *         cdef double y192 = 5041721180160.0*y39 - 104186110149937.0
 */
  __Pyx_TraceLine(945,0,__PYX_ERR(1, 945, __pyx_L1_error))
  __pyx_v_y190 = (__pyx_v_r * __pyx_v_y181);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":946
 *         cdef double y189 = r*y39
 *         cdef double y190 = r*y181
 *         cdef double y191 = y39*y71             # <<<<<<<<<<<<<<
 *         cdef double y192 = 5041721180160.0*y39 - 104186110149937.0
 *         cdef double y193 = -25392914995744.3*nu - r*y182 - y192 - 879923036160.0*y71
 */
  __Pyx_TraceLine(946,0,__PYX_ERR(1, 946, __pyx_L1_error))
  __pyx_v_y191 = (__pyx_v_y39 * __pyx_v_y71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":947
 *         cdef double y190 = r*y181
 *         cdef double y191 = y39*y71
 *         cdef double y192 = 5041721180160.0*y39 - 104186110149937.0             # <<<<<<<<<<<<<<
 *         cdef double y193 = -25392914995744.3*nu - r*y182 - y192 - 879923036160.0*y71
 *         cdef double y194 = y123*y193
 */
  __Pyx_TraceLine(947,0,__PYX_ERR(1, 947, __pyx_L1_error))
  __pyx_v_y192 = ((5041721180160.0 * __pyx_v_y39) - 104186110149937.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":948
 *         cdef double y191 = y39*y71
 *         cdef double y192 = 5041721180160.0*y39 - 104186110149937.0
 *         cdef double y193 = -25392914995744.3*nu - r*y182 - y192 - 879923036160.0*y71             # <<<<<<<<<<<<<<
 *         cdef double y194 = y123*y193
 *         cdef double y195 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0244692826489756*y113 + 0.0210425293255724*y146 + 0.0385795738434214*y187 + 0.00941289164152486*y188 + 0.00662650629087394*y189 - 1.18824456940711e-6*y190 + 0.000486339502879429*y191 - 3.63558293513537e-15*y194 + 0.291062041428379*y39 + 0.0185696317637669*y71 + 1
 */
  __Pyx_TraceLine(948,0,__PYX_ERR(1, 948, __pyx_L1_error))
  __pyx_v_y193 = ((((-25392914995744.3 * __pyx_v_nu) - (__pyx_v_r * __pyx_v_y182)) - __pyx_v_y192) - (879923036160.0 * __pyx_v_y71));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":949
 *         cdef double y192 = 5041721180160.0*y39 - 104186110149937.0
 *         cdef double y193 = -25392914995744.3*nu - r*y182 - y192 - 879923036160.0*y71
 *         cdef double y194 = y123*y193             # <<<<<<<<<<<<<<
 *         cdef double y195 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0244692826489756*y113 + 0.0210425293255724*y146 + 0.0385795738434214*y187 + 0.00941289164152486*y188 + 0.00662650629087394*y189 - 1.18824456940711e-6*y190 + 0.000486339502879429*y191 - 3.63558293513537e-15*y194 + 0.291062041428379*y39 + 0.0185696317637669*y71 + 1
 *         cdef double y196 = y195**(-2)
 */
  __Pyx_TraceLine(949,0,__PYX_ERR(1, 949, __pyx_L1_error))
  __pyx_v_y194 = (__pyx_v_y123 * __pyx_v_y193);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":950
 *         cdef double y193 = -25392914995744.3*nu - r*y182 - y192 - 879923036160.0*y71
 *         cdef double y194 = y123*y193
 *         cdef double y195 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0244692826489756*y113 + 0.0210425293255724*y146 + 0.0385795738434214*y187 + 0.00941289164152486*y188 + 0.00662650629087394*y189 - 1.18824456940711e-6*y190 + 0.000486339502879429*y191 - 3.63558293513537e-15*y194 + 0.291062041428379*y39 + 0.0185696317637669*y71 + 1             # <<<<<<<<<<<<<<
 *         cdef double y196 = y195**(-2)
 *         cdef double y197 = 4.0*y39
 */
  __Pyx_TraceLine(950,0,__PYX_ERR(1, 950, __pyx_L1_error))
  __pyx_v_y195 = (((((((((((((-0.0438084424460039 * __pyx_v_nu) - (0.143521050466841 * __pyx_v_r)) + (0.0244692826489756 * __pyx_v_y113)) + (0.0210425293255724 * __pyx_v_y146)) + (0.0385795738434214 * __pyx_v_y187)) + (0.00941289164152486 * __pyx_v_y188)) + (0.00662650629087394 * __pyx_v_y189)) - (1.18824456940711e-6 * __pyx_v_y190)) + (0.000486339502879429 * __pyx_v_y191)) - (3.63558293513537e-15 * __pyx_v_y194)) + (0.291062041428379 * __pyx_v_y39)) + (0.0185696317637669 * __pyx_v_y71)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":951
 *         cdef double y194 = y123*y193
 *         cdef double y195 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0244692826489756*y113 + 0.0210425293255724*y146 + 0.0385795738434214*y187 + 0.00941289164152486*y188 + 0.00662650629087394*y189 - 1.18824456940711e-6*y190 + 0.000486339502879429*y191 - 3.63558293513537e-15*y194 + 0.291062041428379*y39 + 0.0185696317637669*y71 + 1
 *         cdef double y196 = y195**(-2)             # <<<<<<<<<<<<<<
 *         cdef double y197 = 4.0*y39
 *         cdef double y198 = 0.3125*delta*y67*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*y39 - 5.0)*(5.0*y13 + 5.0*y18 + 5.0*y7) - 0.03125*(-5.0*nu + y197 + 1.0)*(15.0*y13 + 15.0*y18 - 15.0*y7)
 */
  __Pyx_TraceLine(951,0,__PYX_ERR(1, 951, __pyx_L1_error))
  __pyx_v_y196 = pow(__pyx_v_y195, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":952
 *         cdef double y195 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0244692826489756*y113 + 0.0210425293255724*y146 + 0.0385795738434214*y187 + 0.00941289164152486*y188 + 0.00662650629087394*y189 - 1.18824456940711e-6*y190 + 0.000486339502879429*y191 - 3.63558293513537e-15*y194 + 0.291062041428379*y39 + 0.0185696317637669*y71 + 1
 *         cdef double y196 = y195**(-2)
 *         cdef double y197 = 4.0*y39             # <<<<<<<<<<<<<<
 *         cdef double y198 = 0.3125*delta*y67*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*y39 - 5.0)*(5.0*y13 + 5.0*y18 + 5.0*y7) - 0.03125*(-5.0*nu + y197 + 1.0)*(15.0*y13 + 15.0*y18 - 15.0*y7)
 *         cdef double y199 = y171*y186*y196*y198
 */
  __Pyx_TraceLine(952,0,__PYX_ERR(1, 952, __pyx_L1_error))
  __pyx_v_y197 = (4.0 * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":953
 *         cdef double y196 = y195**(-2)
 *         cdef double y197 = 4.0*y39
 *         cdef double y198 = 0.3125*delta*y67*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*y39 - 5.0)*(5.0*y13 + 5.0*y18 + 5.0*y7) - 0.03125*(-5.0*nu + y197 + 1.0)*(15.0*y13 + 15.0*y18 - 15.0*y7)             # <<<<<<<<<<<<<<
 *         cdef double y199 = y171*y186*y196*y198
 *         cdef double y200 = y100*y199
 */
  __Pyx_TraceLine(953,0,__PYX_ERR(1, 953, __pyx_L1_error))
  __pyx_v_y198 = (((((0.3125 * __pyx_v_delta) * __pyx_v_y67) * ((18.0 * __pyx_v_nu) - 1.0)) - ((0.03125 * (((-33.0 * __pyx_v_nu) + (32.0 * __pyx_v_y39)) - 5.0)) * (((5.0 * __pyx_v_y13) + (5.0 * __pyx_v_y18)) + (5.0 * __pyx_v_y7)))) - ((0.03125 * (((-5.0 * __pyx_v_nu) + __pyx_v_y197) + 1.0)) * (((15.0 * __pyx_v_y13) + (15.0 * __pyx_v_y18)) - (15.0 * __pyx_v_y7))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":954
 *         cdef double y197 = 4.0*y39
 *         cdef double y198 = 0.3125*delta*y67*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*y39 - 5.0)*(5.0*y13 + 5.0*y18 + 5.0*y7) - 0.03125*(-5.0*nu + y197 + 1.0)*(15.0*y13 + 15.0*y18 - 15.0*y7)
 *         cdef double y199 = y171*y186*y196*y198             # <<<<<<<<<<<<<<
 *         cdef double y200 = y100*y199
 *         cdef double y201 = r**(-13)
 */
  __Pyx_TraceLine(954,0,__PYX_ERR(1, 954, __pyx_L1_error))
  __pyx_v_y199 = (((__pyx_v_y171 * __pyx_v_y186) * __pyx_v_y196) * __pyx_v_y198);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":955
 *         cdef double y198 = 0.3125*delta*y67*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*y39 - 5.0)*(5.0*y13 + 5.0*y18 + 5.0*y7) - 0.03125*(-5.0*nu + y197 + 1.0)*(15.0*y13 + 15.0*y18 - 15.0*y7)
 *         cdef double y199 = y171*y186*y196*y198
 *         cdef double y200 = y100*y199             # <<<<<<<<<<<<<<
 *         cdef double y201 = r**(-13)
 *         cdef double y202 = y20 + y71
 */
  __Pyx_TraceLine(955,0,__PYX_ERR(1, 955, __pyx_L1_error))
  __pyx_v_y200 = (__pyx_v_y100 * __pyx_v_y199);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":956
 *         cdef double y199 = y171*y186*y196*y198
 *         cdef double y200 = y100*y199
 *         cdef double y201 = r**(-13)             # <<<<<<<<<<<<<<
 *         cdef double y202 = y20 + y71
 *         cdef double y203 = y202**4
 */
  __Pyx_TraceLine(956,0,__PYX_ERR(1, 956, __pyx_L1_error))
  __pyx_v_y201 = pow(__pyx_v_r, -13.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":957
 *         cdef double y200 = y100*y199
 *         cdef double y201 = r**(-13)
 *         cdef double y202 = y20 + y71             # <<<<<<<<<<<<<<
 *         cdef double y203 = y202**4
 *         cdef double y204 = y201*y203
 */
  __Pyx_TraceLine(957,0,__PYX_ERR(1, 957, __pyx_L1_error))
  __pyx_v_y202 = (__pyx_v_y20 + __pyx_v_y71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":958
 *         cdef double y201 = r**(-13)
 *         cdef double y202 = y20 + y71
 *         cdef double y203 = y202**4             # <<<<<<<<<<<<<<
 *         cdef double y204 = y201*y203
 *         cdef double y205 = prst**2
 */
  __Pyx_TraceLine(958,0,__PYX_ERR(1, 958, __pyx_L1_error))
  __pyx_v_y203 = pow(__pyx_v_y202, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":959
 *         cdef double y202 = y20 + y71
 *         cdef double y203 = y202**4
 *         cdef double y204 = y201*y203             # <<<<<<<<<<<<<<
 *         cdef double y205 = prst**2
 *         cdef double y206 = y170**(-2)
 */
  __Pyx_TraceLine(959,0,__PYX_ERR(1, 959, __pyx_L1_error))
  __pyx_v_y204 = (__pyx_v_y201 * __pyx_v_y203);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":960
 *         cdef double y203 = y202**4
 *         cdef double y204 = y201*y203
 *         cdef double y205 = prst**2             # <<<<<<<<<<<<<<
 *         cdef double y206 = y170**(-2)
 *         cdef double y207 = 1822680546449.21*y39
 */
  __Pyx_TraceLine(960,0,__PYX_ERR(1, 960, __pyx_L1_error))
  __pyx_v_y205 = pow(__pyx_v_prst, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":961
 *         cdef double y204 = y201*y203
 *         cdef double y205 = prst**2
 *         cdef double y206 = y170**(-2)             # <<<<<<<<<<<<<<
 *         cdef double y207 = 1822680546449.21*y39
 *         cdef double y208 = 5787938193408.0*y146
 */
  __Pyx_TraceLine(961,0,__PYX_ERR(1, 961, __pyx_L1_error))
  __pyx_v_y206 = pow(__pyx_v_y170, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":962
 *         cdef double y205 = prst**2
 *         cdef double y206 = y170**(-2)
 *         cdef double y207 = 1822680546449.21*y39             # <<<<<<<<<<<<<<
 *         cdef double y208 = 5787938193408.0*y146
 *         cdef double y209 = -12049908701745.2*nu + r*y207 - 39476764256925.6*r + 6730497718123.02*y113 + 10611661054566.2*y187 + 2589101062873.81*y188 - 326837426.241486*y190 + 133772083200.0*y191 - y194 + y208 + 80059249540278.2*y39 + 5107745331375.71*y71 + 275059053208689.0
 */
  __Pyx_TraceLine(962,0,__PYX_ERR(1, 962, __pyx_L1_error))
  __pyx_v_y207 = (1822680546449.21 * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":963
 *         cdef double y206 = y170**(-2)
 *         cdef double y207 = 1822680546449.21*y39
 *         cdef double y208 = 5787938193408.0*y146             # <<<<<<<<<<<<<<
 *         cdef double y209 = -12049908701745.2*nu + r*y207 - 39476764256925.6*r + 6730497718123.02*y113 + 10611661054566.2*y187 + 2589101062873.81*y188 - 326837426.241486*y190 + 133772083200.0*y191 - y194 + y208 + 80059249540278.2*y39 + 5107745331375.71*y71 + 275059053208689.0
 *         cdef double y210 = y209**(-1)
 */
  __Pyx_TraceLine(963,0,__PYX_ERR(1, 963, __pyx_L1_error))
  __pyx_v_y208 = (5787938193408.0 * __pyx_v_y146);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":964
 *         cdef double y207 = 1822680546449.21*y39
 *         cdef double y208 = 5787938193408.0*y146
 *         cdef double y209 = -12049908701745.2*nu + r*y207 - 39476764256925.6*r + 6730497718123.02*y113 + 10611661054566.2*y187 + 2589101062873.81*y188 - 326837426.241486*y190 + 133772083200.0*y191 - y194 + y208 + 80059249540278.2*y39 + 5107745331375.71*y71 + 275059053208689.0             # <<<<<<<<<<<<<<
 *         cdef double y210 = y209**(-1)
 *         cdef double y211 = y205*y206*y210
 */
  __Pyx_TraceLine(964,0,__PYX_ERR(1, 964, __pyx_L1_error))
  __pyx_v_y209 = (((((((((((((-12049908701745.2 * __pyx_v_nu) + (__pyx_v_r * __pyx_v_y207)) - (39476764256925.6 * __pyx_v_r)) + (6730497718123.02 * __pyx_v_y113)) + (10611661054566.2 * __pyx_v_y187)) + (2589101062873.81 * __pyx_v_y188)) - (326837426.241486 * __pyx_v_y190)) + (133772083200.0 * __pyx_v_y191)) - __pyx_v_y194) + __pyx_v_y208) + (80059249540278.2 * __pyx_v_y39)) + (5107745331375.71 * __pyx_v_y71)) + 275059053208689.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":965
 *         cdef double y208 = 5787938193408.0*y146
 *         cdef double y209 = -12049908701745.2*nu + r*y207 - 39476764256925.6*r + 6730497718123.02*y113 + 10611661054566.2*y187 + 2589101062873.81*y188 - 326837426.241486*y190 + 133772083200.0*y191 - y194 + y208 + 80059249540278.2*y39 + 5107745331375.71*y71 + 275059053208689.0
 *         cdef double y210 = y209**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y211 = y205*y206*y210
 *         cdef double y212 = 4.0*nu
 */
  __Pyx_TraceLine(965,0,__PYX_ERR(1, 965, __pyx_L1_error))
  __pyx_v_y210 = pow(__pyx_v_y209, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":966
 *         cdef double y209 = -12049908701745.2*nu + r*y207 - 39476764256925.6*r + 6730497718123.02*y113 + 10611661054566.2*y187 + 2589101062873.81*y188 - 326837426.241486*y190 + 133772083200.0*y191 - y194 + y208 + 80059249540278.2*y39 + 5107745331375.71*y71 + 275059053208689.0
 *         cdef double y210 = y209**(-1)
 *         cdef double y211 = y205*y206*y210             # <<<<<<<<<<<<<<
 *         cdef double y212 = 4.0*nu
 *         cdef double y213 = 3.0*y7
 */
  __Pyx_TraceLine(966,0,__PYX_ERR(1, 966, __pyx_L1_error))
  __pyx_v_y211 = ((__pyx_v_y205 * __pyx_v_y206) * __pyx_v_y210);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":967
 *         cdef double y210 = y209**(-1)
 *         cdef double y211 = y205*y206*y210
 *         cdef double y212 = 4.0*nu             # <<<<<<<<<<<<<<
 *         cdef double y213 = 3.0*y7
 *         cdef double y214 = 3.0*y13 + 3.0*y18
 */
  __Pyx_TraceLine(967,0,__PYX_ERR(1, 967, __pyx_L1_error))
  __pyx_v_y212 = (4.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":968
 *         cdef double y211 = y205*y206*y210
 *         cdef double y212 = 4.0*nu
 *         cdef double y213 = 3.0*y7             # <<<<<<<<<<<<<<
 *         cdef double y214 = 3.0*y13 + 3.0*y18
 *         cdef double y215 = -y213 + y214
 */
  __Pyx_TraceLine(968,0,__PYX_ERR(1, 968, __pyx_L1_error))
  __pyx_v_y213 = (3.0 * __pyx_v_y7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":969
 *         cdef double y212 = 4.0*nu
 *         cdef double y213 = 3.0*y7
 *         cdef double y214 = 3.0*y13 + 3.0*y18             # <<<<<<<<<<<<<<
 *         cdef double y215 = -y213 + y214
 *         cdef double y216 = 0.0625*y215
 */
  __Pyx_TraceLine(969,0,__PYX_ERR(1, 969, __pyx_L1_error))
  __pyx_v_y214 = ((3.0 * __pyx_v_y13) + (3.0 * __pyx_v_y18));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":970
 *         cdef double y213 = 3.0*y7
 *         cdef double y214 = 3.0*y13 + 3.0*y18
 *         cdef double y215 = -y213 + y214             # <<<<<<<<<<<<<<
 *         cdef double y216 = 0.0625*y215
 *         cdef double y217 = y68*(26.0*nu + 449.0)
 */
  __Pyx_TraceLine(970,0,__PYX_ERR(1, 970, __pyx_L1_error))
  __pyx_v_y215 = ((-__pyx_v_y213) + __pyx_v_y214);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":971
 *         cdef double y214 = 3.0*y13 + 3.0*y18
 *         cdef double y215 = -y213 + y214
 *         cdef double y216 = 0.0625*y215             # <<<<<<<<<<<<<<
 *         cdef double y217 = y68*(26.0*nu + 449.0)
 *         cdef double y218 = y20*(-1171.0*nu - 861.0)
 */
  __Pyx_TraceLine(971,0,__PYX_ERR(1, 971, __pyx_L1_error))
  __pyx_v_y216 = (0.0625 * __pyx_v_y215);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":972
 *         cdef double y215 = -y213 + y214
 *         cdef double y216 = 0.0625*y215
 *         cdef double y217 = y68*(26.0*nu + 449.0)             # <<<<<<<<<<<<<<
 *         cdef double y218 = y20*(-1171.0*nu - 861.0)
 *         cdef double y219 = y135*(115.0*nu + y197 - 37.0)
 */
  __Pyx_TraceLine(972,0,__PYX_ERR(1, 972, __pyx_L1_error))
  __pyx_v_y217 = (__pyx_v_y68 * ((26.0 * __pyx_v_nu) + 449.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":973
 *         cdef double y216 = 0.0625*y215
 *         cdef double y217 = y68*(26.0*nu + 449.0)
 *         cdef double y218 = y20*(-1171.0*nu - 861.0)             # <<<<<<<<<<<<<<
 *         cdef double y219 = y135*(115.0*nu + y197 - 37.0)
 *         cdef double y220 = 5787938193408.0*y172 - 9216.0*y174 - 967680.0*y176 + 55296.0*y177 + y180 + y184
 */
  __Pyx_TraceLine(973,0,__PYX_ERR(1, 973, __pyx_L1_error))
  __pyx_v_y218 = (__pyx_v_y20 * ((-1171.0 * __pyx_v_nu) - 861.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":974
 *         cdef double y217 = y68*(26.0*nu + 449.0)
 *         cdef double y218 = y20*(-1171.0*nu - 861.0)
 *         cdef double y219 = y135*(115.0*nu + y197 - 37.0)             # <<<<<<<<<<<<<<
 *         cdef double y220 = 5787938193408.0*y172 - 9216.0*y174 - 967680.0*y176 + 55296.0*y177 + y180 + y184
 *         cdef double y221 = y220**(-1)
 */
  __Pyx_TraceLine(974,0,__PYX_ERR(1, 974, __pyx_L1_error))
  __pyx_v_y219 = (__pyx_v_y135 * (((115.0 * __pyx_v_nu) + __pyx_v_y197) - 37.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":975
 *         cdef double y218 = y20*(-1171.0*nu - 861.0)
 *         cdef double y219 = y135*(115.0*nu + y197 - 37.0)
 *         cdef double y220 = 5787938193408.0*y172 - 9216.0*y174 - 967680.0*y176 + 55296.0*y177 + y180 + y184             # <<<<<<<<<<<<<<
 *         cdef double y221 = y220**(-1)
 *         cdef double y222 = y221*y24
 */
  __Pyx_TraceLine(975,0,__PYX_ERR(1, 975, __pyx_L1_error))
  __pyx_v_y220 = ((((((5787938193408.0 * __pyx_v_y172) - (9216.0 * __pyx_v_y174)) - (967680.0 * __pyx_v_y176)) + (55296.0 * __pyx_v_y177)) + __pyx_v_y180) + __pyx_v_y184);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":976
 *         cdef double y219 = y135*(115.0*nu + y197 - 37.0)
 *         cdef double y220 = 5787938193408.0*y172 - 9216.0*y174 - 967680.0*y176 + 55296.0*y177 + y180 + y184
 *         cdef double y221 = y220**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y222 = y221*y24
 *         cdef double y223 = y209*y222
 */
  __Pyx_TraceLine(976,0,__PYX_ERR(1, 976, __pyx_L1_error))
  __pyx_v_y221 = pow(__pyx_v_y220, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":977
 *         cdef double y220 = 5787938193408.0*y172 - 9216.0*y174 - 967680.0*y176 + 55296.0*y177 + y180 + y184
 *         cdef double y221 = y220**(-1)
 *         cdef double y222 = y221*y24             # <<<<<<<<<<<<<<
 *         cdef double y223 = y209*y222
 *         cdef double y224 = 7680.0*y168
 */
  __Pyx_TraceLine(977,0,__PYX_ERR(1, 977, __pyx_L1_error))
  __pyx_v_y222 = (__pyx_v_y221 * __pyx_v_y24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":978
 *         cdef double y221 = y220**(-1)
 *         cdef double y222 = y221*y24
 *         cdef double y223 = y209*y222             # <<<<<<<<<<<<<<
 *         cdef double y224 = 7680.0*y168
 *         cdef double y225 = y145 + y223*y224 + y32*(y20*(y75 + 2.8125) + y216*(y212 - 1.0) - 2.625*y68) + y36*(0.03125*y217 + 0.015625*y218 + 0.015625*y219)
 */
  __Pyx_TraceLine(978,0,__PYX_ERR(1, 978, __pyx_L1_error))
  __pyx_v_y223 = (__pyx_v_y209 * __pyx_v_y222);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":979
 *         cdef double y222 = y221*y24
 *         cdef double y223 = y209*y222
 *         cdef double y224 = 7680.0*y168             # <<<<<<<<<<<<<<
 *         cdef double y225 = y145 + y223*y224 + y32*(y20*(y75 + 2.8125) + y216*(y212 - 1.0) - 2.625*y68) + y36*(0.03125*y217 + 0.015625*y218 + 0.015625*y219)
 *         cdef double y226 = y220*y225
 */
  __Pyx_TraceLine(979,0,__PYX_ERR(1, 979, __pyx_L1_error))
  __pyx_v_y224 = (7680.0 * __pyx_v_y168);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":980
 *         cdef double y223 = y209*y222
 *         cdef double y224 = 7680.0*y168
 *         cdef double y225 = y145 + y223*y224 + y32*(y20*(y75 + 2.8125) + y216*(y212 - 1.0) - 2.625*y68) + y36*(0.03125*y217 + 0.015625*y218 + 0.015625*y219)             # <<<<<<<<<<<<<<
 *         cdef double y226 = y220*y225
 *         cdef double y227 = y211*y226
 */
  __Pyx_TraceLine(980,0,__PYX_ERR(1, 980, __pyx_L1_error))
  __pyx_v_y225 = (((__pyx_v_y145 + (__pyx_v_y223 * __pyx_v_y224)) + (__pyx_v_y32 * (((__pyx_v_y20 * (__pyx_v_y75 + 2.8125)) + (__pyx_v_y216 * (__pyx_v_y212 - 1.0))) - (2.625 * __pyx_v_y68)))) + (__pyx_v_y36 * (((0.03125 * __pyx_v_y217) + (0.015625 * __pyx_v_y218)) + (0.015625 * __pyx_v_y219))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":981
 *         cdef double y224 = 7680.0*y168
 *         cdef double y225 = y145 + y223*y224 + y32*(y20*(y75 + 2.8125) + y216*(y212 - 1.0) - 2.625*y68) + y36*(0.03125*y217 + 0.015625*y218 + 0.015625*y219)
 *         cdef double y226 = y220*y225             # <<<<<<<<<<<<<<
 *         cdef double y227 = y211*y226
 *         cdef double y228 = y202**2
 */
  __Pyx_TraceLine(981,0,__PYX_ERR(1, 981, __pyx_L1_error))
  __pyx_v_y226 = (__pyx_v_y220 * __pyx_v_y225);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":982
 *         cdef double y225 = y145 + y223*y224 + y32*(y20*(y75 + 2.8125) + y216*(y212 - 1.0) - 2.625*y68) + y36*(0.03125*y217 + 0.015625*y218 + 0.015625*y219)
 *         cdef double y226 = y220*y225
 *         cdef double y227 = y211*y226             # <<<<<<<<<<<<<<
 *         cdef double y228 = y202**2
 *         cdef double y229 = y228*y25
 */
  __Pyx_TraceLine(982,0,__PYX_ERR(1, 982, __pyx_L1_error))
  __pyx_v_y227 = (__pyx_v_y211 * __pyx_v_y226);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":983
 *         cdef double y226 = y220*y225
 *         cdef double y227 = y211*y226
 *         cdef double y228 = y202**2             # <<<<<<<<<<<<<<
 *         cdef double y229 = y228*y25
 *         cdef double y230 = 1.69542100694444e-8*y229
 */
  __Pyx_TraceLine(983,0,__PYX_ERR(1, 983, __pyx_L1_error))
  __pyx_v_y228 = pow(__pyx_v_y202, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":984
 *         cdef double y227 = y211*y226
 *         cdef double y228 = y202**2
 *         cdef double y229 = y228*y25             # <<<<<<<<<<<<<<
 *         cdef double y230 = 1.69542100694444e-8*y229
 *         cdef double y231 = -y1*y131 + 147.443752990146*y100*y102 - 11.3175085791863*y103*y105 + 1.48275342024365*y106*y108 + y109*y110 + y112*y77 + y115*y32 + y117*y77 + y120*y32 + y122*y77 + y124*y125 + y144*y92 + 1.27277314139085e-19*y200*y204 + y227*y230 + 1.0
 */
  __Pyx_TraceLine(984,0,__PYX_ERR(1, 984, __pyx_L1_error))
  __pyx_v_y229 = (__pyx_v_y228 * __pyx_v_y25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":985
 *         cdef double y228 = y202**2
 *         cdef double y229 = y228*y25
 *         cdef double y230 = 1.69542100694444e-8*y229             # <<<<<<<<<<<<<<
 *         cdef double y231 = -y1*y131 + 147.443752990146*y100*y102 - 11.3175085791863*y103*y105 + 1.48275342024365*y106*y108 + y109*y110 + y112*y77 + y115*y32 + y117*y77 + y120*y32 + y122*y77 + y124*y125 + y144*y92 + 1.27277314139085e-19*y200*y204 + y227*y230 + 1.0
 *         cdef double y232 = 2.0*y91
 */
  __Pyx_TraceLine(985,0,__PYX_ERR(1, 985, __pyx_L1_error))
  __pyx_v_y230 = (1.69542100694444e-8 * __pyx_v_y229);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":986
 *         cdef double y229 = y228*y25
 *         cdef double y230 = 1.69542100694444e-8*y229
 *         cdef double y231 = -y1*y131 + 147.443752990146*y100*y102 - 11.3175085791863*y103*y105 + 1.48275342024365*y106*y108 + y109*y110 + y112*y77 + y115*y32 + y117*y77 + y120*y32 + y122*y77 + y124*y125 + y144*y92 + 1.27277314139085e-19*y200*y204 + y227*y230 + 1.0             # <<<<<<<<<<<<<<
 *         cdef double y232 = 2.0*y91
 *         cdef double y233 = y145*(y232 + 1.0) + 1.0
 */
  __Pyx_TraceLine(986,0,__PYX_ERR(1, 986, __pyx_L1_error))
  __pyx_v_y231 = ((((((((((((((((-__pyx_v_y1) * __pyx_v_y131) + ((147.443752990146 * __pyx_v_y100) * __pyx_v_y102)) - ((11.3175085791863 * __pyx_v_y103) * __pyx_v_y105)) + ((1.48275342024365 * __pyx_v_y106) * __pyx_v_y108)) + (__pyx_v_y109 * __pyx_v_y110)) + (__pyx_v_y112 * __pyx_v_y77)) + (__pyx_v_y115 * __pyx_v_y32)) + (__pyx_v_y117 * __pyx_v_y77)) + (__pyx_v_y120 * __pyx_v_y32)) + (__pyx_v_y122 * __pyx_v_y77)) + (__pyx_v_y124 * __pyx_v_y125)) + (__pyx_v_y144 * __pyx_v_y92)) + ((1.27277314139085e-19 * __pyx_v_y200) * __pyx_v_y204)) + (__pyx_v_y227 * __pyx_v_y230)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":987
 *         cdef double y230 = 1.69542100694444e-8*y229
 *         cdef double y231 = -y1*y131 + 147.443752990146*y100*y102 - 11.3175085791863*y103*y105 + 1.48275342024365*y106*y108 + y109*y110 + y112*y77 + y115*y32 + y117*y77 + y120*y32 + y122*y77 + y124*y125 + y144*y92 + 1.27277314139085e-19*y200*y204 + y227*y230 + 1.0
 *         cdef double y232 = 2.0*y91             # <<<<<<<<<<<<<<
 *         cdef double y233 = y145*(y232 + 1.0) + 1.0
 *         cdef double y234 = y233**(-1)
 */
  __Pyx_TraceLine(987,0,__PYX_ERR(1, 987, __pyx_L1_error))
  __pyx_v_y232 = (2.0 * __pyx_v_y91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":988
 *         cdef double y231 = -y1*y131 + 147.443752990146*y100*y102 - 11.3175085791863*y103*y105 + 1.48275342024365*y106*y108 + y109*y110 + y112*y77 + y115*y32 + y117*y77 + y120*y32 + y122*y77 + y124*y125 + y144*y92 + 1.27277314139085e-19*y200*y204 + y227*y230 + 1.0
 *         cdef double y232 = 2.0*y91
 *         cdef double y233 = y145*(y232 + 1.0) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double y234 = y233**(-1)
 *         cdef double y235 = y135*(y212 + 1.0)
 */
  __Pyx_TraceLine(988,0,__PYX_ERR(1, 988, __pyx_L1_error))
  __pyx_v_y233 = ((__pyx_v_y145 * (__pyx_v_y232 + 1.0)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":989
 *         cdef double y232 = 2.0*y91
 *         cdef double y233 = y145*(y232 + 1.0) + 1.0
 *         cdef double y234 = y233**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y235 = y135*(y212 + 1.0)
 *         cdef double y236 = y36*(1.125*y13 + 1.125*y18 + 0.125*y235 + y69 + 1.125*y7)
 */
  __Pyx_TraceLine(989,0,__PYX_ERR(1, 989, __pyx_L1_error))
  __pyx_v_y234 = pow(__pyx_v_y233, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":990
 *         cdef double y233 = y145*(y232 + 1.0) + 1.0
 *         cdef double y234 = y233**(-1)
 *         cdef double y235 = y135*(y212 + 1.0)             # <<<<<<<<<<<<<<
 *         cdef double y236 = y36*(1.125*y13 + 1.125*y18 + 0.125*y235 + y69 + 1.125*y7)
 *         cdef double y237 = y68*(2.0*nu - 3.0)
 */
  __Pyx_TraceLine(990,0,__PYX_ERR(1, 990, __pyx_L1_error))
  __pyx_v_y235 = (__pyx_v_y135 * (__pyx_v_y212 + 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":991
 *         cdef double y234 = y233**(-1)
 *         cdef double y235 = y135*(y212 + 1.0)
 *         cdef double y236 = y36*(1.125*y13 + 1.125*y18 + 0.125*y235 + y69 + 1.125*y7)             # <<<<<<<<<<<<<<
 *         cdef double y237 = y68*(2.0*nu - 3.0)
 *         cdef double y238 = 7.0*nu
 */
  __Pyx_TraceLine(991,0,__PYX_ERR(1, 991, __pyx_L1_error))
  __pyx_v_y236 = (__pyx_v_y36 * (((((1.125 * __pyx_v_y13) + (1.125 * __pyx_v_y18)) + (0.125 * __pyx_v_y235)) + __pyx_v_y69) + (1.125 * __pyx_v_y7)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":992
 *         cdef double y235 = y135*(y212 + 1.0)
 *         cdef double y236 = y36*(1.125*y13 + 1.125*y18 + 0.125*y235 + y69 + 1.125*y7)
 *         cdef double y237 = y68*(2.0*nu - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double y238 = 7.0*nu
 *         cdef double y239 = 25.0*y13 + 25.0*y18 + 25.0*y7
 */
  __Pyx_TraceLine(992,0,__PYX_ERR(1, 992, __pyx_L1_error))
  __pyx_v_y237 = (__pyx_v_y68 * ((2.0 * __pyx_v_nu) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":993
 *         cdef double y236 = y36*(1.125*y13 + 1.125*y18 + 0.125*y235 + y69 + 1.125*y7)
 *         cdef double y237 = y68*(2.0*nu - 3.0)
 *         cdef double y238 = 7.0*nu             # <<<<<<<<<<<<<<
 *         cdef double y239 = 25.0*y13 + 25.0*y18 + 25.0*y7
 *         cdef double y240 = 0.015625*y239
 */
  __Pyx_TraceLine(993,0,__PYX_ERR(1, 993, __pyx_L1_error))
  __pyx_v_y238 = (7.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":994
 *         cdef double y237 = y68*(2.0*nu - 3.0)
 *         cdef double y238 = 7.0*nu
 *         cdef double y239 = 25.0*y13 + 25.0*y18 + 25.0*y7             # <<<<<<<<<<<<<<
 *         cdef double y240 = 0.015625*y239
 *         cdef double y241 = y25*(0.015625*y215*(-27.0*nu + 28.0*y39 - 3.0) - 1.21875*y237 - y240*(y238 + 9.0))
 */
  __Pyx_TraceLine(994,0,__PYX_ERR(1, 994, __pyx_L1_error))
  __pyx_v_y239 = (((25.0 * __pyx_v_y13) + (25.0 * __pyx_v_y18)) + (25.0 * __pyx_v_y7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":995
 *         cdef double y238 = 7.0*nu
 *         cdef double y239 = 25.0*y13 + 25.0*y18 + 25.0*y7
 *         cdef double y240 = 0.015625*y239             # <<<<<<<<<<<<<<
 *         cdef double y241 = y25*(0.015625*y215*(-27.0*nu + 28.0*y39 - 3.0) - 1.21875*y237 - y240*(y238 + 9.0))
 *         cdef double y242 = y59*(y238 - 31.0)
 */
  __Pyx_TraceLine(995,0,__PYX_ERR(1, 995, __pyx_L1_error))
  __pyx_v_y240 = (0.015625 * __pyx_v_y239);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":996
 *         cdef double y239 = 25.0*y13 + 25.0*y18 + 25.0*y7
 *         cdef double y240 = 0.015625*y239
 *         cdef double y241 = y25*(0.015625*y215*(-27.0*nu + 28.0*y39 - 3.0) - 1.21875*y237 - y240*(y238 + 9.0))             # <<<<<<<<<<<<<<
 *         cdef double y242 = y59*(y238 - 31.0)
 *         cdef double y243 = y136*(-y212 - 3.0)
 */
  __Pyx_TraceLine(996,0,__PYX_ERR(1, 996, __pyx_L1_error))
  __pyx_v_y241 = (__pyx_v_y25 * ((((0.015625 * __pyx_v_y215) * (((-27.0 * __pyx_v_nu) + (28.0 * __pyx_v_y39)) - 3.0)) - (1.21875 * __pyx_v_y237)) - (__pyx_v_y240 * (__pyx_v_y238 + 9.0))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":997
 *         cdef double y240 = 0.015625*y239
 *         cdef double y241 = y25*(0.015625*y215*(-27.0*nu + 28.0*y39 - 3.0) - 1.21875*y237 - y240*(y238 + 9.0))
 *         cdef double y242 = y59*(y238 - 31.0)             # <<<<<<<<<<<<<<
 *         cdef double y243 = y136*(-y212 - 3.0)
 *         cdef double y244 = y36*(0.125*y242 + 0.0625*y243 + 2.0625*y70)
 */
  __Pyx_TraceLine(997,0,__PYX_ERR(1, 997, __pyx_L1_error))
  __pyx_v_y242 = (__pyx_v_y59 * (__pyx_v_y238 - 31.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":998
 *         cdef double y241 = y25*(0.015625*y215*(-27.0*nu + 28.0*y39 - 3.0) - 1.21875*y237 - y240*(y238 + 9.0))
 *         cdef double y242 = y59*(y238 - 31.0)
 *         cdef double y243 = y136*(-y212 - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double y244 = y36*(0.125*y242 + 0.0625*y243 + 2.0625*y70)
 *         cdef double y245 = y70*(68.0*nu - 1.0)
 */
  __Pyx_TraceLine(998,0,__PYX_ERR(1, 998, __pyx_L1_error))
  __pyx_v_y243 = (__pyx_v_y136 * ((-__pyx_v_y212) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":999
 *         cdef double y242 = y59*(y238 - 31.0)
 *         cdef double y243 = y136*(-y212 - 3.0)
 *         cdef double y244 = y36*(0.125*y242 + 0.0625*y243 + 2.0625*y70)             # <<<<<<<<<<<<<<
 *         cdef double y245 = y70*(68.0*nu - 1.0)
 *         cdef double y246 = y25*(y141*(2870.0*nu - y139 + 561.0) + y143*(1166.0*nu - 328.0*y39 - 171.0) + 0.125*y245)
 */
  __Pyx_TraceLine(999,0,__PYX_ERR(1, 999, __pyx_L1_error))
  __pyx_v_y244 = (__pyx_v_y36 * (((0.125 * __pyx_v_y242) + (0.0625 * __pyx_v_y243)) + (2.0625 * __pyx_v_y70)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1000
 *         cdef double y243 = y136*(-y212 - 3.0)
 *         cdef double y244 = y36*(0.125*y242 + 0.0625*y243 + 2.0625*y70)
 *         cdef double y245 = y70*(68.0*nu - 1.0)             # <<<<<<<<<<<<<<
 *         cdef double y246 = y25*(y141*(2870.0*nu - y139 + 561.0) + y143*(1166.0*nu - 328.0*y39 - 171.0) + 0.125*y245)
 *         cdef double y247 = y145 + 7680.0*y169 + y236 + y241 + y244 + y246 + y96
 */
  __Pyx_TraceLine(1000,0,__PYX_ERR(1, 1000, __pyx_L1_error))
  __pyx_v_y245 = (__pyx_v_y70 * ((68.0 * __pyx_v_nu) - 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1001
 *         cdef double y244 = y36*(0.125*y242 + 0.0625*y243 + 2.0625*y70)
 *         cdef double y245 = y70*(68.0*nu - 1.0)
 *         cdef double y246 = y25*(y141*(2870.0*nu - y139 + 561.0) + y143*(1166.0*nu - 328.0*y39 - 171.0) + 0.125*y245)             # <<<<<<<<<<<<<<
 *         cdef double y247 = y145 + 7680.0*y169 + y236 + y241 + y244 + y246 + y96
 *         cdef double y248 = y234*y247
 */
  __Pyx_TraceLine(1001,0,__PYX_ERR(1, 1001, __pyx_L1_error))
  __pyx_v_y246 = (__pyx_v_y25 * (((__pyx_v_y141 * (((2870.0 * __pyx_v_nu) - __pyx_v_y139) + 561.0)) + (__pyx_v_y143 * (((1166.0 * __pyx_v_nu) - (328.0 * __pyx_v_y39)) - 171.0))) + (0.125 * __pyx_v_y245)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1002
 *         cdef double y245 = y70*(68.0*nu - 1.0)
 *         cdef double y246 = y25*(y141*(2870.0*nu - y139 + 561.0) + y143*(1166.0*nu - 328.0*y39 - 171.0) + 0.125*y245)
 *         cdef double y247 = y145 + 7680.0*y169 + y236 + y241 + y244 + y246 + y96             # <<<<<<<<<<<<<<
 *         cdef double y248 = y234*y247
 *         cdef double y249 = y231*y248
 */
  __Pyx_TraceLine(1002,0,__PYX_ERR(1, 1002, __pyx_L1_error))
  __pyx_v_y247 = ((((((__pyx_v_y145 + (7680.0 * __pyx_v_y169)) + __pyx_v_y236) + __pyx_v_y241) + __pyx_v_y244) + __pyx_v_y246) + __pyx_v_y96);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1003
 *         cdef double y246 = y25*(y141*(2870.0*nu - y139 + 561.0) + y143*(1166.0*nu - 328.0*y39 - 171.0) + 0.125*y245)
 *         cdef double y247 = y145 + 7680.0*y169 + y236 + y241 + y244 + y246 + y96
 *         cdef double y248 = y234*y247             # <<<<<<<<<<<<<<
 *         cdef double y249 = y231*y248
 *         cdef double y250 = y249**(-1.5)
 */
  __Pyx_TraceLine(1003,0,__PYX_ERR(1, 1003, __pyx_L1_error))
  __pyx_v_y248 = (__pyx_v_y234 * __pyx_v_y247);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1004
 *         cdef double y247 = y145 + 7680.0*y169 + y236 + y241 + y244 + y246 + y96
 *         cdef double y248 = y234*y247
 *         cdef double y249 = y231*y248             # <<<<<<<<<<<<<<
 *         cdef double y250 = y249**(-1.5)
 *         cdef double y251 = -y73
 */
  __Pyx_TraceLine(1004,0,__PYX_ERR(1, 1004, __pyx_L1_error))
  __pyx_v_y249 = (__pyx_v_y231 * __pyx_v_y248);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1005
 *         cdef double y248 = y234*y247
 *         cdef double y249 = y231*y248
 *         cdef double y250 = y249**(-1.5)             # <<<<<<<<<<<<<<
 *         cdef double y251 = -y73
 *         cdef double y252 = 4.0*y91 + 2.0
 */
  __Pyx_TraceLine(1005,0,__PYX_ERR(1, 1005, __pyx_L1_error))
  __pyx_v_y250 = pow(__pyx_v_y249, -1.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1006
 *         cdef double y249 = y231*y248
 *         cdef double y250 = y249**(-1.5)
 *         cdef double y251 = -y73             # <<<<<<<<<<<<<<
 *         cdef double y252 = 4.0*y91 + 2.0
 *         cdef double y253 = y20*y32
 */
  __Pyx_TraceLine(1006,0,__PYX_ERR(1, 1006, __pyx_L1_error))
  __pyx_v_y251 = (-__pyx_v_y73);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1007
 *         cdef double y250 = y249**(-1.5)
 *         cdef double y251 = -y73
 *         cdef double y252 = 4.0*y91 + 2.0             # <<<<<<<<<<<<<<
 *         cdef double y253 = y20*y32
 *         cdef double y254 = y251*y36 - y252*y253
 */
  __Pyx_TraceLine(1007,0,__PYX_ERR(1, 1007, __pyx_L1_error))
  __pyx_v_y252 = ((4.0 * __pyx_v_y91) + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1008
 *         cdef double y251 = -y73
 *         cdef double y252 = 4.0*y91 + 2.0
 *         cdef double y253 = y20*y32             # <<<<<<<<<<<<<<
 *         cdef double y254 = y251*y36 - y252*y253
 *         cdef double y255 = y233**(-2)
 */
  __Pyx_TraceLine(1008,0,__PYX_ERR(1, 1008, __pyx_L1_error))
  __pyx_v_y253 = (__pyx_v_y20 * __pyx_v_y32);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1009
 *         cdef double y252 = 4.0*y91 + 2.0
 *         cdef double y253 = y20*y32
 *         cdef double y254 = y251*y36 - y252*y253             # <<<<<<<<<<<<<<
 *         cdef double y255 = y233**(-2)
 *         cdef double y256 = y247*y255
 */
  __Pyx_TraceLine(1009,0,__PYX_ERR(1, 1009, __pyx_L1_error))
  __pyx_v_y254 = ((__pyx_v_y251 * __pyx_v_y36) - (__pyx_v_y252 * __pyx_v_y253));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1010
 *         cdef double y253 = y20*y32
 *         cdef double y254 = y251*y36 - y252*y253
 *         cdef double y255 = y233**(-2)             # <<<<<<<<<<<<<<
 *         cdef double y256 = y247*y255
 *         cdef double y257 = y231*y256
 */
  __Pyx_TraceLine(1010,0,__PYX_ERR(1, 1010, __pyx_L1_error))
  __pyx_v_y255 = pow(__pyx_v_y233, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1011
 *         cdef double y254 = y251*y36 - y252*y253
 *         cdef double y255 = y233**(-2)
 *         cdef double y256 = y247*y255             # <<<<<<<<<<<<<<
 *         cdef double y257 = y231*y256
 *         cdef double y258 = y213 + y214 - 3.0*y58
 */
  __Pyx_TraceLine(1011,0,__PYX_ERR(1, 1011, __pyx_L1_error))
  __pyx_v_y256 = (__pyx_v_y247 * __pyx_v_y255);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1012
 *         cdef double y255 = y233**(-2)
 *         cdef double y256 = y247*y255
 *         cdef double y257 = y231*y256             # <<<<<<<<<<<<<<
 *         cdef double y258 = y213 + y214 - 3.0*y58
 *         cdef double y259 = 4.5*X_2
 */
  __Pyx_TraceLine(1012,0,__PYX_ERR(1, 1012, __pyx_L1_error))
  __pyx_v_y257 = (__pyx_v_y231 * __pyx_v_y256);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1013
 *         cdef double y256 = y247*y255
 *         cdef double y257 = y231*y256
 *         cdef double y258 = y213 + y214 - 3.0*y58             # <<<<<<<<<<<<<<
 *         cdef double y259 = 4.5*X_2
 *         cdef double y260 = X_1*y5
 */
  __Pyx_TraceLine(1013,0,__PYX_ERR(1, 1013, __pyx_L1_error))
  __pyx_v_y258 = ((__pyx_v_y213 + __pyx_v_y214) - (3.0 * __pyx_v_y58));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1014
 *         cdef double y257 = y231*y256
 *         cdef double y258 = y213 + y214 - 3.0*y58
 *         cdef double y259 = 4.5*X_2             # <<<<<<<<<<<<<<
 *         cdef double y260 = X_1*y5
 *         cdef double y261 = 35.0*nu
 */
  __Pyx_TraceLine(1014,0,__PYX_ERR(1, 1014, __pyx_L1_error))
  __pyx_v_y259 = (4.5 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1015
 *         cdef double y258 = y213 + y214 - 3.0*y58
 *         cdef double y259 = 4.5*X_2
 *         cdef double y260 = X_1*y5             # <<<<<<<<<<<<<<
 *         cdef double y261 = 35.0*nu
 *         cdef double y262 = y261 + 45.0
 */
  __Pyx_TraceLine(1015,0,__PYX_ERR(1, 1015, __pyx_L1_error))
  __pyx_v_y260 = (__pyx_v_X_1 * __pyx_v_y5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1016
 *         cdef double y259 = 4.5*X_2
 *         cdef double y260 = X_1*y5
 *         cdef double y261 = 35.0*nu             # <<<<<<<<<<<<<<
 *         cdef double y262 = y261 + 45.0
 *         cdef double y263 = -135.0*nu + 140.0*y39 - 15.0
 */
  __Pyx_TraceLine(1016,0,__PYX_ERR(1, 1016, __pyx_L1_error))
  __pyx_v_y261 = (35.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1017
 *         cdef double y260 = X_1*y5
 *         cdef double y261 = 35.0*nu
 *         cdef double y262 = y261 + 45.0             # <<<<<<<<<<<<<<
 *         cdef double y263 = -135.0*nu + 140.0*y39 - 15.0
 *         cdef double y264 = 1320.0*y39
 */
  __Pyx_TraceLine(1017,0,__PYX_ERR(1, 1017, __pyx_L1_error))
  __pyx_v_y262 = (__pyx_v_y261 + 45.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1018
 *         cdef double y261 = 35.0*nu
 *         cdef double y262 = y261 + 45.0
 *         cdef double y263 = -135.0*nu + 140.0*y39 - 15.0             # <<<<<<<<<<<<<<
 *         cdef double y264 = 1320.0*y39
 *         cdef double y265 = 14350.0*nu - y264 + 2805.0
 */
  __Pyx_TraceLine(1018,0,__PYX_ERR(1, 1018, __pyx_L1_error))
  __pyx_v_y263 = (((-135.0 * __pyx_v_nu) + (140.0 * __pyx_v_y39)) - 15.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1019
 *         cdef double y262 = y261 + 45.0
 *         cdef double y263 = -135.0*nu + 140.0*y39 - 15.0
 *         cdef double y264 = 1320.0*y39             # <<<<<<<<<<<<<<
 *         cdef double y265 = 14350.0*nu - y264 + 2805.0
 *         cdef double y266 = 5830.0*nu - 1640.0*y39 - 855.0
 */
  __Pyx_TraceLine(1019,0,__PYX_ERR(1, 1019, __pyx_L1_error))
  __pyx_v_y264 = (1320.0 * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1020
 *         cdef double y263 = -135.0*nu + 140.0*y39 - 15.0
 *         cdef double y264 = 1320.0*y39
 *         cdef double y265 = 14350.0*nu - y264 + 2805.0             # <<<<<<<<<<<<<<
 *         cdef double y266 = 5830.0*nu - 1640.0*y39 - 855.0
 *         cdef double y267 = 2048.0*y166
 */
  __Pyx_TraceLine(1020,0,__PYX_ERR(1, 1020, __pyx_L1_error))
  __pyx_v_y265 = (((14350.0 * __pyx_v_nu) - __pyx_v_y264) + 2805.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1021
 *         cdef double y264 = 1320.0*y39
 *         cdef double y265 = 14350.0*nu - y264 + 2805.0
 *         cdef double y266 = 5830.0*nu - 1640.0*y39 - 855.0             # <<<<<<<<<<<<<<
 *         cdef double y267 = 2048.0*y166
 *         cdef double y268 = -6572428.80109422*nu + y109*y267 + 688128.0*y155 + 1300341.64885296*y39 + 1720320.0
 */
  __Pyx_TraceLine(1021,0,__PYX_ERR(1, 1021, __pyx_L1_error))
  __pyx_v_y266 = (((5830.0 * __pyx_v_nu) - (1640.0 * __pyx_v_y39)) - 855.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1022
 *         cdef double y265 = 14350.0*nu - y264 + 2805.0
 *         cdef double y266 = 5830.0*nu - 1640.0*y39 - 855.0
 *         cdef double y267 = 2048.0*y166             # <<<<<<<<<<<<<<
 *         cdef double y268 = -6572428.80109422*nu + y109*y267 + 688128.0*y155 + 1300341.64885296*y39 + 1720320.0
 *         cdef double y269 = y167*y35
 */
  __Pyx_TraceLine(1022,0,__PYX_ERR(1, 1022, __pyx_L1_error))
  __pyx_v_y267 = (2048.0 * __pyx_v_y166);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1023
 *         cdef double y266 = 5830.0*nu - 1640.0*y39 - 855.0
 *         cdef double y267 = 2048.0*y166
 *         cdef double y268 = -6572428.80109422*nu + y109*y267 + 688128.0*y155 + 1300341.64885296*y39 + 1720320.0             # <<<<<<<<<<<<<<
 *         cdef double y269 = y167*y35
 *         cdef double y270 = y118 + 0.28004222119933*y147 + 4.63661586574928e-9*y148 + 2.8978849160933e-11*y149 + 1.35654132757922e-7*y156 + 2.22557561555966e-7*y164 + 0.0546957463279941*y24
 */
  __Pyx_TraceLine(1023,0,__PYX_ERR(1, 1023, __pyx_L1_error))
  __pyx_v_y268 = (((((-6572428.80109422 * __pyx_v_nu) + (__pyx_v_y109 * __pyx_v_y267)) + (688128.0 * __pyx_v_y155)) + (1300341.64885296 * __pyx_v_y39)) + 1720320.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1024
 *         cdef double y267 = 2048.0*y166
 *         cdef double y268 = -6572428.80109422*nu + y109*y267 + 688128.0*y155 + 1300341.64885296*y39 + 1720320.0
 *         cdef double y269 = y167*y35             # <<<<<<<<<<<<<<
 *         cdef double y270 = y118 + 0.28004222119933*y147 + 4.63661586574928e-9*y148 + 2.8978849160933e-11*y149 + 1.35654132757922e-7*y156 + 2.22557561555966e-7*y164 + 0.0546957463279941*y24
 *         cdef double y271 = y270**(-2)
 */
  __Pyx_TraceLine(1024,0,__PYX_ERR(1, 1024, __pyx_L1_error))
  __pyx_v_y269 = (__pyx_v_y167 * __pyx_v_y35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1025
 *         cdef double y268 = -6572428.80109422*nu + y109*y267 + 688128.0*y155 + 1300341.64885296*y39 + 1720320.0
 *         cdef double y269 = y167*y35
 *         cdef double y270 = y118 + 0.28004222119933*y147 + 4.63661586574928e-9*y148 + 2.8978849160933e-11*y149 + 1.35654132757922e-7*y156 + 2.22557561555966e-7*y164 + 0.0546957463279941*y24             # <<<<<<<<<<<<<<
 *         cdef double y271 = y270**(-2)
 *         cdef double y272 = y113*(-5706642207.53758*r - 26189434371.6082)
 */
  __Pyx_TraceLine(1025,0,__PYX_ERR(1, 1025, __pyx_L1_error))
  __pyx_v_y270 = ((((((__pyx_v_y118 + (0.28004222119933 * __pyx_v_y147)) + (4.63661586574928e-9 * __pyx_v_y148)) + (2.8978849160933e-11 * __pyx_v_y149)) + (1.35654132757922e-7 * __pyx_v_y156)) + (2.22557561555966e-7 * __pyx_v_y164)) + (0.0546957463279941 * __pyx_v_y24));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1026
 *         cdef double y269 = y167*y35
 *         cdef double y270 = y118 + 0.28004222119933*y147 + 4.63661586574928e-9*y148 + 2.8978849160933e-11*y149 + 1.35654132757922e-7*y156 + 2.22557561555966e-7*y164 + 0.0546957463279941*y24
 *         cdef double y271 = y270**(-2)             # <<<<<<<<<<<<<<
 *         cdef double y272 = y113*(-5706642207.53758*r - 26189434371.6082)
 *         cdef double y273 = y123*y91
 */
  __Pyx_TraceLine(1026,0,__PYX_ERR(1, 1026, __pyx_L1_error))
  __pyx_v_y271 = pow(__pyx_v_y270, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1027
 *         cdef double y270 = y118 + 0.28004222119933*y147 + 4.63661586574928e-9*y148 + 2.8978849160933e-11*y149 + 1.35654132757922e-7*y156 + 2.22557561555966e-7*y164 + 0.0546957463279941*y24
 *         cdef double y271 = y270**(-2)
 *         cdef double y272 = y113*(-5706642207.53758*r - 26189434371.6082)             # <<<<<<<<<<<<<<
 *         cdef double y273 = y123*y91
 *         cdef double y274 = y273*y39
 */
  __Pyx_TraceLine(1027,0,__PYX_ERR(1, 1027, __pyx_L1_error))
  __pyx_v_y272 = (__pyx_v_y113 * ((-5706642207.53758 * __pyx_v_r) - 26189434371.6082));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1028
 *         cdef double y271 = y270**(-2)
 *         cdef double y272 = y113*(-5706642207.53758*r - 26189434371.6082)
 *         cdef double y273 = y123*y91             # <<<<<<<<<<<<<<
 *         cdef double y274 = y273*y39
 *         cdef double y275 = y39*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*y0 + 7133302759.42198*y35 - 18535504222.6128*y71 + 122635399361.987)
 */
  __Pyx_TraceLine(1028,0,__PYX_ERR(1, 1028, __pyx_L1_error))
  __pyx_v_y273 = (__pyx_v_y123 * __pyx_v_y91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1029
 *         cdef double y272 = y113*(-5706642207.53758*r - 26189434371.6082)
 *         cdef double y273 = y123*y91
 *         cdef double y274 = y273*y39             # <<<<<<<<<<<<<<
 *         cdef double y275 = y39*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*y0 + 7133302759.42198*y35 - 18535504222.6128*y71 + 122635399361.987)
 *         cdef double y276 = y152*y71
 */
  __Pyx_TraceLine(1029,0,__PYX_ERR(1, 1029, __pyx_L1_error))
  __pyx_v_y274 = (__pyx_v_y273 * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1030
 *         cdef double y273 = y123*y91
 *         cdef double y274 = y273*y39
 *         cdef double y275 = y39*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*y0 + 7133302759.42198*y35 - 18535504222.6128*y71 + 122635399361.987)             # <<<<<<<<<<<<<<
 *         cdef double y276 = y152*y71
 *         cdef double y277 = -1791904.9465589*nu + 3840.0*r*y150 + 1920.0*y153 + 2880.0*y276 + 806400.0*y35 - 725760.0*y39 + 4143360.0
 */
  __Pyx_TraceLine(1030,0,__PYX_ERR(1, 1030, __pyx_L1_error))
  __pyx_v_y275 = (__pyx_v_y39 * ((((((-117964800.0 * __pyx_v_a6) + (4129567622.65173 * __pyx_v_r)) - (12357002815.0752 * __pyx_v_y0)) + (7133302759.42198 * __pyx_v_y35)) - (18535504222.6128 * __pyx_v_y71)) + 122635399361.987));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1031
 *         cdef double y274 = y273*y39
 *         cdef double y275 = y39*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*y0 + 7133302759.42198*y35 - 18535504222.6128*y71 + 122635399361.987)
 *         cdef double y276 = y152*y71             # <<<<<<<<<<<<<<
 *         cdef double y277 = -1791904.9465589*nu + 3840.0*r*y150 + 1920.0*y153 + 2880.0*y276 + 806400.0*y35 - 725760.0*y39 + 4143360.0
 *         cdef double y278 = y155*y277
 */
  __Pyx_TraceLine(1031,0,__PYX_ERR(1, 1031, __pyx_L1_error))
  __pyx_v_y276 = (__pyx_v_y152 * __pyx_v_y71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1032
 *         cdef double y275 = y39*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*y0 + 7133302759.42198*y35 - 18535504222.6128*y71 + 122635399361.987)
 *         cdef double y276 = y152*y71
 *         cdef double y277 = -1791904.9465589*nu + 3840.0*r*y150 + 1920.0*y153 + 2880.0*y276 + 806400.0*y35 - 725760.0*y39 + 4143360.0             # <<<<<<<<<<<<<<
 *         cdef double y278 = y155*y277
 *         cdef double y279 = y109*y154
 */
  __Pyx_TraceLine(1032,0,__PYX_ERR(1, 1032, __pyx_L1_error))
  __pyx_v_y277 = (((((((-1791904.9465589 * __pyx_v_nu) + ((3840.0 * __pyx_v_r) * __pyx_v_y150)) + (1920.0 * __pyx_v_y153)) + (2880.0 * __pyx_v_y276)) + (806400.0 * __pyx_v_y35)) - (725760.0 * __pyx_v_y39)) + 4143360.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1033
 *         cdef double y276 = y152*y71
 *         cdef double y277 = -1791904.9465589*nu + 3840.0*r*y150 + 1920.0*y153 + 2880.0*y276 + 806400.0*y35 - 725760.0*y39 + 4143360.0
 *         cdef double y278 = y155*y277             # <<<<<<<<<<<<<<
 *         cdef double y279 = y109*y154
 *         cdef double y280 = 4.0*y0
 */
  __Pyx_TraceLine(1033,0,__PYX_ERR(1, 1033, __pyx_L1_error))
  __pyx_v_y278 = (__pyx_v_y155 * __pyx_v_y277);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1034
 *         cdef double y277 = -1791904.9465589*nu + 3840.0*r*y150 + 1920.0*y153 + 2880.0*y276 + 806400.0*y35 - 725760.0*y39 + 4143360.0
 *         cdef double y278 = y155*y277
 *         cdef double y279 = y109*y154             # <<<<<<<<<<<<<<
 *         cdef double y280 = 4.0*y0
 *         cdef double y281 = y157 + y72 + 8.0
 */
  __Pyx_TraceLine(1034,0,__PYX_ERR(1, 1034, __pyx_L1_error))
  __pyx_v_y279 = (__pyx_v_y109 * __pyx_v_y154);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1035
 *         cdef double y278 = y155*y277
 *         cdef double y279 = y109*y154
 *         cdef double y280 = 4.0*y0             # <<<<<<<<<<<<<<
 *         cdef double y281 = y157 + y72 + 8.0
 *         cdef double y282 = nu*(3740417.71815805*r + 1057984.42953951*y0 + y159*(y280 + y281) + y160*(17058.7851094412*r - 27409.3925547206*y0 + 12794.0888320809*y71 + 13218.7851094412) + y162*(1396.0*y0 + y161 + 5778.0*y71 + 7704.0) + y163*(-y280 + y281) - 938918.400156317*y35 + 2115968.85907902*y71 + 2888096.47013111)
 */
  __Pyx_TraceLine(1035,0,__PYX_ERR(1, 1035, __pyx_L1_error))
  __pyx_v_y280 = (4.0 * __pyx_v_y0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1036
 *         cdef double y279 = y109*y154
 *         cdef double y280 = 4.0*y0
 *         cdef double y281 = y157 + y72 + 8.0             # <<<<<<<<<<<<<<
 *         cdef double y282 = nu*(3740417.71815805*r + 1057984.42953951*y0 + y159*(y280 + y281) + y160*(17058.7851094412*r - 27409.3925547206*y0 + 12794.0888320809*y71 + 13218.7851094412) + y162*(1396.0*y0 + y161 + 5778.0*y71 + 7704.0) + y163*(-y280 + y281) - 938918.400156317*y35 + 2115968.85907902*y71 + 2888096.47013111)
 *         cdef double y283 = y271*(7.0*y272 + 135291469824.0*y274 + 7.0*y275 + 32768.0*y278 + 32768.0*y279 + 53760.0*y282 + 66060288000.0*y35)
 */
  __Pyx_TraceLine(1036,0,__PYX_ERR(1, 1036, __pyx_L1_error))
  __pyx_v_y281 = ((__pyx_v_y157 + __pyx_v_y72) + 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1037
 *         cdef double y280 = 4.0*y0
 *         cdef double y281 = y157 + y72 + 8.0
 *         cdef double y282 = nu*(3740417.71815805*r + 1057984.42953951*y0 + y159*(y280 + y281) + y160*(17058.7851094412*r - 27409.3925547206*y0 + 12794.0888320809*y71 + 13218.7851094412) + y162*(1396.0*y0 + y161 + 5778.0*y71 + 7704.0) + y163*(-y280 + y281) - 938918.400156317*y35 + 2115968.85907902*y71 + 2888096.47013111)             # <<<<<<<<<<<<<<
 *         cdef double y283 = y271*(7.0*y272 + 135291469824.0*y274 + 7.0*y275 + 32768.0*y278 + 32768.0*y279 + 53760.0*y282 + 66060288000.0*y35)
 *         cdef double y284 = y269*y283
 */
  __Pyx_TraceLine(1037,0,__PYX_ERR(1, 1037, __pyx_L1_error))
  __pyx_v_y282 = (__pyx_v_nu * (((((((((3740417.71815805 * __pyx_v_r) + (1057984.42953951 * __pyx_v_y0)) + (__pyx_v_y159 * (__pyx_v_y280 + __pyx_v_y281))) + (__pyx_v_y160 * ((((17058.7851094412 * __pyx_v_r) - (27409.3925547206 * __pyx_v_y0)) + (12794.0888320809 * __pyx_v_y71)) + 13218.7851094412))) + (__pyx_v_y162 * ((((1396.0 * __pyx_v_y0) + __pyx_v_y161) + (5778.0 * __pyx_v_y71)) + 7704.0))) + (__pyx_v_y163 * ((-__pyx_v_y280) + __pyx_v_y281))) - (938918.400156317 * __pyx_v_y35)) + (2115968.85907902 * __pyx_v_y71)) + 2888096.47013111));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1038
 *         cdef double y281 = y157 + y72 + 8.0
 *         cdef double y282 = nu*(3740417.71815805*r + 1057984.42953951*y0 + y159*(y280 + y281) + y160*(17058.7851094412*r - 27409.3925547206*y0 + 12794.0888320809*y71 + 13218.7851094412) + y162*(1396.0*y0 + y161 + 5778.0*y71 + 7704.0) + y163*(-y280 + y281) - 938918.400156317*y35 + 2115968.85907902*y71 + 2888096.47013111)
 *         cdef double y283 = y271*(7.0*y272 + 135291469824.0*y274 + 7.0*y275 + 32768.0*y278 + 32768.0*y279 + 53760.0*y282 + 66060288000.0*y35)             # <<<<<<<<<<<<<<
 *         cdef double y284 = y269*y283
 *         cdef double y285 = 1.31621673590926e-19*y284
 */
  __Pyx_TraceLine(1038,0,__PYX_ERR(1, 1038, __pyx_L1_error))
  __pyx_v_y283 = (__pyx_v_y271 * (((((((7.0 * __pyx_v_y272) + (135291469824.0 * __pyx_v_y274)) + (7.0 * __pyx_v_y275)) + (32768.0 * __pyx_v_y278)) + (32768.0 * __pyx_v_y279)) + (53760.0 * __pyx_v_y282)) + (66060288000.0 * __pyx_v_y35)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1039
 *         cdef double y282 = nu*(3740417.71815805*r + 1057984.42953951*y0 + y159*(y280 + y281) + y160*(17058.7851094412*r - 27409.3925547206*y0 + 12794.0888320809*y71 + 13218.7851094412) + y162*(1396.0*y0 + y161 + 5778.0*y71 + 7704.0) + y163*(-y280 + y281) - 938918.400156317*y35 + 2115968.85907902*y71 + 2888096.47013111)
 *         cdef double y283 = y271*(7.0*y272 + 135291469824.0*y274 + 7.0*y275 + 32768.0*y278 + 32768.0*y279 + 53760.0*y282 + 66060288000.0*y35)
 *         cdef double y284 = y269*y283             # <<<<<<<<<<<<<<
 *         cdef double y285 = 1.31621673590926e-19*y284
 *         cdef double y286 = 30720.0*y0*y165*y167 + 7680.0*y165*y268*y35 - y25*(0.5*y242 + 0.25*y243 + 8.25*y70) - y25*(4.5*y13 + 4.5*y18 + 0.5*y235 + y259*y260 - 5.0*y68) + y251*y32 - y258*y36 - y285 - y43*(y141*y265 + y143*y266 + 0.625*y245) - y43*(0.015625*y215*y263 - 6.09375*y237 - y240*y262)
 */
  __Pyx_TraceLine(1039,0,__PYX_ERR(1, 1039, __pyx_L1_error))
  __pyx_v_y284 = (__pyx_v_y269 * __pyx_v_y283);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1040
 *         cdef double y283 = y271*(7.0*y272 + 135291469824.0*y274 + 7.0*y275 + 32768.0*y278 + 32768.0*y279 + 53760.0*y282 + 66060288000.0*y35)
 *         cdef double y284 = y269*y283
 *         cdef double y285 = 1.31621673590926e-19*y284             # <<<<<<<<<<<<<<
 *         cdef double y286 = 30720.0*y0*y165*y167 + 7680.0*y165*y268*y35 - y25*(0.5*y242 + 0.25*y243 + 8.25*y70) - y25*(4.5*y13 + 4.5*y18 + 0.5*y235 + y259*y260 - 5.0*y68) + y251*y32 - y258*y36 - y285 - y43*(y141*y265 + y143*y266 + 0.625*y245) - y43*(0.015625*y215*y263 - 6.09375*y237 - y240*y262)
 *         cdef double y287 = y231*y234
 */
  __Pyx_TraceLine(1040,0,__PYX_ERR(1, 1040, __pyx_L1_error))
  __pyx_v_y285 = (1.31621673590926e-19 * __pyx_v_y284);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1041
 *         cdef double y284 = y269*y283
 *         cdef double y285 = 1.31621673590926e-19*y284
 *         cdef double y286 = 30720.0*y0*y165*y167 + 7680.0*y165*y268*y35 - y25*(0.5*y242 + 0.25*y243 + 8.25*y70) - y25*(4.5*y13 + 4.5*y18 + 0.5*y235 + y259*y260 - 5.0*y68) + y251*y32 - y258*y36 - y285 - y43*(y141*y265 + y143*y266 + 0.625*y245) - y43*(0.015625*y215*y263 - 6.09375*y237 - y240*y262)             # <<<<<<<<<<<<<<
 *         cdef double y287 = y231*y234
 *         cdef double y288 = r**(-5.5)
 */
  __Pyx_TraceLine(1041,0,__PYX_ERR(1, 1041, __pyx_L1_error))
  __pyx_v_y286 = (((((((((((30720.0 * __pyx_v_y0) * __pyx_v_y165) * __pyx_v_y167) + (((7680.0 * __pyx_v_y165) * __pyx_v_y268) * __pyx_v_y35)) - (__pyx_v_y25 * (((0.5 * __pyx_v_y242) + (0.25 * __pyx_v_y243)) + (8.25 * __pyx_v_y70)))) - (__pyx_v_y25 * (((((4.5 * __pyx_v_y13) + (4.5 * __pyx_v_y18)) + (0.5 * __pyx_v_y235)) + (__pyx_v_y259 * __pyx_v_y260)) - (5.0 * __pyx_v_y68)))) + (__pyx_v_y251 * __pyx_v_y32)) - (__pyx_v_y258 * __pyx_v_y36)) - __pyx_v_y285) - (__pyx_v_y43 * (((__pyx_v_y141 * __pyx_v_y265) + (__pyx_v_y143 * __pyx_v_y266)) + (0.625 * __pyx_v_y245)))) - (__pyx_v_y43 * ((((0.015625 * __pyx_v_y215) * __pyx_v_y263) - (6.09375 * __pyx_v_y237)) - (__pyx_v_y240 * __pyx_v_y262))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1042
 *         cdef double y285 = 1.31621673590926e-19*y284
 *         cdef double y286 = 30720.0*y0*y165*y167 + 7680.0*y165*y268*y35 - y25*(0.5*y242 + 0.25*y243 + 8.25*y70) - y25*(4.5*y13 + 4.5*y18 + 0.5*y235 + y259*y260 - 5.0*y68) + y251*y32 - y258*y36 - y285 - y43*(y141*y265 + y143*y266 + 0.625*y245) - y43*(0.015625*y215*y263 - 6.09375*y237 - y240*y262)
 *         cdef double y287 = y231*y234             # <<<<<<<<<<<<<<
 *         cdef double y288 = r**(-5.5)
 *         cdef double y289 = nu*y100
 */
  __Pyx_TraceLine(1042,0,__PYX_ERR(1, 1042, __pyx_L1_error))
  __pyx_v_y287 = (__pyx_v_y231 * __pyx_v_y234);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1043
 *         cdef double y286 = 30720.0*y0*y165*y167 + 7680.0*y165*y268*y35 - y25*(0.5*y242 + 0.25*y243 + 8.25*y70) - y25*(4.5*y13 + 4.5*y18 + 0.5*y235 + y259*y260 - 5.0*y68) + y251*y32 - y258*y36 - y285 - y43*(y141*y265 + y143*y266 + 0.625*y245) - y43*(0.015625*y215*y263 - 6.09375*y237 - y240*y262)
 *         cdef double y287 = y231*y234
 *         cdef double y288 = r**(-5.5)             # <<<<<<<<<<<<<<
 *         cdef double y289 = nu*y100
 *         cdef double y290 = nu*y77
 */
  __Pyx_TraceLine(1043,0,__PYX_ERR(1, 1043, __pyx_L1_error))
  __pyx_v_y288 = pow(__pyx_v_r, -5.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1044
 *         cdef double y287 = y231*y234
 *         cdef double y288 = r**(-5.5)
 *         cdef double y289 = nu*y100             # <<<<<<<<<<<<<<
 *         cdef double y290 = nu*y77
 *         cdef double y291 = 2.0*y32
 */
  __Pyx_TraceLine(1044,0,__PYX_ERR(1, 1044, __pyx_L1_error))
  __pyx_v_y289 = (__pyx_v_nu * __pyx_v_y100);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1045
 *         cdef double y288 = r**(-5.5)
 *         cdef double y289 = nu*y100
 *         cdef double y290 = nu*y77             # <<<<<<<<<<<<<<
 *         cdef double y291 = 2.0*y32
 *         cdef double y292 = 118.4*y39
 */
  __Pyx_TraceLine(1045,0,__PYX_ERR(1, 1045, __pyx_L1_error))
  __pyx_v_y290 = (__pyx_v_nu * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1046
 *         cdef double y289 = nu*y100
 *         cdef double y290 = nu*y77
 *         cdef double y291 = 2.0*y32             # <<<<<<<<<<<<<<
 *         cdef double y292 = 118.4*y39
 *         cdef double y293 = -51.6952380952381*nu*y91 + y292*y91
 */
  __Pyx_TraceLine(1046,0,__PYX_ERR(1, 1046, __pyx_L1_error))
  __pyx_v_y291 = (2.0 * __pyx_v_y32);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1047
 *         cdef double y290 = nu*y77
 *         cdef double y291 = 2.0*y32
 *         cdef double y292 = 118.4*y39             # <<<<<<<<<<<<<<
 *         cdef double y293 = -51.6952380952381*nu*y91 + y292*y91
 *         cdef double y294 = y100*y124
 */
  __Pyx_TraceLine(1047,0,__PYX_ERR(1, 1047, __pyx_L1_error))
  __pyx_v_y292 = (118.4 * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1048
 *         cdef double y291 = 2.0*y32
 *         cdef double y292 = 118.4*y39
 *         cdef double y293 = -51.6952380952381*nu*y91 + y292*y91             # <<<<<<<<<<<<<<
 *         cdef double y294 = y100*y124
 *         cdef double y295 = y126**(-2)
 */
  __Pyx_TraceLine(1048,0,__PYX_ERR(1, 1048, __pyx_L1_error))
  __pyx_v_y293 = (((-51.6952380952381 * __pyx_v_nu) * __pyx_v_y91) + (__pyx_v_y292 * __pyx_v_y91));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1049
 *         cdef double y292 = 118.4*y39
 *         cdef double y293 = -51.6952380952381*nu*y91 + y292*y91
 *         cdef double y294 = y100*y124             # <<<<<<<<<<<<<<
 *         cdef double y295 = y126**(-2)
 *         cdef double y296 = r*y20
 */
  __Pyx_TraceLine(1049,0,__PYX_ERR(1, 1049, __pyx_L1_error))
  __pyx_v_y294 = (__pyx_v_y100 * __pyx_v_y124);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1050
 *         cdef double y293 = -51.6952380952381*nu*y91 + y292*y91
 *         cdef double y294 = y100*y124
 *         cdef double y295 = y126**(-2)             # <<<<<<<<<<<<<<
 *         cdef double y296 = r*y20
 *         cdef double y297 = y21 + y280 + y296
 */
  __Pyx_TraceLine(1050,0,__PYX_ERR(1, 1050, __pyx_L1_error))
  __pyx_v_y295 = pow(__pyx_v_y126, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1051
 *         cdef double y294 = y100*y124
 *         cdef double y295 = y126**(-2)
 *         cdef double y296 = r*y20             # <<<<<<<<<<<<<<
 *         cdef double y297 = y21 + y280 + y296
 *         cdef double y298 = 0.5*y59
 */
  __Pyx_TraceLine(1051,0,__PYX_ERR(1, 1051, __pyx_L1_error))
  __pyx_v_y296 = (__pyx_v_r * __pyx_v_y20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1052
 *         cdef double y295 = y126**(-2)
 *         cdef double y296 = r*y20
 *         cdef double y297 = y21 + y280 + y296             # <<<<<<<<<<<<<<
 *         cdef double y298 = 0.5*y59
 *         cdef double y299 = 0.0104166666666667*y59
 */
  __Pyx_TraceLine(1052,0,__PYX_ERR(1, 1052, __pyx_L1_error))
  __pyx_v_y297 = ((__pyx_v_y21 + __pyx_v_y280) + __pyx_v_y296);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1053
 *         cdef double y296 = r*y20
 *         cdef double y297 = y21 + y280 + y296
 *         cdef double y298 = 0.5*y59             # <<<<<<<<<<<<<<
 *         cdef double y299 = 0.0104166666666667*y59
 *         cdef double y300 = 0.03125*y136
 */
  __Pyx_TraceLine(1053,0,__PYX_ERR(1, 1053, __pyx_L1_error))
  __pyx_v_y298 = (0.5 * __pyx_v_y59);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1054
 *         cdef double y297 = y21 + y280 + y296
 *         cdef double y298 = 0.5*y59
 *         cdef double y299 = 0.0104166666666667*y59             # <<<<<<<<<<<<<<
 *         cdef double y300 = 0.03125*y136
 *         cdef double y301 = -y25*(-0.25*y138 + y140*y299 + y142*y300) + y32*y59 - y36*(0.5*y136*(0.5625 - 2.25*nu) + y298*(-5.25*nu - 2.8125) + 0.5*y64*(2.25 - y259))
 */
  __Pyx_TraceLine(1054,0,__PYX_ERR(1, 1054, __pyx_L1_error))
  __pyx_v_y299 = (0.0104166666666667 * __pyx_v_y59);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1055
 *         cdef double y298 = 0.5*y59
 *         cdef double y299 = 0.0104166666666667*y59
 *         cdef double y300 = 0.03125*y136             # <<<<<<<<<<<<<<
 *         cdef double y301 = -y25*(-0.25*y138 + y140*y299 + y142*y300) + y32*y59 - y36*(0.5*y136*(0.5625 - 2.25*nu) + y298*(-5.25*nu - 2.8125) + 0.5*y64*(2.25 - y259))
 *         cdef double y302 = r**(-14)
 */
  __Pyx_TraceLine(1055,0,__PYX_ERR(1, 1055, __pyx_L1_error))
  __pyx_v_y300 = (0.03125 * __pyx_v_y136);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1056
 *         cdef double y299 = 0.0104166666666667*y59
 *         cdef double y300 = 0.03125*y136
 *         cdef double y301 = -y25*(-0.25*y138 + y140*y299 + y142*y300) + y32*y59 - y36*(0.5*y136*(0.5625 - 2.25*nu) + y298*(-5.25*nu - 2.8125) + 0.5*y64*(2.25 - y259))             # <<<<<<<<<<<<<<
 *         cdef double y302 = r**(-14)
 *         cdef double y303 = y200*y203
 */
  __Pyx_TraceLine(1056,0,__PYX_ERR(1, 1056, __pyx_L1_error))
  __pyx_v_y301 = ((((-__pyx_v_y25) * (((-0.25 * __pyx_v_y138) + (__pyx_v_y140 * __pyx_v_y299)) + (__pyx_v_y142 * __pyx_v_y300))) + (__pyx_v_y32 * __pyx_v_y59)) - (__pyx_v_y36 * ((((0.5 * __pyx_v_y136) * (0.5625 - (2.25 * __pyx_v_nu))) + (__pyx_v_y298 * ((-5.25 * __pyx_v_nu) - 2.8125))) + ((0.5 * __pyx_v_y64) * (2.25 - __pyx_v_y259)))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1057
 *         cdef double y300 = 0.03125*y136
 *         cdef double y301 = -y25*(-0.25*y138 + y140*y299 + y142*y300) + y32*y59 - y36*(0.5*y136*(0.5625 - 2.25*nu) + y298*(-5.25*nu - 2.8125) + 0.5*y64*(2.25 - y259))
 *         cdef double y302 = r**(-14)             # <<<<<<<<<<<<<<
 *         cdef double y303 = y200*y203
 *         cdef double y304 = r**(-12)
 */
  __Pyx_TraceLine(1057,0,__PYX_ERR(1, 1057, __pyx_L1_error))
  __pyx_v_y302 = pow(__pyx_v_r, -14.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1058
 *         cdef double y301 = -y25*(-0.25*y138 + y140*y299 + y142*y300) + y32*y59 - y36*(0.5*y136*(0.5625 - 2.25*nu) + y298*(-5.25*nu - 2.8125) + 0.5*y64*(2.25 - y259))
 *         cdef double y302 = r**(-14)
 *         cdef double y303 = y200*y203             # <<<<<<<<<<<<<<
 *         cdef double y304 = r**(-12)
 *         cdef double y305 = y202**3
 */
  __Pyx_TraceLine(1058,0,__PYX_ERR(1, 1058, __pyx_L1_error))
  __pyx_v_y303 = (__pyx_v_y200 * __pyx_v_y203);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1059
 *         cdef double y302 = r**(-14)
 *         cdef double y303 = y200*y203
 *         cdef double y304 = r**(-12)             # <<<<<<<<<<<<<<
 *         cdef double y305 = y202**3
 *         cdef double y306 = 5178202125747.62*nu
 */
  __Pyx_TraceLine(1059,0,__PYX_ERR(1, 1059, __pyx_L1_error))
  __pyx_v_y304 = pow(__pyx_v_r, -12.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1060
 *         cdef double y303 = y200*y203
 *         cdef double y304 = r**(-12)
 *         cdef double y305 = y202**3             # <<<<<<<<<<<<<<
 *         cdef double y306 = 5178202125747.62*nu
 *         cdef double y307 = 267544166400.0*y39
 */
  __Pyx_TraceLine(1060,0,__PYX_ERR(1, 1060, __pyx_L1_error))
  __pyx_v_y305 = pow(__pyx_v_y202, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1061
 *         cdef double y304 = r**(-12)
 *         cdef double y305 = y202**3
 *         cdef double y306 = 5178202125747.62*nu             # <<<<<<<<<<<<<<
 *         cdef double y307 = 267544166400.0*y39
 *         cdef double y308 = 11575876386816.0*y91
 */
  __Pyx_TraceLine(1061,0,__PYX_ERR(1, 1061, __pyx_L1_error))
  __pyx_v_y306 = (5178202125747.62 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1062
 *         cdef double y305 = y202**3
 *         cdef double y306 = 5178202125747.62*nu
 *         cdef double y307 = 267544166400.0*y39             # <<<<<<<<<<<<<<
 *         cdef double y308 = 11575876386816.0*y91
 *         cdef double y309 = y123*y308
 */
  __Pyx_TraceLine(1062,0,__PYX_ERR(1, 1062, __pyx_L1_error))
  __pyx_v_y307 = (267544166400.0 * __pyx_v_y39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1063
 *         cdef double y306 = 5178202125747.62*nu
 *         cdef double y307 = 267544166400.0*y39
 *         cdef double y308 = 11575876386816.0*y91             # <<<<<<<<<<<<<<
 *         cdef double y309 = y123*y308
 *         cdef double y310 = y123*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0)
 */
  __Pyx_TraceLine(1063,0,__PYX_ERR(1, 1063, __pyx_L1_error))
  __pyx_v_y308 = (11575876386816.0 * __pyx_v_y91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1064
 *         cdef double y307 = 267544166400.0*y39
 *         cdef double y308 = 11575876386816.0*y91
 *         cdef double y309 = y123*y308             # <<<<<<<<<<<<<<
 *         cdef double y310 = y123*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0)
 *         cdef double y311 = y193*y91
 */
  __Pyx_TraceLine(1064,0,__PYX_ERR(1, 1064, __pyx_L1_error))
  __pyx_v_y309 = (__pyx_v_y123 * __pyx_v_y308);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1065
 *         cdef double y308 = 11575876386816.0*y91
 *         cdef double y309 = y123*y308
 *         cdef double y310 = y123*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0)             # <<<<<<<<<<<<<<
 *         cdef double y311 = y193*y91
 *         cdef double y312 = 5807150888816.34*nu + r*y306 + r*y307 + 10215490662751.4*r + y207 + y309 + y310 - y311 - 53501685054374.1
 */
  __Pyx_TraceLine(1065,0,__PYX_ERR(1, 1065, __pyx_L1_error))
  __pyx_v_y310 = (__pyx_v_y123 * (((4161798144000.0 * __pyx_v_nu) + (1759846072320.0 * __pyx_v_r)) + 12148770078720.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1066
 *         cdef double y309 = y123*y308
 *         cdef double y310 = y123*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0)
 *         cdef double y311 = y193*y91             # <<<<<<<<<<<<<<
 *         cdef double y312 = 5807150888816.34*nu + r*y306 + r*y307 + 10215490662751.4*r + y207 + y309 + y310 - y311 - 53501685054374.1
 *         cdef double y313 = y195**(-3)
 */
  __Pyx_TraceLine(1066,0,__PYX_ERR(1, 1066, __pyx_L1_error))
  __pyx_v_y311 = (__pyx_v_y193 * __pyx_v_y91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1067
 *         cdef double y310 = y123*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0)
 *         cdef double y311 = y193*y91
 *         cdef double y312 = 5807150888816.34*nu + r*y306 + r*y307 + 10215490662751.4*r + y207 + y309 + y310 - y311 - 53501685054374.1             # <<<<<<<<<<<<<<
 *         cdef double y313 = y195**(-3)
 *         cdef double y314 = y312*y313
 */
  __Pyx_TraceLine(1067,0,__PYX_ERR(1, 1067, __pyx_L1_error))
  __pyx_v_y312 = (((((((((5807150888816.34 * __pyx_v_nu) + (__pyx_v_r * __pyx_v_y306)) + (__pyx_v_r * __pyx_v_y307)) + (10215490662751.4 * __pyx_v_r)) + __pyx_v_y207) + __pyx_v_y309) + __pyx_v_y310) - __pyx_v_y311) - 53501685054374.1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1068
 *         cdef double y311 = y193*y91
 *         cdef double y312 = 5807150888816.34*nu + r*y306 + r*y307 + 10215490662751.4*r + y207 + y309 + y310 - y311 - 53501685054374.1
 *         cdef double y313 = y195**(-3)             # <<<<<<<<<<<<<<
 *         cdef double y314 = y312*y313
 *         cdef double y315 = y100*y171*y198
 */
  __Pyx_TraceLine(1068,0,__PYX_ERR(1, 1068, __pyx_L1_error))
  __pyx_v_y313 = pow(__pyx_v_y195, -3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1069
 *         cdef double y312 = 5807150888816.34*nu + r*y306 + r*y307 + 10215490662751.4*r + y207 + y309 + y310 - y311 - 53501685054374.1
 *         cdef double y313 = y195**(-3)
 *         cdef double y314 = y312*y313             # <<<<<<<<<<<<<<
 *         cdef double y315 = y100*y171*y198
 *         cdef double y316 = y186*y315
 */
  __Pyx_TraceLine(1069,0,__PYX_ERR(1, 1069, __pyx_L1_error))
  __pyx_v_y314 = (__pyx_v_y312 * __pyx_v_y313);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1070
 *         cdef double y313 = y195**(-3)
 *         cdef double y314 = y312*y313
 *         cdef double y315 = y100*y171*y198             # <<<<<<<<<<<<<<
 *         cdef double y316 = y186*y315
 *         cdef double y317 = y204*y316
 */
  __Pyx_TraceLine(1070,0,__PYX_ERR(1, 1070, __pyx_L1_error))
  __pyx_v_y315 = ((__pyx_v_y100 * __pyx_v_y171) * __pyx_v_y198);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1071
 *         cdef double y314 = y312*y313
 *         cdef double y315 = y100*y171*y198
 *         cdef double y316 = y186*y315             # <<<<<<<<<<<<<<
 *         cdef double y317 = y204*y316
 *         cdef double y318 = 9.25454462627843e-34*y317
 */
  __Pyx_TraceLine(1071,0,__PYX_ERR(1, 1071, __pyx_L1_error))
  __pyx_v_y316 = (__pyx_v_y186 * __pyx_v_y315);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1072
 *         cdef double y315 = y100*y171*y198
 *         cdef double y316 = y186*y315
 *         cdef double y317 = y204*y316             # <<<<<<<<<<<<<<
 *         cdef double y318 = 9.25454462627843e-34*y317
 *         cdef double y319 = y183*y91
 */
  __Pyx_TraceLine(1072,0,__PYX_ERR(1, 1072, __pyx_L1_error))
  __pyx_v_y317 = (__pyx_v_y204 * __pyx_v_y316);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1073
 *         cdef double y316 = y186*y315
 *         cdef double y317 = y204*y316
 *         cdef double y318 = 9.25454462627843e-34*y317             # <<<<<<<<<<<<<<
 *         cdef double y319 = y183*y91
 *         cdef double y320 = y123*(20113376778784.3*nu + 566231040.0*y190 + y192 + 2639769108480.0*y71)
 */
  __Pyx_TraceLine(1073,0,__PYX_ERR(1, 1073, __pyx_L1_error))
  __pyx_v_y318 = (9.25454462627843e-34 * __pyx_v_y317);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1074
 *         cdef double y317 = y204*y316
 *         cdef double y318 = 9.25454462627843e-34*y317
 *         cdef double y319 = y183*y91             # <<<<<<<<<<<<<<
 *         cdef double y320 = y123*(20113376778784.3*nu + 566231040.0*y190 + y192 + 2639769108480.0*y71)
 *         cdef double y321 = r*y173
 */
  __Pyx_TraceLine(1074,0,__PYX_ERR(1, 1074, __pyx_L1_error))
  __pyx_v_y319 = (__pyx_v_y183 * __pyx_v_y91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1075
 *         cdef double y318 = 9.25454462627843e-34*y317
 *         cdef double y319 = y183*y91
 *         cdef double y320 = y123*(20113376778784.3*nu + 566231040.0*y190 + y192 + 2639769108480.0*y71)             # <<<<<<<<<<<<<<
 *         cdef double y321 = r*y173
 *         cdef double y322 = y175*y71
 */
  __Pyx_TraceLine(1075,0,__PYX_ERR(1, 1075, __pyx_L1_error))
  __pyx_v_y320 = (__pyx_v_y123 * ((((20113376778784.3 * __pyx_v_nu) + (566231040.0 * __pyx_v_y190)) + __pyx_v_y192) + (2639769108480.0 * __pyx_v_y71)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1076
 *         cdef double y319 = y183*y91
 *         cdef double y320 = y123*(20113376778784.3*nu + 566231040.0*y190 + y192 + 2639769108480.0*y71)
 *         cdef double y321 = r*y173             # <<<<<<<<<<<<<<
 *         cdef double y322 = y175*y71
 *         cdef double y323 = 11575876386816.0*y123 + y179 + y208 + y319 + y320 - 18432.0*y321 - 2903040.0*y322
 */
  __Pyx_TraceLine(1076,0,__PYX_ERR(1, 1076, __pyx_L1_error))
  __pyx_v_y321 = (__pyx_v_r * __pyx_v_y173);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1077
 *         cdef double y320 = y123*(20113376778784.3*nu + 566231040.0*y190 + y192 + 2639769108480.0*y71)
 *         cdef double y321 = r*y173
 *         cdef double y322 = y175*y71             # <<<<<<<<<<<<<<
 *         cdef double y323 = 11575876386816.0*y123 + y179 + y208 + y319 + y320 - 18432.0*y321 - 2903040.0*y322
 *         cdef double y324 = y227*y228
 */
  __Pyx_TraceLine(1077,0,__PYX_ERR(1, 1077, __pyx_L1_error))
  __pyx_v_y322 = (__pyx_v_y175 * __pyx_v_y71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1078
 *         cdef double y321 = r*y173
 *         cdef double y322 = y175*y71
 *         cdef double y323 = 11575876386816.0*y123 + y179 + y208 + y319 + y320 - 18432.0*y321 - 2903040.0*y322             # <<<<<<<<<<<<<<
 *         cdef double y324 = y227*y228
 *         cdef double y325 = y206*y226
 */
  __Pyx_TraceLine(1078,0,__PYX_ERR(1, 1078, __pyx_L1_error))
  __pyx_v_y323 = (((((((11575876386816.0 * __pyx_v_y123) + __pyx_v_y179) + __pyx_v_y208) + __pyx_v_y319) + __pyx_v_y320) - (18432.0 * __pyx_v_y321)) - (2903040.0 * __pyx_v_y322));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1079
 *         cdef double y322 = y175*y71
 *         cdef double y323 = 11575876386816.0*y123 + y179 + y208 + y319 + y320 - 18432.0*y321 - 2903040.0*y322
 *         cdef double y324 = y227*y228             # <<<<<<<<<<<<<<
 *         cdef double y325 = y206*y226
 *         cdef double y326 = y205*y325
 */
  __Pyx_TraceLine(1079,0,__PYX_ERR(1, 1079, __pyx_L1_error))
  __pyx_v_y324 = (__pyx_v_y227 * __pyx_v_y228);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1080
 *         cdef double y323 = 11575876386816.0*y123 + y179 + y208 + y319 + y320 - 18432.0*y321 - 2903040.0*y322
 *         cdef double y324 = y227*y228
 *         cdef double y325 = y206*y226             # <<<<<<<<<<<<<<
 *         cdef double y326 = y205*y325
 *         cdef double y327 = y196*y326
 */
  __Pyx_TraceLine(1080,0,__PYX_ERR(1, 1080, __pyx_L1_error))
  __pyx_v_y325 = (__pyx_v_y206 * __pyx_v_y226);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1081
 *         cdef double y324 = y227*y228
 *         cdef double y325 = y206*y226
 *         cdef double y326 = y205*y325             # <<<<<<<<<<<<<<
 *         cdef double y327 = y196*y326
 *         cdef double y328 = y312*y327
 */
  __Pyx_TraceLine(1081,0,__PYX_ERR(1, 1081, __pyx_L1_error))
  __pyx_v_y326 = (__pyx_v_y205 * __pyx_v_y325);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1082
 *         cdef double y325 = y206*y226
 *         cdef double y326 = y205*y325
 *         cdef double y327 = y196*y326             # <<<<<<<<<<<<<<
 *         cdef double y328 = y312*y327
 *         cdef double y329 = 2.24091649004576e-37*y229
 */
  __Pyx_TraceLine(1082,0,__PYX_ERR(1, 1082, __pyx_L1_error))
  __pyx_v_y327 = (__pyx_v_y196 * __pyx_v_y326);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1083
 *         cdef double y326 = y205*y325
 *         cdef double y327 = y196*y326
 *         cdef double y328 = y312*y327             # <<<<<<<<<<<<<<
 *         cdef double y329 = 2.24091649004576e-37*y229
 *         cdef double y330 = y170**(-5)
 */
  __Pyx_TraceLine(1083,0,__PYX_ERR(1, 1083, __pyx_L1_error))
  __pyx_v_y328 = (__pyx_v_y312 * __pyx_v_y327);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1084
 *         cdef double y327 = y196*y326
 *         cdef double y328 = y312*y327
 *         cdef double y329 = 2.24091649004576e-37*y229             # <<<<<<<<<<<<<<
 *         cdef double y330 = y170**(-5)
 *         cdef double y331 = y251*y32
 */
  __Pyx_TraceLine(1084,0,__PYX_ERR(1, 1084, __pyx_L1_error))
  __pyx_v_y329 = (2.24091649004576e-37 * __pyx_v_y229);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1085
 *         cdef double y328 = y312*y327
 *         cdef double y329 = 2.24091649004576e-37*y229
 *         cdef double y330 = y170**(-5)             # <<<<<<<<<<<<<<
 *         cdef double y331 = y251*y32
 *         cdef double y332 = y165*y35
 */
  __Pyx_TraceLine(1085,0,__PYX_ERR(1, 1085, __pyx_L1_error))
  __pyx_v_y330 = pow(__pyx_v_y170, -5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1086
 *         cdef double y329 = 2.24091649004576e-37*y229
 *         cdef double y330 = y170**(-5)
 *         cdef double y331 = y251*y32             # <<<<<<<<<<<<<<
 *         cdef double y332 = y165*y35
 *         cdef double y333 = y268*y332
 */
  __Pyx_TraceLine(1086,0,__PYX_ERR(1, 1086, __pyx_L1_error))
  __pyx_v_y331 = (__pyx_v_y251 * __pyx_v_y32);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1087
 *         cdef double y330 = y170**(-5)
 *         cdef double y331 = y251*y32
 *         cdef double y332 = y165*y35             # <<<<<<<<<<<<<<
 *         cdef double y333 = y268*y332
 *         cdef double y334 = y0*y168
 */
  __Pyx_TraceLine(1087,0,__PYX_ERR(1, 1087, __pyx_L1_error))
  __pyx_v_y332 = (__pyx_v_y165 * __pyx_v_y35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1088
 *         cdef double y331 = y251*y32
 *         cdef double y332 = y165*y35
 *         cdef double y333 = y268*y332             # <<<<<<<<<<<<<<
 *         cdef double y334 = y0*y168
 *         cdef double y335 = -y285 + y331 + 7680.0*y333 + 30720.0*y334
 */
  __Pyx_TraceLine(1088,0,__PYX_ERR(1, 1088, __pyx_L1_error))
  __pyx_v_y333 = (__pyx_v_y268 * __pyx_v_y332);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1089
 *         cdef double y332 = y165*y35
 *         cdef double y333 = y268*y332
 *         cdef double y334 = y0*y168             # <<<<<<<<<<<<<<
 *         cdef double y335 = -y285 + y331 + 7680.0*y333 + 30720.0*y334
 *         cdef double y336 = y330*y335
 */
  __Pyx_TraceLine(1089,0,__PYX_ERR(1, 1089, __pyx_L1_error))
  __pyx_v_y334 = (__pyx_v_y0 * __pyx_v_y168);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1090
 *         cdef double y333 = y268*y332
 *         cdef double y334 = y0*y168
 *         cdef double y335 = -y285 + y331 + 7680.0*y333 + 30720.0*y334             # <<<<<<<<<<<<<<
 *         cdef double y336 = y330*y335
 *         cdef double y337 = y196*y204
 */
  __Pyx_TraceLine(1090,0,__PYX_ERR(1, 1090, __pyx_L1_error))
  __pyx_v_y335 = ((((-__pyx_v_y285) + __pyx_v_y331) + (7680.0 * __pyx_v_y333)) + (30720.0 * __pyx_v_y334));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1091
 *         cdef double y334 = y0*y168
 *         cdef double y335 = -y285 + y331 + 7680.0*y333 + 30720.0*y334
 *         cdef double y336 = y330*y335             # <<<<<<<<<<<<<<
 *         cdef double y337 = y196*y204
 *         cdef double y338 = y198*y337
 */
  __Pyx_TraceLine(1091,0,__PYX_ERR(1, 1091, __pyx_L1_error))
  __pyx_v_y336 = (__pyx_v_y330 * __pyx_v_y335);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1092
 *         cdef double y335 = -y285 + y331 + 7680.0*y333 + 30720.0*y334
 *         cdef double y336 = y330*y335
 *         cdef double y337 = y196*y204             # <<<<<<<<<<<<<<
 *         cdef double y338 = y198*y337
 *         cdef double y339 = y100*y186
 */
  __Pyx_TraceLine(1092,0,__PYX_ERR(1, 1092, __pyx_L1_error))
  __pyx_v_y337 = (__pyx_v_y196 * __pyx_v_y204);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1093
 *         cdef double y336 = y330*y335
 *         cdef double y337 = y196*y204
 *         cdef double y338 = y198*y337             # <<<<<<<<<<<<<<
 *         cdef double y339 = y100*y186
 *         cdef double y340 = y338*y339
 */
  __Pyx_TraceLine(1093,0,__PYX_ERR(1, 1093, __pyx_L1_error))
  __pyx_v_y338 = (__pyx_v_y198 * __pyx_v_y337);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1094
 *         cdef double y337 = y196*y204
 *         cdef double y338 = y198*y337
 *         cdef double y339 = y100*y186             # <<<<<<<<<<<<<<
 *         cdef double y340 = y338*y339
 *         cdef double y341 = 6.62902677807736e-23*y340
 */
  __Pyx_TraceLine(1094,0,__PYX_ERR(1, 1094, __pyx_L1_error))
  __pyx_v_y339 = (__pyx_v_y100 * __pyx_v_y186);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1095
 *         cdef double y338 = y198*y337
 *         cdef double y339 = y100*y186
 *         cdef double y340 = y338*y339             # <<<<<<<<<<<<<<
 *         cdef double y341 = 6.62902677807736e-23*y340
 *         cdef double y342 = y205*y210*y226
 */
  __Pyx_TraceLine(1095,0,__PYX_ERR(1, 1095, __pyx_L1_error))
  __pyx_v_y340 = (__pyx_v_y338 * __pyx_v_y339);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1096
 *         cdef double y339 = y100*y186
 *         cdef double y340 = y338*y339
 *         cdef double y341 = 6.62902677807736e-23*y340             # <<<<<<<<<<<<<<
 *         cdef double y342 = y205*y210*y226
 *         cdef double y343 = y170**(-3)
 */
  __Pyx_TraceLine(1096,0,__PYX_ERR(1, 1096, __pyx_L1_error))
  __pyx_v_y341 = (6.62902677807736e-23 * __pyx_v_y340);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1097
 *         cdef double y340 = y338*y339
 *         cdef double y341 = 6.62902677807736e-23*y340
 *         cdef double y342 = y205*y210*y226             # <<<<<<<<<<<<<<
 *         cdef double y343 = y170**(-3)
 *         cdef double y344 = y335*y343
 */
  __Pyx_TraceLine(1097,0,__PYX_ERR(1, 1097, __pyx_L1_error))
  __pyx_v_y342 = ((__pyx_v_y205 * __pyx_v_y210) * __pyx_v_y226);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1098
 *         cdef double y341 = 6.62902677807736e-23*y340
 *         cdef double y342 = y205*y210*y226
 *         cdef double y343 = y170**(-3)             # <<<<<<<<<<<<<<
 *         cdef double y344 = y335*y343
 *         cdef double y345 = y342*y344
 */
  __Pyx_TraceLine(1098,0,__PYX_ERR(1, 1098, __pyx_L1_error))
  __pyx_v_y343 = pow(__pyx_v_y170, -3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1099
 *         cdef double y342 = y205*y210*y226
 *         cdef double y343 = y170**(-3)
 *         cdef double y344 = y335*y343             # <<<<<<<<<<<<<<
 *         cdef double y345 = y342*y344
 *         cdef double y346 = y26 - 3.0
 */
  __Pyx_TraceLine(1099,0,__PYX_ERR(1, 1099, __pyx_L1_error))
  __pyx_v_y344 = (__pyx_v_y335 * __pyx_v_y343);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1100
 *         cdef double y343 = y170**(-3)
 *         cdef double y344 = y335*y343
 *         cdef double y345 = y342*y344             # <<<<<<<<<<<<<<
 *         cdef double y346 = y26 - 3.0
 *         cdef double y347 = y222*y224
 */
  __Pyx_TraceLine(1100,0,__PYX_ERR(1, 1100, __pyx_L1_error))
  __pyx_v_y345 = (__pyx_v_y342 * __pyx_v_y344);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1101
 *         cdef double y344 = y335*y343
 *         cdef double y345 = y342*y344
 *         cdef double y346 = y26 - 3.0             # <<<<<<<<<<<<<<
 *         cdef double y347 = y222*y224
 *         cdef double y348 = y165*y268
 */
  __Pyx_TraceLine(1101,0,__PYX_ERR(1, 1101, __pyx_L1_error))
  __pyx_v_y346 = (__pyx_v_y26 - 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1102
 *         cdef double y345 = y342*y344
 *         cdef double y346 = y26 - 3.0
 *         cdef double y347 = y222*y224             # <<<<<<<<<<<<<<
 *         cdef double y348 = y165*y268
 *         cdef double y349 = y209*y221
 */
  __Pyx_TraceLine(1102,0,__PYX_ERR(1, 1102, __pyx_L1_error))
  __pyx_v_y347 = (__pyx_v_y222 * __pyx_v_y224);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1103
 *         cdef double y346 = y26 - 3.0
 *         cdef double y347 = y222*y224
 *         cdef double y348 = y165*y268             # <<<<<<<<<<<<<<
 *         cdef double y349 = y209*y221
 *         cdef double y350 = y186**(-1)
 */
  __Pyx_TraceLine(1103,0,__PYX_ERR(1, 1103, __pyx_L1_error))
  __pyx_v_y348 = (__pyx_v_y165 * __pyx_v_y268);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1104
 *         cdef double y347 = y222*y224
 *         cdef double y348 = y165*y268
 *         cdef double y349 = y209*y221             # <<<<<<<<<<<<<<
 *         cdef double y350 = y186**(-1)
 *         cdef double y351 = y323*y350
 */
  __Pyx_TraceLine(1104,0,__PYX_ERR(1, 1104, __pyx_L1_error))
  __pyx_v_y349 = (__pyx_v_y209 * __pyx_v_y221);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1105
 *         cdef double y348 = y165*y268
 *         cdef double y349 = y209*y221
 *         cdef double y350 = y186**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y351 = y323*y350
 *         cdef double y352 = y209*y24
 */
  __Pyx_TraceLine(1105,0,__PYX_ERR(1, 1105, __pyx_L1_error))
  __pyx_v_y350 = pow(__pyx_v_y186, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1106
 *         cdef double y349 = y209*y221
 *         cdef double y350 = y186**(-1)
 *         cdef double y351 = y323*y350             # <<<<<<<<<<<<<<
 *         cdef double y352 = y209*y24
 *         cdef double y353 = y168*y352
 */
  __Pyx_TraceLine(1106,0,__PYX_ERR(1, 1106, __pyx_L1_error))
  __pyx_v_y351 = (__pyx_v_y323 * __pyx_v_y350);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1107
 *         cdef double y350 = y186**(-1)
 *         cdef double y351 = y323*y350
 *         cdef double y352 = y209*y24             # <<<<<<<<<<<<<<
 *         cdef double y353 = y168*y352
 *         cdef double y354 = 2.29252167428035e-22*y353
 */
  __Pyx_TraceLine(1107,0,__PYX_ERR(1, 1107, __pyx_L1_error))
  __pyx_v_y352 = (__pyx_v_y209 * __pyx_v_y24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1108
 *         cdef double y351 = y323*y350
 *         cdef double y352 = y209*y24
 *         cdef double y353 = y168*y352             # <<<<<<<<<<<<<<
 *         cdef double y354 = 2.29252167428035e-22*y353
 *         cdef double y355 = y167*y283
 */
  __Pyx_TraceLine(1108,0,__PYX_ERR(1, 1108, __pyx_L1_error))
  __pyx_v_y353 = (__pyx_v_y168 * __pyx_v_y352);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1109
 *         cdef double y352 = y209*y24
 *         cdef double y353 = y168*y352
 *         cdef double y354 = 2.29252167428035e-22*y353             # <<<<<<<<<<<<<<
 *         cdef double y355 = y167*y283
 *         cdef double y356 = 38400.0*y169*y349 + 7680.0*y223*y348 - 1.31621673590926e-19*y223*y355 - y25*(0.125*y217 + 0.0625*y218 + 0.0625*y219) + y312*y347 + y331 - y351*y354 - y36*(y20*(9.0*nu + 8.4375) + y216*y346 - 7.875*y68)
 */
  __Pyx_TraceLine(1109,0,__PYX_ERR(1, 1109, __pyx_L1_error))
  __pyx_v_y354 = (2.29252167428035e-22 * __pyx_v_y353);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1110
 *         cdef double y353 = y168*y352
 *         cdef double y354 = 2.29252167428035e-22*y353
 *         cdef double y355 = y167*y283             # <<<<<<<<<<<<<<
 *         cdef double y356 = 38400.0*y169*y349 + 7680.0*y223*y348 - 1.31621673590926e-19*y223*y355 - y25*(0.125*y217 + 0.0625*y218 + 0.0625*y219) + y312*y347 + y331 - y351*y354 - y36*(y20*(9.0*nu + 8.4375) + y216*y346 - 7.875*y68)
 *         cdef double y357 = 39.6112800271521*nu*y101*y104 + y1*y127*y128*y34*y77 + y1*y128*y295*y297*y34*y91 + 1.01821851311268e-18*y100*y171*y186*y196*y198*y304*y305 + 7.59859378406358e-45*y100*y171*y196*y198*y201*y203*y220*y323 + y100*y293*y36 - 3.70688355060912*y103*y108 - y110*y290 - y112*y291 - 3.0*y114*y125 - y117*y291 - y120*y81 - y122*y291 - y131 - y144*y80 + 6.78168402777778e-8*y202*y205*y206*y210*y220*y225*y36 + 1.69542100694444e-8*y205*y206*y210*y220*y228*y25*y356 + 1.69542100694444e-8*y205*y206*y210*y225*y228*y25*y323 - 4.41515887225116e-12*y229*y345 - 663.496888455656*y288*y289 - y294*y83 + y301*y34*y77 - 1.65460508380811e-18*y302*y303 - y314*y318 - 8.47710503472222e-8*y324*y43 - y328*y329 - y336*y341
 */
  __Pyx_TraceLine(1110,0,__PYX_ERR(1, 1110, __pyx_L1_error))
  __pyx_v_y355 = (__pyx_v_y167 * __pyx_v_y283);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1111
 *         cdef double y354 = 2.29252167428035e-22*y353
 *         cdef double y355 = y167*y283
 *         cdef double y356 = 38400.0*y169*y349 + 7680.0*y223*y348 - 1.31621673590926e-19*y223*y355 - y25*(0.125*y217 + 0.0625*y218 + 0.0625*y219) + y312*y347 + y331 - y351*y354 - y36*(y20*(9.0*nu + 8.4375) + y216*y346 - 7.875*y68)             # <<<<<<<<<<<<<<
 *         cdef double y357 = 39.6112800271521*nu*y101*y104 + y1*y127*y128*y34*y77 + y1*y128*y295*y297*y34*y91 + 1.01821851311268e-18*y100*y171*y186*y196*y198*y304*y305 + 7.59859378406358e-45*y100*y171*y196*y198*y201*y203*y220*y323 + y100*y293*y36 - 3.70688355060912*y103*y108 - y110*y290 - y112*y291 - 3.0*y114*y125 - y117*y291 - y120*y81 - y122*y291 - y131 - y144*y80 + 6.78168402777778e-8*y202*y205*y206*y210*y220*y225*y36 + 1.69542100694444e-8*y205*y206*y210*y220*y228*y25*y356 + 1.69542100694444e-8*y205*y206*y210*y225*y228*y25*y323 - 4.41515887225116e-12*y229*y345 - 663.496888455656*y288*y289 - y294*y83 + y301*y34*y77 - 1.65460508380811e-18*y302*y303 - y314*y318 - 8.47710503472222e-8*y324*y43 - y328*y329 - y336*y341
 *         cdef double y358 = y248*y357 - y254*y257 + y286*y287
 */
  __Pyx_TraceLine(1111,0,__PYX_ERR(1, 1111, __pyx_L1_error))
  __pyx_v_y356 = (((((((((38400.0 * __pyx_v_y169) * __pyx_v_y349) + ((7680.0 * __pyx_v_y223) * __pyx_v_y348)) - ((1.31621673590926e-19 * __pyx_v_y223) * __pyx_v_y355)) - (__pyx_v_y25 * (((0.125 * __pyx_v_y217) + (0.0625 * __pyx_v_y218)) + (0.0625 * __pyx_v_y219)))) + (__pyx_v_y312 * __pyx_v_y347)) + __pyx_v_y331) - (__pyx_v_y351 * __pyx_v_y354)) - (__pyx_v_y36 * (((__pyx_v_y20 * ((9.0 * __pyx_v_nu) + 8.4375)) + (__pyx_v_y216 * __pyx_v_y346)) - (7.875 * __pyx_v_y68))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1112
 *         cdef double y355 = y167*y283
 *         cdef double y356 = 38400.0*y169*y349 + 7680.0*y223*y348 - 1.31621673590926e-19*y223*y355 - y25*(0.125*y217 + 0.0625*y218 + 0.0625*y219) + y312*y347 + y331 - y351*y354 - y36*(y20*(9.0*nu + 8.4375) + y216*y346 - 7.875*y68)
 *         cdef double y357 = 39.6112800271521*nu*y101*y104 + y1*y127*y128*y34*y77 + y1*y128*y295*y297*y34*y91 + 1.01821851311268e-18*y100*y171*y186*y196*y198*y304*y305 + 7.59859378406358e-45*y100*y171*y196*y198*y201*y203*y220*y323 + y100*y293*y36 - 3.70688355060912*y103*y108 - y110*y290 - y112*y291 - 3.0*y114*y125 - y117*y291 - y120*y81 - y122*y291 - y131 - y144*y80 + 6.78168402777778e-8*y202*y205*y206*y210*y220*y225*y36 + 1.69542100694444e-8*y205*y206*y210*y220*y228*y25*y356 + 1.69542100694444e-8*y205*y206*y210*y225*y228*y25*y323 - 4.41515887225116e-12*y229*y345 - 663.496888455656*y288*y289 - y294*y83 + y301*y34*y77 - 1.65460508380811e-18*y302*y303 - y314*y318 - 8.47710503472222e-8*y324*y43 - y328*y329 - y336*y341             # <<<<<<<<<<<<<<
 *         cdef double y358 = y248*y357 - y254*y257 + y286*y287
 *         cdef double y359 = 12.0*y13 + 12.0*y18 + 12.0*y7
 */
  __Pyx_TraceLine(1112,0,__PYX_ERR(1, 1112, __pyx_L1_error))
  __pyx_v_y357 = (((((((((((((((((((((((((((((39.6112800271521 * __pyx_v_nu) * __pyx_v_y101) * __pyx_v_y104) + ((((__pyx_v_y1 * __pyx_v_y127) * __pyx_v_y128) * __pyx_v_y34) * __pyx_v_y77)) + (((((__pyx_v_y1 * __pyx_v_y128) * __pyx_v_y295) * __pyx_v_y297) * __pyx_v_y34) * __pyx_v_y91)) + (((((((1.01821851311268e-18 * __pyx_v_y100) * __pyx_v_y171) * __pyx_v_y186) * __pyx_v_y196) * __pyx_v_y198) * __pyx_v_y304) * __pyx_v_y305)) + ((((((((7.59859378406358e-45 * __pyx_v_y100) * __pyx_v_y171) * __pyx_v_y196) * __pyx_v_y198) * __pyx_v_y201) * __pyx_v_y203) * __pyx_v_y220) * __pyx_v_y323)) + ((__pyx_v_y100 * __pyx_v_y293) * __pyx_v_y36)) - ((3.70688355060912 * __pyx_v_y103) * __pyx_v_y108)) - (__pyx_v_y110 * __pyx_v_y290)) - (__pyx_v_y112 * __pyx_v_y291)) - ((3.0 * __pyx_v_y114) * __pyx_v_y125)) - (__pyx_v_y117 * __pyx_v_y291)) - (__pyx_v_y120 * __pyx_v_y81)) - (__pyx_v_y122 * __pyx_v_y291)) - __pyx_v_y131) - (__pyx_v_y144 * __pyx_v_y80)) + (((((((6.78168402777778e-8 * __pyx_v_y202) * __pyx_v_y205) * __pyx_v_y206) * __pyx_v_y210) * __pyx_v_y220) * __pyx_v_y225) * __pyx_v_y36)) + (((((((1.69542100694444e-8 * __pyx_v_y205) * __pyx_v_y206) * __pyx_v_y210) * __pyx_v_y220) * __pyx_v_y228) * __pyx_v_y25) * __pyx_v_y356)) + (((((((1.69542100694444e-8 * __pyx_v_y205) * __pyx_v_y206) * __pyx_v_y210) * __pyx_v_y225) * __pyx_v_y228) * __pyx_v_y25) * __pyx_v_y323)) - ((4.41515887225116e-12 * __pyx_v_y229) * __pyx_v_y345)) - ((663.496888455656 * __pyx_v_y288) * __pyx_v_y289)) - (__pyx_v_y294 * __pyx_v_y83)) + ((__pyx_v_y301 * __pyx_v_y34) * __pyx_v_y77)) - ((1.65460508380811e-18 * __pyx_v_y302) * __pyx_v_y303)) - (__pyx_v_y314 * __pyx_v_y318)) - ((8.47710503472222e-8 * __pyx_v_y324) * __pyx_v_y43)) - (__pyx_v_y328 * __pyx_v_y329)) - (__pyx_v_y336 * __pyx_v_y341));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1113
 *         cdef double y356 = 38400.0*y169*y349 + 7680.0*y223*y348 - 1.31621673590926e-19*y223*y355 - y25*(0.125*y217 + 0.0625*y218 + 0.0625*y219) + y312*y347 + y331 - y351*y354 - y36*(y20*(9.0*nu + 8.4375) + y216*y346 - 7.875*y68)
 *         cdef double y357 = 39.6112800271521*nu*y101*y104 + y1*y127*y128*y34*y77 + y1*y128*y295*y297*y34*y91 + 1.01821851311268e-18*y100*y171*y186*y196*y198*y304*y305 + 7.59859378406358e-45*y100*y171*y196*y198*y201*y203*y220*y323 + y100*y293*y36 - 3.70688355060912*y103*y108 - y110*y290 - y112*y291 - 3.0*y114*y125 - y117*y291 - y120*y81 - y122*y291 - y131 - y144*y80 + 6.78168402777778e-8*y202*y205*y206*y210*y220*y225*y36 + 1.69542100694444e-8*y205*y206*y210*y220*y228*y25*y356 + 1.69542100694444e-8*y205*y206*y210*y225*y228*y25*y323 - 4.41515887225116e-12*y229*y345 - 663.496888455656*y288*y289 - y294*y83 + y301*y34*y77 - 1.65460508380811e-18*y302*y303 - y314*y318 - 8.47710503472222e-8*y324*y43 - y328*y329 - y336*y341
 *         cdef double y358 = y248*y357 - y254*y257 + y286*y287             # <<<<<<<<<<<<<<
 *         cdef double y359 = 12.0*y13 + 12.0*y18 + 12.0*y7
 *         cdef double y360 = -4.0*y13 - 4.0*y18 - 4.0*y7
 */
  __Pyx_TraceLine(1113,0,__PYX_ERR(1, 1113, __pyx_L1_error))
  __pyx_v_y358 = (((__pyx_v_y248 * __pyx_v_y357) - (__pyx_v_y254 * __pyx_v_y257)) + (__pyx_v_y286 * __pyx_v_y287));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1114
 *         cdef double y357 = 39.6112800271521*nu*y101*y104 + y1*y127*y128*y34*y77 + y1*y128*y295*y297*y34*y91 + 1.01821851311268e-18*y100*y171*y186*y196*y198*y304*y305 + 7.59859378406358e-45*y100*y171*y196*y198*y201*y203*y220*y323 + y100*y293*y36 - 3.70688355060912*y103*y108 - y110*y290 - y112*y291 - 3.0*y114*y125 - y117*y291 - y120*y81 - y122*y291 - y131 - y144*y80 + 6.78168402777778e-8*y202*y205*y206*y210*y220*y225*y36 + 1.69542100694444e-8*y205*y206*y210*y220*y228*y25*y356 + 1.69542100694444e-8*y205*y206*y210*y225*y228*y25*y323 - 4.41515887225116e-12*y229*y345 - 663.496888455656*y288*y289 - y294*y83 + y301*y34*y77 - 1.65460508380811e-18*y302*y303 - y314*y318 - 8.47710503472222e-8*y324*y43 - y328*y329 - y336*y341
 *         cdef double y358 = y248*y357 - y254*y257 + y286*y287
 *         cdef double y359 = 12.0*y13 + 12.0*y18 + 12.0*y7             # <<<<<<<<<<<<<<
 *         cdef double y360 = -4.0*y13 - 4.0*y18 - 4.0*y7
 *         cdef double y361 = 8.0*y91 + 4.0
 */
  __Pyx_TraceLine(1114,0,__PYX_ERR(1, 1114, __pyx_L1_error))
  __pyx_v_y359 = (((12.0 * __pyx_v_y13) + (12.0 * __pyx_v_y18)) + (12.0 * __pyx_v_y7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1115
 *         cdef double y358 = y248*y357 - y254*y257 + y286*y287
 *         cdef double y359 = 12.0*y13 + 12.0*y18 + 12.0*y7
 *         cdef double y360 = -4.0*y13 - 4.0*y18 - 4.0*y7             # <<<<<<<<<<<<<<
 *         cdef double y361 = 8.0*y91 + 4.0
 *         cdef double y362 = -y253*y361 + y36*y360
 */
  __Pyx_TraceLine(1115,0,__PYX_ERR(1, 1115, __pyx_L1_error))
  __pyx_v_y360 = (((-4.0 * __pyx_v_y13) - (4.0 * __pyx_v_y18)) - (4.0 * __pyx_v_y7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1116
 *         cdef double y359 = 12.0*y13 + 12.0*y18 + 12.0*y7
 *         cdef double y360 = -4.0*y13 - 4.0*y18 - 4.0*y7
 *         cdef double y361 = 8.0*y91 + 4.0             # <<<<<<<<<<<<<<
 *         cdef double y362 = -y253*y361 + y36*y360
 *         cdef double y363 = r**(-7)
 */
  __Pyx_TraceLine(1116,0,__PYX_ERR(1, 1116, __pyx_L1_error))
  __pyx_v_y361 = ((8.0 * __pyx_v_y91) + 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1117
 *         cdef double y360 = -4.0*y13 - 4.0*y18 - 4.0*y7
 *         cdef double y361 = 8.0*y91 + 4.0
 *         cdef double y362 = -y253*y361 + y36*y360             # <<<<<<<<<<<<<<
 *         cdef double y363 = r**(-7)
 *         cdef double y364 = y265*y59
 */
  __Pyx_TraceLine(1117,0,__PYX_ERR(1, 1117, __pyx_L1_error))
  __pyx_v_y362 = (((-__pyx_v_y253) * __pyx_v_y361) + (__pyx_v_y36 * __pyx_v_y360));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1118
 *         cdef double y361 = 8.0*y91 + 4.0
 *         cdef double y362 = -y253*y361 + y36*y360
 *         cdef double y363 = r**(-7)             # <<<<<<<<<<<<<<
 *         cdef double y364 = y265*y59
 *         cdef double y365 = 0.015625*y136
 */
  __Pyx_TraceLine(1118,0,__PYX_ERR(1, 1118, __pyx_L1_error))
  __pyx_v_y363 = pow(__pyx_v_r, -7.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1119
 *         cdef double y362 = -y253*y361 + y36*y360
 *         cdef double y363 = r**(-7)
 *         cdef double y364 = y265*y59             # <<<<<<<<<<<<<<
 *         cdef double y365 = 0.015625*y136
 *         cdef double y366 = 0.03125*y239
 */
  __Pyx_TraceLine(1119,0,__PYX_ERR(1, 1119, __pyx_L1_error))
  __pyx_v_y364 = (__pyx_v_y265 * __pyx_v_y59);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1120
 *         cdef double y363 = r**(-7)
 *         cdef double y364 = y265*y59
 *         cdef double y365 = 0.015625*y136             # <<<<<<<<<<<<<<
 *         cdef double y366 = 0.03125*y239
 *         cdef double y367 = 20.0*nu
 */
  __Pyx_TraceLine(1120,0,__PYX_ERR(1, 1120, __pyx_L1_error))
  __pyx_v_y365 = (0.015625 * __pyx_v_y136);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1121
 *         cdef double y364 = y265*y59
 *         cdef double y365 = 0.015625*y136
 *         cdef double y366 = 0.03125*y239             # <<<<<<<<<<<<<<
 *         cdef double y367 = 20.0*nu
 *         cdef double y368 = 6.0*y13 + 6.0*y18 + 6.0*y7
 */
  __Pyx_TraceLine(1121,0,__PYX_ERR(1, 1121, __pyx_L1_error))
  __pyx_v_y366 = (0.03125 * __pyx_v_y239);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1122
 *         cdef double y365 = 0.015625*y136
 *         cdef double y366 = 0.03125*y239
 *         cdef double y367 = 20.0*nu             # <<<<<<<<<<<<<<
 *         cdef double y368 = 6.0*y13 + 6.0*y18 + 6.0*y7
 *         cdef double y369 = y36*y368
 */
  __Pyx_TraceLine(1122,0,__PYX_ERR(1, 1122, __pyx_L1_error))
  __pyx_v_y367 = (20.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1123
 *         cdef double y366 = 0.03125*y239
 *         cdef double y367 = 20.0*nu
 *         cdef double y368 = 6.0*y13 + 6.0*y18 + 6.0*y7             # <<<<<<<<<<<<<<
 *         cdef double y369 = y36*y368
 *         cdef double y370 = 10569646080.0*y109 - 7680.0*y267*y290
 */
  __Pyx_TraceLine(1123,0,__PYX_ERR(1, 1123, __pyx_L1_error))
  __pyx_v_y368 = (((6.0 * __pyx_v_y13) + (6.0 * __pyx_v_y18)) + (6.0 * __pyx_v_y7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1124
 *         cdef double y367 = 20.0*nu
 *         cdef double y368 = 6.0*y13 + 6.0*y18 + 6.0*y7
 *         cdef double y369 = y36*y368             # <<<<<<<<<<<<<<
 *         cdef double y370 = 10569646080.0*y109 - 7680.0*y267*y290
 *         cdef double y371 = 2.63243347181853e-19*y283
 */
  __Pyx_TraceLine(1124,0,__PYX_ERR(1, 1124, __pyx_L1_error))
  __pyx_v_y369 = (__pyx_v_y36 * __pyx_v_y368);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1125
 *         cdef double y368 = 6.0*y13 + 6.0*y18 + 6.0*y7
 *         cdef double y369 = y36*y368
 *         cdef double y370 = 10569646080.0*y109 - 7680.0*y267*y290             # <<<<<<<<<<<<<<
 *         cdef double y371 = 2.63243347181853e-19*y283
 *         cdef double y372 = y268*y371
 */
  __Pyx_TraceLine(1125,0,__PYX_ERR(1, 1125, __pyx_L1_error))
  __pyx_v_y370 = ((10569646080.0 * __pyx_v_y109) - ((7680.0 * __pyx_v_y267) * __pyx_v_y290));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1126
 *         cdef double y369 = y36*y368
 *         cdef double y370 = 10569646080.0*y109 - 7680.0*y267*y290
 *         cdef double y371 = 2.63243347181853e-19*y283             # <<<<<<<<<<<<<<
 *         cdef double y372 = y268*y371
 *         cdef double y373 = 1.99471718230171e-8*(5.17401430341932e-11*y272 + y274 + 5.17401430341932e-11*y275 + 2.42203000992063e-7*y278 + 2.42203000992063e-7*y279 + 3.97364298502604e-7*y282 + 0.48828125*y35)**2/y270**3
 */
  __Pyx_TraceLine(1126,0,__PYX_ERR(1, 1126, __pyx_L1_error))
  __pyx_v_y371 = (2.63243347181853e-19 * __pyx_v_y283);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1127
 *         cdef double y370 = 10569646080.0*y109 - 7680.0*y267*y290
 *         cdef double y371 = 2.63243347181853e-19*y283
 *         cdef double y372 = y268*y371             # <<<<<<<<<<<<<<
 *         cdef double y373 = 1.99471718230171e-8*(5.17401430341932e-11*y272 + y274 + 5.17401430341932e-11*y275 + 2.42203000992063e-7*y278 + 2.42203000992063e-7*y279 + 3.97364298502604e-7*y282 + 0.48828125*y35)**2/y270**3
 *         cdef double y374 = 135291469824.0*y39*y77
 */
  __Pyx_TraceLine(1127,0,__PYX_ERR(1, 1127, __pyx_L1_error))
  __pyx_v_y372 = (__pyx_v_y268 * __pyx_v_y371);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1128
 *         cdef double y371 = 2.63243347181853e-19*y283
 *         cdef double y372 = y268*y371
 *         cdef double y373 = 1.99471718230171e-8*(5.17401430341932e-11*y272 + y274 + 5.17401430341932e-11*y275 + 2.42203000992063e-7*y278 + 2.42203000992063e-7*y279 + 3.97364298502604e-7*y282 + 0.48828125*y35)**2/y270**3             # <<<<<<<<<<<<<<
 *         cdef double y374 = 135291469824.0*y39*y77
 *         cdef double y375 = 12.0*y71
 */
  __Pyx_TraceLine(1128,0,__PYX_ERR(1, 1128, __pyx_L1_error))
  __pyx_v_y373 = ((1.99471718230171e-8 * pow((((((((5.17401430341932e-11 * __pyx_v_y272) + __pyx_v_y274) + (5.17401430341932e-11 * __pyx_v_y275)) + (2.42203000992063e-7 * __pyx_v_y278)) + (2.42203000992063e-7 * __pyx_v_y279)) + (3.97364298502604e-7 * __pyx_v_y282)) + (0.48828125 * __pyx_v_y35)), 2.0)) / pow(__pyx_v_y270, 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1129
 *         cdef double y372 = y268*y371
 *         cdef double y373 = 1.99471718230171e-8*(5.17401430341932e-11*y272 + y274 + 5.17401430341932e-11*y275 + 2.42203000992063e-7*y278 + 2.42203000992063e-7*y279 + 3.97364298502604e-7*y282 + 0.48828125*y35)**2/y270**3
 *         cdef double y374 = 135291469824.0*y39*y77             # <<<<<<<<<<<<<<
 *         cdef double y375 = 12.0*y71
 *         cdef double y376 = 12.0*r + 8.0
 */
  __Pyx_TraceLine(1129,0,__PYX_ERR(1, 1129, __pyx_L1_error))
  __pyx_v_y374 = ((135291469824.0 * __pyx_v_y39) * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1130
 *         cdef double y373 = 1.99471718230171e-8*(5.17401430341932e-11*y272 + y274 + 5.17401430341932e-11*y275 + 2.42203000992063e-7*y278 + 2.42203000992063e-7*y279 + 3.97364298502604e-7*y282 + 0.48828125*y35)**2/y270**3
 *         cdef double y374 = 135291469824.0*y39*y77
 *         cdef double y375 = 12.0*y71             # <<<<<<<<<<<<<<
 *         cdef double y376 = 12.0*r + 8.0
 *         cdef double y377 = 1.31621673590926e-19*y271*(53760.0*nu*(8463875.43631609*r - 7511347.20125054*y0 + y159*(y375 + y376) + y160*(25588.1776641618*r - 82228.1776641618*y71 + 17058.7851094412) + y162*(11556.0*r + 4188.0*y71 + 7704.0) + y163*(-y375 + y376) + 6347906.57723707*y71 + 7480835.43631609) + 264241152000.0*y0 + 65536.0*y109*y277 - 39946495452.7631*y113 - y123*y374 - 32768.0*y154*y290 + 32768.0*y155*(2257920.0*nu + 5760.0*r*y152 + 3225600.0*y0 + 5760.0*y276 + 4143360.0) + y374 + 7.0*y39*(-37071008445.2255*r + 28533211037.6879*y0 - 37071008445.2255*y71 + 4129567622.65173))
 */
  __Pyx_TraceLine(1130,0,__PYX_ERR(1, 1130, __pyx_L1_error))
  __pyx_v_y375 = (12.0 * __pyx_v_y71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1131
 *         cdef double y374 = 135291469824.0*y39*y77
 *         cdef double y375 = 12.0*y71
 *         cdef double y376 = 12.0*r + 8.0             # <<<<<<<<<<<<<<
 *         cdef double y377 = 1.31621673590926e-19*y271*(53760.0*nu*(8463875.43631609*r - 7511347.20125054*y0 + y159*(y375 + y376) + y160*(25588.1776641618*r - 82228.1776641618*y71 + 17058.7851094412) + y162*(11556.0*r + 4188.0*y71 + 7704.0) + y163*(-y375 + y376) + 6347906.57723707*y71 + 7480835.43631609) + 264241152000.0*y0 + 65536.0*y109*y277 - 39946495452.7631*y113 - y123*y374 - 32768.0*y154*y290 + 32768.0*y155*(2257920.0*nu + 5760.0*r*y152 + 3225600.0*y0 + 5760.0*y276 + 4143360.0) + y374 + 7.0*y39*(-37071008445.2255*r + 28533211037.6879*y0 - 37071008445.2255*y71 + 4129567622.65173))
 *         cdef double y378 = 61440.0*y0*y348 - 1.05297338872741e-18*y0*y355 + 92160.0*y168*y71 + y269*y373 - y269*y377 + y332*y370 - y35*y372 + y369
 */
  __Pyx_TraceLine(1131,0,__PYX_ERR(1, 1131, __pyx_L1_error))
  __pyx_v_y376 = ((12.0 * __pyx_v_r) + 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1132
 *         cdef double y375 = 12.0*y71
 *         cdef double y376 = 12.0*r + 8.0
 *         cdef double y377 = 1.31621673590926e-19*y271*(53760.0*nu*(8463875.43631609*r - 7511347.20125054*y0 + y159*(y375 + y376) + y160*(25588.1776641618*r - 82228.1776641618*y71 + 17058.7851094412) + y162*(11556.0*r + 4188.0*y71 + 7704.0) + y163*(-y375 + y376) + 6347906.57723707*y71 + 7480835.43631609) + 264241152000.0*y0 + 65536.0*y109*y277 - 39946495452.7631*y113 - y123*y374 - 32768.0*y154*y290 + 32768.0*y155*(2257920.0*nu + 5760.0*r*y152 + 3225600.0*y0 + 5760.0*y276 + 4143360.0) + y374 + 7.0*y39*(-37071008445.2255*r + 28533211037.6879*y0 - 37071008445.2255*y71 + 4129567622.65173))             # <<<<<<<<<<<<<<
 *         cdef double y378 = 61440.0*y0*y348 - 1.05297338872741e-18*y0*y355 + 92160.0*y168*y71 + y269*y373 - y269*y377 + y332*y370 - y35*y372 + y369
 *         cdef double y379 = 9.0*X_2
 */
  __Pyx_TraceLine(1132,0,__PYX_ERR(1, 1132, __pyx_L1_error))
  __pyx_v_y377 = ((1.31621673590926e-19 * __pyx_v_y271) * ((((((((((53760.0 * __pyx_v_nu) * ((((((((8463875.43631609 * __pyx_v_r) - (7511347.20125054 * __pyx_v_y0)) + (__pyx_v_y159 * (__pyx_v_y375 + __pyx_v_y376))) + (__pyx_v_y160 * (((25588.1776641618 * __pyx_v_r) - (82228.1776641618 * __pyx_v_y71)) + 17058.7851094412))) + (__pyx_v_y162 * (((11556.0 * __pyx_v_r) + (4188.0 * __pyx_v_y71)) + 7704.0))) + (__pyx_v_y163 * ((-__pyx_v_y375) + __pyx_v_y376))) + (6347906.57723707 * __pyx_v_y71)) + 7480835.43631609)) + (264241152000.0 * __pyx_v_y0)) + ((65536.0 * __pyx_v_y109) * __pyx_v_y277)) - (39946495452.7631 * __pyx_v_y113)) - (__pyx_v_y123 * __pyx_v_y374)) - ((32768.0 * __pyx_v_y154) * __pyx_v_y290)) + ((32768.0 * __pyx_v_y155) * (((((2257920.0 * __pyx_v_nu) + ((5760.0 * __pyx_v_r) * __pyx_v_y152)) + (3225600.0 * __pyx_v_y0)) + (5760.0 * __pyx_v_y276)) + 4143360.0))) + __pyx_v_y374) + ((7.0 * __pyx_v_y39) * ((((-37071008445.2255 * __pyx_v_r) + (28533211037.6879 * __pyx_v_y0)) - (37071008445.2255 * __pyx_v_y71)) + 4129567622.65173))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1133
 *         cdef double y376 = 12.0*r + 8.0
 *         cdef double y377 = 1.31621673590926e-19*y271*(53760.0*nu*(8463875.43631609*r - 7511347.20125054*y0 + y159*(y375 + y376) + y160*(25588.1776641618*r - 82228.1776641618*y71 + 17058.7851094412) + y162*(11556.0*r + 4188.0*y71 + 7704.0) + y163*(-y375 + y376) + 6347906.57723707*y71 + 7480835.43631609) + 264241152000.0*y0 + 65536.0*y109*y277 - 39946495452.7631*y113 - y123*y374 - 32768.0*y154*y290 + 32768.0*y155*(2257920.0*nu + 5760.0*r*y152 + 3225600.0*y0 + 5760.0*y276 + 4143360.0) + y374 + 7.0*y39*(-37071008445.2255*r + 28533211037.6879*y0 - 37071008445.2255*y71 + 4129567622.65173))
 *         cdef double y378 = 61440.0*y0*y348 - 1.05297338872741e-18*y0*y355 + 92160.0*y168*y71 + y269*y373 - y269*y377 + y332*y370 - y35*y372 + y369             # <<<<<<<<<<<<<<
 *         cdef double y379 = 9.0*X_2
 *         cdef double y380 = y1*y129
 */
  __Pyx_TraceLine(1133,0,__PYX_ERR(1, 1133, __pyx_L1_error))
  __pyx_v_y378 = (((((((((61440.0 * __pyx_v_y0) * __pyx_v_y348) - ((1.05297338872741e-18 * __pyx_v_y0) * __pyx_v_y355)) + ((92160.0 * __pyx_v_y168) * __pyx_v_y71)) + (__pyx_v_y269 * __pyx_v_y373)) - (__pyx_v_y269 * __pyx_v_y377)) + (__pyx_v_y332 * __pyx_v_y370)) - (__pyx_v_y35 * __pyx_v_y372)) + __pyx_v_y369);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1134
 *         cdef double y377 = 1.31621673590926e-19*y271*(53760.0*nu*(8463875.43631609*r - 7511347.20125054*y0 + y159*(y375 + y376) + y160*(25588.1776641618*r - 82228.1776641618*y71 + 17058.7851094412) + y162*(11556.0*r + 4188.0*y71 + 7704.0) + y163*(-y375 + y376) + 6347906.57723707*y71 + 7480835.43631609) + 264241152000.0*y0 + 65536.0*y109*y277 - 39946495452.7631*y113 - y123*y374 - 32768.0*y154*y290 + 32768.0*y155*(2257920.0*nu + 5760.0*r*y152 + 3225600.0*y0 + 5760.0*y276 + 4143360.0) + y374 + 7.0*y39*(-37071008445.2255*r + 28533211037.6879*y0 - 37071008445.2255*y71 + 4129567622.65173))
 *         cdef double y378 = 61440.0*y0*y348 - 1.05297338872741e-18*y0*y355 + 92160.0*y168*y71 + y269*y373 - y269*y377 + y332*y370 - y35*y372 + y369
 *         cdef double y379 = 9.0*X_2             # <<<<<<<<<<<<<<
 *         cdef double y380 = y1*y129
 *         cdef double y381 = y128*y34
 */
  __Pyx_TraceLine(1134,0,__PYX_ERR(1, 1134, __pyx_L1_error))
  __pyx_v_y379 = (9.0 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1135
 *         cdef double y378 = 61440.0*y0*y348 - 1.05297338872741e-18*y0*y355 + 92160.0*y168*y71 + y269*y373 - y269*y377 + y332*y370 - y35*y372 + y369
 *         cdef double y379 = 9.0*X_2
 *         cdef double y380 = y1*y129             # <<<<<<<<<<<<<<
 *         cdef double y381 = y128*y34
 *         cdef double y382 = y295*y381
 */
  __Pyx_TraceLine(1135,0,__PYX_ERR(1, 1135, __pyx_L1_error))
  __pyx_v_y380 = (__pyx_v_y1 * __pyx_v_y129);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1136
 *         cdef double y379 = 9.0*X_2
 *         cdef double y380 = y1*y129
 *         cdef double y381 = y128*y34             # <<<<<<<<<<<<<<
 *         cdef double y382 = y295*y381
 *         cdef double y383 = y1*y91
 */
  __Pyx_TraceLine(1136,0,__PYX_ERR(1, 1136, __pyx_L1_error))
  __pyx_v_y381 = (__pyx_v_y128 * __pyx_v_y34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1137
 *         cdef double y380 = y1*y129
 *         cdef double y381 = y128*y34
 *         cdef double y382 = y295*y381             # <<<<<<<<<<<<<<
 *         cdef double y383 = y1*y91
 *         cdef double y384 = y128*y295*y297
 */
  __Pyx_TraceLine(1137,0,__PYX_ERR(1, 1137, __pyx_L1_error))
  __pyx_v_y382 = (__pyx_v_y295 * __pyx_v_y381);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1138
 *         cdef double y381 = y128*y34
 *         cdef double y382 = y295*y381
 *         cdef double y383 = y1*y91             # <<<<<<<<<<<<<<
 *         cdef double y384 = y128*y295*y297
 *         cdef double y385 = (0.230275523363951*nu + 0.031457442188381*y113 + 0.0614297810037367*y123 + 0.0307148905018684*y146 + 0.459657725867658*y178 + 5.30670671930294e-15*y319 + 5.30670671930294e-15*y320 - 9.78132182501918e-11*y321 - 1.54055818744052e-8*y322 + 0.228897162687159*y39 + 1)**2
 */
  __Pyx_TraceLine(1138,0,__PYX_ERR(1, 1138, __pyx_L1_error))
  __pyx_v_y383 = (__pyx_v_y1 * __pyx_v_y91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1139
 *         cdef double y382 = y295*y381
 *         cdef double y383 = y1*y91
 *         cdef double y384 = y128*y295*y297             # <<<<<<<<<<<<<<
 *         cdef double y385 = (0.230275523363951*nu + 0.031457442188381*y113 + 0.0614297810037367*y123 + 0.0307148905018684*y146 + 0.459657725867658*y178 + 5.30670671930294e-15*y319 + 5.30670671930294e-15*y320 - 9.78132182501918e-11*y321 - 1.54055818744052e-8*y322 + 0.228897162687159*y39 + 1)**2
 *         cdef double y386 = y315*y337
 */
  __Pyx_TraceLine(1139,0,__PYX_ERR(1, 1139, __pyx_L1_error))
  __pyx_v_y384 = ((__pyx_v_y128 * __pyx_v_y295) * __pyx_v_y297);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1140
 *         cdef double y383 = y1*y91
 *         cdef double y384 = y128*y295*y297
 *         cdef double y385 = (0.230275523363951*nu + 0.031457442188381*y113 + 0.0614297810037367*y123 + 0.0307148905018684*y146 + 0.459657725867658*y178 + 5.30670671930294e-15*y319 + 5.30670671930294e-15*y320 - 9.78132182501918e-11*y321 - 1.54055818744052e-8*y322 + 0.228897162687159*y39 + 1)**2             # <<<<<<<<<<<<<<
 *         cdef double y386 = y315*y337
 *         cdef double y387 = 11575876386816.0*y77
 */
  __Pyx_TraceLine(1140,0,__PYX_ERR(1, 1140, __pyx_L1_error))
  __pyx_v_y385 = pow((((((((((((0.230275523363951 * __pyx_v_nu) + (0.031457442188381 * __pyx_v_y113)) + (0.0614297810037367 * __pyx_v_y123)) + (0.0307148905018684 * __pyx_v_y146)) + (0.459657725867658 * __pyx_v_y178)) + (5.30670671930294e-15 * __pyx_v_y319)) + (5.30670671930294e-15 * __pyx_v_y320)) - (9.78132182501918e-11 * __pyx_v_y321)) - (1.54055818744052e-8 * __pyx_v_y322)) + (0.228897162687159 * __pyx_v_y39)) + 1.0), 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1141
 *         cdef double y384 = y128*y295*y297
 *         cdef double y385 = (0.230275523363951*nu + 0.031457442188381*y113 + 0.0614297810037367*y123 + 0.0307148905018684*y146 + 0.459657725867658*y178 + 5.30670671930294e-15*y319 + 5.30670671930294e-15*y320 - 9.78132182501918e-11*y321 - 1.54055818744052e-8*y322 + 0.228897162687159*y39 + 1)**2
 *         cdef double y386 = y315*y337             # <<<<<<<<<<<<<<
 *         cdef double y387 = 11575876386816.0*y77
 *         cdef double y388 = 8323596288000.0*nu + 24297540157440.0
 */
  __Pyx_TraceLine(1141,0,__PYX_ERR(1, 1141, __pyx_L1_error))
  __pyx_v_y386 = (__pyx_v_y315 * __pyx_v_y337);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1142
 *         cdef double y385 = (0.230275523363951*nu + 0.031457442188381*y113 + 0.0614297810037367*y123 + 0.0307148905018684*y146 + 0.459657725867658*y178 + 5.30670671930294e-15*y319 + 5.30670671930294e-15*y320 - 9.78132182501918e-11*y321 - 1.54055818744052e-8*y322 + 0.228897162687159*y39 + 1)**2
 *         cdef double y386 = y315*y337
 *         cdef double y387 = 11575876386816.0*y77             # <<<<<<<<<<<<<<
 *         cdef double y388 = 8323596288000.0*nu + 24297540157440.0
 *         cdef double y389 = -y123*y387 + 1759846072320.0*y123 + y193*y77 + y306 + y307 + y387 + y91*(3519692144640.0*r + y388) + 10215490662751.4
 */
  __Pyx_TraceLine(1142,0,__PYX_ERR(1, 1142, __pyx_L1_error))
  __pyx_v_y387 = (11575876386816.0 * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1143
 *         cdef double y386 = y315*y337
 *         cdef double y387 = 11575876386816.0*y77
 *         cdef double y388 = 8323596288000.0*nu + 24297540157440.0             # <<<<<<<<<<<<<<
 *         cdef double y389 = -y123*y387 + 1759846072320.0*y123 + y193*y77 + y306 + y307 + y387 + y91*(3519692144640.0*r + y388) + 10215490662751.4
 *         cdef double y390 = y203*y302
 */
  __Pyx_TraceLine(1143,0,__PYX_ERR(1, 1143, __pyx_L1_error))
  __pyx_v_y388 = ((8323596288000.0 * __pyx_v_nu) + 24297540157440.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1144
 *         cdef double y387 = 11575876386816.0*y77
 *         cdef double y388 = 8323596288000.0*nu + 24297540157440.0
 *         cdef double y389 = -y123*y387 + 1759846072320.0*y123 + y193*y77 + y306 + y307 + y387 + y91*(3519692144640.0*r + y388) + 10215490662751.4             # <<<<<<<<<<<<<<
 *         cdef double y390 = y203*y302
 *         cdef double y391 = y314*y316
 */
  __Pyx_TraceLine(1144,0,__PYX_ERR(1, 1144, __pyx_L1_error))
  __pyx_v_y389 = (((((((((-__pyx_v_y123) * __pyx_v_y387) + (1759846072320.0 * __pyx_v_y123)) + (__pyx_v_y193 * __pyx_v_y77)) + __pyx_v_y306) + __pyx_v_y307) + __pyx_v_y387) + (__pyx_v_y91 * ((3519692144640.0 * __pyx_v_r) + __pyx_v_y388))) + 10215490662751.4);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1145
 *         cdef double y388 = 8323596288000.0*nu + 24297540157440.0
 *         cdef double y389 = -y123*y387 + 1759846072320.0*y123 + y193*y77 + y306 + y307 + y387 + y91*(3519692144640.0*r + y388) + 10215490662751.4
 *         cdef double y390 = y203*y302             # <<<<<<<<<<<<<<
 *         cdef double y391 = y314*y316
 *         cdef double y392 = y304*y305
 */
  __Pyx_TraceLine(1145,0,__PYX_ERR(1, 1145, __pyx_L1_error))
  __pyx_v_y390 = (__pyx_v_y203 * __pyx_v_y302);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1146
 *         cdef double y389 = -y123*y387 + 1759846072320.0*y123 + y193*y77 + y306 + y307 + y387 + y91*(3519692144640.0*r + y388) + 10215490662751.4
 *         cdef double y390 = y203*y302
 *         cdef double y391 = y314*y316             # <<<<<<<<<<<<<<
 *         cdef double y392 = y304*y305
 *         cdef double y393 = (0.108541457767442*nu + 0.190937736865098*r + 0.0967857763822762*y187 + 0.00500066803742898*y189 + 0.216364706551791*y273 + 1.86910000868887e-14*y310 - 1.86910000868887e-14*y311 + 0.0340677222520525*y39 - 1)**2
 */
  __Pyx_TraceLine(1146,0,__PYX_ERR(1, 1146, __pyx_L1_error))
  __pyx_v_y391 = (__pyx_v_y314 * __pyx_v_y316);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1147
 *         cdef double y390 = y203*y302
 *         cdef double y391 = y314*y316
 *         cdef double y392 = y304*y305             # <<<<<<<<<<<<<<
 *         cdef double y393 = (0.108541457767442*nu + 0.190937736865098*r + 0.0967857763822762*y187 + 0.00500066803742898*y189 + 0.216364706551791*y273 + 1.86910000868887e-14*y310 - 1.86910000868887e-14*y311 + 0.0340677222520525*y39 - 1)**2
 *         cdef double y394 = 11614301777632.7*nu - 5806080.0*r*y175 + y123*(5279538216960.0*r + y388) - y183*y77 + y308 + y309 + 3645361092898.41*y39 + y91*(40226753557568.7*nu + 1132462080.0*y190 + 10083442360320.0*y39 + 5279538216960.0*y71 - 208372220299875.0) - 107003370108748.0
 */
  __Pyx_TraceLine(1147,0,__PYX_ERR(1, 1147, __pyx_L1_error))
  __pyx_v_y392 = (__pyx_v_y304 * __pyx_v_y305);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1148
 *         cdef double y391 = y314*y316
 *         cdef double y392 = y304*y305
 *         cdef double y393 = (0.108541457767442*nu + 0.190937736865098*r + 0.0967857763822762*y187 + 0.00500066803742898*y189 + 0.216364706551791*y273 + 1.86910000868887e-14*y310 - 1.86910000868887e-14*y311 + 0.0340677222520525*y39 - 1)**2             # <<<<<<<<<<<<<<
 *         cdef double y394 = 11614301777632.7*nu - 5806080.0*r*y175 + y123*(5279538216960.0*r + y388) - y183*y77 + y308 + y309 + 3645361092898.41*y39 + y91*(40226753557568.7*nu + 1132462080.0*y190 + 10083442360320.0*y39 + 5279538216960.0*y71 - 208372220299875.0) - 107003370108748.0
 *         cdef double y395 = y196*y390
 */
  __Pyx_TraceLine(1148,0,__PYX_ERR(1, 1148, __pyx_L1_error))
  __pyx_v_y393 = pow((((((((((0.108541457767442 * __pyx_v_nu) + (0.190937736865098 * __pyx_v_r)) + (0.0967857763822762 * __pyx_v_y187)) + (0.00500066803742898 * __pyx_v_y189)) + (0.216364706551791 * __pyx_v_y273)) + (1.86910000868887e-14 * __pyx_v_y310)) - (1.86910000868887e-14 * __pyx_v_y311)) + (0.0340677222520525 * __pyx_v_y39)) - 1.0), 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1149
 *         cdef double y392 = y304*y305
 *         cdef double y393 = (0.108541457767442*nu + 0.190937736865098*r + 0.0967857763822762*y187 + 0.00500066803742898*y189 + 0.216364706551791*y273 + 1.86910000868887e-14*y310 - 1.86910000868887e-14*y311 + 0.0340677222520525*y39 - 1)**2
 *         cdef double y394 = 11614301777632.7*nu - 5806080.0*r*y175 + y123*(5279538216960.0*r + y388) - y183*y77 + y308 + y309 + 3645361092898.41*y39 + y91*(40226753557568.7*nu + 1132462080.0*y190 + 10083442360320.0*y39 + 5279538216960.0*y71 - 208372220299875.0) - 107003370108748.0             # <<<<<<<<<<<<<<
 *         cdef double y395 = y196*y390
 *         cdef double y396 = y220*y323
 */
  __Pyx_TraceLine(1149,0,__PYX_ERR(1, 1149, __pyx_L1_error))
  __pyx_v_y394 = (((((((((11614301777632.7 * __pyx_v_nu) - ((5806080.0 * __pyx_v_r) * __pyx_v_y175)) + (__pyx_v_y123 * ((5279538216960.0 * __pyx_v_r) + __pyx_v_y388))) - (__pyx_v_y183 * __pyx_v_y77)) + __pyx_v_y308) + __pyx_v_y309) + (3645361092898.41 * __pyx_v_y39)) + (__pyx_v_y91 * (((((40226753557568.7 * __pyx_v_nu) + (1132462080.0 * __pyx_v_y190)) + (10083442360320.0 * __pyx_v_y39)) + (5279538216960.0 * __pyx_v_y71)) - 208372220299875.0))) - 107003370108748.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1150
 *         cdef double y393 = (0.108541457767442*nu + 0.190937736865098*r + 0.0967857763822762*y187 + 0.00500066803742898*y189 + 0.216364706551791*y273 + 1.86910000868887e-14*y310 - 1.86910000868887e-14*y311 + 0.0340677222520525*y39 - 1)**2
 *         cdef double y394 = 11614301777632.7*nu - 5806080.0*r*y175 + y123*(5279538216960.0*r + y388) - y183*y77 + y308 + y309 + 3645361092898.41*y39 + y91*(40226753557568.7*nu + 1132462080.0*y190 + 10083442360320.0*y39 + 5279538216960.0*y71 - 208372220299875.0) - 107003370108748.0
 *         cdef double y395 = y196*y390             # <<<<<<<<<<<<<<
 *         cdef double y396 = y220*y323
 *         cdef double y397 = y315*y396
 */
  __Pyx_TraceLine(1150,0,__PYX_ERR(1, 1150, __pyx_L1_error))
  __pyx_v_y395 = (__pyx_v_y196 * __pyx_v_y390);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1151
 *         cdef double y394 = 11614301777632.7*nu - 5806080.0*r*y175 + y123*(5279538216960.0*r + y388) - y183*y77 + y308 + y309 + 3645361092898.41*y39 + y91*(40226753557568.7*nu + 1132462080.0*y190 + 10083442360320.0*y39 + 5279538216960.0*y71 - 208372220299875.0) - 107003370108748.0
 *         cdef double y395 = y196*y390
 *         cdef double y396 = y220*y323             # <<<<<<<<<<<<<<
 *         cdef double y397 = y315*y396
 *         cdef double y398 = y196*y392
 */
  __Pyx_TraceLine(1151,0,__PYX_ERR(1, 1151, __pyx_L1_error))
  __pyx_v_y396 = (__pyx_v_y220 * __pyx_v_y323);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1152
 *         cdef double y395 = y196*y390
 *         cdef double y396 = y220*y323
 *         cdef double y397 = y315*y396             # <<<<<<<<<<<<<<
 *         cdef double y398 = y196*y392
 *         cdef double y399 = y204*y314
 */
  __Pyx_TraceLine(1152,0,__PYX_ERR(1, 1152, __pyx_L1_error))
  __pyx_v_y397 = (__pyx_v_y315 * __pyx_v_y396);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1153
 *         cdef double y396 = y220*y323
 *         cdef double y397 = y315*y396
 *         cdef double y398 = y196*y392             # <<<<<<<<<<<<<<
 *         cdef double y399 = y204*y314
 *         cdef double y400 = y211*y230
 */
  __Pyx_TraceLine(1153,0,__PYX_ERR(1, 1153, __pyx_L1_error))
  __pyx_v_y398 = (__pyx_v_y196 * __pyx_v_y392);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1154
 *         cdef double y397 = y315*y396
 *         cdef double y398 = y196*y392
 *         cdef double y399 = y204*y314             # <<<<<<<<<<<<<<
 *         cdef double y400 = y211*y230
 *         cdef double y401 = y211*y323
 */
  __Pyx_TraceLine(1154,0,__PYX_ERR(1, 1154, __pyx_L1_error))
  __pyx_v_y399 = (__pyx_v_y204 * __pyx_v_y314);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1155
 *         cdef double y398 = y196*y392
 *         cdef double y399 = y204*y314
 *         cdef double y400 = y211*y230             # <<<<<<<<<<<<<<
 *         cdef double y401 = y211*y323
 *         cdef double y402 = y225*y401
 */
  __Pyx_TraceLine(1155,0,__PYX_ERR(1, 1155, __pyx_L1_error))
  __pyx_v_y400 = (__pyx_v_y211 * __pyx_v_y230);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1156
 *         cdef double y399 = y204*y314
 *         cdef double y400 = y211*y230
 *         cdef double y401 = y211*y323             # <<<<<<<<<<<<<<
 *         cdef double y402 = y225*y401
 *         cdef double y403 = y202*y36
 */
  __Pyx_TraceLine(1156,0,__PYX_ERR(1, 1156, __pyx_L1_error))
  __pyx_v_y401 = (__pyx_v_y211 * __pyx_v_y323);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1157
 *         cdef double y400 = y211*y230
 *         cdef double y401 = y211*y323
 *         cdef double y402 = y225*y401             # <<<<<<<<<<<<<<
 *         cdef double y403 = y202*y36
 *         cdef double y404 = 1.35633680555556e-7*y403
 */
  __Pyx_TraceLine(1157,0,__PYX_ERR(1, 1157, __pyx_L1_error))
  __pyx_v_y402 = (__pyx_v_y225 * __pyx_v_y401);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1158
 *         cdef double y401 = y211*y323
 *         cdef double y402 = y225*y401
 *         cdef double y403 = y202*y36             # <<<<<<<<<<<<<<
 *         cdef double y404 = 1.35633680555556e-7*y403
 *         cdef double y405 = y228*y43
 */
  __Pyx_TraceLine(1158,0,__PYX_ERR(1, 1158, __pyx_L1_error))
  __pyx_v_y403 = (__pyx_v_y202 * __pyx_v_y36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1159
 *         cdef double y402 = y225*y401
 *         cdef double y403 = y202*y36
 *         cdef double y404 = 1.35633680555556e-7*y403             # <<<<<<<<<<<<<<
 *         cdef double y405 = y228*y43
 *         cdef double y406 = 1.69542100694444e-7*y405
 */
  __Pyx_TraceLine(1159,0,__PYX_ERR(1, 1159, __pyx_L1_error))
  __pyx_v_y404 = (1.35633680555556e-7 * __pyx_v_y403);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1160
 *         cdef double y403 = y202*y36
 *         cdef double y404 = 1.35633680555556e-7*y403
 *         cdef double y405 = y228*y43             # <<<<<<<<<<<<<<
 *         cdef double y406 = 1.69542100694444e-7*y405
 *         cdef double y407 = y205*y229
 */
  __Pyx_TraceLine(1160,0,__PYX_ERR(1, 1160, __pyx_L1_error))
  __pyx_v_y405 = (__pyx_v_y228 * __pyx_v_y43);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1161
 *         cdef double y404 = 1.35633680555556e-7*y403
 *         cdef double y405 = y228*y43
 *         cdef double y406 = 1.69542100694444e-7*y405             # <<<<<<<<<<<<<<
 *         cdef double y407 = y205*y229
 *         cdef double y408 = y225*y323*y407
 */
  __Pyx_TraceLine(1161,0,__PYX_ERR(1, 1161, __pyx_L1_error))
  __pyx_v_y406 = (1.69542100694444e-7 * __pyx_v_y405);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1162
 *         cdef double y405 = y228*y43
 *         cdef double y406 = 1.69542100694444e-7*y405
 *         cdef double y407 = y205*y229             # <<<<<<<<<<<<<<
 *         cdef double y408 = y225*y323*y407
 *         cdef double y409 = 4.48183298009152e-37*y196*y312
 */
  __Pyx_TraceLine(1162,0,__PYX_ERR(1, 1162, __pyx_L1_error))
  __pyx_v_y407 = (__pyx_v_y205 * __pyx_v_y229);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1163
 *         cdef double y406 = 1.69542100694444e-7*y405
 *         cdef double y407 = y205*y229
 *         cdef double y408 = y225*y323*y407             # <<<<<<<<<<<<<<
 *         cdef double y409 = 4.48183298009152e-37*y196*y312
 *         cdef double y410 = y206*y409
 */
  __Pyx_TraceLine(1163,0,__PYX_ERR(1, 1163, __pyx_L1_error))
  __pyx_v_y408 = ((__pyx_v_y225 * __pyx_v_y323) * __pyx_v_y407);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1164
 *         cdef double y407 = y205*y229
 *         cdef double y408 = y225*y323*y407
 *         cdef double y409 = 4.48183298009152e-37*y196*y312             # <<<<<<<<<<<<<<
 *         cdef double y410 = y206*y409
 *         cdef double y411 = y198*y336*y339
 */
  __Pyx_TraceLine(1164,0,__PYX_ERR(1, 1164, __pyx_L1_error))
  __pyx_v_y409 = ((4.48183298009152e-37 * __pyx_v_y196) * __pyx_v_y312);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1165
 *         cdef double y408 = y225*y323*y407
 *         cdef double y409 = 4.48183298009152e-37*y196*y312
 *         cdef double y410 = y206*y409             # <<<<<<<<<<<<<<
 *         cdef double y411 = y198*y336*y339
 *         cdef double y412 = (-4.28455968720463e-24*y284 + 3.25520833333333e-5*y331 + 0.25*y333 + y334)**2
 */
  __Pyx_TraceLine(1165,0,__PYX_ERR(1, 1165, __pyx_L1_error))
  __pyx_v_y410 = (__pyx_v_y206 * __pyx_v_y409);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1166
 *         cdef double y409 = 4.48183298009152e-37*y196*y312
 *         cdef double y410 = y206*y409
 *         cdef double y411 = y198*y336*y339             # <<<<<<<<<<<<<<
 *         cdef double y412 = (-4.28455968720463e-24*y284 + 3.25520833333333e-5*y331 + 0.25*y333 + y334)**2
 *         cdef double y413 = y229*y342
 */
  __Pyx_TraceLine(1166,0,__PYX_ERR(1, 1166, __pyx_L1_error))
  __pyx_v_y411 = ((__pyx_v_y198 * __pyx_v_y336) * __pyx_v_y339);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1167
 *         cdef double y410 = y206*y409
 *         cdef double y411 = y198*y336*y339
 *         cdef double y412 = (-4.28455968720463e-24*y284 + 3.25520833333333e-5*y331 + 0.25*y333 + y334)**2             # <<<<<<<<<<<<<<
 *         cdef double y413 = y229*y342
 *         cdef double y414 = 8.83031774450231e-12*y344
 */
  __Pyx_TraceLine(1167,0,__PYX_ERR(1, 1167, __pyx_L1_error))
  __pyx_v_y412 = pow(((((-4.28455968720463e-24 * __pyx_v_y284) + (3.25520833333333e-5 * __pyx_v_y331)) + (0.25 * __pyx_v_y333)) + __pyx_v_y334), 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1168
 *         cdef double y411 = y198*y336*y339
 *         cdef double y412 = (-4.28455968720463e-24*y284 + 3.25520833333333e-5*y331 + 0.25*y333 + y334)**2
 *         cdef double y413 = y229*y342             # <<<<<<<<<<<<<<
 *         cdef double y414 = 8.83031774450231e-12*y344
 *         cdef double y415 = y210*y414
 */
  __Pyx_TraceLine(1168,0,__PYX_ERR(1, 1168, __pyx_L1_error))
  __pyx_v_y413 = (__pyx_v_y229 * __pyx_v_y342);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1169
 *         cdef double y412 = (-4.28455968720463e-24*y284 + 3.25520833333333e-5*y331 + 0.25*y333 + y334)**2
 *         cdef double y413 = y229*y342
 *         cdef double y414 = 8.83031774450231e-12*y344             # <<<<<<<<<<<<<<
 *         cdef double y415 = y210*y414
 *         cdef double y416 = y220*y356
 */
  __Pyx_TraceLine(1169,0,__PYX_ERR(1, 1169, __pyx_L1_error))
  __pyx_v_y414 = (8.83031774450231e-12 * __pyx_v_y344);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1170
 *         cdef double y413 = y229*y342
 *         cdef double y414 = 8.83031774450231e-12*y344
 *         cdef double y415 = y210*y414             # <<<<<<<<<<<<<<
 *         cdef double y416 = y220*y356
 *         cdef double y417 = y211*y416
 */
  __Pyx_TraceLine(1170,0,__PYX_ERR(1, 1170, __pyx_L1_error))
  __pyx_v_y415 = (__pyx_v_y210 * __pyx_v_y414);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1171
 *         cdef double y414 = 8.83031774450231e-12*y344
 *         cdef double y415 = y210*y414
 *         cdef double y416 = y220*y356             # <<<<<<<<<<<<<<
 *         cdef double y417 = y211*y416
 *         cdef double y418 = 3.39084201388889e-8*y229
 */
  __Pyx_TraceLine(1171,0,__PYX_ERR(1, 1171, __pyx_L1_error))
  __pyx_v_y416 = (__pyx_v_y220 * __pyx_v_y356);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1172
 *         cdef double y415 = y210*y414
 *         cdef double y416 = y220*y356
 *         cdef double y417 = y211*y416             # <<<<<<<<<<<<<<
 *         cdef double y418 = 3.39084201388889e-8*y229
 *         cdef double y419 = y407*y416
 */
  __Pyx_TraceLine(1172,0,__PYX_ERR(1, 1172, __pyx_L1_error))
  __pyx_v_y417 = (__pyx_v_y211 * __pyx_v_y416);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1173
 *         cdef double y416 = y220*y356
 *         cdef double y417 = y211*y416
 *         cdef double y418 = 3.39084201388889e-8*y229             # <<<<<<<<<<<<<<
 *         cdef double y419 = y407*y416
 *         cdef double y420 = y222*y312
 */
  __Pyx_TraceLine(1173,0,__PYX_ERR(1, 1173, __pyx_L1_error))
  __pyx_v_y418 = (3.39084201388889e-8 * __pyx_v_y229);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1174
 *         cdef double y417 = y211*y416
 *         cdef double y418 = 3.39084201388889e-8*y229
 *         cdef double y419 = y407*y416             # <<<<<<<<<<<<<<
 *         cdef double y420 = y222*y312
 *         cdef double y421 = 76800.0*y221
 */
  __Pyx_TraceLine(1174,0,__PYX_ERR(1, 1174, __pyx_L1_error))
  __pyx_v_y419 = (__pyx_v_y407 * __pyx_v_y416);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1175
 *         cdef double y418 = 3.39084201388889e-8*y229
 *         cdef double y419 = y407*y416
 *         cdef double y420 = y222*y312             # <<<<<<<<<<<<<<
 *         cdef double y421 = 76800.0*y221
 *         cdef double y422 = 4.5850433485607e-22*y351
 */
  __Pyx_TraceLine(1175,0,__PYX_ERR(1, 1175, __pyx_L1_error))
  __pyx_v_y420 = (__pyx_v_y222 * __pyx_v_y312);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1176
 *         cdef double y419 = y407*y416
 *         cdef double y420 = y222*y312
 *         cdef double y421 = 76800.0*y221             # <<<<<<<<<<<<<<
 *         cdef double y422 = 4.5850433485607e-22*y351
 *         cdef double y423 = y167*y223
 */
  __Pyx_TraceLine(1176,0,__PYX_ERR(1, 1176, __pyx_L1_error))
  __pyx_v_y421 = (76800.0 * __pyx_v_y221);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1177
 *         cdef double y420 = y222*y312
 *         cdef double y421 = 76800.0*y221
 *         cdef double y422 = 4.5850433485607e-22*y351             # <<<<<<<<<<<<<<
 *         cdef double y423 = y167*y223
 *         cdef double y424 = y249**(-0.5)
 */
  __Pyx_TraceLine(1177,0,__PYX_ERR(1, 1177, __pyx_L1_error))
  __pyx_v_y422 = (4.5850433485607e-22 * __pyx_v_y351);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1178
 *         cdef double y421 = 76800.0*y221
 *         cdef double y422 = 4.5850433485607e-22*y351
 *         cdef double y423 = y167*y223             # <<<<<<<<<<<<<<
 *         cdef double y424 = y249**(-0.5)
 *         cdef double y425 = (y145*y252 + 2.0)**(-1)
 */
  __Pyx_TraceLine(1178,0,__PYX_ERR(1, 1178, __pyx_L1_error))
  __pyx_v_y423 = (__pyx_v_y167 * __pyx_v_y223);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1179
 *         cdef double y422 = 4.5850433485607e-22*y351
 *         cdef double y423 = y167*y223
 *         cdef double y424 = y249**(-0.5)             # <<<<<<<<<<<<<<
 *         cdef double y425 = (y145*y252 + 2.0)**(-1)
 *         cdef double y426 = 12.0*y205
 */
  __Pyx_TraceLine(1179,0,__PYX_ERR(1, 1179, __pyx_L1_error))
  __pyx_v_y424 = pow(__pyx_v_y249, -0.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1180
 *         cdef double y423 = y167*y223
 *         cdef double y424 = y249**(-0.5)
 *         cdef double y425 = (y145*y252 + 2.0)**(-1)             # <<<<<<<<<<<<<<
 *         cdef double y426 = 12.0*y205
 *         cdef double y427 = y111*y77
 */
  __Pyx_TraceLine(1180,0,__PYX_ERR(1, 1180, __pyx_L1_error))
  __pyx_v_y425 = pow(((__pyx_v_y145 * __pyx_v_y252) + 2.0), -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1181
 *         cdef double y424 = y249**(-0.5)
 *         cdef double y425 = (y145*y252 + 2.0)**(-1)
 *         cdef double y426 = 12.0*y205             # <<<<<<<<<<<<<<
 *         cdef double y427 = y111*y77
 *         cdef double y428 = y114*y32
 */
  __Pyx_TraceLine(1181,0,__PYX_ERR(1, 1181, __pyx_L1_error))
  __pyx_v_y426 = (12.0 * __pyx_v_y205);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1182
 *         cdef double y425 = (y145*y252 + 2.0)**(-1)
 *         cdef double y426 = 12.0*y205
 *         cdef double y427 = y111*y77             # <<<<<<<<<<<<<<
 *         cdef double y428 = y114*y32
 *         cdef double y429 = 30.0*y100
 */
  __Pyx_TraceLine(1182,0,__PYX_ERR(1, 1182, __pyx_L1_error))
  __pyx_v_y427 = (__pyx_v_y111 * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1183
 *         cdef double y426 = 12.0*y205
 *         cdef double y427 = y111*y77
 *         cdef double y428 = y114*y32             # <<<<<<<<<<<<<<
 *         cdef double y429 = 30.0*y100
 *         cdef double y430 = y116*y77
 */
  __Pyx_TraceLine(1183,0,__PYX_ERR(1, 1183, __pyx_L1_error))
  __pyx_v_y428 = (__pyx_v_y114 * __pyx_v_y32);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1184
 *         cdef double y427 = y111*y77
 *         cdef double y428 = y114*y32
 *         cdef double y429 = 30.0*y100             # <<<<<<<<<<<<<<
 *         cdef double y430 = y116*y77
 *         cdef double y431 = y119*y32
 */
  __Pyx_TraceLine(1184,0,__PYX_ERR(1, 1184, __pyx_L1_error))
  __pyx_v_y429 = (30.0 * __pyx_v_y100);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1185
 *         cdef double y428 = y114*y32
 *         cdef double y429 = 30.0*y100
 *         cdef double y430 = y116*y77             # <<<<<<<<<<<<<<
 *         cdef double y431 = y119*y32
 *         cdef double y432 = y121*y77
 */
  __Pyx_TraceLine(1185,0,__PYX_ERR(1, 1185, __pyx_L1_error))
  __pyx_v_y430 = (__pyx_v_y116 * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1186
 *         cdef double y429 = 30.0*y100
 *         cdef double y430 = y116*y77
 *         cdef double y431 = y119*y32             # <<<<<<<<<<<<<<
 *         cdef double y432 = y121*y77
 *         cdef double y433 = y124*y36
 */
  __Pyx_TraceLine(1186,0,__PYX_ERR(1, 1186, __pyx_L1_error))
  __pyx_v_y431 = (__pyx_v_y119 * __pyx_v_y32);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1187
 *         cdef double y430 = y116*y77
 *         cdef double y431 = y119*y32
 *         cdef double y432 = y121*y77             # <<<<<<<<<<<<<<
 *         cdef double y433 = y124*y36
 *         cdef double y434 = y199*y204
 */
  __Pyx_TraceLine(1187,0,__PYX_ERR(1, 1187, __pyx_L1_error))
  __pyx_v_y432 = (__pyx_v_y121 * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1188
 *         cdef double y431 = y119*y32
 *         cdef double y432 = y121*y77
 *         cdef double y433 = y124*y36             # <<<<<<<<<<<<<<
 *         cdef double y434 = y199*y204
 *         cdef double y435 = y210*y325
 */
  __Pyx_TraceLine(1188,0,__PYX_ERR(1, 1188, __pyx_L1_error))
  __pyx_v_y433 = (__pyx_v_y124 * __pyx_v_y36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1189
 *         cdef double y432 = y121*y77
 *         cdef double y433 = y124*y36
 *         cdef double y434 = y199*y204             # <<<<<<<<<<<<<<
 *         cdef double y435 = y210*y325
 *         cdef double y436 = y418*y435
 */
  __Pyx_TraceLine(1189,0,__PYX_ERR(1, 1189, __pyx_L1_error))
  __pyx_v_y434 = (__pyx_v_y199 * __pyx_v_y204);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1190
 *         cdef double y433 = y124*y36
 *         cdef double y434 = y199*y204
 *         cdef double y435 = y210*y325             # <<<<<<<<<<<<<<
 *         cdef double y436 = y418*y435
 *         cdef double y437 = prst**3
 */
  __Pyx_TraceLine(1190,0,__PYX_ERR(1, 1190, __pyx_L1_error))
  __pyx_v_y435 = (__pyx_v_y210 * __pyx_v_y325);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1191
 *         cdef double y434 = y199*y204
 *         cdef double y435 = y210*y325
 *         cdef double y436 = y418*y435             # <<<<<<<<<<<<<<
 *         cdef double y437 = prst**3
 *         cdef double y438 = y102*y437
 */
  __Pyx_TraceLine(1191,0,__PYX_ERR(1, 1191, __pyx_L1_error))
  __pyx_v_y436 = (__pyx_v_y418 * __pyx_v_y435);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1192
 *         cdef double y435 = y210*y325
 *         cdef double y436 = y418*y435
 *         cdef double y437 = prst**3             # <<<<<<<<<<<<<<
 *         cdef double y438 = y102*y437
 *         cdef double y439 = prst**5
 */
  __Pyx_TraceLine(1192,0,__PYX_ERR(1, 1192, __pyx_L1_error))
  __pyx_v_y437 = pow(__pyx_v_prst, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1193
 *         cdef double y436 = y418*y435
 *         cdef double y437 = prst**3
 *         cdef double y438 = y102*y437             # <<<<<<<<<<<<<<
 *         cdef double y439 = prst**5
 *         cdef double y440 = nu*y103*y439
 */
  __Pyx_TraceLine(1193,0,__PYX_ERR(1, 1193, __pyx_L1_error))
  __pyx_v_y438 = (__pyx_v_y102 * __pyx_v_y437);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1194
 *         cdef double y437 = prst**3
 *         cdef double y438 = y102*y437
 *         cdef double y439 = prst**5             # <<<<<<<<<<<<<<
 *         cdef double y440 = nu*y103*y439
 *         cdef double y441 = prst**7
 */
  __Pyx_TraceLine(1194,0,__PYX_ERR(1, 1194, __pyx_L1_error))
  __pyx_v_y439 = pow(__pyx_v_prst, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1195
 *         cdef double y438 = y102*y437
 *         cdef double y439 = prst**5
 *         cdef double y440 = nu*y103*y439             # <<<<<<<<<<<<<<
 *         cdef double y441 = prst**7
 *         cdef double y442 = nu*y441
 */
  __Pyx_TraceLine(1195,0,__PYX_ERR(1, 1195, __pyx_L1_error))
  __pyx_v_y440 = ((__pyx_v_nu * __pyx_v_y103) * __pyx_v_y439);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1196
 *         cdef double y439 = prst**5
 *         cdef double y440 = nu*y103*y439
 *         cdef double y441 = prst**7             # <<<<<<<<<<<<<<
 *         cdef double y442 = nu*y441
 *         cdef double y443 = y106*y442
 */
  __Pyx_TraceLine(1196,0,__PYX_ERR(1, 1196, __pyx_L1_error))
  __pyx_v_y441 = pow(__pyx_v_prst, 7.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1197
 *         cdef double y440 = nu*y103*y439
 *         cdef double y441 = prst**7
 *         cdef double y442 = nu*y441             # <<<<<<<<<<<<<<
 *         cdef double y443 = y106*y442
 *         cdef double y444 = y109*y441
 */
  __Pyx_TraceLine(1197,0,__PYX_ERR(1, 1197, __pyx_L1_error))
  __pyx_v_y442 = (__pyx_v_nu * __pyx_v_y441);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1198
 *         cdef double y441 = prst**7
 *         cdef double y442 = nu*y441
 *         cdef double y443 = y106*y442             # <<<<<<<<<<<<<<
 *         cdef double y444 = y109*y441
 *         cdef double y445 = 0.00678224732971111*y437
 */
  __Pyx_TraceLine(1198,0,__PYX_ERR(1, 1198, __pyx_L1_error))
  __pyx_v_y443 = (__pyx_v_y106 * __pyx_v_y442);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1199
 *         cdef double y442 = nu*y441
 *         cdef double y443 = y106*y442
 *         cdef double y444 = y109*y441             # <<<<<<<<<<<<<<
 *         cdef double y445 = 0.00678224732971111*y437
 *         cdef double y446 = 0.0101733709945667*y439
 */
  __Pyx_TraceLine(1199,0,__PYX_ERR(1, 1199, __pyx_L1_error))
  __pyx_v_y444 = (__pyx_v_y109 * __pyx_v_y441);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1200
 *         cdef double y443 = y106*y442
 *         cdef double y444 = y109*y441
 *         cdef double y445 = 0.00678224732971111*y437             # <<<<<<<<<<<<<<
 *         cdef double y446 = 0.0101733709945667*y439
 *         cdef double y447 = y432*y441
 */
  __Pyx_TraceLine(1200,0,__PYX_ERR(1, 1200, __pyx_L1_error))
  __pyx_v_y445 = (0.00678224732971111 * __pyx_v_y437);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1201
 *         cdef double y444 = y109*y441
 *         cdef double y445 = 0.00678224732971111*y437
 *         cdef double y446 = 0.0101733709945667*y439             # <<<<<<<<<<<<<<
 *         cdef double y447 = y432*y441
 *         cdef double y448 = y434*y437
 */
  __Pyx_TraceLine(1201,0,__PYX_ERR(1, 1201, __pyx_L1_error))
  __pyx_v_y446 = (0.0101733709945667 * __pyx_v_y439);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1202
 *         cdef double y445 = 0.00678224732971111*y437
 *         cdef double y446 = 0.0101733709945667*y439
 *         cdef double y447 = y432*y441             # <<<<<<<<<<<<<<
 *         cdef double y448 = y434*y437
 *         cdef double y449 = prst*y435
 */
  __Pyx_TraceLine(1202,0,__PYX_ERR(1, 1202, __pyx_L1_error))
  __pyx_v_y447 = (__pyx_v_y432 * __pyx_v_y441);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1203
 *         cdef double y446 = 0.0101733709945667*y439
 *         cdef double y447 = y432*y441
 *         cdef double y448 = y434*y437             # <<<<<<<<<<<<<<
 *         cdef double y449 = prst*y435
 *         cdef double y450 = y250*y255*(y170 + 0.000130208333333333*y236 + 0.000130208333333333*y241 + 0.000130208333333333*y244 + 0.000130208333333333*y246 + 0.000130208333333333*y96)**2
 */
  __Pyx_TraceLine(1203,0,__PYX_ERR(1, 1203, __pyx_L1_error))
  __pyx_v_y448 = (__pyx_v_y434 * __pyx_v_y437);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1204
 *         cdef double y447 = y432*y441
 *         cdef double y448 = y434*y437
 *         cdef double y449 = prst*y435             # <<<<<<<<<<<<<<
 *         cdef double y450 = y250*y255*(y170 + 0.000130208333333333*y236 + 0.000130208333333333*y241 + 0.000130208333333333*y244 + 0.000130208333333333*y246 + 0.000130208333333333*y96)**2
 *         cdef double y451 = y36*y46
 */
  __Pyx_TraceLine(1204,0,__PYX_ERR(1, 1204, __pyx_L1_error))
  __pyx_v_y449 = (__pyx_v_prst * __pyx_v_y435);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1205
 *         cdef double y448 = y434*y437
 *         cdef double y449 = prst*y435
 *         cdef double y450 = y250*y255*(y170 + 0.000130208333333333*y236 + 0.000130208333333333*y241 + 0.000130208333333333*y244 + 0.000130208333333333*y246 + 0.000130208333333333*y96)**2             # <<<<<<<<<<<<<<
 *         cdef double y451 = y36*y46
 *         cdef double y452 = 12.0*y34
 */
  __Pyx_TraceLine(1205,0,__PYX_ERR(1, 1205, __pyx_L1_error))
  __pyx_v_y450 = ((__pyx_v_y250 * __pyx_v_y255) * pow((((((__pyx_v_y170 + (0.000130208333333333 * __pyx_v_y236)) + (0.000130208333333333 * __pyx_v_y241)) + (0.000130208333333333 * __pyx_v_y244)) + (0.000130208333333333 * __pyx_v_y246)) + (0.000130208333333333 * __pyx_v_y96)), 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1206
 *         cdef double y449 = prst*y435
 *         cdef double y450 = y250*y255*(y170 + 0.000130208333333333*y236 + 0.000130208333333333*y241 + 0.000130208333333333*y244 + 0.000130208333333333*y246 + 0.000130208333333333*y96)**2
 *         cdef double y451 = y36*y46             # <<<<<<<<<<<<<<
 *         cdef double y452 = 12.0*y34
 *         cdef double y453 = y36*y50
 */
  __Pyx_TraceLine(1206,0,__PYX_ERR(1, 1206, __pyx_L1_error))
  __pyx_v_y451 = (__pyx_v_y36 * __pyx_v_y46);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1207
 *         cdef double y450 = y250*y255*(y170 + 0.000130208333333333*y236 + 0.000130208333333333*y241 + 0.000130208333333333*y244 + 0.000130208333333333*y246 + 0.000130208333333333*y96)**2
 *         cdef double y451 = y36*y46
 *         cdef double y452 = 12.0*y34             # <<<<<<<<<<<<<<
 *         cdef double y453 = y36*y50
 *         cdef double y454 = 2.0*pphi*y77
 */
  __Pyx_TraceLine(1207,0,__PYX_ERR(1, 1207, __pyx_L1_error))
  __pyx_v_y452 = (12.0 * __pyx_v_y34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1208
 *         cdef double y451 = y36*y46
 *         cdef double y452 = 12.0*y34
 *         cdef double y453 = y36*y50             # <<<<<<<<<<<<<<
 *         cdef double y454 = 2.0*pphi*y77
 *         cdef double y455 = pphi*y291
 */
  __Pyx_TraceLine(1208,0,__PYX_ERR(1, 1208, __pyx_L1_error))
  __pyx_v_y453 = (__pyx_v_y36 * __pyx_v_y50);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1209
 *         cdef double y452 = 12.0*y34
 *         cdef double y453 = y36*y50
 *         cdef double y454 = 2.0*pphi*y77             # <<<<<<<<<<<<<<
 *         cdef double y455 = pphi*y291
 *         cdef double y456 = pphi**3
 */
  __Pyx_TraceLine(1209,0,__PYX_ERR(1, 1209, __pyx_L1_error))
  __pyx_v_y454 = ((2.0 * __pyx_v_pphi) * __pyx_v_y77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1210
 *         cdef double y453 = y36*y50
 *         cdef double y454 = 2.0*pphi*y77
 *         cdef double y455 = pphi*y291             # <<<<<<<<<<<<<<
 *         cdef double y456 = pphi**3
 *         cdef double y457 = 4.0*y456
 */
  __Pyx_TraceLine(1210,0,__PYX_ERR(1, 1210, __pyx_L1_error))
  __pyx_v_y455 = (__pyx_v_pphi * __pyx_v_y291);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1211
 *         cdef double y454 = 2.0*pphi*y77
 *         cdef double y455 = pphi*y291
 *         cdef double y456 = pphi**3             # <<<<<<<<<<<<<<
 *         cdef double y457 = 4.0*y456
 *         cdef double y458 = y33*y454 + y40*y455 + y451*y457
 */
  __Pyx_TraceLine(1211,0,__PYX_ERR(1, 1211, __pyx_L1_error))
  __pyx_v_y456 = pow(__pyx_v_pphi, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1212
 *         cdef double y455 = pphi*y291
 *         cdef double y456 = pphi**3
 *         cdef double y457 = 4.0*y456             # <<<<<<<<<<<<<<
 *         cdef double y458 = y33*y454 + y40*y455 + y451*y457
 *         cdef double y459 = y453*y457 + y454*y48 + y455*y49
 */
  __Pyx_TraceLine(1212,0,__PYX_ERR(1, 1212, __pyx_L1_error))
  __pyx_v_y457 = (4.0 * __pyx_v_y456);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1213
 *         cdef double y456 = pphi**3
 *         cdef double y457 = 4.0*y456
 *         cdef double y458 = y33*y454 + y40*y455 + y451*y457             # <<<<<<<<<<<<<<
 *         cdef double y459 = y453*y457 + y454*y48 + y455*y49
 *         cdef double y460 = delta*y52
 */
  __Pyx_TraceLine(1213,0,__PYX_ERR(1, 1213, __pyx_L1_error))
  __pyx_v_y458 = (((__pyx_v_y33 * __pyx_v_y454) + (__pyx_v_y40 * __pyx_v_y455)) + (__pyx_v_y451 * __pyx_v_y457));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1214
 *         cdef double y457 = 4.0*y456
 *         cdef double y458 = y33*y454 + y40*y455 + y451*y457
 *         cdef double y459 = y453*y457 + y454*y48 + y455*y49             # <<<<<<<<<<<<<<
 *         cdef double y460 = delta*y52
 *         cdef double y461 = pphi*y1
 */
  __Pyx_TraceLine(1214,0,__PYX_ERR(1, 1214, __pyx_L1_error))
  __pyx_v_y459 = (((__pyx_v_y453 * __pyx_v_y457) + (__pyx_v_y454 * __pyx_v_y48)) + (__pyx_v_y455 * __pyx_v_y49));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1215
 *         cdef double y458 = y33*y454 + y40*y455 + y451*y457
 *         cdef double y459 = y453*y457 + y454*y48 + y455*y49
 *         cdef double y460 = delta*y52             # <<<<<<<<<<<<<<
 *         cdef double y461 = pphi*y1
 *         cdef double y462 = 4.0*y437
 */
  __Pyx_TraceLine(1215,0,__PYX_ERR(1, 1215, __pyx_L1_error))
  __pyx_v_y460 = (__pyx_v_delta * __pyx_v_y52);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1216
 *         cdef double y459 = y453*y457 + y454*y48 + y455*y49
 *         cdef double y460 = delta*y52
 *         cdef double y461 = pphi*y1             # <<<<<<<<<<<<<<
 *         cdef double y462 = 4.0*y437
 *         cdef double y463 = 6.0*y439
 */
  __Pyx_TraceLine(1216,0,__PYX_ERR(1, 1216, __pyx_L1_error))
  __pyx_v_y461 = (__pyx_v_pphi * __pyx_v_y1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1217
 *         cdef double y460 = delta*y52
 *         cdef double y461 = pphi*y1
 *         cdef double y462 = 4.0*y437             # <<<<<<<<<<<<<<
 *         cdef double y463 = 6.0*y439
 *         cdef double y464 = prst*y436 + y427*y462 + y428*y462 + y430*y463 + y431*y463 + y433*y462 + 589.775011960583*y438 - 67.9050514751178*y440 + 11.8620273619492*y443 + 0.975638950243592*y444 + 8.0*y447 + 5.09109256556341e-19*y448
 */
  __Pyx_TraceLine(1217,0,__PYX_ERR(1, 1217, __pyx_L1_error))
  __pyx_v_y462 = (4.0 * __pyx_v_y437);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1218
 *         cdef double y461 = pphi*y1
 *         cdef double y462 = 4.0*y437
 *         cdef double y463 = 6.0*y439             # <<<<<<<<<<<<<<
 *         cdef double y464 = prst*y436 + y427*y462 + y428*y462 + y430*y463 + y431*y463 + y433*y462 + 589.775011960583*y438 - 67.9050514751178*y440 + 11.8620273619492*y443 + 0.975638950243592*y444 + 8.0*y447 + 5.09109256556341e-19*y448
 *         cdef double y465 = y254*y256
 */
  __Pyx_TraceLine(1218,0,__PYX_ERR(1, 1218, __pyx_L1_error))
  __pyx_v_y463 = (6.0 * __pyx_v_y439);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1219
 *         cdef double y462 = 4.0*y437
 *         cdef double y463 = 6.0*y439
 *         cdef double y464 = prst*y436 + y427*y462 + y428*y462 + y430*y463 + y431*y463 + y433*y462 + 589.775011960583*y438 - 67.9050514751178*y440 + 11.8620273619492*y443 + 0.975638950243592*y444 + 8.0*y447 + 5.09109256556341e-19*y448             # <<<<<<<<<<<<<<
 *         cdef double y465 = y254*y256
 *         cdef double y466 = y234*y286
 */
  __Pyx_TraceLine(1219,0,__PYX_ERR(1, 1219, __pyx_L1_error))
  __pyx_v_y464 = ((((((((((((__pyx_v_prst * __pyx_v_y436) + (__pyx_v_y427 * __pyx_v_y462)) + (__pyx_v_y428 * __pyx_v_y462)) + (__pyx_v_y430 * __pyx_v_y463)) + (__pyx_v_y431 * __pyx_v_y463)) + (__pyx_v_y433 * __pyx_v_y462)) + (589.775011960583 * __pyx_v_y438)) - (67.9050514751178 * __pyx_v_y440)) + (11.8620273619492 * __pyx_v_y443)) + (0.975638950243592 * __pyx_v_y444)) + (8.0 * __pyx_v_y447)) + (5.09109256556341e-19 * __pyx_v_y448));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1220
 *         cdef double y463 = 6.0*y439
 *         cdef double y464 = prst*y436 + y427*y462 + y428*y462 + y430*y463 + y431*y463 + y433*y462 + 589.775011960583*y438 - 67.9050514751178*y440 + 11.8620273619492*y443 + 0.975638950243592*y444 + 8.0*y447 + 5.09109256556341e-19*y448
 *         cdef double y465 = y254*y256             # <<<<<<<<<<<<<<
 *         cdef double y466 = y234*y286
 *         cdef double y467 = 16.0*y25
 */
  __Pyx_TraceLine(1220,0,__PYX_ERR(1, 1220, __pyx_L1_error))
  __pyx_v_y465 = (__pyx_v_y254 * __pyx_v_y256);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1221
 *         cdef double y464 = prst*y436 + y427*y462 + y428*y462 + y430*y463 + y431*y463 + y433*y462 + 589.775011960583*y438 - 67.9050514751178*y440 + 11.8620273619492*y443 + 0.975638950243592*y444 + 8.0*y447 + 5.09109256556341e-19*y448
 *         cdef double y465 = y254*y256
 *         cdef double y466 = y234*y286             # <<<<<<<<<<<<<<
 *         cdef double y467 = 16.0*y25
 *         cdef double y468 = y247*y250*y358/(y145*y361 + 4.0)
 */
  __Pyx_TraceLine(1221,0,__PYX_ERR(1, 1221, __pyx_L1_error))
  __pyx_v_y466 = (__pyx_v_y234 * __pyx_v_y286);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1222
 *         cdef double y465 = y254*y256
 *         cdef double y466 = y234*y286
 *         cdef double y467 = 16.0*y25             # <<<<<<<<<<<<<<
 *         cdef double y468 = y247*y250*y358/(y145*y361 + 4.0)
 *         cdef double y469 = dSO*y29
 */
  __Pyx_TraceLine(1222,0,__PYX_ERR(1, 1222, __pyx_L1_error))
  __pyx_v_y467 = (16.0 * __pyx_v_y25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1223
 *         cdef double y466 = y234*y286
 *         cdef double y467 = 16.0*y25
 *         cdef double y468 = y247*y250*y358/(y145*y361 + 4.0)             # <<<<<<<<<<<<<<
 *         cdef double y469 = dSO*y29
 *         cdef double y470 = 4.0*pphi*y32
 */
  __Pyx_TraceLine(1223,0,__PYX_ERR(1, 1223, __pyx_L1_error))
  __pyx_v_y468 = (((__pyx_v_y247 * __pyx_v_y250) * __pyx_v_y358) / ((__pyx_v_y145 * __pyx_v_y361) + 4.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1224
 *         cdef double y467 = 16.0*y25
 *         cdef double y468 = y247*y250*y358/(y145*y361 + 4.0)
 *         cdef double y469 = dSO*y29             # <<<<<<<<<<<<<<
 *         cdef double y470 = 4.0*pphi*y32
 *         cdef double y471 = pphi*y37
 */
  __Pyx_TraceLine(1224,0,__PYX_ERR(1, 1224, __pyx_L1_error))
  __pyx_v_y469 = (__pyx_v_dSO * __pyx_v_y29);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1225
 *         cdef double y468 = y247*y250*y358/(y145*y361 + 4.0)
 *         cdef double y469 = dSO*y29
 *         cdef double y470 = 4.0*pphi*y32             # <<<<<<<<<<<<<<
 *         cdef double y471 = pphi*y37
 *         cdef double y472 = y456*y467
 */
  __Pyx_TraceLine(1225,0,__PYX_ERR(1, 1225, __pyx_L1_error))
  __pyx_v_y470 = ((4.0 * __pyx_v_pphi) * __pyx_v_y32);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1226
 *         cdef double y469 = dSO*y29
 *         cdef double y470 = 4.0*pphi*y32
 *         cdef double y471 = pphi*y37             # <<<<<<<<<<<<<<
 *         cdef double y472 = y456*y467
 *         cdef double y473 = y34*y460
 */
  __Pyx_TraceLine(1226,0,__PYX_ERR(1, 1226, __pyx_L1_error))
  __pyx_v_y471 = (__pyx_v_pphi * __pyx_v_y37);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1227
 *         cdef double y470 = 4.0*pphi*y32
 *         cdef double y471 = pphi*y37
 *         cdef double y472 = y456*y467             # <<<<<<<<<<<<<<
 *         cdef double y473 = y34*y460
 *         cdef double y474 = y129*y232
 */
  __Pyx_TraceLine(1227,0,__PYX_ERR(1, 1227, __pyx_L1_error))
  __pyx_v_y472 = (__pyx_v_y456 * __pyx_v_y467);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1228
 *         cdef double y471 = pphi*y37
 *         cdef double y472 = y456*y467
 *         cdef double y473 = y34*y460             # <<<<<<<<<<<<<<
 *         cdef double y474 = y129*y232
 *         cdef double y475 = -2.0*pphi*y144*y77 + y461*y474
 */
  __Pyx_TraceLine(1228,0,__PYX_ERR(1, 1228, __pyx_L1_error))
  __pyx_v_y473 = (__pyx_v_y34 * __pyx_v_y460);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1229
 *         cdef double y472 = y456*y467
 *         cdef double y473 = y34*y460
 *         cdef double y474 = y129*y232             # <<<<<<<<<<<<<<
 *         cdef double y475 = -2.0*pphi*y144*y77 + y461*y474
 *         cdef double y476 = -y475
 */
  __Pyx_TraceLine(1229,0,__PYX_ERR(1, 1229, __pyx_L1_error))
  __pyx_v_y474 = (__pyx_v_y129 * __pyx_v_y232);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1230
 *         cdef double y473 = y34*y460
 *         cdef double y474 = y129*y232
 *         cdef double y475 = -2.0*pphi*y144*y77 + y461*y474             # <<<<<<<<<<<<<<
 *         cdef double y476 = -y475
 * 
 */
  __Pyx_TraceLine(1230,0,__PYX_ERR(1, 1230, __pyx_L1_error))
  __pyx_v_y475 = ((((-2.0 * __pyx_v_pphi) * __pyx_v_y144) * __pyx_v_y77) + (__pyx_v_y461 * __pyx_v_y474));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1231
 *         cdef double y474 = y129*y232
 *         cdef double y475 = -2.0*pphi*y144*y77 + y461*y474
 *         cdef double y476 = -y475             # <<<<<<<<<<<<<<
 * 
 *         # Evaluate Hamiltonian
 */
  __Pyx_TraceLine(1231,0,__PYX_ERR(1, 1231, __pyx_L1_error))
  __pyx_v_y476 = (-__pyx_v_y475);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1235
 *         # Evaluate Hamiltonian
 *         cdef double H
 *         H,_,_,_,_,_,_,_,_ = self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)             # <<<<<<<<<<<<<<
 * 
 *         # Evaluate Heff Jacobian
 */
  __Pyx_TraceLine(1235,0,__PYX_ERR(1, 1235, __pyx_L1_error))
  __pyx_t_1 = ((struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._call(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 9)) {
      if (size > 9) __Pyx_RaiseTooManyValuesError(9);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 1235, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_16 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_12 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_11 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 8); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_16 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_12 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_11 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_10 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 8); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_7);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[9] = {&__pyx_t_2,&__pyx_t_13,&__pyx_t_16,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_7};
      for (i=0; i < 9; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(1, 1235, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[9] = {&__pyx_t_2,&__pyx_t_13,&__pyx_t_16,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_7};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_19 = Py_TYPE(__pyx_t_6)->tp_iternext;
    for (index=0; index < 9; index++) {
      PyObject* item = __pyx_t_19(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_6), 9) < 0) __PYX_ERR(1, 1235, __pyx_L1_error)
    __pyx_t_19 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_19 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(1, 1235, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_H = __pyx_t_18;
  __pyx_v__ = __pyx_t_13;
  __pyx_t_13 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_16);
  __pyx_t_16 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_12);
  __pyx_t_12 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_11);
  __pyx_t_11 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_8);
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1238
 * 
 *         # Evaluate Heff Jacobian
 *         cdef double dHeffdr = x12*(-dSO*x14*x15*x19 + x19*(x24*(1.25*x3 - 0.75*x48 - 0.416666666666667*x57 + 1.66666666666667*x58 + 1.25*x6 + 1.25*x7) - x31*x55) + x19*(-x21*x22 - x23*x27 - x24*(-11.0625*nu + 1.13541666666667*x28 - 0.15625) - x29*x31 - x32*x37) + x43*(0.375*x14*x49*x50 - x24*(0.0833333333333333*x51 + x52)) + x44*(-x21*x38 - x24*(-0.0625*nu + 1.07291666666667*x28 + 0.15625) - x27*x39 - x31*x40 - x37*x41)) + 0.5*x205*(x192*x195*(30720.0*x0*x119*x138 + 7680.0*x13*x138*x210 - x14*(x177 + x178 - 3.0*x48) + x206*x24 - x209*(x109*(14350.0*nu - 1320.0*x28 + 2805.0) + x111*(5830.0*nu - 1640.0*x28 - 855.0) + 0.625*x202) - x209*(0.015625*x179*(-135.0*nu + 140.0*x28 - 15.0) - 6.09375*x197 - x199*(35.0*nu + 45.0)) - x215 - x35*(0.5*x200 + 0.25*x201 + 8.25*x58) - x35*(0.5*x196 + x2*x208 - 5.0*x57 + 4.5*x6 + 4.5*x7)) - x192*x203*(x14*x206 - x207*x24*x9)/x194**2 + x204*(-663.496888455656*nu*r**(-5.5)*x70 - nu*x21*x80 + 39.6112800271521*nu*x71*x73 + x1*x21*x25*x97*x98 + x1*x25*x60*x98*(r*x9 + x10 + x212)/x96**2 - x100 - x112*x27 + 7.59859378406358e-45*x113*x115*x142*x165*x167*x184*x220*x70 + 6.78168402777778e-8*x114*x14*x170*x171*x175*x184*x190 + x14*x70*(118.4*x211 - 51.6952380952381*x79) - 9.25454462627843e-34*x142*x218*x219/x164**3 - 2.24091649004576e-37*x165*x169*x171*x184*x218*x221*x35 + 1.69542100694444e-8*x169*x170*x171*x175*x184*x35*(x138*x187*x223 - x14*(x180*(12.0*nu - 3.0) - 7.875*x57 + x9*(9.0*nu + 8.4375)) + 38400.0*x140*x186 + x185*x189*x218*x34 - x187*x214 + x222 - x35*(0.125*x181 + 0.0625*x182 + 0.0625*x183) - 2.29252167428035e-22*x174*x188*x220*x34/x155) + 1.69542100694444e-8*x169*x170*x171*x175*x190*x220*x35 - 8.47710503472222e-8*x191*x209 + x21*x25*(-x14*(x101*(-5.25*nu - 2.8125) + x102*(2.25 - x208) + x105*(0.5625 - 2.25*nu)) + x24*x49 - x35*(0.03125*x104*x110 - 0.25*x106 + 0.0104166666666667*x108*x49)) - x216*x92 - x30*x73*x89 - x36*x70*x94 - 3.70688355060912*x74*x78 - 2.0*x81*x85 - 3.0*x84*x95 - 2.0*x86*x90 - 4.41515887225116e-12*x169*x175*x221*x224*x225/x141**3 - 6.62902677807736e-23*x165*x219*x224/x141**5 + 1.01821851311268e-18*x114**3*x142*x155*x165*x167*x70/r**12 - 1.65460508380811e-18*x168/r**14)) - (3.0*x20 + x9)*(pphi*x63 + pphi*x67 + pphi*x69 + x19*x59 + x43*x64)/x11**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double dHeffdphi = 0
 */
  __Pyx_TraceLine(1238,0,__PYX_ERR(1, 1238, __pyx_L1_error))
  __pyx_v_dHeffdr = (((__pyx_v_x12 * ((((((((-__pyx_v_dSO) * __pyx_v_x14) * __pyx_v_x15) * __pyx_v_x19) + (__pyx_v_x19 * ((__pyx_v_x24 * ((((((1.25 * __pyx_v_x3) - (0.75 * __pyx_v_x48)) - (0.416666666666667 * __pyx_v_x57)) + (1.66666666666667 * __pyx_v_x58)) + (1.25 * __pyx_v_x6)) + (1.25 * __pyx_v_x7))) - (__pyx_v_x31 * __pyx_v_x55)))) + (__pyx_v_x19 * ((((((-__pyx_v_x21) * __pyx_v_x22) - (__pyx_v_x23 * __pyx_v_x27)) - (__pyx_v_x24 * (((-11.0625 * __pyx_v_nu) + (1.13541666666667 * __pyx_v_x28)) - 0.15625))) - (__pyx_v_x29 * __pyx_v_x31)) - (__pyx_v_x32 * __pyx_v_x37)))) + (__pyx_v_x43 * ((((0.375 * __pyx_v_x14) * __pyx_v_x49) * __pyx_v_x50) - (__pyx_v_x24 * ((0.0833333333333333 * __pyx_v_x51) + __pyx_v_x52))))) + (__pyx_v_x44 * ((((((-__pyx_v_x21) * __pyx_v_x38) - (__pyx_v_x24 * (((-0.0625 * __pyx_v_nu) + (1.07291666666667 * __pyx_v_x28)) + 0.15625))) - (__pyx_v_x27 * __pyx_v_x39)) - (__pyx_v_x31 * __pyx_v_x40)) - (__pyx_v_x37 * __pyx_v_x41))))) + ((0.5 * __pyx_v_x205) * ((((__pyx_v_x192 * __pyx_v_x195) * (((((((((((30720.0 * __pyx_v_x0) * __pyx_v_x119) * __pyx_v_x138) + (((7680.0 * __pyx_v_x13) * __pyx_v_x138) * __pyx_v_x210)) - (__pyx_v_x14 * ((__pyx_v_x177 + __pyx_v_x178) - (3.0 * __pyx_v_x48)))) + (__pyx_v_x206 * __pyx_v_x24)) - (__pyx_v_x209 * (((__pyx_v_x109 * (((14350.0 * __pyx_v_nu) - (1320.0 * __pyx_v_x28)) + 2805.0)) + (__pyx_v_x111 * (((5830.0 * __pyx_v_nu) - (1640.0 * __pyx_v_x28)) - 855.0))) + (0.625 * __pyx_v_x202)))) - (__pyx_v_x209 * ((((0.015625 * __pyx_v_x179) * (((-135.0 * __pyx_v_nu) + (140.0 * __pyx_v_x28)) - 15.0)) - (6.09375 * __pyx_v_x197)) - (__pyx_v_x199 * ((35.0 * __pyx_v_nu) + 45.0))))) - __pyx_v_x215) - (__pyx_v_x35 * (((0.5 * __pyx_v_x200) + (0.25 * __pyx_v_x201)) + (8.25 * __pyx_v_x58)))) - (__pyx_v_x35 * (((((0.5 * __pyx_v_x196) + (__pyx_v_x2 * __pyx_v_x208)) - (5.0 * __pyx_v_x57)) + (4.5 * __pyx_v_x6)) + (4.5 * __pyx_v_x7))))) - (((__pyx_v_x192 * __pyx_v_x203) * ((__pyx_v_x14 * __pyx_v_x206) - ((__pyx_v_x207 * __pyx_v_x24) * __pyx_v_x9))) / pow(__pyx_v_x194, 2.0))) + (__pyx_v_x204 * (((((((((((((((((((((((((((((-663.496888455656 * __pyx_v_nu) * pow(__pyx_v_r, -5.5)) * __pyx_v_x70) - ((__pyx_v_nu * __pyx_v_x21) * __pyx_v_x80)) + (((39.6112800271521 * __pyx_v_nu) * __pyx_v_x71) * __pyx_v_x73)) + ((((__pyx_v_x1 * __pyx_v_x21) * __pyx_v_x25) * __pyx_v_x97) * __pyx_v_x98)) + (((((__pyx_v_x1 * __pyx_v_x25) * __pyx_v_x60) * __pyx_v_x98) * (((__pyx_v_r * __pyx_v_x9) + __pyx_v_x10) + __pyx_v_x212)) / pow(__pyx_v_x96, 2.0))) - __pyx_v_x100) - (__pyx_v_x112 * __pyx_v_x27)) + ((((((((7.59859378406358e-45 * __pyx_v_x113) * __pyx_v_x115) * __pyx_v_x142) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x184) * __pyx_v_x220) * __pyx_v_x70)) + (((((((6.78168402777778e-8 * __pyx_v_x114) * __pyx_v_x14) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x190)) + ((__pyx_v_x14 * __pyx_v_x70) * ((118.4 * __pyx_v_x211) - (51.6952380952381 * __pyx_v_x79)))) - ((((9.25454462627843e-34 * __pyx_v_x142) * __pyx_v_x218) * __pyx_v_x219) / pow(__pyx_v_x164, 3.0))) - (((((((2.24091649004576e-37 * __pyx_v_x165) * __pyx_v_x169) * __pyx_v_x171) * __pyx_v_x184) * __pyx_v_x218) * __pyx_v_x221) * __pyx_v_x35)) + (((((((1.69542100694444e-8 * __pyx_v_x169) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x35) * (((((((((__pyx_v_x138 * __pyx_v_x187) * __pyx_v_x223) - (__pyx_v_x14 * (((__pyx_v_x180 * ((12.0 * __pyx_v_nu) - 3.0)) - (7.875 * __pyx_v_x57)) + (__pyx_v_x9 * ((9.0 * __pyx_v_nu) + 8.4375))))) + ((38400.0 * __pyx_v_x140) * __pyx_v_x186)) + (((__pyx_v_x185 * __pyx_v_x189) * __pyx_v_x218) * __pyx_v_x34)) - (__pyx_v_x187 * __pyx_v_x214)) + __pyx_v_x222) - (__pyx_v_x35 * (((0.125 * __pyx_v_x181) + (0.0625 * __pyx_v_x182)) + (0.0625 * __pyx_v_x183)))) - (((((2.29252167428035e-22 * __pyx_v_x174) * __pyx_v_x188) * __pyx_v_x220) * __pyx_v_x34) / __pyx_v_x155)))) + (((((((1.69542100694444e-8 * __pyx_v_x169) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x190) * __pyx_v_x220) * __pyx_v_x35)) - ((8.47710503472222e-8 * __pyx_v_x191) * __pyx_v_x209)) + ((__pyx_v_x21 * __pyx_v_x25) * ((((-__pyx_v_x14) * (((__pyx_v_x101 * ((-5.25 * __pyx_v_nu) - 2.8125)) + (__pyx_v_x102 * (2.25 - __pyx_v_x208))) + (__pyx_v_x105 * (0.5625 - (2.25 * __pyx_v_nu))))) + (__pyx_v_x24 * __pyx_v_x49)) - (__pyx_v_x35 * ((((0.03125 * __pyx_v_x104) * __pyx_v_x110) - (0.25 * __pyx_v_x106)) + ((0.0104166666666667 * __pyx_v_x108) * __pyx_v_x49)))))) - (__pyx_v_x216 * __pyx_v_x92)) - ((__pyx_v_x30 * __pyx_v_x73) * __pyx_v_x89)) - ((__pyx_v_x36 * __pyx_v_x70) * __pyx_v_x94)) - ((3.70688355060912 * __pyx_v_x74) * __pyx_v_x78)) - ((2.0 * __pyx_v_x81) * __pyx_v_x85)) - ((3.0 * __pyx_v_x84) * __pyx_v_x95)) - ((2.0 * __pyx_v_x86) * __pyx_v_x90)) - ((((((4.41515887225116e-12 * __pyx_v_x169) * __pyx_v_x175) * __pyx_v_x221) * __pyx_v_x224) * __pyx_v_x225) / pow(__pyx_v_x141, 3.0))) - ((((6.62902677807736e-23 * __pyx_v_x165) * __pyx_v_x219) * __pyx_v_x224) / pow(__pyx_v_x141, 5.0))) + (((((((1.01821851311268e-18 * pow(__pyx_v_x114, 3.0)) * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x70) / pow(__pyx_v_r, 12.0))) - ((1.65460508380811e-18 * __pyx_v_x168) / pow(__pyx_v_r, 14.0))))))) - ((((3.0 * __pyx_v_x20) + __pyx_v_x9) * (((((__pyx_v_pphi * __pyx_v_x63) + (__pyx_v_pphi * __pyx_v_x67)) + (__pyx_v_pphi * __pyx_v_x69)) + (__pyx_v_x19 * __pyx_v_x59)) + (__pyx_v_x43 * __pyx_v_x64))) / pow(__pyx_v_x11, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1240
 *         cdef double dHeffdr = x12*(-dSO*x14*x15*x19 + x19*(x24*(1.25*x3 - 0.75*x48 - 0.416666666666667*x57 + 1.66666666666667*x58 + 1.25*x6 + 1.25*x7) - x31*x55) + x19*(-x21*x22 - x23*x27 - x24*(-11.0625*nu + 1.13541666666667*x28 - 0.15625) - x29*x31 - x32*x37) + x43*(0.375*x14*x49*x50 - x24*(0.0833333333333333*x51 + x52)) + x44*(-x21*x38 - x24*(-0.0625*nu + 1.07291666666667*x28 + 0.15625) - x27*x39 - x31*x40 - x37*x41)) + 0.5*x205*(x192*x195*(30720.0*x0*x119*x138 + 7680.0*x13*x138*x210 - x14*(x177 + x178 - 3.0*x48) + x206*x24 - x209*(x109*(14350.0*nu - 1320.0*x28 + 2805.0) + x111*(5830.0*nu - 1640.0*x28 - 855.0) + 0.625*x202) - x209*(0.015625*x179*(-135.0*nu + 140.0*x28 - 15.0) - 6.09375*x197 - x199*(35.0*nu + 45.0)) - x215 - x35*(0.5*x200 + 0.25*x201 + 8.25*x58) - x35*(0.5*x196 + x2*x208 - 5.0*x57 + 4.5*x6 + 4.5*x7)) - x192*x203*(x14*x206 - x207*x24*x9)/x194**2 + x204*(-663.496888455656*nu*r**(-5.5)*x70 - nu*x21*x80 + 39.6112800271521*nu*x71*x73 + x1*x21*x25*x97*x98 + x1*x25*x60*x98*(r*x9 + x10 + x212)/x96**2 - x100 - x112*x27 + 7.59859378406358e-45*x113*x115*x142*x165*x167*x184*x220*x70 + 6.78168402777778e-8*x114*x14*x170*x171*x175*x184*x190 + x14*x70*(118.4*x211 - 51.6952380952381*x79) - 9.25454462627843e-34*x142*x218*x219/x164**3 - 2.24091649004576e-37*x165*x169*x171*x184*x218*x221*x35 + 1.69542100694444e-8*x169*x170*x171*x175*x184*x35*(x138*x187*x223 - x14*(x180*(12.0*nu - 3.0) - 7.875*x57 + x9*(9.0*nu + 8.4375)) + 38400.0*x140*x186 + x185*x189*x218*x34 - x187*x214 + x222 - x35*(0.125*x181 + 0.0625*x182 + 0.0625*x183) - 2.29252167428035e-22*x174*x188*x220*x34/x155) + 1.69542100694444e-8*x169*x170*x171*x175*x190*x220*x35 - 8.47710503472222e-8*x191*x209 + x21*x25*(-x14*(x101*(-5.25*nu - 2.8125) + x102*(2.25 - x208) + x105*(0.5625 - 2.25*nu)) + x24*x49 - x35*(0.03125*x104*x110 - 0.25*x106 + 0.0104166666666667*x108*x49)) - x216*x92 - x30*x73*x89 - x36*x70*x94 - 3.70688355060912*x74*x78 - 2.0*x81*x85 - 3.0*x84*x95 - 2.0*x86*x90 - 4.41515887225116e-12*x169*x175*x221*x224*x225/x141**3 - 6.62902677807736e-23*x165*x219*x224/x141**5 + 1.01821851311268e-18*x114**3*x142*x155*x165*x167*x70/r**12 - 1.65460508380811e-18*x168/r**14)) - (3.0*x20 + x9)*(pphi*x63 + pphi*x67 + pphi*x69 + x19*x59 + x43*x64)/x11**2
 * 
 *         cdef double dHeffdphi = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)
 */
  __Pyx_TraceLine(1240,0,__PYX_ERR(1, 1240, __pyx_L1_error))
  __pyx_v_dHeffdphi = 0.0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1242
 *         cdef double dHeffdphi = 0
 * 
 *         cdef double dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)             # <<<<<<<<<<<<<<
 * 
 *         cdef double dHeffdpphi = x12*(x18*x59 + 2.0*x18*x68 + x19*(x23*x232 + x233*x29 + x234*x32) + x42*x64 - 0.25*x42*x66 + x44*(x232*x39 + x233*x40 + x234*x41) + x63 + x67 + x69) + x231*(-pphi*x1*x193*x99 + 2.0*pphi*x112*x21)
 */
  __Pyx_TraceLine(1242,0,__PYX_ERR(1, 1242, __pyx_L1_error))
  __pyx_v_dHeffdpr = (__pyx_v_x231 * ((((((((((((((11.8620273619492 * __pyx_v_nu) * __pyx_v_x228) * __pyx_v_x76) + ((((((3.39084201388889e-8 * __pyx_v_prst) * __pyx_v_x169) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x190) * __pyx_v_x225)) + (((((((5.09109256556341e-19 * __pyx_v_x113) * __pyx_v_x115) * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x226)) + ((__pyx_v_x14 * __pyx_v_x229) * __pyx_v_x94)) + (((8.0 * __pyx_v_x21) * __pyx_v_x228) * __pyx_v_x91)) + ((589.775011960583 * __pyx_v_x226) * __pyx_v_x72)) - ((67.9050514751178 * __pyx_v_x227) * __pyx_v_x75)) + ((0.975638950243592 * __pyx_v_x228) * __pyx_v_x79)) + ((__pyx_v_x229 * __pyx_v_x24) * __pyx_v_x84)) + (__pyx_v_x229 * __pyx_v_x82)) + ((__pyx_v_x230 * __pyx_v_x24) * __pyx_v_x89)) + (__pyx_v_x230 * __pyx_v_x87)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1244
 *         cdef double dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)
 * 
 *         cdef double dHeffdpphi = x12*(x18*x59 + 2.0*x18*x68 + x19*(x23*x232 + x233*x29 + x234*x32) + x42*x64 - 0.25*x42*x66 + x44*(x232*x39 + x233*x40 + x234*x41) + x63 + x67 + x69) + x231*(-pphi*x1*x193*x99 + 2.0*pphi*x112*x21)             # <<<<<<<<<<<<<<
 * 
 *         # Evaluate Heff Hessian
 */
  __Pyx_TraceLine(1244,0,__PYX_ERR(1, 1244, __pyx_L1_error))
  __pyx_v_dHeffdpphi = ((__pyx_v_x12 * (((((((((__pyx_v_x18 * __pyx_v_x59) + ((2.0 * __pyx_v_x18) * __pyx_v_x68)) + (__pyx_v_x19 * (((__pyx_v_x23 * __pyx_v_x232) + (__pyx_v_x233 * __pyx_v_x29)) + (__pyx_v_x234 * __pyx_v_x32)))) + (__pyx_v_x42 * __pyx_v_x64)) - ((0.25 * __pyx_v_x42) * __pyx_v_x66)) + (__pyx_v_x44 * (((__pyx_v_x232 * __pyx_v_x39) + (__pyx_v_x233 * __pyx_v_x40)) + (__pyx_v_x234 * __pyx_v_x41)))) + __pyx_v_x63) + __pyx_v_x67) + __pyx_v_x69)) + (__pyx_v_x231 * (((((-__pyx_v_pphi) * __pyx_v_x1) * __pyx_v_x193) * __pyx_v_x99) + (((2.0 * __pyx_v_pphi) * __pyx_v_x112) * __pyx_v_x21))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1247
 * 
 *         # Evaluate Heff Hessian
 *         cdef double d2Heffdr2 = y23*(y25*y26*y31 + y30*(y36*(-3.75*y13 - 3.75*y18 + 2.25*y58 - y69 - 3.75*y7 - 5.0*y70) + y41*y66) + y30*(y32*(1.4375*nu - 0.1875) + y33*y38 + y36*(-33.1875*nu + 3.40625*y39 - 0.46875) + y40*y42 + y44*y47) + y53*(-1.5*y25*y61 + y36*(y59*(1.25 - 2.5*X_2) + 0.25*y62)) + y54*(y32*(0.6875*nu + 0.1875) + y36*(-0.1875*nu + 3.21875*y39 + 0.46875) + y38*y48 + y42*y49 + y44*y51)) - 0.25*y250*y358**2 + 0.5*y424*(-y231*y255*y286*y362 + 2.0*y231*y247*y254**2/y233**3 + y234*y357*(61440.0*y0*y165*y167 + 15360.0*y165*y268*y35 - y25*(y242 + 0.5*y243 + 16.5*y70) - y25*(9.0*y13 + 9.0*y18 + y235 + y260*y379 - 10.0*y68) - 2.63243347181853e-19*y284 + y32*y360 - y36*(y368 - 6.0*y58) - y43*(1.25*y245 + y266*y365 + 0.00520833333333333*y364) - y43*(0.03125*y215*y263 - 12.1875*y237 - y262*y366)) + y248*(3649.23288650611*r**(-6.5)*y289 - 2.0*y1*y384*y92 - 8.0*y100*y25*y293 - 7.91520185839956e-48*y100*y336*y338*y396 + 12.9740924271319*y102*y107 - 178.250760122184*y105*y288 + 0.243909737560898*y108*y32 + 6.0*y111*y125 + y115*y41 + y117*y37 + y120*y41 + y122*y37 + y125*(51.6952380952381*nu*y77 - y292*y77) + 2.0*y129*y92 + y144*y38 + 1.62760416666667e-6*y171*y412*y413 + 1.16714400523217e-40*y196*y226*y312*y344*y407 - 2.5455462827817e-17*y200*y201*y305 - 6.103515625e-7*y202*y227*y25 + 7.59859378406358e-45*y220*y386*y394 + y220*y400*(y165*y223*y370 - y167*y371*y420 - y168*y24*y312*y422 - 2.29252167428035e-21*y169*y209*y351 + y169*y312*y421 + y209*y333*y421 - y223*y372 + y25*(y20*(36.0*nu + 33.75) + 0.25*y215*y346 - 31.5*y68) - 1.31621673590926e-18*y284*y349 + 153600.0*y334*y349 + y347*y389 - y348*y352*y422 + 15360.0*y348*y420 - y350*y354*y394 + 7.85795675813156e-45*y351*y352*y355 + y369 + y373*y423 - y377*y423 + y43*(0.0625*y135*(575.0*nu + 20.0*y39 - 185.0) + 0.0625*y20*(-5855.0*nu - 4305.0) + 0.625*y217) + 2.81299777100766e-6*y353*y385/y185**3) + y225*y394*y400 + 1.35633680555556e-7*y227*y32 + 4.66406554828496e-24*y229*y313*y326*y393 + y232*y297*y382 + y294*y44 - 4.0*y301*y79 - y313*y318*y389 + 5.08626302083333e-7*y324*y363 - y327*y329*y389 - 1.79273319203661e-36*y328*y403 + 2.24091649004576e-36*y328*y405 - y330*y341*y378 - 4.41515887225116e-12*y343*y378*y413 - 3.53212709780093e-11*y345*y403 + 4.41515887225116e-11*y345*y405 + y356*y401*y418 - y380*y80 + y382*y383*(y14*(2.0*y15 + 2.0*y16 + 2.0*y17) + y375 + y6*(4.0*y2 + 4.0*y3 + 4.0*y4) + y8*(2.0*y10 + 2.0*y11 + 2.0*y9)) + 2.69825540021952e-16*y385*y386 + 2.40618160283239e-32*y390*y391 - 1.48072714020455e-32*y391*y392 - 1.97563438385653e-43*y395*y397 + 1.72354696230011e-21*y395*y411 + 1.21577500545017e-43*y397*y398 - 1.10501271569469e-58*y397*y399 - 1.06064428449238e-21*y398*y411 + 9.64015065237337e-37*y399*y411 + y402*y404 - y402*y406 + y404*y417 - y406*y417 - y408*y410 - y408*y415 - y410*y419 - y415*y419 + y92*(y25*(y136*(1.125 - 4.5*nu) + y59*(-10.5*nu - 5.625) + y64*(4.5 - y379)) - y258*y36 + y43*(-1.25*y138 + y299*(-8050.0*nu + y264 + 1875.0) + y300*(-5450.0*nu + 760.0*y39 + 1095.0))) + 2.88925109089694e-20*y317*y393/y195**4 + 4.07287405245073e-17*y340*y412/y170**6 - 32.0*y381*y383*(y0 + 0.25*y21 + 0.25*y296)**2/y126**3 + 6.10931107867609e-18*y200*y228/r**11 + 2.31644711733135e-17*y303/r**15) - y256*y357*y362 - y257*(y20*y36*(12.0*y91 + 6.0) + y25*y359) + y287*(y25*(y359 - 12.0*y58) + y363*(3.75*y245 + 0.015625*y364 + y365*(17490.0*nu - 4920.0*y39 - 2565.0)) + y363*(0.03125*y215*(-405.0*nu + 420.0*y39 - 45.0) - 36.5625*y237 - y366*(105.0*nu + 135.0)) + y378 + y43*(0.25*y136*(-y367 - 15.0) + y298*(y261 - 155.0) + 41.25*y70) + y43*(22.5*y13 + 0.5*y135*(y367 + 5.0) + 22.5*y18 - 25.0*y68 + 22.5*y7))) - y74*(y72 + y73)*(pphi*y84 + pphi*y88 + pphi*y89 - y31*y76 + y54*y86) + (-6.0*r*y74 + 18.0*(0.333333333333333*y13 + 0.333333333333333*y18 + 0.333333333333333*y7 + y71)**2/y22**3)*(pphi*y94 + pphi*y97 + pphi*y99 + y31*y90 + y54*y95)             # <<<<<<<<<<<<<<
 * 
 *         cdef double d2Heffdphi2 = 0
 */
  __Pyx_TraceLine(1247,0,__PYX_ERR(1, 1247, __pyx_L1_error))
  __pyx_v_d2Heffdr2 = (((((__pyx_v_y23 * ((((((__pyx_v_y25 * __pyx_v_y26) * __pyx_v_y31) + (__pyx_v_y30 * ((__pyx_v_y36 * ((((((-3.75 * __pyx_v_y13) - (3.75 * __pyx_v_y18)) + (2.25 * __pyx_v_y58)) - __pyx_v_y69) - (3.75 * __pyx_v_y7)) - (5.0 * __pyx_v_y70))) + (__pyx_v_y41 * __pyx_v_y66)))) + (__pyx_v_y30 * (((((__pyx_v_y32 * ((1.4375 * __pyx_v_nu) - 0.1875)) + (__pyx_v_y33 * __pyx_v_y38)) + (__pyx_v_y36 * (((-33.1875 * __pyx_v_nu) + (3.40625 * __pyx_v_y39)) - 0.46875))) + (__pyx_v_y40 * __pyx_v_y42)) + (__pyx_v_y44 * __pyx_v_y47)))) + (__pyx_v_y53 * (((-1.5 * __pyx_v_y25) * __pyx_v_y61) + (__pyx_v_y36 * ((__pyx_v_y59 * (1.25 - (2.5 * __pyx_v_X_2))) + (0.25 * __pyx_v_y62)))))) + (__pyx_v_y54 * (((((__pyx_v_y32 * ((0.6875 * __pyx_v_nu) + 0.1875)) + (__pyx_v_y36 * (((-0.1875 * __pyx_v_nu) + (3.21875 * __pyx_v_y39)) + 0.46875))) + (__pyx_v_y38 * __pyx_v_y48)) + (__pyx_v_y42 * __pyx_v_y49)) + (__pyx_v_y44 * __pyx_v_y51))))) - ((0.25 * __pyx_v_y250) * pow(__pyx_v_y358, 2.0))) + ((0.5 * __pyx_v_y424) * ((((((((((-__pyx_v_y231) * __pyx_v_y255) * __pyx_v_y286) * __pyx_v_y362) + ((((2.0 * __pyx_v_y231) * __pyx_v_y247) * pow(__pyx_v_y254, 2.0)) / pow(__pyx_v_y233, 3.0))) + ((__pyx_v_y234 * __pyx_v_y357) * (((((((((((61440.0 * __pyx_v_y0) * __pyx_v_y165) * __pyx_v_y167) + (((15360.0 * __pyx_v_y165) * __pyx_v_y268) * __pyx_v_y35)) - (__pyx_v_y25 * ((__pyx_v_y242 + (0.5 * __pyx_v_y243)) + (16.5 * __pyx_v_y70)))) - (__pyx_v_y25 * (((((9.0 * __pyx_v_y13) + (9.0 * __pyx_v_y18)) + __pyx_v_y235) + (__pyx_v_y260 * __pyx_v_y379)) - (10.0 * __pyx_v_y68)))) - (2.63243347181853e-19 * __pyx_v_y284)) + (__pyx_v_y32 * __pyx_v_y360)) - (__pyx_v_y36 * (__pyx_v_y368 - (6.0 * __pyx_v_y58)))) - (__pyx_v_y43 * (((1.25 * __pyx_v_y245) + (__pyx_v_y266 * __pyx_v_y365)) + (0.00520833333333333 * __pyx_v_y364)))) - (__pyx_v_y43 * ((((0.03125 * __pyx_v_y215) * __pyx_v_y263) - (12.1875 * __pyx_v_y237)) - (__pyx_v_y262 * __pyx_v_y366)))))) + (__pyx_v_y248 * (((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((3649.23288650611 * pow(__pyx_v_r, -6.5)) * __pyx_v_y289) - (((2.0 * __pyx_v_y1) * __pyx_v_y384) * __pyx_v_y92)) - (((8.0 * __pyx_v_y100) * __pyx_v_y25) * __pyx_v_y293)) - ((((7.91520185839956e-48 * __pyx_v_y100) * __pyx_v_y336) * __pyx_v_y338) * __pyx_v_y396)) + ((12.9740924271319 * __pyx_v_y102) * __pyx_v_y107)) - ((178.250760122184 * __pyx_v_y105) * __pyx_v_y288)) + ((0.243909737560898 * __pyx_v_y108) * __pyx_v_y32)) + ((6.0 * __pyx_v_y111) * __pyx_v_y125)) + (__pyx_v_y115 * __pyx_v_y41)) + (__pyx_v_y117 * __pyx_v_y37)) + (__pyx_v_y120 * __pyx_v_y41)) + (__pyx_v_y122 * __pyx_v_y37)) + (__pyx_v_y125 * (((51.6952380952381 * __pyx_v_nu) * __pyx_v_y77) - (__pyx_v_y292 * __pyx_v_y77)))) + ((2.0 * __pyx_v_y129) * __pyx_v_y92)) + (__pyx_v_y144 * __pyx_v_y38)) + (((1.62760416666667e-6 * __pyx_v_y171) * __pyx_v_y412) * __pyx_v_y413)) + (((((1.16714400523217e-40 * __pyx_v_y196) * __pyx_v_y226) * __pyx_v_y312) * __pyx_v_y344) * __pyx_v_y407)) - (((2.5455462827817e-17 * __pyx_v_y200) * __pyx_v_y201) * __pyx_v_y305)) - (((6.103515625e-7 * __pyx_v_y202) * __pyx_v_y227) * __pyx_v_y25)) + (((7.59859378406358e-45 * __pyx_v_y220) * __pyx_v_y386) * __pyx_v_y394)) + ((__pyx_v_y220 * __pyx_v_y400) * (((((((((((((((((((((__pyx_v_y165 * __pyx_v_y223) * __pyx_v_y370) - ((__pyx_v_y167 * __pyx_v_y371) * __pyx_v_y420)) - (((__pyx_v_y168 * __pyx_v_y24) * __pyx_v_y312) * __pyx_v_y422)) - (((2.29252167428035e-21 * __pyx_v_y169) * __pyx_v_y209) * __pyx_v_y351)) + ((__pyx_v_y169 * __pyx_v_y312) * __pyx_v_y421)) + ((__pyx_v_y209 * __pyx_v_y333) * __pyx_v_y421)) - (__pyx_v_y223 * __pyx_v_y372)) + (__pyx_v_y25 * (((__pyx_v_y20 * ((36.0 * __pyx_v_nu) + 33.75)) + ((0.25 * __pyx_v_y215) * __pyx_v_y346)) - (31.5 * __pyx_v_y68)))) - ((1.31621673590926e-18 * __pyx_v_y284) * __pyx_v_y349)) + ((153600.0 * __pyx_v_y334) * __pyx_v_y349)) + (__pyx_v_y347 * __pyx_v_y389)) - ((__pyx_v_y348 * __pyx_v_y352) * __pyx_v_y422)) + ((15360.0 * __pyx_v_y348) * __pyx_v_y420)) - ((__pyx_v_y350 * __pyx_v_y354) * __pyx_v_y394)) + (((7.85795675813156e-45 * __pyx_v_y351) * __pyx_v_y352) * __pyx_v_y355)) + __pyx_v_y369) + (__pyx_v_y373 * __pyx_v_y423)) - (__pyx_v_y377 * __pyx_v_y423)) + (__pyx_v_y43 * ((((0.0625 * __pyx_v_y135) * (((575.0 * __pyx_v_nu) + (20.0 * __pyx_v_y39)) - 185.0)) + ((0.0625 * __pyx_v_y20) * ((-5855.0 * __pyx_v_nu) - 4305.0))) + (0.625 * __pyx_v_y217)))) + (((2.81299777100766e-6 * __pyx_v_y353) * __pyx_v_y385) / pow(__pyx_v_y185, 3.0))))) + ((__pyx_v_y225 * __pyx_v_y394) * __pyx_v_y400)) + ((1.35633680555556e-7 * __pyx_v_y227) * __pyx_v_y32)) + ((((4.66406554828496e-24 * __pyx_v_y229) * __pyx_v_y313) * __pyx_v_y326) * __pyx_v_y393)) + ((__pyx_v_y232 * __pyx_v_y297) * __pyx_v_y382)) + (__pyx_v_y294 * __pyx_v_y44)) - ((4.0 * __pyx_v_y301) * __pyx_v_y79)) - ((__pyx_v_y313 * __pyx_v_y318) * __pyx_v_y389)) + ((5.08626302083333e-7 * __pyx_v_y324) * __pyx_v_y363)) - ((__pyx_v_y327 * __pyx_v_y329) * __pyx_v_y389)) - ((1.79273319203661e-36 * __pyx_v_y328) * __pyx_v_y403)) + ((2.24091649004576e-36 * __pyx_v_y328) * __pyx_v_y405)) - ((__pyx_v_y330 * __pyx_v_y341) * __pyx_v_y378)) - (((4.41515887225116e-12 * __pyx_v_y343) * __pyx_v_y378) * __pyx_v_y413)) - ((3.53212709780093e-11 * __pyx_v_y345) * __pyx_v_y403)) + ((4.41515887225116e-11 * __pyx_v_y345) * __pyx_v_y405)) + ((__pyx_v_y356 * __pyx_v_y401) * __pyx_v_y418)) - (__pyx_v_y380 * __pyx_v_y80)) + ((__pyx_v_y382 * __pyx_v_y383) * ((((__pyx_v_y14 * (((2.0 * __pyx_v_y15) + (2.0 * __pyx_v_y16)) + (2.0 * __pyx_v_y17))) + __pyx_v_y375) + (__pyx_v_y6 * (((4.0 * __pyx_v_y2) + (4.0 * __pyx_v_y3)) + (4.0 * __pyx_v_y4)))) + (__pyx_v_y8 * (((2.0 * __pyx_v_y10) + (2.0 * __pyx_v_y11)) + (2.0 * __pyx_v_y9)))))) + ((2.69825540021952e-16 * __pyx_v_y385) * __pyx_v_y386)) + ((2.40618160283239e-32 * __pyx_v_y390) * __pyx_v_y391)) - ((1.48072714020455e-32 * __pyx_v_y391) * __pyx_v_y392)) - ((1.97563438385653e-43 * __pyx_v_y395) * __pyx_v_y397)) + ((1.72354696230011e-21 * __pyx_v_y395) * __pyx_v_y411)) + ((1.21577500545017e-43 * __pyx_v_y397) * __pyx_v_y398)) - ((1.10501271569469e-58 * __pyx_v_y397) * __pyx_v_y399)) - ((1.06064428449238e-21 * __pyx_v_y398) * __pyx_v_y411)) + ((9.64015065237337e-37 * __pyx_v_y399) * __pyx_v_y411)) + (__pyx_v_y402 * __pyx_v_y404)) - (__pyx_v_y402 * __pyx_v_y406)) + (__pyx_v_y404 * __pyx_v_y417)) - (__pyx_v_y406 * __pyx_v_y417)) - (__pyx_v_y408 * __pyx_v_y410)) - (__pyx_v_y408 * __pyx_v_y415)) - (__pyx_v_y410 * __pyx_v_y419)) - (__pyx_v_y415 * __pyx_v_y419)) + (__pyx_v_y92 * (((__pyx_v_y25 * (((__pyx_v_y136 * (1.125 - (4.5 * __pyx_v_nu))) + (__pyx_v_y59 * ((-10.5 * __pyx_v_nu) - 5.625))) + (__pyx_v_y64 * (4.5 - __pyx_v_y379)))) - (__pyx_v_y258 * __pyx_v_y36)) + (__pyx_v_y43 * (((-1.25 * __pyx_v_y138) + (__pyx_v_y299 * (((-8050.0 * __pyx_v_nu) + __pyx_v_y264) + 1875.0))) + (__pyx_v_y300 * (((-5450.0 * __pyx_v_nu) + (760.0 * __pyx_v_y39)) + 1095.0))))))) + (((2.88925109089694e-20 * __pyx_v_y317) * __pyx_v_y393) / pow(__pyx_v_y195, 4.0))) + (((4.07287405245073e-17 * __pyx_v_y340) * __pyx_v_y412) / pow(__pyx_v_y170, 6.0))) - ((((32.0 * __pyx_v_y381) * __pyx_v_y383) * pow(((__pyx_v_y0 + (0.25 * __pyx_v_y21)) + (0.25 * __pyx_v_y296)), 2.0)) / pow(__pyx_v_y126, 3.0))) + (((6.10931107867609e-18 * __pyx_v_y200) * __pyx_v_y228) / pow(__pyx_v_r, 11.0))) + ((2.31644711733135e-17 * __pyx_v_y303) / pow(__pyx_v_r, 15.0))))) - ((__pyx_v_y256 * __pyx_v_y357) * __pyx_v_y362)) - (__pyx_v_y257 * (((__pyx_v_y20 * __pyx_v_y36) * ((12.0 * __pyx_v_y91) + 6.0)) + (__pyx_v_y25 * __pyx_v_y359)))) + (__pyx_v_y287 * ((((((__pyx_v_y25 * (__pyx_v_y359 - (12.0 * __pyx_v_y58))) + (__pyx_v_y363 * (((3.75 * __pyx_v_y245) + (0.015625 * __pyx_v_y364)) + (__pyx_v_y365 * (((17490.0 * __pyx_v_nu) - (4920.0 * __pyx_v_y39)) - 2565.0))))) + (__pyx_v_y363 * ((((0.03125 * __pyx_v_y215) * (((-405.0 * __pyx_v_nu) + (420.0 * __pyx_v_y39)) - 45.0)) - (36.5625 * __pyx_v_y237)) - (__pyx_v_y366 * ((105.0 * __pyx_v_nu) + 135.0))))) + __pyx_v_y378) + (__pyx_v_y43 * ((((0.25 * __pyx_v_y136) * ((-__pyx_v_y367) - 15.0)) + (__pyx_v_y298 * (__pyx_v_y261 - 155.0))) + (41.25 * __pyx_v_y70)))) + (__pyx_v_y43 * (((((22.5 * __pyx_v_y13) + ((0.5 * __pyx_v_y135) * (__pyx_v_y367 + 5.0))) + (22.5 * __pyx_v_y18)) - (25.0 * __pyx_v_y68)) + (22.5 * __pyx_v_y7)))))))) - ((__pyx_v_y74 * (__pyx_v_y72 + __pyx_v_y73)) * (((((__pyx_v_pphi * __pyx_v_y84) + (__pyx_v_pphi * __pyx_v_y88)) + (__pyx_v_pphi * __pyx_v_y89)) - (__pyx_v_y31 * __pyx_v_y76)) + (__pyx_v_y54 * __pyx_v_y86)))) + ((((-6.0 * __pyx_v_r) * __pyx_v_y74) + ((18.0 * pow(((((0.333333333333333 * __pyx_v_y13) + (0.333333333333333 * __pyx_v_y18)) + (0.333333333333333 * __pyx_v_y7)) + __pyx_v_y71), 2.0)) / pow(__pyx_v_y22, 3.0))) * (((((__pyx_v_pphi * __pyx_v_y94) + (__pyx_v_pphi * __pyx_v_y97)) + (__pyx_v_pphi * __pyx_v_y99)) + (__pyx_v_y31 * __pyx_v_y90)) + (__pyx_v_y54 * __pyx_v_y95))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1249
 *         cdef double d2Heffdr2 = y23*(y25*y26*y31 + y30*(y36*(-3.75*y13 - 3.75*y18 + 2.25*y58 - y69 - 3.75*y7 - 5.0*y70) + y41*y66) + y30*(y32*(1.4375*nu - 0.1875) + y33*y38 + y36*(-33.1875*nu + 3.40625*y39 - 0.46875) + y40*y42 + y44*y47) + y53*(-1.5*y25*y61 + y36*(y59*(1.25 - 2.5*X_2) + 0.25*y62)) + y54*(y32*(0.6875*nu + 0.1875) + y36*(-0.1875*nu + 3.21875*y39 + 0.46875) + y38*y48 + y42*y49 + y44*y51)) - 0.25*y250*y358**2 + 0.5*y424*(-y231*y255*y286*y362 + 2.0*y231*y247*y254**2/y233**3 + y234*y357*(61440.0*y0*y165*y167 + 15360.0*y165*y268*y35 - y25*(y242 + 0.5*y243 + 16.5*y70) - y25*(9.0*y13 + 9.0*y18 + y235 + y260*y379 - 10.0*y68) - 2.63243347181853e-19*y284 + y32*y360 - y36*(y368 - 6.0*y58) - y43*(1.25*y245 + y266*y365 + 0.00520833333333333*y364) - y43*(0.03125*y215*y263 - 12.1875*y237 - y262*y366)) + y248*(3649.23288650611*r**(-6.5)*y289 - 2.0*y1*y384*y92 - 8.0*y100*y25*y293 - 7.91520185839956e-48*y100*y336*y338*y396 + 12.9740924271319*y102*y107 - 178.250760122184*y105*y288 + 0.243909737560898*y108*y32 + 6.0*y111*y125 + y115*y41 + y117*y37 + y120*y41 + y122*y37 + y125*(51.6952380952381*nu*y77 - y292*y77) + 2.0*y129*y92 + y144*y38 + 1.62760416666667e-6*y171*y412*y413 + 1.16714400523217e-40*y196*y226*y312*y344*y407 - 2.5455462827817e-17*y200*y201*y305 - 6.103515625e-7*y202*y227*y25 + 7.59859378406358e-45*y220*y386*y394 + y220*y400*(y165*y223*y370 - y167*y371*y420 - y168*y24*y312*y422 - 2.29252167428035e-21*y169*y209*y351 + y169*y312*y421 + y209*y333*y421 - y223*y372 + y25*(y20*(36.0*nu + 33.75) + 0.25*y215*y346 - 31.5*y68) - 1.31621673590926e-18*y284*y349 + 153600.0*y334*y349 + y347*y389 - y348*y352*y422 + 15360.0*y348*y420 - y350*y354*y394 + 7.85795675813156e-45*y351*y352*y355 + y369 + y373*y423 - y377*y423 + y43*(0.0625*y135*(575.0*nu + 20.0*y39 - 185.0) + 0.0625*y20*(-5855.0*nu - 4305.0) + 0.625*y217) + 2.81299777100766e-6*y353*y385/y185**3) + y225*y394*y400 + 1.35633680555556e-7*y227*y32 + 4.66406554828496e-24*y229*y313*y326*y393 + y232*y297*y382 + y294*y44 - 4.0*y301*y79 - y313*y318*y389 + 5.08626302083333e-7*y324*y363 - y327*y329*y389 - 1.79273319203661e-36*y328*y403 + 2.24091649004576e-36*y328*y405 - y330*y341*y378 - 4.41515887225116e-12*y343*y378*y413 - 3.53212709780093e-11*y345*y403 + 4.41515887225116e-11*y345*y405 + y356*y401*y418 - y380*y80 + y382*y383*(y14*(2.0*y15 + 2.0*y16 + 2.0*y17) + y375 + y6*(4.0*y2 + 4.0*y3 + 4.0*y4) + y8*(2.0*y10 + 2.0*y11 + 2.0*y9)) + 2.69825540021952e-16*y385*y386 + 2.40618160283239e-32*y390*y391 - 1.48072714020455e-32*y391*y392 - 1.97563438385653e-43*y395*y397 + 1.72354696230011e-21*y395*y411 + 1.21577500545017e-43*y397*y398 - 1.10501271569469e-58*y397*y399 - 1.06064428449238e-21*y398*y411 + 9.64015065237337e-37*y399*y411 + y402*y404 - y402*y406 + y404*y417 - y406*y417 - y408*y410 - y408*y415 - y410*y419 - y415*y419 + y92*(y25*(y136*(1.125 - 4.5*nu) + y59*(-10.5*nu - 5.625) + y64*(4.5 - y379)) - y258*y36 + y43*(-1.25*y138 + y299*(-8050.0*nu + y264 + 1875.0) + y300*(-5450.0*nu + 760.0*y39 + 1095.0))) + 2.88925109089694e-20*y317*y393/y195**4 + 4.07287405245073e-17*y340*y412/y170**6 - 32.0*y381*y383*(y0 + 0.25*y21 + 0.25*y296)**2/y126**3 + 6.10931107867609e-18*y200*y228/r**11 + 2.31644711733135e-17*y303/r**15) - y256*y357*y362 - y257*(y20*y36*(12.0*y91 + 6.0) + y25*y359) + y287*(y25*(y359 - 12.0*y58) + y363*(3.75*y245 + 0.015625*y364 + y365*(17490.0*nu - 4920.0*y39 - 2565.0)) + y363*(0.03125*y215*(-405.0*nu + 420.0*y39 - 45.0) - 36.5625*y237 - y366*(105.0*nu + 135.0)) + y378 + y43*(0.25*y136*(-y367 - 15.0) + y298*(y261 - 155.0) + 41.25*y70) + y43*(22.5*y13 + 0.5*y135*(y367 + 5.0) + 22.5*y18 - 25.0*y68 + 22.5*y7))) - y74*(y72 + y73)*(pphi*y84 + pphi*y88 + pphi*y89 - y31*y76 + y54*y86) + (-6.0*r*y74 + 18.0*(0.333333333333333*y13 + 0.333333333333333*y18 + 0.333333333333333*y7 + y71)**2/y22**3)*(pphi*y94 + pphi*y97 + pphi*y99 + y31*y90 + y54*y95)
 * 
 *         cdef double d2Heffdphi2 = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double d2Heffdpr2 = y247*y424*y425*(1769.32503588175*y102*y205 - 339.525257375589*y103*y289 + 6.82947265356779*y104*y109 + 56.0*y104*y432 + 83.0341915336443*y105*y106 + 1.52732776966902e-18*y205*y434 + y426*y427 + y426*y428 + y426*y433 + y429*y430 + y429*y431 + y436) - 5129029357728.48*y450*(5.74938229854254e-11*y229*y449 + y427*y445 + y428*y445 + y430*y446 + y431*y446 + y433*y445 + y438 - 0.115137213510253*y440 + 0.020112800850135*y443 + 0.00165425616626294*y444 + 0.0135644946594222*y447 + 8.63226223952612e-22*y448)**2
 */
  __Pyx_TraceLine(1249,0,__PYX_ERR(1, 1249, __pyx_L1_error))
  __pyx_v_d2Heffdphi2 = 0.0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1251
 *         cdef double d2Heffdphi2 = 0
 * 
 *         cdef double d2Heffdpr2 = y247*y424*y425*(1769.32503588175*y102*y205 - 339.525257375589*y103*y289 + 6.82947265356779*y104*y109 + 56.0*y104*y432 + 83.0341915336443*y105*y106 + 1.52732776966902e-18*y205*y434 + y426*y427 + y426*y428 + y426*y433 + y429*y430 + y429*y431 + y436) - 5129029357728.48*y450*(5.74938229854254e-11*y229*y449 + y427*y445 + y428*y445 + y430*y446 + y431*y446 + y433*y445 + y438 - 0.115137213510253*y440 + 0.020112800850135*y443 + 0.00165425616626294*y444 + 0.0135644946594222*y447 + 8.63226223952612e-22*y448)**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double d2Heffdpphi2 = y23*(6.0*y30*y32*y65 + y30*(y32*(-4.171875*nu - 4.14322916666667*y39 + 0.4609375) + y451*y452 + y77*(-2.8125*nu - 0.9375)) + y458*(2.0*y27 + 2.0*y28) + 2.0*y459*y460 - 0.75*y54*y96 + y54*(y32*(-0.546875*nu - 1.59635416666667*y39 - 0.4609375) + y452*y453 + y77*(0.9375 - 0.5625*nu))) + y247*y424*y425*(y130*(-2.0*r - 4.0) + y77*(-y132 + y32*(y133 + y134 + y137) + y36*(-0.125*y138 + 0.00520833333333333*y140*y59 + y142*y365) + 2.0)) - 58982400.0*y450*(pphi*y144*y77 - y130*y461)**2
 */
  __Pyx_TraceLine(1251,0,__PYX_ERR(1, 1251, __pyx_L1_error))
  __pyx_v_d2Heffdpr2 = ((((__pyx_v_y247 * __pyx_v_y424) * __pyx_v_y425) * (((((((((((((1769.32503588175 * __pyx_v_y102) * __pyx_v_y205) - ((339.525257375589 * __pyx_v_y103) * __pyx_v_y289)) + ((6.82947265356779 * __pyx_v_y104) * __pyx_v_y109)) + ((56.0 * __pyx_v_y104) * __pyx_v_y432)) + ((83.0341915336443 * __pyx_v_y105) * __pyx_v_y106)) + ((1.52732776966902e-18 * __pyx_v_y205) * __pyx_v_y434)) + (__pyx_v_y426 * __pyx_v_y427)) + (__pyx_v_y426 * __pyx_v_y428)) + (__pyx_v_y426 * __pyx_v_y433)) + (__pyx_v_y429 * __pyx_v_y430)) + (__pyx_v_y429 * __pyx_v_y431)) + __pyx_v_y436)) - ((5129029357728.48 * __pyx_v_y450) * pow((((((((((((((5.74938229854254e-11 * __pyx_v_y229) * __pyx_v_y449) + (__pyx_v_y427 * __pyx_v_y445)) + (__pyx_v_y428 * __pyx_v_y445)) + (__pyx_v_y430 * __pyx_v_y446)) + (__pyx_v_y431 * __pyx_v_y446)) + (__pyx_v_y433 * __pyx_v_y445)) + __pyx_v_y438) - (0.115137213510253 * __pyx_v_y440)) + (0.020112800850135 * __pyx_v_y443)) + (0.00165425616626294 * __pyx_v_y444)) + (0.0135644946594222 * __pyx_v_y447)) + (8.63226223952612e-22 * __pyx_v_y448)), 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1253
 *         cdef double d2Heffdpr2 = y247*y424*y425*(1769.32503588175*y102*y205 - 339.525257375589*y103*y289 + 6.82947265356779*y104*y109 + 56.0*y104*y432 + 83.0341915336443*y105*y106 + 1.52732776966902e-18*y205*y434 + y426*y427 + y426*y428 + y426*y433 + y429*y430 + y429*y431 + y436) - 5129029357728.48*y450*(5.74938229854254e-11*y229*y449 + y427*y445 + y428*y445 + y430*y446 + y431*y446 + y433*y445 + y438 - 0.115137213510253*y440 + 0.020112800850135*y443 + 0.00165425616626294*y444 + 0.0135644946594222*y447 + 8.63226223952612e-22*y448)**2
 * 
 *         cdef double d2Heffdpphi2 = y23*(6.0*y30*y32*y65 + y30*(y32*(-4.171875*nu - 4.14322916666667*y39 + 0.4609375) + y451*y452 + y77*(-2.8125*nu - 0.9375)) + y458*(2.0*y27 + 2.0*y28) + 2.0*y459*y460 - 0.75*y54*y96 + y54*(y32*(-0.546875*nu - 1.59635416666667*y39 - 0.4609375) + y452*y453 + y77*(0.9375 - 0.5625*nu))) + y247*y424*y425*(y130*(-2.0*r - 4.0) + y77*(-y132 + y32*(y133 + y134 + y137) + y36*(-0.125*y138 + 0.00520833333333333*y140*y59 + y142*y365) + 2.0)) - 58982400.0*y450*(pphi*y144*y77 - y130*y461)**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double  d2Heffdrdphi = 0
 */
  __Pyx_TraceLine(1253,0,__PYX_ERR(1, 1253, __pyx_L1_error))
  __pyx_v_d2Heffdpphi2 = (((__pyx_v_y23 * ((((((((6.0 * __pyx_v_y30) * __pyx_v_y32) * __pyx_v_y65) + (__pyx_v_y30 * (((__pyx_v_y32 * (((-4.171875 * __pyx_v_nu) - (4.14322916666667 * __pyx_v_y39)) + 0.4609375)) + (__pyx_v_y451 * __pyx_v_y452)) + (__pyx_v_y77 * ((-2.8125 * __pyx_v_nu) - 0.9375))))) + (__pyx_v_y458 * ((2.0 * __pyx_v_y27) + (2.0 * __pyx_v_y28)))) + ((2.0 * __pyx_v_y459) * __pyx_v_y460)) - ((0.75 * __pyx_v_y54) * __pyx_v_y96)) + (__pyx_v_y54 * (((__pyx_v_y32 * (((-0.546875 * __pyx_v_nu) - (1.59635416666667 * __pyx_v_y39)) - 0.4609375)) + (__pyx_v_y452 * __pyx_v_y453)) + (__pyx_v_y77 * (0.9375 - (0.5625 * __pyx_v_nu))))))) + (((__pyx_v_y247 * __pyx_v_y424) * __pyx_v_y425) * ((__pyx_v_y130 * ((-2.0 * __pyx_v_r) - 4.0)) + (__pyx_v_y77 * ((((-__pyx_v_y132) + (__pyx_v_y32 * ((__pyx_v_y133 + __pyx_v_y134) + __pyx_v_y137))) + (__pyx_v_y36 * (((-0.125 * __pyx_v_y138) + ((0.00520833333333333 * __pyx_v_y140) * __pyx_v_y59)) + (__pyx_v_y142 * __pyx_v_y365)))) + 2.0))))) - ((58982400.0 * __pyx_v_y450) * pow((((__pyx_v_pphi * __pyx_v_y144) * __pyx_v_y77) - (__pyx_v_y130 * __pyx_v_y461)), 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1255
 *         cdef double d2Heffdpphi2 = y23*(6.0*y30*y32*y65 + y30*(y32*(-4.171875*nu - 4.14322916666667*y39 + 0.4609375) + y451*y452 + y77*(-2.8125*nu - 0.9375)) + y458*(2.0*y27 + 2.0*y28) + 2.0*y459*y460 - 0.75*y54*y96 + y54*(y32*(-0.546875*nu - 1.59635416666667*y39 - 0.4609375) + y452*y453 + y77*(0.9375 - 0.5625*nu))) + y247*y424*y425*(y130*(-2.0*r - 4.0) + y77*(-y132 + y32*(y133 + y134 + y137) + y36*(-0.125*y138 + 0.00520833333333333*y140*y59 + y142*y365) + 2.0)) - 58982400.0*y450*(pphi*y144*y77 - y130*y461)**2
 * 
 *         cdef double  d2Heffdrdphi = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double d2Heffdrdpr = 0.5*y424*(y248*(237.667680162912*nu*y101*y439 - 2653.98755382262*nu*y288*y437 + 1.35633680555556e-7*prst*y202*y206*y210*y220*y225*y36 + 3.39084201388889e-8*prst*y206*y210*y220*y228*y25*y356 + 3.39084201388889e-8*prst*y206*y210*y225*y228*y25*y323 - prst*y210*y226*y229*y414 - prst*y229*y325*y409 - 29.655068404873*y103*y442 - 8.0*y111*y32*y437 - 12.0*y114*y36*y437 - 12.0*y116*y32*y439 - 18.0*y119*y36*y439 - 16.0*y121*y32*y441 - y124*y437*y467 + 4.07287405245073e-18*y171*y186*y196*y198*y304*y305*y437 - 3.70181785051137e-33*y171*y186*y198*y399*y437 + 3.03943751362543e-44*y171*y196*y198*y201*y203*y220*y323*y437 - 2.65161071123094e-22*y186*y336*y338*y437 - 6.61842033523243e-18*y199*y390*y437 - 0.975638950243592*y290*y441 + 4.0*y293*y36*y437 - y406*y449) - y464*y465 + y464*y466) - y464*y468
 */
  __Pyx_TraceLine(1255,0,__PYX_ERR(1, 1255, __pyx_L1_error))
  __pyx_v_d2Heffdrdphi = 0.0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1257
 *         cdef double  d2Heffdrdphi = 0
 * 
 *         cdef double d2Heffdrdpr = 0.5*y424*(y248*(237.667680162912*nu*y101*y439 - 2653.98755382262*nu*y288*y437 + 1.35633680555556e-7*prst*y202*y206*y210*y220*y225*y36 + 3.39084201388889e-8*prst*y206*y210*y220*y228*y25*y356 + 3.39084201388889e-8*prst*y206*y210*y225*y228*y25*y323 - prst*y210*y226*y229*y414 - prst*y229*y325*y409 - 29.655068404873*y103*y442 - 8.0*y111*y32*y437 - 12.0*y114*y36*y437 - 12.0*y116*y32*y439 - 18.0*y119*y36*y439 - 16.0*y121*y32*y441 - y124*y437*y467 + 4.07287405245073e-18*y171*y186*y196*y198*y304*y305*y437 - 3.70181785051137e-33*y171*y186*y198*y399*y437 + 3.03943751362543e-44*y171*y196*y198*y201*y203*y220*y323*y437 - 2.65161071123094e-22*y186*y336*y338*y437 - 6.61842033523243e-18*y199*y390*y437 - 0.975638950243592*y290*y441 + 4.0*y293*y36*y437 - y406*y449) - y464*y465 + y464*y466) - y464*y468             # <<<<<<<<<<<<<<
 * 
 *         cdef double d2Heffdrdpphi = y23*(-y29*y37*y66 + y30*(-y33*y470 - y40*y471 - y46*y472) + 0.75*y36*y473*y59 + y460*y86 - y469*y76 + y54*(-y470*y48 - y471*y49 - y472*y50) + y84 + y88 + y89) + 0.5*y424*(y248*(-pphi*y474 - y144*y470 + y232*y384*y461 + y301*y454 + y380*y454) - y465*y476 + y466*y476) - y468*y476 - y74*(y20 + 3.0*y71)*(2.0*y29*y98 + y30*y458 + y459*y54 + y460*y95 + y469*y90 - 0.25*y473*y96 + y94 + y97 + y99)
 */
  __Pyx_TraceLine(1257,0,__PYX_ERR(1, 1257, __pyx_L1_error))
  __pyx_v_d2Heffdrdpr = (((0.5 * __pyx_v_y424) * (((__pyx_v_y248 * ((((((((((((((((((((((((237.667680162912 * __pyx_v_nu) * __pyx_v_y101) * __pyx_v_y439) - (((2653.98755382262 * __pyx_v_nu) * __pyx_v_y288) * __pyx_v_y437)) + (((((((1.35633680555556e-7 * __pyx_v_prst) * __pyx_v_y202) * __pyx_v_y206) * __pyx_v_y210) * __pyx_v_y220) * __pyx_v_y225) * __pyx_v_y36)) + (((((((3.39084201388889e-8 * __pyx_v_prst) * __pyx_v_y206) * __pyx_v_y210) * __pyx_v_y220) * __pyx_v_y228) * __pyx_v_y25) * __pyx_v_y356)) + (((((((3.39084201388889e-8 * __pyx_v_prst) * __pyx_v_y206) * __pyx_v_y210) * __pyx_v_y225) * __pyx_v_y228) * __pyx_v_y25) * __pyx_v_y323)) - ((((__pyx_v_prst * __pyx_v_y210) * __pyx_v_y226) * __pyx_v_y229) * __pyx_v_y414)) - (((__pyx_v_prst * __pyx_v_y229) * __pyx_v_y325) * __pyx_v_y409)) - ((29.655068404873 * __pyx_v_y103) * __pyx_v_y442)) - (((8.0 * __pyx_v_y111) * __pyx_v_y32) * __pyx_v_y437)) - (((12.0 * __pyx_v_y114) * __pyx_v_y36) * __pyx_v_y437)) - (((12.0 * __pyx_v_y116) * __pyx_v_y32) * __pyx_v_y439)) - (((18.0 * __pyx_v_y119) * __pyx_v_y36) * __pyx_v_y439)) - (((16.0 * __pyx_v_y121) * __pyx_v_y32) * __pyx_v_y441)) - ((__pyx_v_y124 * __pyx_v_y437) * __pyx_v_y467)) + (((((((4.07287405245073e-18 * __pyx_v_y171) * __pyx_v_y186) * __pyx_v_y196) * __pyx_v_y198) * __pyx_v_y304) * __pyx_v_y305) * __pyx_v_y437)) - (((((3.70181785051137e-33 * __pyx_v_y171) * __pyx_v_y186) * __pyx_v_y198) * __pyx_v_y399) * __pyx_v_y437)) + ((((((((3.03943751362543e-44 * __pyx_v_y171) * __pyx_v_y196) * __pyx_v_y198) * __pyx_v_y201) * __pyx_v_y203) * __pyx_v_y220) * __pyx_v_y323) * __pyx_v_y437)) - ((((2.65161071123094e-22 * __pyx_v_y186) * __pyx_v_y336) * __pyx_v_y338) * __pyx_v_y437)) - (((6.61842033523243e-18 * __pyx_v_y199) * __pyx_v_y390) * __pyx_v_y437)) - ((0.975638950243592 * __pyx_v_y290) * __pyx_v_y441)) + (((4.0 * __pyx_v_y293) * __pyx_v_y36) * __pyx_v_y437)) - (__pyx_v_y406 * __pyx_v_y449))) - (__pyx_v_y464 * __pyx_v_y465)) + (__pyx_v_y464 * __pyx_v_y466))) - (__pyx_v_y464 * __pyx_v_y468));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1259
 *         cdef double d2Heffdrdpr = 0.5*y424*(y248*(237.667680162912*nu*y101*y439 - 2653.98755382262*nu*y288*y437 + 1.35633680555556e-7*prst*y202*y206*y210*y220*y225*y36 + 3.39084201388889e-8*prst*y206*y210*y220*y228*y25*y356 + 3.39084201388889e-8*prst*y206*y210*y225*y228*y25*y323 - prst*y210*y226*y229*y414 - prst*y229*y325*y409 - 29.655068404873*y103*y442 - 8.0*y111*y32*y437 - 12.0*y114*y36*y437 - 12.0*y116*y32*y439 - 18.0*y119*y36*y439 - 16.0*y121*y32*y441 - y124*y437*y467 + 4.07287405245073e-18*y171*y186*y196*y198*y304*y305*y437 - 3.70181785051137e-33*y171*y186*y198*y399*y437 + 3.03943751362543e-44*y171*y196*y198*y201*y203*y220*y323*y437 - 2.65161071123094e-22*y186*y336*y338*y437 - 6.61842033523243e-18*y199*y390*y437 - 0.975638950243592*y290*y441 + 4.0*y293*y36*y437 - y406*y449) - y464*y465 + y464*y466) - y464*y468
 * 
 *         cdef double d2Heffdrdpphi = y23*(-y29*y37*y66 + y30*(-y33*y470 - y40*y471 - y46*y472) + 0.75*y36*y473*y59 + y460*y86 - y469*y76 + y54*(-y470*y48 - y471*y49 - y472*y50) + y84 + y88 + y89) + 0.5*y424*(y248*(-pphi*y474 - y144*y470 + y232*y384*y461 + y301*y454 + y380*y454) - y465*y476 + y466*y476) - y468*y476 - y74*(y20 + 3.0*y71)*(2.0*y29*y98 + y30*y458 + y459*y54 + y460*y95 + y469*y90 - 0.25*y473*y96 + y94 + y97 + y99)             # <<<<<<<<<<<<<<
 * 
 *         cdef double d2Heffdphidpr = 0
 */
  __Pyx_TraceLine(1259,0,__PYX_ERR(1, 1259, __pyx_L1_error))
  __pyx_v_d2Heffdrdpphi = ((((__pyx_v_y23 * (((((((((((-__pyx_v_y29) * __pyx_v_y37) * __pyx_v_y66) + (__pyx_v_y30 * ((((-__pyx_v_y33) * __pyx_v_y470) - (__pyx_v_y40 * __pyx_v_y471)) - (__pyx_v_y46 * __pyx_v_y472)))) + (((0.75 * __pyx_v_y36) * __pyx_v_y473) * __pyx_v_y59)) + (__pyx_v_y460 * __pyx_v_y86)) - (__pyx_v_y469 * __pyx_v_y76)) + (__pyx_v_y54 * ((((-__pyx_v_y470) * __pyx_v_y48) - (__pyx_v_y471 * __pyx_v_y49)) - (__pyx_v_y472 * __pyx_v_y50)))) + __pyx_v_y84) + __pyx_v_y88) + __pyx_v_y89)) + ((0.5 * __pyx_v_y424) * (((__pyx_v_y248 * ((((((-__pyx_v_pphi) * __pyx_v_y474) - (__pyx_v_y144 * __pyx_v_y470)) + ((__pyx_v_y232 * __pyx_v_y384) * __pyx_v_y461)) + (__pyx_v_y301 * __pyx_v_y454)) + (__pyx_v_y380 * __pyx_v_y454))) - (__pyx_v_y465 * __pyx_v_y476)) + (__pyx_v_y466 * __pyx_v_y476)))) - (__pyx_v_y468 * __pyx_v_y476)) - ((__pyx_v_y74 * (__pyx_v_y20 + (3.0 * __pyx_v_y71))) * ((((((((((2.0 * __pyx_v_y29) * __pyx_v_y98) + (__pyx_v_y30 * __pyx_v_y458)) + (__pyx_v_y459 * __pyx_v_y54)) + (__pyx_v_y460 * __pyx_v_y95)) + (__pyx_v_y469 * __pyx_v_y90)) - ((0.25 * __pyx_v_y473) * __pyx_v_y96)) + __pyx_v_y94) + __pyx_v_y97) + __pyx_v_y99)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1261
 *         cdef double d2Heffdrdpphi = y23*(-y29*y37*y66 + y30*(-y33*y470 - y40*y471 - y46*y472) + 0.75*y36*y473*y59 + y460*y86 - y469*y76 + y54*(-y470*y48 - y471*y49 - y472*y50) + y84 + y88 + y89) + 0.5*y424*(y248*(-pphi*y474 - y144*y470 + y232*y384*y461 + y301*y454 + y380*y454) - y465*y476 + y466*y476) - y468*y476 - y74*(y20 + 3.0*y71)*(2.0*y29*y98 + y30*y458 + y459*y54 + y460*y95 + y469*y90 - 0.25*y473*y96 + y94 + y97 + y99)
 * 
 *         cdef double d2Heffdphidpr = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double d2Heffdphidpphi = 0
 */
  __Pyx_TraceLine(1261,0,__PYX_ERR(1, 1261, __pyx_L1_error))
  __pyx_v_d2Heffdphidpr = 0.0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1263
 *         cdef double d2Heffdphidpr = 0
 * 
 *         cdef double d2Heffdphidpphi = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double d2Heffdprdpphi = 14745600.0*y450*y464*y475
 */
  __Pyx_TraceLine(1263,0,__PYX_ERR(1, 1263, __pyx_L1_error))
  __pyx_v_d2Heffdphidpphi = 0.0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1265
 *         cdef double d2Heffdphidpphi = 0
 * 
 *         cdef double d2Heffdprdpphi = 14745600.0*y450*y464*y475             # <<<<<<<<<<<<<<
 * 
 *         # Compute H Hessian
 */
  __Pyx_TraceLine(1265,0,__PYX_ERR(1, 1265, __pyx_L1_error))
  __pyx_v_d2Heffdprdpphi = (((14745600.0 * __pyx_v_y450) * __pyx_v_y464) * __pyx_v_y475);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1268
 * 
 *         # Compute H Hessian
 *         cdef double d2Hdr2 = (-(dHeffdr**2/H**3)*(M**2/nu) + d2Heffdr2/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdphi2 = (-(dHeffdphi**2/H**3)*(M**2/nu) + d2Heffdphi2/H) * M*M / nu
 *         cdef double d2Hdpr2 = (-(dHeffdpr**2/H**3)*(M**2/nu) + d2Heffdpr2/H) * M*M / nu
 */
  __Pyx_TraceLine(1268,0,__PYX_ERR(1, 1268, __pyx_L1_error))
  __pyx_v_d2Hdr2 = ((((((-(pow(__pyx_v_dHeffdr, 2.0) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdr2 / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1269
 *         # Compute H Hessian
 *         cdef double d2Hdr2 = (-(dHeffdr**2/H**3)*(M**2/nu) + d2Heffdr2/H) * M*M / nu
 *         cdef double d2Hdphi2 = (-(dHeffdphi**2/H**3)*(M**2/nu) + d2Heffdphi2/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdpr2 = (-(dHeffdpr**2/H**3)*(M**2/nu) + d2Heffdpr2/H) * M*M / nu
 *         cdef double d2Hdpphi2 = (-(dHeffdpphi**2/H**3)*(M**2/nu) + d2Heffdpphi2/H) * M*M / nu
 */
  __Pyx_TraceLine(1269,0,__PYX_ERR(1, 1269, __pyx_L1_error))
  __pyx_v_d2Hdphi2 = ((((((-(pow(__pyx_v_dHeffdphi, 2.0) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdphi2 / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1270
 *         cdef double d2Hdr2 = (-(dHeffdr**2/H**3)*(M**2/nu) + d2Heffdr2/H) * M*M / nu
 *         cdef double d2Hdphi2 = (-(dHeffdphi**2/H**3)*(M**2/nu) + d2Heffdphi2/H) * M*M / nu
 *         cdef double d2Hdpr2 = (-(dHeffdpr**2/H**3)*(M**2/nu) + d2Heffdpr2/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdpphi2 = (-(dHeffdpphi**2/H**3)*(M**2/nu) + d2Heffdpphi2/H) * M*M / nu
 *         cdef double d2Hdrdphi = (-(dHeffdr*dHeffdphi/H**3)*(M**2/nu) + d2Heffdrdphi/H) * M*M / nu
 */
  __Pyx_TraceLine(1270,0,__PYX_ERR(1, 1270, __pyx_L1_error))
  __pyx_v_d2Hdpr2 = ((((((-(pow(__pyx_v_dHeffdpr, 2.0) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdpr2 / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1271
 *         cdef double d2Hdphi2 = (-(dHeffdphi**2/H**3)*(M**2/nu) + d2Heffdphi2/H) * M*M / nu
 *         cdef double d2Hdpr2 = (-(dHeffdpr**2/H**3)*(M**2/nu) + d2Heffdpr2/H) * M*M / nu
 *         cdef double d2Hdpphi2 = (-(dHeffdpphi**2/H**3)*(M**2/nu) + d2Heffdpphi2/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdrdphi = (-(dHeffdr*dHeffdphi/H**3)*(M**2/nu) + d2Heffdrdphi/H) * M*M / nu
 *         cdef double d2Hdrdpr = (-(dHeffdr*dHeffdpr/H**3)*(M**2/nu) + d2Heffdrdpr/H) * M*M / nu
 */
  __Pyx_TraceLine(1271,0,__PYX_ERR(1, 1271, __pyx_L1_error))
  __pyx_v_d2Hdpphi2 = ((((((-(pow(__pyx_v_dHeffdpphi, 2.0) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdpphi2 / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1272
 *         cdef double d2Hdpr2 = (-(dHeffdpr**2/H**3)*(M**2/nu) + d2Heffdpr2/H) * M*M / nu
 *         cdef double d2Hdpphi2 = (-(dHeffdpphi**2/H**3)*(M**2/nu) + d2Heffdpphi2/H) * M*M / nu
 *         cdef double d2Hdrdphi = (-(dHeffdr*dHeffdphi/H**3)*(M**2/nu) + d2Heffdrdphi/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdrdpr = (-(dHeffdr*dHeffdpr/H**3)*(M**2/nu) + d2Heffdrdpr/H) * M*M / nu
 *         cdef double d2Hdrdpphi = (-(dHeffdr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdrdpphi/H) * M*M / nu
 */
  __Pyx_TraceLine(1272,0,__PYX_ERR(1, 1272, __pyx_L1_error))
  __pyx_v_d2Hdrdphi = ((((((-((__pyx_v_dHeffdr * __pyx_v_dHeffdphi) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdrdphi / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1273
 *         cdef double d2Hdpphi2 = (-(dHeffdpphi**2/H**3)*(M**2/nu) + d2Heffdpphi2/H) * M*M / nu
 *         cdef double d2Hdrdphi = (-(dHeffdr*dHeffdphi/H**3)*(M**2/nu) + d2Heffdrdphi/H) * M*M / nu
 *         cdef double d2Hdrdpr = (-(dHeffdr*dHeffdpr/H**3)*(M**2/nu) + d2Heffdrdpr/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdrdpphi = (-(dHeffdr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdrdpphi/H) * M*M / nu
 *         cdef double d2Hdphidpr = (-(dHeffdphi*dHeffdpr/H**3)*(M**2/nu) + d2Heffdphidpr/H) * M*M / nu
 */
  __Pyx_TraceLine(1273,0,__PYX_ERR(1, 1273, __pyx_L1_error))
  __pyx_v_d2Hdrdpr = ((((((-((__pyx_v_dHeffdr * __pyx_v_dHeffdpr) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdrdpr / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1274
 *         cdef double d2Hdrdphi = (-(dHeffdr*dHeffdphi/H**3)*(M**2/nu) + d2Heffdrdphi/H) * M*M / nu
 *         cdef double d2Hdrdpr = (-(dHeffdr*dHeffdpr/H**3)*(M**2/nu) + d2Heffdrdpr/H) * M*M / nu
 *         cdef double d2Hdrdpphi = (-(dHeffdr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdrdpphi/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdphidpr = (-(dHeffdphi*dHeffdpr/H**3)*(M**2/nu) + d2Heffdphidpr/H) * M*M / nu
 *         cdef double d2Hdphidpphi = (-(dHeffdphi*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdphidpphi/H) * M*M / nu
 */
  __Pyx_TraceLine(1274,0,__PYX_ERR(1, 1274, __pyx_L1_error))
  __pyx_v_d2Hdrdpphi = ((((((-((__pyx_v_dHeffdr * __pyx_v_dHeffdpphi) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdrdpphi / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1275
 *         cdef double d2Hdrdpr = (-(dHeffdr*dHeffdpr/H**3)*(M**2/nu) + d2Heffdrdpr/H) * M*M / nu
 *         cdef double d2Hdrdpphi = (-(dHeffdr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdrdpphi/H) * M*M / nu
 *         cdef double d2Hdphidpr = (-(dHeffdphi*dHeffdpr/H**3)*(M**2/nu) + d2Heffdphidpr/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdphidpphi = (-(dHeffdphi*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdphidpphi/H) * M*M / nu
 *         cdef double d2Hdprdpphi = (-(dHeffdpr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdprdpphi/H) * M*M / nu
 */
  __Pyx_TraceLine(1275,0,__PYX_ERR(1, 1275, __pyx_L1_error))
  __pyx_v_d2Hdphidpr = ((((((-((__pyx_v_dHeffdphi * __pyx_v_dHeffdpr) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdphidpr / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1276
 *         cdef double d2Hdrdpphi = (-(dHeffdr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdrdpphi/H) * M*M / nu
 *         cdef double d2Hdphidpr = (-(dHeffdphi*dHeffdpr/H**3)*(M**2/nu) + d2Heffdphidpr/H) * M*M / nu
 *         cdef double d2Hdphidpphi = (-(dHeffdphi*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdphidpphi/H) * M*M / nu             # <<<<<<<<<<<<<<
 *         cdef double d2Hdprdpphi = (-(dHeffdpr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdprdpphi/H) * M*M / nu
 * 
 */
  __Pyx_TraceLine(1276,0,__PYX_ERR(1, 1276, __pyx_L1_error))
  __pyx_v_d2Hdphidpphi = ((((((-((__pyx_v_dHeffdphi * __pyx_v_dHeffdpphi) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdphidpphi / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1277
 *         cdef double d2Hdphidpr = (-(dHeffdphi*dHeffdpr/H**3)*(M**2/nu) + d2Heffdphidpr/H) * M*M / nu
 *         cdef double d2Hdphidpphi = (-(dHeffdphi*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdphidpphi/H) * M*M / nu
 *         cdef double d2Hdprdpphi = (-(dHeffdpr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdprdpphi/H) * M*M / nu             # <<<<<<<<<<<<<<
 * 
 *         return np.array([d2Hdr2, d2Hdrdphi, d2Hdrdpr, d2Hdrdpphi, d2Hdrdphi, d2Hdphi2, d2Hdphidpr, d2Hdphidpphi, d2Hdrdpr, d2Hdphidpr, d2Hdpr2, d2Hdprdpphi, d2Hdrdpphi, d2Hdphidpphi, d2Hdprdpphi, d2Hdpphi2]).reshape(4, 4)
 */
  __Pyx_TraceLine(1277,0,__PYX_ERR(1, 1277, __pyx_L1_error))
  __pyx_v_d2Hdprdpphi = ((((((-((__pyx_v_dHeffdpr * __pyx_v_dHeffdpphi) / pow(__pyx_v_H, 3.0))) * (pow(__pyx_v_M, 2.0) / __pyx_v_nu)) + (__pyx_v_d2Heffdprdpphi / __pyx_v_H)) * __pyx_v_M) * __pyx_v_M) / __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1279
 *         cdef double d2Hdprdpphi = (-(dHeffdpr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdprdpphi/H) * M*M / nu
 * 
 *         return np.array([d2Hdr2, d2Hdrdphi, d2Hdrdpr, d2Hdrdpphi, d2Hdrdphi, d2Hdphi2, d2Hdphidpr, d2Hdphidpphi, d2Hdrdpr, d2Hdphidpr, d2Hdpr2, d2Hdprdpphi, d2Hdrdpphi, d2Hdphidpphi, d2Hdprdpphi, d2Hdpphi2]).reshape(4, 4)             # <<<<<<<<<<<<<<
 * 
 *     cdef double xi(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 */
  __Pyx_TraceLine(1279,0,__PYX_ERR(1, 1279, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_d2Hdr2); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_d2Hdrdphi); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_d2Hdrdpr); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = PyFloat_FromDouble(__pyx_v_d2Hdrdpphi); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyFloat_FromDouble(__pyx_v_d2Hdrdphi); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_16 = PyFloat_FromDouble(__pyx_v_d2Hdphi2); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_d2Hdphidpr); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_d2Hdphidpphi); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_d2Hdrdpr); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_d2Hdphidpr); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_d2Hdpr2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_d2Hdprdpphi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_14 = PyFloat_FromDouble(__pyx_v_d2Hdrdpphi); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_20 = PyFloat_FromDouble(__pyx_v_d2Hdphidpphi); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = PyFloat_FromDouble(__pyx_v_d2Hdprdpphi); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = PyFloat_FromDouble(__pyx_v_d2Hdpphi2); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = PyList_New(16); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_7);
  PyList_SET_ITEM(__pyx_t_23, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_9);
  PyList_SET_ITEM(__pyx_t_23, 1, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyList_SET_ITEM(__pyx_t_23, 2, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyList_SET_ITEM(__pyx_t_23, 3, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyList_SET_ITEM(__pyx_t_23, 4, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_16);
  PyList_SET_ITEM(__pyx_t_23, 5, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_13);
  PyList_SET_ITEM(__pyx_t_23, 6, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_23, 7, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_23, 8, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_23, 9, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_23, 10, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_23, 11, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_14);
  PyList_SET_ITEM(__pyx_t_23, 12, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_20);
  PyList_SET_ITEM(__pyx_t_23, 13, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_21);
  PyList_SET_ITEM(__pyx_t_23, 14, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_22);
  PyList_SET_ITEM(__pyx_t_23, 15, __pyx_t_22);
  __pyx_t_7 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_16 = 0;
  __pyx_t_13 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_14 = 0;
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_22 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_22)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_22);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_1 = (__pyx_t_22) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_22, __pyx_t_23) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_23);
  __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_reshape); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":468
 *         return dHdr, dHdphi, dHdpr, dHdpphi
 * 
 *     cpdef hessian(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.hessian", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c_coeffs);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_9hessian(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_8hessian[] = "\n        Evaluate the Hessian of the Hamiltonian.\n\n        Args:\n          q (double[:]): Canonical positions (r,phi).\n          p (double[:]): Canonical momenta  (prstar,pphi).\n          chi1_v (double[:]): Dimensionless spin vector of the primary.\n          chi2_v (double[:]): Dimensionless spin vector of the secondary.\n          m_1 (double): Primary mass component.\n          m_2 (double): Secondary mass component.\n          chi_1 (double): Projection of chi1_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chi_2 (double): Projection of chi2_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chiL1 (double): Projection of chi1_v onto the orbital angular momentum unit vector (l).\n          chiL2 (double): Projection of chi2_v onto the orbital angular momentum unit vector (l).\n\n        Returns:\n           (np.array)  d2Hdr2, d2Hdrdphi, d2Hdrdpr, d2Hdrdpphi, d2Hdrdphi, d2Hdphi2, d2Hdphidpr, d2Hdphidpphi, d2Hdrdpr, d2Hdphidpr, d2Hdpr2, d2Hdprdpphi, d2Hdrdpphi, d2Hdphidpphi, d2Hdprdpphi, d2Hdpphi2\n\n        ";
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_9hessian(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_q = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi1_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi2_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_m_1;
  double __pyx_v_m_2;
  double __pyx_v_chi_1;
  double __pyx_v_chi_2;
  double __pyx_v_chiL1;
  double __pyx_v_chiL2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hessian (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_q,&__pyx_n_s_p,&__pyx_n_s_chi1_v,&__pyx_n_s_chi2_v,&__pyx_n_s_m_1,&__pyx_n_s_m_2,&__pyx_n_s_chi_1,&__pyx_n_s_chi_2,&__pyx_n_s_chiL1,&__pyx_n_s_chiL2,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_q)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 1); __PYX_ERR(1, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi1_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 2); __PYX_ERR(1, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi2_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 3); __PYX_ERR(1, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 4); __PYX_ERR(1, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 5); __PYX_ERR(1, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 6); __PYX_ERR(1, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 7); __PYX_ERR(1, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 8); __PYX_ERR(1, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, 9); __PYX_ERR(1, 468, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "hessian") < 0)) __PYX_ERR(1, 468, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_q = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_q.memview)) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_p.memview)) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_chi1_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi1_v.memview)) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_chi2_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi2_v.memview)) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_m_1 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_m_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_m_2 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_m_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_chi_1 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_chi_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_chi_2 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_chi_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_chiL1 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_chiL1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 468, __pyx_L3_error)
    __pyx_v_chiL2 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_chiL2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 468, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("hessian", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 468, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.hessian", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_8hessian(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_8hessian(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hessian", 0);
  __Pyx_TraceCall("hessian (wrapper)", __pyx_f[1], 468, 0, __PYX_ERR(1, 468, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 468, __pyx_L1_error) }
  if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 468, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 468, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 468, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_hessian(__pyx_v_self, __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.hessian", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_q, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi1_v, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi2_v, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1281
 *         return np.array([d2Hdr2, d2Hdrdphi, d2Hdrdpr, d2Hdrdpphi, d2Hdrdphi, d2Hdphi2, d2Hdphidpr, d2Hdphidpphi, d2Hdrdpr, d2Hdphidpr, d2Hdpr2, d2Hdprdpphi, d2Hdrdpphi, d2Hdphidpphi, d2Hdprdpphi, d2Hdpphi2]).reshape(4, 4)
 * 
 *     cdef double xi(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

static double __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_xi(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, CYTHON_UNUSED double __pyx_v_m_1, CYTHON_UNUSED double __pyx_v_m_2, CYTHON_UNUSED double __pyx_v_chi_1, CYTHON_UNUSED double __pyx_v_chi_2, CYTHON_UNUSED double __pyx_v_chiL1, CYTHON_UNUSED double __pyx_v_chiL2) {
  double __pyx_v_r;
  CYTHON_UNUSED double __pyx_v_phi;
  CYTHON_UNUSED double __pyx_v_prst;
  double __pyx_v_L;
  double __pyx_v_chix1;
  double __pyx_v_chiy1;
  double __pyx_v_chiz1;
  double __pyx_v_chix2;
  double __pyx_v_chiy2;
  double __pyx_v_chiz2;
  CYTHON_UNUSED double __pyx_v_pphi;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *__pyx_v_c_coeffs = 0;
  double __pyx_v_a6;
  CYTHON_UNUSED double __pyx_v_dSO;
  CYTHON_UNUSED double __pyx_v_M;
  double __pyx_v_nu;
  double __pyx_v_X_1;
  double __pyx_v_X_2;
  CYTHON_UNUSED double __pyx_v_delta;
  double __pyx_v_Dbpm;
  double __pyx_v_Apm;
  double __pyx_v_t2;
  double __pyx_v_t1;
  double __pyx_v_ap2;
  double __pyx_v_xi;
  double __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("xi", 0);
  __Pyx_TraceCall("xi", __pyx_f[1], 1281, 0, __PYX_ERR(1, 1281, __pyx_L1_error));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1304
 *         # Coordinate definitions
 * 
 *         cdef double r = q[0]             # <<<<<<<<<<<<<<
 *         cdef double phi = q[1]
 * 
 */
  __Pyx_TraceLine(1304,0,__PYX_ERR(1, 1304, __pyx_L1_error))
  __pyx_t_1 = 0;
  __pyx_v_r = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_1 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1305
 * 
 *         cdef double r = q[0]
 *         cdef double phi = q[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double prst = p[0]
 */
  __Pyx_TraceLine(1305,0,__PYX_ERR(1, 1305, __pyx_L1_error))
  __pyx_t_1 = 1;
  __pyx_v_phi = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_1 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1307
 *         cdef double phi = q[1]
 * 
 *         cdef double prst = p[0]             # <<<<<<<<<<<<<<
 *         cdef double L = p[1]
 * 
 */
  __Pyx_TraceLine(1307,0,__PYX_ERR(1, 1307, __pyx_L1_error))
  __pyx_t_1 = 0;
  __pyx_v_prst = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_1 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1308
 * 
 *         cdef double prst = p[0]
 *         cdef double L = p[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix1 = chi1_v[0]
 */
  __Pyx_TraceLine(1308,0,__PYX_ERR(1, 1308, __pyx_L1_error))
  __pyx_t_1 = 1;
  __pyx_v_L = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_1 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1310
 *         cdef double L = p[1]
 * 
 *         cdef double chix1 = chi1_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]
 */
  __Pyx_TraceLine(1310,0,__PYX_ERR(1, 1310, __pyx_L1_error))
  __pyx_t_1 = 0;
  __pyx_v_chix1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_1 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1311
 * 
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz1 = chi1_v[2]
 * 
 */
  __Pyx_TraceLine(1311,0,__PYX_ERR(1, 1311, __pyx_L1_error))
  __pyx_t_1 = 1;
  __pyx_v_chiy1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_1 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1312
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix2 = chi2_v[0]
 */
  __Pyx_TraceLine(1312,0,__PYX_ERR(1, 1312, __pyx_L1_error))
  __pyx_t_1 = 2;
  __pyx_v_chiz1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_1 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1314
 *         cdef double chiz1 = chi1_v[2]
 * 
 *         cdef double chix2 = chi2_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]
 */
  __Pyx_TraceLine(1314,0,__PYX_ERR(1, 1314, __pyx_L1_error))
  __pyx_t_1 = 0;
  __pyx_v_chix2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_1 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1315
 * 
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz2 = chi2_v[2]
 * 
 */
  __Pyx_TraceLine(1315,0,__PYX_ERR(1, 1315, __pyx_L1_error))
  __pyx_t_1 = 1;
  __pyx_v_chiy2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_1 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1316
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double pphi = L
 */
  __Pyx_TraceLine(1316,0,__PYX_ERR(1, 1316, __pyx_L1_error))
  __pyx_t_1 = 2;
  __pyx_v_chiz2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_1 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1318
 *         cdef double chiz2 = chi2_v[2]
 * 
 *         cdef double pphi = L             # <<<<<<<<<<<<<<
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 */
  __Pyx_TraceLine(1318,0,__PYX_ERR(1, 1318, __pyx_L1_error))
  __pyx_v_pphi = __pyx_v_L;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1320
 *         cdef double pphi = L
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs             # <<<<<<<<<<<<<<
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']
 */
  __Pyx_TraceLine(1320,0,__PYX_ERR(1, 1320, __pyx_L1_error))
  __pyx_t_2 = ((PyObject *)__pyx_v_self->calibration_coeffs);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_c_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1321
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']             # <<<<<<<<<<<<<<
 *         cdef double dSO = c_coeffs['dSO']
 * 
 */
  __Pyx_TraceLine(1321,0,__PYX_ERR(1, 1321, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_a6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_a6 = __pyx_t_3;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1322
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1322,0,__PYX_ERR(1, 1322, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_dSO); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_dSO = __pyx_t_3;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1327
 * 
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M             # <<<<<<<<<<<<<<
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 */
  __Pyx_TraceLine(1327,0,__PYX_ERR(1, 1327, __pyx_L1_error))
  __pyx_t_3 = __pyx_v_self->__pyx_base.EOBpars->p_params->M;
  __pyx_v_M = __pyx_t_3;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1328
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu             # <<<<<<<<<<<<<<
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 */
  __Pyx_TraceLine(1328,0,__PYX_ERR(1, 1328, __pyx_L1_error))
  __pyx_t_3 = __pyx_v_self->__pyx_base.EOBpars->p_params->nu;
  __pyx_v_nu = __pyx_t_3;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1329
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1             # <<<<<<<<<<<<<<
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 */
  __Pyx_TraceLine(1329,0,__PYX_ERR(1, 1329, __pyx_L1_error))
  __pyx_t_3 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_1;
  __pyx_v_X_1 = __pyx_t_3;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1330
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2             # <<<<<<<<<<<<<<
 *         cdef double delta = self.EOBpars.p_params.delta
 * 
 */
  __Pyx_TraceLine(1330,0,__PYX_ERR(1, 1330, __pyx_L1_error))
  __pyx_t_3 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_2;
  __pyx_v_X_2 = __pyx_t_3;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1331
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta             # <<<<<<<<<<<<<<
 * 
 *         # Actual Hamiltonian expressions
 */
  __Pyx_TraceLine(1331,0,__PYX_ERR(1, 1331, __pyx_L1_error))
  __pyx_t_3 = __pyx_v_self->__pyx_base.EOBpars->p_params->delta;
  __pyx_v_delta = __pyx_t_3;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1334
 * 
 *         # Actual Hamiltonian expressions
 *         cdef double Dbpm = r*(6730497718123.02*nu**3 + 133772083200.0*nu**2*r**2 + 1822680546449.21*nu**2*r + 80059249540278.2*nu**2 + 2589101062873.81*nu*r**2 + 10611661054566.2*nu*r - 12049908701745.2*nu + 5107745331375.71*r**2 - 326837426.241486*r*(14700.0*nu + 42911.0) - 39476764256925.6*r - (-5041721180160.0*nu**2 - 25392914995744.3*nu - 879923036160.0*r**2 - 283115520.0*r*(14700.0*nu + 42911.0) + 104186110149937.0)*log(r) + 5787938193408.0*log(r)**2 + 275059053208689.0)/(55296.0*nu*(14515200.0*nu**3 - 42636451.6032331*nu**2 - 2510664218.28128*nu + 1002013764.01019) - 967680.0*r**3*(-138240.0*nu**2 - 2675575.66847905*nu - 5278341.3229329) - 9216.0*r**2*(-197773496.793534*nu**2 - 630116198.873299*nu + 5805304367.87913) + r*(5927865218923.02*nu**3 + 43133561885859.3*nu**2 + 43393301259014.8*nu + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0) + 5787938193408.0*r*log(r)**2 + (-1698693120.0*nu*(11592.0*nu + 69847.0) + 879923036160.0*r**3 + 283115520.0*r**2*(14700.0*nu + 42911.0) + 49152.0*r*(102574080.0*nu**2 + 409207698.136075*nu - 2119671837.36038))*log(r))             # <<<<<<<<<<<<<<
 * 
 *         cdef double Apm = 7680.0*r**4*(-5416406.59541186*nu**2 + 28.0*nu*(1920.0*a6 + 733955.307463037) + 2048.0*nu*(756.0*nu + 336.0*r + 407.0)*log(r) - 7.0*r*(-185763.092693281*nu**2 + 938918.400156317*nu - 245760.0) - 3440640.0)/(241555486248.807*nu**4 + 1120.0*nu**3*(-17833256.898555*r**2 - 163683964.822551*r - 1188987459.03162) + 7.0*nu**2*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 1426660551.8844*r**5 - 3089250703.76879*r**4 - 6178501407.53758*r**3 + 2064783811.32587*r**2 + 122635399361.987*r + 276057889687.011) + 67645734912.0*nu**2*log(r)**2 + 53760.0*nu*(7680.0*a6*(r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 128.0*r*(-6852.34813868015*r**4 + 4264.6962773603*r**3 + 8529.39255472061*r**2 + 13218.7851094412*r - 33722.4297811176) + 113485.217444961*r*(-r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 148.04406601634*r*(349.0*r**4 + 1926.0*r**3 + 3852.0*r**2 + 7704.0*r + 36400.0)) + 32768.0*nu*(-1882456.23663972*nu**2 - 38842241.4769507*nu + 161280.0*r**5 + 480.0*r**4*(756.0*nu + 1079.0) + 960.0*r**3*(756.0*nu + 1079.0) + 1920.0*r**2*(588.0*nu + 1079.0) + 240.0*r*(-3024.0*nu**2 - 7466.27061066206*nu + 17264.0) + 13447680.0)*log(r) + 13212057600.0*r**5)
 */
  __Pyx_TraceLine(1334,0,__PYX_ERR(1, 1334, __pyx_L1_error))
  __pyx_v_Dbpm = ((__pyx_v_r * (((((((((((((6730497718123.02 * pow(__pyx_v_nu, 3.0)) + ((133772083200.0 * pow(__pyx_v_nu, 2.0)) * pow(__pyx_v_r, 2.0))) + ((1822680546449.21 * pow(__pyx_v_nu, 2.0)) * __pyx_v_r)) + (80059249540278.2 * pow(__pyx_v_nu, 2.0))) + ((2589101062873.81 * __pyx_v_nu) * pow(__pyx_v_r, 2.0))) + ((10611661054566.2 * __pyx_v_nu) * __pyx_v_r)) - (12049908701745.2 * __pyx_v_nu)) + (5107745331375.71 * pow(__pyx_v_r, 2.0))) - ((326837426.241486 * __pyx_v_r) * ((14700.0 * __pyx_v_nu) + 42911.0))) - (39476764256925.6 * __pyx_v_r)) - ((((((-5041721180160.0 * pow(__pyx_v_nu, 2.0)) - (25392914995744.3 * __pyx_v_nu)) - (879923036160.0 * pow(__pyx_v_r, 2.0))) - ((283115520.0 * __pyx_v_r) * ((14700.0 * __pyx_v_nu) + 42911.0))) + 104186110149937.0) * log(__pyx_v_r))) + (5787938193408.0 * pow(log(__pyx_v_r), 2.0))) + 275059053208689.0)) / (((((((55296.0 * __pyx_v_nu) * ((((14515200.0 * pow(__pyx_v_nu, 3.0)) - (42636451.6032331 * pow(__pyx_v_nu, 2.0))) - (2510664218.28128 * __pyx_v_nu)) + 1002013764.01019)) - ((967680.0 * pow(__pyx_v_r, 3.0)) * (((-138240.0 * pow(__pyx_v_nu, 2.0)) - (2675575.66847905 * __pyx_v_nu)) - 5278341.3229329))) - ((9216.0 * pow(__pyx_v_r, 2.0)) * (((-197773496.793534 * pow(__pyx_v_nu, 2.0)) - (630116198.873299 * __pyx_v_nu)) + 5805304367.87913))) + (__pyx_v_r * (((((5927865218923.02 * pow(__pyx_v_nu, 3.0)) + (43133561885859.3 * pow(__pyx_v_nu, 2.0))) + (43393301259014.8 * __pyx_v_nu)) + (86618264430493.3 * pow((1.0 - (0.496948781616935 * __pyx_v_nu)), 2.0))) + 188440788778196.0))) + ((5787938193408.0 * __pyx_v_r) * pow(log(__pyx_v_r), 2.0))) + ((((((-1698693120.0 * __pyx_v_nu) * ((11592.0 * __pyx_v_nu) + 69847.0)) + (879923036160.0 * pow(__pyx_v_r, 3.0))) + ((283115520.0 * pow(__pyx_v_r, 2.0)) * ((14700.0 * __pyx_v_nu) + 42911.0))) + ((49152.0 * __pyx_v_r) * (((102574080.0 * pow(__pyx_v_nu, 2.0)) + (409207698.136075 * __pyx_v_nu)) - 2119671837.36038))) * log(__pyx_v_r))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1336
 *         cdef double Dbpm = r*(6730497718123.02*nu**3 + 133772083200.0*nu**2*r**2 + 1822680546449.21*nu**2*r + 80059249540278.2*nu**2 + 2589101062873.81*nu*r**2 + 10611661054566.2*nu*r - 12049908701745.2*nu + 5107745331375.71*r**2 - 326837426.241486*r*(14700.0*nu + 42911.0) - 39476764256925.6*r - (-5041721180160.0*nu**2 - 25392914995744.3*nu - 879923036160.0*r**2 - 283115520.0*r*(14700.0*nu + 42911.0) + 104186110149937.0)*log(r) + 5787938193408.0*log(r)**2 + 275059053208689.0)/(55296.0*nu*(14515200.0*nu**3 - 42636451.6032331*nu**2 - 2510664218.28128*nu + 1002013764.01019) - 967680.0*r**3*(-138240.0*nu**2 - 2675575.66847905*nu - 5278341.3229329) - 9216.0*r**2*(-197773496.793534*nu**2 - 630116198.873299*nu + 5805304367.87913) + r*(5927865218923.02*nu**3 + 43133561885859.3*nu**2 + 43393301259014.8*nu + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0) + 5787938193408.0*r*log(r)**2 + (-1698693120.0*nu*(11592.0*nu + 69847.0) + 879923036160.0*r**3 + 283115520.0*r**2*(14700.0*nu + 42911.0) + 49152.0*r*(102574080.0*nu**2 + 409207698.136075*nu - 2119671837.36038))*log(r))
 * 
 *         cdef double Apm = 7680.0*r**4*(-5416406.59541186*nu**2 + 28.0*nu*(1920.0*a6 + 733955.307463037) + 2048.0*nu*(756.0*nu + 336.0*r + 407.0)*log(r) - 7.0*r*(-185763.092693281*nu**2 + 938918.400156317*nu - 245760.0) - 3440640.0)/(241555486248.807*nu**4 + 1120.0*nu**3*(-17833256.898555*r**2 - 163683964.822551*r - 1188987459.03162) + 7.0*nu**2*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 1426660551.8844*r**5 - 3089250703.76879*r**4 - 6178501407.53758*r**3 + 2064783811.32587*r**2 + 122635399361.987*r + 276057889687.011) + 67645734912.0*nu**2*log(r)**2 + 53760.0*nu*(7680.0*a6*(r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 128.0*r*(-6852.34813868015*r**4 + 4264.6962773603*r**3 + 8529.39255472061*r**2 + 13218.7851094412*r - 33722.4297811176) + 113485.217444961*r*(-r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 148.04406601634*r*(349.0*r**4 + 1926.0*r**3 + 3852.0*r**2 + 7704.0*r + 36400.0)) + 32768.0*nu*(-1882456.23663972*nu**2 - 38842241.4769507*nu + 161280.0*r**5 + 480.0*r**4*(756.0*nu + 1079.0) + 960.0*r**3*(756.0*nu + 1079.0) + 1920.0*r**2*(588.0*nu + 1079.0) + 240.0*r*(-3024.0*nu**2 - 7466.27061066206*nu + 17264.0) + 13447680.0)*log(r) + 13212057600.0*r**5)             # <<<<<<<<<<<<<<
 * 
 *         cdef double t2 = chix2**2 + chiy2**2 + chiz2**2
 */
  __Pyx_TraceLine(1336,0,__PYX_ERR(1, 1336, __pyx_L1_error))
  __pyx_v_Apm = (((7680.0 * pow(__pyx_v_r, 4.0)) * (((((-5416406.59541186 * pow(__pyx_v_nu, 2.0)) + ((28.0 * __pyx_v_nu) * ((1920.0 * __pyx_v_a6) + 733955.307463037))) + (((2048.0 * __pyx_v_nu) * (((756.0 * __pyx_v_nu) + (336.0 * __pyx_v_r)) + 407.0)) * log(__pyx_v_r))) - ((7.0 * __pyx_v_r) * (((-185763.092693281 * pow(__pyx_v_nu, 2.0)) + (938918.400156317 * __pyx_v_nu)) - 245760.0))) - 3440640.0)) / (((((((241555486248.807 * pow(__pyx_v_nu, 4.0)) + ((1120.0 * pow(__pyx_v_nu, 3.0)) * (((-17833256.898555 * pow(__pyx_v_r, 2.0)) - (163683964.822551 * __pyx_v_r)) - 1188987459.03162))) + ((7.0 * pow(__pyx_v_nu, 2.0)) * (((((((((-39321600.0 * __pyx_v_a6) * ((3.0 * __pyx_v_r) + 59.0)) + (745857848.115604 * __pyx_v_a6)) + (1426660551.8844 * pow(__pyx_v_r, 5.0))) - (3089250703.76879 * pow(__pyx_v_r, 4.0))) - (6178501407.53758 * pow(__pyx_v_r, 3.0))) + (2064783811.32587 * pow(__pyx_v_r, 2.0))) + (122635399361.987 * __pyx_v_r)) + 276057889687.011))) + ((67645734912.0 * pow(__pyx_v_nu, 2.0)) * pow(log(__pyx_v_r), 2.0))) + ((53760.0 * __pyx_v_nu) * (((((7680.0 * __pyx_v_a6) * ((((pow(__pyx_v_r, 4.0) + (2.0 * pow(__pyx_v_r, 3.0))) + (4.0 * pow(__pyx_v_r, 2.0))) + (8.0 * __pyx_v_r)) + 16.0)) + ((128.0 * __pyx_v_r) * (((((-6852.34813868015 * pow(__pyx_v_r, 4.0)) + (4264.6962773603 * pow(__pyx_v_r, 3.0))) + (8529.39255472061 * pow(__pyx_v_r, 2.0))) + (13218.7851094412 * __pyx_v_r)) - 33722.4297811176))) + ((113485.217444961 * __pyx_v_r) * (((((-pow(__pyx_v_r, 4.0)) + (2.0 * pow(__pyx_v_r, 3.0))) + (4.0 * pow(__pyx_v_r, 2.0))) + (8.0 * __pyx_v_r)) + 16.0))) + ((148.04406601634 * __pyx_v_r) * (((((349.0 * pow(__pyx_v_r, 4.0)) + (1926.0 * pow(__pyx_v_r, 3.0))) + (3852.0 * pow(__pyx_v_r, 2.0))) + (7704.0 * __pyx_v_r)) + 36400.0))))) + (((32768.0 * __pyx_v_nu) * ((((((((-1882456.23663972 * pow(__pyx_v_nu, 2.0)) - (38842241.4769507 * __pyx_v_nu)) + (161280.0 * pow(__pyx_v_r, 5.0))) + ((480.0 * pow(__pyx_v_r, 4.0)) * ((756.0 * __pyx_v_nu) + 1079.0))) + ((960.0 * pow(__pyx_v_r, 3.0)) * ((756.0 * __pyx_v_nu) + 1079.0))) + ((1920.0 * pow(__pyx_v_r, 2.0)) * ((588.0 * __pyx_v_nu) + 1079.0))) + ((240.0 * __pyx_v_r) * (((-3024.0 * pow(__pyx_v_nu, 2.0)) - (7466.27061066206 * __pyx_v_nu)) + 17264.0))) + 13447680.0)) * log(__pyx_v_r))) + (13212057600.0 * pow(__pyx_v_r, 5.0))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1338
 *         cdef double Apm = 7680.0*r**4*(-5416406.59541186*nu**2 + 28.0*nu*(1920.0*a6 + 733955.307463037) + 2048.0*nu*(756.0*nu + 336.0*r + 407.0)*log(r) - 7.0*r*(-185763.092693281*nu**2 + 938918.400156317*nu - 245760.0) - 3440640.0)/(241555486248.807*nu**4 + 1120.0*nu**3*(-17833256.898555*r**2 - 163683964.822551*r - 1188987459.03162) + 7.0*nu**2*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 1426660551.8844*r**5 - 3089250703.76879*r**4 - 6178501407.53758*r**3 + 2064783811.32587*r**2 + 122635399361.987*r + 276057889687.011) + 67645734912.0*nu**2*log(r)**2 + 53760.0*nu*(7680.0*a6*(r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 128.0*r*(-6852.34813868015*r**4 + 4264.6962773603*r**3 + 8529.39255472061*r**2 + 13218.7851094412*r - 33722.4297811176) + 113485.217444961*r*(-r**4 + 2.0*r**3 + 4.0*r**2 + 8.0*r + 16.0) + 148.04406601634*r*(349.0*r**4 + 1926.0*r**3 + 3852.0*r**2 + 7704.0*r + 36400.0)) + 32768.0*nu*(-1882456.23663972*nu**2 - 38842241.4769507*nu + 161280.0*r**5 + 480.0*r**4*(756.0*nu + 1079.0) + 960.0*r**3*(756.0*nu + 1079.0) + 1920.0*r**2*(588.0*nu + 1079.0) + 240.0*r*(-3024.0*nu**2 - 7466.27061066206*nu + 17264.0) + 13447680.0)*log(r) + 13212057600.0*r**5)
 * 
 *         cdef double t2 = chix2**2 + chiy2**2 + chiz2**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double t1 = chix1**2 + chiy1**2 + chiz1**2
 */
  __Pyx_TraceLine(1338,0,__PYX_ERR(1, 1338, __pyx_L1_error))
  __pyx_v_t2 = ((pow(__pyx_v_chix2, 2.0) + pow(__pyx_v_chiy2, 2.0)) + pow(__pyx_v_chiz2, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1340
 *         cdef double t2 = chix2**2 + chiy2**2 + chiz2**2
 * 
 *         cdef double t1 = chix1**2 + chiy1**2 + chiz1**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double ap2 = X_1**2*t1 + X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2
 */
  __Pyx_TraceLine(1340,0,__PYX_ERR(1, 1340, __pyx_L1_error))
  __pyx_v_t1 = ((pow(__pyx_v_chix1, 2.0) + pow(__pyx_v_chiy1, 2.0)) + pow(__pyx_v_chiz1, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1342
 *         cdef double t1 = chix1**2 + chiy1**2 + chiz1**2
 * 
 *         cdef double ap2 = X_1**2*t1 + X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2             # <<<<<<<<<<<<<<
 * 
 *         cdef double xi = Dbpm**0.5*r**2*(Apm + ap2/r**2)/(ap2 + r**2)
 */
  __Pyx_TraceLine(1342,0,__PYX_ERR(1, 1342, __pyx_L1_error))
  __pyx_v_ap2 = (((pow(__pyx_v_X_1, 2.0) * __pyx_v_t1) + ((__pyx_v_X_1 * __pyx_v_X_2) * ((((2.0 * __pyx_v_chix1) * __pyx_v_chix2) + ((2.0 * __pyx_v_chiy1) * __pyx_v_chiy2)) + ((2.0 * __pyx_v_chiz1) * __pyx_v_chiz2)))) + (pow(__pyx_v_X_2, 2.0) * __pyx_v_t2));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1344
 *         cdef double ap2 = X_1**2*t1 + X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2) + X_2**2*t2
 * 
 *         cdef double xi = Dbpm**0.5*r**2*(Apm + ap2/r**2)/(ap2 + r**2)             # <<<<<<<<<<<<<<
 * 
 *         return xi
 */
  __Pyx_TraceLine(1344,0,__PYX_ERR(1, 1344, __pyx_L1_error))
  __pyx_v_xi = (((pow(__pyx_v_Dbpm, 0.5) * pow(__pyx_v_r, 2.0)) * (__pyx_v_Apm + (__pyx_v_ap2 / pow(__pyx_v_r, 2.0)))) / (__pyx_v_ap2 + pow(__pyx_v_r, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1346
 *         cdef double xi = Dbpm**0.5*r**2*(Apm + ap2/r**2)/(ap2 + r**2)
 * 
 *         return xi             # <<<<<<<<<<<<<<
 * 
 *     cpdef dynamics(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 */
  __Pyx_TraceLine(1346,0,__PYX_ERR(1, 1346, __pyx_L1_error))
  __pyx_r = __pyx_v_xi;
  goto __pyx_L0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1281
 *         return np.array([d2Hdr2, d2Hdrdphi, d2Hdrdpr, d2Hdrdpphi, d2Hdrdphi, d2Hdphi2, d2Hdphidpr, d2Hdphidpphi, d2Hdrdpr, d2Hdphidpr, d2Hdpr2, d2Hdprdpphi, d2Hdrdpphi, d2Hdphidpphi, d2Hdprdpphi, d2Hdpphi2]).reshape(4, 4)
 * 
 *     cdef double xi(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.xi", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c_coeffs);
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1348
 *         return xi
 * 
 *     cpdef dynamics(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_11dynamics(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_dynamics(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch) {
  double __pyx_v_r;
  CYTHON_UNUSED double __pyx_v_phi;
  double __pyx_v_prst;
  double __pyx_v_L;
  double __pyx_v_chix1;
  double __pyx_v_chiy1;
  double __pyx_v_chiz1;
  double __pyx_v_chix2;
  double __pyx_v_chiy2;
  double __pyx_v_chiz2;
  double __pyx_v_pphi;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *__pyx_v_c_coeffs = 0;
  double __pyx_v_a6;
  double __pyx_v_dSO;
  double __pyx_v_M;
  double __pyx_v_nu;
  double __pyx_v_X_1;
  double __pyx_v_X_2;
  double __pyx_v_delta;
  double __pyx_v_x0;
  double __pyx_v_x1;
  double __pyx_v_x2;
  double __pyx_v_x3;
  double __pyx_v_x4;
  double __pyx_v_x5;
  double __pyx_v_x6;
  double __pyx_v_x7;
  double __pyx_v_x8;
  double __pyx_v_x9;
  double __pyx_v_x10;
  double __pyx_v_x11;
  double __pyx_v_x12;
  double __pyx_v_x13;
  double __pyx_v_x14;
  double __pyx_v_x15;
  double __pyx_v_x16;
  double __pyx_v_x17;
  double __pyx_v_x18;
  double __pyx_v_x19;
  double __pyx_v_x20;
  double __pyx_v_x21;
  double __pyx_v_x22;
  double __pyx_v_x23;
  double __pyx_v_x24;
  double __pyx_v_x25;
  double __pyx_v_x26;
  double __pyx_v_x27;
  double __pyx_v_x28;
  double __pyx_v_x29;
  double __pyx_v_x30;
  double __pyx_v_x31;
  double __pyx_v_x32;
  double __pyx_v_x33;
  double __pyx_v_x34;
  double __pyx_v_x35;
  double __pyx_v_x36;
  double __pyx_v_x37;
  double __pyx_v_x38;
  double __pyx_v_x39;
  double __pyx_v_x40;
  double __pyx_v_x41;
  double __pyx_v_x42;
  double __pyx_v_x43;
  double __pyx_v_x44;
  double __pyx_v_x45;
  double __pyx_v_x46;
  double __pyx_v_x47;
  double __pyx_v_x48;
  double __pyx_v_x49;
  double __pyx_v_x50;
  double __pyx_v_x51;
  double __pyx_v_x52;
  double __pyx_v_x53;
  double __pyx_v_x54;
  double __pyx_v_x55;
  double __pyx_v_x56;
  double __pyx_v_x57;
  double __pyx_v_x58;
  double __pyx_v_x59;
  double __pyx_v_x60;
  double __pyx_v_x61;
  double __pyx_v_x62;
  double __pyx_v_x63;
  double __pyx_v_x64;
  double __pyx_v_x65;
  double __pyx_v_x66;
  double __pyx_v_x67;
  double __pyx_v_x68;
  double __pyx_v_x69;
  double __pyx_v_x70;
  double __pyx_v_x71;
  double __pyx_v_x72;
  double __pyx_v_x73;
  double __pyx_v_x74;
  double __pyx_v_x75;
  double __pyx_v_x76;
  double __pyx_v_x77;
  double __pyx_v_x78;
  double __pyx_v_x79;
  double __pyx_v_x80;
  double __pyx_v_x81;
  double __pyx_v_x82;
  double __pyx_v_x83;
  double __pyx_v_x84;
  double __pyx_v_x85;
  double __pyx_v_x86;
  double __pyx_v_x87;
  double __pyx_v_x88;
  double __pyx_v_x89;
  double __pyx_v_x90;
  double __pyx_v_x91;
  double __pyx_v_x92;
  double __pyx_v_x93;
  double __pyx_v_x94;
  double __pyx_v_x95;
  double __pyx_v_x96;
  double __pyx_v_x97;
  double __pyx_v_x98;
  double __pyx_v_x99;
  double __pyx_v_x100;
  double __pyx_v_x101;
  double __pyx_v_x102;
  double __pyx_v_x103;
  double __pyx_v_x104;
  double __pyx_v_x105;
  double __pyx_v_x106;
  double __pyx_v_x107;
  double __pyx_v_x108;
  double __pyx_v_x109;
  double __pyx_v_x110;
  double __pyx_v_x111;
  double __pyx_v_x112;
  double __pyx_v_x113;
  double __pyx_v_x114;
  double __pyx_v_x115;
  double __pyx_v_x116;
  double __pyx_v_x117;
  double __pyx_v_x118;
  double __pyx_v_x119;
  double __pyx_v_x120;
  double __pyx_v_x121;
  double __pyx_v_x122;
  double __pyx_v_x123;
  double __pyx_v_x124;
  double __pyx_v_x125;
  double __pyx_v_x126;
  double __pyx_v_x127;
  double __pyx_v_x128;
  double __pyx_v_x129;
  double __pyx_v_x130;
  double __pyx_v_x131;
  double __pyx_v_x132;
  double __pyx_v_x133;
  double __pyx_v_x134;
  double __pyx_v_x135;
  double __pyx_v_x136;
  double __pyx_v_x137;
  double __pyx_v_x138;
  double __pyx_v_x139;
  double __pyx_v_x140;
  double __pyx_v_x141;
  double __pyx_v_x142;
  double __pyx_v_x143;
  double __pyx_v_x144;
  double __pyx_v_x145;
  double __pyx_v_x146;
  double __pyx_v_x147;
  double __pyx_v_x148;
  double __pyx_v_x149;
  double __pyx_v_x150;
  double __pyx_v_x151;
  double __pyx_v_x152;
  double __pyx_v_x153;
  double __pyx_v_x154;
  double __pyx_v_x155;
  double __pyx_v_x156;
  double __pyx_v_x157;
  double __pyx_v_x158;
  double __pyx_v_x159;
  double __pyx_v_x160;
  double __pyx_v_x161;
  double __pyx_v_x162;
  double __pyx_v_x163;
  double __pyx_v_x164;
  double __pyx_v_x165;
  double __pyx_v_x166;
  double __pyx_v_x167;
  double __pyx_v_x168;
  double __pyx_v_x169;
  double __pyx_v_x170;
  double __pyx_v_x171;
  double __pyx_v_x172;
  double __pyx_v_x173;
  double __pyx_v_x174;
  double __pyx_v_x175;
  double __pyx_v_x176;
  double __pyx_v_x177;
  double __pyx_v_x178;
  double __pyx_v_x179;
  double __pyx_v_x180;
  double __pyx_v_x181;
  double __pyx_v_x182;
  double __pyx_v_x183;
  double __pyx_v_x184;
  double __pyx_v_x185;
  double __pyx_v_x186;
  double __pyx_v_x187;
  double __pyx_v_x188;
  double __pyx_v_x189;
  double __pyx_v_x190;
  double __pyx_v_x191;
  double __pyx_v_x192;
  double __pyx_v_x193;
  double __pyx_v_x194;
  double __pyx_v_x195;
  double __pyx_v_x196;
  double __pyx_v_x197;
  double __pyx_v_x198;
  double __pyx_v_x199;
  double __pyx_v_x200;
  double __pyx_v_x201;
  double __pyx_v_x202;
  double __pyx_v_x203;
  double __pyx_v_x204;
  double __pyx_v_x205;
  double __pyx_v_x206;
  double __pyx_v_x207;
  double __pyx_v_x208;
  double __pyx_v_x209;
  double __pyx_v_x210;
  double __pyx_v_x211;
  double __pyx_v_x212;
  double __pyx_v_x213;
  double __pyx_v_x214;
  double __pyx_v_x215;
  double __pyx_v_x216;
  double __pyx_v_x217;
  double __pyx_v_x218;
  double __pyx_v_x219;
  double __pyx_v_x220;
  double __pyx_v_x221;
  double __pyx_v_x222;
  double __pyx_v_x223;
  double __pyx_v_x224;
  double __pyx_v_x225;
  double __pyx_v_x226;
  double __pyx_v_x227;
  double __pyx_v_x228;
  double __pyx_v_x229;
  double __pyx_v_x230;
  double __pyx_v_x231;
  double __pyx_v_x232;
  double __pyx_v_x233;
  double __pyx_v_x234;
  double __pyx_v_H;
  double __pyx_v_xi;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  double __pyx_v_dHeffdr;
  double __pyx_v_dHeffdphi;
  double __pyx_v_dHeffdpr;
  double __pyx_v_dHeffdpphi;
  double __pyx_v_M2;
  double __pyx_v_nuH;
  double __pyx_v_dHdr;
  double __pyx_v_dHdphi;
  double __pyx_v_dHdpr;
  double __pyx_v_dHdpphi;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  double __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  double __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dynamics", 0);
  __Pyx_TraceCall("dynamics", __pyx_f[1], 1348, 0, __PYX_ERR(1, 1348, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dynamics); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_11dynamics)) {
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 1348, __pyx_L1_error) }
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_q, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 1348, __pyx_L1_error) }
        __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_p, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 1348, __pyx_L1_error) }
        __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_chi1_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 1348, __pyx_L1_error) }
        __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_chi2_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_m_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyFloat_FromDouble(__pyx_v_m_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_chi_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyFloat_FromDouble(__pyx_v_chi_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_chiL1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = PyFloat_FromDouble(__pyx_v_chiL2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_13 = __pyx_t_1; __pyx_t_14 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1348, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1348, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(10+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 1348, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_15, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_15, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_15, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_15, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_15, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_15, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_16, 8+__pyx_t_15, __pyx_t_11);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_16, 9+__pyx_t_15, __pyx_t_12);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_12 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1348, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1371
 *         # Coordinate definitions
 * 
 *         cdef double r = q[0]             # <<<<<<<<<<<<<<
 *         cdef double phi = q[1]
 * 
 */
  __Pyx_TraceLine(1371,0,__PYX_ERR(1, 1371, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_r = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1372
 * 
 *         cdef double r = q[0]
 *         cdef double phi = q[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double prst = p[0]
 */
  __Pyx_TraceLine(1372,0,__PYX_ERR(1, 1372, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_phi = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1374
 *         cdef double phi = q[1]
 * 
 *         cdef double prst = p[0]             # <<<<<<<<<<<<<<
 *         cdef double L = p[1]
 * 
 */
  __Pyx_TraceLine(1374,0,__PYX_ERR(1, 1374, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_prst = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1375
 * 
 *         cdef double prst = p[0]
 *         cdef double L = p[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix1 = chi1_v[0]
 */
  __Pyx_TraceLine(1375,0,__PYX_ERR(1, 1375, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_L = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1377
 *         cdef double L = p[1]
 * 
 *         cdef double chix1 = chi1_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]
 */
  __Pyx_TraceLine(1377,0,__PYX_ERR(1, 1377, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1378
 * 
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz1 = chi1_v[2]
 * 
 */
  __Pyx_TraceLine(1378,0,__PYX_ERR(1, 1378, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1379
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix2 = chi2_v[0]
 */
  __Pyx_TraceLine(1379,0,__PYX_ERR(1, 1379, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1381
 *         cdef double chiz1 = chi1_v[2]
 * 
 *         cdef double chix2 = chi2_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]
 */
  __Pyx_TraceLine(1381,0,__PYX_ERR(1, 1381, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1382
 * 
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz2 = chi2_v[2]
 * 
 */
  __Pyx_TraceLine(1382,0,__PYX_ERR(1, 1382, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1383
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double pphi = L
 */
  __Pyx_TraceLine(1383,0,__PYX_ERR(1, 1383, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1385
 *         cdef double chiz2 = chi2_v[2]
 * 
 *         cdef double pphi = L             # <<<<<<<<<<<<<<
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 */
  __Pyx_TraceLine(1385,0,__PYX_ERR(1, 1385, __pyx_L1_error))
  __pyx_v_pphi = __pyx_v_L;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1387
 *         cdef double pphi = L
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs             # <<<<<<<<<<<<<<
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']
 */
  __Pyx_TraceLine(1387,0,__PYX_ERR(1, 1387, __pyx_L1_error))
  __pyx_t_1 = ((PyObject *)__pyx_v_self->calibration_coeffs);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_c_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1388
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']             # <<<<<<<<<<<<<<
 *         cdef double dSO = c_coeffs['dSO']
 * 
 */
  __Pyx_TraceLine(1388,0,__PYX_ERR(1, 1388, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_a6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_a6 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1389
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1389,0,__PYX_ERR(1, 1389, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_dSO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1389, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dSO = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1393
 * 
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M             # <<<<<<<<<<<<<<
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 */
  __Pyx_TraceLine(1393,0,__PYX_ERR(1, 1393, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->M;
  __pyx_v_M = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1394
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu             # <<<<<<<<<<<<<<
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 */
  __Pyx_TraceLine(1394,0,__PYX_ERR(1, 1394, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->nu;
  __pyx_v_nu = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1395
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1             # <<<<<<<<<<<<<<
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 */
  __Pyx_TraceLine(1395,0,__PYX_ERR(1, 1395, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_1;
  __pyx_v_X_1 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1396
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2             # <<<<<<<<<<<<<<
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3
 */
  __Pyx_TraceLine(1396,0,__PYX_ERR(1, 1396, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_2;
  __pyx_v_X_2 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1397
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta             # <<<<<<<<<<<<<<
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0
 */
  __Pyx_TraceLine(1397,0,__PYX_ERR(1, 1397, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->delta;
  __pyx_v_delta = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1398
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3             # <<<<<<<<<<<<<<
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 */
  __Pyx_TraceLine(1398,0,__PYX_ERR(1, 1398, __pyx_L1_error))
  __pyx_v_x0 = pow(__pyx_v_r, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1399
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0             # <<<<<<<<<<<<<<
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2
 */
  __Pyx_TraceLine(1399,0,__PYX_ERR(1, 1399, __pyx_L1_error))
  __pyx_v_x1 = (__pyx_v_r + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1400
 *         cdef double x0 = r**3
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)             # <<<<<<<<<<<<<<
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2
 */
  __Pyx_TraceLine(1400,0,__PYX_ERR(1, 1400, __pyx_L1_error))
  __pyx_v_x2 = (__pyx_v_X_1 * ((((2.0 * __pyx_v_chix1) * __pyx_v_chix2) + ((2.0 * __pyx_v_chiy1) * __pyx_v_chiy2)) + ((2.0 * __pyx_v_chiz1) * __pyx_v_chiz2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1401
 *         cdef double x1 = r + 2.0
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2             # <<<<<<<<<<<<<<
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 */
  __Pyx_TraceLine(1401,0,__PYX_ERR(1, 1401, __pyx_L1_error))
  __pyx_v_x3 = (__pyx_v_X_2 * __pyx_v_x2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1402
 *         cdef double x2 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2             # <<<<<<<<<<<<<<
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5
 */
  __Pyx_TraceLine(1402,0,__PYX_ERR(1, 1402, __pyx_L1_error))
  __pyx_v_x4 = pow(__pyx_v_X_1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1403
 *         cdef double x3 = X_2*x2
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2             # <<<<<<<<<<<<<<
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 */
  __Pyx_TraceLine(1403,0,__PYX_ERR(1, 1403, __pyx_L1_error))
  __pyx_v_x5 = ((pow(__pyx_v_chix1, 2.0) + pow(__pyx_v_chiy1, 2.0)) + pow(__pyx_v_chiz1, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1404
 *         cdef double x4 = X_1**2
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5             # <<<<<<<<<<<<<<
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7
 */
  __Pyx_TraceLine(1404,0,__PYX_ERR(1, 1404, __pyx_L1_error))
  __pyx_v_x6 = (__pyx_v_x4 * __pyx_v_x5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1405
 *         cdef double x5 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)             # <<<<<<<<<<<<<<
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8
 */
  __Pyx_TraceLine(1405,0,__PYX_ERR(1, 1405, __pyx_L1_error))
  __pyx_v_x7 = (pow(__pyx_v_X_2, 2.0) * ((pow(__pyx_v_chix2, 2.0) + pow(__pyx_v_chiy2, 2.0)) + pow(__pyx_v_chiz2, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1406
 *         cdef double x6 = x4*x5
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7             # <<<<<<<<<<<<<<
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9
 */
  __Pyx_TraceLine(1406,0,__PYX_ERR(1, 1406, __pyx_L1_error))
  __pyx_v_x8 = (__pyx_v_x6 + __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1407
 *         cdef double x7 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8             # <<<<<<<<<<<<<<
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10
 */
  __Pyx_TraceLine(1407,0,__PYX_ERR(1, 1407, __pyx_L1_error))
  __pyx_v_x9 = (__pyx_v_x3 + __pyx_v_x8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1408
 *         cdef double x8 = x6 + x7
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9             # <<<<<<<<<<<<<<
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)
 */
  __Pyx_TraceLine(1408,0,__PYX_ERR(1, 1408, __pyx_L1_error))
  __pyx_v_x10 = (__pyx_v_x1 * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1409
 *         cdef double x9 = x3 + x8
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10             # <<<<<<<<<<<<<<
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4
 */
  __Pyx_TraceLine(1409,0,__PYX_ERR(1, 1409, __pyx_L1_error))
  __pyx_v_x11 = (__pyx_v_x0 + __pyx_v_x10);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1410
 *         cdef double x10 = x1*x9
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)
 */
  __Pyx_TraceLine(1410,0,__PYX_ERR(1, 1410, __pyx_L1_error))
  __pyx_v_x12 = pow(__pyx_v_x11, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1411
 *         cdef double x11 = x0 + x10
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4             # <<<<<<<<<<<<<<
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu
 */
  __Pyx_TraceLine(1411,0,__PYX_ERR(1, 1411, __pyx_L1_error))
  __pyx_v_x13 = pow(__pyx_v_r, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1412
 *         cdef double x12 = x11**(-1)
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1
 */
  __Pyx_TraceLine(1412,0,__PYX_ERR(1, 1412, __pyx_L1_error))
  __pyx_v_x14 = pow(__pyx_v_x13, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1413
 *         cdef double x13 = r**4
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2
 */
  __Pyx_TraceLine(1413,0,__PYX_ERR(1, 1413, __pyx_L1_error))
  __pyx_v_x15 = (3.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1414
 *         cdef double x14 = x13**(-1)
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1             # <<<<<<<<<<<<<<
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17
 */
  __Pyx_TraceLine(1414,0,__PYX_ERR(1, 1414, __pyx_L1_error))
  __pyx_v_x16 = (__pyx_v_X_1 * __pyx_v_chiL1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1415
 *         cdef double x15 = 3.0*nu
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2             # <<<<<<<<<<<<<<
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18
 */
  __Pyx_TraceLine(1415,0,__PYX_ERR(1, 1415, __pyx_L1_error))
  __pyx_v_x17 = (__pyx_v_X_2 * __pyx_v_chiL2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1416
 *         cdef double x16 = X_1*chiL1
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17             # <<<<<<<<<<<<<<
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2
 */
  __Pyx_TraceLine(1416,0,__PYX_ERR(1, 1416, __pyx_L1_error))
  __pyx_v_x18 = (__pyx_v_x16 + __pyx_v_x17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1417
 *         cdef double x17 = X_2*chiL2
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18             # <<<<<<<<<<<<<<
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)
 */
  __Pyx_TraceLine(1417,0,__PYX_ERR(1, 1417, __pyx_L1_error))
  __pyx_v_x19 = (__pyx_v_pphi * __pyx_v_x18);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1418
 *         cdef double x18 = x16 + x17
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2             # <<<<<<<<<<<<<<
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375
 */
  __Pyx_TraceLine(1418,0,__PYX_ERR(1, 1418, __pyx_L1_error))
  __pyx_v_x20 = pow(__pyx_v_r, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1419
 *         cdef double x19 = pphi*x18
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875
 */
  __Pyx_TraceLine(1419,0,__PYX_ERR(1, 1419, __pyx_L1_error))
  __pyx_v_x21 = pow(__pyx_v_x20, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1420
 *         cdef double x20 = r**2
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375             # <<<<<<<<<<<<<<
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)
 */
  __Pyx_TraceLine(1420,0,__PYX_ERR(1, 1420, __pyx_L1_error))
  __pyx_v_x22 = ((0.71875 * __pyx_v_nu) - 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1421
 *         cdef double x21 = x20**(-1)
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875             # <<<<<<<<<<<<<<
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2
 */
  __Pyx_TraceLine(1421,0,__PYX_ERR(1, 1421, __pyx_L1_error))
  __pyx_v_x23 = ((-1.40625 * __pyx_v_nu) - 0.46875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1422
 *         cdef double x22 = 0.71875*nu - 0.09375
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25
 */
  __Pyx_TraceLine(1422,0,__PYX_ERR(1, 1422, __pyx_L1_error))
  __pyx_v_x24 = pow(__pyx_v_x0, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1423
 *         cdef double x23 = -1.40625*nu - 0.46875
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2             # <<<<<<<<<<<<<<
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26
 */
  __Pyx_TraceLine(1423,0,__PYX_ERR(1, 1423, __pyx_L1_error))
  __pyx_v_x25 = pow(__pyx_v_pphi, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1424
 *         cdef double x24 = x0**(-1)
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25             # <<<<<<<<<<<<<<
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2
 */
  __Pyx_TraceLine(1424,0,__PYX_ERR(1, 1424, __pyx_L1_error))
  __pyx_v_x26 = (__pyx_v_x24 * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1425
 *         cdef double x25 = pphi**2
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26             # <<<<<<<<<<<<<<
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 */
  __Pyx_TraceLine(1425,0,__PYX_ERR(1, 1425, __pyx_L1_error))
  __pyx_v_x27 = (2.0 * __pyx_v_x26);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1426
 *         cdef double x26 = x24*x25
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2             # <<<<<<<<<<<<<<
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14
 */
  __Pyx_TraceLine(1426,0,__PYX_ERR(1, 1426, __pyx_L1_error))
  __pyx_v_x28 = pow(__pyx_v_nu, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1427
 *         cdef double x27 = 2.0*x26
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30
 */
  __Pyx_TraceLine(1427,0,__PYX_ERR(1, 1427, __pyx_L1_error))
  __pyx_v_x29 = (((-2.0859375 * __pyx_v_nu) - (2.07161458333333 * __pyx_v_x28)) + 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1428
 *         cdef double x28 = nu**2
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14             # <<<<<<<<<<<<<<
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 */
  __Pyx_TraceLine(1428,0,__PYX_ERR(1, 1428, __pyx_L1_error))
  __pyx_v_x30 = (3.0 * __pyx_v_x14);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1429
 *         cdef double x29 = -2.0859375*nu - 2.07161458333333*x28 + 0.23046875
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30             # <<<<<<<<<<<<<<
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4
 */
  __Pyx_TraceLine(1429,0,__PYX_ERR(1, 1429, __pyx_L1_error))
  __pyx_v_x31 = (__pyx_v_x25 * __pyx_v_x30);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1430
 *         cdef double x30 = 3.0*x14
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5
 */
  __Pyx_TraceLine(1430,0,__PYX_ERR(1, 1430, __pyx_L1_error))
  __pyx_v_x32 = (((0.5859375 * __pyx_v_nu) + (1.34765625 * __pyx_v_x28)) + 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1431
 *         cdef double x31 = x25*x30
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4             # <<<<<<<<<<<<<<
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)
 */
  __Pyx_TraceLine(1431,0,__PYX_ERR(1, 1431, __pyx_L1_error))
  __pyx_v_x33 = pow(__pyx_v_pphi, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1432
 *         cdef double x32 = 0.5859375*nu + 1.34765625*x28 + 0.41015625
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5             # <<<<<<<<<<<<<<
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35
 */
  __Pyx_TraceLine(1432,0,__PYX_ERR(1, 1432, __pyx_L1_error))
  __pyx_v_x34 = pow(__pyx_v_r, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1433
 *         cdef double x33 = pphi**4
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36
 */
  __Pyx_TraceLine(1433,0,__PYX_ERR(1, 1433, __pyx_L1_error))
  __pyx_v_x35 = pow(__pyx_v_x34, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1434
 *         cdef double x34 = r**5
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35             # <<<<<<<<<<<<<<
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375
 */
  __Pyx_TraceLine(1434,0,__PYX_ERR(1, 1434, __pyx_L1_error))
  __pyx_v_x36 = (4.0 * __pyx_v_x35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1435
 *         cdef double x35 = x34**(-1)
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36             # <<<<<<<<<<<<<<
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu
 */
  __Pyx_TraceLine(1435,0,__PYX_ERR(1, 1435, __pyx_L1_error))
  __pyx_v_x37 = (__pyx_v_x33 * __pyx_v_x36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1436
 *         cdef double x36 = 4.0*x35
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375             # <<<<<<<<<<<<<<
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 */
  __Pyx_TraceLine(1436,0,__PYX_ERR(1, 1436, __pyx_L1_error))
  __pyx_v_x38 = ((0.34375 * __pyx_v_nu) + 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1437
 *         cdef double x37 = x33*x36
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu             # <<<<<<<<<<<<<<
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 */
  __Pyx_TraceLine(1437,0,__PYX_ERR(1, 1437, __pyx_L1_error))
  __pyx_v_x39 = (0.46875 - (0.28125 * __pyx_v_nu));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1438
 *         cdef double x38 = 0.34375*nu + 0.09375
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17
 */
  __Pyx_TraceLine(1438,0,__PYX_ERR(1, 1438, __pyx_L1_error))
  __pyx_v_x40 = (((-0.2734375 * __pyx_v_nu) - (0.798177083333333 * __pyx_v_x28)) - 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1439
 *         cdef double x39 = 0.46875 - 0.28125*nu
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42
 */
  __Pyx_TraceLine(1439,0,__PYX_ERR(1, 1439, __pyx_L1_error))
  __pyx_v_x41 = (((-0.3515625 * __pyx_v_nu) + (0.29296875 * __pyx_v_x28)) - 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1440
 *         cdef double x40 = -0.2734375*nu - 0.798177083333333*x28 - 0.23046875
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17             # <<<<<<<<<<<<<<
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43
 */
  __Pyx_TraceLine(1440,0,__PYX_ERR(1, 1440, __pyx_L1_error))
  __pyx_v_x42 = (__pyx_v_x16 - __pyx_v_x17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1441
 *         cdef double x41 = -0.3515625*nu + 0.29296875*x28 - 0.41015625
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42             # <<<<<<<<<<<<<<
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1
 */
  __Pyx_TraceLine(1441,0,__PYX_ERR(1, 1441, __pyx_L1_error))
  __pyx_v_x43 = (__pyx_v_pphi * __pyx_v_x42);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1442
 *         cdef double x42 = x16 - x17
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43             # <<<<<<<<<<<<<<
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2
 */
  __Pyx_TraceLine(1442,0,__PYX_ERR(1, 1442, __pyx_L1_error))
  __pyx_v_x44 = (__pyx_v_delta * __pyx_v_x43);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1443
 *         cdef double x43 = pphi*x42
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1             # <<<<<<<<<<<<<<
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46
 */
  __Pyx_TraceLine(1443,0,__PYX_ERR(1, 1443, __pyx_L1_error))
  __pyx_v_x45 = (__pyx_v_chi_1 * __pyx_v_X_1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1444
 *         cdef double x44 = delta*x43
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2             # <<<<<<<<<<<<<<
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2
 */
  __Pyx_TraceLine(1444,0,__PYX_ERR(1, 1444, __pyx_L1_error))
  __pyx_v_x46 = (__pyx_v_chi_2 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1445
 *         cdef double x45 = chi_1*X_1
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46             # <<<<<<<<<<<<<<
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9
 */
  __Pyx_TraceLine(1445,0,__PYX_ERR(1, 1445, __pyx_L1_error))
  __pyx_v_x47 = (__pyx_v_x45 + __pyx_v_x46);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1446
 *         cdef double x46 = chi_2*X_2
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2             # <<<<<<<<<<<<<<
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25
 */
  __Pyx_TraceLine(1446,0,__PYX_ERR(1, 1446, __pyx_L1_error))
  __pyx_v_x48 = pow(__pyx_v_x47, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1447
 *         cdef double x47 = x45 + x46
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9             # <<<<<<<<<<<<<<
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9
 */
  __Pyx_TraceLine(1447,0,__PYX_ERR(1, 1447, __pyx_L1_error))
  __pyx_v_x49 = ((-__pyx_v_x48) + __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1448
 *         cdef double x48 = x47**2
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25             # <<<<<<<<<<<<<<
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 */
  __Pyx_TraceLine(1448,0,__PYX_ERR(1, 1448, __pyx_L1_error))
  __pyx_v_x50 = (__pyx_v_delta * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1449
 *         cdef double x49 = -x48 + x9
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9             # <<<<<<<<<<<<<<
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46
 */
  __Pyx_TraceLine(1449,0,__PYX_ERR(1, 1449, __pyx_L1_error))
  __pyx_v_x51 = (__pyx_v_delta * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1450
 *         cdef double x50 = delta*x25
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)             # <<<<<<<<<<<<<<
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 */
  __Pyx_TraceLine(1450,0,__PYX_ERR(1, 1450, __pyx_L1_error))
  __pyx_v_x52 = (__pyx_v_x49 * (0.416666666666667 - (0.833333333333333 * __pyx_v_X_2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1451
 *         cdef double x51 = delta*x9
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46             # <<<<<<<<<<<<<<
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 */
  __Pyx_TraceLine(1451,0,__PYX_ERR(1, 1451, __pyx_L1_error))
  __pyx_v_x53 = (__pyx_v_x45 - __pyx_v_x46);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1452
 *         cdef double x52 = x49*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7             # <<<<<<<<<<<<<<
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7
 */
  __Pyx_TraceLine(1452,0,__PYX_ERR(1, 1452, __pyx_L1_error))
  __pyx_v_x54 = (((__pyx_v_x4 * __pyx_v_x5) - (__pyx_v_x47 * __pyx_v_x53)) - __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1453
 *         cdef double x53 = x45 - x46
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7             # <<<<<<<<<<<<<<
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56
 */
  __Pyx_TraceLine(1453,0,__PYX_ERR(1, 1453, __pyx_L1_error))
  __pyx_v_x55 = (((((-0.125 * __pyx_v_x3) + (0.125 * __pyx_v_x48)) + ((0.5 * __pyx_v_x54) * (0.5 - __pyx_v_X_2))) - (0.125 * __pyx_v_x6)) - (0.125 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1454
 *         cdef double x54 = x4*x5 - x47*x53 - x7
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7             # <<<<<<<<<<<<<<
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54
 */
  __Pyx_TraceLine(1454,0,__PYX_ERR(1, 1454, __pyx_L1_error))
  __pyx_v_x56 = (__pyx_v_x6 - __pyx_v_x7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1455
 *         cdef double x55 = -0.125*x3 + 0.125*x48 + 0.5*x54*(0.5 - X_2) - 0.125*x6 - 0.125*x7
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56             # <<<<<<<<<<<<<<
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24
 */
  __Pyx_TraceLine(1455,0,__PYX_ERR(1, 1455, __pyx_L1_error))
  __pyx_v_x57 = (__pyx_v_delta * __pyx_v_x56);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1456
 *         cdef double x56 = x6 - x7
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54             # <<<<<<<<<<<<<<
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)
 */
  __Pyx_TraceLine(1456,0,__PYX_ERR(1, 1456, __pyx_L1_error))
  __pyx_v_x58 = (__pyx_v_delta * __pyx_v_x54);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1457
 *         cdef double x57 = delta*x56
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24             # <<<<<<<<<<<<<<
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25
 */
  __Pyx_TraceLine(1457,0,__PYX_ERR(1, 1457, __pyx_L1_error))
  __pyx_v_x59 = ((__pyx_v_nu * __pyx_v_dSO) * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1458
 *         cdef double x58 = delta*x54
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33
 */
  __Pyx_TraceLine(1458,0,__PYX_ERR(1, 1458, __pyx_L1_error))
  __pyx_v_x60 = pow(__pyx_v_r, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1459
 *         cdef double x59 = nu*dSO*x24
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25             # <<<<<<<<<<<<<<
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 */
  __Pyx_TraceLine(1459,0,__PYX_ERR(1, 1459, __pyx_L1_error))
  __pyx_v_x61 = (__pyx_v_x21 * __pyx_v_x25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1460
 *         cdef double x60 = r**(-1)
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33             # <<<<<<<<<<<<<<
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 */
  __Pyx_TraceLine(1460,0,__PYX_ERR(1, 1460, __pyx_L1_error))
  __pyx_v_x62 = (__pyx_v_x14 * __pyx_v_x33);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1461
 *         cdef double x61 = x21*x25
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)             # <<<<<<<<<<<<<<
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49
 */
  __Pyx_TraceLine(1461,0,__PYX_ERR(1, 1461, __pyx_L1_error))
  __pyx_v_x63 = (__pyx_v_x18 * ((((((__pyx_v_x21 * (((-5.53125 * __pyx_v_nu) + (0.567708333333333 * __pyx_v_x28)) - 0.078125)) + (__pyx_v_x22 * __pyx_v_x60)) + (__pyx_v_x23 * __pyx_v_x61)) + (__pyx_v_x26 * __pyx_v_x29)) + (__pyx_v_x32 * __pyx_v_x62)) + 1.75));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1462
 *         cdef double x62 = x14*x33
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)             # <<<<<<<<<<<<<<
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65
 */
  __Pyx_TraceLine(1462,0,__PYX_ERR(1, 1462, __pyx_L1_error))
  __pyx_v_x64 = (__pyx_v_delta * ((((((__pyx_v_x21 * (((-0.03125 * __pyx_v_nu) + (0.536458333333333 * __pyx_v_x28)) + 0.078125)) + (__pyx_v_x26 * __pyx_v_x40)) + (__pyx_v_x38 * __pyx_v_x60)) + (__pyx_v_x39 * __pyx_v_x61)) + (__pyx_v_x41 * __pyx_v_x62)) + 0.25));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1463
 *         cdef double x63 = x18*(x21*(-5.53125*nu + 0.567708333333333*x28 - 0.078125) + x22*x60 + x23*x61 + x26*x29 + x32*x62 + 1.75)
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49             # <<<<<<<<<<<<<<
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 */
  __Pyx_TraceLine(1463,0,__PYX_ERR(1, 1463, __pyx_L1_error))
  __pyx_v_x65 = (__pyx_v_x24 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1464
 *         cdef double x64 = delta*(x21*(-0.03125*nu + 0.536458333333333*x28 + 0.078125) + x26*x40 + x38*x60 + x39*x61 + x41*x62 + 0.25)
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65             # <<<<<<<<<<<<<<
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55
 */
  __Pyx_TraceLine(1464,0,__PYX_ERR(1, 1464, __pyx_L1_error))
  __pyx_v_x66 = (__pyx_v_x50 * __pyx_v_x65);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1465
 *         cdef double x65 = x24*x49
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)             # <<<<<<<<<<<<<<
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 */
  __Pyx_TraceLine(1465,0,__PYX_ERR(1, 1465, __pyx_L1_error))
  __pyx_v_x67 = (__pyx_v_x42 * ((__pyx_v_x21 * ((0.0416666666666667 * __pyx_v_x51) + (0.5 * __pyx_v_x52))) - (0.125 * __pyx_v_x66)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1466
 *         cdef double x66 = x50*x65
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55             # <<<<<<<<<<<<<<
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4
 */
  __Pyx_TraceLine(1466,0,__PYX_ERR(1, 1466, __pyx_L1_error))
  __pyx_v_x68 = (__pyx_v_x26 * __pyx_v_x55);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1467
 *         cdef double x67 = x42*(x21*(0.0416666666666667*x51 + 0.5*x52) - 0.125*x66)
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)             # <<<<<<<<<<<<<<
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)
 */
  __Pyx_TraceLine(1467,0,__PYX_ERR(1, 1467, __pyx_L1_error))
  __pyx_v_x69 = (__pyx_v_x18 * ((__pyx_v_x21 * (((((((0.208333333333333 * __pyx_v_delta) * __pyx_v_x56) - (0.625 * __pyx_v_x3)) + (0.375 * __pyx_v_x48)) - (0.833333333333333 * __pyx_v_x58)) - (0.625 * __pyx_v_x6)) - (0.625 * __pyx_v_x7))) + __pyx_v_x68));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1468
 *         cdef double x68 = x26*x55
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4             # <<<<<<<<<<<<<<
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71
 */
  __Pyx_TraceLine(1468,0,__PYX_ERR(1, 1468, __pyx_L1_error))
  __pyx_v_x70 = pow(__pyx_v_prst, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1469
 *         cdef double x69 = x18*(x21*(0.208333333333333*delta*x56 - 0.625*x3 + 0.375*x48 - 0.833333333333333*x58 - 0.625*x6 - 0.625*x7) + x68)
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)             # <<<<<<<<<<<<<<
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6
 */
  __Pyx_TraceLine(1469,0,__PYX_ERR(1, 1469, __pyx_L1_error))
  __pyx_v_x71 = pow(__pyx_v_r, -4.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1470
 *         cdef double x70 = prst**4
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71             # <<<<<<<<<<<<<<
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)
 */
  __Pyx_TraceLine(1470,0,__PYX_ERR(1, 1470, __pyx_L1_error))
  __pyx_v_x72 = (__pyx_v_nu * __pyx_v_x71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1471
 *         cdef double x71 = r**(-4.5)
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6             # <<<<<<<<<<<<<<
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74
 */
  __Pyx_TraceLine(1471,0,__PYX_ERR(1, 1471, __pyx_L1_error))
  __pyx_v_x73 = pow(__pyx_v_prst, 6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1472
 *         cdef double x72 = nu*x71
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)             # <<<<<<<<<<<<<<
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)
 */
  __Pyx_TraceLine(1472,0,__PYX_ERR(1, 1472, __pyx_L1_error))
  __pyx_v_x74 = pow(__pyx_v_r, -3.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1473
 *         cdef double x73 = prst**6
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74             # <<<<<<<<<<<<<<
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8
 */
  __Pyx_TraceLine(1473,0,__PYX_ERR(1, 1473, __pyx_L1_error))
  __pyx_v_x75 = (__pyx_v_nu * __pyx_v_x74);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1474
 *         cdef double x74 = r**(-3.5)
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)             # <<<<<<<<<<<<<<
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77
 */
  __Pyx_TraceLine(1474,0,__PYX_ERR(1, 1474, __pyx_L1_error))
  __pyx_v_x76 = pow(__pyx_v_r, -2.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1475
 *         cdef double x75 = nu*x74
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8             # <<<<<<<<<<<<<<
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60
 */
  __Pyx_TraceLine(1475,0,__PYX_ERR(1, 1475, __pyx_L1_error))
  __pyx_v_x77 = pow(__pyx_v_prst, 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1476
 *         cdef double x76 = r**(-2.5)
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77             # <<<<<<<<<<<<<<
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77
 */
  __Pyx_TraceLine(1476,0,__PYX_ERR(1, 1476, __pyx_L1_error))
  __pyx_v_x78 = (__pyx_v_nu * __pyx_v_x77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1477
 *         cdef double x77 = prst**8
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60             # <<<<<<<<<<<<<<
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28
 */
  __Pyx_TraceLine(1477,0,__PYX_ERR(1, 1477, __pyx_L1_error))
  __pyx_v_x79 = (__pyx_v_nu * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1478
 *         cdef double x78 = nu*x77
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77             # <<<<<<<<<<<<<<
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81
 */
  __Pyx_TraceLine(1478,0,__PYX_ERR(1, 1478, __pyx_L1_error))
  __pyx_v_x80 = (0.121954868780449 * __pyx_v_x77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1479
 *         cdef double x79 = nu*x60
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3
 */
  __Pyx_TraceLine(1479,0,__PYX_ERR(1, 1479, __pyx_L1_error))
  __pyx_v_x81 = ((8.0 * __pyx_v_nu) - (6.0 * __pyx_v_x28));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1480
 *         cdef double x80 = 0.121954868780449*x77
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81             # <<<<<<<<<<<<<<
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 */
  __Pyx_TraceLine(1480,0,__PYX_ERR(1, 1480, __pyx_L1_error))
  __pyx_v_x82 = (__pyx_v_x21 * __pyx_v_x81);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1481
 *         cdef double x81 = 8.0*nu - 6.0*x28
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3             # <<<<<<<<<<<<<<
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70
 */
  __Pyx_TraceLine(1481,0,__PYX_ERR(1, 1481, __pyx_L1_error))
  __pyx_v_x83 = pow(__pyx_v_nu, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1482
 *         cdef double x82 = x21*x81
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83             # <<<<<<<<<<<<<<
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 */
  __Pyx_TraceLine(1482,0,__PYX_ERR(1, 1482, __pyx_L1_error))
  __pyx_v_x84 = (((92.7110442849544 * __pyx_v_nu) - (131.0 * __pyx_v_x28)) + (10.0 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1483
 *         cdef double x83 = nu**3
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70             # <<<<<<<<<<<<<<
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86
 */
  __Pyx_TraceLine(1483,0,__PYX_ERR(1, 1483, __pyx_L1_error))
  __pyx_v_x85 = (__pyx_v_x24 * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1484
 *         cdef double x84 = 92.7110442849544*nu - 131.0*x28 + 10.0*x83
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83             # <<<<<<<<<<<<<<
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4
 */
  __Pyx_TraceLine(1484,0,__PYX_ERR(1, 1484, __pyx_L1_error))
  __pyx_v_x86 = (((-2.78300763695006 * __pyx_v_nu) - (5.4 * __pyx_v_x28)) + (6.0 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1485
 *         cdef double x85 = x24*x70
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86             # <<<<<<<<<<<<<<
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 */
  __Pyx_TraceLine(1485,0,__PYX_ERR(1, 1485, __pyx_L1_error))
  __pyx_v_x87 = (__pyx_v_x21 * __pyx_v_x86);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1486
 *         cdef double x86 = -2.78300763695006*nu - 5.4*x28 + 6.0*x83
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4             # <<<<<<<<<<<<<<
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73
 */
  __Pyx_TraceLine(1486,0,__PYX_ERR(1, 1486, __pyx_L1_error))
  __pyx_v_x88 = pow(__pyx_v_nu, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1487
 *         cdef double x87 = x21*x86
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88             # <<<<<<<<<<<<<<
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 */
  __Pyx_TraceLine(1487,0,__PYX_ERR(1, 1487, __pyx_L1_error))
  __pyx_v_x89 = ((((-33.9782122170436 * __pyx_v_nu) - (89.5298327361234 * __pyx_v_x28)) + (188.0 * __pyx_v_x83)) - (14.0 * __pyx_v_x88));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1488
 *         cdef double x88 = nu**4
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73             # <<<<<<<<<<<<<<
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91
 */
  __Pyx_TraceLine(1488,0,__PYX_ERR(1, 1488, __pyx_L1_error))
  __pyx_v_x90 = (__pyx_v_x24 * __pyx_v_x73);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1489
 *         cdef double x89 = -33.9782122170436*nu - 89.5298327361234*x28 + 188.0*x83 - 14.0*x88
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88             # <<<<<<<<<<<<<<
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)
 */
  __Pyx_TraceLine(1489,0,__PYX_ERR(1, 1489, __pyx_L1_error))
  __pyx_v_x91 = ((((1.38977750996128 * __pyx_v_nu) + (3.33842023648322 * __pyx_v_x28)) + (3.42857142857143 * __pyx_v_x83)) - (6.0 * __pyx_v_x88));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1490
 *         cdef double x90 = x24*x73
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91             # <<<<<<<<<<<<<<
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 */
  __Pyx_TraceLine(1490,0,__PYX_ERR(1, 1490, __pyx_L1_error))
  __pyx_v_x92 = (__pyx_v_x77 * __pyx_v_x91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1491
 *         cdef double x91 = 1.38977750996128*nu + 3.33842023648322*x28 + 3.42857142857143*x83 - 6.0*x88
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)             # <<<<<<<<<<<<<<
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70
 */
  __Pyx_TraceLine(1491,0,__PYX_ERR(1, 1491, __pyx_L1_error))
  __pyx_v_x93 = log(__pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1492
 *         cdef double x92 = x77*x91
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83             # <<<<<<<<<<<<<<
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13
 */
  __Pyx_TraceLine(1492,0,__PYX_ERR(1, 1492, __pyx_L1_error))
  __pyx_v_x94 = (((__pyx_v_nu * (452.542166996693 - (51.6952380952381 * __pyx_v_x93))) + (__pyx_v_x28 * ((118.4 * __pyx_v_x93) - 1796.13660498019))) + (602.318540416564 * __pyx_v_x83));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1493
 *         cdef double x93 = log(r)
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70             # <<<<<<<<<<<<<<
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)
 */
  __Pyx_TraceLine(1493,0,__PYX_ERR(1, 1493, __pyx_L1_error))
  __pyx_v_x95 = (__pyx_v_x14 * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1494
 *         cdef double x94 = nu*(452.542166996693 - 51.6952380952381*x93) + x28*(118.4*x93 - 1796.13660498019) + 602.318540416564*x83
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13             # <<<<<<<<<<<<<<
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2
 */
  __Pyx_TraceLine(1494,0,__PYX_ERR(1, 1494, __pyx_L1_error))
  __pyx_v_x96 = ((__pyx_v_r * __pyx_v_x10) + __pyx_v_x13);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1495
 *         cdef double x95 = x14*x70
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98
 */
  __Pyx_TraceLine(1495,0,__PYX_ERR(1, 1495, __pyx_L1_error))
  __pyx_v_x97 = pow(__pyx_v_x96, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1496
 *         cdef double x96 = r*x10 + x13
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2             # <<<<<<<<<<<<<<
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99
 */
  __Pyx_TraceLine(1496,0,__PYX_ERR(1, 1496, __pyx_L1_error))
  __pyx_v_x98 = pow(__pyx_v_x18, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1497
 *         cdef double x97 = x96**(-1)
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98             # <<<<<<<<<<<<<<
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49
 */
  __Pyx_TraceLine(1497,0,__PYX_ERR(1, 1497, __pyx_L1_error))
  __pyx_v_x99 = (__pyx_v_x97 * __pyx_v_x98);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1498
 *         cdef double x98 = x18**2
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99             # <<<<<<<<<<<<<<
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54
 */
  __Pyx_TraceLine(1498,0,__PYX_ERR(1, 1498, __pyx_L1_error))
  __pyx_v_x100 = ((__pyx_v_x25 * __pyx_v_x60) * __pyx_v_x99);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1499
 *         cdef double x99 = x97*x98
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49             # <<<<<<<<<<<<<<
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8
 */
  __Pyx_TraceLine(1499,0,__PYX_ERR(1, 1499, __pyx_L1_error))
  __pyx_v_x101 = (0.5 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1500
 *         cdef double x100 = x25*x60*x99
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54             # <<<<<<<<<<<<<<
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2
 */
  __Pyx_TraceLine(1500,0,__PYX_ERR(1, 1500, __pyx_L1_error))
  __pyx_v_x102 = (0.5 * __pyx_v_x54);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1501
 *         cdef double x101 = 0.5*x49
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8             # <<<<<<<<<<<<<<
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104
 */
  __Pyx_TraceLine(1501,0,__PYX_ERR(1, 1501, __pyx_L1_error))
  __pyx_v_x103 = ((-__pyx_v_x3) + __pyx_v_x8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1502
 *         cdef double x102 = 0.5*x54
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2             # <<<<<<<<<<<<<<
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 */
  __Pyx_TraceLine(1502,0,__PYX_ERR(1, 1502, __pyx_L1_error))
  __pyx_v_x104 = (__pyx_v_x103 - pow(__pyx_v_x53, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1503
 *         cdef double x103 = -x3 + x8
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104             # <<<<<<<<<<<<<<
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28
 */
  __Pyx_TraceLine(1503,0,__PYX_ERR(1, 1503, __pyx_L1_error))
  __pyx_v_x105 = (0.5 * __pyx_v_x104);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1504
 *         cdef double x104 = x103 - x53**2
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)             # <<<<<<<<<<<<<<
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 */
  __Pyx_TraceLine(1504,0,__PYX_ERR(1, 1504, __pyx_L1_error))
  __pyx_v_x106 = (__pyx_v_x58 * ((98.0 * __pyx_v_nu) + 43.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1505
 *         cdef double x105 = 0.5*x104
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49
 */
  __Pyx_TraceLine(1505,0,__PYX_ERR(1, 1505, __pyx_L1_error))
  __pyx_v_x107 = (264.0 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1506
 *         cdef double x106 = x58*(98.0*nu + 43.0)
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0             # <<<<<<<<<<<<<<
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 */
  __Pyx_TraceLine(1506,0,__PYX_ERR(1, 1506, __pyx_L1_error))
  __pyx_v_x108 = (((-1610.0 * __pyx_v_nu) + __pyx_v_x107) + 375.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1507
 *         cdef double x107 = 264.0*x28
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49             # <<<<<<<<<<<<<<
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104
 */
  __Pyx_TraceLine(1507,0,__PYX_ERR(1, 1507, __pyx_L1_error))
  __pyx_v_x109 = (0.00260416666666667 * __pyx_v_x49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1508
 *         cdef double x108 = -1610.0*nu + x107 + 375.0
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0             # <<<<<<<<<<<<<<
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 */
  __Pyx_TraceLine(1508,0,__PYX_ERR(1, 1508, __pyx_L1_error))
  __pyx_v_x110 = (((-1090.0 * __pyx_v_nu) + (152.0 * __pyx_v_x28)) + 219.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1509
 *         cdef double x109 = 0.00260416666666667*x49
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104             # <<<<<<<<<<<<<<
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)
 */
  __Pyx_TraceLine(1509,0,__PYX_ERR(1, 1509, __pyx_L1_error))
  __pyx_v_x111 = (0.0078125 * __pyx_v_x104);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1510
 *         cdef double x110 = -1090.0*nu + 152.0*x28 + 219.0
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9
 */
  __Pyx_TraceLine(1510,0,__PYX_ERR(1, 1510, __pyx_L1_error))
  __pyx_v_x112 = (((((-__pyx_v_x101) * __pyx_v_x21) + (__pyx_v_x14 * (((-0.0625 * __pyx_v_x106) + (__pyx_v_x108 * __pyx_v_x109)) + (__pyx_v_x110 * __pyx_v_x111)))) + (__pyx_v_x24 * (((__pyx_v_x101 * ((-1.75 * __pyx_v_nu) - 0.9375)) + (__pyx_v_x102 * (0.75 - (1.5 * __pyx_v_X_2)))) + (__pyx_v_x105 * (0.1875 - (0.75 * __pyx_v_nu)))))) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1511
 *         cdef double x111 = 0.0078125*x104
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)             # <<<<<<<<<<<<<<
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4
 */
  __Pyx_TraceLine(1511,0,__PYX_ERR(1, 1511, __pyx_L1_error))
  __pyx_v_x113 = pow(__pyx_v_r, -13.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1512
 *         cdef double x112 = -x101*x21 + x14*(-0.0625*x106 + x108*x109 + x110*x111) + x24*(x101*(-1.75*nu - 0.9375) + x102*(0.75 - 1.5*X_2) + x105*(0.1875 - 0.75*nu)) + 1.0
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9             # <<<<<<<<<<<<<<
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9
 */
  __Pyx_TraceLine(1512,0,__PYX_ERR(1, 1512, __pyx_L1_error))
  __pyx_v_x114 = (__pyx_v_x20 + __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1513
 *         cdef double x113 = r**(-13)
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4             # <<<<<<<<<<<<<<
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu
 */
  __Pyx_TraceLine(1513,0,__PYX_ERR(1, 1513, __pyx_L1_error))
  __pyx_v_x115 = pow(__pyx_v_x114, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1514
 *         cdef double x114 = x20 + x9
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9             # <<<<<<<<<<<<<<
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0
 */
  __Pyx_TraceLine(1514,0,__PYX_ERR(1, 1514, __pyx_L1_error))
  __pyx_v_x116 = (__pyx_v_x21 * __pyx_v_x9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1515
 *         cdef double x115 = x114**4
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 */
  __Pyx_TraceLine(1515,0,__PYX_ERR(1, 1515, __pyx_L1_error))
  __pyx_v_x117 = (756.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1516
 *         cdef double x116 = x21*x9
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0             # <<<<<<<<<<<<<<
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2
 */
  __Pyx_TraceLine(1516,0,__PYX_ERR(1, 1516, __pyx_L1_error))
  __pyx_v_x118 = (((336.0 * __pyx_v_r) + __pyx_v_x117) + 407.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1517
 *         cdef double x117 = 756.0*nu
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0             # <<<<<<<<<<<<<<
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28
 */
  __Pyx_TraceLine(1517,0,__PYX_ERR(1, 1517, __pyx_L1_error))
  __pyx_v_x119 = (((((((2048.0 * __pyx_v_nu) * __pyx_v_x118) * __pyx_v_x93) + ((28.0 * __pyx_v_nu) * ((1920.0 * __pyx_v_a6) + 733955.307463037))) - ((7.0 * __pyx_v_r) * (((938918.400156317 * __pyx_v_nu) - (185763.092693281 * __pyx_v_x28)) - 245760.0))) - (5416406.59541186 * __pyx_v_x28)) - 3440640.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1518
 *         cdef double x118 = 336.0*r + x117 + 407.0
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2             # <<<<<<<<<<<<<<
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 */
  __Pyx_TraceLine(1518,0,__PYX_ERR(1, 1518, __pyx_L1_error))
  __pyx_v_x120 = pow(__pyx_v_x93, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1519
 *         cdef double x119 = 2048.0*nu*x118*x93 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*x28 - 245760.0) - 5416406.59541186*x28 - 3440640.0
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28             # <<<<<<<<<<<<<<
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 */
  __Pyx_TraceLine(1519,0,__PYX_ERR(1, 1519, __pyx_L1_error))
  __pyx_v_x121 = (__pyx_v_x120 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1520
 *         cdef double x120 = x93**2
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)             # <<<<<<<<<<<<<<
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0
 */
  __Pyx_TraceLine(1520,0,__PYX_ERR(1, 1520, __pyx_L1_error))
  __pyx_v_x122 = (__pyx_v_x83 * (((-163683964.822551 * __pyx_v_r) - (17833256.898555 * __pyx_v_x20)) - 1188987459.03162));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1521
 *         cdef double x121 = x120*x28
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)             # <<<<<<<<<<<<<<
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0
 */
  __Pyx_TraceLine(1521,0,__PYX_ERR(1, 1521, __pyx_L1_error))
  __pyx_v_x123 = (__pyx_v_x28 * (((((((((-39321600.0 * __pyx_v_a6) * ((3.0 * __pyx_v_r) + 59.0)) + (745857848.115604 * __pyx_v_a6)) + (122635399361.987 * __pyx_v_r)) - (6178501407.53758 * __pyx_v_x0)) - (3089250703.76879 * __pyx_v_x13)) + (2064783811.32587 * __pyx_v_x20)) + (1426660551.8844 * __pyx_v_x34)) + 276057889687.011));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1522
 *         cdef double x122 = x83*(-163683964.822551*r - 17833256.898555*x20 - 1188987459.03162)
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125
 */
  __Pyx_TraceLine(1522,0,__PYX_ERR(1, 1522, __pyx_L1_error))
  __pyx_v_x124 = ((588.0 * __pyx_v_nu) + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1523
 *         cdef double x123 = x28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*x0 - 3089250703.76879*x13 + 2064783811.32587*x20 + 1426660551.8844*x34 + 276057889687.011)
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 */
  __Pyx_TraceLine(1523,0,__PYX_ERR(1, 1523, __pyx_L1_error))
  __pyx_v_x125 = (__pyx_v_x117 + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1524
 *         cdef double x124 = 588.0*nu + 1079.0
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125             # <<<<<<<<<<<<<<
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93
 */
  __Pyx_TraceLine(1524,0,__PYX_ERR(1, 1524, __pyx_L1_error))
  __pyx_v_x126 = (__pyx_v_x0 * __pyx_v_x125);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1525
 *         cdef double x125 = x117 + 1079.0
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0             # <<<<<<<<<<<<<<
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128
 */
  __Pyx_TraceLine(1525,0,__PYX_ERR(1, 1525, __pyx_L1_error))
  __pyx_v_x127 = ((((((((-38842241.4769507 * __pyx_v_nu) + ((240.0 * __pyx_v_r) * (((-7466.27061066206 * __pyx_v_nu) - (3024.0 * __pyx_v_x28)) + 17264.0))) + ((1920.0 * __pyx_v_x124) * __pyx_v_x20)) + ((480.0 * __pyx_v_x125) * __pyx_v_x13)) + (960.0 * __pyx_v_x126)) - (1882456.23663972 * __pyx_v_x28)) + (161280.0 * __pyx_v_x34)) + 13447680.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1526
 *         cdef double x126 = x0*x125
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93             # <<<<<<<<<<<<<<
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r
 */
  __Pyx_TraceLine(1526,0,__PYX_ERR(1, 1526, __pyx_L1_error))
  __pyx_v_x128 = (__pyx_v_nu * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1527
 *         cdef double x127 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*x28 + 17264.0) + 1920.0*x124*x20 + 480.0*x125*x13 + 960.0*x126 - 1882456.23663972*x28 + 161280.0*x34 + 13447680.0
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128             # <<<<<<<<<<<<<<
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 */
  __Pyx_TraceLine(1527,0,__PYX_ERR(1, 1527, __pyx_L1_error))
  __pyx_v_x129 = (__pyx_v_x127 * __pyx_v_x128);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1528
 *         cdef double x128 = nu*x93
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r             # <<<<<<<<<<<<<<
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6
 */
  __Pyx_TraceLine(1528,0,__PYX_ERR(1, 1528, __pyx_L1_error))
  __pyx_v_x130 = (8.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1529
 *         cdef double x129 = x127*x128
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0             # <<<<<<<<<<<<<<
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r
 */
  __Pyx_TraceLine(1529,0,__PYX_ERR(1, 1529, __pyx_L1_error))
  __pyx_v_x131 = ((((2.0 * __pyx_v_x0) + __pyx_v_x130) + (4.0 * __pyx_v_x20)) + 16.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1530
 *         cdef double x130 = 8.0*r
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6             # <<<<<<<<<<<<<<
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r
 */
  __Pyx_TraceLine(1530,0,__PYX_ERR(1, 1530, __pyx_L1_error))
  __pyx_v_x132 = (7680.0 * __pyx_v_a6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1531
 *         cdef double x131 = 2.0*x0 + x130 + 4.0*x20 + 16.0
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r             # <<<<<<<<<<<<<<
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r
 */
  __Pyx_TraceLine(1531,0,__PYX_ERR(1, 1531, __pyx_L1_error))
  __pyx_v_x133 = (128.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1532
 *         cdef double x132 = 7680.0*a6
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r             # <<<<<<<<<<<<<<
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r
 */
  __Pyx_TraceLine(1532,0,__PYX_ERR(1, 1532, __pyx_L1_error))
  __pyx_v_x134 = (7704.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1533
 *         cdef double x133 = 128.0*r
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r             # <<<<<<<<<<<<<<
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 */
  __Pyx_TraceLine(1533,0,__PYX_ERR(1, 1533, __pyx_L1_error))
  __pyx_v_x135 = (148.04406601634 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1534
 *         cdef double x134 = 7704.0*r
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r             # <<<<<<<<<<<<<<
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 */
  __Pyx_TraceLine(1534,0,__PYX_ERR(1, 1534, __pyx_L1_error))
  __pyx_v_x136 = (113485.217444961 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1535
 *         cdef double x135 = 148.04406601634*r
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))             # <<<<<<<<<<<<<<
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138
 */
  __Pyx_TraceLine(1535,0,__PYX_ERR(1, 1535, __pyx_L1_error))
  __pyx_v_x137 = (__pyx_v_nu * ((((__pyx_v_x132 * (__pyx_v_x13 + __pyx_v_x131)) + (__pyx_v_x133 * (((((13218.7851094412 * __pyx_v_r) + (4264.6962773603 * __pyx_v_x0)) - (6852.34813868015 * __pyx_v_x13)) + (8529.39255472061 * __pyx_v_x20)) - 33722.4297811176))) + (__pyx_v_x135 * (((((1926.0 * __pyx_v_x0) + (349.0 * __pyx_v_x13)) + __pyx_v_x134) + (3852.0 * __pyx_v_x20)) + 36400.0))) + (__pyx_v_x136 * ((-__pyx_v_x13) + __pyx_v_x131))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1536
 *         cdef double x136 = 113485.217444961*r
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139
 */
  __Pyx_TraceLine(1536,0,__PYX_ERR(1, 1536, __pyx_L1_error))
  __pyx_v_x138 = pow((((((((67645734912.0 * __pyx_v_x121) + (1120.0 * __pyx_v_x122)) + (7.0 * __pyx_v_x123)) + (32768.0 * __pyx_v_x129)) + (53760.0 * __pyx_v_x137)) + (13212057600.0 * __pyx_v_x34)) + (241555486248.807 * __pyx_v_x88)), -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1537
 *         cdef double x137 = nu*(x132*(x13 + x131) + x133*(13218.7851094412*r + 4264.6962773603*x0 - 6852.34813868015*x13 + 8529.39255472061*x20 - 33722.4297811176) + x135*(1926.0*x0 + 349.0*x13 + x134 + 3852.0*x20 + 36400.0) + x136*(-x13 + x131))
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138             # <<<<<<<<<<<<<<
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 */
  __Pyx_TraceLine(1537,0,__PYX_ERR(1, 1537, __pyx_L1_error))
  __pyx_v_x139 = (__pyx_v_x13 * __pyx_v_x138);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1538
 *         cdef double x138 = (67645734912.0*x121 + 1120.0*x122 + 7.0*x123 + 32768.0*x129 + 53760.0*x137 + 13212057600.0*x34 + 241555486248.807*x88)**(-1)
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139             # <<<<<<<<<<<<<<
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)
 */
  __Pyx_TraceLine(1538,0,__PYX_ERR(1, 1538, __pyx_L1_error))
  __pyx_v_x140 = (__pyx_v_x119 * __pyx_v_x139);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1539
 *         cdef double x139 = x13*x138
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140             # <<<<<<<<<<<<<<
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120
 */
  __Pyx_TraceLine(1539,0,__PYX_ERR(1, 1539, __pyx_L1_error))
  __pyx_v_x141 = ((0.000130208333333333 * __pyx_v_x116) + __pyx_v_x140);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1540
 *         cdef double x140 = x119*x139
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)             # <<<<<<<<<<<<<<
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 */
  __Pyx_TraceLine(1540,0,__PYX_ERR(1, 1540, __pyx_L1_error))
  __pyx_v_x142 = pow(__pyx_v_x141, -4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1541
 *         cdef double x141 = 0.000130208333333333*x116 + x140
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120             # <<<<<<<<<<<<<<
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20
 */
  __Pyx_TraceLine(1541,0,__PYX_ERR(1, 1541, __pyx_L1_error))
  __pyx_v_x143 = (__pyx_v_r * __pyx_v_x120);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1542
 *         cdef double x142 = x141**(-4)
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913             # <<<<<<<<<<<<<<
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 */
  __Pyx_TraceLine(1542,0,__PYX_ERR(1, 1542, __pyx_L1_error))
  __pyx_v_x144 = (((-630116198.873299 * __pyx_v_nu) - (197773496.793534 * __pyx_v_x28)) + 5805304367.87913);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1543
 *         cdef double x143 = r*x120
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20             # <<<<<<<<<<<<<<
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146
 */
  __Pyx_TraceLine(1543,0,__PYX_ERR(1, 1543, __pyx_L1_error))
  __pyx_v_x145 = (__pyx_v_x144 * __pyx_v_x20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1544
 *         cdef double x144 = -630116198.873299*nu - 197773496.793534*x28 + 5805304367.87913
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329             # <<<<<<<<<<<<<<
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 */
  __Pyx_TraceLine(1544,0,__PYX_ERR(1, 1544, __pyx_L1_error))
  __pyx_v_x146 = (((-2675575.66847905 * __pyx_v_nu) - (138240.0 * __pyx_v_x28)) - 5278341.3229329);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1545
 *         cdef double x145 = x144*x20
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146             # <<<<<<<<<<<<<<
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 */
  __Pyx_TraceLine(1545,0,__PYX_ERR(1, 1545, __pyx_L1_error))
  __pyx_v_x147 = (__pyx_v_x0 * __pyx_v_x146);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1546
 *         cdef double x146 = -2675575.66847905*nu - 138240.0*x28 - 5278341.3229329
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)             # <<<<<<<<<<<<<<
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149
 */
  __Pyx_TraceLine(1546,0,__PYX_ERR(1, 1546, __pyx_L1_error))
  __pyx_v_x148 = (__pyx_v_nu * ((((-2510664218.28128 * __pyx_v_nu) - (42636451.6032331 * __pyx_v_x28)) + (14515200.0 * __pyx_v_x83)) + 1002013764.01019));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1547
 *         cdef double x147 = x0*x146
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0             # <<<<<<<<<<<<<<
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0
 */
  __Pyx_TraceLine(1547,0,__PYX_ERR(1, 1547, __pyx_L1_error))
  __pyx_v_x149 = (((((43393301259014.8 * __pyx_v_nu) + (43133561885859.3 * __pyx_v_x28)) + (5927865218923.02 * __pyx_v_x83)) + (86618264430493.3 * pow((1.0 - (0.496948781616935 * __pyx_v_nu)), 2.0))) + 188440788778196.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1548
 *         cdef double x148 = nu*(-2510664218.28128*nu - 42636451.6032331*x28 + 14515200.0*x83 + 1002013764.01019)
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149             # <<<<<<<<<<<<<<
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151
 */
  __Pyx_TraceLine(1548,0,__PYX_ERR(1, 1548, __pyx_L1_error))
  __pyx_v_x150 = (__pyx_v_r * __pyx_v_x149);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1549
 *         cdef double x149 = 43393301259014.8*nu + 43133561885859.3*x28 + 5927865218923.02*x83 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0             # <<<<<<<<<<<<<<
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 */
  __Pyx_TraceLine(1549,0,__PYX_ERR(1, 1549, __pyx_L1_error))
  __pyx_v_x151 = ((14700.0 * __pyx_v_nu) + 42911.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1550
 *         cdef double x150 = r*x149
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151             # <<<<<<<<<<<<<<
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93
 */
  __Pyx_TraceLine(1550,0,__PYX_ERR(1, 1550, __pyx_L1_error))
  __pyx_v_x152 = (283115520.0 * __pyx_v_x151);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1551
 *         cdef double x151 = 14700.0*nu + 42911.0
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20             # <<<<<<<<<<<<<<
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 */
  __Pyx_TraceLine(1551,0,__PYX_ERR(1, 1551, __pyx_L1_error))
  __pyx_v_x153 = (((((-1698693120.0 * __pyx_v_nu) * ((11592.0 * __pyx_v_nu) + 69847.0)) + ((49152.0 * __pyx_v_r) * (((409207698.136075 * __pyx_v_nu) + (102574080.0 * __pyx_v_x28)) - 2119671837.36038))) + (879923036160.0 * __pyx_v_x0)) + (__pyx_v_x152 * __pyx_v_x20));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1552
 *         cdef double x152 = 283115520.0*x151
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93             # <<<<<<<<<<<<<<
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r
 */
  __Pyx_TraceLine(1552,0,__PYX_ERR(1, 1552, __pyx_L1_error))
  __pyx_v_x154 = (__pyx_v_x153 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1553
 *         cdef double x153 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*x28 - 2119671837.36038) + 879923036160.0*x0 + x152*x20
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2             # <<<<<<<<<<<<<<
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20
 */
  __Pyx_TraceLine(1553,0,__PYX_ERR(1, 1553, __pyx_L1_error))
  __pyx_v_x155 = pow((((((__pyx_v_x143 - (1.59227685093395e-9 * __pyx_v_x145)) - (1.67189069348064e-7 * __pyx_v_x147)) + (9.55366110560367e-9 * __pyx_v_x148)) + (1.72773095804465e-13 * __pyx_v_x150)) + (1.72773095804465e-13 * __pyx_v_x154)), 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1554
 *         cdef double x154 = x153*x93
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r             # <<<<<<<<<<<<<<
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28
 */
  __Pyx_TraceLine(1554,0,__PYX_ERR(1, 1554, __pyx_L1_error))
  __pyx_v_x156 = (__pyx_v_nu * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1555
 *         cdef double x155 = (x143 - 1.59227685093395e-9*x145 - 1.67189069348064e-7*x147 + 9.55366110560367e-9*x148 + 1.72773095804465e-13*x150 + 1.72773095804465e-13*x154)**2
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20             # <<<<<<<<<<<<<<
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151
 */
  __Pyx_TraceLine(1555,0,__PYX_ERR(1, 1555, __pyx_L1_error))
  __pyx_v_x157 = (__pyx_v_nu * __pyx_v_x20);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1556
 *         cdef double x156 = nu*r
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28             # <<<<<<<<<<<<<<
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28
 */
  __Pyx_TraceLine(1556,0,__PYX_ERR(1, 1556, __pyx_L1_error))
  __pyx_v_x158 = (__pyx_v_r * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1557
 *         cdef double x157 = nu*x20
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151             # <<<<<<<<<<<<<<
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 */
  __Pyx_TraceLine(1557,0,__PYX_ERR(1, 1557, __pyx_L1_error))
  __pyx_v_x159 = (__pyx_v_r * __pyx_v_x151);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1558
 *         cdef double x158 = r*x28
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28             # <<<<<<<<<<<<<<
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 */
  __Pyx_TraceLine(1558,0,__PYX_ERR(1, 1558, __pyx_L1_error))
  __pyx_v_x160 = (__pyx_v_x20 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1559
 *         cdef double x159 = r*x151
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0             # <<<<<<<<<<<<<<
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93
 */
  __Pyx_TraceLine(1559,0,__PYX_ERR(1, 1559, __pyx_L1_error))
  __pyx_v_x161 = ((5041721180160.0 * __pyx_v_x28) - 104186110149937.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1560
 *         cdef double x160 = x20*x28
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20             # <<<<<<<<<<<<<<
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 */
  __Pyx_TraceLine(1560,0,__PYX_ERR(1, 1560, __pyx_L1_error))
  __pyx_v_x162 = ((((-25392914995744.3 * __pyx_v_nu) - (__pyx_v_r * __pyx_v_x152)) - __pyx_v_x161) - (879923036160.0 * __pyx_v_x20));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1561
 *         cdef double x161 = 5041721180160.0*x28 - 104186110149937.0
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93             # <<<<<<<<<<<<<<
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)
 */
  __Pyx_TraceLine(1561,0,__PYX_ERR(1, 1561, __pyx_L1_error))
  __pyx_v_x163 = (__pyx_v_x162 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1562
 *         cdef double x162 = -25392914995744.3*nu - r*x152 - x161 - 879923036160.0*x20
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1             # <<<<<<<<<<<<<<
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28
 */
  __Pyx_TraceLine(1562,0,__PYX_ERR(1, 1562, __pyx_L1_error))
  __pyx_v_x164 = (((((((((((((-0.0438084424460039 * __pyx_v_nu) - (0.143521050466841 * __pyx_v_r)) + (0.0210425293255724 * __pyx_v_x120)) + (0.0385795738434214 * __pyx_v_x156)) + (0.00941289164152486 * __pyx_v_x157)) + (0.00662650629087394 * __pyx_v_x158)) - (1.18824456940711e-6 * __pyx_v_x159)) + (0.000486339502879429 * __pyx_v_x160)) - (3.63558293513537e-15 * __pyx_v_x163)) + (0.0185696317637669 * __pyx_v_x20)) + (0.291062041428379 * __pyx_v_x28)) + (0.0244692826489756 * __pyx_v_x83)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1563
 *         cdef double x163 = x162*x93
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)             # <<<<<<<<<<<<<<
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 */
  __Pyx_TraceLine(1563,0,__PYX_ERR(1, 1563, __pyx_L1_error))
  __pyx_v_x165 = pow(__pyx_v_x164, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1564
 *         cdef double x164 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0210425293255724*x120 + 0.0385795738434214*x156 + 0.00941289164152486*x157 + 0.00662650629087394*x158 - 1.18824456940711e-6*x159 + 0.000486339502879429*x160 - 3.63558293513537e-15*x163 + 0.0185696317637669*x20 + 0.291062041428379*x28 + 0.0244692826489756*x83 + 1
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28             # <<<<<<<<<<<<<<
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 */
  __Pyx_TraceLine(1564,0,__PYX_ERR(1, 1564, __pyx_L1_error))
  __pyx_v_x166 = (4.0 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1565
 *         cdef double x165 = x164**(-2)
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)             # <<<<<<<<<<<<<<
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2
 */
  __Pyx_TraceLine(1565,0,__PYX_ERR(1, 1565, __pyx_L1_error))
  __pyx_v_x167 = (((((0.3125 * __pyx_v_delta) * __pyx_v_x56) * ((18.0 * __pyx_v_nu) - 1.0)) - ((0.03125 * (((-33.0 * __pyx_v_nu) + (32.0 * __pyx_v_x28)) - 5.0)) * (((5.0 * __pyx_v_x3) + (5.0 * __pyx_v_x6)) + (5.0 * __pyx_v_x7)))) - ((0.03125 * (((-5.0 * __pyx_v_nu) + __pyx_v_x166) + 1.0)) * (((-15.0 * __pyx_v_x3) + (15.0 * __pyx_v_x6)) + (15.0 * __pyx_v_x7))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1566
 *         cdef double x166 = 4.0*x28
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70             # <<<<<<<<<<<<<<
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2
 */
  __Pyx_TraceLine(1566,0,__PYX_ERR(1, 1566, __pyx_L1_error))
  __pyx_v_x168 = (((((__pyx_v_x115 * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1567
 *         cdef double x167 = 0.3125*delta*x56*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*x28 - 5.0)*(5.0*x3 + 5.0*x6 + 5.0*x7) - 0.03125*(-5.0*nu + x166 + 1.0)*(-15.0*x3 + 15.0*x6 + 15.0*x7)
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2             # <<<<<<<<<<<<<<
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)
 */
  __Pyx_TraceLine(1567,0,__PYX_ERR(1, 1567, __pyx_L1_error))
  __pyx_v_x169 = pow(__pyx_v_x114, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1568
 *         cdef double x168 = x115*x142*x155*x165*x167*x70
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2             # <<<<<<<<<<<<<<
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28
 */
  __Pyx_TraceLine(1568,0,__PYX_ERR(1, 1568, __pyx_L1_error))
  __pyx_v_x170 = pow(__pyx_v_prst, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1569
 *         cdef double x169 = x114**2
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)             # <<<<<<<<<<<<<<
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120
 */
  __Pyx_TraceLine(1569,0,__PYX_ERR(1, 1569, __pyx_L1_error))
  __pyx_v_x171 = pow(__pyx_v_x141, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1570
 *         cdef double x170 = prst**2
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28             # <<<<<<<<<<<<<<
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 */
  __Pyx_TraceLine(1570,0,__PYX_ERR(1, 1570, __pyx_L1_error))
  __pyx_v_x172 = (1822680546449.21 * __pyx_v_x28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1571
 *         cdef double x171 = x141**(-2)
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120             # <<<<<<<<<<<<<<
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)
 */
  __Pyx_TraceLine(1571,0,__PYX_ERR(1, 1571, __pyx_L1_error))
  __pyx_v_x173 = (5787938193408.0 * __pyx_v_x120);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1572
 *         cdef double x172 = 1822680546449.21*x28
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0             # <<<<<<<<<<<<<<
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu
 */
  __Pyx_TraceLine(1572,0,__PYX_ERR(1, 1572, __pyx_L1_error))
  __pyx_v_x174 = (((((((((((((-12049908701745.2 * __pyx_v_nu) + (__pyx_v_r * __pyx_v_x172)) - (39476764256925.6 * __pyx_v_r)) + (10611661054566.2 * __pyx_v_x156)) + (2589101062873.81 * __pyx_v_x157)) - (326837426.241486 * __pyx_v_x159)) + (133772083200.0 * __pyx_v_x160)) - __pyx_v_x163) + __pyx_v_x173) + (5107745331375.71 * __pyx_v_x20)) + (80059249540278.2 * __pyx_v_x28)) + (6730497718123.02 * __pyx_v_x83)) + 275059053208689.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1573
 *         cdef double x173 = 5787938193408.0*x120
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3
 */
  __Pyx_TraceLine(1573,0,__PYX_ERR(1, 1573, __pyx_L1_error))
  __pyx_v_x175 = pow(__pyx_v_x174, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1574
 *         cdef double x174 = -12049908701745.2*nu + r*x172 - 39476764256925.6*r + 10611661054566.2*x156 + 2589101062873.81*x157 - 326837426.241486*x159 + 133772083200.0*x160 - x163 + x173 + 5107745331375.71*x20 + 80059249540278.2*x28 + 6730497718123.02*x83 + 275059053208689.0
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 */
  __Pyx_TraceLine(1574,0,__PYX_ERR(1, 1574, __pyx_L1_error))
  __pyx_v_x176 = (4.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1575
 *         cdef double x175 = x174**(-1)
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3             # <<<<<<<<<<<<<<
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178
 */
  __Pyx_TraceLine(1575,0,__PYX_ERR(1, 1575, __pyx_L1_error))
  __pyx_v_x177 = (3.0 * __pyx_v_x3);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1576
 *         cdef double x176 = 4.0*nu
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7             # <<<<<<<<<<<<<<
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179
 */
  __Pyx_TraceLine(1576,0,__PYX_ERR(1, 1576, __pyx_L1_error))
  __pyx_v_x178 = ((3.0 * __pyx_v_x6) + (3.0 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1577
 *         cdef double x177 = 3.0*x3
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178             # <<<<<<<<<<<<<<
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 */
  __Pyx_TraceLine(1577,0,__PYX_ERR(1, 1577, __pyx_L1_error))
  __pyx_v_x179 = ((-__pyx_v_x177) + __pyx_v_x178);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1578
 *         cdef double x178 = 3.0*x6 + 3.0*x7
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179             # <<<<<<<<<<<<<<
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 */
  __Pyx_TraceLine(1578,0,__PYX_ERR(1, 1578, __pyx_L1_error))
  __pyx_v_x180 = (0.0625 * __pyx_v_x179);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1579
 *         cdef double x179 = -x177 + x178
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)             # <<<<<<<<<<<<<<
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 */
  __Pyx_TraceLine(1579,0,__PYX_ERR(1, 1579, __pyx_L1_error))
  __pyx_v_x181 = (__pyx_v_x57 * ((26.0 * __pyx_v_nu) + 449.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1580
 *         cdef double x180 = 0.0625*x179
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)             # <<<<<<<<<<<<<<
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 */
  __Pyx_TraceLine(1580,0,__PYX_ERR(1, 1580, __pyx_L1_error))
  __pyx_v_x182 = (__pyx_v_x9 * ((-1171.0 * __pyx_v_nu) - 861.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1581
 *         cdef double x181 = x57*(26.0*nu + 449.0)
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)             # <<<<<<<<<<<<<<
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)
 */
  __Pyx_TraceLine(1581,0,__PYX_ERR(1, 1581, __pyx_L1_error))
  __pyx_v_x183 = (__pyx_v_x103 * (((115.0 * __pyx_v_nu) + __pyx_v_x166) - 37.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1582
 *         cdef double x182 = x9*(-1171.0*nu - 861.0)
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154             # <<<<<<<<<<<<<<
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185
 */
  __Pyx_TraceLine(1582,0,__PYX_ERR(1, 1582, __pyx_L1_error))
  __pyx_v_x184 = ((((((5787938193408.0 * __pyx_v_x143) - (9216.0 * __pyx_v_x145)) - (967680.0 * __pyx_v_x147)) + (55296.0 * __pyx_v_x148)) + __pyx_v_x150) + __pyx_v_x154);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1583
 *         cdef double x183 = x103*(115.0*nu + x166 - 37.0)
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34
 */
  __Pyx_TraceLine(1583,0,__PYX_ERR(1, 1583, __pyx_L1_error))
  __pyx_v_x185 = pow(__pyx_v_x184, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1584
 *         cdef double x184 = 5787938193408.0*x143 - 9216.0*x145 - 967680.0*x147 + 55296.0*x148 + x150 + x154
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185             # <<<<<<<<<<<<<<
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138
 */
  __Pyx_TraceLine(1584,0,__PYX_ERR(1, 1584, __pyx_L1_error))
  __pyx_v_x186 = (__pyx_v_x174 * __pyx_v_x185);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1585
 *         cdef double x185 = x184**(-1)
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34             # <<<<<<<<<<<<<<
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188
 */
  __Pyx_TraceLine(1585,0,__PYX_ERR(1, 1585, __pyx_L1_error))
  __pyx_v_x187 = (__pyx_v_x186 * __pyx_v_x34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1586
 *         cdef double x186 = x174*x185
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138             # <<<<<<<<<<<<<<
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 */
  __Pyx_TraceLine(1586,0,__PYX_ERR(1, 1586, __pyx_L1_error))
  __pyx_v_x188 = (__pyx_v_x119 * __pyx_v_x138);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1587
 *         cdef double x187 = x186*x34
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188             # <<<<<<<<<<<<<<
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 */
  __Pyx_TraceLine(1587,0,__PYX_ERR(1, 1587, __pyx_L1_error))
  __pyx_v_x189 = (7680.0 * __pyx_v_x188);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1588
 *         cdef double x188 = x119*x138
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))             # <<<<<<<<<<<<<<
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 */
  __Pyx_TraceLine(1588,0,__PYX_ERR(1, 1588, __pyx_L1_error))
  __pyx_v_x190 = (((__pyx_v_x116 + (__pyx_v_x14 * (((0.03125 * __pyx_v_x181) + (0.015625 * __pyx_v_x182)) + (0.015625 * __pyx_v_x183)))) + (__pyx_v_x187 * __pyx_v_x189)) + (__pyx_v_x24 * (((__pyx_v_x180 * (__pyx_v_x176 - 1.0)) - (2.625 * __pyx_v_x57)) + (__pyx_v_x9 * (__pyx_v_x15 + 2.8125)))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1589
 *         cdef double x189 = 7680.0*x188
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190             # <<<<<<<<<<<<<<
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60
 */
  __Pyx_TraceLine(1589,0,__PYX_ERR(1, 1589, __pyx_L1_error))
  __pyx_v_x191 = (((((__pyx_v_x169 * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x190);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1590
 *         cdef double x190 = x116 + x14*(0.03125*x181 + 0.015625*x182 + 0.015625*x183) + x187*x189 + x24*(x180*(x176 - 1.0) - 2.625*x57 + x9*(x15 + 2.8125))
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 */
  __Pyx_TraceLine(1590,0,__PYX_ERR(1, 1590, __pyx_L1_error))
  __pyx_v_x192 = ((((((((((((((((-__pyx_v_x1) * __pyx_v_x100) + (__pyx_v_x112 * __pyx_v_x61)) + ((1.27277314139085e-19 * __pyx_v_x113) * __pyx_v_x168)) + ((1.69542100694444e-8 * __pyx_v_x191) * __pyx_v_x35)) + (__pyx_v_x21 * __pyx_v_x92)) + ((147.443752990146 * __pyx_v_x70) * __pyx_v_x72)) + (__pyx_v_x70 * __pyx_v_x82)) - ((11.3175085791863 * __pyx_v_x73) * __pyx_v_x75)) + (__pyx_v_x73 * __pyx_v_x87)) + ((1.48275342024365 * __pyx_v_x76) * __pyx_v_x78)) + (__pyx_v_x79 * __pyx_v_x80)) + (__pyx_v_x84 * __pyx_v_x85)) + (__pyx_v_x89 * __pyx_v_x90)) + (__pyx_v_x94 * __pyx_v_x95)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1591
 *         cdef double x191 = x169*x170*x171*x175*x184*x190
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60             # <<<<<<<<<<<<<<
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)
 */
  __Pyx_TraceLine(1591,0,__PYX_ERR(1, 1591, __pyx_L1_error))
  __pyx_v_x193 = (2.0 * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1592
 *         cdef double x192 = -x1*x100 + x112*x61 + 1.27277314139085e-19*x113*x168 + 1.69542100694444e-8*x191*x35 + x21*x92 + 147.443752990146*x70*x72 + x70*x82 - 11.3175085791863*x73*x75 + x73*x87 + 1.48275342024365*x76*x78 + x79*x80 + x84*x85 + x89*x90 + x94*x95 + 1.0
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)
 */
  __Pyx_TraceLine(1592,0,__PYX_ERR(1, 1592, __pyx_L1_error))
  __pyx_v_x194 = ((__pyx_v_x116 * (__pyx_v_x193 + 1.0)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1593
 *         cdef double x193 = 2.0*x60
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)             # <<<<<<<<<<<<<<
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 */
  __Pyx_TraceLine(1593,0,__PYX_ERR(1, 1593, __pyx_L1_error))
  __pyx_v_x195 = pow(__pyx_v_x194, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1594
 *         cdef double x194 = x116*(x193 + 1.0) + 1.0
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)             # <<<<<<<<<<<<<<
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu
 */
  __Pyx_TraceLine(1594,0,__PYX_ERR(1, 1594, __pyx_L1_error))
  __pyx_v_x196 = (__pyx_v_x103 * (__pyx_v_x176 + 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1595
 *         cdef double x195 = x194**(-1)
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 */
  __Pyx_TraceLine(1595,0,__PYX_ERR(1, 1595, __pyx_L1_error))
  __pyx_v_x197 = (__pyx_v_x57 * ((2.0 * __pyx_v_nu) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1596
 *         cdef double x196 = x103*(x176 + 1.0)
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu             # <<<<<<<<<<<<<<
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)
 */
  __Pyx_TraceLine(1596,0,__PYX_ERR(1, 1596, __pyx_L1_error))
  __pyx_v_x198 = (7.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1597
 *         cdef double x197 = x57*(2.0*nu - 3.0)
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7             # <<<<<<<<<<<<<<
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)
 */
  __Pyx_TraceLine(1597,0,__PYX_ERR(1, 1597, __pyx_L1_error))
  __pyx_v_x199 = (((0.390625 * __pyx_v_x3) + (0.390625 * __pyx_v_x6)) + (0.390625 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1598
 *         cdef double x198 = 7.0*nu
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)             # <<<<<<<<<<<<<<
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 */
  __Pyx_TraceLine(1598,0,__PYX_ERR(1, 1598, __pyx_L1_error))
  __pyx_v_x200 = (__pyx_v_x49 * (__pyx_v_x198 - 31.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1599
 *         cdef double x199 = 0.390625*x3 + 0.390625*x6 + 0.390625*x7
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 */
  __Pyx_TraceLine(1599,0,__PYX_ERR(1, 1599, __pyx_L1_error))
  __pyx_v_x201 = (__pyx_v_x104 * ((-__pyx_v_x176) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1600
 *         cdef double x200 = x49*(x198 - 31.0)
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)             # <<<<<<<<<<<<<<
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203
 */
  __Pyx_TraceLine(1600,0,__PYX_ERR(1, 1600, __pyx_L1_error))
  __pyx_v_x202 = (__pyx_v_x58 * ((68.0 * __pyx_v_nu) - 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1601
 *         cdef double x201 = x104*(-x176 - 3.0)
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65             # <<<<<<<<<<<<<<
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)
 */
  __Pyx_TraceLine(1601,0,__PYX_ERR(1, 1601, __pyx_L1_error))
  __pyx_v_x203 = ((((((__pyx_v_x116 + (__pyx_v_x14 * (((0.125 * __pyx_v_x200) + (0.0625 * __pyx_v_x201)) + (2.0625 * __pyx_v_x58)))) + (__pyx_v_x14 * (((((0.125 * __pyx_v_x196) + (1.125 * __pyx_v_x3)) - (1.25 * __pyx_v_x57)) + (1.125 * __pyx_v_x6)) + (1.125 * __pyx_v_x7)))) + (7680.0 * __pyx_v_x140)) + (__pyx_v_x35 * (((__pyx_v_x109 * (((2870.0 * __pyx_v_nu) - __pyx_v_x107) + 561.0)) + (__pyx_v_x111 * (((1166.0 * __pyx_v_nu) - (328.0 * __pyx_v_x28)) - 171.0))) + (0.125 * __pyx_v_x202)))) + (__pyx_v_x35 * ((((0.015625 * __pyx_v_x179) * (((-27.0 * __pyx_v_nu) + (28.0 * __pyx_v_x28)) - 3.0)) - (1.21875 * __pyx_v_x197)) - (__pyx_v_x199 * (__pyx_v_x198 + 9.0))))) + __pyx_v_x65);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1602
 *         cdef double x202 = x58*(68.0*nu - 1.0)
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203             # <<<<<<<<<<<<<<
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 */
  __Pyx_TraceLine(1602,0,__PYX_ERR(1, 1602, __pyx_L1_error))
  __pyx_v_x204 = (__pyx_v_x195 * __pyx_v_x203);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1603
 *         cdef double x203 = x116 + x14*(0.125*x200 + 0.0625*x201 + 2.0625*x58) + x14*(0.125*x196 + 1.125*x3 - 1.25*x57 + 1.125*x6 + 1.125*x7) + 7680.0*x140 + x35*(x109*(2870.0*nu - x107 + 561.0) + x111*(1166.0*nu - 328.0*x28 - 171.0) + 0.125*x202) + x35*(0.015625*x179*(-27.0*nu + 28.0*x28 - 3.0) - 1.21875*x197 - x199*(x198 + 9.0)) + x65
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)             # <<<<<<<<<<<<<<
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0
 */
  __Pyx_TraceLine(1603,0,__PYX_ERR(1, 1603, __pyx_L1_error))
  __pyx_v_x205 = pow((__pyx_v_x192 * __pyx_v_x204), -0.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1604
 *         cdef double x204 = x195*x203
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7             # <<<<<<<<<<<<<<
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2
 */
  __Pyx_TraceLine(1604,0,__PYX_ERR(1, 1604, __pyx_L1_error))
  __pyx_v_x206 = (((-2.0 * __pyx_v_x3) - (2.0 * __pyx_v_x6)) - (2.0 * __pyx_v_x7));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1605
 *         cdef double x205 = (x192*x204)**(-0.5)
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0             # <<<<<<<<<<<<<<
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)
 */
  __Pyx_TraceLine(1605,0,__PYX_ERR(1, 1605, __pyx_L1_error))
  __pyx_v_x207 = ((4.0 * __pyx_v_x60) + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1606
 *         cdef double x206 = -2.0*x3 - 2.0*x6 - 2.0*x7
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2             # <<<<<<<<<<<<<<
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 */
  __Pyx_TraceLine(1606,0,__PYX_ERR(1, 1606, __pyx_L1_error))
  __pyx_v_x208 = (4.5 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1607
 *         cdef double x207 = 4.0*x60 + 2.0
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)             # <<<<<<<<<<<<<<
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60
 */
  __Pyx_TraceLine(1607,0,__PYX_ERR(1, 1607, __pyx_L1_error))
  __pyx_v_x209 = pow(__pyx_v_r, -6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1608
 *         cdef double x208 = 4.5*X_2
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0             # <<<<<<<<<<<<<<
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0
 */
  __Pyx_TraceLine(1608,0,__PYX_ERR(1, 1608, __pyx_L1_error))
  __pyx_v_x210 = (((((-6572428.80109422 * __pyx_v_nu) + ((2048.0 * __pyx_v_x118) * __pyx_v_x79)) + (688128.0 * __pyx_v_x128)) + (1300341.64885296 * __pyx_v_x28)) + 1720320.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1609
 *         cdef double x209 = r**(-6)
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60             # <<<<<<<<<<<<<<
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 */
  __Pyx_TraceLine(1609,0,__PYX_ERR(1, 1609, __pyx_L1_error))
  __pyx_v_x211 = (__pyx_v_x28 * __pyx_v_x60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1610
 *         cdef double x210 = -6572428.80109422*nu + 2048.0*x118*x79 + 688128.0*x128 + 1300341.64885296*x28 + 1720320.0
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0             # <<<<<<<<<<<<<<
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 */
  __Pyx_TraceLine(1610,0,__PYX_ERR(1, 1610, __pyx_L1_error))
  __pyx_v_x212 = (4.0 * __pyx_v_x0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1611
 *         cdef double x211 = x28*x60
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0             # <<<<<<<<<<<<<<
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214
 */
  __Pyx_TraceLine(1611,0,__PYX_ERR(1, 1611, __pyx_L1_error))
  __pyx_v_x213 = ((__pyx_v_x130 + (6.0 * __pyx_v_x20)) + 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1612
 *         cdef double x212 = 4.0*x0
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2             # <<<<<<<<<<<<<<
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24
 */
  __Pyx_TraceLine(1612,0,__PYX_ERR(1, 1612, __pyx_L1_error))
  __pyx_v_x214 = (((1.31621673590926e-19 * __pyx_v_x119) * ((((((((53760.0 * __pyx_v_nu) * (((((((((3740417.71815805 * __pyx_v_r) + (1057984.42953951 * __pyx_v_x0)) - (938918.400156317 * __pyx_v_x13)) + (__pyx_v_x132 * (__pyx_v_x212 + __pyx_v_x213))) + (__pyx_v_x133 * ((((17058.7851094412 * __pyx_v_r) - (27409.3925547206 * __pyx_v_x0)) + (12794.0888320809 * __pyx_v_x20)) + 13218.7851094412))) + (__pyx_v_x135 * ((((1396.0 * __pyx_v_x0) + __pyx_v_x134) + (5778.0 * __pyx_v_x20)) + 7704.0))) + (__pyx_v_x136 * ((-__pyx_v_x212) + __pyx_v_x213))) + (2115968.85907902 * __pyx_v_x20)) + 2888096.47013111)) + ((32768.0 * __pyx_v_x127) * __pyx_v_x79)) + ((32768.0 * __pyx_v_x128) * (((((((-1791904.9465589 * __pyx_v_nu) + ((3840.0 * __pyx_v_r) * __pyx_v_x124)) + ((2880.0 * __pyx_v_x125) * __pyx_v_x20)) + (1920.0 * __pyx_v_x126)) + (806400.0 * __pyx_v_x13)) - (725760.0 * __pyx_v_x28)) + 4143360.0))) + (66060288000.0 * __pyx_v_x13)) + ((135291469824.0 * __pyx_v_x211) * __pyx_v_x93)) + ((7.0 * __pyx_v_x28) * ((((((-117964800.0 * __pyx_v_a6) + (4129567622.65173 * __pyx_v_r)) - (12357002815.0752 * __pyx_v_x0)) + (7133302759.42198 * __pyx_v_x13)) - (18535504222.6128 * __pyx_v_x20)) + 122635399361.987))) + ((7.0 * __pyx_v_x83) * ((-5706642207.53758 * __pyx_v_r) - 26189434371.6082)))) / pow((((((((0.28004222119933 * __pyx_v_x121) + (4.63661586574928e-9 * __pyx_v_x122)) + (2.8978849160933e-11 * __pyx_v_x123)) + (1.35654132757922e-7 * __pyx_v_x129)) + (2.22557561555966e-7 * __pyx_v_x137)) + (0.0546957463279941 * __pyx_v_x34)) + __pyx_v_x88), 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1613
 *         cdef double x213 = x130 + 6.0*x20 + 8.0
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214             # <<<<<<<<<<<<<<
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93
 */
  __Pyx_TraceLine(1613,0,__PYX_ERR(1, 1613, __pyx_L1_error))
  __pyx_v_x215 = (__pyx_v_x13 * __pyx_v_x214);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1614
 *         cdef double x214 = 1.31621673590926e-19*x119*(53760.0*nu*(3740417.71815805*r + 1057984.42953951*x0 - 938918.400156317*x13 + x132*(x212 + x213) + x133*(17058.7851094412*r - 27409.3925547206*x0 + 12794.0888320809*x20 + 13218.7851094412) + x135*(1396.0*x0 + x134 + 5778.0*x20 + 7704.0) + x136*(-x212 + x213) + 2115968.85907902*x20 + 2888096.47013111) + 32768.0*x127*x79 + 32768.0*x128*(-1791904.9465589*nu + 3840.0*r*x124 + 2880.0*x125*x20 + 1920.0*x126 + 806400.0*x13 - 725760.0*x28 + 4143360.0) + 66060288000.0*x13 + 135291469824.0*x211*x93 + 7.0*x28*(-117964800.0*a6 + 4129567622.65173*r - 12357002815.0752*x0 + 7133302759.42198*x13 - 18535504222.6128*x20 + 122635399361.987) + 7.0*x83*(-5706642207.53758*r - 26189434371.6082))/(0.28004222119933*x121 + 4.63661586574928e-9*x122 + 2.8978849160933e-11*x123 + 1.35654132757922e-7*x129 + 2.22557561555966e-7*x137 + 0.0546957463279941*x34 + x88)**2
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24             # <<<<<<<<<<<<<<
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 */
  __Pyx_TraceLine(1614,0,__PYX_ERR(1, 1614, __pyx_L1_error))
  __pyx_v_x216 = (2.0 * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1615
 *         cdef double x215 = x13*x214
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93             # <<<<<<<<<<<<<<
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70
 */
  __Pyx_TraceLine(1615,0,__PYX_ERR(1, 1615, __pyx_L1_error))
  __pyx_v_x217 = (11575876386816.0 * __pyx_v_x93);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1616
 *         cdef double x216 = 2.0*x24
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1             # <<<<<<<<<<<<<<
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 */
  __Pyx_TraceLine(1616,0,__PYX_ERR(1, 1616, __pyx_L1_error))
  __pyx_v_x218 = (((((((((5807150888816.34 * __pyx_v_nu) + (10215490662751.4 * __pyx_v_r)) + (5178202125747.62 * __pyx_v_x156)) + (267544166400.0 * __pyx_v_x158)) - (__pyx_v_x162 * __pyx_v_x60)) + __pyx_v_x172) + (__pyx_v_x217 * __pyx_v_x60)) + (__pyx_v_x93 * (((4161798144000.0 * __pyx_v_nu) + (1759846072320.0 * __pyx_v_r)) + 12148770078720.0))) - 53501685054374.1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1617
 *         cdef double x217 = 11575876386816.0*x93
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70             # <<<<<<<<<<<<<<
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190
 */
  __Pyx_TraceLine(1617,0,__PYX_ERR(1, 1617, __pyx_L1_error))
  __pyx_v_x219 = ((((__pyx_v_x113 * __pyx_v_x115) * __pyx_v_x155) * __pyx_v_x167) * __pyx_v_x70);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1618
 *         cdef double x218 = 5807150888816.34*nu + 10215490662751.4*r + 5178202125747.62*x156 + 267544166400.0*x158 - x162*x60 + x172 + x217*x60 + x93*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)             # <<<<<<<<<<<<<<
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24
 */
  __Pyx_TraceLine(1618,0,__PYX_ERR(1, 1618, __pyx_L1_error))
  __pyx_v_x220 = ((((((((-18432.0 * __pyx_v_r) * __pyx_v_x144) - ((2903040.0 * __pyx_v_x146) * __pyx_v_x20)) + __pyx_v_x149) + (__pyx_v_x153 * __pyx_v_x60)) + __pyx_v_x173) + __pyx_v_x217) + (__pyx_v_x93 * ((((20113376778784.3 * __pyx_v_nu) + (566231040.0 * __pyx_v_x159)) + __pyx_v_x161) + (2639769108480.0 * __pyx_v_x20))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1619
 *         cdef double x219 = x113*x115*x155*x167*x70
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190             # <<<<<<<<<<<<<<
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210
 */
  __Pyx_TraceLine(1619,0,__PYX_ERR(1, 1619, __pyx_L1_error))
  __pyx_v_x221 = (__pyx_v_x170 * __pyx_v_x190);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1620
 *         cdef double x220 = -18432.0*r*x144 - 2903040.0*x146*x20 + x149 + x153*x60 + x173 + x217 + x93*(20113376778784.3*nu + 566231040.0*x159 + x161 + 2639769108480.0*x20)
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24             # <<<<<<<<<<<<<<
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 */
  __Pyx_TraceLine(1620,0,__PYX_ERR(1, 1620, __pyx_L1_error))
  __pyx_v_x222 = (__pyx_v_x206 * __pyx_v_x24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1621
 *         cdef double x221 = x170*x190
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210             # <<<<<<<<<<<<<<
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35
 */
  __Pyx_TraceLine(1621,0,__PYX_ERR(1, 1621, __pyx_L1_error))
  __pyx_v_x223 = (7680.0 * __pyx_v_x210);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1622
 *         cdef double x222 = x206*x24
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222             # <<<<<<<<<<<<<<
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3
 */
  __Pyx_TraceLine(1622,0,__PYX_ERR(1, 1622, __pyx_L1_error))
  __pyx_v_x224 = (((((30720.0 * __pyx_v_x0) * __pyx_v_x188) + (__pyx_v_x139 * __pyx_v_x223)) - __pyx_v_x215) + __pyx_v_x222);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1623
 *         cdef double x223 = 7680.0*x210
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35             # <<<<<<<<<<<<<<
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5
 */
  __Pyx_TraceLine(1623,0,__PYX_ERR(1, 1623, __pyx_L1_error))
  __pyx_v_x225 = (__pyx_v_x184 * __pyx_v_x35);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1624
 *         cdef double x224 = 30720.0*x0*x188 + x139*x223 - x215 + x222
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3             # <<<<<<<<<<<<<<
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7
 */
  __Pyx_TraceLine(1624,0,__PYX_ERR(1, 1624, __pyx_L1_error))
  __pyx_v_x226 = pow(__pyx_v_prst, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1625
 *         cdef double x225 = x184*x35
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5             # <<<<<<<<<<<<<<
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226
 */
  __Pyx_TraceLine(1625,0,__PYX_ERR(1, 1625, __pyx_L1_error))
  __pyx_v_x227 = pow(__pyx_v_prst, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1626
 *         cdef double x226 = prst**3
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7             # <<<<<<<<<<<<<<
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227
 */
  __Pyx_TraceLine(1626,0,__PYX_ERR(1, 1626, __pyx_L1_error))
  __pyx_v_x228 = pow(__pyx_v_prst, 7.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1627
 *         cdef double x227 = prst**5
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226             # <<<<<<<<<<<<<<
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 */
  __Pyx_TraceLine(1627,0,__PYX_ERR(1, 1627, __pyx_L1_error))
  __pyx_v_x229 = (4.0 * __pyx_v_x226);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1628
 *         cdef double x228 = prst**7
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227             # <<<<<<<<<<<<<<
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21
 */
  __Pyx_TraceLine(1628,0,__PYX_ERR(1, 1628, __pyx_L1_error))
  __pyx_v_x230 = (6.0 * __pyx_v_x227);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1629
 *         cdef double x229 = 4.0*x226
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)             # <<<<<<<<<<<<<<
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216
 */
  __Pyx_TraceLine(1629,0,__PYX_ERR(1, 1629, __pyx_L1_error))
  __pyx_v_x231 = ((__pyx_v_x203 * __pyx_v_x205) / ((__pyx_v_x116 * __pyx_v_x207) + 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1630
 *         cdef double x230 = 6.0*x227
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21             # <<<<<<<<<<<<<<
 *         cdef double x233 = pphi*x216
 *         cdef double x234 = 4.0*pphi**3*x14
 */
  __Pyx_TraceLine(1630,0,__PYX_ERR(1, 1630, __pyx_L1_error))
  __pyx_v_x232 = ((2.0 * __pyx_v_pphi) * __pyx_v_x21);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1631
 *         cdef double x231 = x203*x205/(x116*x207 + 2.0)
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216             # <<<<<<<<<<<<<<
 *         cdef double x234 = 4.0*pphi**3*x14
 * 
 */
  __Pyx_TraceLine(1631,0,__PYX_ERR(1, 1631, __pyx_L1_error))
  __pyx_v_x233 = (__pyx_v_pphi * __pyx_v_x216);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1632
 *         cdef double x232 = 2.0*pphi*x21
 *         cdef double x233 = pphi*x216
 *         cdef double x234 = 4.0*pphi**3*x14             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1632,0,__PYX_ERR(1, 1632, __pyx_L1_error))
  __pyx_v_x234 = ((4.0 * pow(__pyx_v_pphi, 3.0)) * __pyx_v_x14);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1637
 *         # Evaluate Hamiltonian
 *         cdef double H,xi
 *         H,xi,_,_,_,_,_,_,_  = self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)             # <<<<<<<<<<<<<<
 * 
 *         # Heff Jacobian expressions
 */
  __Pyx_TraceLine(1637,0,__PYX_ERR(1, 1637, __pyx_L1_error))
  __pyx_t_1 = ((struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._call(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 9)) {
      if (size > 9) __Pyx_RaiseTooManyValuesError(9);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 1637, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_16 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_12 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_11 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 8); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_16 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_12 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_11 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_10 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 8); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_7);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[9] = {&__pyx_t_2,&__pyx_t_13,&__pyx_t_16,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_7};
      for (i=0; i < 9; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(1, 1637, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[9] = {&__pyx_t_2,&__pyx_t_13,&__pyx_t_16,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_7};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_19 = Py_TYPE(__pyx_t_6)->tp_iternext;
    for (index=0; index < 9; index++) {
      PyObject* item = __pyx_t_19(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_6), 9) < 0) __PYX_ERR(1, 1637, __pyx_L1_error)
    __pyx_t_19 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_19 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(1, 1637, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_20 = __pyx_PyFloat_AsDouble(__pyx_t_13); if (unlikely((__pyx_t_20 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_v_H = __pyx_t_18;
  __pyx_v_xi = __pyx_t_20;
  __pyx_v__ = __pyx_t_16;
  __pyx_t_16 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_12);
  __pyx_t_12 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_11);
  __pyx_t_11 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_8);
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1640
 * 
 *         # Heff Jacobian expressions
 *         cdef double dHeffdr = x12*(-dSO*x14*x15*x19 + x19*(x24*(1.25*x3 - 0.75*x48 - 0.416666666666667*x57 + 1.66666666666667*x58 + 1.25*x6 + 1.25*x7) - x31*x55) + x19*(-x21*x22 - x23*x27 - x24*(-11.0625*nu + 1.13541666666667*x28 - 0.15625) - x29*x31 - x32*x37) + x43*(0.375*x14*x49*x50 - x24*(0.0833333333333333*x51 + x52)) + x44*(-x21*x38 - x24*(-0.0625*nu + 1.07291666666667*x28 + 0.15625) - x27*x39 - x31*x40 - x37*x41)) + 0.5*x205*(x192*x195*(30720.0*x0*x119*x138 + 7680.0*x13*x138*x210 - x14*(x177 + x178 - 3.0*x48) + x206*x24 - x209*(x109*(14350.0*nu - 1320.0*x28 + 2805.0) + x111*(5830.0*nu - 1640.0*x28 - 855.0) + 0.625*x202) - x209*(0.015625*x179*(-135.0*nu + 140.0*x28 - 15.0) - 6.09375*x197 - x199*(35.0*nu + 45.0)) - x215 - x35*(0.5*x200 + 0.25*x201 + 8.25*x58) - x35*(0.5*x196 + x2*x208 - 5.0*x57 + 4.5*x6 + 4.5*x7)) - x192*x203*(x14*x206 - x207*x24*x9)/x194**2 + x204*(-663.496888455656*nu*r**(-5.5)*x70 - nu*x21*x80 + 39.6112800271521*nu*x71*x73 + x1*x21*x25*x97*x98 + x1*x25*x60*x98*(r*x9 + x10 + x212)/x96**2 - x100 - x112*x27 + 7.59859378406358e-45*x113*x115*x142*x165*x167*x184*x220*x70 + 6.78168402777778e-8*x114*x14*x170*x171*x175*x184*x190 + x14*x70*(118.4*x211 - 51.6952380952381*x79) - 9.25454462627843e-34*x142*x218*x219/x164**3 - 2.24091649004576e-37*x165*x169*x171*x184*x218*x221*x35 + 1.69542100694444e-8*x169*x170*x171*x175*x184*x35*(x138*x187*x223 - x14*(x180*(12.0*nu - 3.0) - 7.875*x57 + x9*(9.0*nu + 8.4375)) + 38400.0*x140*x186 + x185*x189*x218*x34 - x187*x214 + x222 - x35*(0.125*x181 + 0.0625*x182 + 0.0625*x183) - 2.29252167428035e-22*x174*x188*x220*x34/x155) + 1.69542100694444e-8*x169*x170*x171*x175*x190*x220*x35 - 8.47710503472222e-8*x191*x209 + x21*x25*(-x14*(x101*(-5.25*nu - 2.8125) + x102*(2.25 - x208) + x105*(0.5625 - 2.25*nu)) + x24*x49 - x35*(0.03125*x104*x110 - 0.25*x106 + 0.0104166666666667*x108*x49)) - x216*x92 - x30*x73*x89 - x36*x70*x94 - 3.70688355060912*x74*x78 - 2.0*x81*x85 - 3.0*x84*x95 - 2.0*x86*x90 - 4.41515887225116e-12*x169*x175*x221*x224*x225/x141**3 - 6.62902677807736e-23*x165*x219*x224/x141**5 + 1.01821851311268e-18*x114**3*x142*x155*x165*x167*x70/r**12 - 1.65460508380811e-18*x168/r**14)) - (3.0*x20 + x9)*(pphi*x63 + pphi*x67 + pphi*x69 + x19*x59 + x43*x64)/x11**2             # <<<<<<<<<<<<<<
 * 
 *         cdef double  dHeffdphi = 0
 */
  __Pyx_TraceLine(1640,0,__PYX_ERR(1, 1640, __pyx_L1_error))
  __pyx_v_dHeffdr = (((__pyx_v_x12 * ((((((((-__pyx_v_dSO) * __pyx_v_x14) * __pyx_v_x15) * __pyx_v_x19) + (__pyx_v_x19 * ((__pyx_v_x24 * ((((((1.25 * __pyx_v_x3) - (0.75 * __pyx_v_x48)) - (0.416666666666667 * __pyx_v_x57)) + (1.66666666666667 * __pyx_v_x58)) + (1.25 * __pyx_v_x6)) + (1.25 * __pyx_v_x7))) - (__pyx_v_x31 * __pyx_v_x55)))) + (__pyx_v_x19 * ((((((-__pyx_v_x21) * __pyx_v_x22) - (__pyx_v_x23 * __pyx_v_x27)) - (__pyx_v_x24 * (((-11.0625 * __pyx_v_nu) + (1.13541666666667 * __pyx_v_x28)) - 0.15625))) - (__pyx_v_x29 * __pyx_v_x31)) - (__pyx_v_x32 * __pyx_v_x37)))) + (__pyx_v_x43 * ((((0.375 * __pyx_v_x14) * __pyx_v_x49) * __pyx_v_x50) - (__pyx_v_x24 * ((0.0833333333333333 * __pyx_v_x51) + __pyx_v_x52))))) + (__pyx_v_x44 * ((((((-__pyx_v_x21) * __pyx_v_x38) - (__pyx_v_x24 * (((-0.0625 * __pyx_v_nu) + (1.07291666666667 * __pyx_v_x28)) + 0.15625))) - (__pyx_v_x27 * __pyx_v_x39)) - (__pyx_v_x31 * __pyx_v_x40)) - (__pyx_v_x37 * __pyx_v_x41))))) + ((0.5 * __pyx_v_x205) * ((((__pyx_v_x192 * __pyx_v_x195) * (((((((((((30720.0 * __pyx_v_x0) * __pyx_v_x119) * __pyx_v_x138) + (((7680.0 * __pyx_v_x13) * __pyx_v_x138) * __pyx_v_x210)) - (__pyx_v_x14 * ((__pyx_v_x177 + __pyx_v_x178) - (3.0 * __pyx_v_x48)))) + (__pyx_v_x206 * __pyx_v_x24)) - (__pyx_v_x209 * (((__pyx_v_x109 * (((14350.0 * __pyx_v_nu) - (1320.0 * __pyx_v_x28)) + 2805.0)) + (__pyx_v_x111 * (((5830.0 * __pyx_v_nu) - (1640.0 * __pyx_v_x28)) - 855.0))) + (0.625 * __pyx_v_x202)))) - (__pyx_v_x209 * ((((0.015625 * __pyx_v_x179) * (((-135.0 * __pyx_v_nu) + (140.0 * __pyx_v_x28)) - 15.0)) - (6.09375 * __pyx_v_x197)) - (__pyx_v_x199 * ((35.0 * __pyx_v_nu) + 45.0))))) - __pyx_v_x215) - (__pyx_v_x35 * (((0.5 * __pyx_v_x200) + (0.25 * __pyx_v_x201)) + (8.25 * __pyx_v_x58)))) - (__pyx_v_x35 * (((((0.5 * __pyx_v_x196) + (__pyx_v_x2 * __pyx_v_x208)) - (5.0 * __pyx_v_x57)) + (4.5 * __pyx_v_x6)) + (4.5 * __pyx_v_x7))))) - (((__pyx_v_x192 * __pyx_v_x203) * ((__pyx_v_x14 * __pyx_v_x206) - ((__pyx_v_x207 * __pyx_v_x24) * __pyx_v_x9))) / pow(__pyx_v_x194, 2.0))) + (__pyx_v_x204 * (((((((((((((((((((((((((((((-663.496888455656 * __pyx_v_nu) * pow(__pyx_v_r, -5.5)) * __pyx_v_x70) - ((__pyx_v_nu * __pyx_v_x21) * __pyx_v_x80)) + (((39.6112800271521 * __pyx_v_nu) * __pyx_v_x71) * __pyx_v_x73)) + ((((__pyx_v_x1 * __pyx_v_x21) * __pyx_v_x25) * __pyx_v_x97) * __pyx_v_x98)) + (((((__pyx_v_x1 * __pyx_v_x25) * __pyx_v_x60) * __pyx_v_x98) * (((__pyx_v_r * __pyx_v_x9) + __pyx_v_x10) + __pyx_v_x212)) / pow(__pyx_v_x96, 2.0))) - __pyx_v_x100) - (__pyx_v_x112 * __pyx_v_x27)) + ((((((((7.59859378406358e-45 * __pyx_v_x113) * __pyx_v_x115) * __pyx_v_x142) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x184) * __pyx_v_x220) * __pyx_v_x70)) + (((((((6.78168402777778e-8 * __pyx_v_x114) * __pyx_v_x14) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x190)) + ((__pyx_v_x14 * __pyx_v_x70) * ((118.4 * __pyx_v_x211) - (51.6952380952381 * __pyx_v_x79)))) - ((((9.25454462627843e-34 * __pyx_v_x142) * __pyx_v_x218) * __pyx_v_x219) / pow(__pyx_v_x164, 3.0))) - (((((((2.24091649004576e-37 * __pyx_v_x165) * __pyx_v_x169) * __pyx_v_x171) * __pyx_v_x184) * __pyx_v_x218) * __pyx_v_x221) * __pyx_v_x35)) + (((((((1.69542100694444e-8 * __pyx_v_x169) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x184) * __pyx_v_x35) * (((((((((__pyx_v_x138 * __pyx_v_x187) * __pyx_v_x223) - (__pyx_v_x14 * (((__pyx_v_x180 * ((12.0 * __pyx_v_nu) - 3.0)) - (7.875 * __pyx_v_x57)) + (__pyx_v_x9 * ((9.0 * __pyx_v_nu) + 8.4375))))) + ((38400.0 * __pyx_v_x140) * __pyx_v_x186)) + (((__pyx_v_x185 * __pyx_v_x189) * __pyx_v_x218) * __pyx_v_x34)) - (__pyx_v_x187 * __pyx_v_x214)) + __pyx_v_x222) - (__pyx_v_x35 * (((0.125 * __pyx_v_x181) + (0.0625 * __pyx_v_x182)) + (0.0625 * __pyx_v_x183)))) - (((((2.29252167428035e-22 * __pyx_v_x174) * __pyx_v_x188) * __pyx_v_x220) * __pyx_v_x34) / __pyx_v_x155)))) + (((((((1.69542100694444e-8 * __pyx_v_x169) * __pyx_v_x170) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x190) * __pyx_v_x220) * __pyx_v_x35)) - ((8.47710503472222e-8 * __pyx_v_x191) * __pyx_v_x209)) + ((__pyx_v_x21 * __pyx_v_x25) * ((((-__pyx_v_x14) * (((__pyx_v_x101 * ((-5.25 * __pyx_v_nu) - 2.8125)) + (__pyx_v_x102 * (2.25 - __pyx_v_x208))) + (__pyx_v_x105 * (0.5625 - (2.25 * __pyx_v_nu))))) + (__pyx_v_x24 * __pyx_v_x49)) - (__pyx_v_x35 * ((((0.03125 * __pyx_v_x104) * __pyx_v_x110) - (0.25 * __pyx_v_x106)) + ((0.0104166666666667 * __pyx_v_x108) * __pyx_v_x49)))))) - (__pyx_v_x216 * __pyx_v_x92)) - ((__pyx_v_x30 * __pyx_v_x73) * __pyx_v_x89)) - ((__pyx_v_x36 * __pyx_v_x70) * __pyx_v_x94)) - ((3.70688355060912 * __pyx_v_x74) * __pyx_v_x78)) - ((2.0 * __pyx_v_x81) * __pyx_v_x85)) - ((3.0 * __pyx_v_x84) * __pyx_v_x95)) - ((2.0 * __pyx_v_x86) * __pyx_v_x90)) - ((((((4.41515887225116e-12 * __pyx_v_x169) * __pyx_v_x175) * __pyx_v_x221) * __pyx_v_x224) * __pyx_v_x225) / pow(__pyx_v_x141, 3.0))) - ((((6.62902677807736e-23 * __pyx_v_x165) * __pyx_v_x219) * __pyx_v_x224) / pow(__pyx_v_x141, 5.0))) + (((((((1.01821851311268e-18 * pow(__pyx_v_x114, 3.0)) * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x70) / pow(__pyx_v_r, 12.0))) - ((1.65460508380811e-18 * __pyx_v_x168) / pow(__pyx_v_r, 14.0))))))) - ((((3.0 * __pyx_v_x20) + __pyx_v_x9) * (((((__pyx_v_pphi * __pyx_v_x63) + (__pyx_v_pphi * __pyx_v_x67)) + (__pyx_v_pphi * __pyx_v_x69)) + (__pyx_v_x19 * __pyx_v_x59)) + (__pyx_v_x43 * __pyx_v_x64))) / pow(__pyx_v_x11, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1642
 *         cdef double dHeffdr = x12*(-dSO*x14*x15*x19 + x19*(x24*(1.25*x3 - 0.75*x48 - 0.416666666666667*x57 + 1.66666666666667*x58 + 1.25*x6 + 1.25*x7) - x31*x55) + x19*(-x21*x22 - x23*x27 - x24*(-11.0625*nu + 1.13541666666667*x28 - 0.15625) - x29*x31 - x32*x37) + x43*(0.375*x14*x49*x50 - x24*(0.0833333333333333*x51 + x52)) + x44*(-x21*x38 - x24*(-0.0625*nu + 1.07291666666667*x28 + 0.15625) - x27*x39 - x31*x40 - x37*x41)) + 0.5*x205*(x192*x195*(30720.0*x0*x119*x138 + 7680.0*x13*x138*x210 - x14*(x177 + x178 - 3.0*x48) + x206*x24 - x209*(x109*(14350.0*nu - 1320.0*x28 + 2805.0) + x111*(5830.0*nu - 1640.0*x28 - 855.0) + 0.625*x202) - x209*(0.015625*x179*(-135.0*nu + 140.0*x28 - 15.0) - 6.09375*x197 - x199*(35.0*nu + 45.0)) - x215 - x35*(0.5*x200 + 0.25*x201 + 8.25*x58) - x35*(0.5*x196 + x2*x208 - 5.0*x57 + 4.5*x6 + 4.5*x7)) - x192*x203*(x14*x206 - x207*x24*x9)/x194**2 + x204*(-663.496888455656*nu*r**(-5.5)*x70 - nu*x21*x80 + 39.6112800271521*nu*x71*x73 + x1*x21*x25*x97*x98 + x1*x25*x60*x98*(r*x9 + x10 + x212)/x96**2 - x100 - x112*x27 + 7.59859378406358e-45*x113*x115*x142*x165*x167*x184*x220*x70 + 6.78168402777778e-8*x114*x14*x170*x171*x175*x184*x190 + x14*x70*(118.4*x211 - 51.6952380952381*x79) - 9.25454462627843e-34*x142*x218*x219/x164**3 - 2.24091649004576e-37*x165*x169*x171*x184*x218*x221*x35 + 1.69542100694444e-8*x169*x170*x171*x175*x184*x35*(x138*x187*x223 - x14*(x180*(12.0*nu - 3.0) - 7.875*x57 + x9*(9.0*nu + 8.4375)) + 38400.0*x140*x186 + x185*x189*x218*x34 - x187*x214 + x222 - x35*(0.125*x181 + 0.0625*x182 + 0.0625*x183) - 2.29252167428035e-22*x174*x188*x220*x34/x155) + 1.69542100694444e-8*x169*x170*x171*x175*x190*x220*x35 - 8.47710503472222e-8*x191*x209 + x21*x25*(-x14*(x101*(-5.25*nu - 2.8125) + x102*(2.25 - x208) + x105*(0.5625 - 2.25*nu)) + x24*x49 - x35*(0.03125*x104*x110 - 0.25*x106 + 0.0104166666666667*x108*x49)) - x216*x92 - x30*x73*x89 - x36*x70*x94 - 3.70688355060912*x74*x78 - 2.0*x81*x85 - 3.0*x84*x95 - 2.0*x86*x90 - 4.41515887225116e-12*x169*x175*x221*x224*x225/x141**3 - 6.62902677807736e-23*x165*x219*x224/x141**5 + 1.01821851311268e-18*x114**3*x142*x155*x165*x167*x70/r**12 - 1.65460508380811e-18*x168/r**14)) - (3.0*x20 + x9)*(pphi*x63 + pphi*x67 + pphi*x69 + x19*x59 + x43*x64)/x11**2
 * 
 *         cdef double  dHeffdphi = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double  dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)
 */
  __Pyx_TraceLine(1642,0,__PYX_ERR(1, 1642, __pyx_L1_error))
  __pyx_v_dHeffdphi = 0.0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1644
 *         cdef double  dHeffdphi = 0
 * 
 *         cdef double  dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)             # <<<<<<<<<<<<<<
 * 
 *         cdef double  dHeffdpphi = x12*(x18*x59 + 2.0*x18*x68 + x19*(x23*x232 + x233*x29 + x234*x32) + x42*x64 - 0.25*x42*x66 + x44*(x232*x39 + x233*x40 + x234*x41) + x63 + x67 + x69) + x231*(-pphi*x1*x193*x99 + 2.0*pphi*x112*x21)
 */
  __Pyx_TraceLine(1644,0,__PYX_ERR(1, 1644, __pyx_L1_error))
  __pyx_v_dHeffdpr = (__pyx_v_x231 * ((((((((((((((11.8620273619492 * __pyx_v_nu) * __pyx_v_x228) * __pyx_v_x76) + ((((((3.39084201388889e-8 * __pyx_v_prst) * __pyx_v_x169) * __pyx_v_x171) * __pyx_v_x175) * __pyx_v_x190) * __pyx_v_x225)) + (((((((5.09109256556341e-19 * __pyx_v_x113) * __pyx_v_x115) * __pyx_v_x142) * __pyx_v_x155) * __pyx_v_x165) * __pyx_v_x167) * __pyx_v_x226)) + ((__pyx_v_x14 * __pyx_v_x229) * __pyx_v_x94)) + (((8.0 * __pyx_v_x21) * __pyx_v_x228) * __pyx_v_x91)) + ((589.775011960583 * __pyx_v_x226) * __pyx_v_x72)) - ((67.9050514751178 * __pyx_v_x227) * __pyx_v_x75)) + ((0.975638950243592 * __pyx_v_x228) * __pyx_v_x79)) + ((__pyx_v_x229 * __pyx_v_x24) * __pyx_v_x84)) + (__pyx_v_x229 * __pyx_v_x82)) + ((__pyx_v_x230 * __pyx_v_x24) * __pyx_v_x89)) + (__pyx_v_x230 * __pyx_v_x87)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1646
 *         cdef double  dHeffdpr = x231*(11.8620273619492*nu*x228*x76 + 3.39084201388889e-8*prst*x169*x171*x175*x190*x225 + 5.09109256556341e-19*x113*x115*x142*x155*x165*x167*x226 + x14*x229*x94 + 8.0*x21*x228*x91 + 589.775011960583*x226*x72 - 67.9050514751178*x227*x75 + 0.975638950243592*x228*x79 + x229*x24*x84 + x229*x82 + x230*x24*x89 + x230*x87)
 * 
 *         cdef double  dHeffdpphi = x12*(x18*x59 + 2.0*x18*x68 + x19*(x23*x232 + x233*x29 + x234*x32) + x42*x64 - 0.25*x42*x66 + x44*(x232*x39 + x233*x40 + x234*x41) + x63 + x67 + x69) + x231*(-pphi*x1*x193*x99 + 2.0*pphi*x112*x21)             # <<<<<<<<<<<<<<
 *         cdef double  M2 = M*M
 *         cdef double  nuH = nu*H
 */
  __Pyx_TraceLine(1646,0,__PYX_ERR(1, 1646, __pyx_L1_error))
  __pyx_v_dHeffdpphi = ((__pyx_v_x12 * (((((((((__pyx_v_x18 * __pyx_v_x59) + ((2.0 * __pyx_v_x18) * __pyx_v_x68)) + (__pyx_v_x19 * (((__pyx_v_x23 * __pyx_v_x232) + (__pyx_v_x233 * __pyx_v_x29)) + (__pyx_v_x234 * __pyx_v_x32)))) + (__pyx_v_x42 * __pyx_v_x64)) - ((0.25 * __pyx_v_x42) * __pyx_v_x66)) + (__pyx_v_x44 * (((__pyx_v_x232 * __pyx_v_x39) + (__pyx_v_x233 * __pyx_v_x40)) + (__pyx_v_x234 * __pyx_v_x41)))) + __pyx_v_x63) + __pyx_v_x67) + __pyx_v_x69)) + (__pyx_v_x231 * (((((-__pyx_v_pphi) * __pyx_v_x1) * __pyx_v_x193) * __pyx_v_x99) + (((2.0 * __pyx_v_pphi) * __pyx_v_x112) * __pyx_v_x21))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1647
 * 
 *         cdef double  dHeffdpphi = x12*(x18*x59 + 2.0*x18*x68 + x19*(x23*x232 + x233*x29 + x234*x32) + x42*x64 - 0.25*x42*x66 + x44*(x232*x39 + x233*x40 + x234*x41) + x63 + x67 + x69) + x231*(-pphi*x1*x193*x99 + 2.0*pphi*x112*x21)
 *         cdef double  M2 = M*M             # <<<<<<<<<<<<<<
 *         cdef double  nuH = nu*H
 *         # Compute H Jacobian
 */
  __Pyx_TraceLine(1647,0,__PYX_ERR(1, 1647, __pyx_L1_error))
  __pyx_v_M2 = (__pyx_v_M * __pyx_v_M);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1648
 *         cdef double  dHeffdpphi = x12*(x18*x59 + 2.0*x18*x68 + x19*(x23*x232 + x233*x29 + x234*x32) + x42*x64 - 0.25*x42*x66 + x44*(x232*x39 + x233*x40 + x234*x41) + x63 + x67 + x69) + x231*(-pphi*x1*x193*x99 + 2.0*pphi*x112*x21)
 *         cdef double  M2 = M*M
 *         cdef double  nuH = nu*H             # <<<<<<<<<<<<<<
 *         # Compute H Jacobian
 *         cdef double  dHdr = M2 * dHeffdr / nuH
 */
  __Pyx_TraceLine(1648,0,__PYX_ERR(1, 1648, __pyx_L1_error))
  __pyx_v_nuH = (__pyx_v_nu * __pyx_v_H);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1650
 *         cdef double  nuH = nu*H
 *         # Compute H Jacobian
 *         cdef double  dHdr = M2 * dHeffdr / nuH             # <<<<<<<<<<<<<<
 *         cdef double  dHdphi = M2 * dHeffdphi / nuH
 *         cdef double  dHdpr = M2 * dHeffdpr / nuH
 */
  __Pyx_TraceLine(1650,0,__PYX_ERR(1, 1650, __pyx_L1_error))
  __pyx_v_dHdr = ((__pyx_v_M2 * __pyx_v_dHeffdr) / __pyx_v_nuH);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1651
 *         # Compute H Jacobian
 *         cdef double  dHdr = M2 * dHeffdr / nuH
 *         cdef double  dHdphi = M2 * dHeffdphi / nuH             # <<<<<<<<<<<<<<
 *         cdef double  dHdpr = M2 * dHeffdpr / nuH
 *         cdef double  dHdpphi = M2 * dHeffdpphi / nuH
 */
  __Pyx_TraceLine(1651,0,__PYX_ERR(1, 1651, __pyx_L1_error))
  __pyx_v_dHdphi = ((__pyx_v_M2 * __pyx_v_dHeffdphi) / __pyx_v_nuH);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1652
 *         cdef double  dHdr = M2 * dHeffdr / nuH
 *         cdef double  dHdphi = M2 * dHeffdphi / nuH
 *         cdef double  dHdpr = M2 * dHeffdpr / nuH             # <<<<<<<<<<<<<<
 *         cdef double  dHdpphi = M2 * dHeffdpphi / nuH
 * 
 */
  __Pyx_TraceLine(1652,0,__PYX_ERR(1, 1652, __pyx_L1_error))
  __pyx_v_dHdpr = ((__pyx_v_M2 * __pyx_v_dHeffdpr) / __pyx_v_nuH);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1653
 *         cdef double  dHdphi = M2 * dHeffdphi / nuH
 *         cdef double  dHdpr = M2 * dHeffdpr / nuH
 *         cdef double  dHdpphi = M2 * dHeffdpphi / nuH             # <<<<<<<<<<<<<<
 * 
 *         return dHdr, dHdphi, dHdpr, dHdpphi,H,xi
 */
  __Pyx_TraceLine(1653,0,__PYX_ERR(1, 1653, __pyx_L1_error))
  __pyx_v_dHdpphi = ((__pyx_v_M2 * __pyx_v_dHeffdpphi) / __pyx_v_nuH);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1655
 *         cdef double  dHdpphi = M2 * dHeffdpphi / nuH
 * 
 *         return dHdr, dHdphi, dHdpr, dHdpphi,H,xi             # <<<<<<<<<<<<<<
 * 
 *     cpdef double omega(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 */
  __Pyx_TraceLine(1655,0,__PYX_ERR(1, 1655, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dHdr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_dHdphi); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_dHdpr); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_dHdpphi); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_H); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = PyFloat_FromDouble(__pyx_v_xi); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyTuple_New(6); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_12, 4, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_12, 5, __pyx_t_11);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_r = __pyx_t_12;
  __pyx_t_12 = 0;
  goto __pyx_L0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1348
 *         return xi
 * 
 *     cpdef dynamics(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.dynamics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c_coeffs);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_11dynamics(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_10dynamics[] = "\n        Compute the dynamics from the Hamiltonian,i.e., dHdr, dHdphi, dHdpr, dHdpphi,H and xi.\n\n        Args:\n          q (double[:]): Canonical positions (r,phi).\n          p (double[:]): Canonical momenta  (prstar,pphi).\n          chi1_v (double[:]): Dimensionless spin vector of the primary.\n          chi2_v (double[:]): Dimensionless spin vector of the secondary.\n          m_1 (double): Primary mass component.\n          m_2 (double): Secondary mass component.\n          chi_1 (double): Projection of chi1_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chi_2 (double): Projection of chi2_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chiL1 (double): Projection of chi1_v onto the orbital angular momentum unit vector (l).\n          chiL2 (double): Projection of chi2_v onto the orbital angular momentum unit vector (l).\n\n        Returns:\n           (tuple) dHdr, dHdphi, dHdpr, dHdpphi,H and xi\n        ";
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_11dynamics(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_q = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi1_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi2_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_m_1;
  double __pyx_v_m_2;
  double __pyx_v_chi_1;
  double __pyx_v_chi_2;
  double __pyx_v_chiL1;
  double __pyx_v_chiL2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dynamics (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_q,&__pyx_n_s_p,&__pyx_n_s_chi1_v,&__pyx_n_s_chi2_v,&__pyx_n_s_m_1,&__pyx_n_s_m_2,&__pyx_n_s_chi_1,&__pyx_n_s_chi_2,&__pyx_n_s_chiL1,&__pyx_n_s_chiL2,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_q)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 1); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi1_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 2); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi2_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 3); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 4); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 5); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 6); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 7); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 8); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, 9); __PYX_ERR(1, 1348, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dynamics") < 0)) __PYX_ERR(1, 1348, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_q = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_q.memview)) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_p.memview)) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_chi1_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi1_v.memview)) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_chi2_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi2_v.memview)) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_m_1 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_m_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_m_2 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_m_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_chi_1 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_chi_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_chi_2 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_chi_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_chiL1 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_chiL1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1348, __pyx_L3_error)
    __pyx_v_chiL2 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_chiL2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1348, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dynamics", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1348, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.dynamics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_10dynamics(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_10dynamics(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dynamics", 0);
  __Pyx_TraceCall("dynamics (wrapper)", __pyx_f[1], 1348, 0, __PYX_ERR(1, 1348, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 1348, __pyx_L1_error) }
  if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 1348, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 1348, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 1348, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_dynamics(__pyx_v_self, __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.dynamics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_q, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi1_v, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi2_v, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1657
 *         return dHdr, dHdphi, dHdpr, dHdpphi,H,xi
 * 
 *     cpdef double omega(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_13omega(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_omega(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch) {
  double __pyx_v_r;
  CYTHON_UNUSED double __pyx_v_phi;
  double __pyx_v_prst;
  double __pyx_v_L;
  double __pyx_v_chix1;
  double __pyx_v_chiy1;
  double __pyx_v_chiz1;
  double __pyx_v_chix2;
  double __pyx_v_chiy2;
  double __pyx_v_chiz2;
  double __pyx_v_pphi;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *__pyx_v_c_coeffs = 0;
  double __pyx_v_a6;
  double __pyx_v_dSO;
  double __pyx_v_M;
  double __pyx_v_nu;
  double __pyx_v_X_1;
  double __pyx_v_X_2;
  double __pyx_v_delta;
  double __pyx_v_z0;
  double __pyx_v_z1;
  double __pyx_v_z2;
  double __pyx_v_z3;
  double __pyx_v_z4;
  double __pyx_v_z5;
  double __pyx_v_z6;
  double __pyx_v_z7;
  double __pyx_v_z8;
  double __pyx_v_z9;
  double __pyx_v_z10;
  double __pyx_v_z11;
  double __pyx_v_z12;
  double __pyx_v_z13;
  double __pyx_v_z14;
  double __pyx_v_z15;
  double __pyx_v_z16;
  double __pyx_v_z17;
  double __pyx_v_z18;
  double __pyx_v_z19;
  double __pyx_v_z20;
  double __pyx_v_z21;
  double __pyx_v_z22;
  double __pyx_v_z23;
  double __pyx_v_z24;
  double __pyx_v_z25;
  double __pyx_v_z26;
  double __pyx_v_z27;
  double __pyx_v_z28;
  double __pyx_v_z29;
  double __pyx_v_z30;
  double __pyx_v_z31;
  double __pyx_v_z32;
  double __pyx_v_z33;
  double __pyx_v_z34;
  double __pyx_v_z35;
  double __pyx_v_z36;
  double __pyx_v_z37;
  double __pyx_v_z38;
  double __pyx_v_z39;
  double __pyx_v_z40;
  double __pyx_v_z41;
  double __pyx_v_z42;
  double __pyx_v_z43;
  double __pyx_v_z44;
  double __pyx_v_z45;
  double __pyx_v_z46;
  double __pyx_v_z47;
  double __pyx_v_z48;
  double __pyx_v_z49;
  double __pyx_v_z50;
  double __pyx_v_z51;
  double __pyx_v_z52;
  double __pyx_v_z53;
  double __pyx_v_z54;
  double __pyx_v_z55;
  double __pyx_v_z56;
  double __pyx_v_z57;
  double __pyx_v_z58;
  double __pyx_v_z59;
  double __pyx_v_z60;
  double __pyx_v_z61;
  double __pyx_v_z62;
  double __pyx_v_z63;
  double __pyx_v_z64;
  double __pyx_v_z65;
  double __pyx_v_z66;
  double __pyx_v_z67;
  double __pyx_v_z68;
  double __pyx_v_z69;
  double __pyx_v_z70;
  double __pyx_v_z71;
  double __pyx_v_z72;
  double __pyx_v_z73;
  double __pyx_v_z74;
  double __pyx_v_z75;
  double __pyx_v_z76;
  double __pyx_v_z77;
  double __pyx_v_z78;
  double __pyx_v_z79;
  double __pyx_v_z80;
  double __pyx_v_z81;
  double __pyx_v_z82;
  double __pyx_v_z83;
  double __pyx_v_z84;
  double __pyx_v_z85;
  double __pyx_v_z86;
  double __pyx_v_z87;
  double __pyx_v_z88;
  double __pyx_v_z89;
  double __pyx_v_z90;
  double __pyx_v_z91;
  double __pyx_v_z92;
  double __pyx_v_z93;
  double __pyx_v_z94;
  double __pyx_v_z95;
  double __pyx_v_z96;
  double __pyx_v_H;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  double __pyx_v_dHeffdpphi;
  double __pyx_v_omega;
  double __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  double __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("omega", 0);
  __Pyx_TraceCall("omega", __pyx_f[1], 1657, 0, __PYX_ERR(1, 1657, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_omega); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1657, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_13omega)) {
        if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 1657, __pyx_L1_error) }
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_q, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 1657, __pyx_L1_error) }
        __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_p, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 1657, __pyx_L1_error) }
        __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_chi1_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 1657, __pyx_L1_error) }
        __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_chi2_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_m_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyFloat_FromDouble(__pyx_v_m_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_chi_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyFloat_FromDouble(__pyx_v_chi_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_chiL1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = PyFloat_FromDouble(__pyx_v_chiL2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_13 = __pyx_t_1; __pyx_t_14 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1657, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1657, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(10+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 1657, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_15, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_15, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_15, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_15, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_15, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_15, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_16, 8+__pyx_t_15, __pyx_t_11);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_16, 9+__pyx_t_15, __pyx_t_12);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_12 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1657, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1657, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_17;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1680
 *         # Coordinate definitions
 * 
 *         cdef double r = q[0]             # <<<<<<<<<<<<<<
 *         cdef double phi = q[1]
 * 
 */
  __Pyx_TraceLine(1680,0,__PYX_ERR(1, 1680, __pyx_L1_error))
  __pyx_t_18 = 0;
  __pyx_v_r = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_18 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1681
 * 
 *         cdef double r = q[0]
 *         cdef double phi = q[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double prst = p[0]
 */
  __Pyx_TraceLine(1681,0,__PYX_ERR(1, 1681, __pyx_L1_error))
  __pyx_t_18 = 1;
  __pyx_v_phi = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_18 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1683
 *         cdef double phi = q[1]
 * 
 *         cdef double prst = p[0]             # <<<<<<<<<<<<<<
 *         cdef double L = p[1]
 * 
 */
  __Pyx_TraceLine(1683,0,__PYX_ERR(1, 1683, __pyx_L1_error))
  __pyx_t_18 = 0;
  __pyx_v_prst = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_18 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1684
 * 
 *         cdef double prst = p[0]
 *         cdef double L = p[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix1 = chi1_v[0]
 */
  __Pyx_TraceLine(1684,0,__PYX_ERR(1, 1684, __pyx_L1_error))
  __pyx_t_18 = 1;
  __pyx_v_L = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_18 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1686
 *         cdef double L = p[1]
 * 
 *         cdef double chix1 = chi1_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]
 */
  __Pyx_TraceLine(1686,0,__PYX_ERR(1, 1686, __pyx_L1_error))
  __pyx_t_18 = 0;
  __pyx_v_chix1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_18 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1687
 * 
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz1 = chi1_v[2]
 * 
 */
  __Pyx_TraceLine(1687,0,__PYX_ERR(1, 1687, __pyx_L1_error))
  __pyx_t_18 = 1;
  __pyx_v_chiy1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_18 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1688
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix2 = chi2_v[0]
 */
  __Pyx_TraceLine(1688,0,__PYX_ERR(1, 1688, __pyx_L1_error))
  __pyx_t_18 = 2;
  __pyx_v_chiz1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_18 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1690
 *         cdef double chiz1 = chi1_v[2]
 * 
 *         cdef double chix2 = chi2_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]
 */
  __Pyx_TraceLine(1690,0,__PYX_ERR(1, 1690, __pyx_L1_error))
  __pyx_t_18 = 0;
  __pyx_v_chix2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_18 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1691
 * 
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz2 = chi2_v[2]
 * 
 */
  __Pyx_TraceLine(1691,0,__PYX_ERR(1, 1691, __pyx_L1_error))
  __pyx_t_18 = 1;
  __pyx_v_chiy2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_18 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1692
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double pphi = L
 */
  __Pyx_TraceLine(1692,0,__PYX_ERR(1, 1692, __pyx_L1_error))
  __pyx_t_18 = 2;
  __pyx_v_chiz2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_18 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1694
 *         cdef double chiz2 = chi2_v[2]
 * 
 *         cdef double pphi = L             # <<<<<<<<<<<<<<
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 */
  __Pyx_TraceLine(1694,0,__PYX_ERR(1, 1694, __pyx_L1_error))
  __pyx_v_pphi = __pyx_v_L;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1696
 *         cdef double pphi = L
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs             # <<<<<<<<<<<<<<
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']
 */
  __Pyx_TraceLine(1696,0,__PYX_ERR(1, 1696, __pyx_L1_error))
  __pyx_t_1 = ((PyObject *)__pyx_v_self->calibration_coeffs);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_c_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1697
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']             # <<<<<<<<<<<<<<
 *         cdef double dSO = c_coeffs['dSO']
 * 
 */
  __Pyx_TraceLine(1697,0,__PYX_ERR(1, 1697, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_a6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_a6 = __pyx_t_17;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1698
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1698,0,__PYX_ERR(1, 1698, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_dSO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1698, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dSO = __pyx_t_17;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1702
 * 
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M             # <<<<<<<<<<<<<<
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 */
  __Pyx_TraceLine(1702,0,__PYX_ERR(1, 1702, __pyx_L1_error))
  __pyx_t_17 = __pyx_v_self->__pyx_base.EOBpars->p_params->M;
  __pyx_v_M = __pyx_t_17;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1703
 *         # Extra quantities used in the Jacobian
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu             # <<<<<<<<<<<<<<
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 */
  __Pyx_TraceLine(1703,0,__PYX_ERR(1, 1703, __pyx_L1_error))
  __pyx_t_17 = __pyx_v_self->__pyx_base.EOBpars->p_params->nu;
  __pyx_v_nu = __pyx_t_17;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1704
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1             # <<<<<<<<<<<<<<
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 */
  __Pyx_TraceLine(1704,0,__PYX_ERR(1, 1704, __pyx_L1_error))
  __pyx_t_17 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_1;
  __pyx_v_X_1 = __pyx_t_17;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1705
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2             # <<<<<<<<<<<<<<
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double z0 = r**3
 */
  __Pyx_TraceLine(1705,0,__PYX_ERR(1, 1705, __pyx_L1_error))
  __pyx_t_17 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_2;
  __pyx_v_X_2 = __pyx_t_17;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1706
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta             # <<<<<<<<<<<<<<
 *         cdef double z0 = r**3
 *         cdef double z1 = r + 2.0
 */
  __Pyx_TraceLine(1706,0,__PYX_ERR(1, 1706, __pyx_L1_error))
  __pyx_t_17 = __pyx_v_self->__pyx_base.EOBpars->p_params->delta;
  __pyx_v_delta = __pyx_t_17;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1707
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double z0 = r**3             # <<<<<<<<<<<<<<
 *         cdef double z1 = r + 2.0
 *         cdef double z2 = X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 */
  __Pyx_TraceLine(1707,0,__PYX_ERR(1, 1707, __pyx_L1_error))
  __pyx_v_z0 = pow(__pyx_v_r, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1708
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double z0 = r**3
 *         cdef double z1 = r + 2.0             # <<<<<<<<<<<<<<
 *         cdef double z2 = X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double z3 = X_1**2
 */
  __Pyx_TraceLine(1708,0,__PYX_ERR(1, 1708, __pyx_L1_error))
  __pyx_v_z1 = (__pyx_v_r + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1709
 *         cdef double z0 = r**3
 *         cdef double z1 = r + 2.0
 *         cdef double z2 = X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)             # <<<<<<<<<<<<<<
 *         cdef double z3 = X_1**2
 *         cdef double z4 = chix1**2 + chiy1**2 + chiz1**2
 */
  __Pyx_TraceLine(1709,0,__PYX_ERR(1, 1709, __pyx_L1_error))
  __pyx_v_z2 = ((__pyx_v_X_1 * __pyx_v_X_2) * ((((2.0 * __pyx_v_chix1) * __pyx_v_chix2) + ((2.0 * __pyx_v_chiy1) * __pyx_v_chiy2)) + ((2.0 * __pyx_v_chiz1) * __pyx_v_chiz2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1710
 *         cdef double z1 = r + 2.0
 *         cdef double z2 = X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double z3 = X_1**2             # <<<<<<<<<<<<<<
 *         cdef double z4 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double z5 = z3*z4
 */
  __Pyx_TraceLine(1710,0,__PYX_ERR(1, 1710, __pyx_L1_error))
  __pyx_v_z3 = pow(__pyx_v_X_1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1711
 *         cdef double z2 = X_1*X_2*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double z3 = X_1**2
 *         cdef double z4 = chix1**2 + chiy1**2 + chiz1**2             # <<<<<<<<<<<<<<
 *         cdef double z5 = z3*z4
 *         cdef double z6 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 */
  __Pyx_TraceLine(1711,0,__PYX_ERR(1, 1711, __pyx_L1_error))
  __pyx_v_z4 = ((pow(__pyx_v_chix1, 2.0) + pow(__pyx_v_chiy1, 2.0)) + pow(__pyx_v_chiz1, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1712
 *         cdef double z3 = X_1**2
 *         cdef double z4 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double z5 = z3*z4             # <<<<<<<<<<<<<<
 *         cdef double z6 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double z7 = z5 + z6
 */
  __Pyx_TraceLine(1712,0,__PYX_ERR(1, 1712, __pyx_L1_error))
  __pyx_v_z5 = (__pyx_v_z3 * __pyx_v_z4);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1713
 *         cdef double z4 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double z5 = z3*z4
 *         cdef double z6 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)             # <<<<<<<<<<<<<<
 *         cdef double z7 = z5 + z6
 *         cdef double z8 = z2 + z7
 */
  __Pyx_TraceLine(1713,0,__PYX_ERR(1, 1713, __pyx_L1_error))
  __pyx_v_z6 = (pow(__pyx_v_X_2, 2.0) * ((pow(__pyx_v_chix2, 2.0) + pow(__pyx_v_chiy2, 2.0)) + pow(__pyx_v_chiz2, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1714
 *         cdef double z5 = z3*z4
 *         cdef double z6 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double z7 = z5 + z6             # <<<<<<<<<<<<<<
 *         cdef double z8 = z2 + z7
 *         cdef double z9 = z1*z8
 */
  __Pyx_TraceLine(1714,0,__PYX_ERR(1, 1714, __pyx_L1_error))
  __pyx_v_z7 = (__pyx_v_z5 + __pyx_v_z6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1715
 *         cdef double z6 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double z7 = z5 + z6
 *         cdef double z8 = z2 + z7             # <<<<<<<<<<<<<<
 *         cdef double z9 = z1*z8
 *         cdef double z10 = z0**(-1)
 */
  __Pyx_TraceLine(1715,0,__PYX_ERR(1, 1715, __pyx_L1_error))
  __pyx_v_z8 = (__pyx_v_z2 + __pyx_v_z7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1716
 *         cdef double z7 = z5 + z6
 *         cdef double z8 = z2 + z7
 *         cdef double z9 = z1*z8             # <<<<<<<<<<<<<<
 *         cdef double z10 = z0**(-1)
 *         cdef double z11 = X_1*chiL1
 */
  __Pyx_TraceLine(1716,0,__PYX_ERR(1, 1716, __pyx_L1_error))
  __pyx_v_z9 = (__pyx_v_z1 * __pyx_v_z8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1717
 *         cdef double z8 = z2 + z7
 *         cdef double z9 = z1*z8
 *         cdef double z10 = z0**(-1)             # <<<<<<<<<<<<<<
 *         cdef double z11 = X_1*chiL1
 *         cdef double z12 = X_2*chiL2
 */
  __Pyx_TraceLine(1717,0,__PYX_ERR(1, 1717, __pyx_L1_error))
  __pyx_v_z10 = pow(__pyx_v_z0, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1718
 *         cdef double z9 = z1*z8
 *         cdef double z10 = z0**(-1)
 *         cdef double z11 = X_1*chiL1             # <<<<<<<<<<<<<<
 *         cdef double z12 = X_2*chiL2
 *         cdef double z13 = z11 + z12
 */
  __Pyx_TraceLine(1718,0,__PYX_ERR(1, 1718, __pyx_L1_error))
  __pyx_v_z11 = (__pyx_v_X_1 * __pyx_v_chiL1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1719
 *         cdef double z10 = z0**(-1)
 *         cdef double z11 = X_1*chiL1
 *         cdef double z12 = X_2*chiL2             # <<<<<<<<<<<<<<
 *         cdef double z13 = z11 + z12
 *         cdef double z14 = -1.40625*nu - 0.46875
 */
  __Pyx_TraceLine(1719,0,__PYX_ERR(1, 1719, __pyx_L1_error))
  __pyx_v_z12 = (__pyx_v_X_2 * __pyx_v_chiL2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1720
 *         cdef double z11 = X_1*chiL1
 *         cdef double z12 = X_2*chiL2
 *         cdef double z13 = z11 + z12             # <<<<<<<<<<<<<<
 *         cdef double z14 = -1.40625*nu - 0.46875
 *         cdef double z15 = r**2
 */
  __Pyx_TraceLine(1720,0,__PYX_ERR(1, 1720, __pyx_L1_error))
  __pyx_v_z13 = (__pyx_v_z11 + __pyx_v_z12);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1721
 *         cdef double z12 = X_2*chiL2
 *         cdef double z13 = z11 + z12
 *         cdef double z14 = -1.40625*nu - 0.46875             # <<<<<<<<<<<<<<
 *         cdef double z15 = r**2
 *         cdef double z16 = z15**(-1)
 */
  __Pyx_TraceLine(1721,0,__PYX_ERR(1, 1721, __pyx_L1_error))
  __pyx_v_z14 = ((-1.40625 * __pyx_v_nu) - 0.46875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1722
 *         cdef double z13 = z11 + z12
 *         cdef double z14 = -1.40625*nu - 0.46875
 *         cdef double z15 = r**2             # <<<<<<<<<<<<<<
 *         cdef double z16 = z15**(-1)
 *         cdef double z17 = 2.0*pphi
 */
  __Pyx_TraceLine(1722,0,__PYX_ERR(1, 1722, __pyx_L1_error))
  __pyx_v_z15 = pow(__pyx_v_r, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1723
 *         cdef double z14 = -1.40625*nu - 0.46875
 *         cdef double z15 = r**2
 *         cdef double z16 = z15**(-1)             # <<<<<<<<<<<<<<
 *         cdef double z17 = 2.0*pphi
 *         cdef double z18 = z16*z17
 */
  __Pyx_TraceLine(1723,0,__PYX_ERR(1, 1723, __pyx_L1_error))
  __pyx_v_z16 = pow(__pyx_v_z15, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1724
 *         cdef double z15 = r**2
 *         cdef double z16 = z15**(-1)
 *         cdef double z17 = 2.0*pphi             # <<<<<<<<<<<<<<
 *         cdef double z18 = z16*z17
 *         cdef double z19 = nu**2
 */
  __Pyx_TraceLine(1724,0,__PYX_ERR(1, 1724, __pyx_L1_error))
  __pyx_v_z17 = (2.0 * __pyx_v_pphi);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1725
 *         cdef double z16 = z15**(-1)
 *         cdef double z17 = 2.0*pphi
 *         cdef double z18 = z16*z17             # <<<<<<<<<<<<<<
 *         cdef double z19 = nu**2
 *         cdef double z20 = -2.0859375*nu - 2.07161458333333*z19 + 0.23046875
 */
  __Pyx_TraceLine(1725,0,__PYX_ERR(1, 1725, __pyx_L1_error))
  __pyx_v_z18 = (__pyx_v_z16 * __pyx_v_z17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1726
 *         cdef double z17 = 2.0*pphi
 *         cdef double z18 = z16*z17
 *         cdef double z19 = nu**2             # <<<<<<<<<<<<<<
 *         cdef double z20 = -2.0859375*nu - 2.07161458333333*z19 + 0.23046875
 *         cdef double z21 = z10*z17
 */
  __Pyx_TraceLine(1726,0,__PYX_ERR(1, 1726, __pyx_L1_error))
  __pyx_v_z19 = pow(__pyx_v_nu, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1727
 *         cdef double z18 = z16*z17
 *         cdef double z19 = nu**2
 *         cdef double z20 = -2.0859375*nu - 2.07161458333333*z19 + 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double z21 = z10*z17
 *         cdef double z22 = 0.5859375*nu + 1.34765625*z19 + 0.41015625
 */
  __Pyx_TraceLine(1727,0,__PYX_ERR(1, 1727, __pyx_L1_error))
  __pyx_v_z20 = (((-2.0859375 * __pyx_v_nu) - (2.07161458333333 * __pyx_v_z19)) + 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1728
 *         cdef double z19 = nu**2
 *         cdef double z20 = -2.0859375*nu - 2.07161458333333*z19 + 0.23046875
 *         cdef double z21 = z10*z17             # <<<<<<<<<<<<<<
 *         cdef double z22 = 0.5859375*nu + 1.34765625*z19 + 0.41015625
 *         cdef double z23 = r**4
 */
  __Pyx_TraceLine(1728,0,__PYX_ERR(1, 1728, __pyx_L1_error))
  __pyx_v_z21 = (__pyx_v_z10 * __pyx_v_z17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1729
 *         cdef double z20 = -2.0859375*nu - 2.07161458333333*z19 + 0.23046875
 *         cdef double z21 = z10*z17
 *         cdef double z22 = 0.5859375*nu + 1.34765625*z19 + 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double z23 = r**4
 *         cdef double z24 = z23**(-1)
 */
  __Pyx_TraceLine(1729,0,__PYX_ERR(1, 1729, __pyx_L1_error))
  __pyx_v_z22 = (((0.5859375 * __pyx_v_nu) + (1.34765625 * __pyx_v_z19)) + 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1730
 *         cdef double z21 = z10*z17
 *         cdef double z22 = 0.5859375*nu + 1.34765625*z19 + 0.41015625
 *         cdef double z23 = r**4             # <<<<<<<<<<<<<<
 *         cdef double z24 = z23**(-1)
 *         cdef double z25 = 4.0*pphi**3*z24
 */
  __Pyx_TraceLine(1730,0,__PYX_ERR(1, 1730, __pyx_L1_error))
  __pyx_v_z23 = pow(__pyx_v_r, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1731
 *         cdef double z22 = 0.5859375*nu + 1.34765625*z19 + 0.41015625
 *         cdef double z23 = r**4
 *         cdef double z24 = z23**(-1)             # <<<<<<<<<<<<<<
 *         cdef double z25 = 4.0*pphi**3*z24
 *         cdef double z26 = 0.46875 - 0.28125*nu
 */
  __Pyx_TraceLine(1731,0,__PYX_ERR(1, 1731, __pyx_L1_error))
  __pyx_v_z24 = pow(__pyx_v_z23, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1732
 *         cdef double z23 = r**4
 *         cdef double z24 = z23**(-1)
 *         cdef double z25 = 4.0*pphi**3*z24             # <<<<<<<<<<<<<<
 *         cdef double z26 = 0.46875 - 0.28125*nu
 *         cdef double z27 = -0.2734375*nu - 0.798177083333333*z19 - 0.23046875
 */
  __Pyx_TraceLine(1732,0,__PYX_ERR(1, 1732, __pyx_L1_error))
  __pyx_v_z25 = ((4.0 * pow(__pyx_v_pphi, 3.0)) * __pyx_v_z24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1733
 *         cdef double z24 = z23**(-1)
 *         cdef double z25 = 4.0*pphi**3*z24
 *         cdef double z26 = 0.46875 - 0.28125*nu             # <<<<<<<<<<<<<<
 *         cdef double z27 = -0.2734375*nu - 0.798177083333333*z19 - 0.23046875
 *         cdef double z28 = -0.3515625*nu + 0.29296875*z19 - 0.41015625
 */
  __Pyx_TraceLine(1733,0,__PYX_ERR(1, 1733, __pyx_L1_error))
  __pyx_v_z26 = (0.46875 - (0.28125 * __pyx_v_nu));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1734
 *         cdef double z25 = 4.0*pphi**3*z24
 *         cdef double z26 = 0.46875 - 0.28125*nu
 *         cdef double z27 = -0.2734375*nu - 0.798177083333333*z19 - 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double z28 = -0.3515625*nu + 0.29296875*z19 - 0.41015625
 *         cdef double z29 = z11 - z12
 */
  __Pyx_TraceLine(1734,0,__PYX_ERR(1, 1734, __pyx_L1_error))
  __pyx_v_z27 = (((-0.2734375 * __pyx_v_nu) - (0.798177083333333 * __pyx_v_z19)) - 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1735
 *         cdef double z26 = 0.46875 - 0.28125*nu
 *         cdef double z27 = -0.2734375*nu - 0.798177083333333*z19 - 0.23046875
 *         cdef double z28 = -0.3515625*nu + 0.29296875*z19 - 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double z29 = z11 - z12
 *         cdef double z30 = delta*z29
 */
  __Pyx_TraceLine(1735,0,__PYX_ERR(1, 1735, __pyx_L1_error))
  __pyx_v_z28 = (((-0.3515625 * __pyx_v_nu) + (0.29296875 * __pyx_v_z19)) - 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1736
 *         cdef double z27 = -0.2734375*nu - 0.798177083333333*z19 - 0.23046875
 *         cdef double z28 = -0.3515625*nu + 0.29296875*z19 - 0.41015625
 *         cdef double z29 = z11 - z12             # <<<<<<<<<<<<<<
 *         cdef double z30 = delta*z29
 *         cdef double z31 = pphi**2
 */
  __Pyx_TraceLine(1736,0,__PYX_ERR(1, 1736, __pyx_L1_error))
  __pyx_v_z29 = (__pyx_v_z11 - __pyx_v_z12);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1737
 *         cdef double z28 = -0.3515625*nu + 0.29296875*z19 - 0.41015625
 *         cdef double z29 = z11 - z12
 *         cdef double z30 = delta*z29             # <<<<<<<<<<<<<<
 *         cdef double z31 = pphi**2
 *         cdef double z32 = chi_1*X_1
 */
  __Pyx_TraceLine(1737,0,__PYX_ERR(1, 1737, __pyx_L1_error))
  __pyx_v_z30 = (__pyx_v_delta * __pyx_v_z29);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1738
 *         cdef double z29 = z11 - z12
 *         cdef double z30 = delta*z29
 *         cdef double z31 = pphi**2             # <<<<<<<<<<<<<<
 *         cdef double z32 = chi_1*X_1
 *         cdef double z33 = chi_2*X_2
 */
  __Pyx_TraceLine(1738,0,__PYX_ERR(1, 1738, __pyx_L1_error))
  __pyx_v_z31 = pow(__pyx_v_pphi, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1739
 *         cdef double z30 = delta*z29
 *         cdef double z31 = pphi**2
 *         cdef double z32 = chi_1*X_1             # <<<<<<<<<<<<<<
 *         cdef double z33 = chi_2*X_2
 *         cdef double z34 = z32 + z33
 */
  __Pyx_TraceLine(1739,0,__PYX_ERR(1, 1739, __pyx_L1_error))
  __pyx_v_z32 = (__pyx_v_chi_1 * __pyx_v_X_1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1740
 *         cdef double z31 = pphi**2
 *         cdef double z32 = chi_1*X_1
 *         cdef double z33 = chi_2*X_2             # <<<<<<<<<<<<<<
 *         cdef double z34 = z32 + z33
 *         cdef double z35 = z34**2
 */
  __Pyx_TraceLine(1740,0,__PYX_ERR(1, 1740, __pyx_L1_error))
  __pyx_v_z33 = (__pyx_v_chi_2 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1741
 *         cdef double z32 = chi_1*X_1
 *         cdef double z33 = chi_2*X_2
 *         cdef double z34 = z32 + z33             # <<<<<<<<<<<<<<
 *         cdef double z35 = z34**2
 *         cdef double z36 = -z35 + z8
 */
  __Pyx_TraceLine(1741,0,__PYX_ERR(1, 1741, __pyx_L1_error))
  __pyx_v_z34 = (__pyx_v_z32 + __pyx_v_z33);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1742
 *         cdef double z33 = chi_2*X_2
 *         cdef double z34 = z32 + z33
 *         cdef double z35 = z34**2             # <<<<<<<<<<<<<<
 *         cdef double z36 = -z35 + z8
 *         cdef double z37 = z10*z36
 */
  __Pyx_TraceLine(1742,0,__PYX_ERR(1, 1742, __pyx_L1_error))
  __pyx_v_z35 = pow(__pyx_v_z34, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1743
 *         cdef double z34 = z32 + z33
 *         cdef double z35 = z34**2
 *         cdef double z36 = -z35 + z8             # <<<<<<<<<<<<<<
 *         cdef double z37 = z10*z36
 *         cdef double z38 = z31*z37
 */
  __Pyx_TraceLine(1743,0,__PYX_ERR(1, 1743, __pyx_L1_error))
  __pyx_v_z36 = ((-__pyx_v_z35) + __pyx_v_z8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1744
 *         cdef double z35 = z34**2
 *         cdef double z36 = -z35 + z8
 *         cdef double z37 = z10*z36             # <<<<<<<<<<<<<<
 *         cdef double z38 = z31*z37
 *         cdef double z39 = r**(-1)
 */
  __Pyx_TraceLine(1744,0,__PYX_ERR(1, 1744, __pyx_L1_error))
  __pyx_v_z37 = (__pyx_v_z10 * __pyx_v_z36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1745
 *         cdef double z36 = -z35 + z8
 *         cdef double z37 = z10*z36
 *         cdef double z38 = z31*z37             # <<<<<<<<<<<<<<
 *         cdef double z39 = r**(-1)
 *         cdef double z40 = z16*z31
 */
  __Pyx_TraceLine(1745,0,__PYX_ERR(1, 1745, __pyx_L1_error))
  __pyx_v_z38 = (__pyx_v_z31 * __pyx_v_z37);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1746
 *         cdef double z37 = z10*z36
 *         cdef double z38 = z31*z37
 *         cdef double z39 = r**(-1)             # <<<<<<<<<<<<<<
 *         cdef double z40 = z16*z31
 *         cdef double z41 = z10*z31
 */
  __Pyx_TraceLine(1746,0,__PYX_ERR(1, 1746, __pyx_L1_error))
  __pyx_v_z39 = pow(__pyx_v_r, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1747
 *         cdef double z38 = z31*z37
 *         cdef double z39 = r**(-1)
 *         cdef double z40 = z16*z31             # <<<<<<<<<<<<<<
 *         cdef double z41 = z10*z31
 *         cdef double z42 = pphi**4*z24
 */
  __Pyx_TraceLine(1747,0,__PYX_ERR(1, 1747, __pyx_L1_error))
  __pyx_v_z40 = (__pyx_v_z16 * __pyx_v_z31);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1748
 *         cdef double z39 = r**(-1)
 *         cdef double z40 = z16*z31
 *         cdef double z41 = z10*z31             # <<<<<<<<<<<<<<
 *         cdef double z42 = pphi**4*z24
 *         cdef double z43 = z32 - z33
 */
  __Pyx_TraceLine(1748,0,__PYX_ERR(1, 1748, __pyx_L1_error))
  __pyx_v_z41 = (__pyx_v_z10 * __pyx_v_z31);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1749
 *         cdef double z40 = z16*z31
 *         cdef double z41 = z10*z31
 *         cdef double z42 = pphi**4*z24             # <<<<<<<<<<<<<<
 *         cdef double z43 = z32 - z33
 *         cdef double z44 = z3*z4 - z34*z43 - z6
 */
  __Pyx_TraceLine(1749,0,__PYX_ERR(1, 1749, __pyx_L1_error))
  __pyx_v_z42 = (pow(__pyx_v_pphi, 4.0) * __pyx_v_z24);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1750
 *         cdef double z41 = z10*z31
 *         cdef double z42 = pphi**4*z24
 *         cdef double z43 = z32 - z33             # <<<<<<<<<<<<<<
 *         cdef double z44 = z3*z4 - z34*z43 - z6
 *         cdef double z45 = z41*(-0.125*z2 + 0.125*z35 + 0.5*z44*(0.5 - X_2) - 0.125*z5 - 0.125*z6)
 */
  __Pyx_TraceLine(1750,0,__PYX_ERR(1, 1750, __pyx_L1_error))
  __pyx_v_z43 = (__pyx_v_z32 - __pyx_v_z33);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1751
 *         cdef double z42 = pphi**4*z24
 *         cdef double z43 = z32 - z33
 *         cdef double z44 = z3*z4 - z34*z43 - z6             # <<<<<<<<<<<<<<
 *         cdef double z45 = z41*(-0.125*z2 + 0.125*z35 + 0.5*z44*(0.5 - X_2) - 0.125*z5 - 0.125*z6)
 *         cdef double z46 = 0.5*z36
 */
  __Pyx_TraceLine(1751,0,__PYX_ERR(1, 1751, __pyx_L1_error))
  __pyx_v_z44 = (((__pyx_v_z3 * __pyx_v_z4) - (__pyx_v_z34 * __pyx_v_z43)) - __pyx_v_z6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1752
 *         cdef double z43 = z32 - z33
 *         cdef double z44 = z3*z4 - z34*z43 - z6
 *         cdef double z45 = z41*(-0.125*z2 + 0.125*z35 + 0.5*z44*(0.5 - X_2) - 0.125*z5 - 0.125*z6)             # <<<<<<<<<<<<<<
 *         cdef double z46 = 0.5*z36
 *         cdef double z47 = z5 - z6
 */
  __Pyx_TraceLine(1752,0,__PYX_ERR(1, 1752, __pyx_L1_error))
  __pyx_v_z45 = (__pyx_v_z41 * (((((-0.125 * __pyx_v_z2) + (0.125 * __pyx_v_z35)) + ((0.5 * __pyx_v_z44) * (0.5 - __pyx_v_X_2))) - (0.125 * __pyx_v_z5)) - (0.125 * __pyx_v_z6)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1753
 *         cdef double z44 = z3*z4 - z34*z43 - z6
 *         cdef double z45 = z41*(-0.125*z2 + 0.125*z35 + 0.5*z44*(0.5 - X_2) - 0.125*z5 - 0.125*z6)
 *         cdef double z46 = 0.5*z36             # <<<<<<<<<<<<<<
 *         cdef double z47 = z5 - z6
 *         cdef double z48 = delta*z44
 */
  __Pyx_TraceLine(1753,0,__PYX_ERR(1, 1753, __pyx_L1_error))
  __pyx_v_z46 = (0.5 * __pyx_v_z36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1754
 *         cdef double z45 = z41*(-0.125*z2 + 0.125*z35 + 0.5*z44*(0.5 - X_2) - 0.125*z5 - 0.125*z6)
 *         cdef double z46 = 0.5*z36
 *         cdef double z47 = z5 - z6             # <<<<<<<<<<<<<<
 *         cdef double z48 = delta*z44
 *         cdef double z49 = z16*z8
 */
  __Pyx_TraceLine(1754,0,__PYX_ERR(1, 1754, __pyx_L1_error))
  __pyx_v_z47 = (__pyx_v_z5 - __pyx_v_z6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1755
 *         cdef double z46 = 0.5*z36
 *         cdef double z47 = z5 - z6
 *         cdef double z48 = delta*z44             # <<<<<<<<<<<<<<
 *         cdef double z49 = z16*z8
 *         cdef double z50 = 2.0*z39
 */
  __Pyx_TraceLine(1755,0,__PYX_ERR(1, 1755, __pyx_L1_error))
  __pyx_v_z48 = (__pyx_v_delta * __pyx_v_z44);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1756
 *         cdef double z47 = z5 - z6
 *         cdef double z48 = delta*z44
 *         cdef double z49 = z16*z8             # <<<<<<<<<<<<<<
 *         cdef double z50 = 2.0*z39
 *         cdef double z51 = z1*z13**2/(r*z9 + z23)
 */
  __Pyx_TraceLine(1756,0,__PYX_ERR(1, 1756, __pyx_L1_error))
  __pyx_v_z49 = (__pyx_v_z16 * __pyx_v_z8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1757
 *         cdef double z48 = delta*z44
 *         cdef double z49 = z16*z8
 *         cdef double z50 = 2.0*z39             # <<<<<<<<<<<<<<
 *         cdef double z51 = z1*z13**2/(r*z9 + z23)
 *         cdef double z52 = -z2 + z7
 */
  __Pyx_TraceLine(1757,0,__PYX_ERR(1, 1757, __pyx_L1_error))
  __pyx_v_z50 = (2.0 * __pyx_v_z39);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1758
 *         cdef double z49 = z16*z8
 *         cdef double z50 = 2.0*z39
 *         cdef double z51 = z1*z13**2/(r*z9 + z23)             # <<<<<<<<<<<<<<
 *         cdef double z52 = -z2 + z7
 *         cdef double z53 = -z43**2 + z52
 */
  __Pyx_TraceLine(1758,0,__PYX_ERR(1, 1758, __pyx_L1_error))
  __pyx_v_z51 = ((__pyx_v_z1 * pow(__pyx_v_z13, 2.0)) / ((__pyx_v_r * __pyx_v_z9) + __pyx_v_z23));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1759
 *         cdef double z50 = 2.0*z39
 *         cdef double z51 = z1*z13**2/(r*z9 + z23)
 *         cdef double z52 = -z2 + z7             # <<<<<<<<<<<<<<
 *         cdef double z53 = -z43**2 + z52
 *         cdef double z54 = 264.0*z19
 */
  __Pyx_TraceLine(1759,0,__PYX_ERR(1, 1759, __pyx_L1_error))
  __pyx_v_z52 = ((-__pyx_v_z2) + __pyx_v_z7);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1760
 *         cdef double z51 = z1*z13**2/(r*z9 + z23)
 *         cdef double z52 = -z2 + z7
 *         cdef double z53 = -z43**2 + z52             # <<<<<<<<<<<<<<
 *         cdef double z54 = 264.0*z19
 *         cdef double z55 = 0.00260416666666667*z36
 */
  __Pyx_TraceLine(1760,0,__PYX_ERR(1, 1760, __pyx_L1_error))
  __pyx_v_z53 = ((-pow(__pyx_v_z43, 2.0)) + __pyx_v_z52);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1761
 *         cdef double z52 = -z2 + z7
 *         cdef double z53 = -z43**2 + z52
 *         cdef double z54 = 264.0*z19             # <<<<<<<<<<<<<<
 *         cdef double z55 = 0.00260416666666667*z36
 *         cdef double z56 = 0.0078125*z53
 */
  __Pyx_TraceLine(1761,0,__PYX_ERR(1, 1761, __pyx_L1_error))
  __pyx_v_z54 = (264.0 * __pyx_v_z19);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1762
 *         cdef double z53 = -z43**2 + z52
 *         cdef double z54 = 264.0*z19
 *         cdef double z55 = 0.00260416666666667*z36             # <<<<<<<<<<<<<<
 *         cdef double z56 = 0.0078125*z53
 *         cdef double z57 = z10*(0.5*z44*(0.75 - 1.5*X_2) + z46*(-1.75*nu - 0.9375) + 0.5*z53*(0.1875 - 0.75*nu)) - z16*z46 + z24*(-0.0625*z48*(98.0*nu + 43.0) + z55*(-1610.0*nu + z54 + 375.0) + z56*(-1090.0*nu + 152.0*z19 + 219.0)) + 1.0
 */
  __Pyx_TraceLine(1762,0,__PYX_ERR(1, 1762, __pyx_L1_error))
  __pyx_v_z55 = (0.00260416666666667 * __pyx_v_z36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1763
 *         cdef double z54 = 264.0*z19
 *         cdef double z55 = 0.00260416666666667*z36
 *         cdef double z56 = 0.0078125*z53             # <<<<<<<<<<<<<<
 *         cdef double z57 = z10*(0.5*z44*(0.75 - 1.5*X_2) + z46*(-1.75*nu - 0.9375) + 0.5*z53*(0.1875 - 0.75*nu)) - z16*z46 + z24*(-0.0625*z48*(98.0*nu + 43.0) + z55*(-1610.0*nu + z54 + 375.0) + z56*(-1090.0*nu + 152.0*z19 + 219.0)) + 1.0
 *         cdef double z58 = delta*z47
 */
  __Pyx_TraceLine(1763,0,__PYX_ERR(1, 1763, __pyx_L1_error))
  __pyx_v_z56 = (0.0078125 * __pyx_v_z53);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1764
 *         cdef double z55 = 0.00260416666666667*z36
 *         cdef double z56 = 0.0078125*z53
 *         cdef double z57 = z10*(0.5*z44*(0.75 - 1.5*X_2) + z46*(-1.75*nu - 0.9375) + 0.5*z53*(0.1875 - 0.75*nu)) - z16*z46 + z24*(-0.0625*z48*(98.0*nu + 43.0) + z55*(-1610.0*nu + z54 + 375.0) + z56*(-1090.0*nu + 152.0*z19 + 219.0)) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double z58 = delta*z47
 *         cdef double z59 = 4.0*nu
 */
  __Pyx_TraceLine(1764,0,__PYX_ERR(1, 1764, __pyx_L1_error))
  __pyx_v_z57 = ((((__pyx_v_z10 * ((((0.5 * __pyx_v_z44) * (0.75 - (1.5 * __pyx_v_X_2))) + (__pyx_v_z46 * ((-1.75 * __pyx_v_nu) - 0.9375))) + ((0.5 * __pyx_v_z53) * (0.1875 - (0.75 * __pyx_v_nu))))) - (__pyx_v_z16 * __pyx_v_z46)) + (__pyx_v_z24 * ((((-0.0625 * __pyx_v_z48) * ((98.0 * __pyx_v_nu) + 43.0)) + (__pyx_v_z55 * (((-1610.0 * __pyx_v_nu) + __pyx_v_z54) + 375.0))) + (__pyx_v_z56 * (((-1090.0 * __pyx_v_nu) + (152.0 * __pyx_v_z19)) + 219.0))))) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1765
 *         cdef double z56 = 0.0078125*z53
 *         cdef double z57 = z10*(0.5*z44*(0.75 - 1.5*X_2) + z46*(-1.75*nu - 0.9375) + 0.5*z53*(0.1875 - 0.75*nu)) - z16*z46 + z24*(-0.0625*z48*(98.0*nu + 43.0) + z55*(-1610.0*nu + z54 + 375.0) + z56*(-1090.0*nu + 152.0*z19 + 219.0)) + 1.0
 *         cdef double z58 = delta*z47             # <<<<<<<<<<<<<<
 *         cdef double z59 = 4.0*nu
 *         cdef double z60 = r**5
 */
  __Pyx_TraceLine(1765,0,__PYX_ERR(1, 1765, __pyx_L1_error))
  __pyx_v_z58 = (__pyx_v_delta * __pyx_v_z47);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1766
 *         cdef double z57 = z10*(0.5*z44*(0.75 - 1.5*X_2) + z46*(-1.75*nu - 0.9375) + 0.5*z53*(0.1875 - 0.75*nu)) - z16*z46 + z24*(-0.0625*z48*(98.0*nu + 43.0) + z55*(-1610.0*nu + z54 + 375.0) + z56*(-1090.0*nu + 152.0*z19 + 219.0)) + 1.0
 *         cdef double z58 = delta*z47
 *         cdef double z59 = 4.0*nu             # <<<<<<<<<<<<<<
 *         cdef double z60 = r**5
 *         cdef double z61 = z60**(-1)
 */
  __Pyx_TraceLine(1766,0,__PYX_ERR(1, 1766, __pyx_L1_error))
  __pyx_v_z59 = (4.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1767
 *         cdef double z58 = delta*z47
 *         cdef double z59 = 4.0*nu
 *         cdef double z60 = r**5             # <<<<<<<<<<<<<<
 *         cdef double z61 = z60**(-1)
 *         cdef double z62 = 7.0*nu
 */
  __Pyx_TraceLine(1767,0,__PYX_ERR(1, 1767, __pyx_L1_error))
  __pyx_v_z60 = pow(__pyx_v_r, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1768
 *         cdef double z59 = 4.0*nu
 *         cdef double z60 = r**5
 *         cdef double z61 = z60**(-1)             # <<<<<<<<<<<<<<
 *         cdef double z62 = 7.0*nu
 *         cdef double z63 = -3.0*z2 + 3.0*z5 + 3.0*z6
 */
  __Pyx_TraceLine(1768,0,__PYX_ERR(1, 1768, __pyx_L1_error))
  __pyx_v_z61 = pow(__pyx_v_z60, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1769
 *         cdef double z60 = r**5
 *         cdef double z61 = z60**(-1)
 *         cdef double z62 = 7.0*nu             # <<<<<<<<<<<<<<
 *         cdef double z63 = -3.0*z2 + 3.0*z5 + 3.0*z6
 *         cdef double z64 = nu**4
 */
  __Pyx_TraceLine(1769,0,__PYX_ERR(1, 1769, __pyx_L1_error))
  __pyx_v_z62 = (7.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1770
 *         cdef double z61 = z60**(-1)
 *         cdef double z62 = 7.0*nu
 *         cdef double z63 = -3.0*z2 + 3.0*z5 + 3.0*z6             # <<<<<<<<<<<<<<
 *         cdef double z64 = nu**4
 *         cdef double z65 = log(r)
 */
  __Pyx_TraceLine(1770,0,__PYX_ERR(1, 1770, __pyx_L1_error))
  __pyx_v_z63 = (((-3.0 * __pyx_v_z2) + (3.0 * __pyx_v_z5)) + (3.0 * __pyx_v_z6));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1771
 *         cdef double z62 = 7.0*nu
 *         cdef double z63 = -3.0*z2 + 3.0*z5 + 3.0*z6
 *         cdef double z64 = nu**4             # <<<<<<<<<<<<<<
 *         cdef double z65 = log(r)
 *         cdef double z66 = z65**2
 */
  __Pyx_TraceLine(1771,0,__PYX_ERR(1, 1771, __pyx_L1_error))
  __pyx_v_z64 = pow(__pyx_v_nu, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1772
 *         cdef double z63 = -3.0*z2 + 3.0*z5 + 3.0*z6
 *         cdef double z64 = nu**4
 *         cdef double z65 = log(r)             # <<<<<<<<<<<<<<
 *         cdef double z66 = z65**2
 *         cdef double z67 = nu**3
 */
  __Pyx_TraceLine(1772,0,__PYX_ERR(1, 1772, __pyx_L1_error))
  __pyx_v_z65 = log(__pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1773
 *         cdef double z64 = nu**4
 *         cdef double z65 = log(r)
 *         cdef double z66 = z65**2             # <<<<<<<<<<<<<<
 *         cdef double z67 = nu**3
 *         cdef double z68 = 756.0*nu
 */
  __Pyx_TraceLine(1773,0,__PYX_ERR(1, 1773, __pyx_L1_error))
  __pyx_v_z66 = pow(__pyx_v_z65, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1774
 *         cdef double z65 = log(r)
 *         cdef double z66 = z65**2
 *         cdef double z67 = nu**3             # <<<<<<<<<<<<<<
 *         cdef double z68 = 756.0*nu
 *         cdef double z69 = z68 + 1079.0
 */
  __Pyx_TraceLine(1774,0,__PYX_ERR(1, 1774, __pyx_L1_error))
  __pyx_v_z67 = pow(__pyx_v_nu, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1775
 *         cdef double z66 = z65**2
 *         cdef double z67 = nu**3
 *         cdef double z68 = 756.0*nu             # <<<<<<<<<<<<<<
 *         cdef double z69 = z68 + 1079.0
 *         cdef double z70 = 8.0*r + 2.0*z0 + 4.0*z15 + 16.0
 */
  __Pyx_TraceLine(1775,0,__PYX_ERR(1, 1775, __pyx_L1_error))
  __pyx_v_z68 = (756.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1776
 *         cdef double z67 = nu**3
 *         cdef double z68 = 756.0*nu
 *         cdef double z69 = z68 + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double z70 = 8.0*r + 2.0*z0 + 4.0*z15 + 16.0
 *         cdef double z71 = (2048.0*nu*z65*(336.0*r + z68 + 407.0) + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*z19 - 245760.0) - 5416406.59541186*z19 - 3440640.0)/(32768.0*nu*z65*(-38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*z19 + 17264.0) + 960.0*z0*z69 + 1920.0*z15*(588.0*nu + 1079.0) - 1882456.23663972*z19 + 480.0*z23*z69 + 161280.0*z60 + 13447680.0) + 53760.0*nu*(7680.0*a6*(z23 + z70) + 113485.217444961*r*(-z23 + z70) + 148.04406601634*r*(7704.0*r + 1926.0*z0 + 3852.0*z15 + 349.0*z23 + 36400.0) + 128.0*r*(13218.7851094412*r + 4264.6962773603*z0 + 8529.39255472061*z15 - 6852.34813868015*z23 - 33722.4297811176)) + 67645734912.0*z19*z66 + 7.0*z19*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*z0 + 2064783811.32587*z15 - 3089250703.76879*z23 + 1426660551.8844*z60 + 276057889687.011) + 13212057600.0*z60 + 241555486248.807*z64 + 1120.0*z67*(-163683964.822551*r - 17833256.898555*z15 - 1188987459.03162))
 */
  __Pyx_TraceLine(1776,0,__PYX_ERR(1, 1776, __pyx_L1_error))
  __pyx_v_z69 = (__pyx_v_z68 + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1777
 *         cdef double z68 = 756.0*nu
 *         cdef double z69 = z68 + 1079.0
 *         cdef double z70 = 8.0*r + 2.0*z0 + 4.0*z15 + 16.0             # <<<<<<<<<<<<<<
 *         cdef double z71 = (2048.0*nu*z65*(336.0*r + z68 + 407.0) + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*z19 - 245760.0) - 5416406.59541186*z19 - 3440640.0)/(32768.0*nu*z65*(-38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*z19 + 17264.0) + 960.0*z0*z69 + 1920.0*z15*(588.0*nu + 1079.0) - 1882456.23663972*z19 + 480.0*z23*z69 + 161280.0*z60 + 13447680.0) + 53760.0*nu*(7680.0*a6*(z23 + z70) + 113485.217444961*r*(-z23 + z70) + 148.04406601634*r*(7704.0*r + 1926.0*z0 + 3852.0*z15 + 349.0*z23 + 36400.0) + 128.0*r*(13218.7851094412*r + 4264.6962773603*z0 + 8529.39255472061*z15 - 6852.34813868015*z23 - 33722.4297811176)) + 67645734912.0*z19*z66 + 7.0*z19*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*z0 + 2064783811.32587*z15 - 3089250703.76879*z23 + 1426660551.8844*z60 + 276057889687.011) + 13212057600.0*z60 + 241555486248.807*z64 + 1120.0*z67*(-163683964.822551*r - 17833256.898555*z15 - 1188987459.03162))
 *         cdef double z72 = z23*z71
 */
  __Pyx_TraceLine(1777,0,__PYX_ERR(1, 1777, __pyx_L1_error))
  __pyx_v_z70 = ((((8.0 * __pyx_v_r) + (2.0 * __pyx_v_z0)) + (4.0 * __pyx_v_z15)) + 16.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1778
 *         cdef double z69 = z68 + 1079.0
 *         cdef double z70 = 8.0*r + 2.0*z0 + 4.0*z15 + 16.0
 *         cdef double z71 = (2048.0*nu*z65*(336.0*r + z68 + 407.0) + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*z19 - 245760.0) - 5416406.59541186*z19 - 3440640.0)/(32768.0*nu*z65*(-38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*z19 + 17264.0) + 960.0*z0*z69 + 1920.0*z15*(588.0*nu + 1079.0) - 1882456.23663972*z19 + 480.0*z23*z69 + 161280.0*z60 + 13447680.0) + 53760.0*nu*(7680.0*a6*(z23 + z70) + 113485.217444961*r*(-z23 + z70) + 148.04406601634*r*(7704.0*r + 1926.0*z0 + 3852.0*z15 + 349.0*z23 + 36400.0) + 128.0*r*(13218.7851094412*r + 4264.6962773603*z0 + 8529.39255472061*z15 - 6852.34813868015*z23 - 33722.4297811176)) + 67645734912.0*z19*z66 + 7.0*z19*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*z0 + 2064783811.32587*z15 - 3089250703.76879*z23 + 1426660551.8844*z60 + 276057889687.011) + 13212057600.0*z60 + 241555486248.807*z64 + 1120.0*z67*(-163683964.822551*r - 17833256.898555*z15 - 1188987459.03162))             # <<<<<<<<<<<<<<
 *         cdef double z72 = z23*z71
 *         cdef double z73 = z24*(0.125*z36*(z62 - 31.0) + 2.0625*z48 + 0.0625*z53*(-z59 - 3.0)) + z24*(1.125*z2 + 1.125*z5 + 0.125*z52*(z59 + 1.0) - 1.25*z58 + 1.125*z6) + z37 + z49 + z61*(0.125*z48*(68.0*nu - 1.0) + z55*(2870.0*nu - z54 + 561.0) + z56*(1166.0*nu - 328.0*z19 - 171.0)) + z61*(-1.21875*z58*(2.0*nu - 3.0) + 0.015625*z63*(-27.0*nu + 28.0*z19 - 3.0) - 0.015625*(z62 + 9.0)*(25.0*z2 + 25.0*z5 + 25.0*z6)) + 7680.0*z72
 */
  __Pyx_TraceLine(1778,0,__PYX_ERR(1, 1778, __pyx_L1_error))
  __pyx_v_z71 = ((((((((2048.0 * __pyx_v_nu) * __pyx_v_z65) * (((336.0 * __pyx_v_r) + __pyx_v_z68) + 407.0)) + ((28.0 * __pyx_v_nu) * ((1920.0 * __pyx_v_a6) + 733955.307463037))) - ((7.0 * __pyx_v_r) * (((938918.400156317 * __pyx_v_nu) - (185763.092693281 * __pyx_v_z19)) - 245760.0))) - (5416406.59541186 * __pyx_v_z19)) - 3440640.0) / (((((((((32768.0 * __pyx_v_nu) * __pyx_v_z65) * ((((((((-38842241.4769507 * __pyx_v_nu) + ((240.0 * __pyx_v_r) * (((-7466.27061066206 * __pyx_v_nu) - (3024.0 * __pyx_v_z19)) + 17264.0))) + ((960.0 * __pyx_v_z0) * __pyx_v_z69)) + ((1920.0 * __pyx_v_z15) * ((588.0 * __pyx_v_nu) + 1079.0))) - (1882456.23663972 * __pyx_v_z19)) + ((480.0 * __pyx_v_z23) * __pyx_v_z69)) + (161280.0 * __pyx_v_z60)) + 13447680.0)) + ((53760.0 * __pyx_v_nu) * (((((7680.0 * __pyx_v_a6) * (__pyx_v_z23 + __pyx_v_z70)) + ((113485.217444961 * __pyx_v_r) * ((-__pyx_v_z23) + __pyx_v_z70))) + ((148.04406601634 * __pyx_v_r) * (((((7704.0 * __pyx_v_r) + (1926.0 * __pyx_v_z0)) + (3852.0 * __pyx_v_z15)) + (349.0 * __pyx_v_z23)) + 36400.0))) + ((128.0 * __pyx_v_r) * (((((13218.7851094412 * __pyx_v_r) + (4264.6962773603 * __pyx_v_z0)) + (8529.39255472061 * __pyx_v_z15)) - (6852.34813868015 * __pyx_v_z23)) - 33722.4297811176))))) + ((67645734912.0 * __pyx_v_z19) * __pyx_v_z66)) + ((7.0 * __pyx_v_z19) * (((((((((-39321600.0 * __pyx_v_a6) * ((3.0 * __pyx_v_r) + 59.0)) + (745857848.115604 * __pyx_v_a6)) + (122635399361.987 * __pyx_v_r)) - (6178501407.53758 * __pyx_v_z0)) + (2064783811.32587 * __pyx_v_z15)) - (3089250703.76879 * __pyx_v_z23)) + (1426660551.8844 * __pyx_v_z60)) + 276057889687.011))) + (13212057600.0 * __pyx_v_z60)) + (241555486248.807 * __pyx_v_z64)) + ((1120.0 * __pyx_v_z67) * (((-163683964.822551 * __pyx_v_r) - (17833256.898555 * __pyx_v_z15)) - 1188987459.03162))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1779
 *         cdef double z70 = 8.0*r + 2.0*z0 + 4.0*z15 + 16.0
 *         cdef double z71 = (2048.0*nu*z65*(336.0*r + z68 + 407.0) + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*z19 - 245760.0) - 5416406.59541186*z19 - 3440640.0)/(32768.0*nu*z65*(-38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*z19 + 17264.0) + 960.0*z0*z69 + 1920.0*z15*(588.0*nu + 1079.0) - 1882456.23663972*z19 + 480.0*z23*z69 + 161280.0*z60 + 13447680.0) + 53760.0*nu*(7680.0*a6*(z23 + z70) + 113485.217444961*r*(-z23 + z70) + 148.04406601634*r*(7704.0*r + 1926.0*z0 + 3852.0*z15 + 349.0*z23 + 36400.0) + 128.0*r*(13218.7851094412*r + 4264.6962773603*z0 + 8529.39255472061*z15 - 6852.34813868015*z23 - 33722.4297811176)) + 67645734912.0*z19*z66 + 7.0*z19*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*z0 + 2064783811.32587*z15 - 3089250703.76879*z23 + 1426660551.8844*z60 + 276057889687.011) + 13212057600.0*z60 + 241555486248.807*z64 + 1120.0*z67*(-163683964.822551*r - 17833256.898555*z15 - 1188987459.03162))
 *         cdef double z72 = z23*z71             # <<<<<<<<<<<<<<
 *         cdef double z73 = z24*(0.125*z36*(z62 - 31.0) + 2.0625*z48 + 0.0625*z53*(-z59 - 3.0)) + z24*(1.125*z2 + 1.125*z5 + 0.125*z52*(z59 + 1.0) - 1.25*z58 + 1.125*z6) + z37 + z49 + z61*(0.125*z48*(68.0*nu - 1.0) + z55*(2870.0*nu - z54 + 561.0) + z56*(1166.0*nu - 328.0*z19 - 171.0)) + z61*(-1.21875*z58*(2.0*nu - 3.0) + 0.015625*z63*(-27.0*nu + 28.0*z19 - 3.0) - 0.015625*(z62 + 9.0)*(25.0*z2 + 25.0*z5 + 25.0*z6)) + 7680.0*z72
 *         cdef double z74 = prst**4
 */
  __Pyx_TraceLine(1779,0,__PYX_ERR(1, 1779, __pyx_L1_error))
  __pyx_v_z72 = (__pyx_v_z23 * __pyx_v_z71);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1780
 *         cdef double z71 = (2048.0*nu*z65*(336.0*r + z68 + 407.0) + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*z19 - 245760.0) - 5416406.59541186*z19 - 3440640.0)/(32768.0*nu*z65*(-38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*z19 + 17264.0) + 960.0*z0*z69 + 1920.0*z15*(588.0*nu + 1079.0) - 1882456.23663972*z19 + 480.0*z23*z69 + 161280.0*z60 + 13447680.0) + 53760.0*nu*(7680.0*a6*(z23 + z70) + 113485.217444961*r*(-z23 + z70) + 148.04406601634*r*(7704.0*r + 1926.0*z0 + 3852.0*z15 + 349.0*z23 + 36400.0) + 128.0*r*(13218.7851094412*r + 4264.6962773603*z0 + 8529.39255472061*z15 - 6852.34813868015*z23 - 33722.4297811176)) + 67645734912.0*z19*z66 + 7.0*z19*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r - 6178501407.53758*z0 + 2064783811.32587*z15 - 3089250703.76879*z23 + 1426660551.8844*z60 + 276057889687.011) + 13212057600.0*z60 + 241555486248.807*z64 + 1120.0*z67*(-163683964.822551*r - 17833256.898555*z15 - 1188987459.03162))
 *         cdef double z72 = z23*z71
 *         cdef double z73 = z24*(0.125*z36*(z62 - 31.0) + 2.0625*z48 + 0.0625*z53*(-z59 - 3.0)) + z24*(1.125*z2 + 1.125*z5 + 0.125*z52*(z59 + 1.0) - 1.25*z58 + 1.125*z6) + z37 + z49 + z61*(0.125*z48*(68.0*nu - 1.0) + z55*(2870.0*nu - z54 + 561.0) + z56*(1166.0*nu - 328.0*z19 - 171.0)) + z61*(-1.21875*z58*(2.0*nu - 3.0) + 0.015625*z63*(-27.0*nu + 28.0*z19 - 3.0) - 0.015625*(z62 + 9.0)*(25.0*z2 + 25.0*z5 + 25.0*z6)) + 7680.0*z72             # <<<<<<<<<<<<<<
 *         cdef double z74 = prst**4
 *         cdef double z75 = prst**6
 */
  __Pyx_TraceLine(1780,0,__PYX_ERR(1, 1780, __pyx_L1_error))
  __pyx_v_z73 = (((((((__pyx_v_z24 * ((((0.125 * __pyx_v_z36) * (__pyx_v_z62 - 31.0)) + (2.0625 * __pyx_v_z48)) + ((0.0625 * __pyx_v_z53) * ((-__pyx_v_z59) - 3.0)))) + (__pyx_v_z24 * (((((1.125 * __pyx_v_z2) + (1.125 * __pyx_v_z5)) + ((0.125 * __pyx_v_z52) * (__pyx_v_z59 + 1.0))) - (1.25 * __pyx_v_z58)) + (1.125 * __pyx_v_z6)))) + __pyx_v_z37) + __pyx_v_z49) + (__pyx_v_z61 * ((((0.125 * __pyx_v_z48) * ((68.0 * __pyx_v_nu) - 1.0)) + (__pyx_v_z55 * (((2870.0 * __pyx_v_nu) - __pyx_v_z54) + 561.0))) + (__pyx_v_z56 * (((1166.0 * __pyx_v_nu) - (328.0 * __pyx_v_z19)) - 171.0))))) + (__pyx_v_z61 * ((((-1.21875 * __pyx_v_z58) * ((2.0 * __pyx_v_nu) - 3.0)) + ((0.015625 * __pyx_v_z63) * (((-27.0 * __pyx_v_nu) + (28.0 * __pyx_v_z19)) - 3.0))) - ((0.015625 * (__pyx_v_z62 + 9.0)) * (((25.0 * __pyx_v_z2) + (25.0 * __pyx_v_z5)) + (25.0 * __pyx_v_z6)))))) + (7680.0 * __pyx_v_z72));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1781
 *         cdef double z72 = z23*z71
 *         cdef double z73 = z24*(0.125*z36*(z62 - 31.0) + 2.0625*z48 + 0.0625*z53*(-z59 - 3.0)) + z24*(1.125*z2 + 1.125*z5 + 0.125*z52*(z59 + 1.0) - 1.25*z58 + 1.125*z6) + z37 + z49 + z61*(0.125*z48*(68.0*nu - 1.0) + z55*(2870.0*nu - z54 + 561.0) + z56*(1166.0*nu - 328.0*z19 - 171.0)) + z61*(-1.21875*z58*(2.0*nu - 3.0) + 0.015625*z63*(-27.0*nu + 28.0*z19 - 3.0) - 0.015625*(z62 + 9.0)*(25.0*z2 + 25.0*z5 + 25.0*z6)) + 7680.0*z72
 *         cdef double z74 = prst**4             # <<<<<<<<<<<<<<
 *         cdef double z75 = prst**6
 *         cdef double z76 = prst**8
 */
  __Pyx_TraceLine(1781,0,__PYX_ERR(1, 1781, __pyx_L1_error))
  __pyx_v_z74 = pow(__pyx_v_prst, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1782
 *         cdef double z73 = z24*(0.125*z36*(z62 - 31.0) + 2.0625*z48 + 0.0625*z53*(-z59 - 3.0)) + z24*(1.125*z2 + 1.125*z5 + 0.125*z52*(z59 + 1.0) - 1.25*z58 + 1.125*z6) + z37 + z49 + z61*(0.125*z48*(68.0*nu - 1.0) + z55*(2870.0*nu - z54 + 561.0) + z56*(1166.0*nu - 328.0*z19 - 171.0)) + z61*(-1.21875*z58*(2.0*nu - 3.0) + 0.015625*z63*(-27.0*nu + 28.0*z19 - 3.0) - 0.015625*(z62 + 9.0)*(25.0*z2 + 25.0*z5 + 25.0*z6)) + 7680.0*z72
 *         cdef double z74 = prst**4
 *         cdef double z75 = prst**6             # <<<<<<<<<<<<<<
 *         cdef double z76 = prst**8
 *         cdef double z77 = nu*z76
 */
  __Pyx_TraceLine(1782,0,__PYX_ERR(1, 1782, __pyx_L1_error))
  __pyx_v_z75 = pow(__pyx_v_prst, 6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1783
 *         cdef double z74 = prst**4
 *         cdef double z75 = prst**6
 *         cdef double z76 = prst**8             # <<<<<<<<<<<<<<
 *         cdef double z77 = nu*z76
 *         cdef double z78 = z15 + z8
 */
  __Pyx_TraceLine(1783,0,__PYX_ERR(1, 1783, __pyx_L1_error))
  __pyx_v_z76 = pow(__pyx_v_prst, 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1784
 *         cdef double z75 = prst**6
 *         cdef double z76 = prst**8
 *         cdef double z77 = nu*z76             # <<<<<<<<<<<<<<
 *         cdef double z78 = z15 + z8
 *         cdef double z79 = nu*r
 */
  __Pyx_TraceLine(1784,0,__PYX_ERR(1, 1784, __pyx_L1_error))
  __pyx_v_z77 = (__pyx_v_nu * __pyx_v_z76);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1785
 *         cdef double z76 = prst**8
 *         cdef double z77 = nu*z76
 *         cdef double z78 = z15 + z8             # <<<<<<<<<<<<<<
 *         cdef double z79 = nu*r
 *         cdef double z80 = nu*z15
 */
  __Pyx_TraceLine(1785,0,__PYX_ERR(1, 1785, __pyx_L1_error))
  __pyx_v_z78 = (__pyx_v_z15 + __pyx_v_z8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1786
 *         cdef double z77 = nu*z76
 *         cdef double z78 = z15 + z8
 *         cdef double z79 = nu*r             # <<<<<<<<<<<<<<
 *         cdef double z80 = nu*z15
 *         cdef double z81 = r*z19
 */
  __Pyx_TraceLine(1786,0,__PYX_ERR(1, 1786, __pyx_L1_error))
  __pyx_v_z79 = (__pyx_v_nu * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1787
 *         cdef double z78 = z15 + z8
 *         cdef double z79 = nu*r
 *         cdef double z80 = nu*z15             # <<<<<<<<<<<<<<
 *         cdef double z81 = r*z19
 *         cdef double z82 = 14700.0*nu + 42911.0
 */
  __Pyx_TraceLine(1787,0,__PYX_ERR(1, 1787, __pyx_L1_error))
  __pyx_v_z80 = (__pyx_v_nu * __pyx_v_z15);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1788
 *         cdef double z79 = nu*r
 *         cdef double z80 = nu*z15
 *         cdef double z81 = r*z19             # <<<<<<<<<<<<<<
 *         cdef double z82 = 14700.0*nu + 42911.0
 *         cdef double z83 = r*z82
 */
  __Pyx_TraceLine(1788,0,__PYX_ERR(1, 1788, __pyx_L1_error))
  __pyx_v_z81 = (__pyx_v_r * __pyx_v_z19);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1789
 *         cdef double z80 = nu*z15
 *         cdef double z81 = r*z19
 *         cdef double z82 = 14700.0*nu + 42911.0             # <<<<<<<<<<<<<<
 *         cdef double z83 = r*z82
 *         cdef double z84 = z15*z19
 */
  __Pyx_TraceLine(1789,0,__PYX_ERR(1, 1789, __pyx_L1_error))
  __pyx_v_z82 = ((14700.0 * __pyx_v_nu) + 42911.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1790
 *         cdef double z81 = r*z19
 *         cdef double z82 = 14700.0*nu + 42911.0
 *         cdef double z83 = r*z82             # <<<<<<<<<<<<<<
 *         cdef double z84 = z15*z19
 *         cdef double z85 = 283115520.0*z82
 */
  __Pyx_TraceLine(1790,0,__PYX_ERR(1, 1790, __pyx_L1_error))
  __pyx_v_z83 = (__pyx_v_r * __pyx_v_z82);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1791
 *         cdef double z82 = 14700.0*nu + 42911.0
 *         cdef double z83 = r*z82
 *         cdef double z84 = z15*z19             # <<<<<<<<<<<<<<
 *         cdef double z85 = 283115520.0*z82
 *         cdef double z86 = z65*(-25392914995744.3*nu - r*z85 - 879923036160.0*z15 - 5041721180160.0*z19 + 104186110149937.0)
 */
  __Pyx_TraceLine(1791,0,__PYX_ERR(1, 1791, __pyx_L1_error))
  __pyx_v_z84 = (__pyx_v_z15 * __pyx_v_z19);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1792
 *         cdef double z83 = r*z82
 *         cdef double z84 = z15*z19
 *         cdef double z85 = 283115520.0*z82             # <<<<<<<<<<<<<<
 *         cdef double z86 = z65*(-25392914995744.3*nu - r*z85 - 879923036160.0*z15 - 5041721180160.0*z19 + 104186110149937.0)
 *         cdef double z87 = r*z66
 */
  __Pyx_TraceLine(1792,0,__PYX_ERR(1, 1792, __pyx_L1_error))
  __pyx_v_z85 = (283115520.0 * __pyx_v_z82);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1793
 *         cdef double z84 = z15*z19
 *         cdef double z85 = 283115520.0*z82
 *         cdef double z86 = z65*(-25392914995744.3*nu - r*z85 - 879923036160.0*z15 - 5041721180160.0*z19 + 104186110149937.0)             # <<<<<<<<<<<<<<
 *         cdef double z87 = r*z66
 *         cdef double z88 = z15*(-630116198.873299*nu - 197773496.793534*z19 + 5805304367.87913)
 */
  __Pyx_TraceLine(1793,0,__PYX_ERR(1, 1793, __pyx_L1_error))
  __pyx_v_z86 = (__pyx_v_z65 * (((((-25392914995744.3 * __pyx_v_nu) - (__pyx_v_r * __pyx_v_z85)) - (879923036160.0 * __pyx_v_z15)) - (5041721180160.0 * __pyx_v_z19)) + 104186110149937.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1794
 *         cdef double z85 = 283115520.0*z82
 *         cdef double z86 = z65*(-25392914995744.3*nu - r*z85 - 879923036160.0*z15 - 5041721180160.0*z19 + 104186110149937.0)
 *         cdef double z87 = r*z66             # <<<<<<<<<<<<<<
 *         cdef double z88 = z15*(-630116198.873299*nu - 197773496.793534*z19 + 5805304367.87913)
 *         cdef double z89 = z0*(-2675575.66847905*nu - 138240.0*z19 - 5278341.3229329)
 */
  __Pyx_TraceLine(1794,0,__PYX_ERR(1, 1794, __pyx_L1_error))
  __pyx_v_z87 = (__pyx_v_r * __pyx_v_z66);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1795
 *         cdef double z86 = z65*(-25392914995744.3*nu - r*z85 - 879923036160.0*z15 - 5041721180160.0*z19 + 104186110149937.0)
 *         cdef double z87 = r*z66
 *         cdef double z88 = z15*(-630116198.873299*nu - 197773496.793534*z19 + 5805304367.87913)             # <<<<<<<<<<<<<<
 *         cdef double z89 = z0*(-2675575.66847905*nu - 138240.0*z19 - 5278341.3229329)
 *         cdef double z90 = nu*(-2510664218.28128*nu - 42636451.6032331*z19 + 14515200.0*z67 + 1002013764.01019)
 */
  __Pyx_TraceLine(1795,0,__PYX_ERR(1, 1795, __pyx_L1_error))
  __pyx_v_z88 = (__pyx_v_z15 * (((-630116198.873299 * __pyx_v_nu) - (197773496.793534 * __pyx_v_z19)) + 5805304367.87913));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1796
 *         cdef double z87 = r*z66
 *         cdef double z88 = z15*(-630116198.873299*nu - 197773496.793534*z19 + 5805304367.87913)
 *         cdef double z89 = z0*(-2675575.66847905*nu - 138240.0*z19 - 5278341.3229329)             # <<<<<<<<<<<<<<
 *         cdef double z90 = nu*(-2510664218.28128*nu - 42636451.6032331*z19 + 14515200.0*z67 + 1002013764.01019)
 *         cdef double z91 = r*(43393301259014.8*nu + 43133561885859.3*z19 + 5927865218923.02*z67 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0)
 */
  __Pyx_TraceLine(1796,0,__PYX_ERR(1, 1796, __pyx_L1_error))
  __pyx_v_z89 = (__pyx_v_z0 * (((-2675575.66847905 * __pyx_v_nu) - (138240.0 * __pyx_v_z19)) - 5278341.3229329));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1797
 *         cdef double z88 = z15*(-630116198.873299*nu - 197773496.793534*z19 + 5805304367.87913)
 *         cdef double z89 = z0*(-2675575.66847905*nu - 138240.0*z19 - 5278341.3229329)
 *         cdef double z90 = nu*(-2510664218.28128*nu - 42636451.6032331*z19 + 14515200.0*z67 + 1002013764.01019)             # <<<<<<<<<<<<<<
 *         cdef double z91 = r*(43393301259014.8*nu + 43133561885859.3*z19 + 5927865218923.02*z67 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0)
 *         cdef double z92 = z65*(-1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*z19 - 2119671837.36038) + 879923036160.0*z0 + z15*z85)
 */
  __Pyx_TraceLine(1797,0,__PYX_ERR(1, 1797, __pyx_L1_error))
  __pyx_v_z90 = (__pyx_v_nu * ((((-2510664218.28128 * __pyx_v_nu) - (42636451.6032331 * __pyx_v_z19)) + (14515200.0 * __pyx_v_z67)) + 1002013764.01019));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1798
 *         cdef double z89 = z0*(-2675575.66847905*nu - 138240.0*z19 - 5278341.3229329)
 *         cdef double z90 = nu*(-2510664218.28128*nu - 42636451.6032331*z19 + 14515200.0*z67 + 1002013764.01019)
 *         cdef double z91 = r*(43393301259014.8*nu + 43133561885859.3*z19 + 5927865218923.02*z67 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0)             # <<<<<<<<<<<<<<
 *         cdef double z92 = z65*(-1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*z19 - 2119671837.36038) + 879923036160.0*z0 + z15*z85)
 *         cdef double z93 = 4.0*z19
 */
  __Pyx_TraceLine(1798,0,__PYX_ERR(1, 1798, __pyx_L1_error))
  __pyx_v_z91 = (__pyx_v_r * (((((43393301259014.8 * __pyx_v_nu) + (43133561885859.3 * __pyx_v_z19)) + (5927865218923.02 * __pyx_v_z67)) + (86618264430493.3 * pow((1.0 - (0.496948781616935 * __pyx_v_nu)), 2.0))) + 188440788778196.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1799
 *         cdef double z90 = nu*(-2510664218.28128*nu - 42636451.6032331*z19 + 14515200.0*z67 + 1002013764.01019)
 *         cdef double z91 = r*(43393301259014.8*nu + 43133561885859.3*z19 + 5927865218923.02*z67 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0)
 *         cdef double z92 = z65*(-1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*z19 - 2119671837.36038) + 879923036160.0*z0 + z15*z85)             # <<<<<<<<<<<<<<
 *         cdef double z93 = 4.0*z19
 *         cdef double z94 = 0.000130208333333333*z49 + z72
 */
  __Pyx_TraceLine(1799,0,__PYX_ERR(1, 1799, __pyx_L1_error))
  __pyx_v_z92 = (__pyx_v_z65 * (((((-1698693120.0 * __pyx_v_nu) * ((11592.0 * __pyx_v_nu) + 69847.0)) + ((49152.0 * __pyx_v_r) * (((409207698.136075 * __pyx_v_nu) + (102574080.0 * __pyx_v_z19)) - 2119671837.36038))) + (879923036160.0 * __pyx_v_z0)) + (__pyx_v_z15 * __pyx_v_z85)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1800
 *         cdef double z91 = r*(43393301259014.8*nu + 43133561885859.3*z19 + 5927865218923.02*z67 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0)
 *         cdef double z92 = z65*(-1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*z19 - 2119671837.36038) + 879923036160.0*z0 + z15*z85)
 *         cdef double z93 = 4.0*z19             # <<<<<<<<<<<<<<
 *         cdef double z94 = 0.000130208333333333*z49 + z72
 *         cdef double z95 = -12049908701745.2*nu - 39476764256925.6*r + 5107745331375.71*z15 + 80059249540278.2*z19 + 5787938193408.0*z66 + 6730497718123.02*z67 + 10611661054566.2*z79 + 2589101062873.81*z80 + 1822680546449.21*z81 - 326837426.241486*z83 + 133772083200.0*z84 - z86 + 275059053208689.0
 */
  __Pyx_TraceLine(1800,0,__PYX_ERR(1, 1800, __pyx_L1_error))
  __pyx_v_z93 = (4.0 * __pyx_v_z19);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1801
 *         cdef double z92 = z65*(-1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*z19 - 2119671837.36038) + 879923036160.0*z0 + z15*z85)
 *         cdef double z93 = 4.0*z19
 *         cdef double z94 = 0.000130208333333333*z49 + z72             # <<<<<<<<<<<<<<
 *         cdef double z95 = -12049908701745.2*nu - 39476764256925.6*r + 5107745331375.71*z15 + 80059249540278.2*z19 + 5787938193408.0*z66 + 6730497718123.02*z67 + 10611661054566.2*z79 + 2589101062873.81*z80 + 1822680546449.21*z81 - 326837426.241486*z83 + 133772083200.0*z84 - z86 + 275059053208689.0
 *         cdef double z96 = 5787938193408.0*z87 - 9216.0*z88 - 967680.0*z89 + 55296.0*z90 + z91 + z92
 */
  __Pyx_TraceLine(1801,0,__PYX_ERR(1, 1801, __pyx_L1_error))
  __pyx_v_z94 = ((0.000130208333333333 * __pyx_v_z49) + __pyx_v_z72);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1802
 *         cdef double z93 = 4.0*z19
 *         cdef double z94 = 0.000130208333333333*z49 + z72
 *         cdef double z95 = -12049908701745.2*nu - 39476764256925.6*r + 5107745331375.71*z15 + 80059249540278.2*z19 + 5787938193408.0*z66 + 6730497718123.02*z67 + 10611661054566.2*z79 + 2589101062873.81*z80 + 1822680546449.21*z81 - 326837426.241486*z83 + 133772083200.0*z84 - z86 + 275059053208689.0             # <<<<<<<<<<<<<<
 *         cdef double z96 = 5787938193408.0*z87 - 9216.0*z88 - 967680.0*z89 + 55296.0*z90 + z91 + z92
 * 
 */
  __Pyx_TraceLine(1802,0,__PYX_ERR(1, 1802, __pyx_L1_error))
  __pyx_v_z95 = (((((((((((((-12049908701745.2 * __pyx_v_nu) - (39476764256925.6 * __pyx_v_r)) + (5107745331375.71 * __pyx_v_z15)) + (80059249540278.2 * __pyx_v_z19)) + (5787938193408.0 * __pyx_v_z66)) + (6730497718123.02 * __pyx_v_z67)) + (10611661054566.2 * __pyx_v_z79)) + (2589101062873.81 * __pyx_v_z80)) + (1822680546449.21 * __pyx_v_z81)) - (326837426.241486 * __pyx_v_z83)) + (133772083200.0 * __pyx_v_z84)) - __pyx_v_z86) + 275059053208689.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1803
 *         cdef double z94 = 0.000130208333333333*z49 + z72
 *         cdef double z95 = -12049908701745.2*nu - 39476764256925.6*r + 5107745331375.71*z15 + 80059249540278.2*z19 + 5787938193408.0*z66 + 6730497718123.02*z67 + 10611661054566.2*z79 + 2589101062873.81*z80 + 1822680546449.21*z81 - 326837426.241486*z83 + 133772083200.0*z84 - z86 + 275059053208689.0
 *         cdef double z96 = 5787938193408.0*z87 - 9216.0*z88 - 967680.0*z89 + 55296.0*z90 + z91 + z92             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1803,0,__PYX_ERR(1, 1803, __pyx_L1_error))
  __pyx_v_z96 = ((((((5787938193408.0 * __pyx_v_z87) - (9216.0 * __pyx_v_z88)) - (967680.0 * __pyx_v_z89)) + (55296.0 * __pyx_v_z90)) + __pyx_v_z91) + __pyx_v_z92);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1808
 *         # Evaluate Hamiltonian
 *         cdef double H
 *         H,_,_,_,_,_,_,_,_ = self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)             # <<<<<<<<<<<<<<
 * 
 *         # Heff Jacobian expressions
 */
  __Pyx_TraceLine(1808,0,__PYX_ERR(1, 1808, __pyx_L1_error))
  __pyx_t_1 = ((struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._call(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 9)) {
      if (size > 9) __Pyx_RaiseTooManyValuesError(9);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 1808, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_16 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_12 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_11 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 8); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_16 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_12 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_11 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_10 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 8); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_7);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[9] = {&__pyx_t_2,&__pyx_t_13,&__pyx_t_16,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_7};
      for (i=0; i < 9; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(1, 1808, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[9] = {&__pyx_t_2,&__pyx_t_13,&__pyx_t_16,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_7};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_19 = Py_TYPE(__pyx_t_6)->tp_iternext;
    for (index=0; index < 9; index++) {
      PyObject* item = __pyx_t_19(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_6), 9) < 0) __PYX_ERR(1, 1808, __pyx_L1_error)
    __pyx_t_19 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_19 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(1, 1808, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_H = __pyx_t_17;
  __pyx_v__ = __pyx_t_13;
  __pyx_t_13 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_16);
  __pyx_t_16 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_12);
  __pyx_t_12 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_11);
  __pyx_t_11 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_8);
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1812
 *         # Heff Jacobian expressions
 * 
 *         cdef double dHeffdpphi = z73*(z73*(147.443752990146*nu*r**(-4.5)*z74 - 11.3175085791863*nu*r**(-3.5)*z75 + 1.69542100694444e-8*prst**2*z61*z78**2*z96*(z10*(-2.625*z58 + 0.0625*z63*(z59 - 1.0) + z8*(3.0*nu + 2.8125)) + z24*(0.015625*z52*(115.0*nu + z93 - 37.0) + 0.03125*z58*(26.0*nu + 449.0) + 0.015625*z8*(-1171.0*nu - 861.0)) + z49 + 7680.0*z60*z71*z95/z96)/(z94**2*z95) + 1.48275342024365*r**(-2.5)*z77 + z10*z74*(92.7110442849544*nu - 131.0*z19 + 10.0*z67) + z10*z75*(-33.9782122170436*nu - 89.5298327361234*z19 - 14.0*z64 + 188.0*z67) + z16*z74*(8.0*nu - 6.0*z19) + z16*z75*(-2.78300763695006*nu - 5.4*z19 + 6.0*z67) + z16*z76*(1.38977750996128*nu + 3.33842023648322*z19 - 6.0*z64 + 3.42857142857143*z67) + z24*z74*(nu*(452.542166996693 - 51.6952380952381*z65) + z19*(118.4*z65 - 1796.13660498019) + 602.318540416564*z67) - z31*z39*z51 + 0.121954868780449*z39*z77 + z40*z57 + 1.0 + 1.27277314139085e-19*z74*z78**4*(0.3125*delta*z47*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*z19 - 5.0)*(5.0*z2 + 5.0*z5 + 5.0*z6) - 0.03125*(-5.0*nu + z93 + 1.0)*(-15.0*z2 + 15.0*z5 + 15.0*z6))*(z87 - 1.59227685093395e-9*z88 - 1.67189069348064e-7*z89 + 9.55366110560367e-9*z90 + 1.72773095804465e-13*z91 + 1.72773095804465e-13*z92)**2/(r**13*z94**4*(-0.0438084424460039*nu - 0.143521050466841*r + 0.0185696317637669*z15 + 0.291062041428379*z19 + 0.0210425293255724*z66 + 0.0244692826489756*z67 + 0.0385795738434214*z79 + 0.00941289164152486*z80 + 0.00662650629087394*z81 - 1.18824456940711e-6*z83 + 0.000486339502879429*z84 - 3.63558293513537e-15*z86 + 1)**2))/(z49*(z50 + 1.0) + 1.0))**(-0.5)*(2.0*pphi*z16*z57 - pphi*z50*z51)/(z49*(4.0*z39 + 2.0) + 2.0) + (nu*dSO*z10*z13 + pphi*z13*(z14*z18 + z20*z21 + z22*z25) + pphi*z30*(z18*z26 + z21*z27 + z25*z28) + 2.0*z13*z45 + z13*(z16*(0.208333333333333*delta*z47 - 0.625*z2 + 0.375*z35 - 0.833333333333333*z48 - 0.625*z5 - 0.625*z6) + z45) + z13*(z14*z40 + z16*(-5.53125*nu + 0.567708333333333*z19 - 0.078125) + z20*z41 + z22*z42 + z39*(0.71875*nu - 0.09375) + 1.75) + z29*(-0.125*delta*z38 + z16*(0.0416666666666667*delta*z8 + z46*(0.416666666666667 - 0.833333333333333*X_2))) - 0.25*z30*z38 + z30*(z16*(-0.03125*nu + 0.536458333333333*z19 + 0.078125) + z26*z40 + z27*z41 + z28*z42 + z39*(0.34375*nu + 0.09375) + 0.25))/(z0 + z9)             # <<<<<<<<<<<<<<
 * 
 *         # Compute H Jacobian
 */
  __Pyx_TraceLine(1812,0,__PYX_ERR(1, 1812, __pyx_L1_error))
  __pyx_v_dHeffdpphi = ((((__pyx_v_z73 * pow(((__pyx_v_z73 * (((((((((((((((((147.443752990146 * __pyx_v_nu) * pow(__pyx_v_r, -4.5)) * __pyx_v_z74) - (((11.3175085791863 * __pyx_v_nu) * pow(__pyx_v_r, -3.5)) * __pyx_v_z75)) + ((((((1.69542100694444e-8 * pow(__pyx_v_prst, 2.0)) * __pyx_v_z61) * pow(__pyx_v_z78, 2.0)) * __pyx_v_z96) * ((((__pyx_v_z10 * (((-2.625 * __pyx_v_z58) + ((0.0625 * __pyx_v_z63) * (__pyx_v_z59 - 1.0))) + (__pyx_v_z8 * ((3.0 * __pyx_v_nu) + 2.8125)))) + (__pyx_v_z24 * ((((0.015625 * __pyx_v_z52) * (((115.0 * __pyx_v_nu) + __pyx_v_z93) - 37.0)) + ((0.03125 * __pyx_v_z58) * ((26.0 * __pyx_v_nu) + 449.0))) + ((0.015625 * __pyx_v_z8) * ((-1171.0 * __pyx_v_nu) - 861.0))))) + __pyx_v_z49) + ((((7680.0 * __pyx_v_z60) * __pyx_v_z71) * __pyx_v_z95) / __pyx_v_z96))) / (pow(__pyx_v_z94, 2.0) * __pyx_v_z95))) + ((1.48275342024365 * pow(__pyx_v_r, -2.5)) * __pyx_v_z77)) + ((__pyx_v_z10 * __pyx_v_z74) * (((92.7110442849544 * __pyx_v_nu) - (131.0 * __pyx_v_z19)) + (10.0 * __pyx_v_z67)))) + ((__pyx_v_z10 * __pyx_v_z75) * ((((-33.9782122170436 * __pyx_v_nu) - (89.5298327361234 * __pyx_v_z19)) - (14.0 * __pyx_v_z64)) + (188.0 * __pyx_v_z67)))) + ((__pyx_v_z16 * __pyx_v_z74) * ((8.0 * __pyx_v_nu) - (6.0 * __pyx_v_z19)))) + ((__pyx_v_z16 * __pyx_v_z75) * (((-2.78300763695006 * __pyx_v_nu) - (5.4 * __pyx_v_z19)) + (6.0 * __pyx_v_z67)))) + ((__pyx_v_z16 * __pyx_v_z76) * ((((1.38977750996128 * __pyx_v_nu) + (3.33842023648322 * __pyx_v_z19)) - (6.0 * __pyx_v_z64)) + (3.42857142857143 * __pyx_v_z67)))) + ((__pyx_v_z24 * __pyx_v_z74) * (((__pyx_v_nu * (452.542166996693 - (51.6952380952381 * __pyx_v_z65))) + (__pyx_v_z19 * ((118.4 * __pyx_v_z65) - 1796.13660498019))) + (602.318540416564 * __pyx_v_z67)))) - ((__pyx_v_z31 * __pyx_v_z39) * __pyx_v_z51)) + ((0.121954868780449 * __pyx_v_z39) * __pyx_v_z77)) + (__pyx_v_z40 * __pyx_v_z57)) + 1.0) + (((((1.27277314139085e-19 * __pyx_v_z74) * pow(__pyx_v_z78, 4.0)) * (((((0.3125 * __pyx_v_delta) * __pyx_v_z47) * ((18.0 * __pyx_v_nu) - 1.0)) - ((0.03125 * (((-33.0 * __pyx_v_nu) + (32.0 * __pyx_v_z19)) - 5.0)) * (((5.0 * __pyx_v_z2) + (5.0 * __pyx_v_z5)) + (5.0 * __pyx_v_z6)))) - ((0.03125 * (((-5.0 * __pyx_v_nu) + __pyx_v_z93) + 1.0)) * (((-15.0 * __pyx_v_z2) + (15.0 * __pyx_v_z5)) + (15.0 * __pyx_v_z6))))) * pow((((((__pyx_v_z87 - (1.59227685093395e-9 * __pyx_v_z88)) - (1.67189069348064e-7 * __pyx_v_z89)) + (9.55366110560367e-9 * __pyx_v_z90)) + (1.72773095804465e-13 * __pyx_v_z91)) + (1.72773095804465e-13 * __pyx_v_z92)), 2.0)) / ((pow(__pyx_v_r, 13.0) * pow(__pyx_v_z94, 4.0)) * pow((((((((((((((-0.0438084424460039 * __pyx_v_nu) - (0.143521050466841 * __pyx_v_r)) + (0.0185696317637669 * __pyx_v_z15)) + (0.291062041428379 * __pyx_v_z19)) + (0.0210425293255724 * __pyx_v_z66)) + (0.0244692826489756 * __pyx_v_z67)) + (0.0385795738434214 * __pyx_v_z79)) + (0.00941289164152486 * __pyx_v_z80)) + (0.00662650629087394 * __pyx_v_z81)) - (1.18824456940711e-6 * __pyx_v_z83)) + (0.000486339502879429 * __pyx_v_z84)) - (3.63558293513537e-15 * __pyx_v_z86)) + 1.0), 2.0))))) / ((__pyx_v_z49 * (__pyx_v_z50 + 1.0)) + 1.0)), -0.5)) * ((((2.0 * __pyx_v_pphi) * __pyx_v_z16) * __pyx_v_z57) - ((__pyx_v_pphi * __pyx_v_z50) * __pyx_v_z51))) / ((__pyx_v_z49 * ((4.0 * __pyx_v_z39) + 2.0)) + 2.0)) + ((((((((((((__pyx_v_nu * __pyx_v_dSO) * __pyx_v_z10) * __pyx_v_z13) + ((__pyx_v_pphi * __pyx_v_z13) * (((__pyx_v_z14 * __pyx_v_z18) + (__pyx_v_z20 * __pyx_v_z21)) + (__pyx_v_z22 * __pyx_v_z25)))) + ((__pyx_v_pphi * __pyx_v_z30) * (((__pyx_v_z18 * __pyx_v_z26) + (__pyx_v_z21 * __pyx_v_z27)) + (__pyx_v_z25 * __pyx_v_z28)))) + ((2.0 * __pyx_v_z13) * __pyx_v_z45)) + (__pyx_v_z13 * ((__pyx_v_z16 * (((((((0.208333333333333 * __pyx_v_delta) * __pyx_v_z47) - (0.625 * __pyx_v_z2)) + (0.375 * __pyx_v_z35)) - (0.833333333333333 * __pyx_v_z48)) - (0.625 * __pyx_v_z5)) - (0.625 * __pyx_v_z6))) + __pyx_v_z45))) + (__pyx_v_z13 * ((((((__pyx_v_z14 * __pyx_v_z40) + (__pyx_v_z16 * (((-5.53125 * __pyx_v_nu) + (0.567708333333333 * __pyx_v_z19)) - 0.078125))) + (__pyx_v_z20 * __pyx_v_z41)) + (__pyx_v_z22 * __pyx_v_z42)) + (__pyx_v_z39 * ((0.71875 * __pyx_v_nu) - 0.09375))) + 1.75))) + (__pyx_v_z29 * (((-0.125 * __pyx_v_delta) * __pyx_v_z38) + (__pyx_v_z16 * (((0.0416666666666667 * __pyx_v_delta) * __pyx_v_z8) + (__pyx_v_z46 * (0.416666666666667 - (0.833333333333333 * __pyx_v_X_2)))))))) - ((0.25 * __pyx_v_z30) * __pyx_v_z38)) + (__pyx_v_z30 * ((((((__pyx_v_z16 * (((-0.03125 * __pyx_v_nu) + (0.536458333333333 * __pyx_v_z19)) + 0.078125)) + (__pyx_v_z26 * __pyx_v_z40)) + (__pyx_v_z27 * __pyx_v_z41)) + (__pyx_v_z28 * __pyx_v_z42)) + (__pyx_v_z39 * ((0.34375 * __pyx_v_nu) + 0.09375))) + 0.25))) / (__pyx_v_z0 + __pyx_v_z9)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1816
 *         # Compute H Jacobian
 * 
 *         cdef double omega = M * M * dHeffdpphi / (nu*H)             # <<<<<<<<<<<<<<
 * 
 *         return omega
 */
  __Pyx_TraceLine(1816,0,__PYX_ERR(1, 1816, __pyx_L1_error))
  __pyx_v_omega = (((__pyx_v_M * __pyx_v_M) * __pyx_v_dHeffdpphi) / (__pyx_v_nu * __pyx_v_H));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1818
 *         cdef double omega = M * M * dHeffdpphi / (nu*H)
 * 
 *         return omega             # <<<<<<<<<<<<<<
 * 
 *     cpdef auxderivs(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 */
  __Pyx_TraceLine(1818,0,__PYX_ERR(1, 1818, __pyx_L1_error))
  __pyx_r = __pyx_v_omega;
  goto __pyx_L0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1657
 *         return dHdr, dHdphi, dHdpr, dHdpphi,H,xi
 * 
 *     cpdef double omega(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_WriteUnraisable("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.omega", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c_coeffs);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_13omega(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_12omega[] = "\n        Compute the orbital frequency from the Hamiltonian.\n\n        Args:\n          q (double[:]): Canonical positions (r,phi).\n          p (double[:]): Canonical momenta  (prstar,pphi).\n          chi1_v (double[:]): Dimensionless spin vector of the primary.\n          chi2_v (double[:]): Dimensionless spin vector of the secondary.\n          m_1 (double): Primary mass component.\n          m_2 (double): Secondary mass component.\n          chi_1 (double): Projection of chi1_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chi_2 (double): Projection of chi2_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chiL1 (double): Projection of chi1_v onto the orbital angular momentum unit vector (l).\n          chiL2 (double): Projection of chi2_v onto the orbital angular momentum unit vector (l).\n\n        Returns:\n           (double) dHdpphi\n        ";
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_13omega(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_q = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi1_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi2_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_m_1;
  double __pyx_v_m_2;
  double __pyx_v_chi_1;
  double __pyx_v_chi_2;
  double __pyx_v_chiL1;
  double __pyx_v_chiL2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("omega (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_q,&__pyx_n_s_p,&__pyx_n_s_chi1_v,&__pyx_n_s_chi2_v,&__pyx_n_s_m_1,&__pyx_n_s_m_2,&__pyx_n_s_chi_1,&__pyx_n_s_chi_2,&__pyx_n_s_chiL1,&__pyx_n_s_chiL2,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_q)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 1); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi1_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 2); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi2_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 3); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 4); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 5); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 6); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 7); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 8); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, 9); __PYX_ERR(1, 1657, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "omega") < 0)) __PYX_ERR(1, 1657, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_q = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_q.memview)) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_p.memview)) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_chi1_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi1_v.memview)) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_chi2_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi2_v.memview)) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_m_1 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_m_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_m_2 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_m_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_chi_1 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_chi_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_chi_2 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_chi_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_chiL1 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_chiL1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1657, __pyx_L3_error)
    __pyx_v_chiL2 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_chiL2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1657, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("omega", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1657, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.omega", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_12omega(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_12omega(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("omega", 0);
  __Pyx_TraceCall("omega (wrapper)", __pyx_f[1], 1657, 0, __PYX_ERR(1, 1657, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 1657, __pyx_L1_error) }
  if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 1657, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 1657, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 1657, __pyx_L1_error) }
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_omega(__pyx_v_self, __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.omega", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_q, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi1_v, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi2_v, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1820
 *         return omega
 * 
 *     cpdef auxderivs(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_15auxderivs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_auxderivs(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, CYTHON_UNUSED double __pyx_v_m_1, CYTHON_UNUSED double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2, int __pyx_skip_dispatch) {
  double __pyx_v_r;
  CYTHON_UNUSED double __pyx_v_phi;
  double __pyx_v_prst;
  double __pyx_v_L;
  double __pyx_v_chix1;
  double __pyx_v_chiy1;
  double __pyx_v_chiz1;
  double __pyx_v_chix2;
  double __pyx_v_chiy2;
  double __pyx_v_chiz2;
  double __pyx_v_pphi;
  struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *__pyx_v_c_coeffs = 0;
  double __pyx_v_a6;
  double __pyx_v_dSO;
  CYTHON_UNUSED double __pyx_v_M;
  double __pyx_v_nu;
  double __pyx_v_X_1;
  double __pyx_v_X_2;
  double __pyx_v_delta;
  double __pyx_v_w0;
  double __pyx_v_w1;
  double __pyx_v_w2;
  double __pyx_v_w3;
  double __pyx_v_w4;
  double __pyx_v_w5;
  double __pyx_v_w6;
  double __pyx_v_w7;
  double __pyx_v_w8;
  double __pyx_v_w9;
  double __pyx_v_w10;
  double __pyx_v_w11;
  double __pyx_v_w12;
  double __pyx_v_w13;
  double __pyx_v_w14;
  double __pyx_v_w15;
  double __pyx_v_w16;
  double __pyx_v_w17;
  double __pyx_v_w18;
  double __pyx_v_w19;
  double __pyx_v_w20;
  double __pyx_v_w21;
  double __pyx_v_w22;
  double __pyx_v_w23;
  double __pyx_v_w24;
  double __pyx_v_w25;
  double __pyx_v_w26;
  double __pyx_v_w27;
  double __pyx_v_w28;
  double __pyx_v_w29;
  double __pyx_v_w30;
  double __pyx_v_w31;
  double __pyx_v_w32;
  double __pyx_v_w33;
  double __pyx_v_w34;
  double __pyx_v_w35;
  double __pyx_v_w36;
  double __pyx_v_w37;
  double __pyx_v_w38;
  double __pyx_v_w39;
  double __pyx_v_w40;
  double __pyx_v_w41;
  double __pyx_v_w42;
  double __pyx_v_w43;
  double __pyx_v_w44;
  double __pyx_v_w45;
  double __pyx_v_w46;
  double __pyx_v_w47;
  double __pyx_v_w48;
  double __pyx_v_w49;
  double __pyx_v_w50;
  double __pyx_v_w51;
  double __pyx_v_w52;
  double __pyx_v_w53;
  double __pyx_v_w54;
  double __pyx_v_w55;
  double __pyx_v_w56;
  double __pyx_v_w57;
  double __pyx_v_w58;
  double __pyx_v_w59;
  double __pyx_v_w60;
  double __pyx_v_w61;
  double __pyx_v_w62;
  double __pyx_v_w63;
  double __pyx_v_w64;
  double __pyx_v_w65;
  double __pyx_v_w66;
  double __pyx_v_w67;
  double __pyx_v_w68;
  double __pyx_v_w69;
  double __pyx_v_w70;
  double __pyx_v_w71;
  double __pyx_v_w72;
  double __pyx_v_w73;
  double __pyx_v_w74;
  double __pyx_v_w75;
  double __pyx_v_w76;
  double __pyx_v_w77;
  double __pyx_v_w78;
  double __pyx_v_w79;
  double __pyx_v_w80;
  double __pyx_v_w81;
  double __pyx_v_w82;
  double __pyx_v_w83;
  double __pyx_v_w84;
  double __pyx_v_w85;
  double __pyx_v_w86;
  double __pyx_v_w87;
  double __pyx_v_w88;
  double __pyx_v_w89;
  double __pyx_v_w90;
  double __pyx_v_w91;
  double __pyx_v_w92;
  double __pyx_v_w93;
  double __pyx_v_w94;
  double __pyx_v_w95;
  double __pyx_v_w96;
  double __pyx_v_w97;
  double __pyx_v_w98;
  double __pyx_v_w99;
  double __pyx_v_w100;
  double __pyx_v_w101;
  double __pyx_v_w102;
  double __pyx_v_w103;
  double __pyx_v_w104;
  double __pyx_v_w105;
  double __pyx_v_w106;
  double __pyx_v_w107;
  double __pyx_v_w108;
  double __pyx_v_w109;
  double __pyx_v_w110;
  double __pyx_v_w111;
  double __pyx_v_w112;
  double __pyx_v_w113;
  double __pyx_v_w114;
  double __pyx_v_w115;
  double __pyx_v_w116;
  double __pyx_v_w117;
  double __pyx_v_w118;
  double __pyx_v_w119;
  double __pyx_v_w120;
  double __pyx_v_w121;
  double __pyx_v_w122;
  double __pyx_v_w123;
  double __pyx_v_w124;
  double __pyx_v_w125;
  double __pyx_v_w126;
  double __pyx_v_w127;
  double __pyx_v_w128;
  double __pyx_v_w129;
  double __pyx_v_w130;
  double __pyx_v_w131;
  double __pyx_v_w132;
  double __pyx_v_w133;
  double __pyx_v_w134;
  double __pyx_v_w135;
  double __pyx_v_w136;
  double __pyx_v_w137;
  double __pyx_v_w138;
  double __pyx_v_w139;
  double __pyx_v_w140;
  double __pyx_v_w141;
  double __pyx_v_w142;
  double __pyx_v_w143;
  double __pyx_v_w144;
  double __pyx_v_w145;
  double __pyx_v_w146;
  double __pyx_v_w147;
  double __pyx_v_w148;
  double __pyx_v_w149;
  double __pyx_v_w150;
  double __pyx_v_w151;
  double __pyx_v_w152;
  double __pyx_v_w153;
  double __pyx_v_w154;
  double __pyx_v_w155;
  double __pyx_v_w156;
  double __pyx_v_w157;
  double __pyx_v_w158;
  double __pyx_v_w159;
  double __pyx_v_w160;
  double __pyx_v_w161;
  double __pyx_v_w162;
  double __pyx_v_w163;
  double __pyx_v_w164;
  double __pyx_v_w165;
  double __pyx_v_w166;
  double __pyx_v_w167;
  double __pyx_v_w168;
  double __pyx_v_w169;
  double __pyx_v_w170;
  double __pyx_v_w171;
  double __pyx_v_w172;
  double __pyx_v_w173;
  double __pyx_v_w174;
  double __pyx_v_w175;
  double __pyx_v_w176;
  double __pyx_v_w177;
  double __pyx_v_w178;
  double __pyx_v_w179;
  double __pyx_v_w180;
  double __pyx_v_w181;
  double __pyx_v_w182;
  double __pyx_v_w183;
  double __pyx_v_w184;
  double __pyx_v_w185;
  double __pyx_v_w186;
  double __pyx_v_w187;
  double __pyx_v_w188;
  double __pyx_v_w189;
  double __pyx_v_w190;
  double __pyx_v_w191;
  double __pyx_v_w192;
  double __pyx_v_w193;
  double __pyx_v_w194;
  double __pyx_v_w195;
  double __pyx_v_w196;
  double __pyx_v_w197;
  double __pyx_v_w198;
  double __pyx_v_w199;
  double __pyx_v_w200;
  double __pyx_v_w201;
  double __pyx_v_w202;
  double __pyx_v_w203;
  double __pyx_v_w204;
  double __pyx_v_w205;
  double __pyx_v_w206;
  double __pyx_v_w207;
  double __pyx_v_w208;
  double __pyx_v_w209;
  double __pyx_v_w210;
  double __pyx_v_w211;
  double __pyx_v_w212;
  double __pyx_v_w213;
  double __pyx_v_w214;
  double __pyx_v_w215;
  double __pyx_v_w216;
  double __pyx_v_w217;
  double __pyx_v_w218;
  double __pyx_v_w219;
  double __pyx_v_w220;
  double __pyx_v_w221;
  double __pyx_v_w222;
  double __pyx_v_w223;
  double __pyx_v_w224;
  double __pyx_v_w225;
  double __pyx_v_w226;
  double __pyx_v_w227;
  double __pyx_v_w228;
  double __pyx_v_w229;
  double __pyx_v_w230;
  double __pyx_v_dAdr;
  double __pyx_v_dBnpdr;
  double __pyx_v_dBnpadr;
  double __pyx_v_dxidr;
  double __pyx_v_dQdr;
  double __pyx_v_dQdprst;
  double __pyx_v_dHodddr;
  double __pyx_v_dBpdr;
  double __pyx_v_dHevendr;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  double __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("auxderivs", 0);
  __Pyx_TraceCall("auxderivs", __pyx_f[1], 1820, 0, __PYX_ERR(1, 1820, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_auxderivs); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1820, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_15auxderivs)) {
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 1820, __pyx_L1_error) }
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_q, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 1820, __pyx_L1_error) }
        __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_p, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 1820, __pyx_L1_error) }
        __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_chi1_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 1820, __pyx_L1_error) }
        __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_chi2_v, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_m_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyFloat_FromDouble(__pyx_v_m_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_chi_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyFloat_FromDouble(__pyx_v_chi_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_chiL1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = PyFloat_FromDouble(__pyx_v_chiL2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_13 = __pyx_t_1; __pyx_t_14 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1820, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[11] = {__pyx_t_14, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 10+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1820, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(10+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 1820, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_15, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_15, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_15, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_15, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_15, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_15, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_16, 8+__pyx_t_15, __pyx_t_11);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_16, 9+__pyx_t_15, __pyx_t_12);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_12 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1820, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1844
 *         # Coordinate definitions
 * 
 *         cdef double r = q[0]             # <<<<<<<<<<<<<<
 *         cdef double phi = q[1]
 * 
 */
  __Pyx_TraceLine(1844,0,__PYX_ERR(1, 1844, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_r = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1845
 * 
 *         cdef double r = q[0]
 *         cdef double phi = q[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double prst = p[0]
 */
  __Pyx_TraceLine(1845,0,__PYX_ERR(1, 1845, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_phi = (*((double *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_17 * __pyx_v_q.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1847
 *         cdef double phi = q[1]
 * 
 *         cdef double prst = p[0]             # <<<<<<<<<<<<<<
 *         cdef double L = p[1]
 * 
 */
  __Pyx_TraceLine(1847,0,__PYX_ERR(1, 1847, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_prst = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1848
 * 
 *         cdef double prst = p[0]
 *         cdef double L = p[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix1 = chi1_v[0]
 */
  __Pyx_TraceLine(1848,0,__PYX_ERR(1, 1848, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_L = (*((double *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_17 * __pyx_v_p.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1850
 *         cdef double L = p[1]
 * 
 *         cdef double chix1 = chi1_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]
 */
  __Pyx_TraceLine(1850,0,__PYX_ERR(1, 1850, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1851
 * 
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz1 = chi1_v[2]
 * 
 */
  __Pyx_TraceLine(1851,0,__PYX_ERR(1, 1851, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1852
 *         cdef double chix1 = chi1_v[0]
 *         cdef double chiy1 = chi1_v[1]
 *         cdef double chiz1 = chi1_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double chix2 = chi2_v[0]
 */
  __Pyx_TraceLine(1852,0,__PYX_ERR(1, 1852, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz1 = (*((double *) ( /* dim=0 */ (__pyx_v_chi1_v.data + __pyx_t_17 * __pyx_v_chi1_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1854
 *         cdef double chiz1 = chi1_v[2]
 * 
 *         cdef double chix2 = chi2_v[0]             # <<<<<<<<<<<<<<
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]
 */
  __Pyx_TraceLine(1854,0,__PYX_ERR(1, 1854, __pyx_L1_error))
  __pyx_t_17 = 0;
  __pyx_v_chix2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1855
 * 
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]             # <<<<<<<<<<<<<<
 *         cdef double chiz2 = chi2_v[2]
 * 
 */
  __Pyx_TraceLine(1855,0,__PYX_ERR(1, 1855, __pyx_L1_error))
  __pyx_t_17 = 1;
  __pyx_v_chiy2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1856
 *         cdef double chix2 = chi2_v[0]
 *         cdef double chiy2 = chi2_v[1]
 *         cdef double chiz2 = chi2_v[2]             # <<<<<<<<<<<<<<
 * 
 *         cdef double pphi = L
 */
  __Pyx_TraceLine(1856,0,__PYX_ERR(1, 1856, __pyx_L1_error))
  __pyx_t_17 = 2;
  __pyx_v_chiz2 = (*((double *) ( /* dim=0 */ (__pyx_v_chi2_v.data + __pyx_t_17 * __pyx_v_chi2_v.strides[0]) )));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1858
 *         cdef double chiz2 = chi2_v[2]
 * 
 *         cdef double pphi = L             # <<<<<<<<<<<<<<
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 */
  __Pyx_TraceLine(1858,0,__PYX_ERR(1, 1858, __pyx_L1_error))
  __pyx_v_pphi = __pyx_v_L;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1860
 *         cdef double pphi = L
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs             # <<<<<<<<<<<<<<
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']
 */
  __Pyx_TraceLine(1860,0,__PYX_ERR(1, 1860, __pyx_L1_error))
  __pyx_t_1 = ((PyObject *)__pyx_v_self->calibration_coeffs);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_c_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1861
 * 
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']             # <<<<<<<<<<<<<<
 *         cdef double dSO = c_coeffs['dSO']
 * 
 */
  __Pyx_TraceLine(1861,0,__PYX_ERR(1, 1861, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_a6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1861, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_a6 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1862
 *         cdef CalibCoeffs c_coeffs = self.calibration_coeffs
 *         cdef double a6 = c_coeffs['a6']
 *         cdef double dSO = c_coeffs['dSO']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1862,0,__PYX_ERR(1, 1862, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_c_coeffs), __pyx_n_u_dSO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dSO = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1866
 * 
 *         # Extra quantities used in the auxiliary derivatives
 *         cdef double M = self.EOBpars.p_params.M             # <<<<<<<<<<<<<<
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 */
  __Pyx_TraceLine(1866,0,__PYX_ERR(1, 1866, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->M;
  __pyx_v_M = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1867
 *         # Extra quantities used in the auxiliary derivatives
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu             # <<<<<<<<<<<<<<
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 */
  __Pyx_TraceLine(1867,0,__PYX_ERR(1, 1867, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->nu;
  __pyx_v_nu = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1868
 *         cdef double M = self.EOBpars.p_params.M
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1             # <<<<<<<<<<<<<<
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 */
  __Pyx_TraceLine(1868,0,__PYX_ERR(1, 1868, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_1;
  __pyx_v_X_1 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1869
 *         cdef double nu = self.EOBpars.p_params.nu
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2             # <<<<<<<<<<<<<<
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double w0 = r**(-1)
 */
  __Pyx_TraceLine(1869,0,__PYX_ERR(1, 1869, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->X_2;
  __pyx_v_X_2 = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1870
 *         cdef double X_1 = self.EOBpars.p_params.X_1
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta             # <<<<<<<<<<<<<<
 *         cdef double w0 = r**(-1)
 *         cdef double w1 = r**2
 */
  __Pyx_TraceLine(1870,0,__PYX_ERR(1, 1870, __pyx_L1_error))
  __pyx_t_18 = __pyx_v_self->__pyx_base.EOBpars->p_params->delta;
  __pyx_v_delta = __pyx_t_18;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1871
 *         cdef double X_2 = self.EOBpars.p_params.X_2
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double w0 = r**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w1 = r**2
 *         cdef double w2 = w1**(-1)
 */
  __Pyx_TraceLine(1871,0,__PYX_ERR(1, 1871, __pyx_L1_error))
  __pyx_v_w0 = pow(__pyx_v_r, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1872
 *         cdef double delta = self.EOBpars.p_params.delta
 *         cdef double w0 = r**(-1)
 *         cdef double w1 = r**2             # <<<<<<<<<<<<<<
 *         cdef double w2 = w1**(-1)
 *         cdef double w3 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 */
  __Pyx_TraceLine(1872,0,__PYX_ERR(1, 1872, __pyx_L1_error))
  __pyx_v_w1 = pow(__pyx_v_r, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1873
 *         cdef double w0 = r**(-1)
 *         cdef double w1 = r**2
 *         cdef double w2 = w1**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w3 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double w4 = X_2*w3
 */
  __Pyx_TraceLine(1873,0,__PYX_ERR(1, 1873, __pyx_L1_error))
  __pyx_v_w2 = pow(__pyx_v_w1, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1874
 *         cdef double w1 = r**2
 *         cdef double w2 = w1**(-1)
 *         cdef double w3 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)             # <<<<<<<<<<<<<<
 *         cdef double w4 = X_2*w3
 *         cdef double w5 = X_1**2
 */
  __Pyx_TraceLine(1874,0,__PYX_ERR(1, 1874, __pyx_L1_error))
  __pyx_v_w3 = (__pyx_v_X_1 * ((((2.0 * __pyx_v_chix1) * __pyx_v_chix2) + ((2.0 * __pyx_v_chiy1) * __pyx_v_chiy2)) + ((2.0 * __pyx_v_chiz1) * __pyx_v_chiz2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1875
 *         cdef double w2 = w1**(-1)
 *         cdef double w3 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double w4 = X_2*w3             # <<<<<<<<<<<<<<
 *         cdef double w5 = X_1**2
 *         cdef double w6 = chix1**2 + chiy1**2 + chiz1**2
 */
  __Pyx_TraceLine(1875,0,__PYX_ERR(1, 1875, __pyx_L1_error))
  __pyx_v_w4 = (__pyx_v_X_2 * __pyx_v_w3);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1876
 *         cdef double w3 = X_1*(2.0*chix1*chix2 + 2.0*chiy1*chiy2 + 2.0*chiz1*chiz2)
 *         cdef double w4 = X_2*w3
 *         cdef double w5 = X_1**2             # <<<<<<<<<<<<<<
 *         cdef double w6 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double w7 = w5*w6
 */
  __Pyx_TraceLine(1876,0,__PYX_ERR(1, 1876, __pyx_L1_error))
  __pyx_v_w5 = pow(__pyx_v_X_1, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1877
 *         cdef double w4 = X_2*w3
 *         cdef double w5 = X_1**2
 *         cdef double w6 = chix1**2 + chiy1**2 + chiz1**2             # <<<<<<<<<<<<<<
 *         cdef double w7 = w5*w6
 *         cdef double w8 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 */
  __Pyx_TraceLine(1877,0,__PYX_ERR(1, 1877, __pyx_L1_error))
  __pyx_v_w6 = ((pow(__pyx_v_chix1, 2.0) + pow(__pyx_v_chiy1, 2.0)) + pow(__pyx_v_chiz1, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1878
 *         cdef double w5 = X_1**2
 *         cdef double w6 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double w7 = w5*w6             # <<<<<<<<<<<<<<
 *         cdef double w8 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double w9 = w7 + w8
 */
  __Pyx_TraceLine(1878,0,__PYX_ERR(1, 1878, __pyx_L1_error))
  __pyx_v_w7 = (__pyx_v_w5 * __pyx_v_w6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1879
 *         cdef double w6 = chix1**2 + chiy1**2 + chiz1**2
 *         cdef double w7 = w5*w6
 *         cdef double w8 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)             # <<<<<<<<<<<<<<
 *         cdef double w9 = w7 + w8
 *         cdef double w10 = w4 + w9
 */
  __Pyx_TraceLine(1879,0,__PYX_ERR(1, 1879, __pyx_L1_error))
  __pyx_v_w8 = (pow(__pyx_v_X_2, 2.0) * ((pow(__pyx_v_chix2, 2.0) + pow(__pyx_v_chiy2, 2.0)) + pow(__pyx_v_chiz2, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1880
 *         cdef double w7 = w5*w6
 *         cdef double w8 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double w9 = w7 + w8             # <<<<<<<<<<<<<<
 *         cdef double w10 = w4 + w9
 *         cdef double w11 = w10*w2
 */
  __Pyx_TraceLine(1880,0,__PYX_ERR(1, 1880, __pyx_L1_error))
  __pyx_v_w9 = (__pyx_v_w7 + __pyx_v_w8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1881
 *         cdef double w8 = X_2**2*(chix2**2 + chiy2**2 + chiz2**2)
 *         cdef double w9 = w7 + w8
 *         cdef double w10 = w4 + w9             # <<<<<<<<<<<<<<
 *         cdef double w11 = w10*w2
 *         cdef double w12 = w11*(2.0*w0 + 1.0) + 1.0
 */
  __Pyx_TraceLine(1881,0,__PYX_ERR(1, 1881, __pyx_L1_error))
  __pyx_v_w10 = (__pyx_v_w4 + __pyx_v_w9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1882
 *         cdef double w9 = w7 + w8
 *         cdef double w10 = w4 + w9
 *         cdef double w11 = w10*w2             # <<<<<<<<<<<<<<
 *         cdef double w12 = w11*(2.0*w0 + 1.0) + 1.0
 *         cdef double w13 = r**4
 */
  __Pyx_TraceLine(1882,0,__PYX_ERR(1, 1882, __pyx_L1_error))
  __pyx_v_w11 = (__pyx_v_w10 * __pyx_v_w2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1883
 *         cdef double w10 = w4 + w9
 *         cdef double w11 = w10*w2
 *         cdef double w12 = w11*(2.0*w0 + 1.0) + 1.0             # <<<<<<<<<<<<<<
 *         cdef double w13 = r**4
 *         cdef double w14 = w13**(-1)
 */
  __Pyx_TraceLine(1883,0,__PYX_ERR(1, 1883, __pyx_L1_error))
  __pyx_v_w12 = ((__pyx_v_w11 * ((2.0 * __pyx_v_w0) + 1.0)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1884
 *         cdef double w11 = w10*w2
 *         cdef double w12 = w11*(2.0*w0 + 1.0) + 1.0
 *         cdef double w13 = r**4             # <<<<<<<<<<<<<<
 *         cdef double w14 = w13**(-1)
 *         cdef double w15 = 2.0*w4 + 2.0*w7 + 2.0*w8
 */
  __Pyx_TraceLine(1884,0,__PYX_ERR(1, 1884, __pyx_L1_error))
  __pyx_v_w13 = pow(__pyx_v_r, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1885
 *         cdef double w12 = w11*(2.0*w0 + 1.0) + 1.0
 *         cdef double w13 = r**4
 *         cdef double w14 = w13**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w15 = 2.0*w4 + 2.0*w7 + 2.0*w8
 *         cdef double w16 = -w15
 */
  __Pyx_TraceLine(1885,0,__PYX_ERR(1, 1885, __pyx_L1_error))
  __pyx_v_w14 = pow(__pyx_v_w13, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1886
 *         cdef double w13 = r**4
 *         cdef double w14 = w13**(-1)
 *         cdef double w15 = 2.0*w4 + 2.0*w7 + 2.0*w8             # <<<<<<<<<<<<<<
 *         cdef double w16 = -w15
 *         cdef double w17 = r**3
 */
  __Pyx_TraceLine(1886,0,__PYX_ERR(1, 1886, __pyx_L1_error))
  __pyx_v_w15 = (((2.0 * __pyx_v_w4) + (2.0 * __pyx_v_w7)) + (2.0 * __pyx_v_w8));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1887
 *         cdef double w14 = w13**(-1)
 *         cdef double w15 = 2.0*w4 + 2.0*w7 + 2.0*w8
 *         cdef double w16 = -w15             # <<<<<<<<<<<<<<
 *         cdef double w17 = r**3
 *         cdef double w18 = w17**(-1)
 */
  __Pyx_TraceLine(1887,0,__PYX_ERR(1, 1887, __pyx_L1_error))
  __pyx_v_w16 = (-__pyx_v_w15);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1888
 *         cdef double w15 = 2.0*w4 + 2.0*w7 + 2.0*w8
 *         cdef double w16 = -w15
 *         cdef double w17 = r**3             # <<<<<<<<<<<<<<
 *         cdef double w18 = w17**(-1)
 *         cdef double w19 = r**5
 */
  __Pyx_TraceLine(1888,0,__PYX_ERR(1, 1888, __pyx_L1_error))
  __pyx_v_w17 = pow(__pyx_v_r, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1889
 *         cdef double w16 = -w15
 *         cdef double w17 = r**3
 *         cdef double w18 = w17**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w19 = r**5
 *         cdef double w20 = w19**(-1)
 */
  __Pyx_TraceLine(1889,0,__PYX_ERR(1, 1889, __pyx_L1_error))
  __pyx_v_w18 = pow(__pyx_v_w17, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1890
 *         cdef double w17 = r**3
 *         cdef double w18 = w17**(-1)
 *         cdef double w19 = r**5             # <<<<<<<<<<<<<<
 *         cdef double w20 = w19**(-1)
 *         cdef double w21 = chi_1*X_1
 */
  __Pyx_TraceLine(1890,0,__PYX_ERR(1, 1890, __pyx_L1_error))
  __pyx_v_w19 = pow(__pyx_v_r, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1891
 *         cdef double w18 = w17**(-1)
 *         cdef double w19 = r**5
 *         cdef double w20 = w19**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w21 = chi_1*X_1
 *         cdef double w22 = chi_2*X_2
 */
  __Pyx_TraceLine(1891,0,__PYX_ERR(1, 1891, __pyx_L1_error))
  __pyx_v_w20 = pow(__pyx_v_w19, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1892
 *         cdef double w19 = r**5
 *         cdef double w20 = w19**(-1)
 *         cdef double w21 = chi_1*X_1             # <<<<<<<<<<<<<<
 *         cdef double w22 = chi_2*X_2
 *         cdef double w23 = w21 + w22
 */
  __Pyx_TraceLine(1892,0,__PYX_ERR(1, 1892, __pyx_L1_error))
  __pyx_v_w21 = (__pyx_v_chi_1 * __pyx_v_X_1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1893
 *         cdef double w20 = w19**(-1)
 *         cdef double w21 = chi_1*X_1
 *         cdef double w22 = chi_2*X_2             # <<<<<<<<<<<<<<
 *         cdef double w23 = w21 + w22
 *         cdef double w24 = w21 - w22
 */
  __Pyx_TraceLine(1893,0,__PYX_ERR(1, 1893, __pyx_L1_error))
  __pyx_v_w22 = (__pyx_v_chi_2 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1894
 *         cdef double w21 = chi_1*X_1
 *         cdef double w22 = chi_2*X_2
 *         cdef double w23 = w21 + w22             # <<<<<<<<<<<<<<
 *         cdef double w24 = w21 - w22
 *         cdef double w25 = -w23*w24 + w5*w6 - w8
 */
  __Pyx_TraceLine(1894,0,__PYX_ERR(1, 1894, __pyx_L1_error))
  __pyx_v_w23 = (__pyx_v_w21 + __pyx_v_w22);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1895
 *         cdef double w22 = chi_2*X_2
 *         cdef double w23 = w21 + w22
 *         cdef double w24 = w21 - w22             # <<<<<<<<<<<<<<
 *         cdef double w25 = -w23*w24 + w5*w6 - w8
 *         cdef double w26 = delta*w25
 */
  __Pyx_TraceLine(1895,0,__PYX_ERR(1, 1895, __pyx_L1_error))
  __pyx_v_w24 = (__pyx_v_w21 - __pyx_v_w22);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1896
 *         cdef double w23 = w21 + w22
 *         cdef double w24 = w21 - w22
 *         cdef double w25 = -w23*w24 + w5*w6 - w8             # <<<<<<<<<<<<<<
 *         cdef double w26 = delta*w25
 *         cdef double w27 = w26*(68.0*nu - 1.0)
 */
  __Pyx_TraceLine(1896,0,__PYX_ERR(1, 1896, __pyx_L1_error))
  __pyx_v_w25 = ((((-__pyx_v_w23) * __pyx_v_w24) + (__pyx_v_w5 * __pyx_v_w6)) - __pyx_v_w8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1897
 *         cdef double w24 = w21 - w22
 *         cdef double w25 = -w23*w24 + w5*w6 - w8
 *         cdef double w26 = delta*w25             # <<<<<<<<<<<<<<
 *         cdef double w27 = w26*(68.0*nu - 1.0)
 *         cdef double w28 = nu**2
 */
  __Pyx_TraceLine(1897,0,__PYX_ERR(1, 1897, __pyx_L1_error))
  __pyx_v_w26 = (__pyx_v_delta * __pyx_v_w25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1898
 *         cdef double w25 = -w23*w24 + w5*w6 - w8
 *         cdef double w26 = delta*w25
 *         cdef double w27 = w26*(68.0*nu - 1.0)             # <<<<<<<<<<<<<<
 *         cdef double w28 = nu**2
 *         cdef double w29 = 264.0*w28
 */
  __Pyx_TraceLine(1898,0,__PYX_ERR(1, 1898, __pyx_L1_error))
  __pyx_v_w27 = (__pyx_v_w26 * ((68.0 * __pyx_v_nu) - 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1899
 *         cdef double w26 = delta*w25
 *         cdef double w27 = w26*(68.0*nu - 1.0)
 *         cdef double w28 = nu**2             # <<<<<<<<<<<<<<
 *         cdef double w29 = 264.0*w28
 *         cdef double w30 = w23**2
 */
  __Pyx_TraceLine(1899,0,__PYX_ERR(1, 1899, __pyx_L1_error))
  __pyx_v_w28 = pow(__pyx_v_nu, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1900
 *         cdef double w27 = w26*(68.0*nu - 1.0)
 *         cdef double w28 = nu**2
 *         cdef double w29 = 264.0*w28             # <<<<<<<<<<<<<<
 *         cdef double w30 = w23**2
 *         cdef double w31 = w10 - w30
 */
  __Pyx_TraceLine(1900,0,__PYX_ERR(1, 1900, __pyx_L1_error))
  __pyx_v_w29 = (264.0 * __pyx_v_w28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1901
 *         cdef double w28 = nu**2
 *         cdef double w29 = 264.0*w28
 *         cdef double w30 = w23**2             # <<<<<<<<<<<<<<
 *         cdef double w31 = w10 - w30
 *         cdef double w32 = 0.00260416666666667*w31
 */
  __Pyx_TraceLine(1901,0,__PYX_ERR(1, 1901, __pyx_L1_error))
  __pyx_v_w30 = pow(__pyx_v_w23, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1902
 *         cdef double w29 = 264.0*w28
 *         cdef double w30 = w23**2
 *         cdef double w31 = w10 - w30             # <<<<<<<<<<<<<<
 *         cdef double w32 = 0.00260416666666667*w31
 *         cdef double w33 = -w4 + w9
 */
  __Pyx_TraceLine(1902,0,__PYX_ERR(1, 1902, __pyx_L1_error))
  __pyx_v_w31 = (__pyx_v_w10 - __pyx_v_w30);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1903
 *         cdef double w30 = w23**2
 *         cdef double w31 = w10 - w30
 *         cdef double w32 = 0.00260416666666667*w31             # <<<<<<<<<<<<<<
 *         cdef double w33 = -w4 + w9
 *         cdef double w34 = -w24**2 + w33
 */
  __Pyx_TraceLine(1903,0,__PYX_ERR(1, 1903, __pyx_L1_error))
  __pyx_v_w32 = (0.00260416666666667 * __pyx_v_w31);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1904
 *         cdef double w31 = w10 - w30
 *         cdef double w32 = 0.00260416666666667*w31
 *         cdef double w33 = -w4 + w9             # <<<<<<<<<<<<<<
 *         cdef double w34 = -w24**2 + w33
 *         cdef double w35 = 0.0078125*w34
 */
  __Pyx_TraceLine(1904,0,__PYX_ERR(1, 1904, __pyx_L1_error))
  __pyx_v_w33 = ((-__pyx_v_w4) + __pyx_v_w9);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1905
 *         cdef double w32 = 0.00260416666666667*w31
 *         cdef double w33 = -w4 + w9
 *         cdef double w34 = -w24**2 + w33             # <<<<<<<<<<<<<<
 *         cdef double w35 = 0.0078125*w34
 *         cdef double w36 = w7 - w8
 */
  __Pyx_TraceLine(1905,0,__PYX_ERR(1, 1905, __pyx_L1_error))
  __pyx_v_w34 = ((-pow(__pyx_v_w24, 2.0)) + __pyx_v_w33);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1906
 *         cdef double w33 = -w4 + w9
 *         cdef double w34 = -w24**2 + w33
 *         cdef double w35 = 0.0078125*w34             # <<<<<<<<<<<<<<
 *         cdef double w36 = w7 - w8
 *         cdef double w37 = delta*w36
 */
  __Pyx_TraceLine(1906,0,__PYX_ERR(1, 1906, __pyx_L1_error))
  __pyx_v_w35 = (0.0078125 * __pyx_v_w34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1907
 *         cdef double w34 = -w24**2 + w33
 *         cdef double w35 = 0.0078125*w34
 *         cdef double w36 = w7 - w8             # <<<<<<<<<<<<<<
 *         cdef double w37 = delta*w36
 *         cdef double w38 = w37*(2.0*nu - 3.0)
 */
  __Pyx_TraceLine(1907,0,__PYX_ERR(1, 1907, __pyx_L1_error))
  __pyx_v_w36 = (__pyx_v_w7 - __pyx_v_w8);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1908
 *         cdef double w35 = 0.0078125*w34
 *         cdef double w36 = w7 - w8
 *         cdef double w37 = delta*w36             # <<<<<<<<<<<<<<
 *         cdef double w38 = w37*(2.0*nu - 3.0)
 *         cdef double w39 = 7.0*nu
 */
  __Pyx_TraceLine(1908,0,__PYX_ERR(1, 1908, __pyx_L1_error))
  __pyx_v_w37 = (__pyx_v_delta * __pyx_v_w36);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1909
 *         cdef double w36 = w7 - w8
 *         cdef double w37 = delta*w36
 *         cdef double w38 = w37*(2.0*nu - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double w39 = 7.0*nu
 *         cdef double w40 = 0.390625*w4 + 0.390625*w7 + 0.390625*w8
 */
  __Pyx_TraceLine(1909,0,__PYX_ERR(1, 1909, __pyx_L1_error))
  __pyx_v_w38 = (__pyx_v_w37 * ((2.0 * __pyx_v_nu) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1910
 *         cdef double w37 = delta*w36
 *         cdef double w38 = w37*(2.0*nu - 3.0)
 *         cdef double w39 = 7.0*nu             # <<<<<<<<<<<<<<
 *         cdef double w40 = 0.390625*w4 + 0.390625*w7 + 0.390625*w8
 *         cdef double w41 = 3.0*w4
 */
  __Pyx_TraceLine(1910,0,__PYX_ERR(1, 1910, __pyx_L1_error))
  __pyx_v_w39 = (7.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1911
 *         cdef double w38 = w37*(2.0*nu - 3.0)
 *         cdef double w39 = 7.0*nu
 *         cdef double w40 = 0.390625*w4 + 0.390625*w7 + 0.390625*w8             # <<<<<<<<<<<<<<
 *         cdef double w41 = 3.0*w4
 *         cdef double w42 = 3.0*w7 + 3.0*w8
 */
  __Pyx_TraceLine(1911,0,__PYX_ERR(1, 1911, __pyx_L1_error))
  __pyx_v_w40 = (((0.390625 * __pyx_v_w4) + (0.390625 * __pyx_v_w7)) + (0.390625 * __pyx_v_w8));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1912
 *         cdef double w39 = 7.0*nu
 *         cdef double w40 = 0.390625*w4 + 0.390625*w7 + 0.390625*w8
 *         cdef double w41 = 3.0*w4             # <<<<<<<<<<<<<<
 *         cdef double w42 = 3.0*w7 + 3.0*w8
 *         cdef double w43 = -w41 + w42
 */
  __Pyx_TraceLine(1912,0,__PYX_ERR(1, 1912, __pyx_L1_error))
  __pyx_v_w41 = (3.0 * __pyx_v_w4);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1913
 *         cdef double w40 = 0.390625*w4 + 0.390625*w7 + 0.390625*w8
 *         cdef double w41 = 3.0*w4
 *         cdef double w42 = 3.0*w7 + 3.0*w8             # <<<<<<<<<<<<<<
 *         cdef double w43 = -w41 + w42
 *         cdef double w44 = w31*(w39 - 31.0)
 */
  __Pyx_TraceLine(1913,0,__PYX_ERR(1, 1913, __pyx_L1_error))
  __pyx_v_w42 = ((3.0 * __pyx_v_w7) + (3.0 * __pyx_v_w8));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1914
 *         cdef double w41 = 3.0*w4
 *         cdef double w42 = 3.0*w7 + 3.0*w8
 *         cdef double w43 = -w41 + w42             # <<<<<<<<<<<<<<
 *         cdef double w44 = w31*(w39 - 31.0)
 *         cdef double w45 = 4.0*nu
 */
  __Pyx_TraceLine(1914,0,__PYX_ERR(1, 1914, __pyx_L1_error))
  __pyx_v_w43 = ((-__pyx_v_w41) + __pyx_v_w42);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1915
 *         cdef double w42 = 3.0*w7 + 3.0*w8
 *         cdef double w43 = -w41 + w42
 *         cdef double w44 = w31*(w39 - 31.0)             # <<<<<<<<<<<<<<
 *         cdef double w45 = 4.0*nu
 *         cdef double w46 = w34*(-w45 - 3.0)
 */
  __Pyx_TraceLine(1915,0,__PYX_ERR(1, 1915, __pyx_L1_error))
  __pyx_v_w44 = (__pyx_v_w31 * (__pyx_v_w39 - 31.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1916
 *         cdef double w43 = -w41 + w42
 *         cdef double w44 = w31*(w39 - 31.0)
 *         cdef double w45 = 4.0*nu             # <<<<<<<<<<<<<<
 *         cdef double w46 = w34*(-w45 - 3.0)
 *         cdef double w47 = w33*(w45 + 1.0)
 */
  __Pyx_TraceLine(1916,0,__PYX_ERR(1, 1916, __pyx_L1_error))
  __pyx_v_w45 = (4.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1917
 *         cdef double w44 = w31*(w39 - 31.0)
 *         cdef double w45 = 4.0*nu
 *         cdef double w46 = w34*(-w45 - 3.0)             # <<<<<<<<<<<<<<
 *         cdef double w47 = w33*(w45 + 1.0)
 *         cdef double w48 = w18*w31
 */
  __Pyx_TraceLine(1917,0,__PYX_ERR(1, 1917, __pyx_L1_error))
  __pyx_v_w46 = (__pyx_v_w34 * ((-__pyx_v_w45) - 3.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1918
 *         cdef double w45 = 4.0*nu
 *         cdef double w46 = w34*(-w45 - 3.0)
 *         cdef double w47 = w33*(w45 + 1.0)             # <<<<<<<<<<<<<<
 *         cdef double w48 = w18*w31
 *         cdef double w49 = log(r)
 */
  __Pyx_TraceLine(1918,0,__PYX_ERR(1, 1918, __pyx_L1_error))
  __pyx_v_w47 = (__pyx_v_w33 * (__pyx_v_w45 + 1.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1919
 *         cdef double w46 = w34*(-w45 - 3.0)
 *         cdef double w47 = w33*(w45 + 1.0)
 *         cdef double w48 = w18*w31             # <<<<<<<<<<<<<<
 *         cdef double w49 = log(r)
 *         cdef double w50 = 756.0*nu
 */
  __Pyx_TraceLine(1919,0,__PYX_ERR(1, 1919, __pyx_L1_error))
  __pyx_v_w48 = (__pyx_v_w18 * __pyx_v_w31);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1920
 *         cdef double w47 = w33*(w45 + 1.0)
 *         cdef double w48 = w18*w31
 *         cdef double w49 = log(r)             # <<<<<<<<<<<<<<
 *         cdef double w50 = 756.0*nu
 *         cdef double w51 = 336.0*r + w50 + 407.0
 */
  __Pyx_TraceLine(1920,0,__PYX_ERR(1, 1920, __pyx_L1_error))
  __pyx_v_w49 = log(__pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1921
 *         cdef double w48 = w18*w31
 *         cdef double w49 = log(r)
 *         cdef double w50 = 756.0*nu             # <<<<<<<<<<<<<<
 *         cdef double w51 = 336.0*r + w50 + 407.0
 *         cdef double w52 = 2048.0*nu*w49*w51 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*w28 - 245760.0) - 5416406.59541186*w28 - 3440640.0
 */
  __Pyx_TraceLine(1921,0,__PYX_ERR(1, 1921, __pyx_L1_error))
  __pyx_v_w50 = (756.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1922
 *         cdef double w49 = log(r)
 *         cdef double w50 = 756.0*nu
 *         cdef double w51 = 336.0*r + w50 + 407.0             # <<<<<<<<<<<<<<
 *         cdef double w52 = 2048.0*nu*w49*w51 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*w28 - 245760.0) - 5416406.59541186*w28 - 3440640.0
 *         cdef double w53 = nu**4
 */
  __Pyx_TraceLine(1922,0,__PYX_ERR(1, 1922, __pyx_L1_error))
  __pyx_v_w51 = (((336.0 * __pyx_v_r) + __pyx_v_w50) + 407.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1923
 *         cdef double w50 = 756.0*nu
 *         cdef double w51 = 336.0*r + w50 + 407.0
 *         cdef double w52 = 2048.0*nu*w49*w51 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*w28 - 245760.0) - 5416406.59541186*w28 - 3440640.0             # <<<<<<<<<<<<<<
 *         cdef double w53 = nu**4
 *         cdef double w54 = w49**2
 */
  __Pyx_TraceLine(1923,0,__PYX_ERR(1, 1923, __pyx_L1_error))
  __pyx_v_w52 = (((((((2048.0 * __pyx_v_nu) * __pyx_v_w49) * __pyx_v_w51) + ((28.0 * __pyx_v_nu) * ((1920.0 * __pyx_v_a6) + 733955.307463037))) - ((7.0 * __pyx_v_r) * (((938918.400156317 * __pyx_v_nu) - (185763.092693281 * __pyx_v_w28)) - 245760.0))) - (5416406.59541186 * __pyx_v_w28)) - 3440640.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1924
 *         cdef double w51 = 336.0*r + w50 + 407.0
 *         cdef double w52 = 2048.0*nu*w49*w51 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*w28 - 245760.0) - 5416406.59541186*w28 - 3440640.0
 *         cdef double w53 = nu**4             # <<<<<<<<<<<<<<
 *         cdef double w54 = w49**2
 *         cdef double w55 = w28*w54
 */
  __Pyx_TraceLine(1924,0,__PYX_ERR(1, 1924, __pyx_L1_error))
  __pyx_v_w53 = pow(__pyx_v_nu, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1925
 *         cdef double w52 = 2048.0*nu*w49*w51 + 28.0*nu*(1920.0*a6 + 733955.307463037) - 7.0*r*(938918.400156317*nu - 185763.092693281*w28 - 245760.0) - 5416406.59541186*w28 - 3440640.0
 *         cdef double w53 = nu**4
 *         cdef double w54 = w49**2             # <<<<<<<<<<<<<<
 *         cdef double w55 = w28*w54
 *         cdef double w56 = nu**3
 */
  __Pyx_TraceLine(1925,0,__PYX_ERR(1, 1925, __pyx_L1_error))
  __pyx_v_w54 = pow(__pyx_v_w49, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1926
 *         cdef double w53 = nu**4
 *         cdef double w54 = w49**2
 *         cdef double w55 = w28*w54             # <<<<<<<<<<<<<<
 *         cdef double w56 = nu**3
 *         cdef double w57 = w56*(-163683964.822551*r - 17833256.898555*w1 - 1188987459.03162)
 */
  __Pyx_TraceLine(1926,0,__PYX_ERR(1, 1926, __pyx_L1_error))
  __pyx_v_w55 = (__pyx_v_w28 * __pyx_v_w54);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1927
 *         cdef double w54 = w49**2
 *         cdef double w55 = w28*w54
 *         cdef double w56 = nu**3             # <<<<<<<<<<<<<<
 *         cdef double w57 = w56*(-163683964.822551*r - 17833256.898555*w1 - 1188987459.03162)
 *         cdef double w58 = w28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r + 2064783811.32587*w1 - 3089250703.76879*w13 - 6178501407.53758*w17 + 1426660551.8844*w19 + 276057889687.011)
 */
  __Pyx_TraceLine(1927,0,__PYX_ERR(1, 1927, __pyx_L1_error))
  __pyx_v_w56 = pow(__pyx_v_nu, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1928
 *         cdef double w55 = w28*w54
 *         cdef double w56 = nu**3
 *         cdef double w57 = w56*(-163683964.822551*r - 17833256.898555*w1 - 1188987459.03162)             # <<<<<<<<<<<<<<
 *         cdef double w58 = w28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r + 2064783811.32587*w1 - 3089250703.76879*w13 - 6178501407.53758*w17 + 1426660551.8844*w19 + 276057889687.011)
 *         cdef double w59 = 588.0*nu + 1079.0
 */
  __Pyx_TraceLine(1928,0,__PYX_ERR(1, 1928, __pyx_L1_error))
  __pyx_v_w57 = (__pyx_v_w56 * (((-163683964.822551 * __pyx_v_r) - (17833256.898555 * __pyx_v_w1)) - 1188987459.03162));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1929
 *         cdef double w56 = nu**3
 *         cdef double w57 = w56*(-163683964.822551*r - 17833256.898555*w1 - 1188987459.03162)
 *         cdef double w58 = w28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r + 2064783811.32587*w1 - 3089250703.76879*w13 - 6178501407.53758*w17 + 1426660551.8844*w19 + 276057889687.011)             # <<<<<<<<<<<<<<
 *         cdef double w59 = 588.0*nu + 1079.0
 *         cdef double w60 = w50 + 1079.0
 */
  __Pyx_TraceLine(1929,0,__PYX_ERR(1, 1929, __pyx_L1_error))
  __pyx_v_w58 = (__pyx_v_w28 * (((((((((-39321600.0 * __pyx_v_a6) * ((3.0 * __pyx_v_r) + 59.0)) + (745857848.115604 * __pyx_v_a6)) + (122635399361.987 * __pyx_v_r)) + (2064783811.32587 * __pyx_v_w1)) - (3089250703.76879 * __pyx_v_w13)) - (6178501407.53758 * __pyx_v_w17)) + (1426660551.8844 * __pyx_v_w19)) + 276057889687.011));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1930
 *         cdef double w57 = w56*(-163683964.822551*r - 17833256.898555*w1 - 1188987459.03162)
 *         cdef double w58 = w28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r + 2064783811.32587*w1 - 3089250703.76879*w13 - 6178501407.53758*w17 + 1426660551.8844*w19 + 276057889687.011)
 *         cdef double w59 = 588.0*nu + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double w60 = w50 + 1079.0
 *         cdef double w61 = w17*w60
 */
  __Pyx_TraceLine(1930,0,__PYX_ERR(1, 1930, __pyx_L1_error))
  __pyx_v_w59 = ((588.0 * __pyx_v_nu) + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1931
 *         cdef double w58 = w28*(-39321600.0*a6*(3.0*r + 59.0) + 745857848.115604*a6 + 122635399361.987*r + 2064783811.32587*w1 - 3089250703.76879*w13 - 6178501407.53758*w17 + 1426660551.8844*w19 + 276057889687.011)
 *         cdef double w59 = 588.0*nu + 1079.0
 *         cdef double w60 = w50 + 1079.0             # <<<<<<<<<<<<<<
 *         cdef double w61 = w17*w60
 *         cdef double w62 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*w28 + 17264.0) + 1920.0*w1*w59 + 480.0*w13*w60 + 161280.0*w19 - 1882456.23663972*w28 + 960.0*w61 + 13447680.0
 */
  __Pyx_TraceLine(1931,0,__PYX_ERR(1, 1931, __pyx_L1_error))
  __pyx_v_w60 = (__pyx_v_w50 + 1079.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1932
 *         cdef double w59 = 588.0*nu + 1079.0
 *         cdef double w60 = w50 + 1079.0
 *         cdef double w61 = w17*w60             # <<<<<<<<<<<<<<
 *         cdef double w62 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*w28 + 17264.0) + 1920.0*w1*w59 + 480.0*w13*w60 + 161280.0*w19 - 1882456.23663972*w28 + 960.0*w61 + 13447680.0
 *         cdef double w63 = nu*w49
 */
  __Pyx_TraceLine(1932,0,__PYX_ERR(1, 1932, __pyx_L1_error))
  __pyx_v_w61 = (__pyx_v_w17 * __pyx_v_w60);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1933
 *         cdef double w60 = w50 + 1079.0
 *         cdef double w61 = w17*w60
 *         cdef double w62 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*w28 + 17264.0) + 1920.0*w1*w59 + 480.0*w13*w60 + 161280.0*w19 - 1882456.23663972*w28 + 960.0*w61 + 13447680.0             # <<<<<<<<<<<<<<
 *         cdef double w63 = nu*w49
 *         cdef double w64 = w62*w63
 */
  __Pyx_TraceLine(1933,0,__PYX_ERR(1, 1933, __pyx_L1_error))
  __pyx_v_w62 = ((((((((-38842241.4769507 * __pyx_v_nu) + ((240.0 * __pyx_v_r) * (((-7466.27061066206 * __pyx_v_nu) - (3024.0 * __pyx_v_w28)) + 17264.0))) + ((1920.0 * __pyx_v_w1) * __pyx_v_w59)) + ((480.0 * __pyx_v_w13) * __pyx_v_w60)) + (161280.0 * __pyx_v_w19)) - (1882456.23663972 * __pyx_v_w28)) + (960.0 * __pyx_v_w61)) + 13447680.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1934
 *         cdef double w61 = w17*w60
 *         cdef double w62 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*w28 + 17264.0) + 1920.0*w1*w59 + 480.0*w13*w60 + 161280.0*w19 - 1882456.23663972*w28 + 960.0*w61 + 13447680.0
 *         cdef double w63 = nu*w49             # <<<<<<<<<<<<<<
 *         cdef double w64 = w62*w63
 *         cdef double w65 = 8.0*r
 */
  __Pyx_TraceLine(1934,0,__PYX_ERR(1, 1934, __pyx_L1_error))
  __pyx_v_w63 = (__pyx_v_nu * __pyx_v_w49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1935
 *         cdef double w62 = -38842241.4769507*nu + 240.0*r*(-7466.27061066206*nu - 3024.0*w28 + 17264.0) + 1920.0*w1*w59 + 480.0*w13*w60 + 161280.0*w19 - 1882456.23663972*w28 + 960.0*w61 + 13447680.0
 *         cdef double w63 = nu*w49
 *         cdef double w64 = w62*w63             # <<<<<<<<<<<<<<
 *         cdef double w65 = 8.0*r
 *         cdef double w66 = 2.0*w17
 */
  __Pyx_TraceLine(1935,0,__PYX_ERR(1, 1935, __pyx_L1_error))
  __pyx_v_w64 = (__pyx_v_w62 * __pyx_v_w63);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1936
 *         cdef double w63 = nu*w49
 *         cdef double w64 = w62*w63
 *         cdef double w65 = 8.0*r             # <<<<<<<<<<<<<<
 *         cdef double w66 = 2.0*w17
 *         cdef double w67 = 4.0*w1 + w65 + w66 + 16.0
 */
  __Pyx_TraceLine(1936,0,__PYX_ERR(1, 1936, __pyx_L1_error))
  __pyx_v_w65 = (8.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1937
 *         cdef double w64 = w62*w63
 *         cdef double w65 = 8.0*r
 *         cdef double w66 = 2.0*w17             # <<<<<<<<<<<<<<
 *         cdef double w67 = 4.0*w1 + w65 + w66 + 16.0
 *         cdef double w68 = 7680.0*a6
 */
  __Pyx_TraceLine(1937,0,__PYX_ERR(1, 1937, __pyx_L1_error))
  __pyx_v_w66 = (2.0 * __pyx_v_w17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1938
 *         cdef double w65 = 8.0*r
 *         cdef double w66 = 2.0*w17
 *         cdef double w67 = 4.0*w1 + w65 + w66 + 16.0             # <<<<<<<<<<<<<<
 *         cdef double w68 = 7680.0*a6
 *         cdef double w69 = 128.0*r
 */
  __Pyx_TraceLine(1938,0,__PYX_ERR(1, 1938, __pyx_L1_error))
  __pyx_v_w67 = ((((4.0 * __pyx_v_w1) + __pyx_v_w65) + __pyx_v_w66) + 16.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1939
 *         cdef double w66 = 2.0*w17
 *         cdef double w67 = 4.0*w1 + w65 + w66 + 16.0
 *         cdef double w68 = 7680.0*a6             # <<<<<<<<<<<<<<
 *         cdef double w69 = 128.0*r
 *         cdef double w70 = 7704.0*r
 */
  __Pyx_TraceLine(1939,0,__PYX_ERR(1, 1939, __pyx_L1_error))
  __pyx_v_w68 = (7680.0 * __pyx_v_a6);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1940
 *         cdef double w67 = 4.0*w1 + w65 + w66 + 16.0
 *         cdef double w68 = 7680.0*a6
 *         cdef double w69 = 128.0*r             # <<<<<<<<<<<<<<
 *         cdef double w70 = 7704.0*r
 *         cdef double w71 = 148.04406601634*r
 */
  __Pyx_TraceLine(1940,0,__PYX_ERR(1, 1940, __pyx_L1_error))
  __pyx_v_w69 = (128.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1941
 *         cdef double w68 = 7680.0*a6
 *         cdef double w69 = 128.0*r
 *         cdef double w70 = 7704.0*r             # <<<<<<<<<<<<<<
 *         cdef double w71 = 148.04406601634*r
 *         cdef double w72 = 113485.217444961*r
 */
  __Pyx_TraceLine(1941,0,__PYX_ERR(1, 1941, __pyx_L1_error))
  __pyx_v_w70 = (7704.0 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1942
 *         cdef double w69 = 128.0*r
 *         cdef double w70 = 7704.0*r
 *         cdef double w71 = 148.04406601634*r             # <<<<<<<<<<<<<<
 *         cdef double w72 = 113485.217444961*r
 *         cdef double w73 = nu*(w68*(w13 + w67) + w69*(13218.7851094412*r + 8529.39255472061*w1 - 6852.34813868015*w13 + 4264.6962773603*w17 - 33722.4297811176) + w71*(3852.0*w1 + 349.0*w13 + 1926.0*w17 + w70 + 36400.0) + w72*(-w13 + w67))
 */
  __Pyx_TraceLine(1942,0,__PYX_ERR(1, 1942, __pyx_L1_error))
  __pyx_v_w71 = (148.04406601634 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1943
 *         cdef double w70 = 7704.0*r
 *         cdef double w71 = 148.04406601634*r
 *         cdef double w72 = 113485.217444961*r             # <<<<<<<<<<<<<<
 *         cdef double w73 = nu*(w68*(w13 + w67) + w69*(13218.7851094412*r + 8529.39255472061*w1 - 6852.34813868015*w13 + 4264.6962773603*w17 - 33722.4297811176) + w71*(3852.0*w1 + 349.0*w13 + 1926.0*w17 + w70 + 36400.0) + w72*(-w13 + w67))
 *         cdef double w74 = (13212057600.0*w19 + 241555486248.807*w53 + 67645734912.0*w55 + 1120.0*w57 + 7.0*w58 + 32768.0*w64 + 53760.0*w73)**(-1)
 */
  __Pyx_TraceLine(1943,0,__PYX_ERR(1, 1943, __pyx_L1_error))
  __pyx_v_w72 = (113485.217444961 * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1944
 *         cdef double w71 = 148.04406601634*r
 *         cdef double w72 = 113485.217444961*r
 *         cdef double w73 = nu*(w68*(w13 + w67) + w69*(13218.7851094412*r + 8529.39255472061*w1 - 6852.34813868015*w13 + 4264.6962773603*w17 - 33722.4297811176) + w71*(3852.0*w1 + 349.0*w13 + 1926.0*w17 + w70 + 36400.0) + w72*(-w13 + w67))             # <<<<<<<<<<<<<<
 *         cdef double w74 = (13212057600.0*w19 + 241555486248.807*w53 + 67645734912.0*w55 + 1120.0*w57 + 7.0*w58 + 32768.0*w64 + 53760.0*w73)**(-1)
 *         cdef double w75 = w13*w74
 */
  __Pyx_TraceLine(1944,0,__PYX_ERR(1, 1944, __pyx_L1_error))
  __pyx_v_w73 = (__pyx_v_nu * ((((__pyx_v_w68 * (__pyx_v_w13 + __pyx_v_w67)) + (__pyx_v_w69 * (((((13218.7851094412 * __pyx_v_r) + (8529.39255472061 * __pyx_v_w1)) - (6852.34813868015 * __pyx_v_w13)) + (4264.6962773603 * __pyx_v_w17)) - 33722.4297811176))) + (__pyx_v_w71 * (((((3852.0 * __pyx_v_w1) + (349.0 * __pyx_v_w13)) + (1926.0 * __pyx_v_w17)) + __pyx_v_w70) + 36400.0))) + (__pyx_v_w72 * ((-__pyx_v_w13) + __pyx_v_w67))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1945
 *         cdef double w72 = 113485.217444961*r
 *         cdef double w73 = nu*(w68*(w13 + w67) + w69*(13218.7851094412*r + 8529.39255472061*w1 - 6852.34813868015*w13 + 4264.6962773603*w17 - 33722.4297811176) + w71*(3852.0*w1 + 349.0*w13 + 1926.0*w17 + w70 + 36400.0) + w72*(-w13 + w67))
 *         cdef double w74 = (13212057600.0*w19 + 241555486248.807*w53 + 67645734912.0*w55 + 1120.0*w57 + 7.0*w58 + 32768.0*w64 + 53760.0*w73)**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w75 = w13*w74
 *         cdef double w76 = w52*w75
 */
  __Pyx_TraceLine(1945,0,__PYX_ERR(1, 1945, __pyx_L1_error))
  __pyx_v_w74 = pow((((((((13212057600.0 * __pyx_v_w19) + (241555486248.807 * __pyx_v_w53)) + (67645734912.0 * __pyx_v_w55)) + (1120.0 * __pyx_v_w57)) + (7.0 * __pyx_v_w58)) + (32768.0 * __pyx_v_w64)) + (53760.0 * __pyx_v_w73)), -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1946
 *         cdef double w73 = nu*(w68*(w13 + w67) + w69*(13218.7851094412*r + 8529.39255472061*w1 - 6852.34813868015*w13 + 4264.6962773603*w17 - 33722.4297811176) + w71*(3852.0*w1 + 349.0*w13 + 1926.0*w17 + w70 + 36400.0) + w72*(-w13 + w67))
 *         cdef double w74 = (13212057600.0*w19 + 241555486248.807*w53 + 67645734912.0*w55 + 1120.0*w57 + 7.0*w58 + 32768.0*w64 + 53760.0*w73)**(-1)
 *         cdef double w75 = w13*w74             # <<<<<<<<<<<<<<
 *         cdef double w76 = w52*w75
 *         cdef double w77 = w11 + 7680.0*w76
 */
  __Pyx_TraceLine(1946,0,__PYX_ERR(1, 1946, __pyx_L1_error))
  __pyx_v_w75 = (__pyx_v_w13 * __pyx_v_w74);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1947
 *         cdef double w74 = (13212057600.0*w19 + 241555486248.807*w53 + 67645734912.0*w55 + 1120.0*w57 + 7.0*w58 + 32768.0*w64 + 53760.0*w73)**(-1)
 *         cdef double w75 = w13*w74
 *         cdef double w76 = w52*w75             # <<<<<<<<<<<<<<
 *         cdef double w77 = w11 + 7680.0*w76
 *         cdef double w78 = w14*(2.0625*w26 + 0.125*w44 + 0.0625*w46) + w14*(-1.25*w37 + 1.125*w4 + 0.125*w47 + 1.125*w7 + 1.125*w8) + w20*(0.125*w27 + w32*(2870.0*nu - w29 + 561.0) + w35*(1166.0*nu - 328.0*w28 - 171.0)) + w20*(-1.21875*w38 - w40*(w39 + 9.0) + 0.015625*w43*(-27.0*nu + 28.0*w28 - 3.0)) + w48 + w77
 */
  __Pyx_TraceLine(1947,0,__PYX_ERR(1, 1947, __pyx_L1_error))
  __pyx_v_w76 = (__pyx_v_w52 * __pyx_v_w75);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1948
 *         cdef double w75 = w13*w74
 *         cdef double w76 = w52*w75
 *         cdef double w77 = w11 + 7680.0*w76             # <<<<<<<<<<<<<<
 *         cdef double w78 = w14*(2.0625*w26 + 0.125*w44 + 0.0625*w46) + w14*(-1.25*w37 + 1.125*w4 + 0.125*w47 + 1.125*w7 + 1.125*w8) + w20*(0.125*w27 + w32*(2870.0*nu - w29 + 561.0) + w35*(1166.0*nu - 328.0*w28 - 171.0)) + w20*(-1.21875*w38 - w40*(w39 + 9.0) + 0.015625*w43*(-27.0*nu + 28.0*w28 - 3.0)) + w48 + w77
 *         cdef double w79 = w78*(w10*w18*(-4.0*w0 - 2.0) + w14*w16)/w12**2
 */
  __Pyx_TraceLine(1948,0,__PYX_ERR(1, 1948, __pyx_L1_error))
  __pyx_v_w77 = (__pyx_v_w11 + (7680.0 * __pyx_v_w76));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1949
 *         cdef double w76 = w52*w75
 *         cdef double w77 = w11 + 7680.0*w76
 *         cdef double w78 = w14*(2.0625*w26 + 0.125*w44 + 0.0625*w46) + w14*(-1.25*w37 + 1.125*w4 + 0.125*w47 + 1.125*w7 + 1.125*w8) + w20*(0.125*w27 + w32*(2870.0*nu - w29 + 561.0) + w35*(1166.0*nu - 328.0*w28 - 171.0)) + w20*(-1.21875*w38 - w40*(w39 + 9.0) + 0.015625*w43*(-27.0*nu + 28.0*w28 - 3.0)) + w48 + w77             # <<<<<<<<<<<<<<
 *         cdef double w79 = w78*(w10*w18*(-4.0*w0 - 2.0) + w14*w16)/w12**2
 *         cdef double w80 = w12**(-1)
 */
  __Pyx_TraceLine(1949,0,__PYX_ERR(1, 1949, __pyx_L1_error))
  __pyx_v_w78 = ((((((__pyx_v_w14 * (((2.0625 * __pyx_v_w26) + (0.125 * __pyx_v_w44)) + (0.0625 * __pyx_v_w46))) + (__pyx_v_w14 * (((((-1.25 * __pyx_v_w37) + (1.125 * __pyx_v_w4)) + (0.125 * __pyx_v_w47)) + (1.125 * __pyx_v_w7)) + (1.125 * __pyx_v_w8)))) + (__pyx_v_w20 * (((0.125 * __pyx_v_w27) + (__pyx_v_w32 * (((2870.0 * __pyx_v_nu) - __pyx_v_w29) + 561.0))) + (__pyx_v_w35 * (((1166.0 * __pyx_v_nu) - (328.0 * __pyx_v_w28)) - 171.0))))) + (__pyx_v_w20 * (((-1.21875 * __pyx_v_w38) - (__pyx_v_w40 * (__pyx_v_w39 + 9.0))) + ((0.015625 * __pyx_v_w43) * (((-27.0 * __pyx_v_nu) + (28.0 * __pyx_v_w28)) - 3.0))))) + __pyx_v_w48) + __pyx_v_w77);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1950
 *         cdef double w77 = w11 + 7680.0*w76
 *         cdef double w78 = w14*(2.0625*w26 + 0.125*w44 + 0.0625*w46) + w14*(-1.25*w37 + 1.125*w4 + 0.125*w47 + 1.125*w7 + 1.125*w8) + w20*(0.125*w27 + w32*(2870.0*nu - w29 + 561.0) + w35*(1166.0*nu - 328.0*w28 - 171.0)) + w20*(-1.21875*w38 - w40*(w39 + 9.0) + 0.015625*w43*(-27.0*nu + 28.0*w28 - 3.0)) + w48 + w77
 *         cdef double w79 = w78*(w10*w18*(-4.0*w0 - 2.0) + w14*w16)/w12**2             # <<<<<<<<<<<<<<
 *         cdef double w80 = w12**(-1)
 *         cdef double w81 = 4.5*X_2
 */
  __Pyx_TraceLine(1950,0,__PYX_ERR(1, 1950, __pyx_L1_error))
  __pyx_v_w79 = ((__pyx_v_w78 * (((__pyx_v_w10 * __pyx_v_w18) * ((-4.0 * __pyx_v_w0) - 2.0)) + (__pyx_v_w14 * __pyx_v_w16))) / pow(__pyx_v_w12, 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1951
 *         cdef double w78 = w14*(2.0625*w26 + 0.125*w44 + 0.0625*w46) + w14*(-1.25*w37 + 1.125*w4 + 0.125*w47 + 1.125*w7 + 1.125*w8) + w20*(0.125*w27 + w32*(2870.0*nu - w29 + 561.0) + w35*(1166.0*nu - 328.0*w28 - 171.0)) + w20*(-1.21875*w38 - w40*(w39 + 9.0) + 0.015625*w43*(-27.0*nu + 28.0*w28 - 3.0)) + w48 + w77
 *         cdef double w79 = w78*(w10*w18*(-4.0*w0 - 2.0) + w14*w16)/w12**2
 *         cdef double w80 = w12**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w81 = 4.5*X_2
 *         cdef double w82 = r**(-6)
 */
  __Pyx_TraceLine(1951,0,__PYX_ERR(1, 1951, __pyx_L1_error))
  __pyx_v_w80 = pow(__pyx_v_w12, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1952
 *         cdef double w79 = w78*(w10*w18*(-4.0*w0 - 2.0) + w14*w16)/w12**2
 *         cdef double w80 = w12**(-1)
 *         cdef double w81 = 4.5*X_2             # <<<<<<<<<<<<<<
 *         cdef double w82 = r**(-6)
 *         cdef double w83 = nu*w0
 */
  __Pyx_TraceLine(1952,0,__PYX_ERR(1, 1952, __pyx_L1_error))
  __pyx_v_w81 = (4.5 * __pyx_v_X_2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1953
 *         cdef double w80 = w12**(-1)
 *         cdef double w81 = 4.5*X_2
 *         cdef double w82 = r**(-6)             # <<<<<<<<<<<<<<
 *         cdef double w83 = nu*w0
 *         cdef double w84 = -6572428.80109422*nu + 1300341.64885296*w28 + 2048.0*w51*w83 + 688128.0*w63 + 1720320.0
 */
  __Pyx_TraceLine(1953,0,__PYX_ERR(1, 1953, __pyx_L1_error))
  __pyx_v_w82 = pow(__pyx_v_r, -6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1954
 *         cdef double w81 = 4.5*X_2
 *         cdef double w82 = r**(-6)
 *         cdef double w83 = nu*w0             # <<<<<<<<<<<<<<
 *         cdef double w84 = -6572428.80109422*nu + 1300341.64885296*w28 + 2048.0*w51*w83 + 688128.0*w63 + 1720320.0
 *         cdef double w85 = w0*w28
 */
  __Pyx_TraceLine(1954,0,__PYX_ERR(1, 1954, __pyx_L1_error))
  __pyx_v_w83 = (__pyx_v_nu * __pyx_v_w0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1955
 *         cdef double w82 = r**(-6)
 *         cdef double w83 = nu*w0
 *         cdef double w84 = -6572428.80109422*nu + 1300341.64885296*w28 + 2048.0*w51*w83 + 688128.0*w63 + 1720320.0             # <<<<<<<<<<<<<<
 *         cdef double w85 = w0*w28
 *         cdef double w86 = 4.0*w17
 */
  __Pyx_TraceLine(1955,0,__PYX_ERR(1, 1955, __pyx_L1_error))
  __pyx_v_w84 = (((((-6572428.80109422 * __pyx_v_nu) + (1300341.64885296 * __pyx_v_w28)) + ((2048.0 * __pyx_v_w51) * __pyx_v_w83)) + (688128.0 * __pyx_v_w63)) + 1720320.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1956
 *         cdef double w83 = nu*w0
 *         cdef double w84 = -6572428.80109422*nu + 1300341.64885296*w28 + 2048.0*w51*w83 + 688128.0*w63 + 1720320.0
 *         cdef double w85 = w0*w28             # <<<<<<<<<<<<<<
 *         cdef double w86 = 4.0*w17
 *         cdef double w87 = 6.0*w1 + w65 + 8.0
 */
  __Pyx_TraceLine(1956,0,__PYX_ERR(1, 1956, __pyx_L1_error))
  __pyx_v_w85 = (__pyx_v_w0 * __pyx_v_w28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1957
 *         cdef double w84 = -6572428.80109422*nu + 1300341.64885296*w28 + 2048.0*w51*w83 + 688128.0*w63 + 1720320.0
 *         cdef double w85 = w0*w28
 *         cdef double w86 = 4.0*w17             # <<<<<<<<<<<<<<
 *         cdef double w87 = 6.0*w1 + w65 + 8.0
 *         cdef double w88 = 1.31621673590926e-19*w52*(53760.0*nu*(3740417.71815805*r + 2115968.85907902*w1 - 938918.400156317*w13 + 1057984.42953951*w17 + w68*(w86 + w87) + w69*(17058.7851094412*r + 12794.0888320809*w1 - 27409.3925547206*w17 + 13218.7851094412) + w71*(5778.0*w1 + 1396.0*w17 + w70 + 7704.0) + w72*(-w86 + w87) + 2888096.47013111) + 66060288000.0*w13 + 7.0*w28*(-117964800.0*a6 + 4129567622.65173*r - 18535504222.6128*w1 + 7133302759.42198*w13 - 12357002815.0752*w17 + 122635399361.987) + 135291469824.0*w49*w85 + 7.0*w56*(-5706642207.53758*r - 26189434371.6082) + 32768.0*w62*w83 + 32768.0*w63*(-1791904.9465589*nu + 3840.0*r*w59 + 2880.0*w1*w60 + 806400.0*w13 - 725760.0*w28 + 1920.0*w61 + 4143360.0))/(0.0546957463279941*w19 + w53 + 0.28004222119933*w55 + 4.63661586574928e-9*w57 + 2.8978849160933e-11*w58 + 1.35654132757922e-7*w64 + 2.22557561555966e-7*w73)**2
 */
  __Pyx_TraceLine(1957,0,__PYX_ERR(1, 1957, __pyx_L1_error))
  __pyx_v_w86 = (4.0 * __pyx_v_w17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1958
 *         cdef double w85 = w0*w28
 *         cdef double w86 = 4.0*w17
 *         cdef double w87 = 6.0*w1 + w65 + 8.0             # <<<<<<<<<<<<<<
 *         cdef double w88 = 1.31621673590926e-19*w52*(53760.0*nu*(3740417.71815805*r + 2115968.85907902*w1 - 938918.400156317*w13 + 1057984.42953951*w17 + w68*(w86 + w87) + w69*(17058.7851094412*r + 12794.0888320809*w1 - 27409.3925547206*w17 + 13218.7851094412) + w71*(5778.0*w1 + 1396.0*w17 + w70 + 7704.0) + w72*(-w86 + w87) + 2888096.47013111) + 66060288000.0*w13 + 7.0*w28*(-117964800.0*a6 + 4129567622.65173*r - 18535504222.6128*w1 + 7133302759.42198*w13 - 12357002815.0752*w17 + 122635399361.987) + 135291469824.0*w49*w85 + 7.0*w56*(-5706642207.53758*r - 26189434371.6082) + 32768.0*w62*w83 + 32768.0*w63*(-1791904.9465589*nu + 3840.0*r*w59 + 2880.0*w1*w60 + 806400.0*w13 - 725760.0*w28 + 1920.0*w61 + 4143360.0))/(0.0546957463279941*w19 + w53 + 0.28004222119933*w55 + 4.63661586574928e-9*w57 + 2.8978849160933e-11*w58 + 1.35654132757922e-7*w64 + 2.22557561555966e-7*w73)**2
 *         cdef double w89 = w13*w88
 */
  __Pyx_TraceLine(1958,0,__PYX_ERR(1, 1958, __pyx_L1_error))
  __pyx_v_w87 = (((6.0 * __pyx_v_w1) + __pyx_v_w65) + 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1959
 *         cdef double w86 = 4.0*w17
 *         cdef double w87 = 6.0*w1 + w65 + 8.0
 *         cdef double w88 = 1.31621673590926e-19*w52*(53760.0*nu*(3740417.71815805*r + 2115968.85907902*w1 - 938918.400156317*w13 + 1057984.42953951*w17 + w68*(w86 + w87) + w69*(17058.7851094412*r + 12794.0888320809*w1 - 27409.3925547206*w17 + 13218.7851094412) + w71*(5778.0*w1 + 1396.0*w17 + w70 + 7704.0) + w72*(-w86 + w87) + 2888096.47013111) + 66060288000.0*w13 + 7.0*w28*(-117964800.0*a6 + 4129567622.65173*r - 18535504222.6128*w1 + 7133302759.42198*w13 - 12357002815.0752*w17 + 122635399361.987) + 135291469824.0*w49*w85 + 7.0*w56*(-5706642207.53758*r - 26189434371.6082) + 32768.0*w62*w83 + 32768.0*w63*(-1791904.9465589*nu + 3840.0*r*w59 + 2880.0*w1*w60 + 806400.0*w13 - 725760.0*w28 + 1920.0*w61 + 4143360.0))/(0.0546957463279941*w19 + w53 + 0.28004222119933*w55 + 4.63661586574928e-9*w57 + 2.8978849160933e-11*w58 + 1.35654132757922e-7*w64 + 2.22557561555966e-7*w73)**2             # <<<<<<<<<<<<<<
 *         cdef double w89 = w13*w88
 *         cdef double w90 = w80*(7680.0*w13*w74*w84 - w14*(-3.0*w30 + w41 + w42) + w16*w18 + 30720.0*w17*w52*w74 - w20*(8.25*w26 + 0.5*w44 + 0.25*w46) - w20*(w3*w81 - 5.0*w37 + 0.5*w47 + 4.5*w7 + 4.5*w8) - w82*(0.625*w27 + w32*(14350.0*nu - 1320.0*w28 + 2805.0) + w35*(5830.0*nu - 1640.0*w28 - 855.0)) - w82*(-6.09375*w38 - w40*(35.0*nu + 45.0) + 0.015625*w43*(-135.0*nu + 140.0*w28 - 15.0)) - w89)
 */
  __Pyx_TraceLine(1959,0,__PYX_ERR(1, 1959, __pyx_L1_error))
  __pyx_v_w88 = (((1.31621673590926e-19 * __pyx_v_w52) * ((((((((53760.0 * __pyx_v_nu) * (((((((((3740417.71815805 * __pyx_v_r) + (2115968.85907902 * __pyx_v_w1)) - (938918.400156317 * __pyx_v_w13)) + (1057984.42953951 * __pyx_v_w17)) + (__pyx_v_w68 * (__pyx_v_w86 + __pyx_v_w87))) + (__pyx_v_w69 * ((((17058.7851094412 * __pyx_v_r) + (12794.0888320809 * __pyx_v_w1)) - (27409.3925547206 * __pyx_v_w17)) + 13218.7851094412))) + (__pyx_v_w71 * ((((5778.0 * __pyx_v_w1) + (1396.0 * __pyx_v_w17)) + __pyx_v_w70) + 7704.0))) + (__pyx_v_w72 * ((-__pyx_v_w86) + __pyx_v_w87))) + 2888096.47013111)) + (66060288000.0 * __pyx_v_w13)) + ((7.0 * __pyx_v_w28) * ((((((-117964800.0 * __pyx_v_a6) + (4129567622.65173 * __pyx_v_r)) - (18535504222.6128 * __pyx_v_w1)) + (7133302759.42198 * __pyx_v_w13)) - (12357002815.0752 * __pyx_v_w17)) + 122635399361.987))) + ((135291469824.0 * __pyx_v_w49) * __pyx_v_w85)) + ((7.0 * __pyx_v_w56) * ((-5706642207.53758 * __pyx_v_r) - 26189434371.6082))) + ((32768.0 * __pyx_v_w62) * __pyx_v_w83)) + ((32768.0 * __pyx_v_w63) * (((((((-1791904.9465589 * __pyx_v_nu) + ((3840.0 * __pyx_v_r) * __pyx_v_w59)) + ((2880.0 * __pyx_v_w1) * __pyx_v_w60)) + (806400.0 * __pyx_v_w13)) - (725760.0 * __pyx_v_w28)) + (1920.0 * __pyx_v_w61)) + 4143360.0)))) / pow((((((((0.0546957463279941 * __pyx_v_w19) + __pyx_v_w53) + (0.28004222119933 * __pyx_v_w55)) + (4.63661586574928e-9 * __pyx_v_w57)) + (2.8978849160933e-11 * __pyx_v_w58)) + (1.35654132757922e-7 * __pyx_v_w64)) + (2.22557561555966e-7 * __pyx_v_w73)), 2.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1960
 *         cdef double w87 = 6.0*w1 + w65 + 8.0
 *         cdef double w88 = 1.31621673590926e-19*w52*(53760.0*nu*(3740417.71815805*r + 2115968.85907902*w1 - 938918.400156317*w13 + 1057984.42953951*w17 + w68*(w86 + w87) + w69*(17058.7851094412*r + 12794.0888320809*w1 - 27409.3925547206*w17 + 13218.7851094412) + w71*(5778.0*w1 + 1396.0*w17 + w70 + 7704.0) + w72*(-w86 + w87) + 2888096.47013111) + 66060288000.0*w13 + 7.0*w28*(-117964800.0*a6 + 4129567622.65173*r - 18535504222.6128*w1 + 7133302759.42198*w13 - 12357002815.0752*w17 + 122635399361.987) + 135291469824.0*w49*w85 + 7.0*w56*(-5706642207.53758*r - 26189434371.6082) + 32768.0*w62*w83 + 32768.0*w63*(-1791904.9465589*nu + 3840.0*r*w59 + 2880.0*w1*w60 + 806400.0*w13 - 725760.0*w28 + 1920.0*w61 + 4143360.0))/(0.0546957463279941*w19 + w53 + 0.28004222119933*w55 + 4.63661586574928e-9*w57 + 2.8978849160933e-11*w58 + 1.35654132757922e-7*w64 + 2.22557561555966e-7*w73)**2
 *         cdef double w89 = w13*w88             # <<<<<<<<<<<<<<
 *         cdef double w90 = w80*(7680.0*w13*w74*w84 - w14*(-3.0*w30 + w41 + w42) + w16*w18 + 30720.0*w17*w52*w74 - w20*(8.25*w26 + 0.5*w44 + 0.25*w46) - w20*(w3*w81 - 5.0*w37 + 0.5*w47 + 4.5*w7 + 4.5*w8) - w82*(0.625*w27 + w32*(14350.0*nu - 1320.0*w28 + 2805.0) + w35*(5830.0*nu - 1640.0*w28 - 855.0)) - w82*(-6.09375*w38 - w40*(35.0*nu + 45.0) + 0.015625*w43*(-135.0*nu + 140.0*w28 - 15.0)) - w89)
 *         cdef double w91 = w16*w18
 */
  __Pyx_TraceLine(1960,0,__PYX_ERR(1, 1960, __pyx_L1_error))
  __pyx_v_w89 = (__pyx_v_w13 * __pyx_v_w88);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1961
 *         cdef double w88 = 1.31621673590926e-19*w52*(53760.0*nu*(3740417.71815805*r + 2115968.85907902*w1 - 938918.400156317*w13 + 1057984.42953951*w17 + w68*(w86 + w87) + w69*(17058.7851094412*r + 12794.0888320809*w1 - 27409.3925547206*w17 + 13218.7851094412) + w71*(5778.0*w1 + 1396.0*w17 + w70 + 7704.0) + w72*(-w86 + w87) + 2888096.47013111) + 66060288000.0*w13 + 7.0*w28*(-117964800.0*a6 + 4129567622.65173*r - 18535504222.6128*w1 + 7133302759.42198*w13 - 12357002815.0752*w17 + 122635399361.987) + 135291469824.0*w49*w85 + 7.0*w56*(-5706642207.53758*r - 26189434371.6082) + 32768.0*w62*w83 + 32768.0*w63*(-1791904.9465589*nu + 3840.0*r*w59 + 2880.0*w1*w60 + 806400.0*w13 - 725760.0*w28 + 1920.0*w61 + 4143360.0))/(0.0546957463279941*w19 + w53 + 0.28004222119933*w55 + 4.63661586574928e-9*w57 + 2.8978849160933e-11*w58 + 1.35654132757922e-7*w64 + 2.22557561555966e-7*w73)**2
 *         cdef double w89 = w13*w88
 *         cdef double w90 = w80*(7680.0*w13*w74*w84 - w14*(-3.0*w30 + w41 + w42) + w16*w18 + 30720.0*w17*w52*w74 - w20*(8.25*w26 + 0.5*w44 + 0.25*w46) - w20*(w3*w81 - 5.0*w37 + 0.5*w47 + 4.5*w7 + 4.5*w8) - w82*(0.625*w27 + w32*(14350.0*nu - 1320.0*w28 + 2805.0) + w35*(5830.0*nu - 1640.0*w28 - 855.0)) - w82*(-6.09375*w38 - w40*(35.0*nu + 45.0) + 0.015625*w43*(-135.0*nu + 140.0*w28 - 15.0)) - w89)             # <<<<<<<<<<<<<<
 *         cdef double w91 = w16*w18
 *         cdef double w92 = 0.0625*w43
 */
  __Pyx_TraceLine(1961,0,__PYX_ERR(1, 1961, __pyx_L1_error))
  __pyx_v_w90 = (__pyx_v_w80 * (((((((((((7680.0 * __pyx_v_w13) * __pyx_v_w74) * __pyx_v_w84) - (__pyx_v_w14 * (((-3.0 * __pyx_v_w30) + __pyx_v_w41) + __pyx_v_w42))) + (__pyx_v_w16 * __pyx_v_w18)) + (((30720.0 * __pyx_v_w17) * __pyx_v_w52) * __pyx_v_w74)) - (__pyx_v_w20 * (((8.25 * __pyx_v_w26) + (0.5 * __pyx_v_w44)) + (0.25 * __pyx_v_w46)))) - (__pyx_v_w20 * (((((__pyx_v_w3 * __pyx_v_w81) - (5.0 * __pyx_v_w37)) + (0.5 * __pyx_v_w47)) + (4.5 * __pyx_v_w7)) + (4.5 * __pyx_v_w8)))) - (__pyx_v_w82 * (((0.625 * __pyx_v_w27) + (__pyx_v_w32 * (((14350.0 * __pyx_v_nu) - (1320.0 * __pyx_v_w28)) + 2805.0))) + (__pyx_v_w35 * (((5830.0 * __pyx_v_nu) - (1640.0 * __pyx_v_w28)) - 855.0))))) - (__pyx_v_w82 * (((-6.09375 * __pyx_v_w38) - (__pyx_v_w40 * ((35.0 * __pyx_v_nu) + 45.0))) + ((0.015625 * __pyx_v_w43) * (((-135.0 * __pyx_v_nu) + (140.0 * __pyx_v_w28)) - 15.0))))) - __pyx_v_w89));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1962
 *         cdef double w89 = w13*w88
 *         cdef double w90 = w80*(7680.0*w13*w74*w84 - w14*(-3.0*w30 + w41 + w42) + w16*w18 + 30720.0*w17*w52*w74 - w20*(8.25*w26 + 0.5*w44 + 0.25*w46) - w20*(w3*w81 - 5.0*w37 + 0.5*w47 + 4.5*w7 + 4.5*w8) - w82*(0.625*w27 + w32*(14350.0*nu - 1320.0*w28 + 2805.0) + w35*(5830.0*nu - 1640.0*w28 - 855.0)) - w82*(-6.09375*w38 - w40*(35.0*nu + 45.0) + 0.015625*w43*(-135.0*nu + 140.0*w28 - 15.0)) - w89)
 *         cdef double w91 = w16*w18             # <<<<<<<<<<<<<<
 *         cdef double w92 = 0.0625*w43
 *         cdef double w93 = w37*(26.0*nu + 449.0)
 */
  __Pyx_TraceLine(1962,0,__PYX_ERR(1, 1962, __pyx_L1_error))
  __pyx_v_w91 = (__pyx_v_w16 * __pyx_v_w18);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1963
 *         cdef double w90 = w80*(7680.0*w13*w74*w84 - w14*(-3.0*w30 + w41 + w42) + w16*w18 + 30720.0*w17*w52*w74 - w20*(8.25*w26 + 0.5*w44 + 0.25*w46) - w20*(w3*w81 - 5.0*w37 + 0.5*w47 + 4.5*w7 + 4.5*w8) - w82*(0.625*w27 + w32*(14350.0*nu - 1320.0*w28 + 2805.0) + w35*(5830.0*nu - 1640.0*w28 - 855.0)) - w82*(-6.09375*w38 - w40*(35.0*nu + 45.0) + 0.015625*w43*(-135.0*nu + 140.0*w28 - 15.0)) - w89)
 *         cdef double w91 = w16*w18
 *         cdef double w92 = 0.0625*w43             # <<<<<<<<<<<<<<
 *         cdef double w93 = w37*(26.0*nu + 449.0)
 *         cdef double w94 = w10*(-1171.0*nu - 861.0)
 */
  __Pyx_TraceLine(1963,0,__PYX_ERR(1, 1963, __pyx_L1_error))
  __pyx_v_w92 = (0.0625 * __pyx_v_w43);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1964
 *         cdef double w91 = w16*w18
 *         cdef double w92 = 0.0625*w43
 *         cdef double w93 = w37*(26.0*nu + 449.0)             # <<<<<<<<<<<<<<
 *         cdef double w94 = w10*(-1171.0*nu - 861.0)
 *         cdef double w95 = 4.0*w28
 */
  __Pyx_TraceLine(1964,0,__PYX_ERR(1, 1964, __pyx_L1_error))
  __pyx_v_w93 = (__pyx_v_w37 * ((26.0 * __pyx_v_nu) + 449.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1965
 *         cdef double w92 = 0.0625*w43
 *         cdef double w93 = w37*(26.0*nu + 449.0)
 *         cdef double w94 = w10*(-1171.0*nu - 861.0)             # <<<<<<<<<<<<<<
 *         cdef double w95 = 4.0*w28
 *         cdef double w96 = w33*(115.0*nu + w95 - 37.0)
 */
  __Pyx_TraceLine(1965,0,__PYX_ERR(1, 1965, __pyx_L1_error))
  __pyx_v_w94 = (__pyx_v_w10 * ((-1171.0 * __pyx_v_nu) - 861.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1966
 *         cdef double w93 = w37*(26.0*nu + 449.0)
 *         cdef double w94 = w10*(-1171.0*nu - 861.0)
 *         cdef double w95 = 4.0*w28             # <<<<<<<<<<<<<<
 *         cdef double w96 = w33*(115.0*nu + w95 - 37.0)
 *         cdef double w97 = 5787938193408.0*w54
 */
  __Pyx_TraceLine(1966,0,__PYX_ERR(1, 1966, __pyx_L1_error))
  __pyx_v_w95 = (4.0 * __pyx_v_w28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1967
 *         cdef double w94 = w10*(-1171.0*nu - 861.0)
 *         cdef double w95 = 4.0*w28
 *         cdef double w96 = w33*(115.0*nu + w95 - 37.0)             # <<<<<<<<<<<<<<
 *         cdef double w97 = 5787938193408.0*w54
 *         cdef double w98 = -630116198.873299*nu - 197773496.793534*w28 + 5805304367.87913
 */
  __Pyx_TraceLine(1967,0,__PYX_ERR(1, 1967, __pyx_L1_error))
  __pyx_v_w96 = (__pyx_v_w33 * (((115.0 * __pyx_v_nu) + __pyx_v_w95) - 37.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1968
 *         cdef double w95 = 4.0*w28
 *         cdef double w96 = w33*(115.0*nu + w95 - 37.0)
 *         cdef double w97 = 5787938193408.0*w54             # <<<<<<<<<<<<<<
 *         cdef double w98 = -630116198.873299*nu - 197773496.793534*w28 + 5805304367.87913
 *         cdef double w99 = w1*w98
 */
  __Pyx_TraceLine(1968,0,__PYX_ERR(1, 1968, __pyx_L1_error))
  __pyx_v_w97 = (5787938193408.0 * __pyx_v_w54);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1969
 *         cdef double w96 = w33*(115.0*nu + w95 - 37.0)
 *         cdef double w97 = 5787938193408.0*w54
 *         cdef double w98 = -630116198.873299*nu - 197773496.793534*w28 + 5805304367.87913             # <<<<<<<<<<<<<<
 *         cdef double w99 = w1*w98
 *         cdef double w100 = -2675575.66847905*nu - 138240.0*w28 - 5278341.3229329
 */
  __Pyx_TraceLine(1969,0,__PYX_ERR(1, 1969, __pyx_L1_error))
  __pyx_v_w98 = (((-630116198.873299 * __pyx_v_nu) - (197773496.793534 * __pyx_v_w28)) + 5805304367.87913);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1970
 *         cdef double w97 = 5787938193408.0*w54
 *         cdef double w98 = -630116198.873299*nu - 197773496.793534*w28 + 5805304367.87913
 *         cdef double w99 = w1*w98             # <<<<<<<<<<<<<<
 *         cdef double w100 = -2675575.66847905*nu - 138240.0*w28 - 5278341.3229329
 *         cdef double w101 = w100*w17
 */
  __Pyx_TraceLine(1970,0,__PYX_ERR(1, 1970, __pyx_L1_error))
  __pyx_v_w99 = (__pyx_v_w1 * __pyx_v_w98);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1971
 *         cdef double w98 = -630116198.873299*nu - 197773496.793534*w28 + 5805304367.87913
 *         cdef double w99 = w1*w98
 *         cdef double w100 = -2675575.66847905*nu - 138240.0*w28 - 5278341.3229329             # <<<<<<<<<<<<<<
 *         cdef double w101 = w100*w17
 *         cdef double w102 = nu*(-2510664218.28128*nu - 42636451.6032331*w28 + 14515200.0*w56 + 1002013764.01019)
 */
  __Pyx_TraceLine(1971,0,__PYX_ERR(1, 1971, __pyx_L1_error))
  __pyx_v_w100 = (((-2675575.66847905 * __pyx_v_nu) - (138240.0 * __pyx_v_w28)) - 5278341.3229329);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1972
 *         cdef double w99 = w1*w98
 *         cdef double w100 = -2675575.66847905*nu - 138240.0*w28 - 5278341.3229329
 *         cdef double w101 = w100*w17             # <<<<<<<<<<<<<<
 *         cdef double w102 = nu*(-2510664218.28128*nu - 42636451.6032331*w28 + 14515200.0*w56 + 1002013764.01019)
 *         cdef double w103 = 43393301259014.8*nu + 43133561885859.3*w28 + 5927865218923.02*w56 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 */
  __Pyx_TraceLine(1972,0,__PYX_ERR(1, 1972, __pyx_L1_error))
  __pyx_v_w101 = (__pyx_v_w100 * __pyx_v_w17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1973
 *         cdef double w100 = -2675575.66847905*nu - 138240.0*w28 - 5278341.3229329
 *         cdef double w101 = w100*w17
 *         cdef double w102 = nu*(-2510664218.28128*nu - 42636451.6032331*w28 + 14515200.0*w56 + 1002013764.01019)             # <<<<<<<<<<<<<<
 *         cdef double w103 = 43393301259014.8*nu + 43133561885859.3*w28 + 5927865218923.02*w56 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double w104 = r*w103
 */
  __Pyx_TraceLine(1973,0,__PYX_ERR(1, 1973, __pyx_L1_error))
  __pyx_v_w102 = (__pyx_v_nu * ((((-2510664218.28128 * __pyx_v_nu) - (42636451.6032331 * __pyx_v_w28)) + (14515200.0 * __pyx_v_w56)) + 1002013764.01019));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1974
 *         cdef double w101 = w100*w17
 *         cdef double w102 = nu*(-2510664218.28128*nu - 42636451.6032331*w28 + 14515200.0*w56 + 1002013764.01019)
 *         cdef double w103 = 43393301259014.8*nu + 43133561885859.3*w28 + 5927865218923.02*w56 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0             # <<<<<<<<<<<<<<
 *         cdef double w104 = r*w103
 *         cdef double w105 = 14700.0*nu + 42911.0
 */
  __Pyx_TraceLine(1974,0,__PYX_ERR(1, 1974, __pyx_L1_error))
  __pyx_v_w103 = (((((43393301259014.8 * __pyx_v_nu) + (43133561885859.3 * __pyx_v_w28)) + (5927865218923.02 * __pyx_v_w56)) + (86618264430493.3 * pow((1.0 - (0.496948781616935 * __pyx_v_nu)), 2.0))) + 188440788778196.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1975
 *         cdef double w102 = nu*(-2510664218.28128*nu - 42636451.6032331*w28 + 14515200.0*w56 + 1002013764.01019)
 *         cdef double w103 = 43393301259014.8*nu + 43133561885859.3*w28 + 5927865218923.02*w56 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double w104 = r*w103             # <<<<<<<<<<<<<<
 *         cdef double w105 = 14700.0*nu + 42911.0
 *         cdef double w106 = 283115520.0*w105
 */
  __Pyx_TraceLine(1975,0,__PYX_ERR(1, 1975, __pyx_L1_error))
  __pyx_v_w104 = (__pyx_v_r * __pyx_v_w103);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1976
 *         cdef double w103 = 43393301259014.8*nu + 43133561885859.3*w28 + 5927865218923.02*w56 + 86618264430493.3*(1 - 0.496948781616935*nu)**2 + 188440788778196.0
 *         cdef double w104 = r*w103
 *         cdef double w105 = 14700.0*nu + 42911.0             # <<<<<<<<<<<<<<
 *         cdef double w106 = 283115520.0*w105
 *         cdef double w107 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*w28 - 2119671837.36038) + w1*w106 + 879923036160.0*w17
 */
  __Pyx_TraceLine(1976,0,__PYX_ERR(1, 1976, __pyx_L1_error))
  __pyx_v_w105 = ((14700.0 * __pyx_v_nu) + 42911.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1977
 *         cdef double w104 = r*w103
 *         cdef double w105 = 14700.0*nu + 42911.0
 *         cdef double w106 = 283115520.0*w105             # <<<<<<<<<<<<<<
 *         cdef double w107 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*w28 - 2119671837.36038) + w1*w106 + 879923036160.0*w17
 *         cdef double w108 = w107*w49
 */
  __Pyx_TraceLine(1977,0,__PYX_ERR(1, 1977, __pyx_L1_error))
  __pyx_v_w106 = (283115520.0 * __pyx_v_w105);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1978
 *         cdef double w105 = 14700.0*nu + 42911.0
 *         cdef double w106 = 283115520.0*w105
 *         cdef double w107 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*w28 - 2119671837.36038) + w1*w106 + 879923036160.0*w17             # <<<<<<<<<<<<<<
 *         cdef double w108 = w107*w49
 *         cdef double w109 = r*w97 - 967680.0*w101 + 55296.0*w102 + w104 + w108 - 9216.0*w99
 */
  __Pyx_TraceLine(1978,0,__PYX_ERR(1, 1978, __pyx_L1_error))
  __pyx_v_w107 = (((((-1698693120.0 * __pyx_v_nu) * ((11592.0 * __pyx_v_nu) + 69847.0)) + ((49152.0 * __pyx_v_r) * (((409207698.136075 * __pyx_v_nu) + (102574080.0 * __pyx_v_w28)) - 2119671837.36038))) + (__pyx_v_w1 * __pyx_v_w106)) + (879923036160.0 * __pyx_v_w17));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1979
 *         cdef double w106 = 283115520.0*w105
 *         cdef double w107 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*w28 - 2119671837.36038) + w1*w106 + 879923036160.0*w17
 *         cdef double w108 = w107*w49             # <<<<<<<<<<<<<<
 *         cdef double w109 = r*w97 - 967680.0*w101 + 55296.0*w102 + w104 + w108 - 9216.0*w99
 *         cdef double w110 = w109**(-1)
 */
  __Pyx_TraceLine(1979,0,__PYX_ERR(1, 1979, __pyx_L1_error))
  __pyx_v_w108 = (__pyx_v_w107 * __pyx_v_w49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1980
 *         cdef double w107 = -1698693120.0*nu*(11592.0*nu + 69847.0) + 49152.0*r*(409207698.136075*nu + 102574080.0*w28 - 2119671837.36038) + w1*w106 + 879923036160.0*w17
 *         cdef double w108 = w107*w49
 *         cdef double w109 = r*w97 - 967680.0*w101 + 55296.0*w102 + w104 + w108 - 9216.0*w99             # <<<<<<<<<<<<<<
 *         cdef double w110 = w109**(-1)
 *         cdef double w111 = nu*r
 */
  __Pyx_TraceLine(1980,0,__PYX_ERR(1, 1980, __pyx_L1_error))
  __pyx_v_w109 = ((((((__pyx_v_r * __pyx_v_w97) - (967680.0 * __pyx_v_w101)) + (55296.0 * __pyx_v_w102)) + __pyx_v_w104) + __pyx_v_w108) - (9216.0 * __pyx_v_w99));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1981
 *         cdef double w108 = w107*w49
 *         cdef double w109 = r*w97 - 967680.0*w101 + 55296.0*w102 + w104 + w108 - 9216.0*w99
 *         cdef double w110 = w109**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w111 = nu*r
 *         cdef double w112 = 1822680546449.21*w28
 */
  __Pyx_TraceLine(1981,0,__PYX_ERR(1, 1981, __pyx_L1_error))
  __pyx_v_w110 = pow(__pyx_v_w109, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1982
 *         cdef double w109 = r*w97 - 967680.0*w101 + 55296.0*w102 + w104 + w108 - 9216.0*w99
 *         cdef double w110 = w109**(-1)
 *         cdef double w111 = nu*r             # <<<<<<<<<<<<<<
 *         cdef double w112 = 1822680546449.21*w28
 *         cdef double w113 = r*w28
 */
  __Pyx_TraceLine(1982,0,__PYX_ERR(1, 1982, __pyx_L1_error))
  __pyx_v_w111 = (__pyx_v_nu * __pyx_v_r);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1983
 *         cdef double w110 = w109**(-1)
 *         cdef double w111 = nu*r
 *         cdef double w112 = 1822680546449.21*w28             # <<<<<<<<<<<<<<
 *         cdef double w113 = r*w28
 *         cdef double w114 = 11575876386816.0*w49
 */
  __Pyx_TraceLine(1983,0,__PYX_ERR(1, 1983, __pyx_L1_error))
  __pyx_v_w112 = (1822680546449.21 * __pyx_v_w28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1984
 *         cdef double w111 = nu*r
 *         cdef double w112 = 1822680546449.21*w28
 *         cdef double w113 = r*w28             # <<<<<<<<<<<<<<
 *         cdef double w114 = 11575876386816.0*w49
 *         cdef double w115 = 5041721180160.0*w28 - 104186110149937.0
 */
  __Pyx_TraceLine(1984,0,__PYX_ERR(1, 1984, __pyx_L1_error))
  __pyx_v_w113 = (__pyx_v_r * __pyx_v_w28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1985
 *         cdef double w112 = 1822680546449.21*w28
 *         cdef double w113 = r*w28
 *         cdef double w114 = 11575876386816.0*w49             # <<<<<<<<<<<<<<
 *         cdef double w115 = 5041721180160.0*w28 - 104186110149937.0
 *         cdef double w116 = -25392914995744.3*nu - r*w106 - 879923036160.0*w1 - w115
 */
  __Pyx_TraceLine(1985,0,__PYX_ERR(1, 1985, __pyx_L1_error))
  __pyx_v_w114 = (11575876386816.0 * __pyx_v_w49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1986
 *         cdef double w113 = r*w28
 *         cdef double w114 = 11575876386816.0*w49
 *         cdef double w115 = 5041721180160.0*w28 - 104186110149937.0             # <<<<<<<<<<<<<<
 *         cdef double w116 = -25392914995744.3*nu - r*w106 - 879923036160.0*w1 - w115
 *         cdef double w117 = 5807150888816.34*nu + 10215490662751.4*r + w0*w114 - w0*w116 + 5178202125747.62*w111 + w112 + 267544166400.0*w113 + w49*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 */
  __Pyx_TraceLine(1986,0,__PYX_ERR(1, 1986, __pyx_L1_error))
  __pyx_v_w115 = ((5041721180160.0 * __pyx_v_w28) - 104186110149937.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1987
 *         cdef double w114 = 11575876386816.0*w49
 *         cdef double w115 = 5041721180160.0*w28 - 104186110149937.0
 *         cdef double w116 = -25392914995744.3*nu - r*w106 - 879923036160.0*w1 - w115             # <<<<<<<<<<<<<<
 *         cdef double w117 = 5807150888816.34*nu + 10215490662751.4*r + w0*w114 - w0*w116 + 5178202125747.62*w111 + w112 + 267544166400.0*w113 + w49*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double w118 = w110*w117
 */
  __Pyx_TraceLine(1987,0,__PYX_ERR(1, 1987, __pyx_L1_error))
  __pyx_v_w116 = ((((-25392914995744.3 * __pyx_v_nu) - (__pyx_v_r * __pyx_v_w106)) - (879923036160.0 * __pyx_v_w1)) - __pyx_v_w115);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1988
 *         cdef double w115 = 5041721180160.0*w28 - 104186110149937.0
 *         cdef double w116 = -25392914995744.3*nu - r*w106 - 879923036160.0*w1 - w115
 *         cdef double w117 = 5807150888816.34*nu + 10215490662751.4*r + w0*w114 - w0*w116 + 5178202125747.62*w111 + w112 + 267544166400.0*w113 + w49*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1             # <<<<<<<<<<<<<<
 *         cdef double w118 = w110*w117
 *         cdef double w119 = w52*w74
 */
  __Pyx_TraceLine(1988,0,__PYX_ERR(1, 1988, __pyx_L1_error))
  __pyx_v_w117 = (((((((((5807150888816.34 * __pyx_v_nu) + (10215490662751.4 * __pyx_v_r)) + (__pyx_v_w0 * __pyx_v_w114)) - (__pyx_v_w0 * __pyx_v_w116)) + (5178202125747.62 * __pyx_v_w111)) + __pyx_v_w112) + (267544166400.0 * __pyx_v_w113)) + (__pyx_v_w49 * (((4161798144000.0 * __pyx_v_nu) + (1759846072320.0 * __pyx_v_r)) + 12148770078720.0))) - 53501685054374.1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1989
 *         cdef double w116 = -25392914995744.3*nu - r*w106 - 879923036160.0*w1 - w115
 *         cdef double w117 = 5807150888816.34*nu + 10215490662751.4*r + w0*w114 - w0*w116 + 5178202125747.62*w111 + w112 + 267544166400.0*w113 + w49*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double w118 = w110*w117             # <<<<<<<<<<<<<<
 *         cdef double w119 = w52*w74
 *         cdef double w120 = 7680.0*w119
 */
  __Pyx_TraceLine(1989,0,__PYX_ERR(1, 1989, __pyx_L1_error))
  __pyx_v_w118 = (__pyx_v_w110 * __pyx_v_w117);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1990
 *         cdef double w117 = 5807150888816.34*nu + 10215490662751.4*r + w0*w114 - w0*w116 + 5178202125747.62*w111 + w112 + 267544166400.0*w113 + w49*(4161798144000.0*nu + 1759846072320.0*r + 12148770078720.0) - 53501685054374.1
 *         cdef double w118 = w110*w117
 *         cdef double w119 = w52*w74             # <<<<<<<<<<<<<<
 *         cdef double w120 = 7680.0*w119
 *         cdef double w121 = 7680.0*w84
 */
  __Pyx_TraceLine(1990,0,__PYX_ERR(1, 1990, __pyx_L1_error))
  __pyx_v_w119 = (__pyx_v_w52 * __pyx_v_w74);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1991
 *         cdef double w118 = w110*w117
 *         cdef double w119 = w52*w74
 *         cdef double w120 = 7680.0*w119             # <<<<<<<<<<<<<<
 *         cdef double w121 = 7680.0*w84
 *         cdef double w122 = nu*w1
 */
  __Pyx_TraceLine(1991,0,__PYX_ERR(1, 1991, __pyx_L1_error))
  __pyx_v_w120 = (7680.0 * __pyx_v_w119);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1992
 *         cdef double w119 = w52*w74
 *         cdef double w120 = 7680.0*w119
 *         cdef double w121 = 7680.0*w84             # <<<<<<<<<<<<<<
 *         cdef double w122 = nu*w1
 *         cdef double w123 = r*w105
 */
  __Pyx_TraceLine(1992,0,__PYX_ERR(1, 1992, __pyx_L1_error))
  __pyx_v_w121 = (7680.0 * __pyx_v_w84);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1993
 *         cdef double w120 = 7680.0*w119
 *         cdef double w121 = 7680.0*w84
 *         cdef double w122 = nu*w1             # <<<<<<<<<<<<<<
 *         cdef double w123 = r*w105
 *         cdef double w124 = w1*w28
 */
  __Pyx_TraceLine(1993,0,__PYX_ERR(1, 1993, __pyx_L1_error))
  __pyx_v_w122 = (__pyx_v_nu * __pyx_v_w1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1994
 *         cdef double w121 = 7680.0*w84
 *         cdef double w122 = nu*w1
 *         cdef double w123 = r*w105             # <<<<<<<<<<<<<<
 *         cdef double w124 = w1*w28
 *         cdef double w125 = w116*w49
 */
  __Pyx_TraceLine(1994,0,__PYX_ERR(1, 1994, __pyx_L1_error))
  __pyx_v_w123 = (__pyx_v_r * __pyx_v_w105);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1995
 *         cdef double w122 = nu*w1
 *         cdef double w123 = r*w105
 *         cdef double w124 = w1*w28             # <<<<<<<<<<<<<<
 *         cdef double w125 = w116*w49
 *         cdef double w126 = -12049908701745.2*nu + r*w112 - 39476764256925.6*r + 5107745331375.71*w1 + 10611661054566.2*w111 + 2589101062873.81*w122 - 326837426.241486*w123 + 133772083200.0*w124 - w125 + 80059249540278.2*w28 + 6730497718123.02*w56 + w97 + 275059053208689.0
 */
  __Pyx_TraceLine(1995,0,__PYX_ERR(1, 1995, __pyx_L1_error))
  __pyx_v_w124 = (__pyx_v_w1 * __pyx_v_w28);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1996
 *         cdef double w123 = r*w105
 *         cdef double w124 = w1*w28
 *         cdef double w125 = w116*w49             # <<<<<<<<<<<<<<
 *         cdef double w126 = -12049908701745.2*nu + r*w112 - 39476764256925.6*r + 5107745331375.71*w1 + 10611661054566.2*w111 + 2589101062873.81*w122 - 326837426.241486*w123 + 133772083200.0*w124 - w125 + 80059249540278.2*w28 + 6730497718123.02*w56 + w97 + 275059053208689.0
 *         cdef double w127 = w110*w126
 */
  __Pyx_TraceLine(1996,0,__PYX_ERR(1, 1996, __pyx_L1_error))
  __pyx_v_w125 = (__pyx_v_w116 * __pyx_v_w49);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1997
 *         cdef double w124 = w1*w28
 *         cdef double w125 = w116*w49
 *         cdef double w126 = -12049908701745.2*nu + r*w112 - 39476764256925.6*r + 5107745331375.71*w1 + 10611661054566.2*w111 + 2589101062873.81*w122 - 326837426.241486*w123 + 133772083200.0*w124 - w125 + 80059249540278.2*w28 + 6730497718123.02*w56 + w97 + 275059053208689.0             # <<<<<<<<<<<<<<
 *         cdef double w127 = w110*w126
 *         cdef double w128 = w127*w19
 */
  __Pyx_TraceLine(1997,0,__PYX_ERR(1, 1997, __pyx_L1_error))
  __pyx_v_w126 = (((((((((((((-12049908701745.2 * __pyx_v_nu) + (__pyx_v_r * __pyx_v_w112)) - (39476764256925.6 * __pyx_v_r)) + (5107745331375.71 * __pyx_v_w1)) + (10611661054566.2 * __pyx_v_w111)) + (2589101062873.81 * __pyx_v_w122)) - (326837426.241486 * __pyx_v_w123)) + (133772083200.0 * __pyx_v_w124)) - __pyx_v_w125) + (80059249540278.2 * __pyx_v_w28)) + (6730497718123.02 * __pyx_v_w56)) + __pyx_v_w97) + 275059053208689.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1998
 *         cdef double w125 = w116*w49
 *         cdef double w126 = -12049908701745.2*nu + r*w112 - 39476764256925.6*r + 5107745331375.71*w1 + 10611661054566.2*w111 + 2589101062873.81*w122 - 326837426.241486*w123 + 133772083200.0*w124 - w125 + 80059249540278.2*w28 + 6730497718123.02*w56 + w97 + 275059053208689.0
 *         cdef double w127 = w110*w126             # <<<<<<<<<<<<<<
 *         cdef double w128 = w127*w19
 *         cdef double w129 = (r*w54 - 1.67189069348064e-7*w101 + 9.55366110560367e-9*w102 + 1.72773095804465e-13*w104 + 1.72773095804465e-13*w108 - 1.59227685093395e-9*w99)**2
 */
  __Pyx_TraceLine(1998,0,__PYX_ERR(1, 1998, __pyx_L1_error))
  __pyx_v_w127 = (__pyx_v_w110 * __pyx_v_w126);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1999
 *         cdef double w126 = -12049908701745.2*nu + r*w112 - 39476764256925.6*r + 5107745331375.71*w1 + 10611661054566.2*w111 + 2589101062873.81*w122 - 326837426.241486*w123 + 133772083200.0*w124 - w125 + 80059249540278.2*w28 + 6730497718123.02*w56 + w97 + 275059053208689.0
 *         cdef double w127 = w110*w126
 *         cdef double w128 = w127*w19             # <<<<<<<<<<<<<<
 *         cdef double w129 = (r*w54 - 1.67189069348064e-7*w101 + 9.55366110560367e-9*w102 + 1.72773095804465e-13*w104 + 1.72773095804465e-13*w108 - 1.59227685093395e-9*w99)**2
 *         cdef double w130 = -18432.0*r*w98 + w0*w107 - 2903040.0*w1*w100 + w103 + w114 + w49*(20113376778784.3*nu + 2639769108480.0*w1 + w115 + 566231040.0*w123) + w97
 */
  __Pyx_TraceLine(1999,0,__PYX_ERR(1, 1999, __pyx_L1_error))
  __pyx_v_w128 = (__pyx_v_w127 * __pyx_v_w19);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2000
 *         cdef double w127 = w110*w126
 *         cdef double w128 = w127*w19
 *         cdef double w129 = (r*w54 - 1.67189069348064e-7*w101 + 9.55366110560367e-9*w102 + 1.72773095804465e-13*w104 + 1.72773095804465e-13*w108 - 1.59227685093395e-9*w99)**2             # <<<<<<<<<<<<<<
 *         cdef double w130 = -18432.0*r*w98 + w0*w107 - 2903040.0*w1*w100 + w103 + w114 + w49*(20113376778784.3*nu + 2639769108480.0*w1 + w115 + 566231040.0*w123) + w97
 *         cdef double w131 = w126*w130/w129
 */
  __Pyx_TraceLine(2000,0,__PYX_ERR(1, 2000, __pyx_L1_error))
  __pyx_v_w129 = pow(((((((__pyx_v_r * __pyx_v_w54) - (1.67189069348064e-7 * __pyx_v_w101)) + (9.55366110560367e-9 * __pyx_v_w102)) + (1.72773095804465e-13 * __pyx_v_w104)) + (1.72773095804465e-13 * __pyx_v_w108)) - (1.59227685093395e-9 * __pyx_v_w99)), 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2001
 *         cdef double w128 = w127*w19
 *         cdef double w129 = (r*w54 - 1.67189069348064e-7*w101 + 9.55366110560367e-9*w102 + 1.72773095804465e-13*w104 + 1.72773095804465e-13*w108 - 1.59227685093395e-9*w99)**2
 *         cdef double w130 = -18432.0*r*w98 + w0*w107 - 2903040.0*w1*w100 + w103 + w114 + w49*(20113376778784.3*nu + 2639769108480.0*w1 + w115 + 566231040.0*w123) + w97             # <<<<<<<<<<<<<<
 *         cdef double w131 = w126*w130/w129
 *         cdef double w132 = w118*w120*w19 - 2.29252167428035e-22*w119*w131*w19 + w121*w128*w74 + 38400.0*w127*w76 - w128*w88 - w14*(w10*(9.0*nu + 8.4375) - 7.875*w37 + w92*(12.0*nu - 3.0)) - w20*(0.125*w93 + 0.0625*w94 + 0.0625*w96) + w91
 */
  __Pyx_TraceLine(2001,0,__PYX_ERR(1, 2001, __pyx_L1_error))
  __pyx_v_w130 = ((((((((-18432.0 * __pyx_v_r) * __pyx_v_w98) + (__pyx_v_w0 * __pyx_v_w107)) - ((2903040.0 * __pyx_v_w1) * __pyx_v_w100)) + __pyx_v_w103) + __pyx_v_w114) + (__pyx_v_w49 * ((((20113376778784.3 * __pyx_v_nu) + (2639769108480.0 * __pyx_v_w1)) + __pyx_v_w115) + (566231040.0 * __pyx_v_w123)))) + __pyx_v_w97);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2002
 *         cdef double w129 = (r*w54 - 1.67189069348064e-7*w101 + 9.55366110560367e-9*w102 + 1.72773095804465e-13*w104 + 1.72773095804465e-13*w108 - 1.59227685093395e-9*w99)**2
 *         cdef double w130 = -18432.0*r*w98 + w0*w107 - 2903040.0*w1*w100 + w103 + w114 + w49*(20113376778784.3*nu + 2639769108480.0*w1 + w115 + 566231040.0*w123) + w97
 *         cdef double w131 = w126*w130/w129             # <<<<<<<<<<<<<<
 *         cdef double w132 = w118*w120*w19 - 2.29252167428035e-22*w119*w131*w19 + w121*w128*w74 + 38400.0*w127*w76 - w128*w88 - w14*(w10*(9.0*nu + 8.4375) - 7.875*w37 + w92*(12.0*nu - 3.0)) - w20*(0.125*w93 + 0.0625*w94 + 0.0625*w96) + w91
 *         cdef double w133 = r + 2.0
 */
  __Pyx_TraceLine(2002,0,__PYX_ERR(1, 2002, __pyx_L1_error))
  __pyx_v_w131 = ((__pyx_v_w126 * __pyx_v_w130) / __pyx_v_w129);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2003
 *         cdef double w130 = -18432.0*r*w98 + w0*w107 - 2903040.0*w1*w100 + w103 + w114 + w49*(20113376778784.3*nu + 2639769108480.0*w1 + w115 + 566231040.0*w123) + w97
 *         cdef double w131 = w126*w130/w129
 *         cdef double w132 = w118*w120*w19 - 2.29252167428035e-22*w119*w131*w19 + w121*w128*w74 + 38400.0*w127*w76 - w128*w88 - w14*(w10*(9.0*nu + 8.4375) - 7.875*w37 + w92*(12.0*nu - 3.0)) - w20*(0.125*w93 + 0.0625*w94 + 0.0625*w96) + w91             # <<<<<<<<<<<<<<
 *         cdef double w133 = r + 2.0
 *         cdef double w134 = r*w10
 */
  __Pyx_TraceLine(2003,0,__PYX_ERR(1, 2003, __pyx_L1_error))
  __pyx_v_w132 = (((((((((__pyx_v_w118 * __pyx_v_w120) * __pyx_v_w19) - (((2.29252167428035e-22 * __pyx_v_w119) * __pyx_v_w131) * __pyx_v_w19)) + ((__pyx_v_w121 * __pyx_v_w128) * __pyx_v_w74)) + ((38400.0 * __pyx_v_w127) * __pyx_v_w76)) - (__pyx_v_w128 * __pyx_v_w88)) - (__pyx_v_w14 * (((__pyx_v_w10 * ((9.0 * __pyx_v_nu) + 8.4375)) - (7.875 * __pyx_v_w37)) + (__pyx_v_w92 * ((12.0 * __pyx_v_nu) - 3.0))))) - (__pyx_v_w20 * (((0.125 * __pyx_v_w93) + (0.0625 * __pyx_v_w94)) + (0.0625 * __pyx_v_w96)))) + __pyx_v_w91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2004
 *         cdef double w131 = w126*w130/w129
 *         cdef double w132 = w118*w120*w19 - 2.29252167428035e-22*w119*w131*w19 + w121*w128*w74 + 38400.0*w127*w76 - w128*w88 - w14*(w10*(9.0*nu + 8.4375) - 7.875*w37 + w92*(12.0*nu - 3.0)) - w20*(0.125*w93 + 0.0625*w94 + 0.0625*w96) + w91
 *         cdef double w133 = r + 2.0             # <<<<<<<<<<<<<<
 *         cdef double w134 = r*w10
 *         cdef double w135 = w13 + w133*w134
 */
  __Pyx_TraceLine(2004,0,__PYX_ERR(1, 2004, __pyx_L1_error))
  __pyx_v_w133 = (__pyx_v_r + 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2005
 *         cdef double w132 = w118*w120*w19 - 2.29252167428035e-22*w119*w131*w19 + w121*w128*w74 + 38400.0*w127*w76 - w128*w88 - w14*(w10*(9.0*nu + 8.4375) - 7.875*w37 + w92*(12.0*nu - 3.0)) - w20*(0.125*w93 + 0.0625*w94 + 0.0625*w96) + w91
 *         cdef double w133 = r + 2.0
 *         cdef double w134 = r*w10             # <<<<<<<<<<<<<<
 *         cdef double w135 = w13 + w133*w134
 *         cdef double w136 = w135**(-1)
 */
  __Pyx_TraceLine(2005,0,__PYX_ERR(1, 2005, __pyx_L1_error))
  __pyx_v_w134 = (__pyx_v_r * __pyx_v_w10);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2006
 *         cdef double w133 = r + 2.0
 *         cdef double w134 = r*w10
 *         cdef double w135 = w13 + w133*w134             # <<<<<<<<<<<<<<
 *         cdef double w136 = w135**(-1)
 *         cdef double w137 = w133*w136
 */
  __Pyx_TraceLine(2006,0,__PYX_ERR(1, 2006, __pyx_L1_error))
  __pyx_v_w135 = (__pyx_v_w13 + (__pyx_v_w133 * __pyx_v_w134));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2007
 *         cdef double w134 = r*w10
 *         cdef double w135 = w13 + w133*w134
 *         cdef double w136 = w135**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w137 = w133*w136
 *         cdef double w138 = w135**(-2)
 */
  __Pyx_TraceLine(2007,0,__PYX_ERR(1, 2007, __pyx_L1_error))
  __pyx_v_w136 = pow(__pyx_v_w135, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2008
 *         cdef double w135 = w13 + w133*w134
 *         cdef double w136 = w135**(-1)
 *         cdef double w137 = w133*w136             # <<<<<<<<<<<<<<
 *         cdef double w138 = w135**(-2)
 *         cdef double w139 = w10*w133
 */
  __Pyx_TraceLine(2008,0,__PYX_ERR(1, 2008, __pyx_L1_error))
  __pyx_v_w137 = (__pyx_v_w133 * __pyx_v_w136);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2009
 *         cdef double w136 = w135**(-1)
 *         cdef double w137 = w133*w136
 *         cdef double w138 = w135**(-2)             # <<<<<<<<<<<<<<
 *         cdef double w139 = w10*w133
 *         cdef double w140 = w134 + w139 + w86
 */
  __Pyx_TraceLine(2009,0,__PYX_ERR(1, 2009, __pyx_L1_error))
  __pyx_v_w138 = pow(__pyx_v_w135, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2010
 *         cdef double w137 = w133*w136
 *         cdef double w138 = w135**(-2)
 *         cdef double w139 = w10*w133             # <<<<<<<<<<<<<<
 *         cdef double w140 = w134 + w139 + w86
 *         cdef double w141 = sqrt(r*w127)
 */
  __Pyx_TraceLine(2010,0,__PYX_ERR(1, 2010, __pyx_L1_error))
  __pyx_v_w139 = (__pyx_v_w10 * __pyx_v_w133);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2011
 *         cdef double w138 = w135**(-2)
 *         cdef double w139 = w10*w133
 *         cdef double w140 = w134 + w139 + w86             # <<<<<<<<<<<<<<
 *         cdef double w141 = sqrt(r*w127)
 *         cdef double w142 = w1 + w10
 */
  __Pyx_TraceLine(2011,0,__PYX_ERR(1, 2011, __pyx_L1_error))
  __pyx_v_w140 = ((__pyx_v_w134 + __pyx_v_w139) + __pyx_v_w86);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2012
 *         cdef double w139 = w10*w133
 *         cdef double w140 = w134 + w139 + w86
 *         cdef double w141 = sqrt(r*w127)             # <<<<<<<<<<<<<<
 *         cdef double w142 = w1 + w10
 *         cdef double w143 = w141/w142
 */
  __Pyx_TraceLine(2012,0,__PYX_ERR(1, 2012, __pyx_L1_error))
  __pyx_v_w141 = sqrt((__pyx_v_r * __pyx_v_w127));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2013
 *         cdef double w140 = w134 + w139 + w86
 *         cdef double w141 = sqrt(r*w127)
 *         cdef double w142 = w1 + w10             # <<<<<<<<<<<<<<
 *         cdef double w143 = w141/w142
 *         cdef double w144 = w142**2
 */
  __Pyx_TraceLine(2013,0,__PYX_ERR(1, 2013, __pyx_L1_error))
  __pyx_v_w142 = (__pyx_v_w1 + __pyx_v_w10);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2014
 *         cdef double w141 = sqrt(r*w127)
 *         cdef double w142 = w1 + w10
 *         cdef double w143 = w141/w142             # <<<<<<<<<<<<<<
 *         cdef double w144 = w142**2
 *         cdef double w145 = 30720.0*w119*w17 + w121*w75 - w89 + w91
 */
  __Pyx_TraceLine(2014,0,__PYX_ERR(1, 2014, __pyx_L1_error))
  __pyx_v_w143 = (__pyx_v_w141 / __pyx_v_w142);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2015
 *         cdef double w142 = w1 + w10
 *         cdef double w143 = w141/w142
 *         cdef double w144 = w142**2             # <<<<<<<<<<<<<<
 *         cdef double w145 = 30720.0*w119*w17 + w121*w75 - w89 + w91
 *         cdef double w146 = prst**4
 */
  __Pyx_TraceLine(2015,0,__PYX_ERR(1, 2015, __pyx_L1_error))
  __pyx_v_w144 = pow(__pyx_v_w142, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2016
 *         cdef double w143 = w141/w142
 *         cdef double w144 = w142**2
 *         cdef double w145 = 30720.0*w119*w17 + w121*w75 - w89 + w91             # <<<<<<<<<<<<<<
 *         cdef double w146 = prst**4
 *         cdef double w147 = r**(-4.5)
 */
  __Pyx_TraceLine(2016,0,__PYX_ERR(1, 2016, __pyx_L1_error))
  __pyx_v_w145 = (((((30720.0 * __pyx_v_w119) * __pyx_v_w17) + (__pyx_v_w121 * __pyx_v_w75)) - __pyx_v_w89) + __pyx_v_w91);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2017
 *         cdef double w144 = w142**2
 *         cdef double w145 = 30720.0*w119*w17 + w121*w75 - w89 + w91
 *         cdef double w146 = prst**4             # <<<<<<<<<<<<<<
 *         cdef double w147 = r**(-4.5)
 *         cdef double w148 = prst**6
 */
  __Pyx_TraceLine(2017,0,__PYX_ERR(1, 2017, __pyx_L1_error))
  __pyx_v_w146 = pow(__pyx_v_prst, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2018
 *         cdef double w145 = 30720.0*w119*w17 + w121*w75 - w89 + w91
 *         cdef double w146 = prst**4
 *         cdef double w147 = r**(-4.5)             # <<<<<<<<<<<<<<
 *         cdef double w148 = prst**6
 *         cdef double w149 = r**(-3.5)
 */
  __Pyx_TraceLine(2018,0,__PYX_ERR(1, 2018, __pyx_L1_error))
  __pyx_v_w147 = pow(__pyx_v_r, -4.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2019
 *         cdef double w146 = prst**4
 *         cdef double w147 = r**(-4.5)
 *         cdef double w148 = prst**6             # <<<<<<<<<<<<<<
 *         cdef double w149 = r**(-3.5)
 *         cdef double w150 = prst**8
 */
  __Pyx_TraceLine(2019,0,__PYX_ERR(1, 2019, __pyx_L1_error))
  __pyx_v_w148 = pow(__pyx_v_prst, 6.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2020
 *         cdef double w147 = r**(-4.5)
 *         cdef double w148 = prst**6
 *         cdef double w149 = r**(-3.5)             # <<<<<<<<<<<<<<
 *         cdef double w150 = prst**8
 *         cdef double w151 = nu*w150
 */
  __Pyx_TraceLine(2020,0,__PYX_ERR(1, 2020, __pyx_L1_error))
  __pyx_v_w149 = pow(__pyx_v_r, -3.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2021
 *         cdef double w148 = prst**6
 *         cdef double w149 = r**(-3.5)
 *         cdef double w150 = prst**8             # <<<<<<<<<<<<<<
 *         cdef double w151 = nu*w150
 *         cdef double w152 = 8.0*nu - 6.0*w28
 */
  __Pyx_TraceLine(2021,0,__PYX_ERR(1, 2021, __pyx_L1_error))
  __pyx_v_w150 = pow(__pyx_v_prst, 8.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2022
 *         cdef double w149 = r**(-3.5)
 *         cdef double w150 = prst**8
 *         cdef double w151 = nu*w150             # <<<<<<<<<<<<<<
 *         cdef double w152 = 8.0*nu - 6.0*w28
 *         cdef double w153 = w146*w18
 */
  __Pyx_TraceLine(2022,0,__PYX_ERR(1, 2022, __pyx_L1_error))
  __pyx_v_w151 = (__pyx_v_nu * __pyx_v_w150);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2023
 *         cdef double w150 = prst**8
 *         cdef double w151 = nu*w150
 *         cdef double w152 = 8.0*nu - 6.0*w28             # <<<<<<<<<<<<<<
 *         cdef double w153 = w146*w18
 *         cdef double w154 = 92.7110442849544*nu - 131.0*w28 + 10.0*w56
 */
  __Pyx_TraceLine(2023,0,__PYX_ERR(1, 2023, __pyx_L1_error))
  __pyx_v_w152 = ((8.0 * __pyx_v_nu) - (6.0 * __pyx_v_w28));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2024
 *         cdef double w151 = nu*w150
 *         cdef double w152 = 8.0*nu - 6.0*w28
 *         cdef double w153 = w146*w18             # <<<<<<<<<<<<<<
 *         cdef double w154 = 92.7110442849544*nu - 131.0*w28 + 10.0*w56
 *         cdef double w155 = w14*w146
 */
  __Pyx_TraceLine(2024,0,__PYX_ERR(1, 2024, __pyx_L1_error))
  __pyx_v_w153 = (__pyx_v_w146 * __pyx_v_w18);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2025
 *         cdef double w152 = 8.0*nu - 6.0*w28
 *         cdef double w153 = w146*w18
 *         cdef double w154 = 92.7110442849544*nu - 131.0*w28 + 10.0*w56             # <<<<<<<<<<<<<<
 *         cdef double w155 = w14*w146
 *         cdef double w156 = -2.78300763695006*nu - 5.4*w28 + 6.0*w56
 */
  __Pyx_TraceLine(2025,0,__PYX_ERR(1, 2025, __pyx_L1_error))
  __pyx_v_w154 = (((92.7110442849544 * __pyx_v_nu) - (131.0 * __pyx_v_w28)) + (10.0 * __pyx_v_w56));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2026
 *         cdef double w153 = w146*w18
 *         cdef double w154 = 92.7110442849544*nu - 131.0*w28 + 10.0*w56
 *         cdef double w155 = w14*w146             # <<<<<<<<<<<<<<
 *         cdef double w156 = -2.78300763695006*nu - 5.4*w28 + 6.0*w56
 *         cdef double w157 = w148*w18
 */
  __Pyx_TraceLine(2026,0,__PYX_ERR(1, 2026, __pyx_L1_error))
  __pyx_v_w155 = (__pyx_v_w14 * __pyx_v_w146);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2027
 *         cdef double w154 = 92.7110442849544*nu - 131.0*w28 + 10.0*w56
 *         cdef double w155 = w14*w146
 *         cdef double w156 = -2.78300763695006*nu - 5.4*w28 + 6.0*w56             # <<<<<<<<<<<<<<
 *         cdef double w157 = w148*w18
 *         cdef double w158 = -33.9782122170436*nu - 89.5298327361234*w28 - 14.0*w53 + 188.0*w56
 */
  __Pyx_TraceLine(2027,0,__PYX_ERR(1, 2027, __pyx_L1_error))
  __pyx_v_w156 = (((-2.78300763695006 * __pyx_v_nu) - (5.4 * __pyx_v_w28)) + (6.0 * __pyx_v_w56));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2028
 *         cdef double w155 = w14*w146
 *         cdef double w156 = -2.78300763695006*nu - 5.4*w28 + 6.0*w56
 *         cdef double w157 = w148*w18             # <<<<<<<<<<<<<<
 *         cdef double w158 = -33.9782122170436*nu - 89.5298327361234*w28 - 14.0*w53 + 188.0*w56
 *         cdef double w159 = 3.0*w14
 */
  __Pyx_TraceLine(2028,0,__PYX_ERR(1, 2028, __pyx_L1_error))
  __pyx_v_w157 = (__pyx_v_w148 * __pyx_v_w18);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2029
 *         cdef double w156 = -2.78300763695006*nu - 5.4*w28 + 6.0*w56
 *         cdef double w157 = w148*w18
 *         cdef double w158 = -33.9782122170436*nu - 89.5298327361234*w28 - 14.0*w53 + 188.0*w56             # <<<<<<<<<<<<<<
 *         cdef double w159 = 3.0*w14
 *         cdef double w160 = 1.38977750996128*nu + 3.33842023648322*w28 - 6.0*w53 + 3.42857142857143*w56
 */
  __Pyx_TraceLine(2029,0,__PYX_ERR(1, 2029, __pyx_L1_error))
  __pyx_v_w158 = ((((-33.9782122170436 * __pyx_v_nu) - (89.5298327361234 * __pyx_v_w28)) - (14.0 * __pyx_v_w53)) + (188.0 * __pyx_v_w56));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2030
 *         cdef double w157 = w148*w18
 *         cdef double w158 = -33.9782122170436*nu - 89.5298327361234*w28 - 14.0*w53 + 188.0*w56
 *         cdef double w159 = 3.0*w14             # <<<<<<<<<<<<<<
 *         cdef double w160 = 1.38977750996128*nu + 3.33842023648322*w28 - 6.0*w53 + 3.42857142857143*w56
 *         cdef double w161 = w150*w160
 */
  __Pyx_TraceLine(2030,0,__PYX_ERR(1, 2030, __pyx_L1_error))
  __pyx_v_w159 = (3.0 * __pyx_v_w14);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2031
 *         cdef double w158 = -33.9782122170436*nu - 89.5298327361234*w28 - 14.0*w53 + 188.0*w56
 *         cdef double w159 = 3.0*w14
 *         cdef double w160 = 1.38977750996128*nu + 3.33842023648322*w28 - 6.0*w53 + 3.42857142857143*w56             # <<<<<<<<<<<<<<
 *         cdef double w161 = w150*w160
 *         cdef double w162 = nu*(452.542166996693 - 51.6952380952381*w49) + w28*(118.4*w49 - 1796.13660498019) + 602.318540416564*w56
 */
  __Pyx_TraceLine(2031,0,__PYX_ERR(1, 2031, __pyx_L1_error))
  __pyx_v_w160 = ((((1.38977750996128 * __pyx_v_nu) + (3.33842023648322 * __pyx_v_w28)) - (6.0 * __pyx_v_w53)) + (3.42857142857143 * __pyx_v_w56));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2032
 *         cdef double w159 = 3.0*w14
 *         cdef double w160 = 1.38977750996128*nu + 3.33842023648322*w28 - 6.0*w53 + 3.42857142857143*w56
 *         cdef double w161 = w150*w160             # <<<<<<<<<<<<<<
 *         cdef double w162 = nu*(452.542166996693 - 51.6952380952381*w49) + w28*(118.4*w49 - 1796.13660498019) + 602.318540416564*w56
 *         cdef double w163 = 4.0*w162
 */
  __Pyx_TraceLine(2032,0,__PYX_ERR(1, 2032, __pyx_L1_error))
  __pyx_v_w161 = (__pyx_v_w150 * __pyx_v_w160);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2033
 *         cdef double w160 = 1.38977750996128*nu + 3.33842023648322*w28 - 6.0*w53 + 3.42857142857143*w56
 *         cdef double w161 = w150*w160
 *         cdef double w162 = nu*(452.542166996693 - 51.6952380952381*w49) + w28*(118.4*w49 - 1796.13660498019) + 602.318540416564*w56             # <<<<<<<<<<<<<<
 *         cdef double w163 = 4.0*w162
 *         cdef double w164 = w142**4
 */
  __Pyx_TraceLine(2033,0,__PYX_ERR(1, 2033, __pyx_L1_error))
  __pyx_v_w162 = (((__pyx_v_nu * (452.542166996693 - (51.6952380952381 * __pyx_v_w49))) + (__pyx_v_w28 * ((118.4 * __pyx_v_w49) - 1796.13660498019))) + (602.318540416564 * __pyx_v_w56));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2034
 *         cdef double w161 = w150*w160
 *         cdef double w162 = nu*(452.542166996693 - 51.6952380952381*w49) + w28*(118.4*w49 - 1796.13660498019) + 602.318540416564*w56
 *         cdef double w163 = 4.0*w162             # <<<<<<<<<<<<<<
 *         cdef double w164 = w142**4
 *         cdef double w165 = 0.000130208333333333*w11 + w76
 */
  __Pyx_TraceLine(2034,0,__PYX_ERR(1, 2034, __pyx_L1_error))
  __pyx_v_w163 = (4.0 * __pyx_v_w162);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2035
 *         cdef double w162 = nu*(452.542166996693 - 51.6952380952381*w49) + w28*(118.4*w49 - 1796.13660498019) + 602.318540416564*w56
 *         cdef double w163 = 4.0*w162
 *         cdef double w164 = w142**4             # <<<<<<<<<<<<<<
 *         cdef double w165 = 0.000130208333333333*w11 + w76
 *         cdef double w166 = w165**(-4)
 */
  __Pyx_TraceLine(2035,0,__PYX_ERR(1, 2035, __pyx_L1_error))
  __pyx_v_w164 = pow(__pyx_v_w142, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2036
 *         cdef double w163 = 4.0*w162
 *         cdef double w164 = w142**4
 *         cdef double w165 = 0.000130208333333333*w11 + w76             # <<<<<<<<<<<<<<
 *         cdef double w166 = w165**(-4)
 *         cdef double w167 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0185696317637669*w1 + 0.0385795738434214*w111 + 0.00662650629087394*w113 + 0.00941289164152486*w122 - 1.18824456940711e-6*w123 + 0.000486339502879429*w124 - 3.63558293513537e-15*w125 + 0.291062041428379*w28 + 0.0210425293255724*w54 + 0.0244692826489756*w56 + 1
 */
  __Pyx_TraceLine(2036,0,__PYX_ERR(1, 2036, __pyx_L1_error))
  __pyx_v_w165 = ((0.000130208333333333 * __pyx_v_w11) + __pyx_v_w76);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2037
 *         cdef double w164 = w142**4
 *         cdef double w165 = 0.000130208333333333*w11 + w76
 *         cdef double w166 = w165**(-4)             # <<<<<<<<<<<<<<
 *         cdef double w167 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0185696317637669*w1 + 0.0385795738434214*w111 + 0.00662650629087394*w113 + 0.00941289164152486*w122 - 1.18824456940711e-6*w123 + 0.000486339502879429*w124 - 3.63558293513537e-15*w125 + 0.291062041428379*w28 + 0.0210425293255724*w54 + 0.0244692826489756*w56 + 1
 *         cdef double w168 = w167**(-2)
 */
  __Pyx_TraceLine(2037,0,__PYX_ERR(1, 2037, __pyx_L1_error))
  __pyx_v_w166 = pow(__pyx_v_w165, -4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2038
 *         cdef double w165 = 0.000130208333333333*w11 + w76
 *         cdef double w166 = w165**(-4)
 *         cdef double w167 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0185696317637669*w1 + 0.0385795738434214*w111 + 0.00662650629087394*w113 + 0.00941289164152486*w122 - 1.18824456940711e-6*w123 + 0.000486339502879429*w124 - 3.63558293513537e-15*w125 + 0.291062041428379*w28 + 0.0210425293255724*w54 + 0.0244692826489756*w56 + 1             # <<<<<<<<<<<<<<
 *         cdef double w168 = w167**(-2)
 *         cdef double w169 = 0.3125*delta*w36*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*w28 - 5.0)*(5.0*w4 + 5.0*w7 + 5.0*w8) - 0.03125*(-5.0*nu + w95 + 1.0)*(-15.0*w4 + 15.0*w7 + 15.0*w8)
 */
  __Pyx_TraceLine(2038,0,__PYX_ERR(1, 2038, __pyx_L1_error))
  __pyx_v_w167 = (((((((((((((-0.0438084424460039 * __pyx_v_nu) - (0.143521050466841 * __pyx_v_r)) + (0.0185696317637669 * __pyx_v_w1)) + (0.0385795738434214 * __pyx_v_w111)) + (0.00662650629087394 * __pyx_v_w113)) + (0.00941289164152486 * __pyx_v_w122)) - (1.18824456940711e-6 * __pyx_v_w123)) + (0.000486339502879429 * __pyx_v_w124)) - (3.63558293513537e-15 * __pyx_v_w125)) + (0.291062041428379 * __pyx_v_w28)) + (0.0210425293255724 * __pyx_v_w54)) + (0.0244692826489756 * __pyx_v_w56)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2039
 *         cdef double w166 = w165**(-4)
 *         cdef double w167 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0185696317637669*w1 + 0.0385795738434214*w111 + 0.00662650629087394*w113 + 0.00941289164152486*w122 - 1.18824456940711e-6*w123 + 0.000486339502879429*w124 - 3.63558293513537e-15*w125 + 0.291062041428379*w28 + 0.0210425293255724*w54 + 0.0244692826489756*w56 + 1
 *         cdef double w168 = w167**(-2)             # <<<<<<<<<<<<<<
 *         cdef double w169 = 0.3125*delta*w36*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*w28 - 5.0)*(5.0*w4 + 5.0*w7 + 5.0*w8) - 0.03125*(-5.0*nu + w95 + 1.0)*(-15.0*w4 + 15.0*w7 + 15.0*w8)
 *         cdef double w170 = w129*w166*w168*w169
 */
  __Pyx_TraceLine(2039,0,__PYX_ERR(1, 2039, __pyx_L1_error))
  __pyx_v_w168 = pow(__pyx_v_w167, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2040
 *         cdef double w167 = -0.0438084424460039*nu - 0.143521050466841*r + 0.0185696317637669*w1 + 0.0385795738434214*w111 + 0.00662650629087394*w113 + 0.00941289164152486*w122 - 1.18824456940711e-6*w123 + 0.000486339502879429*w124 - 3.63558293513537e-15*w125 + 0.291062041428379*w28 + 0.0210425293255724*w54 + 0.0244692826489756*w56 + 1
 *         cdef double w168 = w167**(-2)
 *         cdef double w169 = 0.3125*delta*w36*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*w28 - 5.0)*(5.0*w4 + 5.0*w7 + 5.0*w8) - 0.03125*(-5.0*nu + w95 + 1.0)*(-15.0*w4 + 15.0*w7 + 15.0*w8)             # <<<<<<<<<<<<<<
 *         cdef double w170 = w129*w166*w168*w169
 *         cdef double w171 = w146*w164*w170
 */
  __Pyx_TraceLine(2040,0,__PYX_ERR(1, 2040, __pyx_L1_error))
  __pyx_v_w169 = (((((0.3125 * __pyx_v_delta) * __pyx_v_w36) * ((18.0 * __pyx_v_nu) - 1.0)) - ((0.03125 * (((-33.0 * __pyx_v_nu) + (32.0 * __pyx_v_w28)) - 5.0)) * (((5.0 * __pyx_v_w4) + (5.0 * __pyx_v_w7)) + (5.0 * __pyx_v_w8)))) - ((0.03125 * (((-5.0 * __pyx_v_nu) + __pyx_v_w95) + 1.0)) * (((-15.0 * __pyx_v_w4) + (15.0 * __pyx_v_w7)) + (15.0 * __pyx_v_w8))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2041
 *         cdef double w168 = w167**(-2)
 *         cdef double w169 = 0.3125*delta*w36*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*w28 - 5.0)*(5.0*w4 + 5.0*w7 + 5.0*w8) - 0.03125*(-5.0*nu + w95 + 1.0)*(-15.0*w4 + 15.0*w7 + 15.0*w8)
 *         cdef double w170 = w129*w166*w168*w169             # <<<<<<<<<<<<<<
 *         cdef double w171 = w146*w164*w170
 *         cdef double w172 = r**(-13)
 */
  __Pyx_TraceLine(2041,0,__PYX_ERR(1, 2041, __pyx_L1_error))
  __pyx_v_w170 = (((__pyx_v_w129 * __pyx_v_w166) * __pyx_v_w168) * __pyx_v_w169);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2042
 *         cdef double w169 = 0.3125*delta*w36*(18.0*nu - 1.0) - 0.03125*(-33.0*nu + 32.0*w28 - 5.0)*(5.0*w4 + 5.0*w7 + 5.0*w8) - 0.03125*(-5.0*nu + w95 + 1.0)*(-15.0*w4 + 15.0*w7 + 15.0*w8)
 *         cdef double w170 = w129*w166*w168*w169
 *         cdef double w171 = w146*w164*w170             # <<<<<<<<<<<<<<
 *         cdef double w172 = r**(-13)
 *         cdef double w173 = w164*w172
 */
  __Pyx_TraceLine(2042,0,__PYX_ERR(1, 2042, __pyx_L1_error))
  __pyx_v_w171 = ((__pyx_v_w146 * __pyx_v_w164) * __pyx_v_w170);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2043
 *         cdef double w170 = w129*w166*w168*w169
 *         cdef double w171 = w146*w164*w170
 *         cdef double w172 = r**(-13)             # <<<<<<<<<<<<<<
 *         cdef double w173 = w164*w172
 *         cdef double w174 = 663.496888455656*nu*r**(-5.5)*w146 - 39.6112800271521*nu*w147*w148 - 7.59859378406358e-45*w109*w130*w146*w164*w166*w168*w169*w172 + 9.25454462627843e-34*w117*w129*w146*w166*w169*w173/w167**3 + 6.62902677807736e-23*w129*w145*w146*w168*w169*w173/w165**5 - w14*w146*(-51.6952380952381*w83 + 118.4*w85) + w146*w163*w20 + w148*w158*w159 + 3.70688355060912*w149*w151 + 0.121954868780449*w151*w2 + 2.0*w152*w153 + 3.0*w154*w155 + 2.0*w156*w157 + 2.0*w161*w18 - 1.01821851311268e-18*w129*w142**3*w146*w166*w168*w169/r**12 + 1.65460508380811e-18*w171/r**14
 */
  __Pyx_TraceLine(2043,0,__PYX_ERR(1, 2043, __pyx_L1_error))
  __pyx_v_w172 = pow(__pyx_v_r, -13.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2044
 *         cdef double w171 = w146*w164*w170
 *         cdef double w172 = r**(-13)
 *         cdef double w173 = w164*w172             # <<<<<<<<<<<<<<
 *         cdef double w174 = 663.496888455656*nu*r**(-5.5)*w146 - 39.6112800271521*nu*w147*w148 - 7.59859378406358e-45*w109*w130*w146*w164*w166*w168*w169*w172 + 9.25454462627843e-34*w117*w129*w146*w166*w169*w173/w167**3 + 6.62902677807736e-23*w129*w145*w146*w168*w169*w173/w165**5 - w14*w146*(-51.6952380952381*w83 + 118.4*w85) + w146*w163*w20 + w148*w158*w159 + 3.70688355060912*w149*w151 + 0.121954868780449*w151*w2 + 2.0*w152*w153 + 3.0*w154*w155 + 2.0*w156*w157 + 2.0*w161*w18 - 1.01821851311268e-18*w129*w142**3*w146*w166*w168*w169/r**12 + 1.65460508380811e-18*w171/r**14
 *         cdef double w175 = prst**3
 */
  __Pyx_TraceLine(2044,0,__PYX_ERR(1, 2044, __pyx_L1_error))
  __pyx_v_w173 = (__pyx_v_w164 * __pyx_v_w172);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2045
 *         cdef double w172 = r**(-13)
 *         cdef double w173 = w164*w172
 *         cdef double w174 = 663.496888455656*nu*r**(-5.5)*w146 - 39.6112800271521*nu*w147*w148 - 7.59859378406358e-45*w109*w130*w146*w164*w166*w168*w169*w172 + 9.25454462627843e-34*w117*w129*w146*w166*w169*w173/w167**3 + 6.62902677807736e-23*w129*w145*w146*w168*w169*w173/w165**5 - w14*w146*(-51.6952380952381*w83 + 118.4*w85) + w146*w163*w20 + w148*w158*w159 + 3.70688355060912*w149*w151 + 0.121954868780449*w151*w2 + 2.0*w152*w153 + 3.0*w154*w155 + 2.0*w156*w157 + 2.0*w161*w18 - 1.01821851311268e-18*w129*w142**3*w146*w166*w168*w169/r**12 + 1.65460508380811e-18*w171/r**14             # <<<<<<<<<<<<<<
 *         cdef double w175 = prst**3
 *         cdef double w176 = nu*w147
 */
  __Pyx_TraceLine(2045,0,__PYX_ERR(1, 2045, __pyx_L1_error))
  __pyx_v_w174 = ((((((((((((((((((663.496888455656 * __pyx_v_nu) * pow(__pyx_v_r, -5.5)) * __pyx_v_w146) - (((39.6112800271521 * __pyx_v_nu) * __pyx_v_w147) * __pyx_v_w148)) - ((((((((7.59859378406358e-45 * __pyx_v_w109) * __pyx_v_w130) * __pyx_v_w146) * __pyx_v_w164) * __pyx_v_w166) * __pyx_v_w168) * __pyx_v_w169) * __pyx_v_w172)) + (((((((9.25454462627843e-34 * __pyx_v_w117) * __pyx_v_w129) * __pyx_v_w146) * __pyx_v_w166) * __pyx_v_w169) * __pyx_v_w173) / pow(__pyx_v_w167, 3.0))) + (((((((6.62902677807736e-23 * __pyx_v_w129) * __pyx_v_w145) * __pyx_v_w146) * __pyx_v_w168) * __pyx_v_w169) * __pyx_v_w173) / pow(__pyx_v_w165, 5.0))) - ((__pyx_v_w14 * __pyx_v_w146) * ((-51.6952380952381 * __pyx_v_w83) + (118.4 * __pyx_v_w85)))) + ((__pyx_v_w146 * __pyx_v_w163) * __pyx_v_w20)) + ((__pyx_v_w148 * __pyx_v_w158) * __pyx_v_w159)) + ((3.70688355060912 * __pyx_v_w149) * __pyx_v_w151)) + ((0.121954868780449 * __pyx_v_w151) * __pyx_v_w2)) + ((2.0 * __pyx_v_w152) * __pyx_v_w153)) + ((3.0 * __pyx_v_w154) * __pyx_v_w155)) + ((2.0 * __pyx_v_w156) * __pyx_v_w157)) + ((2.0 * __pyx_v_w161) * __pyx_v_w18)) - (((((((1.01821851311268e-18 * __pyx_v_w129) * pow(__pyx_v_w142, 3.0)) * __pyx_v_w146) * __pyx_v_w166) * __pyx_v_w168) * __pyx_v_w169) / pow(__pyx_v_r, 12.0))) + ((1.65460508380811e-18 * __pyx_v_w171) / pow(__pyx_v_r, 14.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2046
 *         cdef double w173 = w164*w172
 *         cdef double w174 = 663.496888455656*nu*r**(-5.5)*w146 - 39.6112800271521*nu*w147*w148 - 7.59859378406358e-45*w109*w130*w146*w164*w166*w168*w169*w172 + 9.25454462627843e-34*w117*w129*w146*w166*w169*w173/w167**3 + 6.62902677807736e-23*w129*w145*w146*w168*w169*w173/w165**5 - w14*w146*(-51.6952380952381*w83 + 118.4*w85) + w146*w163*w20 + w148*w158*w159 + 3.70688355060912*w149*w151 + 0.121954868780449*w151*w2 + 2.0*w152*w153 + 3.0*w154*w155 + 2.0*w156*w157 + 2.0*w161*w18 - 1.01821851311268e-18*w129*w142**3*w146*w166*w168*w169/r**12 + 1.65460508380811e-18*w171/r**14
 *         cdef double w175 = prst**3             # <<<<<<<<<<<<<<
 *         cdef double w176 = nu*w147
 *         cdef double w177 = prst**5
 */
  __Pyx_TraceLine(2046,0,__PYX_ERR(1, 2046, __pyx_L1_error))
  __pyx_v_w175 = pow(__pyx_v_prst, 3.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2047
 *         cdef double w174 = 663.496888455656*nu*r**(-5.5)*w146 - 39.6112800271521*nu*w147*w148 - 7.59859378406358e-45*w109*w130*w146*w164*w166*w168*w169*w172 + 9.25454462627843e-34*w117*w129*w146*w166*w169*w173/w167**3 + 6.62902677807736e-23*w129*w145*w146*w168*w169*w173/w165**5 - w14*w146*(-51.6952380952381*w83 + 118.4*w85) + w146*w163*w20 + w148*w158*w159 + 3.70688355060912*w149*w151 + 0.121954868780449*w151*w2 + 2.0*w152*w153 + 3.0*w154*w155 + 2.0*w156*w157 + 2.0*w161*w18 - 1.01821851311268e-18*w129*w142**3*w146*w166*w168*w169/r**12 + 1.65460508380811e-18*w171/r**14
 *         cdef double w175 = prst**3
 *         cdef double w176 = nu*w147             # <<<<<<<<<<<<<<
 *         cdef double w177 = prst**5
 *         cdef double w178 = nu*w149
 */
  __Pyx_TraceLine(2047,0,__PYX_ERR(1, 2047, __pyx_L1_error))
  __pyx_v_w176 = (__pyx_v_nu * __pyx_v_w147);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2048
 *         cdef double w175 = prst**3
 *         cdef double w176 = nu*w147
 *         cdef double w177 = prst**5             # <<<<<<<<<<<<<<
 *         cdef double w178 = nu*w149
 *         cdef double w179 = r**(-2.5)
 */
  __Pyx_TraceLine(2048,0,__PYX_ERR(1, 2048, __pyx_L1_error))
  __pyx_v_w177 = pow(__pyx_v_prst, 5.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2049
 *         cdef double w176 = nu*w147
 *         cdef double w177 = prst**5
 *         cdef double w178 = nu*w149             # <<<<<<<<<<<<<<
 *         cdef double w179 = r**(-2.5)
 *         cdef double w180 = prst**7
 */
  __Pyx_TraceLine(2049,0,__PYX_ERR(1, 2049, __pyx_L1_error))
  __pyx_v_w178 = (__pyx_v_nu * __pyx_v_w149);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2050
 *         cdef double w177 = prst**5
 *         cdef double w178 = nu*w149
 *         cdef double w179 = r**(-2.5)             # <<<<<<<<<<<<<<
 *         cdef double w180 = prst**7
 *         cdef double w181 = w152*w2
 */
  __Pyx_TraceLine(2050,0,__PYX_ERR(1, 2050, __pyx_L1_error))
  __pyx_v_w179 = pow(__pyx_v_r, -2.5);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2051
 *         cdef double w178 = nu*w149
 *         cdef double w179 = r**(-2.5)
 *         cdef double w180 = prst**7             # <<<<<<<<<<<<<<
 *         cdef double w181 = w152*w2
 *         cdef double w182 = 4.0*w175
 */
  __Pyx_TraceLine(2051,0,__PYX_ERR(1, 2051, __pyx_L1_error))
  __pyx_v_w180 = pow(__pyx_v_prst, 7.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2052
 *         cdef double w179 = r**(-2.5)
 *         cdef double w180 = prst**7
 *         cdef double w181 = w152*w2             # <<<<<<<<<<<<<<
 *         cdef double w182 = 4.0*w175
 *         cdef double w183 = w156*w2
 */
  __Pyx_TraceLine(2052,0,__PYX_ERR(1, 2052, __pyx_L1_error))
  __pyx_v_w181 = (__pyx_v_w152 * __pyx_v_w2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2053
 *         cdef double w180 = prst**7
 *         cdef double w181 = w152*w2
 *         cdef double w182 = 4.0*w175             # <<<<<<<<<<<<<<
 *         cdef double w183 = w156*w2
 *         cdef double w184 = 6.0*w177
 */
  __Pyx_TraceLine(2053,0,__PYX_ERR(1, 2053, __pyx_L1_error))
  __pyx_v_w182 = (4.0 * __pyx_v_w175);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2054
 *         cdef double w181 = w152*w2
 *         cdef double w182 = 4.0*w175
 *         cdef double w183 = w156*w2             # <<<<<<<<<<<<<<
 *         cdef double w184 = 6.0*w177
 *         cdef double w185 = w139 + w17
 */
  __Pyx_TraceLine(2054,0,__PYX_ERR(1, 2054, __pyx_L1_error))
  __pyx_v_w183 = (__pyx_v_w156 * __pyx_v_w2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2055
 *         cdef double w182 = 4.0*w175
 *         cdef double w183 = w156*w2
 *         cdef double w184 = 6.0*w177             # <<<<<<<<<<<<<<
 *         cdef double w185 = w139 + w17
 *         cdef double w186 = 3.0*nu
 */
  __Pyx_TraceLine(2055,0,__PYX_ERR(1, 2055, __pyx_L1_error))
  __pyx_v_w184 = (6.0 * __pyx_v_w177);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2056
 *         cdef double w183 = w156*w2
 *         cdef double w184 = 6.0*w177
 *         cdef double w185 = w139 + w17             # <<<<<<<<<<<<<<
 *         cdef double w186 = 3.0*nu
 *         cdef double w187 = X_1*chiL1
 */
  __Pyx_TraceLine(2056,0,__PYX_ERR(1, 2056, __pyx_L1_error))
  __pyx_v_w185 = (__pyx_v_w139 + __pyx_v_w17);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2057
 *         cdef double w184 = 6.0*w177
 *         cdef double w185 = w139 + w17
 *         cdef double w186 = 3.0*nu             # <<<<<<<<<<<<<<
 *         cdef double w187 = X_1*chiL1
 *         cdef double w188 = X_2*chiL2
 */
  __Pyx_TraceLine(2057,0,__PYX_ERR(1, 2057, __pyx_L1_error))
  __pyx_v_w186 = (3.0 * __pyx_v_nu);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2058
 *         cdef double w185 = w139 + w17
 *         cdef double w186 = 3.0*nu
 *         cdef double w187 = X_1*chiL1             # <<<<<<<<<<<<<<
 *         cdef double w188 = X_2*chiL2
 *         cdef double w189 = w187 + w188
 */
  __Pyx_TraceLine(2058,0,__PYX_ERR(1, 2058, __pyx_L1_error))
  __pyx_v_w187 = (__pyx_v_X_1 * __pyx_v_chiL1);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2059
 *         cdef double w186 = 3.0*nu
 *         cdef double w187 = X_1*chiL1
 *         cdef double w188 = X_2*chiL2             # <<<<<<<<<<<<<<
 *         cdef double w189 = w187 + w188
 *         cdef double w190 = pphi*w189
 */
  __Pyx_TraceLine(2059,0,__PYX_ERR(1, 2059, __pyx_L1_error))
  __pyx_v_w188 = (__pyx_v_X_2 * __pyx_v_chiL2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2060
 *         cdef double w187 = X_1*chiL1
 *         cdef double w188 = X_2*chiL2
 *         cdef double w189 = w187 + w188             # <<<<<<<<<<<<<<
 *         cdef double w190 = pphi*w189
 *         cdef double w191 = dSO*w190
 */
  __Pyx_TraceLine(2060,0,__PYX_ERR(1, 2060, __pyx_L1_error))
  __pyx_v_w189 = (__pyx_v_w187 + __pyx_v_w188);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2061
 *         cdef double w188 = X_2*chiL2
 *         cdef double w189 = w187 + w188
 *         cdef double w190 = pphi*w189             # <<<<<<<<<<<<<<
 *         cdef double w191 = dSO*w190
 *         cdef double w192 = 0.71875*nu - 0.09375
 */
  __Pyx_TraceLine(2061,0,__PYX_ERR(1, 2061, __pyx_L1_error))
  __pyx_v_w190 = (__pyx_v_pphi * __pyx_v_w189);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2062
 *         cdef double w189 = w187 + w188
 *         cdef double w190 = pphi*w189
 *         cdef double w191 = dSO*w190             # <<<<<<<<<<<<<<
 *         cdef double w192 = 0.71875*nu - 0.09375
 *         cdef double w193 = -1.40625*nu - 0.46875
 */
  __Pyx_TraceLine(2062,0,__PYX_ERR(1, 2062, __pyx_L1_error))
  __pyx_v_w191 = (__pyx_v_dSO * __pyx_v_w190);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2063
 *         cdef double w190 = pphi*w189
 *         cdef double w191 = dSO*w190
 *         cdef double w192 = 0.71875*nu - 0.09375             # <<<<<<<<<<<<<<
 *         cdef double w193 = -1.40625*nu - 0.46875
 *         cdef double w194 = pphi**2
 */
  __Pyx_TraceLine(2063,0,__PYX_ERR(1, 2063, __pyx_L1_error))
  __pyx_v_w192 = ((0.71875 * __pyx_v_nu) - 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2064
 *         cdef double w191 = dSO*w190
 *         cdef double w192 = 0.71875*nu - 0.09375
 *         cdef double w193 = -1.40625*nu - 0.46875             # <<<<<<<<<<<<<<
 *         cdef double w194 = pphi**2
 *         cdef double w195 = w18*w194
 */
  __Pyx_TraceLine(2064,0,__PYX_ERR(1, 2064, __pyx_L1_error))
  __pyx_v_w193 = ((-1.40625 * __pyx_v_nu) - 0.46875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2065
 *         cdef double w192 = 0.71875*nu - 0.09375
 *         cdef double w193 = -1.40625*nu - 0.46875
 *         cdef double w194 = pphi**2             # <<<<<<<<<<<<<<
 *         cdef double w195 = w18*w194
 *         cdef double w196 = 2.0*w195
 */
  __Pyx_TraceLine(2065,0,__PYX_ERR(1, 2065, __pyx_L1_error))
  __pyx_v_w194 = pow(__pyx_v_pphi, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2066
 *         cdef double w193 = -1.40625*nu - 0.46875
 *         cdef double w194 = pphi**2
 *         cdef double w195 = w18*w194             # <<<<<<<<<<<<<<
 *         cdef double w196 = 2.0*w195
 *         cdef double w197 = -2.0859375*nu - 2.07161458333333*w28 + 0.23046875
 */
  __Pyx_TraceLine(2066,0,__PYX_ERR(1, 2066, __pyx_L1_error))
  __pyx_v_w195 = (__pyx_v_w18 * __pyx_v_w194);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2067
 *         cdef double w194 = pphi**2
 *         cdef double w195 = w18*w194
 *         cdef double w196 = 2.0*w195             # <<<<<<<<<<<<<<
 *         cdef double w197 = -2.0859375*nu - 2.07161458333333*w28 + 0.23046875
 *         cdef double w198 = w159*w194
 */
  __Pyx_TraceLine(2067,0,__PYX_ERR(1, 2067, __pyx_L1_error))
  __pyx_v_w196 = (2.0 * __pyx_v_w195);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2068
 *         cdef double w195 = w18*w194
 *         cdef double w196 = 2.0*w195
 *         cdef double w197 = -2.0859375*nu - 2.07161458333333*w28 + 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double w198 = w159*w194
 *         cdef double w199 = 0.5859375*nu + 1.34765625*w28 + 0.41015625
 */
  __Pyx_TraceLine(2068,0,__PYX_ERR(1, 2068, __pyx_L1_error))
  __pyx_v_w197 = (((-2.0859375 * __pyx_v_nu) - (2.07161458333333 * __pyx_v_w28)) + 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2069
 *         cdef double w196 = 2.0*w195
 *         cdef double w197 = -2.0859375*nu - 2.07161458333333*w28 + 0.23046875
 *         cdef double w198 = w159*w194             # <<<<<<<<<<<<<<
 *         cdef double w199 = 0.5859375*nu + 1.34765625*w28 + 0.41015625
 *         cdef double w200 = pphi**4
 */
  __Pyx_TraceLine(2069,0,__PYX_ERR(1, 2069, __pyx_L1_error))
  __pyx_v_w198 = (__pyx_v_w159 * __pyx_v_w194);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2070
 *         cdef double w197 = -2.0859375*nu - 2.07161458333333*w28 + 0.23046875
 *         cdef double w198 = w159*w194
 *         cdef double w199 = 0.5859375*nu + 1.34765625*w28 + 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double w200 = pphi**4
 *         cdef double w201 = 4.0*w20*w200
 */
  __Pyx_TraceLine(2070,0,__PYX_ERR(1, 2070, __pyx_L1_error))
  __pyx_v_w199 = (((0.5859375 * __pyx_v_nu) + (1.34765625 * __pyx_v_w28)) + 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2071
 *         cdef double w198 = w159*w194
 *         cdef double w199 = 0.5859375*nu + 1.34765625*w28 + 0.41015625
 *         cdef double w200 = pphi**4             # <<<<<<<<<<<<<<
 *         cdef double w201 = 4.0*w20*w200
 *         cdef double w202 = 0.34375*nu + 0.09375
 */
  __Pyx_TraceLine(2071,0,__PYX_ERR(1, 2071, __pyx_L1_error))
  __pyx_v_w200 = pow(__pyx_v_pphi, 4.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2072
 *         cdef double w199 = 0.5859375*nu + 1.34765625*w28 + 0.41015625
 *         cdef double w200 = pphi**4
 *         cdef double w201 = 4.0*w20*w200             # <<<<<<<<<<<<<<
 *         cdef double w202 = 0.34375*nu + 0.09375
 *         cdef double w203 = 0.46875 - 0.28125*nu
 */
  __Pyx_TraceLine(2072,0,__PYX_ERR(1, 2072, __pyx_L1_error))
  __pyx_v_w201 = ((4.0 * __pyx_v_w20) * __pyx_v_w200);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2073
 *         cdef double w200 = pphi**4
 *         cdef double w201 = 4.0*w20*w200
 *         cdef double w202 = 0.34375*nu + 0.09375             # <<<<<<<<<<<<<<
 *         cdef double w203 = 0.46875 - 0.28125*nu
 *         cdef double w204 = -0.2734375*nu - 0.798177083333333*w28 - 0.23046875
 */
  __Pyx_TraceLine(2073,0,__PYX_ERR(1, 2073, __pyx_L1_error))
  __pyx_v_w202 = ((0.34375 * __pyx_v_nu) + 0.09375);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2074
 *         cdef double w201 = 4.0*w20*w200
 *         cdef double w202 = 0.34375*nu + 0.09375
 *         cdef double w203 = 0.46875 - 0.28125*nu             # <<<<<<<<<<<<<<
 *         cdef double w204 = -0.2734375*nu - 0.798177083333333*w28 - 0.23046875
 *         cdef double w205 = -0.3515625*nu + 0.29296875*w28 - 0.41015625
 */
  __Pyx_TraceLine(2074,0,__PYX_ERR(1, 2074, __pyx_L1_error))
  __pyx_v_w203 = (0.46875 - (0.28125 * __pyx_v_nu));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2075
 *         cdef double w202 = 0.34375*nu + 0.09375
 *         cdef double w203 = 0.46875 - 0.28125*nu
 *         cdef double w204 = -0.2734375*nu - 0.798177083333333*w28 - 0.23046875             # <<<<<<<<<<<<<<
 *         cdef double w205 = -0.3515625*nu + 0.29296875*w28 - 0.41015625
 *         cdef double w206 = pphi*(w187 - w188)
 */
  __Pyx_TraceLine(2075,0,__PYX_ERR(1, 2075, __pyx_L1_error))
  __pyx_v_w204 = (((-0.2734375 * __pyx_v_nu) - (0.798177083333333 * __pyx_v_w28)) - 0.23046875);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2076
 *         cdef double w203 = 0.46875 - 0.28125*nu
 *         cdef double w204 = -0.2734375*nu - 0.798177083333333*w28 - 0.23046875
 *         cdef double w205 = -0.3515625*nu + 0.29296875*w28 - 0.41015625             # <<<<<<<<<<<<<<
 *         cdef double w206 = pphi*(w187 - w188)
 *         cdef double w207 = delta*w206
 */
  __Pyx_TraceLine(2076,0,__PYX_ERR(1, 2076, __pyx_L1_error))
  __pyx_v_w205 = (((-0.3515625 * __pyx_v_nu) + (0.29296875 * __pyx_v_w28)) - 0.41015625);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2077
 *         cdef double w204 = -0.2734375*nu - 0.798177083333333*w28 - 0.23046875
 *         cdef double w205 = -0.3515625*nu + 0.29296875*w28 - 0.41015625
 *         cdef double w206 = pphi*(w187 - w188)             # <<<<<<<<<<<<<<
 *         cdef double w207 = delta*w206
 *         cdef double w208 = delta*w194
 */
  __Pyx_TraceLine(2077,0,__PYX_ERR(1, 2077, __pyx_L1_error))
  __pyx_v_w206 = (__pyx_v_pphi * (__pyx_v_w187 - __pyx_v_w188));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2078
 *         cdef double w205 = -0.3515625*nu + 0.29296875*w28 - 0.41015625
 *         cdef double w206 = pphi*(w187 - w188)
 *         cdef double w207 = delta*w206             # <<<<<<<<<<<<<<
 *         cdef double w208 = delta*w194
 *         cdef double w209 = delta*w10
 */
  __Pyx_TraceLine(2078,0,__PYX_ERR(1, 2078, __pyx_L1_error))
  __pyx_v_w207 = (__pyx_v_delta * __pyx_v_w206);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2079
 *         cdef double w206 = pphi*(w187 - w188)
 *         cdef double w207 = delta*w206
 *         cdef double w208 = delta*w194             # <<<<<<<<<<<<<<
 *         cdef double w209 = delta*w10
 *         cdef double w210 = w31*(0.416666666666667 - 0.833333333333333*X_2)
 */
  __Pyx_TraceLine(2079,0,__PYX_ERR(1, 2079, __pyx_L1_error))
  __pyx_v_w208 = (__pyx_v_delta * __pyx_v_w194);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2080
 *         cdef double w207 = delta*w206
 *         cdef double w208 = delta*w194
 *         cdef double w209 = delta*w10             # <<<<<<<<<<<<<<
 *         cdef double w210 = w31*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double w211 = 0.5*w25*(0.5 - X_2) + 0.125*w30 - 0.125*w4 - 0.125*w7 - 0.125*w8
 */
  __Pyx_TraceLine(2080,0,__PYX_ERR(1, 2080, __pyx_L1_error))
  __pyx_v_w209 = (__pyx_v_delta * __pyx_v_w10);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2081
 *         cdef double w208 = delta*w194
 *         cdef double w209 = delta*w10
 *         cdef double w210 = w31*(0.416666666666667 - 0.833333333333333*X_2)             # <<<<<<<<<<<<<<
 *         cdef double w211 = 0.5*w25*(0.5 - X_2) + 0.125*w30 - 0.125*w4 - 0.125*w7 - 0.125*w8
 *         cdef double w212 = w194*w2
 */
  __Pyx_TraceLine(2081,0,__PYX_ERR(1, 2081, __pyx_L1_error))
  __pyx_v_w210 = (__pyx_v_w31 * (0.416666666666667 - (0.833333333333333 * __pyx_v_X_2)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2082
 *         cdef double w209 = delta*w10
 *         cdef double w210 = w31*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double w211 = 0.5*w25*(0.5 - X_2) + 0.125*w30 - 0.125*w4 - 0.125*w7 - 0.125*w8             # <<<<<<<<<<<<<<
 *         cdef double w212 = w194*w2
 *         cdef double w213 = w14*w200
 */
  __Pyx_TraceLine(2082,0,__PYX_ERR(1, 2082, __pyx_L1_error))
  __pyx_v_w211 = ((((((0.5 * __pyx_v_w25) * (0.5 - __pyx_v_X_2)) + (0.125 * __pyx_v_w30)) - (0.125 * __pyx_v_w4)) - (0.125 * __pyx_v_w7)) - (0.125 * __pyx_v_w8));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2083
 *         cdef double w210 = w31*(0.416666666666667 - 0.833333333333333*X_2)
 *         cdef double w211 = 0.5*w25*(0.5 - X_2) + 0.125*w30 - 0.125*w4 - 0.125*w7 - 0.125*w8
 *         cdef double w212 = w194*w2             # <<<<<<<<<<<<<<
 *         cdef double w213 = w14*w200
 *         cdef double w214 = 0.5*w25
 */
  __Pyx_TraceLine(2083,0,__PYX_ERR(1, 2083, __pyx_L1_error))
  __pyx_v_w212 = (__pyx_v_w194 * __pyx_v_w2);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2084
 *         cdef double w211 = 0.5*w25*(0.5 - X_2) + 0.125*w30 - 0.125*w4 - 0.125*w7 - 0.125*w8
 *         cdef double w212 = w194*w2
 *         cdef double w213 = w14*w200             # <<<<<<<<<<<<<<
 *         cdef double w214 = 0.5*w25
 *         cdef double w215 = 0.5*w31
 */
  __Pyx_TraceLine(2084,0,__PYX_ERR(1, 2084, __pyx_L1_error))
  __pyx_v_w213 = (__pyx_v_w14 * __pyx_v_w200);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2085
 *         cdef double w212 = w194*w2
 *         cdef double w213 = w14*w200
 *         cdef double w214 = 0.5*w25             # <<<<<<<<<<<<<<
 *         cdef double w215 = 0.5*w31
 *         cdef double w216 = 0.5*w34
 */
  __Pyx_TraceLine(2085,0,__PYX_ERR(1, 2085, __pyx_L1_error))
  __pyx_v_w214 = (0.5 * __pyx_v_w25);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2086
 *         cdef double w213 = w14*w200
 *         cdef double w214 = 0.5*w25
 *         cdef double w215 = 0.5*w31             # <<<<<<<<<<<<<<
 *         cdef double w216 = 0.5*w34
 *         cdef double w217 = w26*(98.0*nu + 43.0)
 */
  __Pyx_TraceLine(2086,0,__PYX_ERR(1, 2086, __pyx_L1_error))
  __pyx_v_w215 = (0.5 * __pyx_v_w31);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2087
 *         cdef double w214 = 0.5*w25
 *         cdef double w215 = 0.5*w31
 *         cdef double w216 = 0.5*w34             # <<<<<<<<<<<<<<
 *         cdef double w217 = w26*(98.0*nu + 43.0)
 *         cdef double w218 = -1610.0*nu + w29 + 375.0
 */
  __Pyx_TraceLine(2087,0,__PYX_ERR(1, 2087, __pyx_L1_error))
  __pyx_v_w216 = (0.5 * __pyx_v_w34);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2088
 *         cdef double w215 = 0.5*w31
 *         cdef double w216 = 0.5*w34
 *         cdef double w217 = w26*(98.0*nu + 43.0)             # <<<<<<<<<<<<<<
 *         cdef double w218 = -1610.0*nu + w29 + 375.0
 *         cdef double w219 = -1090.0*nu + 152.0*w28 + 219.0
 */
  __Pyx_TraceLine(2088,0,__PYX_ERR(1, 2088, __pyx_L1_error))
  __pyx_v_w217 = (__pyx_v_w26 * ((98.0 * __pyx_v_nu) + 43.0));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2089
 *         cdef double w216 = 0.5*w34
 *         cdef double w217 = w26*(98.0*nu + 43.0)
 *         cdef double w218 = -1610.0*nu + w29 + 375.0             # <<<<<<<<<<<<<<
 *         cdef double w219 = -1090.0*nu + 152.0*w28 + 219.0
 *         cdef double w220 = -w14*(w214*(2.25 - w81) + w215*(-5.25*nu - 2.8125) + w216*(0.5625 - 2.25*nu)) + w18*w31 - w20*(-0.25*w217 + 0.0104166666666667*w218*w31 + 0.03125*w219*w34)
 */
  __Pyx_TraceLine(2089,0,__PYX_ERR(1, 2089, __pyx_L1_error))
  __pyx_v_w218 = (((-1610.0 * __pyx_v_nu) + __pyx_v_w29) + 375.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2090
 *         cdef double w217 = w26*(98.0*nu + 43.0)
 *         cdef double w218 = -1610.0*nu + w29 + 375.0
 *         cdef double w219 = -1090.0*nu + 152.0*w28 + 219.0             # <<<<<<<<<<<<<<
 *         cdef double w220 = -w14*(w214*(2.25 - w81) + w215*(-5.25*nu - 2.8125) + w216*(0.5625 - 2.25*nu)) + w18*w31 - w20*(-0.25*w217 + 0.0104166666666667*w218*w31 + 0.03125*w219*w34)
 *         cdef double w221 = w189**2
 */
  __Pyx_TraceLine(2090,0,__PYX_ERR(1, 2090, __pyx_L1_error))
  __pyx_v_w219 = (((-1090.0 * __pyx_v_nu) + (152.0 * __pyx_v_w28)) + 219.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2091
 *         cdef double w218 = -1610.0*nu + w29 + 375.0
 *         cdef double w219 = -1090.0*nu + 152.0*w28 + 219.0
 *         cdef double w220 = -w14*(w214*(2.25 - w81) + w215*(-5.25*nu - 2.8125) + w216*(0.5625 - 2.25*nu)) + w18*w31 - w20*(-0.25*w217 + 0.0104166666666667*w218*w31 + 0.03125*w219*w34)             # <<<<<<<<<<<<<<
 *         cdef double w221 = w189**2
 *         cdef double w222 = w0*w194*w221
 */
  __Pyx_TraceLine(2091,0,__PYX_ERR(1, 2091, __pyx_L1_error))
  __pyx_v_w220 = ((((-__pyx_v_w14) * (((__pyx_v_w214 * (2.25 - __pyx_v_w81)) + (__pyx_v_w215 * ((-5.25 * __pyx_v_nu) - 2.8125))) + (__pyx_v_w216 * (0.5625 - (2.25 * __pyx_v_nu))))) + (__pyx_v_w18 * __pyx_v_w31)) - (__pyx_v_w20 * (((-0.25 * __pyx_v_w217) + ((0.0104166666666667 * __pyx_v_w218) * __pyx_v_w31)) + ((0.03125 * __pyx_v_w219) * __pyx_v_w34))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2092
 *         cdef double w219 = -1090.0*nu + 152.0*w28 + 219.0
 *         cdef double w220 = -w14*(w214*(2.25 - w81) + w215*(-5.25*nu - 2.8125) + w216*(0.5625 - 2.25*nu)) + w18*w31 - w20*(-0.25*w217 + 0.0104166666666667*w218*w31 + 0.03125*w219*w34)
 *         cdef double w221 = w189**2             # <<<<<<<<<<<<<<
 *         cdef double w222 = w0*w194*w221
 *         cdef double w223 = w14*(-0.0625*w217 + w218*w32 + w219*w35) + w18*(w214*(0.75 - 1.5*X_2) + w215*(-1.75*nu - 0.9375) + w216*(0.1875 - 0.75*nu)) - w2*w215 + 1.0
 */
  __Pyx_TraceLine(2092,0,__PYX_ERR(1, 2092, __pyx_L1_error))
  __pyx_v_w221 = pow(__pyx_v_w189, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2093
 *         cdef double w220 = -w14*(w214*(2.25 - w81) + w215*(-5.25*nu - 2.8125) + w216*(0.5625 - 2.25*nu)) + w18*w31 - w20*(-0.25*w217 + 0.0104166666666667*w218*w31 + 0.03125*w219*w34)
 *         cdef double w221 = w189**2
 *         cdef double w222 = w0*w194*w221             # <<<<<<<<<<<<<<
 *         cdef double w223 = w14*(-0.0625*w217 + w218*w32 + w219*w35) + w18*(w214*(0.75 - 1.5*X_2) + w215*(-1.75*nu - 0.9375) + w216*(0.1875 - 0.75*nu)) - w2*w215 + 1.0
 *         cdef double w224 = prst**2
 */
  __Pyx_TraceLine(2093,0,__PYX_ERR(1, 2093, __pyx_L1_error))
  __pyx_v_w222 = ((__pyx_v_w0 * __pyx_v_w194) * __pyx_v_w221);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2094
 *         cdef double w221 = w189**2
 *         cdef double w222 = w0*w194*w221
 *         cdef double w223 = w14*(-0.0625*w217 + w218*w32 + w219*w35) + w18*(w214*(0.75 - 1.5*X_2) + w215*(-1.75*nu - 0.9375) + w216*(0.1875 - 0.75*nu)) - w2*w215 + 1.0             # <<<<<<<<<<<<<<
 *         cdef double w224 = prst**2
 *         cdef double w225 = w165**(-2)
 */
  __Pyx_TraceLine(2094,0,__PYX_ERR(1, 2094, __pyx_L1_error))
  __pyx_v_w223 = ((((__pyx_v_w14 * (((-0.0625 * __pyx_v_w217) + (__pyx_v_w218 * __pyx_v_w32)) + (__pyx_v_w219 * __pyx_v_w35))) + (__pyx_v_w18 * (((__pyx_v_w214 * (0.75 - (1.5 * __pyx_v_X_2))) + (__pyx_v_w215 * ((-1.75 * __pyx_v_nu) - 0.9375))) + (__pyx_v_w216 * (0.1875 - (0.75 * __pyx_v_nu)))))) - (__pyx_v_w2 * __pyx_v_w215)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2095
 *         cdef double w222 = w0*w194*w221
 *         cdef double w223 = w14*(-0.0625*w217 + w218*w32 + w219*w35) + w18*(w214*(0.75 - 1.5*X_2) + w215*(-1.75*nu - 0.9375) + w216*(0.1875 - 0.75*nu)) - w2*w215 + 1.0
 *         cdef double w224 = prst**2             # <<<<<<<<<<<<<<
 *         cdef double w225 = w165**(-2)
 *         cdef double w226 = w126**(-1)
 */
  __Pyx_TraceLine(2095,0,__PYX_ERR(1, 2095, __pyx_L1_error))
  __pyx_v_w224 = pow(__pyx_v_prst, 2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2096
 *         cdef double w223 = w14*(-0.0625*w217 + w218*w32 + w219*w35) + w18*(w214*(0.75 - 1.5*X_2) + w215*(-1.75*nu - 0.9375) + w216*(0.1875 - 0.75*nu)) - w2*w215 + 1.0
 *         cdef double w224 = prst**2
 *         cdef double w225 = w165**(-2)             # <<<<<<<<<<<<<<
 *         cdef double w226 = w126**(-1)
 *         cdef double w227 = w11 + w120*w128 + w14*(0.03125*w93 + 0.015625*w94 + 0.015625*w96) + w18*(w10*(w186 + 2.8125) - 2.625*w37 + w92*(w45 - 1.0))
 */
  __Pyx_TraceLine(2096,0,__PYX_ERR(1, 2096, __pyx_L1_error))
  __pyx_v_w225 = pow(__pyx_v_w165, -2.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2097
 *         cdef double w224 = prst**2
 *         cdef double w225 = w165**(-2)
 *         cdef double w226 = w126**(-1)             # <<<<<<<<<<<<<<
 *         cdef double w227 = w11 + w120*w128 + w14*(0.03125*w93 + 0.015625*w94 + 0.015625*w96) + w18*(w10*(w186 + 2.8125) - 2.625*w37 + w92*(w45 - 1.0))
 *         cdef double w228 = w109*w144*w224*w225*w226*w227
 */
  __Pyx_TraceLine(2097,0,__PYX_ERR(1, 2097, __pyx_L1_error))
  __pyx_v_w226 = pow(__pyx_v_w126, -1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2098
 *         cdef double w225 = w165**(-2)
 *         cdef double w226 = w126**(-1)
 *         cdef double w227 = w11 + w120*w128 + w14*(0.03125*w93 + 0.015625*w94 + 0.015625*w96) + w18*(w10*(w186 + 2.8125) - 2.625*w37 + w92*(w45 - 1.0))             # <<<<<<<<<<<<<<
 *         cdef double w228 = w109*w144*w224*w225*w226*w227
 *         cdef double w229 = -w137*w222 + 147.443752990146*w146*w176 + w146*w181 - 11.3175085791863*w148*w178 + w148*w183 + 0.121954868780449*w150*w83 + 1.48275342024365*w151*w179 + w153*w154 + w155*w162 + w157*w158 + w161*w2 + 1.27277314139085e-19*w171*w172 + 1.69542100694444e-8*w20*w228 + w212*w223 + 1.0
 */
  __Pyx_TraceLine(2098,0,__PYX_ERR(1, 2098, __pyx_L1_error))
  __pyx_v_w227 = (((__pyx_v_w11 + (__pyx_v_w120 * __pyx_v_w128)) + (__pyx_v_w14 * (((0.03125 * __pyx_v_w93) + (0.015625 * __pyx_v_w94)) + (0.015625 * __pyx_v_w96)))) + (__pyx_v_w18 * (((__pyx_v_w10 * (__pyx_v_w186 + 2.8125)) - (2.625 * __pyx_v_w37)) + (__pyx_v_w92 * (__pyx_v_w45 - 1.0)))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2099
 *         cdef double w226 = w126**(-1)
 *         cdef double w227 = w11 + w120*w128 + w14*(0.03125*w93 + 0.015625*w94 + 0.015625*w96) + w18*(w10*(w186 + 2.8125) - 2.625*w37 + w92*(w45 - 1.0))
 *         cdef double w228 = w109*w144*w224*w225*w226*w227             # <<<<<<<<<<<<<<
 *         cdef double w229 = -w137*w222 + 147.443752990146*w146*w176 + w146*w181 - 11.3175085791863*w148*w178 + w148*w183 + 0.121954868780449*w150*w83 + 1.48275342024365*w151*w179 + w153*w154 + w155*w162 + w157*w158 + w161*w2 + 1.27277314139085e-19*w171*w172 + 1.69542100694444e-8*w20*w228 + w212*w223 + 1.0
 *         cdef double w230 = w78*w80
 */
  __Pyx_TraceLine(2099,0,__PYX_ERR(1, 2099, __pyx_L1_error))
  __pyx_v_w228 = (((((__pyx_v_w109 * __pyx_v_w144) * __pyx_v_w224) * __pyx_v_w225) * __pyx_v_w226) * __pyx_v_w227);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2100
 *         cdef double w227 = w11 + w120*w128 + w14*(0.03125*w93 + 0.015625*w94 + 0.015625*w96) + w18*(w10*(w186 + 2.8125) - 2.625*w37 + w92*(w45 - 1.0))
 *         cdef double w228 = w109*w144*w224*w225*w226*w227
 *         cdef double w229 = -w137*w222 + 147.443752990146*w146*w176 + w146*w181 - 11.3175085791863*w148*w178 + w148*w183 + 0.121954868780449*w150*w83 + 1.48275342024365*w151*w179 + w153*w154 + w155*w162 + w157*w158 + w161*w2 + 1.27277314139085e-19*w171*w172 + 1.69542100694444e-8*w20*w228 + w212*w223 + 1.0             # <<<<<<<<<<<<<<
 *         cdef double w230 = w78*w80
 * 
 */
  __Pyx_TraceLine(2100,0,__PYX_ERR(1, 2100, __pyx_L1_error))
  __pyx_v_w229 = ((((((((((((((((-__pyx_v_w137) * __pyx_v_w222) + ((147.443752990146 * __pyx_v_w146) * __pyx_v_w176)) + (__pyx_v_w146 * __pyx_v_w181)) - ((11.3175085791863 * __pyx_v_w148) * __pyx_v_w178)) + (__pyx_v_w148 * __pyx_v_w183)) + ((0.121954868780449 * __pyx_v_w150) * __pyx_v_w83)) + ((1.48275342024365 * __pyx_v_w151) * __pyx_v_w179)) + (__pyx_v_w153 * __pyx_v_w154)) + (__pyx_v_w155 * __pyx_v_w162)) + (__pyx_v_w157 * __pyx_v_w158)) + (__pyx_v_w161 * __pyx_v_w2)) + ((1.27277314139085e-19 * __pyx_v_w171) * __pyx_v_w172)) + ((1.69542100694444e-8 * __pyx_v_w20) * __pyx_v_w228)) + (__pyx_v_w212 * __pyx_v_w223)) + 1.0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2101
 *         cdef double w228 = w109*w144*w224*w225*w226*w227
 *         cdef double w229 = -w137*w222 + 147.443752990146*w146*w176 + w146*w181 - 11.3175085791863*w148*w178 + w148*w183 + 0.121954868780449*w150*w83 + 1.48275342024365*w151*w179 + w153*w154 + w155*w162 + w157*w158 + w161*w2 + 1.27277314139085e-19*w171*w172 + 1.69542100694444e-8*w20*w228 + w212*w223 + 1.0
 *         cdef double w230 = w78*w80             # <<<<<<<<<<<<<<
 * 
 *         cdef double dAdr = -w79 + w90
 */
  __Pyx_TraceLine(2101,0,__PYX_ERR(1, 2101, __pyx_L1_error))
  __pyx_v_w230 = (__pyx_v_w78 * __pyx_v_w80);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2103
 *         cdef double w230 = w78*w80
 * 
 *         cdef double dAdr = -w79 + w90             # <<<<<<<<<<<<<<
 *         cdef double dBnpdr = w132
 *         cdef double dBnpadr = r*w133*w138*w140 - r*w136 - w137
 */
  __Pyx_TraceLine(2103,0,__PYX_ERR(1, 2103, __pyx_L1_error))
  __pyx_v_dAdr = ((-__pyx_v_w79) + __pyx_v_w90);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2104
 * 
 *         cdef double dAdr = -w79 + w90
 *         cdef double dBnpdr = w132             # <<<<<<<<<<<<<<
 *         cdef double dBnpadr = r*w133*w138*w140 - r*w136 - w137
 *         cdef double dxidr = 2.0*r*w143*w77 + w1*w143*w145 + w1*w77*(r*w118 - 2.98505426338587e-26*r*w131 + w127)/(w141*(2.0*w1 + w15)) - w141*w66*w77/w144
 */
  __Pyx_TraceLine(2104,0,__PYX_ERR(1, 2104, __pyx_L1_error))
  __pyx_v_dBnpdr = __pyx_v_w132;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2105
 *         cdef double dAdr = -w79 + w90
 *         cdef double dBnpdr = w132
 *         cdef double dBnpadr = r*w133*w138*w140 - r*w136 - w137             # <<<<<<<<<<<<<<
 *         cdef double dxidr = 2.0*r*w143*w77 + w1*w143*w145 + w1*w77*(r*w118 - 2.98505426338587e-26*r*w131 + w127)/(w141*(2.0*w1 + w15)) - w141*w66*w77/w144
 *         cdef double dQdr = -w174
 */
  __Pyx_TraceLine(2105,0,__PYX_ERR(1, 2105, __pyx_L1_error))
  __pyx_v_dBnpadr = (((((__pyx_v_r * __pyx_v_w133) * __pyx_v_w138) * __pyx_v_w140) - (__pyx_v_r * __pyx_v_w136)) - __pyx_v_w137);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2106
 *         cdef double dBnpdr = w132
 *         cdef double dBnpadr = r*w133*w138*w140 - r*w136 - w137
 *         cdef double dxidr = 2.0*r*w143*w77 + w1*w143*w145 + w1*w77*(r*w118 - 2.98505426338587e-26*r*w131 + w127)/(w141*(2.0*w1 + w15)) - w141*w66*w77/w144             # <<<<<<<<<<<<<<
 *         cdef double dQdr = -w174
 *         cdef double dQdprst = 11.8620273619492*nu*w179*w180 + w14*w163*w175 + w154*w18*w182 + w158*w18*w184 + 8.0*w160*w180*w2 + 5.09109256556341e-19*w170*w173*w175 + 589.775011960583*w175*w176 - 67.9050514751178*w177*w178 + 0.975638950243592*w180*w83 + w181*w182 + w183*w184
 */
  __Pyx_TraceLine(2106,0,__PYX_ERR(1, 2106, __pyx_L1_error))
  __pyx_v_dxidr = ((((((2.0 * __pyx_v_r) * __pyx_v_w143) * __pyx_v_w77) + ((__pyx_v_w1 * __pyx_v_w143) * __pyx_v_w145)) + (((__pyx_v_w1 * __pyx_v_w77) * (((__pyx_v_r * __pyx_v_w118) - ((2.98505426338587e-26 * __pyx_v_r) * __pyx_v_w131)) + __pyx_v_w127)) / (__pyx_v_w141 * ((2.0 * __pyx_v_w1) + __pyx_v_w15)))) - (((__pyx_v_w141 * __pyx_v_w66) * __pyx_v_w77) / __pyx_v_w144));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2107
 *         cdef double dBnpadr = r*w133*w138*w140 - r*w136 - w137
 *         cdef double dxidr = 2.0*r*w143*w77 + w1*w143*w145 + w1*w77*(r*w118 - 2.98505426338587e-26*r*w131 + w127)/(w141*(2.0*w1 + w15)) - w141*w66*w77/w144
 *         cdef double dQdr = -w174             # <<<<<<<<<<<<<<
 *         cdef double dQdprst = 11.8620273619492*nu*w179*w180 + w14*w163*w175 + w154*w18*w182 + w158*w18*w184 + 8.0*w160*w180*w2 + 5.09109256556341e-19*w170*w173*w175 + 589.775011960583*w175*w176 - 67.9050514751178*w177*w178 + 0.975638950243592*w180*w83 + w181*w182 + w183*w184
 *         cdef double dHodddr = (-w14*w186*w191 + w190*(w18*(1.66666666666667*w26 - 0.75*w30 - 0.416666666666667*w37 + 1.25*w4 + 1.25*w7 + 1.25*w8) - w198*w211) + w190*(-w18*(-11.0625*nu + 1.13541666666667*w28 - 0.15625) - w192*w2 - w193*w196 - w197*w198 - w199*w201) + w206*(0.375*w14*w208*w31 - w18*(0.0833333333333333*w209 + w210)) + w207*(-w18*(-0.0625*nu + 1.07291666666667*w28 + 0.15625) - w196*w203 - w198*w204 - w2*w202 - w201*w205))/w185 - (3.0*w1 + w10)*(nu*w18*w191 + w190*(w195*w211 + w2*(0.208333333333333*delta*w36 - 0.833333333333333*w26 + 0.375*w30 - 0.625*w4 - 0.625*w7 - 0.625*w8)) + w190*(w0*w192 + w193*w212 + w195*w197 + w199*w213 + w2*(-5.53125*nu + 0.567708333333333*w28 - 0.078125) + 1.75) + w206*(w2*(0.0416666666666667*w209 + 0.5*w210) - 0.125*w208*w48) + w207*(w0*w202 + w195*w204 + w2*(-0.03125*nu + 0.536458333333333*w28 + 0.078125) + w203*w212 + w205*w213 + 0.25))/w185**2
 */
  __Pyx_TraceLine(2107,0,__PYX_ERR(1, 2107, __pyx_L1_error))
  __pyx_v_dQdr = (-__pyx_v_w174);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2108
 *         cdef double dxidr = 2.0*r*w143*w77 + w1*w143*w145 + w1*w77*(r*w118 - 2.98505426338587e-26*r*w131 + w127)/(w141*(2.0*w1 + w15)) - w141*w66*w77/w144
 *         cdef double dQdr = -w174
 *         cdef double dQdprst = 11.8620273619492*nu*w179*w180 + w14*w163*w175 + w154*w18*w182 + w158*w18*w184 + 8.0*w160*w180*w2 + 5.09109256556341e-19*w170*w173*w175 + 589.775011960583*w175*w176 - 67.9050514751178*w177*w178 + 0.975638950243592*w180*w83 + w181*w182 + w183*w184             # <<<<<<<<<<<<<<
 *         cdef double dHodddr = (-w14*w186*w191 + w190*(w18*(1.66666666666667*w26 - 0.75*w30 - 0.416666666666667*w37 + 1.25*w4 + 1.25*w7 + 1.25*w8) - w198*w211) + w190*(-w18*(-11.0625*nu + 1.13541666666667*w28 - 0.15625) - w192*w2 - w193*w196 - w197*w198 - w199*w201) + w206*(0.375*w14*w208*w31 - w18*(0.0833333333333333*w209 + w210)) + w207*(-w18*(-0.0625*nu + 1.07291666666667*w28 + 0.15625) - w196*w203 - w198*w204 - w2*w202 - w201*w205))/w185 - (3.0*w1 + w10)*(nu*w18*w191 + w190*(w195*w211 + w2*(0.208333333333333*delta*w36 - 0.833333333333333*w26 + 0.375*w30 - 0.625*w4 - 0.625*w7 - 0.625*w8)) + w190*(w0*w192 + w193*w212 + w195*w197 + w199*w213 + w2*(-5.53125*nu + 0.567708333333333*w28 - 0.078125) + 1.75) + w206*(w2*(0.0416666666666667*w209 + 0.5*w210) - 0.125*w208*w48) + w207*(w0*w202 + w195*w204 + w2*(-0.03125*nu + 0.536458333333333*w28 + 0.078125) + w203*w212 + w205*w213 + 0.25))/w185**2
 *         cdef double dBpdr = w220
 */
  __Pyx_TraceLine(2108,0,__PYX_ERR(1, 2108, __pyx_L1_error))
  __pyx_v_dQdprst = (((((((((((((11.8620273619492 * __pyx_v_nu) * __pyx_v_w179) * __pyx_v_w180) + ((__pyx_v_w14 * __pyx_v_w163) * __pyx_v_w175)) + ((__pyx_v_w154 * __pyx_v_w18) * __pyx_v_w182)) + ((__pyx_v_w158 * __pyx_v_w18) * __pyx_v_w184)) + (((8.0 * __pyx_v_w160) * __pyx_v_w180) * __pyx_v_w2)) + (((5.09109256556341e-19 * __pyx_v_w170) * __pyx_v_w173) * __pyx_v_w175)) + ((589.775011960583 * __pyx_v_w175) * __pyx_v_w176)) - ((67.9050514751178 * __pyx_v_w177) * __pyx_v_w178)) + ((0.975638950243592 * __pyx_v_w180) * __pyx_v_w83)) + (__pyx_v_w181 * __pyx_v_w182)) + (__pyx_v_w183 * __pyx_v_w184));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2109
 *         cdef double dQdr = -w174
 *         cdef double dQdprst = 11.8620273619492*nu*w179*w180 + w14*w163*w175 + w154*w18*w182 + w158*w18*w184 + 8.0*w160*w180*w2 + 5.09109256556341e-19*w170*w173*w175 + 589.775011960583*w175*w176 - 67.9050514751178*w177*w178 + 0.975638950243592*w180*w83 + w181*w182 + w183*w184
 *         cdef double dHodddr = (-w14*w186*w191 + w190*(w18*(1.66666666666667*w26 - 0.75*w30 - 0.416666666666667*w37 + 1.25*w4 + 1.25*w7 + 1.25*w8) - w198*w211) + w190*(-w18*(-11.0625*nu + 1.13541666666667*w28 - 0.15625) - w192*w2 - w193*w196 - w197*w198 - w199*w201) + w206*(0.375*w14*w208*w31 - w18*(0.0833333333333333*w209 + w210)) + w207*(-w18*(-0.0625*nu + 1.07291666666667*w28 + 0.15625) - w196*w203 - w198*w204 - w2*w202 - w201*w205))/w185 - (3.0*w1 + w10)*(nu*w18*w191 + w190*(w195*w211 + w2*(0.208333333333333*delta*w36 - 0.833333333333333*w26 + 0.375*w30 - 0.625*w4 - 0.625*w7 - 0.625*w8)) + w190*(w0*w192 + w193*w212 + w195*w197 + w199*w213 + w2*(-5.53125*nu + 0.567708333333333*w28 - 0.078125) + 1.75) + w206*(w2*(0.0416666666666667*w209 + 0.5*w210) - 0.125*w208*w48) + w207*(w0*w202 + w195*w204 + w2*(-0.03125*nu + 0.536458333333333*w28 + 0.078125) + w203*w212 + w205*w213 + 0.25))/w185**2             # <<<<<<<<<<<<<<
 *         cdef double dBpdr = w220
 *         cdef double dHevendr = 0.5*(w229*w230)**(-0.5)*(-w229*w79 + w229*w90 + w230*(w0*w133*w138*w140*w194*w221 - 2.24091649004576e-37*w109*w117*w144*w168*w20*w224*w225*w227 + 1.69542100694444e-8*w109*w132*w144*w20*w224*w225*w226 + 6.78168402777778e-8*w109*w14*w142*w224*w225*w226*w227 - 4.41515887225116e-12*w109*w144*w145*w20*w224*w226*w227/w165**3 + 1.69542100694444e-8*w130*w144*w20*w224*w225*w226*w227 + w133*w136*w194*w2*w221 - w136*w222 - w174 + w194*w2*w220 - w196*w223 - 8.47710503472222e-8*w228*w82))
 */
  __Pyx_TraceLine(2109,0,__PYX_ERR(1, 2109, __pyx_L1_error))
  __pyx_v_dHodddr = (((((((((-__pyx_v_w14) * __pyx_v_w186) * __pyx_v_w191) + (__pyx_v_w190 * ((__pyx_v_w18 * ((((((1.66666666666667 * __pyx_v_w26) - (0.75 * __pyx_v_w30)) - (0.416666666666667 * __pyx_v_w37)) + (1.25 * __pyx_v_w4)) + (1.25 * __pyx_v_w7)) + (1.25 * __pyx_v_w8))) - (__pyx_v_w198 * __pyx_v_w211)))) + (__pyx_v_w190 * ((((((-__pyx_v_w18) * (((-11.0625 * __pyx_v_nu) + (1.13541666666667 * __pyx_v_w28)) - 0.15625)) - (__pyx_v_w192 * __pyx_v_w2)) - (__pyx_v_w193 * __pyx_v_w196)) - (__pyx_v_w197 * __pyx_v_w198)) - (__pyx_v_w199 * __pyx_v_w201)))) + (__pyx_v_w206 * ((((0.375 * __pyx_v_w14) * __pyx_v_w208) * __pyx_v_w31) - (__pyx_v_w18 * ((0.0833333333333333 * __pyx_v_w209) + __pyx_v_w210))))) + (__pyx_v_w207 * ((((((-__pyx_v_w18) * (((-0.0625 * __pyx_v_nu) + (1.07291666666667 * __pyx_v_w28)) + 0.15625)) - (__pyx_v_w196 * __pyx_v_w203)) - (__pyx_v_w198 * __pyx_v_w204)) - (__pyx_v_w2 * __pyx_v_w202)) - (__pyx_v_w201 * __pyx_v_w205)))) / __pyx_v_w185) - ((((3.0 * __pyx_v_w1) + __pyx_v_w10) * ((((((__pyx_v_nu * __pyx_v_w18) * __pyx_v_w191) + (__pyx_v_w190 * ((__pyx_v_w195 * __pyx_v_w211) + (__pyx_v_w2 * (((((((0.208333333333333 * __pyx_v_delta) * __pyx_v_w36) - (0.833333333333333 * __pyx_v_w26)) + (0.375 * __pyx_v_w30)) - (0.625 * __pyx_v_w4)) - (0.625 * __pyx_v_w7)) - (0.625 * __pyx_v_w8)))))) + (__pyx_v_w190 * ((((((__pyx_v_w0 * __pyx_v_w192) + (__pyx_v_w193 * __pyx_v_w212)) + (__pyx_v_w195 * __pyx_v_w197)) + (__pyx_v_w199 * __pyx_v_w213)) + (__pyx_v_w2 * (((-5.53125 * __pyx_v_nu) + (0.567708333333333 * __pyx_v_w28)) - 0.078125))) + 1.75))) + (__pyx_v_w206 * ((__pyx_v_w2 * ((0.0416666666666667 * __pyx_v_w209) + (0.5 * __pyx_v_w210))) - ((0.125 * __pyx_v_w208) * __pyx_v_w48)))) + (__pyx_v_w207 * ((((((__pyx_v_w0 * __pyx_v_w202) + (__pyx_v_w195 * __pyx_v_w204)) + (__pyx_v_w2 * (((-0.03125 * __pyx_v_nu) + (0.536458333333333 * __pyx_v_w28)) + 0.078125))) + (__pyx_v_w203 * __pyx_v_w212)) + (__pyx_v_w205 * __pyx_v_w213)) + 0.25)))) / pow(__pyx_v_w185, 2.0)));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2110
 *         cdef double dQdprst = 11.8620273619492*nu*w179*w180 + w14*w163*w175 + w154*w18*w182 + w158*w18*w184 + 8.0*w160*w180*w2 + 5.09109256556341e-19*w170*w173*w175 + 589.775011960583*w175*w176 - 67.9050514751178*w177*w178 + 0.975638950243592*w180*w83 + w181*w182 + w183*w184
 *         cdef double dHodddr = (-w14*w186*w191 + w190*(w18*(1.66666666666667*w26 - 0.75*w30 - 0.416666666666667*w37 + 1.25*w4 + 1.25*w7 + 1.25*w8) - w198*w211) + w190*(-w18*(-11.0625*nu + 1.13541666666667*w28 - 0.15625) - w192*w2 - w193*w196 - w197*w198 - w199*w201) + w206*(0.375*w14*w208*w31 - w18*(0.0833333333333333*w209 + w210)) + w207*(-w18*(-0.0625*nu + 1.07291666666667*w28 + 0.15625) - w196*w203 - w198*w204 - w2*w202 - w201*w205))/w185 - (3.0*w1 + w10)*(nu*w18*w191 + w190*(w195*w211 + w2*(0.208333333333333*delta*w36 - 0.833333333333333*w26 + 0.375*w30 - 0.625*w4 - 0.625*w7 - 0.625*w8)) + w190*(w0*w192 + w193*w212 + w195*w197 + w199*w213 + w2*(-5.53125*nu + 0.567708333333333*w28 - 0.078125) + 1.75) + w206*(w2*(0.0416666666666667*w209 + 0.5*w210) - 0.125*w208*w48) + w207*(w0*w202 + w195*w204 + w2*(-0.03125*nu + 0.536458333333333*w28 + 0.078125) + w203*w212 + w205*w213 + 0.25))/w185**2
 *         cdef double dBpdr = w220             # <<<<<<<<<<<<<<
 *         cdef double dHevendr = 0.5*(w229*w230)**(-0.5)*(-w229*w79 + w229*w90 + w230*(w0*w133*w138*w140*w194*w221 - 2.24091649004576e-37*w109*w117*w144*w168*w20*w224*w225*w227 + 1.69542100694444e-8*w109*w132*w144*w20*w224*w225*w226 + 6.78168402777778e-8*w109*w14*w142*w224*w225*w226*w227 - 4.41515887225116e-12*w109*w144*w145*w20*w224*w226*w227/w165**3 + 1.69542100694444e-8*w130*w144*w20*w224*w225*w226*w227 + w133*w136*w194*w2*w221 - w136*w222 - w174 + w194*w2*w220 - w196*w223 - 8.47710503472222e-8*w228*w82))
 * 
 */
  __Pyx_TraceLine(2110,0,__PYX_ERR(1, 2110, __pyx_L1_error))
  __pyx_v_dBpdr = __pyx_v_w220;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2111
 *         cdef double dHodddr = (-w14*w186*w191 + w190*(w18*(1.66666666666667*w26 - 0.75*w30 - 0.416666666666667*w37 + 1.25*w4 + 1.25*w7 + 1.25*w8) - w198*w211) + w190*(-w18*(-11.0625*nu + 1.13541666666667*w28 - 0.15625) - w192*w2 - w193*w196 - w197*w198 - w199*w201) + w206*(0.375*w14*w208*w31 - w18*(0.0833333333333333*w209 + w210)) + w207*(-w18*(-0.0625*nu + 1.07291666666667*w28 + 0.15625) - w196*w203 - w198*w204 - w2*w202 - w201*w205))/w185 - (3.0*w1 + w10)*(nu*w18*w191 + w190*(w195*w211 + w2*(0.208333333333333*delta*w36 - 0.833333333333333*w26 + 0.375*w30 - 0.625*w4 - 0.625*w7 - 0.625*w8)) + w190*(w0*w192 + w193*w212 + w195*w197 + w199*w213 + w2*(-5.53125*nu + 0.567708333333333*w28 - 0.078125) + 1.75) + w206*(w2*(0.0416666666666667*w209 + 0.5*w210) - 0.125*w208*w48) + w207*(w0*w202 + w195*w204 + w2*(-0.03125*nu + 0.536458333333333*w28 + 0.078125) + w203*w212 + w205*w213 + 0.25))/w185**2
 *         cdef double dBpdr = w220
 *         cdef double dHevendr = 0.5*(w229*w230)**(-0.5)*(-w229*w79 + w229*w90 + w230*(w0*w133*w138*w140*w194*w221 - 2.24091649004576e-37*w109*w117*w144*w168*w20*w224*w225*w227 + 1.69542100694444e-8*w109*w132*w144*w20*w224*w225*w226 + 6.78168402777778e-8*w109*w14*w142*w224*w225*w226*w227 - 4.41515887225116e-12*w109*w144*w145*w20*w224*w226*w227/w165**3 + 1.69542100694444e-8*w130*w144*w20*w224*w225*w226*w227 + w133*w136*w194*w2*w221 - w136*w222 - w174 + w194*w2*w220 - w196*w223 - 8.47710503472222e-8*w228*w82))             # <<<<<<<<<<<<<<
 * 
 *         return dAdr, dBnpdr, dBnpadr, dxidr, dQdr, dQdprst, dHodddr, dBpdr, dHevendr
 */
  __Pyx_TraceLine(2111,0,__PYX_ERR(1, 2111, __pyx_L1_error))
  __pyx_v_dHevendr = ((0.5 * pow((__pyx_v_w229 * __pyx_v_w230), -0.5)) * ((((-__pyx_v_w229) * __pyx_v_w79) + (__pyx_v_w229 * __pyx_v_w90)) + (__pyx_v_w230 * ((((((((((((((((__pyx_v_w0 * __pyx_v_w133) * __pyx_v_w138) * __pyx_v_w140) * __pyx_v_w194) * __pyx_v_w221) - ((((((((2.24091649004576e-37 * __pyx_v_w109) * __pyx_v_w117) * __pyx_v_w144) * __pyx_v_w168) * __pyx_v_w20) * __pyx_v_w224) * __pyx_v_w225) * __pyx_v_w227)) + (((((((1.69542100694444e-8 * __pyx_v_w109) * __pyx_v_w132) * __pyx_v_w144) * __pyx_v_w20) * __pyx_v_w224) * __pyx_v_w225) * __pyx_v_w226)) + (((((((6.78168402777778e-8 * __pyx_v_w109) * __pyx_v_w14) * __pyx_v_w142) * __pyx_v_w224) * __pyx_v_w225) * __pyx_v_w226) * __pyx_v_w227)) - ((((((((4.41515887225116e-12 * __pyx_v_w109) * __pyx_v_w144) * __pyx_v_w145) * __pyx_v_w20) * __pyx_v_w224) * __pyx_v_w226) * __pyx_v_w227) / pow(__pyx_v_w165, 3.0))) + (((((((1.69542100694444e-8 * __pyx_v_w130) * __pyx_v_w144) * __pyx_v_w20) * __pyx_v_w224) * __pyx_v_w225) * __pyx_v_w226) * __pyx_v_w227)) + ((((__pyx_v_w133 * __pyx_v_w136) * __pyx_v_w194) * __pyx_v_w2) * __pyx_v_w221)) - (__pyx_v_w136 * __pyx_v_w222)) - __pyx_v_w174) + ((__pyx_v_w194 * __pyx_v_w2) * __pyx_v_w220)) - (__pyx_v_w196 * __pyx_v_w223)) - ((8.47710503472222e-8 * __pyx_v_w228) * __pyx_v_w82)))));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2113
 *         cdef double dHevendr = 0.5*(w229*w230)**(-0.5)*(-w229*w79 + w229*w90 + w230*(w0*w133*w138*w140*w194*w221 - 2.24091649004576e-37*w109*w117*w144*w168*w20*w224*w225*w227 + 1.69542100694444e-8*w109*w132*w144*w20*w224*w225*w226 + 6.78168402777778e-8*w109*w14*w142*w224*w225*w226*w227 - 4.41515887225116e-12*w109*w144*w145*w20*w224*w226*w227/w165**3 + 1.69542100694444e-8*w130*w144*w20*w224*w225*w226*w227 + w133*w136*w194*w2*w221 - w136*w222 - w174 + w194*w2*w220 - w196*w223 - 8.47710503472222e-8*w228*w82))
 * 
 *         return dAdr, dBnpdr, dBnpadr, dxidr, dQdr, dQdprst, dHodddr, dBpdr, dHevendr             # <<<<<<<<<<<<<<
 */
  __Pyx_TraceLine(2113,0,__PYX_ERR(1, 2113, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dAdr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_dBnpdr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_dBnpadr); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_16 = PyFloat_FromDouble(__pyx_v_dxidr); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_12 = PyFloat_FromDouble(__pyx_v_dQdr); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_11 = PyFloat_FromDouble(__pyx_v_dQdprst); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_dHodddr); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_dBpdr); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_dHevendr); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = PyTuple_New(9); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_7, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_7, 8, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_13 = 0;
  __pyx_t_16 = 0;
  __pyx_t_12 = 0;
  __pyx_t_11 = 0;
  __pyx_t_10 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1820
 *         return omega
 * 
 *     cpdef auxderivs(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.auxderivs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c_coeffs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_15auxderivs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_14auxderivs[] = "\n        Compute derivatives of the potentials which are used in the post-adiabatic approximation.\n\n        Args:\n          q (double[:]): Canonical positions (r,phi).\n          p (double[:]): Canonical momenta  (prstar,pphi).\n          chi1_v (double[:]): Dimensionless spin vector of the primary.\n          chi2_v (double[:]): Dimensionless spin vector of the secondary.\n          m_1 (double): Primary mass component.\n          m_2 (double): Secondary mass component.\n          chi_1 (double): Projection of chi1_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chi_2 (double): Projection of chi2_v onto the Newtonian orbital angular momentum unit vector (lN).\n          chiL1 (double): Projection of chi1_v onto the orbital angular momentum unit vector (l).\n          chiL2 (double): Projection of chi2_v onto the orbital angular momentum unit vector (l).\n\n        Returns:\n           (tuple) dAdr, dBnpdr, dBnpadr, dxidr, dQdr, dQdprst, dHodddr, dBpdr, dHevendr\n\n        ";
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_15auxderivs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_q = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi1_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_chi2_v = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_m_1;
  double __pyx_v_m_2;
  double __pyx_v_chi_1;
  double __pyx_v_chi_2;
  double __pyx_v_chiL1;
  double __pyx_v_chiL2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("auxderivs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_q,&__pyx_n_s_p,&__pyx_n_s_chi1_v,&__pyx_n_s_chi2_v,&__pyx_n_s_m_1,&__pyx_n_s_m_2,&__pyx_n_s_chi_1,&__pyx_n_s_chi_2,&__pyx_n_s_chiL1,&__pyx_n_s_chiL2,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_q)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 1); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi1_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 2); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi2_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 3); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 4); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 5); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 6); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chi_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 7); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 8); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chiL2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, 9); __PYX_ERR(1, 1820, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "auxderivs") < 0)) __PYX_ERR(1, 1820, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_q = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_q.memview)) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_p.memview)) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_chi1_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi1_v.memview)) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_chi2_v = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_chi2_v.memview)) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_m_1 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_m_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_m_2 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_m_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_chi_1 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_chi_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_chi_2 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_chi_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_chiL1 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_chiL1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1820, __pyx_L3_error)
    __pyx_v_chiL2 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_chiL2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 1820, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("auxderivs", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1820, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.auxderivs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_14auxderivs(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_14auxderivs(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, __Pyx_memviewslice __pyx_v_q, __Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_chi1_v, __Pyx_memviewslice __pyx_v_chi2_v, double __pyx_v_m_1, double __pyx_v_m_2, double __pyx_v_chi_1, double __pyx_v_chi_2, double __pyx_v_chiL1, double __pyx_v_chiL2) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("auxderivs", 0);
  __Pyx_TraceCall("auxderivs (wrapper)", __pyx_f[1], 1820, 0, __PYX_ERR(1, 1820, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_q.memview)) { __Pyx_RaiseUnboundLocalError("q"); __PYX_ERR(1, 1820, __pyx_L1_error) }
  if (unlikely(!__pyx_v_p.memview)) { __Pyx_RaiseUnboundLocalError("p"); __PYX_ERR(1, 1820, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi1_v.memview)) { __Pyx_RaiseUnboundLocalError("chi1_v"); __PYX_ERR(1, 1820, __pyx_L1_error) }
  if (unlikely(!__pyx_v_chi2_v.memview)) { __Pyx_RaiseUnboundLocalError("chi2_v"); __PYX_ERR(1, 1820, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_auxderivs(__pyx_v_self, __pyx_v_q, __pyx_v_p, __pyx_v_chi1_v, __pyx_v_chi2_v, __pyx_v_m_1, __pyx_v_m_2, __pyx_v_chi_1, __pyx_v_chi_2, __pyx_v_chiL1, __pyx_v_chiL2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.auxderivs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_q, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi1_v, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_chi2_v, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":18
 * 
 * cdef class Ham_AvgS2precess_simple_cython_PA_AD(Hamiltonian_v5PHM_C):
 *     cdef public CalibCoeffs calibration_coeffs             # <<<<<<<<<<<<<<
 *     def __cinit__(self,EOBParams params):
 *         self.EOBpars = params
 */

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs___get__(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs___get__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[1], 18, 0, __PYX_ERR(1, 18, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->calibration_coeffs));
  __pyx_r = ((PyObject *)__pyx_v_self->calibration_coeffs);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.calibration_coeffs.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_2__set__(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_2__set__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_TraceCall("__set__", __pyx_f[1], 18, 0, __PYX_ERR(1, 18, __pyx_L1_error));
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_8pyseobnr_3eob_5utils_10containers_CalibCoeffs))))) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->calibration_coeffs);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->calibration_coeffs));
  __pyx_v_self->calibration_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.calibration_coeffs.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_4__del__(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_4__del__(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_TraceCall("__del__", __pyx_f[1], 18, 0, __PYX_ERR(1, 18, __pyx_L1_error));
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->calibration_coeffs);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->calibration_coeffs));
  __pyx_v_self->calibration_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.calibration_coeffs.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_16__reduce_cython__(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceCall("__reduce_cython__", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __Pyx_TraceLine(2,0,__PYX_ERR(0, 2, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18__setstate_cython__(((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceCall("__setstate_cython__", __pyx_f[0], 3, 0, __PYX_ERR(0, 3, __pyx_L1_error));

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __Pyx_TraceLine(4,0,__PYX_ERR(0, 4, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);
  __Pyx_TraceCall("PyArray_MultiIterNew1", __pyx_f[2], 734, 0, __PYX_ERR(2, 734, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_TraceLine(735,0,__PYX_ERR(2, 735, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);
  __Pyx_TraceCall("PyArray_MultiIterNew2", __pyx_f[2], 737, 0, __PYX_ERR(2, 737, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_TraceLine(738,0,__PYX_ERR(2, 738, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);
  __Pyx_TraceCall("PyArray_MultiIterNew3", __pyx_f[2], 740, 0, __PYX_ERR(2, 740, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_TraceLine(741,0,__PYX_ERR(2, 741, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);
  __Pyx_TraceCall("PyArray_MultiIterNew4", __pyx_f[2], 743, 0, __PYX_ERR(2, 743, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_TraceLine(744,0,__PYX_ERR(2, 744, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);
  __Pyx_TraceCall("PyArray_MultiIterNew5", __pyx_f[2], 746, 0, __PYX_ERR(2, 746, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_TraceLine(747,0,__PYX_ERR(2, 747, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);
  __Pyx_TraceCall("PyDataType_SHAPE", __pyx_f[2], 749, 0, __PYX_ERR(2, 749, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __Pyx_TraceLine(750,0,__PYX_ERR(2, 750, __pyx_L1_error))
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_TraceLine(751,0,__PYX_ERR(2, 751, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(753,0,__PYX_ERR(2, 753, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.PyDataType_SHAPE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_array_base", 0);
  __Pyx_TraceCall("set_array_base", __pyx_f[2], 928, 0, __PYX_ERR(2, 928, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  __Pyx_TraceLine(929,0,__PYX_ERR(2, 929, __pyx_L1_error))
  Py_INCREF(__pyx_v_base);

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __Pyx_TraceLine(930,0,__PYX_ERR(2, 930, __pyx_L1_error))
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_array_base", 0);
  __Pyx_TraceCall("get_array_base", __pyx_f[2], 932, 0, __PYX_ERR(2, 932, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __Pyx_TraceLine(933,0,__PYX_ERR(2, 933, __pyx_L1_error))
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __Pyx_TraceLine(934,0,__PYX_ERR(2, 934, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_TraceLine(935,0,__PYX_ERR(2, 935, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_TraceLine(936,0,__PYX_ERR(2, 936, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.get_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);
  __Pyx_TraceCall("import_array", __pyx_f[2], 940, 0, __PYX_ERR(2, 940, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  __Pyx_TraceLine(941,0,__PYX_ERR(2, 941, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __Pyx_TraceLine(942,0,__PYX_ERR(2, 942, __pyx_L3_error))
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 942, __pyx_L3_error)

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __Pyx_TraceLine(943,0,__PYX_ERR(2, 943, __pyx_L5_except_error))
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 943, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __Pyx_TraceLine(944,0,__PYX_ERR(2, 944, __pyx_L5_except_error))
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 944, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);
  __Pyx_TraceCall("import_umath", __pyx_f[2], 946, 0, __PYX_ERR(2, 946, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  __Pyx_TraceLine(947,0,__PYX_ERR(2, 947, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __Pyx_TraceLine(948,0,__PYX_ERR(2, 948, __pyx_L3_error))
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 948, __pyx_L3_error)

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __Pyx_TraceLine(949,0,__PYX_ERR(2, 949, __pyx_L5_except_error))
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 949, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __Pyx_TraceLine(950,0,__PYX_ERR(2, 950, __pyx_L5_except_error))
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 950, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);
  __Pyx_TraceCall("import_ufunc", __pyx_f[2], 952, 0, __PYX_ERR(2, 952, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  __Pyx_TraceLine(953,0,__PYX_ERR(2, 953, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __Pyx_TraceLine(954,0,__PYX_ERR(2, 954, __pyx_L3_error))
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 954, __pyx_L3_error)

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __Pyx_TraceLine(955,0,__PYX_ERR(2, 955, __pyx_L5_except_error))
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 955, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __Pyx_TraceLine(956,0,__PYX_ERR(2, 956, __pyx_L5_except_error))
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 956, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);
  __Pyx_TraceCall("is_timedelta64_object", __pyx_f[2], 966, 0, __PYX_ERR(2, 966, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(978,0,__PYX_ERR(2, 978, __pyx_L1_error))
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("numpy.is_timedelta64_object", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);
  __Pyx_TraceCall("is_datetime64_object", __pyx_f[2], 981, 0, __PYX_ERR(2, 981, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(993,0,__PYX_ERR(2, 993, __pyx_L1_error))
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("numpy.is_datetime64_object", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("get_datetime64_value", __pyx_f[2], 996, 1, __PYX_ERR(2, 996, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1003,1,__PYX_ERR(2, 1003, __pyx_L1_error))
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("numpy.get_datetime64_value", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("get_timedelta64_value", __pyx_f[2], 1006, 1, __PYX_ERR(2, 1006, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1010,1,__PYX_ERR(2, 1010, __pyx_L1_error))
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("numpy.get_timedelta64_value", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("get_datetime64_unit", __pyx_f[2], 1013, 1, __PYX_ERR(2, 1013, __pyx_L1_error));

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __Pyx_TraceLine(1017,1,__PYX_ERR(2, 1017, __pyx_L1_error))
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("numpy.get_datetime64_unit", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = (NPY_DATETIMEUNIT) 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(0, 123, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(0, 123, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 123, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L3_error)
    } else {

      /* "View.MemoryView":124
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 123, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(0, 123, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(0, 123, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_TraceCall("__cinit__", __pyx_f[0], 123, 0, __PYX_ERR(0, 123, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":130
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  __Pyx_TraceLine(130,0,__PYX_ERR(0, 130, __pyx_L1_error))
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 130, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 130, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":131
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __Pyx_TraceLine(131,0,__PYX_ERR(0, 131, __pyx_L1_error))
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __Pyx_TraceLine(133,0,__PYX_ERR(0, 133, __pyx_L1_error))
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __Pyx_TraceLine(134,0,__PYX_ERR(0, 134, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 134, __pyx_L1_error)

    /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __Pyx_TraceLine(136,0,__PYX_ERR(0, 136, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __Pyx_TraceLine(137,0,__PYX_ERR(0, 137, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 137, __pyx_L1_error)

    /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __Pyx_TraceLine(139,0,__PYX_ERR(0, 139, __pyx_L1_error))
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":140
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __Pyx_TraceLine(140,0,__PYX_ERR(0, 140, __pyx_L1_error))
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":141
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  __Pyx_TraceLine(141,0,__PYX_ERR(0, 141, __pyx_L1_error))
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(0, 141, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":142
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(142,0,__PYX_ERR(0, 142, __pyx_L1_error))
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 142, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 142, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":145
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __Pyx_TraceLine(145,0,__PYX_ERR(0, 145, __pyx_L1_error))
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":146
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __Pyx_TraceLine(146,0,__PYX_ERR(0, 146, __pyx_L1_error))
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __Pyx_TraceLine(148,0,__PYX_ERR(0, 148, __pyx_L1_error))
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(149,0,__PYX_ERR(0, 149, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 149, __pyx_L1_error)

    /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __Pyx_TraceLine(152,0,__PYX_ERR(0, 152, __pyx_L1_error))
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 152, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __Pyx_TraceLine(153,0,__PYX_ERR(0, 153, __pyx_L1_error))
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":154
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __Pyx_TraceLine(154,0,__PYX_ERR(0, 154, __pyx_L1_error))
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(0, 154, __pyx_L1_error)

      /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":155
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    __Pyx_TraceLine(155,0,__PYX_ERR(0, 155, __pyx_L1_error))
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
    __Pyx_TraceLine(152,0,__PYX_ERR(0, 152, __pyx_L1_error))
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __Pyx_TraceLine(158,0,__PYX_ERR(0, 158, __pyx_L1_error))
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 158, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":159
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __Pyx_TraceLine(159,0,__PYX_ERR(0, 159, __pyx_L1_error))
    __pyx_v_order = 'F';

    /* "View.MemoryView":160
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_TraceLine(160,0,__PYX_ERR(0, 160, __pyx_L1_error))
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __Pyx_TraceLine(161,0,__PYX_ERR(0, 161, __pyx_L1_error))
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 161, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":162
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __Pyx_TraceLine(162,0,__PYX_ERR(0, 162, __pyx_L1_error))
    __pyx_v_order = 'C';

    /* "View.MemoryView":163
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_TraceLine(163,0,__PYX_ERR(0, 163, __pyx_L1_error))
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":165
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  __Pyx_TraceLine(165,0,__PYX_ERR(0, 165, __pyx_L1_error))
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(0, 165, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":167
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __Pyx_TraceLine(167,0,__PYX_ERR(0, 167, __pyx_L1_error))
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":170
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __Pyx_TraceLine(170,0,__PYX_ERR(0, 170, __pyx_L1_error))
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":171
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __Pyx_TraceLine(171,0,__PYX_ERR(0, 171, __pyx_L1_error))
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 171, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(172,0,__PYX_ERR(0, 172, __pyx_L1_error))
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":175
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __Pyx_TraceLine(175,0,__PYX_ERR(0, 175, __pyx_L1_error))
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __Pyx_TraceLine(176,0,__PYX_ERR(0, 176, __pyx_L1_error))
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __Pyx_TraceLine(177,0,__PYX_ERR(0, 177, __pyx_L1_error))
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(0, 177, __pyx_L1_error)

      /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __Pyx_TraceLine(179,0,__PYX_ERR(0, 179, __pyx_L1_error))
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":180
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __Pyx_TraceLine(180,0,__PYX_ERR(0, 180, __pyx_L1_error))
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":181
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      __Pyx_TraceLine(181,0,__PYX_ERR(0, 181, __pyx_L1_error))
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(0, 181, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(0, 181, __pyx_L1_error)
      }
      __pyx_t_1 = (__pyx_v_self->len / __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":182
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        __Pyx_TraceLine(182,0,__PYX_ERR(0, 182, __pyx_L1_error))
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":183
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        __Pyx_TraceLine(183,0,__PYX_ERR(0, 183, __pyx_L1_error))
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);
  __Pyx_TraceCall("__getbuffer__", __pyx_f[0], 186, 0, __PYX_ERR(0, 186, __pyx_L1_error));

  /* "View.MemoryView":187
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __Pyx_TraceLine(187,0,__PYX_ERR(0, 187, __pyx_L1_error))
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __Pyx_TraceLine(188,0,__PYX_ERR(0, 188, __pyx_L1_error))
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 188, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":189
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __Pyx_TraceLine(189,0,__PYX_ERR(0, 189, __pyx_L1_error))
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __Pyx_TraceLine(190,0,__PYX_ERR(0, 190, __pyx_L1_error))
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 190, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":191
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __Pyx_TraceLine(191,0,__PYX_ERR(0, 191, __pyx_L1_error))
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __Pyx_TraceLine(192,0,__PYX_ERR(0, 192, __pyx_L1_error))
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __Pyx_TraceLine(193,0,__PYX_ERR(0, 193, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 193, __pyx_L1_error)

    /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":194
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __Pyx_TraceLine(194,0,__PYX_ERR(0, 194, __pyx_L1_error))
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":195
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __Pyx_TraceLine(195,0,__PYX_ERR(0, 195, __pyx_L1_error))
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":196
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __Pyx_TraceLine(196,0,__PYX_ERR(0, 196, __pyx_L1_error))
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":197
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __Pyx_TraceLine(197,0,__PYX_ERR(0, 197, __pyx_L1_error))
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __Pyx_TraceLine(198,0,__PYX_ERR(0, 198, __pyx_L1_error))
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":199
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __Pyx_TraceLine(199,0,__PYX_ERR(0, 199, __pyx_L1_error))
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":200
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __Pyx_TraceLine(200,0,__PYX_ERR(0, 200, __pyx_L1_error))
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":201
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __Pyx_TraceLine(201,0,__PYX_ERR(0, 201, __pyx_L1_error))
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __Pyx_TraceLine(203,0,__PYX_ERR(0, 203, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":204
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __Pyx_TraceLine(204,0,__PYX_ERR(0, 204, __pyx_L1_error))
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":206
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  __Pyx_TraceLine(206,0,__PYX_ERR(0, 206, __pyx_L1_error))
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":208
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_TraceLine(208,0,__PYX_ERR(0, 208, __pyx_L1_error))
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);
  __Pyx_TraceCall("__dealloc__", __pyx_f[0], 212, 0, __PYX_ERR(0, 212, __pyx_L1_error));

  /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __Pyx_TraceLine(213,0,__PYX_ERR(0, 213, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":214
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __Pyx_TraceLine(214,0,__PYX_ERR(0, 214, __pyx_L1_error))
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __Pyx_TraceLine(215,0,__PYX_ERR(0, 215, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __Pyx_TraceLine(216,0,__PYX_ERR(0, 216, __pyx_L1_error))
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":217
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __Pyx_TraceLine(217,0,__PYX_ERR(0, 217, __pyx_L1_error))
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":219
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    __Pyx_TraceLine(219,0,__PYX_ERR(0, 219, __pyx_L1_error))
    free(__pyx_v_self->data);

    /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":220
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(220,0,__PYX_ERR(0, 220, __pyx_L1_error))
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.array.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 223, 0, __PYX_ERR(0, 223, __pyx_L1_error));

  /* "View.MemoryView":224
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_TraceLine(224,0,__PYX_ERR(0, 224, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 0);
  __Pyx_TraceCall("get_memview", __pyx_f[0], 227, 0, __PYX_ERR(0, 227, __pyx_L1_error));

  /* "View.MemoryView":228
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __Pyx_TraceLine(228,0,__PYX_ERR(0, 228, __pyx_L1_error))
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":229
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_TraceLine(229,0,__PYX_ERR(0, 229, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 0);
  __Pyx_TraceCall("__len__", __pyx_f[0], 231, 0, __PYX_ERR(0, 231, __pyx_L1_error));

  /* "View.MemoryView":232
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __Pyx_TraceLine(232,0,__PYX_ERR(0, 232, __pyx_L1_error))
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);
  __Pyx_TraceCall("__getattr__", __pyx_f[0], 234, 0, __PYX_ERR(0, 234, __pyx_L1_error));

  /* "View.MemoryView":235
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_TraceLine(235,0,__PYX_ERR(0, 235, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);
  __Pyx_TraceCall("__getitem__", __pyx_f[0], 237, 0, __PYX_ERR(0, 237, __pyx_L1_error));

  /* "View.MemoryView":238
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_TraceLine(238,0,__PYX_ERR(0, 238, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_TraceCall("__setitem__", __pyx_f[0], 240, 0, __PYX_ERR(0, 240, __pyx_L1_error));

  /* "View.MemoryView":241
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(241,0,__PYX_ERR(0, 241, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceCall("__reduce_cython__", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __Pyx_TraceLine(2,0,__PYX_ERR(0, 2, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceCall("__setstate_cython__", __pyx_f[0], 3, 0, __PYX_ERR(0, 3, __pyx_L1_error));

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __Pyx_TraceLine(4,0,__PYX_ERR(0, 4, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);
  __Pyx_TraceCall("array_cwrapper", __pyx_f[0], 245, 0, __PYX_ERR(0, 245, __pyx_L1_error));

  /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __Pyx_TraceLine(249,0,__PYX_ERR(0, 249, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":250
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __Pyx_TraceLine(250,0,__PYX_ERR(0, 250, __pyx_L1_error))
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  __Pyx_TraceLine(252,0,__PYX_ERR(0, 252, __pyx_L1_error))
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":253
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __Pyx_TraceLine(253,0,__PYX_ERR(0, 253, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(0, 253, __pyx_L1_error)

    /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __Pyx_TraceLine(252,0,__PYX_ERR(0, 252, __pyx_L1_error))
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":254
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __Pyx_TraceLine(254,0,__PYX_ERR(0, 254, __pyx_L1_error))
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":256
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(256,0,__PYX_ERR(0, 256, __pyx_L1_error))
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 282, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 282, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_TraceCall("__init__", __pyx_f[0], 282, 0, __PYX_ERR(0, 282, __pyx_L1_error));

  /* "View.MemoryView":283
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_TraceLine(283,0,__PYX_ERR(0, 283, __pyx_L1_error))
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);
  __Pyx_TraceCall("__repr__", __pyx_f[0], 284, 0, __PYX_ERR(0, 284, __pyx_L1_error));

  /* "View.MemoryView":285
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_TraceLine(285,0,__PYX_ERR(0, 285, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceCall("__reduce_cython__", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_TraceLine(5,0,__PYX_ERR(0, 5, __pyx_L1_error))
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __Pyx_TraceLine(6,0,__PYX_ERR(0, 6, __pyx_L1_error))
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __Pyx_TraceLine(7,0,__PYX_ERR(0, 7, __pyx_L1_error))
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __Pyx_TraceLine(8,0,__PYX_ERR(0, 8, __pyx_L1_error))
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __Pyx_TraceLine(9,0,__PYX_ERR(0, 9, __pyx_L1_error))
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  __Pyx_TraceLine(11,0,__PYX_ERR(0, 11, __pyx_L1_error))
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __Pyx_TraceLine(12,0,__PYX_ERR(0, 12, __pyx_L1_error))
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_TraceLine(13,0,__PYX_ERR(0, 13, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  __Pyx_TraceLine(15,0,__PYX_ERR(0, 15, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceCall("__setstate_cython__", __pyx_f[0], 16, 0, __PYX_ERR(0, 16, __pyx_L1_error));

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  __Pyx_TraceLine(17,0,__PYX_ERR(0, 17, __pyx_L1_error))
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("align_pointer", __pyx_f[0], 299, 1, __PYX_ERR(0, 299, __pyx_L1_error));

  /* "View.MemoryView":301
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __Pyx_TraceLine(301,1,__PYX_ERR(0, 301, __pyx_L1_error))
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":305
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __Pyx_TraceLine(307,1,__PYX_ERR(0, 307, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":308
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __Pyx_TraceLine(308,1,__PYX_ERR(0, 308, __pyx_L1_error))
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":310
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(310,1,__PYX_ERR(0, 310, __pyx_L1_error))
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.align_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(0, 346, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 346, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 346, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 346, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 346, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_TraceCall("__cinit__", __pyx_f[0], 346, 0, __PYX_ERR(0, 346, __pyx_L1_error));

  /* "View.MemoryView":347
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_TraceLine(347,0,__PYX_ERR(0, 347, __pyx_L1_error))
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":348
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __Pyx_TraceLine(348,0,__PYX_ERR(0, 348, __pyx_L1_error))
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __Pyx_TraceLine(349,0,__PYX_ERR(0, 349, __pyx_L1_error))
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":350
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __Pyx_TraceLine(350,0,__PYX_ERR(0, 350, __pyx_L1_error))
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 350, __pyx_L1_error)

    /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __Pyx_TraceLine(351,0,__PYX_ERR(0, 351, __pyx_L1_error))
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":352
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      __Pyx_TraceLine(352,0,__PYX_ERR(0, 352, __pyx_L1_error))
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":353
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 */
      __Pyx_TraceLine(353,0,__PYX_ERR(0, 353, __pyx_L1_error))
      Py_INCREF(Py_None);

      /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */
  __Pyx_TraceLine(355,0,__PYX_ERR(0, 355, __pyx_L1_error))
  __pyx_t_1 = ((!(__PYX_CYTHON_ATOMICS_ENABLED() != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */
    __Pyx_TraceLine(357,0,__PYX_ERR(0, 357, __pyx_L1_error))
    __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":358
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 */
      __Pyx_TraceLine(358,0,__PYX_ERR(0, 358, __pyx_L1_error))
      __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

      /* "View.MemoryView":359
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 */
      __Pyx_TraceLine(359,0,__PYX_ERR(0, 359, __pyx_L1_error))
      __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

      /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */
    }

    /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */
    __Pyx_TraceLine(360,0,__PYX_ERR(0, 360, __pyx_L1_error))
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":361
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *                 if self.lock is NULL:
 *                     raise MemoryError
 */
      __Pyx_TraceLine(361,0,__PYX_ERR(0, 361, __pyx_L1_error))
      __pyx_v_self->lock = PyThread_allocate_lock();

      /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      __Pyx_TraceLine(362,0,__PYX_ERR(0, 362, __pyx_L1_error))
      __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":363
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
        __Pyx_TraceLine(363,0,__PYX_ERR(0, 363, __pyx_L1_error))
        PyErr_NoMemory(); __PYX_ERR(0, 363, __pyx_L1_error)

        /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      }

      /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */
    }

    /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */
  }

  /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __Pyx_TraceLine(365,0,__PYX_ERR(0, 365, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":366
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __Pyx_TraceLine(366,0,__PYX_ERR(0, 366, __pyx_L1_error))
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L12_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L11;
  }

  /* "View.MemoryView":368
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  __Pyx_TraceLine(368,0,__PYX_ERR(0, 368, __pyx_L1_error))
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L11:;

  /* "View.MemoryView":370
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __Pyx_TraceLine(370,0,__PYX_ERR(0, 370, __pyx_L1_error))
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":372
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __Pyx_TraceLine(372,0,__PYX_ERR(0, 372, __pyx_L1_error))
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);
  __Pyx_TraceCall("__dealloc__", __pyx_f[0], 374, 0, __PYX_ERR(0, 374, __pyx_L1_error));

  /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __Pyx_TraceLine(375,0,__PYX_ERR(0, 375, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":376
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_TraceLine(376,0,__PYX_ERR(0, 376, __pyx_L1_error))
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __Pyx_TraceLine(377,0,__PYX_ERR(0, 377, __pyx_L1_error))
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":379
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    __Pyx_TraceLine(379,0,__PYX_ERR(0, 379, __pyx_L1_error))
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":380
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    __Pyx_TraceLine(380,0,__PYX_ERR(0, 380, __pyx_L1_error))
    Py_DECREF(Py_None);

    /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __Pyx_TraceLine(384,0,__PYX_ERR(0, 384, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":385
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __Pyx_TraceLine(385,0,__PYX_ERR(0, 385, __pyx_L1_error))
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __Pyx_TraceLine(386,0,__PYX_ERR(0, 386, __pyx_L1_error))
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":387
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __Pyx_TraceLine(387,0,__PYX_ERR(0, 387, __pyx_L1_error))
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __Pyx_TraceLine(388,0,__PYX_ERR(0, 388, __pyx_L1_error))
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":390
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __Pyx_TraceLine(390,0,__PYX_ERR(0, 390, __pyx_L1_error))
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":389
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          __Pyx_TraceLine(389,0,__PYX_ERR(0, 389, __pyx_L1_error))
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":391
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        __Pyx_TraceLine(391,0,__PYX_ERR(0, 391, __pyx_L1_error))
        goto __pyx_L6_break;

        /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":393
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      __Pyx_TraceLine(393,0,__PYX_ERR(0, 393, __pyx_L1_error))
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.memoryview.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);
  __Pyx_TraceCall("get_item_pointer", __pyx_f[0], 395, 0, __PYX_ERR(0, 395, __pyx_L1_error));

  /* "View.MemoryView":397
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __Pyx_TraceLine(397,0,__PYX_ERR(0, 397, __pyx_L1_error))
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __Pyx_TraceLine(399,0,__PYX_ERR(0, 399, __pyx_L1_error))
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 399, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 399, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 399, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":400
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __Pyx_TraceLine(400,0,__PYX_ERR(0, 400, __pyx_L1_error))
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 400, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(0, 400, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
    __Pyx_TraceLine(399,0,__PYX_ERR(0, 399, __pyx_L1_error))
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":402
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(402,0,__PYX_ERR(0, 402, __pyx_L1_error))
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);
  __Pyx_TraceCall("__getitem__", __pyx_f[0], 405, 0, __PYX_ERR(0, 405, __pyx_L1_error));

  /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __Pyx_TraceLine(406,0,__PYX_ERR(0, 406, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":407
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_TraceLine(407,0,__PYX_ERR(0, 407, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":409
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __Pyx_TraceLine(409,0,__PYX_ERR(0, 409, __pyx_L1_error))
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 409, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 409, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __Pyx_TraceLine(412,0,__PYX_ERR(0, 412, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 412, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":413
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_TraceLine(413,0,__PYX_ERR(0, 413, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":415
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  __Pyx_TraceLine(415,0,__PYX_ERR(0, 415, __pyx_L1_error))
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(0, 415, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":416
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_TraceLine(416,0,__PYX_ERR(0, 416, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_TraceCall("__setitem__", __pyx_f[0], 418, 0, __PYX_ERR(0, 418, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __Pyx_TraceLine(419,0,__PYX_ERR(0, 419, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __Pyx_TraceLine(420,0,__PYX_ERR(0, 420, __pyx_L1_error))
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 420, __pyx_L1_error)

    /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":422
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __Pyx_TraceLine(422,0,__PYX_ERR(0, 422, __pyx_L1_error))
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 422, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 422, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __Pyx_TraceLine(424,0,__PYX_ERR(0, 424, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 424, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":425
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __Pyx_TraceLine(425,0,__PYX_ERR(0, 425, __pyx_L1_error))
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __Pyx_TraceLine(426,0,__PYX_ERR(0, 426, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 426, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":427
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __Pyx_TraceLine(427,0,__PYX_ERR(0, 427, __pyx_L1_error))
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":429
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    __Pyx_TraceLine(429,0,__PYX_ERR(0, 429, __pyx_L1_error))
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(0, 429, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":431
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  __Pyx_TraceLine(431,0,__PYX_ERR(0, 431, __pyx_L1_error))
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_TraceCall("is_slice", __pyx_f[0], 433, 0, __PYX_ERR(0, 433, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __Pyx_TraceLine(434,0,__PYX_ERR(0, 434, __pyx_L1_error))
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    __Pyx_TraceLine(435,0,__PYX_ERR(0, 435, __pyx_L1_error))
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __Pyx_TraceLine(436,0,__PYX_ERR(0, 436, __pyx_L4_error))
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 436, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __Pyx_TraceLine(437,0,__PYX_ERR(0, 437, __pyx_L4_error))
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 437, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __Pyx_TraceLine(436,0,__PYX_ERR(0, 436, __pyx_L4_error))
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 436, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 436, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":438
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __Pyx_TraceLine(438,0,__PYX_ERR(0, 438, __pyx_L6_except_error))
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(0, 438, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":439
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_TraceLine(439,0,__PYX_ERR(0, 439, __pyx_L6_except_error))
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":441
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_TraceLine(441,0,__PYX_ERR(0, 441, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  __Pyx_memviewslice *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);
  __Pyx_TraceCall("setitem_slice_assignment", __pyx_f[0], 443, 0, __PYX_ERR(0, 443, __pyx_L1_error));

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __Pyx_TraceLine(447,0,__PYX_ERR(0, 447, __pyx_L1_error))
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(0, 447, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(0, 447, __pyx_L1_error)

  /* "View.MemoryView":448
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  __Pyx_TraceLine(448,0,__PYX_ERR(0, 448, __pyx_L1_error))
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(0, 448, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(0, 448, __pyx_L1_error)

  /* "View.MemoryView":449
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __Pyx_TraceLine(449,0,__PYX_ERR(0, 449, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __Pyx_TraceLine(447,0,__PYX_ERR(0, 447, __pyx_L1_error))
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 447, __pyx_L1_error)

  /* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);
  __Pyx_TraceCall("setitem_slice_assign_scalar", __pyx_f[0], 451, 0, __PYX_ERR(0, 451, __pyx_L1_error));

  /* "View.MemoryView":453
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __Pyx_TraceLine(453,0,__PYX_ERR(0, 453, __pyx_L1_error))
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":458
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __Pyx_TraceLine(458,0,__PYX_ERR(0, 458, __pyx_L1_error))
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(0, 458, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __Pyx_TraceLine(460,0,__PYX_ERR(0, 460, __pyx_L1_error))
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":461
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __Pyx_TraceLine(461,0,__PYX_ERR(0, 461, __pyx_L1_error))
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __Pyx_TraceLine(462,0,__PYX_ERR(0, 462, __pyx_L1_error))
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":463
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      __Pyx_TraceLine(463,0,__PYX_ERR(0, 463, __pyx_L1_error))
      PyErr_NoMemory(); __PYX_ERR(0, 463, __pyx_L1_error)

      /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":464
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __Pyx_TraceLine(464,0,__PYX_ERR(0, 464, __pyx_L1_error))
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":466
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __Pyx_TraceLine(466,0,__PYX_ERR(0, 466, __pyx_L1_error))
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":468
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  __Pyx_TraceLine(468,0,__PYX_ERR(0, 468, __pyx_L1_error))
  /*try:*/ {

    /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __Pyx_TraceLine(469,0,__PYX_ERR(0, 469, __pyx_L6_error))
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":470
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      __Pyx_TraceLine(470,0,__PYX_ERR(0, 470, __pyx_L6_error))
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":472
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(472,0,__PYX_ERR(0, 472, __pyx_L6_error))
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __Pyx_TraceLine(476,0,__PYX_ERR(0, 476, __pyx_L6_error))
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":477
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __Pyx_TraceLine(477,0,__PYX_ERR(0, 477, __pyx_L6_error))
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 477, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":478
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __Pyx_TraceLine(478,0,__PYX_ERR(0, 478, __pyx_L6_error))
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":481
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  __Pyx_TraceLine(481,0,__PYX_ERR(0, 481, __pyx_L6_error))
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);
  __Pyx_TraceCall("setitem_indexed", __pyx_f[0], 483, 0, __PYX_ERR(0, 483, __pyx_L1_error));

  /* "View.MemoryView":484
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __Pyx_TraceLine(484,0,__PYX_ERR(0, 484, __pyx_L1_error))
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(0, 484, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":485
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __Pyx_TraceLine(485,0,__PYX_ERR(0, 485, __pyx_L1_error))
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);
  __Pyx_TraceCall("convert_item_to_object", __pyx_f[0], 487, 0, __PYX_ERR(0, 487, __pyx_L1_error));

  /* "View.MemoryView":490
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __Pyx_TraceLine(490,0,__PYX_ERR(0, 490, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":493
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __Pyx_TraceLine(493,0,__PYX_ERR(0, 493, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  __Pyx_TraceLine(494,0,__PYX_ERR(0, 494, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":495
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __Pyx_TraceLine(495,0,__PYX_ERR(0, 495, __pyx_L3_error))
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 495, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 495, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 495, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    __Pyx_TraceLine(499,0,__PYX_ERR(0, 499, __pyx_L5_except_error))
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":500
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_TraceLine(500,0,__PYX_ERR(0, 500, __pyx_L5_except_error))
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":501
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_TraceLine(501,0,__PYX_ERR(0, 501, __pyx_L5_except_error))
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":496
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_TraceLine(496,0,__PYX_ERR(0, 496, __pyx_L5_except_error))
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 496, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(0, 496, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __Pyx_TraceLine(497,0,__PYX_ERR(0, 497, __pyx_L5_except_error))
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 497, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 497, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);
  __Pyx_TraceCall("assign_item_from_object", __pyx_f[0], 503, 0, __PYX_ERR(0, 503, __pyx_L1_error));

  /* "View.MemoryView":506
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __Pyx_TraceLine(506,0,__PYX_ERR(0, 506, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __Pyx_TraceLine(511,0,__PYX_ERR(0, 511, __pyx_L1_error))
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":512
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __Pyx_TraceLine(512,0,__PYX_ERR(0, 512, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  __Pyx_TraceLine(514,0,__PYX_ERR(0, 514, __pyx_L1_error))
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 514, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 514, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 514, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 514, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 514, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __Pyx_TraceLine(516,0,__PYX_ERR(0, 516, __pyx_L1_error))
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(0, 516, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __Pyx_TraceLine(517,0,__PYX_ERR(0, 517, __pyx_L1_error))
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __Pyx_TraceLine(516,0,__PYX_ERR(0, 516, __pyx_L1_error))
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __Pyx_TraceLine(517,0,__PYX_ERR(0, 517, __pyx_L1_error))
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);
  __Pyx_TraceCall("__getbuffer__", __pyx_f[0], 520, 0, __PYX_ERR(0, 520, __pyx_L1_error));

  /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __Pyx_TraceLine(521,0,__PYX_ERR(0, 521, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __Pyx_TraceLine(522,0,__PYX_ERR(0, 522, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 522, __pyx_L1_error)

    /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __Pyx_TraceLine(524,0,__PYX_ERR(0, 524, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":525
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __Pyx_TraceLine(525,0,__PYX_ERR(0, 525, __pyx_L1_error))
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":527
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  __Pyx_TraceLine(527,0,__PYX_ERR(0, 527, __pyx_L1_error))
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __Pyx_TraceLine(529,0,__PYX_ERR(0, 529, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":530
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __Pyx_TraceLine(530,0,__PYX_ERR(0, 530, __pyx_L1_error))
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":532
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  __Pyx_TraceLine(532,0,__PYX_ERR(0, 532, __pyx_L1_error))
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __Pyx_TraceLine(534,0,__PYX_ERR(0, 534, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":535
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __Pyx_TraceLine(535,0,__PYX_ERR(0, 535, __pyx_L1_error))
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":537
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __Pyx_TraceLine(537,0,__PYX_ERR(0, 537, __pyx_L1_error))
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __Pyx_TraceLine(539,0,__PYX_ERR(0, 539, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":540
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __Pyx_TraceLine(540,0,__PYX_ERR(0, 540, __pyx_L1_error))
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":542
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  __Pyx_TraceLine(542,0,__PYX_ERR(0, 542, __pyx_L1_error))
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":544
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __Pyx_TraceLine(544,0,__PYX_ERR(0, 544, __pyx_L1_error))
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":545
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __Pyx_TraceLine(545,0,__PYX_ERR(0, 545, __pyx_L1_error))
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":546
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __Pyx_TraceLine(546,0,__PYX_ERR(0, 546, __pyx_L1_error))
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":547
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __Pyx_TraceLine(547,0,__PYX_ERR(0, 547, __pyx_L1_error))
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":548
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __Pyx_TraceLine(548,0,__PYX_ERR(0, 548, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":549
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_TraceLine(549,0,__PYX_ERR(0, 549, __pyx_L1_error))
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 555, 0, __PYX_ERR(0, 555, __pyx_L1_error));

  /* "View.MemoryView":556
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __Pyx_TraceLine(556,0,__PYX_ERR(0, 556, __pyx_L1_error))
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(0, 556, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":557
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __Pyx_TraceLine(557,0,__PYX_ERR(0, 557, __pyx_L1_error))
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(0, 557, __pyx_L1_error)

  /* "View.MemoryView":558
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(558,0,__PYX_ERR(0, 558, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 561, 0, __PYX_ERR(0, 561, __pyx_L1_error));

  /* "View.MemoryView":562
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(562,0,__PYX_ERR(0, 562, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.base.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 565, 0, __PYX_ERR(0, 565, __pyx_L1_error));

  /* "View.MemoryView":566
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(566,0,__PYX_ERR(0, 566, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 569, 0, __PYX_ERR(0, 569, __pyx_L1_error));

  /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __Pyx_TraceLine(570,0,__PYX_ERR(0, 570, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __Pyx_TraceLine(572,0,__PYX_ERR(0, 572, __pyx_L1_error))
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 572, __pyx_L1_error)

    /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":574
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(574,0,__PYX_ERR(0, 574, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 577, 0, __PYX_ERR(0, 577, __pyx_L1_error));

  /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __Pyx_TraceLine(578,0,__PYX_ERR(0, 578, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_TraceLine(579,0,__PYX_ERR(0, 579, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__17, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":581
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(581,0,__PYX_ERR(0, 581, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 584, 0, __PYX_ERR(0, 584, __pyx_L1_error));

  /* "View.MemoryView":585
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(585,0,__PYX_ERR(0, 585, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 588, 0, __PYX_ERR(0, 588, __pyx_L1_error));

  /* "View.MemoryView":589
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(589,0,__PYX_ERR(0, 589, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 592, 0, __PYX_ERR(0, 592, __pyx_L1_error));

  /* "View.MemoryView":593
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(593,0,__PYX_ERR(0, 593, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 596, 0, __PYX_ERR(0, 596, __pyx_L1_error));

  /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __Pyx_TraceLine(597,0,__PYX_ERR(0, 597, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":598
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_TraceLine(598,0,__PYX_ERR(0, 598, __pyx_L1_error))
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":600
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __Pyx_TraceLine(600,0,__PYX_ERR(0, 600, __pyx_L1_error))
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":601
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __Pyx_TraceLine(601,0,__PYX_ERR(0, 601, __pyx_L1_error))
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":603
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_TraceLine(603,0,__PYX_ERR(0, 603, __pyx_L1_error))
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":605
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_TraceLine(605,0,__PYX_ERR(0, 605, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 0);
  __Pyx_TraceCall("__len__", __pyx_f[0], 607, 0, __PYX_ERR(0, 607, __pyx_L1_error));

  /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __Pyx_TraceLine(608,0,__PYX_ERR(0, 608, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":609
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __Pyx_TraceLine(609,0,__PYX_ERR(0, 609, __pyx_L1_error))
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":611
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_TraceLine(611,0,__PYX_ERR(0, 611, __pyx_L1_error))
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);
  __Pyx_TraceCall("__repr__", __pyx_f[0], 613, 0, __PYX_ERR(0, 613, __pyx_L1_error));

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_TraceLine(614,0,__PYX_ERR(0, 614, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":615
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_TraceLine(615,0,__PYX_ERR(0, 615, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_TraceLine(614,0,__PYX_ERR(0, 614, __pyx_L1_error))
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);
  __Pyx_TraceCall("__str__", __pyx_f[0], 617, 0, __PYX_ERR(0, 617, __pyx_L1_error));

  /* "View.MemoryView":618
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(618,0,__PYX_ERR(0, 618, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 0);
  __Pyx_TraceCall("is_c_contig", __pyx_f[0], 621, 0, __PYX_ERR(0, 621, __pyx_L1_error));

  /* "View.MemoryView":624
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __Pyx_TraceLine(624,0,__PYX_ERR(0, 624, __pyx_L1_error))
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(0, 624, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":625
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_TraceLine(625,0,__PYX_ERR(0, 625, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 0);
  __Pyx_TraceCall("is_f_contig", __pyx_f[0], 627, 0, __PYX_ERR(0, 627, __pyx_L1_error));

  /* "View.MemoryView":630
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __Pyx_TraceLine(630,0,__PYX_ERR(0, 630, __pyx_L1_error))
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(0, 630, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":631
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_TraceLine(631,0,__PYX_ERR(0, 631, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  __Pyx_TraceCall("copy", __pyx_f[0], 633, 0, __PYX_ERR(0, 633, __pyx_L1_error));

  /* "View.MemoryView":635
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __Pyx_TraceLine(635,0,__PYX_ERR(0, 635, __pyx_L1_error))
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":637
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __Pyx_TraceLine(637,0,__PYX_ERR(0, 637, __pyx_L1_error))
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":638
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __Pyx_TraceLine(638,0,__PYX_ERR(0, 638, __pyx_L1_error))
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 638, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":643
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_TraceLine(643,0,__PYX_ERR(0, 643, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 0);
  __Pyx_TraceCall("copy_fortran", __pyx_f[0], 645, 0, __PYX_ERR(0, 645, __pyx_L1_error));

  /* "View.MemoryView":647
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __Pyx_TraceLine(647,0,__PYX_ERR(0, 647, __pyx_L1_error))
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":649
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __Pyx_TraceLine(649,0,__PYX_ERR(0, 649, __pyx_L1_error))
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":650
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __Pyx_TraceLine(650,0,__PYX_ERR(0, 650, __pyx_L1_error))
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":655
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(655,0,__PYX_ERR(0, 655, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceCall("__reduce_cython__", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __Pyx_TraceLine(2,0,__PYX_ERR(0, 2, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceCall("__setstate_cython__", __pyx_f[0], 3, 0, __PYX_ERR(0, 3, __pyx_L1_error));

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __Pyx_TraceLine(4,0,__PYX_ERR(0, 4, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);
  __Pyx_TraceCall("memoryview_cwrapper", __pyx_f[0], 659, 0, __PYX_ERR(0, 659, __pyx_L1_error));

  /* "View.MemoryView":660
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __Pyx_TraceLine(660,0,__PYX_ERR(0, 660, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":661
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __Pyx_TraceLine(661,0,__PYX_ERR(0, 661, __pyx_L1_error))
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":662
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_TraceLine(662,0,__PYX_ERR(0, 662, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_check", 0);
  __Pyx_TraceCall("memoryview_check", __pyx_f[0], 665, 0, __PYX_ERR(0, 665, __pyx_L1_error));

  /* "View.MemoryView":666
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __Pyx_TraceLine(666,0,__PYX_ERR(0, 666, __pyx_L1_error))
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.memoryview_check", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 0);
  __Pyx_TraceCall("_unellipsify", __pyx_f[0], 668, 0, __PYX_ERR(0, 668, __pyx_L1_error));

  /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __Pyx_TraceLine(673,0,__PYX_ERR(0, 673, __pyx_L1_error))
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":674
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __Pyx_TraceLine(674,0,__PYX_ERR(0, 674, __pyx_L1_error))
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":676
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  __Pyx_TraceLine(676,0,__PYX_ERR(0, 676, __pyx_L1_error))
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":678
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __Pyx_TraceLine(678,0,__PYX_ERR(0, 678, __pyx_L1_error))
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":679
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __Pyx_TraceLine(679,0,__PYX_ERR(0, 679, __pyx_L1_error))
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":680
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __Pyx_TraceLine(680,0,__PYX_ERR(0, 680, __pyx_L1_error))
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_TraceLine(681,0,__PYX_ERR(0, 681, __pyx_L1_error))
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 681, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 681, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 681, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 681, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 681, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 681, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __Pyx_TraceLine(682,0,__PYX_ERR(0, 682, __pyx_L1_error))
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __Pyx_TraceLine(683,0,__PYX_ERR(0, 683, __pyx_L1_error))
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __Pyx_TraceLine(684,0,__PYX_ERR(0, 684, __pyx_L1_error))
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 684, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__20);
            __Pyx_GIVEREF(__pyx_slice__20);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__20);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":685
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __Pyx_TraceLine(685,0,__PYX_ERR(0, 685, __pyx_L1_error))
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":687
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      __Pyx_TraceLine(687,0,__PYX_ERR(0, 687, __pyx_L1_error))
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__20); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 687, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":688
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __Pyx_TraceLine(688,0,__PYX_ERR(0, 688, __pyx_L1_error))
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    __Pyx_TraceLine(690,0,__PYX_ERR(0, 690, __pyx_L1_error))
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":691
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __Pyx_TraceLine(691,0,__PYX_ERR(0, 691, __pyx_L1_error))
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 691, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 691, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(0, 691, __pyx_L1_error)

        /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":693
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __Pyx_TraceLine(693,0,__PYX_ERR(0, 693, __pyx_L1_error))
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":694
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __Pyx_TraceLine(694,0,__PYX_ERR(0, 694, __pyx_L1_error))
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 694, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
    __Pyx_TraceLine(681,0,__PYX_ERR(0, 681, __pyx_L1_error))
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":696
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __Pyx_TraceLine(696,0,__PYX_ERR(0, 696, __pyx_L1_error))
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 696, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __Pyx_TraceLine(697,0,__PYX_ERR(0, 697, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":698
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __Pyx_TraceLine(698,0,__PYX_ERR(0, 698, __pyx_L1_error))
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__20);
        __Pyx_GIVEREF(__pyx_slice__20);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__20);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":700
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_TraceLine(700,0,__PYX_ERR(0, 700, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);
  __Pyx_TraceCall("assert_direct_dimensions", __pyx_f[0], 702, 0, __PYX_ERR(0, 702, __pyx_L1_error));

  /* "View.MemoryView":703
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __Pyx_TraceLine(703,0,__PYX_ERR(0, 703, __pyx_L1_error))
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __Pyx_TraceLine(704,0,__PYX_ERR(0, 704, __pyx_L1_error))
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_TraceLine(705,0,__PYX_ERR(0, 705, __pyx_L1_error))
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 705, __pyx_L1_error)

      /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 0);
  __Pyx_TraceCall("memview_slice", __pyx_f[0], 712, 0, __PYX_ERR(0, 712, __pyx_L1_error));

  /* "View.MemoryView":713
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __Pyx_TraceLine(713,0,__PYX_ERR(0, 713, __pyx_L1_error))
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":720
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  __Pyx_TraceLine(720,0,__PYX_ERR(0, 720, __pyx_L1_error))
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":724
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  __Pyx_TraceLine(724,0,__PYX_ERR(0, 724, __pyx_L1_error))
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 724, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __Pyx_TraceLine(726,0,__PYX_ERR(0, 726, __pyx_L1_error))
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":727
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    __Pyx_TraceLine(727,0,__PYX_ERR(0, 727, __pyx_L1_error))
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(0, 727, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":728
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __Pyx_TraceLine(728,0,__PYX_ERR(0, 728, __pyx_L1_error))
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":730
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  __Pyx_TraceLine(730,0,__PYX_ERR(0, 730, __pyx_L1_error))
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":731
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(731,0,__PYX_ERR(0, 731, __pyx_L1_error))
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":737
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __Pyx_TraceLine(737,0,__PYX_ERR(0, 737, __pyx_L1_error))
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":738
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(738,0,__PYX_ERR(0, 738, __pyx_L1_error))
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":743
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __Pyx_TraceLine(743,0,__PYX_ERR(0, 743, __pyx_L1_error))
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":744
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __Pyx_TraceLine(744,0,__PYX_ERR(0, 744, __pyx_L1_error))
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __Pyx_TraceLine(748,0,__PYX_ERR(0, 748, __pyx_L1_error))
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 748, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 748, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 748, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 748, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __Pyx_TraceLine(749,0,__PYX_ERR(0, 749, __pyx_L1_error))
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":753
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __Pyx_TraceLine(753,0,__PYX_ERR(0, 753, __pyx_L1_error))
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 753, __pyx_L1_error)

      /* "View.MemoryView":750
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __Pyx_TraceLine(750,0,__PYX_ERR(0, 750, __pyx_L1_error))
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 750, __pyx_L1_error)

      /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __Pyx_TraceLine(756,0,__PYX_ERR(0, 756, __pyx_L1_error))
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":757
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      __Pyx_TraceLine(757,0,__PYX_ERR(0, 757, __pyx_L1_error))
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":758
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      __Pyx_TraceLine(758,0,__PYX_ERR(0, 758, __pyx_L1_error))
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":759
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      __Pyx_TraceLine(759,0,__PYX_ERR(0, 759, __pyx_L1_error))
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":760
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __Pyx_TraceLine(760,0,__PYX_ERR(0, 760, __pyx_L1_error))
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":762
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    __Pyx_TraceLine(762,0,__PYX_ERR(0, 762, __pyx_L1_error))
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":763
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __Pyx_TraceLine(763,0,__PYX_ERR(0, 763, __pyx_L1_error))
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 763, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 763, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 763, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":764
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __Pyx_TraceLine(764,0,__PYX_ERR(0, 764, __pyx_L1_error))
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 764, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 764, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":766
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __Pyx_TraceLine(766,0,__PYX_ERR(0, 766, __pyx_L1_error))
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":767
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __Pyx_TraceLine(767,0,__PYX_ERR(0, 767, __pyx_L1_error))
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __Pyx_TraceLine(768,0,__PYX_ERR(0, 768, __pyx_L1_error))
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 768, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":770
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __Pyx_TraceLine(770,0,__PYX_ERR(0, 770, __pyx_L1_error))
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 770, __pyx_L1_error)

      /* "View.MemoryView":776
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __Pyx_TraceLine(776,0,__PYX_ERR(0, 776, __pyx_L1_error))
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
    __Pyx_TraceLine(748,0,__PYX_ERR(0, 748, __pyx_L1_error))
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __Pyx_TraceLine(778,0,__PYX_ERR(0, 778, __pyx_L1_error))
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_TraceLine(779,0,__PYX_ERR(0, 779, __pyx_L1_error))
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":780
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    __Pyx_TraceLine(780,0,__PYX_ERR(0, 780, __pyx_L1_error))
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(0, 780, __pyx_L1_error) }

    /* "View.MemoryView":781
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    __Pyx_TraceLine(781,0,__PYX_ERR(0, 781, __pyx_L1_error))
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(0, 781, __pyx_L1_error) }

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_TraceLine(779,0,__PYX_ERR(0, 779, __pyx_L1_error))
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(0, 779, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  __Pyx_TraceLine(784,0,__PYX_ERR(0, 784, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":785
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(785,0,__PYX_ERR(0, 785, __pyx_L1_error))
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 784, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    __Pyx_TraceLine(784,0,__PYX_ERR(0, 784, __pyx_L1_error))
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(0, 784, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("slice_memviewslice", __pyx_f[0], 809, 1, __PYX_ERR(0, 809, __pyx_L1_error));

  /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __Pyx_TraceLine(829,1,__PYX_ERR(0, 829, __pyx_L1_error))
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __Pyx_TraceLine(831,1,__PYX_ERR(0, 831, __pyx_L1_error))
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":832
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __Pyx_TraceLine(832,1,__PYX_ERR(0, 832, __pyx_L1_error))
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __Pyx_TraceLine(833,1,__PYX_ERR(0, 833, __pyx_L1_error))
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":834
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __Pyx_TraceLine(834,1,__PYX_ERR(0, 834, __pyx_L1_error))
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 834, __pyx_L1_error)

      /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":837
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  __Pyx_TraceLine(837,1,__PYX_ERR(0, 837, __pyx_L1_error))
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __Pyx_TraceLine(839,1,__PYX_ERR(0, 839, __pyx_L1_error))
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":840
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_TraceLine(840,1,__PYX_ERR(0, 840, __pyx_L1_error))
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 840, __pyx_L1_error)

      /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __Pyx_TraceLine(843,1,__PYX_ERR(0, 843, __pyx_L1_error))
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __Pyx_TraceLine(844,1,__PYX_ERR(0, 844, __pyx_L1_error))
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":845
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __Pyx_TraceLine(845,1,__PYX_ERR(0, 845, __pyx_L1_error))
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __Pyx_TraceLine(846,1,__PYX_ERR(0, 846, __pyx_L1_error))
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":847
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __Pyx_TraceLine(847,1,__PYX_ERR(0, 847, __pyx_L1_error))
          __pyx_v_start = 0;

          /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __Pyx_TraceLine(848,1,__PYX_ERR(0, 848, __pyx_L1_error))
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __Pyx_TraceLine(849,1,__PYX_ERR(0, 849, __pyx_L1_error))
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":850
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __Pyx_TraceLine(850,1,__PYX_ERR(0, 850, __pyx_L1_error))
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":852
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __Pyx_TraceLine(852,1,__PYX_ERR(0, 852, __pyx_L1_error))
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    __Pyx_TraceLine(854,1,__PYX_ERR(0, 854, __pyx_L1_error))
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":855
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __Pyx_TraceLine(855,1,__PYX_ERR(0, 855, __pyx_L1_error))
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":857
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      __Pyx_TraceLine(857,1,__PYX_ERR(0, 857, __pyx_L1_error))
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __Pyx_TraceLine(859,1,__PYX_ERR(0, 859, __pyx_L1_error))
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __Pyx_TraceLine(860,1,__PYX_ERR(0, 860, __pyx_L1_error))
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":861
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __Pyx_TraceLine(861,1,__PYX_ERR(0, 861, __pyx_L1_error))
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __Pyx_TraceLine(862,1,__PYX_ERR(0, 862, __pyx_L1_error))
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":863
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __Pyx_TraceLine(863,1,__PYX_ERR(0, 863, __pyx_L1_error))
          __pyx_v_stop = 0;

          /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __Pyx_TraceLine(864,1,__PYX_ERR(0, 864, __pyx_L1_error))
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":865
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __Pyx_TraceLine(865,1,__PYX_ERR(0, 865, __pyx_L1_error))
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    __Pyx_TraceLine(867,1,__PYX_ERR(0, 867, __pyx_L1_error))
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":868
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __Pyx_TraceLine(868,1,__PYX_ERR(0, 868, __pyx_L1_error))
        __pyx_v_stop = -1L;

        /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":870
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      __Pyx_TraceLine(870,1,__PYX_ERR(0, 870, __pyx_L1_error))
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __Pyx_TraceLine(872,1,__PYX_ERR(0, 872, __pyx_L1_error))
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":873
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_TraceLine(873,1,__PYX_ERR(0, 873, __pyx_L1_error))
      __pyx_v_step = 1;

      /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":877
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":880
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __Pyx_TraceLine(882,1,__PYX_ERR(0, 882, __pyx_L1_error))
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":883
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_TraceLine(883,1,__PYX_ERR(0, 883, __pyx_L1_error))
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":886
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    __Pyx_TraceLine(886,1,__PYX_ERR(0, 886, __pyx_L1_error))
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":887
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    __Pyx_TraceLine(887,1,__PYX_ERR(0, 887, __pyx_L1_error))
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":888
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(888,1,__PYX_ERR(0, 888, __pyx_L1_error))
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __Pyx_TraceLine(891,1,__PYX_ERR(0, 891, __pyx_L1_error))
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":892
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __Pyx_TraceLine(892,1,__PYX_ERR(0, 892, __pyx_L1_error))
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":894
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  __Pyx_TraceLine(894,1,__PYX_ERR(0, 894, __pyx_L1_error))
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __Pyx_TraceLine(896,1,__PYX_ERR(0, 896, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __Pyx_TraceLine(897,1,__PYX_ERR(0, 897, __pyx_L1_error))
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __Pyx_TraceLine(898,1,__PYX_ERR(0, 898, __pyx_L1_error))
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":899
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __Pyx_TraceLine(899,1,__PYX_ERR(0, 899, __pyx_L1_error))
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":901
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      __Pyx_TraceLine(901,1,__PYX_ERR(0, 901, __pyx_L1_error))
      /*else*/ {

        /* "View.MemoryView":902
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __Pyx_TraceLine(902,1,__PYX_ERR(0, 902, __pyx_L1_error))
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 901, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":904
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    __Pyx_TraceLine(904,1,__PYX_ERR(0, 904, __pyx_L1_error))
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":906
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(906,1,__PYX_ERR(0, 906, __pyx_L1_error))
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);
  __Pyx_TraceCall("pybuffer_index", __pyx_f[0], 912, 0, __PYX_ERR(0, 912, __pyx_L1_error));

  /* "View.MemoryView":914
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __Pyx_TraceLine(914,0,__PYX_ERR(0, 914, __pyx_L1_error))
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":915
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __Pyx_TraceLine(915,0,__PYX_ERR(0, 915, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __Pyx_TraceLine(918,0,__PYX_ERR(0, 918, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":919
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    __Pyx_TraceLine(919,0,__PYX_ERR(0, 919, __pyx_L1_error))
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(0, 919, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(0, 919, __pyx_L1_error)
    }
    __pyx_v_shape = (__pyx_v_view->len / __pyx_v_itemsize);

    /* "View.MemoryView":920
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __Pyx_TraceLine(920,0,__PYX_ERR(0, 920, __pyx_L1_error))
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":922
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  __Pyx_TraceLine(922,0,__PYX_ERR(0, 922, __pyx_L1_error))
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":923
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __Pyx_TraceLine(923,0,__PYX_ERR(0, 923, __pyx_L1_error))
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __Pyx_TraceLine(924,0,__PYX_ERR(0, 924, __pyx_L1_error))
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":925
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __Pyx_TraceLine(925,0,__PYX_ERR(0, 925, __pyx_L1_error))
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __Pyx_TraceLine(927,0,__PYX_ERR(0, 927, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":928
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __Pyx_TraceLine(928,0,__PYX_ERR(0, 928, __pyx_L1_error))
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __Pyx_TraceLine(929,0,__PYX_ERR(0, 929, __pyx_L1_error))
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":930
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __Pyx_TraceLine(930,0,__PYX_ERR(0, 930, __pyx_L1_error))
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 930, __pyx_L1_error)

      /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __Pyx_TraceLine(932,0,__PYX_ERR(0, 932, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":933
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __Pyx_TraceLine(933,0,__PYX_ERR(0, 933, __pyx_L1_error))
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 933, __pyx_L1_error)

    /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":935
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __Pyx_TraceLine(935,0,__PYX_ERR(0, 935, __pyx_L1_error))
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __Pyx_TraceLine(936,0,__PYX_ERR(0, 936, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":937
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __Pyx_TraceLine(937,0,__PYX_ERR(0, 937, __pyx_L1_error))
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":939
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(939,0,__PYX_ERR(0, 939, __pyx_L1_error))
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("transpose_memslice", __pyx_f[0], 945, 1, __PYX_ERR(0, 945, __pyx_L1_error));

  /* "View.MemoryView":946
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __Pyx_TraceLine(946,1,__PYX_ERR(0, 946, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":948
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __Pyx_TraceLine(948,1,__PYX_ERR(0, 948, __pyx_L1_error))
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":949
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(949,1,__PYX_ERR(0, 949, __pyx_L1_error))
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":953
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __Pyx_TraceLine(953,1,__PYX_ERR(0, 953, __pyx_L1_error))
  __pyx_t_3 = (__pyx_v_ndim / 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":954
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __Pyx_TraceLine(954,1,__PYX_ERR(0, 954, __pyx_L1_error))
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":955
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __Pyx_TraceLine(955,1,__PYX_ERR(0, 955, __pyx_L1_error))
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":956
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __Pyx_TraceLine(956,1,__PYX_ERR(0, 956, __pyx_L1_error))
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __Pyx_TraceLine(958,1,__PYX_ERR(0, 958, __pyx_L1_error))
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":959
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __Pyx_TraceLine(959,1,__PYX_ERR(0, 959, __pyx_L1_error))
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 959, __pyx_L1_error)

      /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":961
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(961,1,__PYX_ERR(0, 961, __pyx_L1_error))
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);
  __Pyx_TraceCall("__dealloc__", __pyx_f[0], 978, 0, __PYX_ERR(0, 978, __pyx_L1_error));

  /* "View.MemoryView":979
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __Pyx_TraceLine(979,0,__PYX_ERR(0, 979, __pyx_L1_error))
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView._memoryviewslice.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);
  __Pyx_TraceCall("convert_item_to_object", __pyx_f[0], 981, 0, __PYX_ERR(0, 981, __pyx_L1_error));

  /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __Pyx_TraceLine(982,0,__PYX_ERR(0, 982, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":983
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_TraceLine(983,0,__PYX_ERR(0, 983, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":985
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  __Pyx_TraceLine(985,0,__PYX_ERR(0, 985, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);
  __Pyx_TraceCall("assign_item_from_object", __pyx_f[0], 987, 0, __PYX_ERR(0, 987, __pyx_L1_error));

  /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __Pyx_TraceLine(988,0,__PYX_ERR(0, 988, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":989
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __Pyx_TraceLine(989,0,__PYX_ERR(0, 989, __pyx_L1_error))
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(0, 989, __pyx_L1_error)

    /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":991
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_TraceLine(991,0,__PYX_ERR(0, 991, __pyx_L1_error))
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 991, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_TraceCall("__get__", __pyx_f[0], 994, 0, __PYX_ERR(0, 994, __pyx_L1_error));

  /* "View.MemoryView":995
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_TraceLine(995,0,__PYX_ERR(0, 995, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.base.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceCall("__reduce_cython__", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __Pyx_TraceLine(2,0,__PYX_ERR(0, 2, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceCall("__setstate_cython__", __pyx_f[0], 3, 0, __PYX_ERR(0, 3, __pyx_L1_error));

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __Pyx_TraceLine(4,0,__PYX_ERR(0, 4, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);
  __Pyx_TraceCall("memoryview_fromslice", __pyx_f[0], 1001, 0, __PYX_ERR(0, 1001, __pyx_L1_error));

  /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __Pyx_TraceLine(1009,0,__PYX_ERR(0, 1009, __pyx_L1_error))
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1010
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(1010,0,__PYX_ERR(0, 1010, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1015
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __Pyx_TraceLine(1015,0,__PYX_ERR(0, 1015, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1017
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __Pyx_TraceLine(1017,0,__PYX_ERR(0, 1017, __pyx_L1_error))
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1018
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __Pyx_TraceLine(1018,0,__PYX_ERR(0, 1018, __pyx_L1_error))
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1020
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __Pyx_TraceLine(1020,0,__PYX_ERR(0, 1020, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1021
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __Pyx_TraceLine(1021,0,__PYX_ERR(0, 1021, __pyx_L1_error))
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1023
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __Pyx_TraceLine(1023,0,__PYX_ERR(0, 1023, __pyx_L1_error))
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1024
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __Pyx_TraceLine(1024,0,__PYX_ERR(0, 1024, __pyx_L1_error))
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1025
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __Pyx_TraceLine(1025,0,__PYX_ERR(0, 1025, __pyx_L1_error))
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1026
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  __Pyx_TraceLine(1026,0,__PYX_ERR(0, 1026, __pyx_L1_error))
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1027
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  __Pyx_TraceLine(1027,0,__PYX_ERR(0, 1027, __pyx_L1_error))
  Py_INCREF(Py_None);

  /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __Pyx_TraceLine(1029,0,__PYX_ERR(0, 1029, __pyx_L1_error))
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1030
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __Pyx_TraceLine(1030,0,__PYX_ERR(0, 1030, __pyx_L1_error))
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  __Pyx_TraceLine(1032,0,__PYX_ERR(0, 1032, __pyx_L1_error))
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1034
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __Pyx_TraceLine(1034,0,__PYX_ERR(0, 1034, __pyx_L1_error))
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1035
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1035,0,__PYX_ERR(0, 1035, __pyx_L1_error))
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1038
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __Pyx_TraceLine(1038,0,__PYX_ERR(0, 1038, __pyx_L1_error))
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1039
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __Pyx_TraceLine(1039,0,__PYX_ERR(0, 1039, __pyx_L1_error))
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __Pyx_TraceLine(1040,0,__PYX_ERR(0, 1040, __pyx_L1_error))
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1041
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __Pyx_TraceLine(1041,0,__PYX_ERR(0, 1041, __pyx_L1_error))
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1042
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      __Pyx_TraceLine(1042,0,__PYX_ERR(0, 1042, __pyx_L1_error))
      goto __pyx_L6_break;

      /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1044
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __Pyx_TraceLine(1044,0,__PYX_ERR(0, 1044, __pyx_L1_error))
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1045
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __Pyx_TraceLine(1045,0,__PYX_ERR(0, 1045, __pyx_L1_error))
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1046
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __Pyx_TraceLine(1046,0,__PYX_ERR(0, 1046, __pyx_L1_error))
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1048
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __Pyx_TraceLine(1048,0,__PYX_ERR(0, 1048, __pyx_L1_error))
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1049
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __Pyx_TraceLine(1049,0,__PYX_ERR(0, 1049, __pyx_L1_error))
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1051
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_TraceLine(1051,0,__PYX_ERR(0, 1051, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);
  __Pyx_TraceCall("get_slice_from_memview", __pyx_f[0], 1054, 0, __PYX_ERR(0, 1054, __pyx_L1_error));

  /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __Pyx_TraceLine(1057,0,__PYX_ERR(0, 1057, __pyx_L1_error))
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1058
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    __Pyx_TraceLine(1058,0,__PYX_ERR(0, 1058, __pyx_L1_error))
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(0, 1058, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1059
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __Pyx_TraceLine(1059,0,__PYX_ERR(0, 1059, __pyx_L1_error))
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1061
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  __Pyx_TraceLine(1061,0,__PYX_ERR(0, 1061, __pyx_L1_error))
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1062
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __Pyx_TraceLine(1062,0,__PYX_ERR(0, 1062, __pyx_L1_error))
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("slice_copy", 0);
  __Pyx_TraceCall("slice_copy", __pyx_f[0], 1065, 0, __PYX_ERR(0, 1065, __pyx_L1_error));

  /* "View.MemoryView":1069
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __Pyx_TraceLine(1069,0,__PYX_ERR(0, 1069, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1070
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __Pyx_TraceLine(1070,0,__PYX_ERR(0, 1070, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1071
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __Pyx_TraceLine(1071,0,__PYX_ERR(0, 1071, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1073
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __Pyx_TraceLine(1073,0,__PYX_ERR(0, 1073, __pyx_L1_error))
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1074
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __Pyx_TraceLine(1074,0,__PYX_ERR(0, 1074, __pyx_L1_error))
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1076
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __Pyx_TraceLine(1076,0,__PYX_ERR(0, 1076, __pyx_L1_error))
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1077
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    __Pyx_TraceLine(1077,0,__PYX_ERR(0, 1077, __pyx_L1_error))
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1078
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    __Pyx_TraceLine(1078,0,__PYX_ERR(0, 1078, __pyx_L1_error))
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1079
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    __Pyx_TraceLine(1079,0,__PYX_ERR(0, 1079, __pyx_L1_error))
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.slice_copy", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);
  __Pyx_TraceCall("memoryview_copy", __pyx_f[0], 1082, 0, __PYX_ERR(0, 1082, __pyx_L1_error));

  /* "View.MemoryView":1085
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __Pyx_TraceLine(1085,0,__PYX_ERR(0, 1085, __pyx_L1_error))
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1086
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_TraceLine(1086,0,__PYX_ERR(0, 1086, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);
  __Pyx_TraceCall("memoryview_copy_from_slice", __pyx_f[0], 1089, 0, __PYX_ERR(0, 1089, __pyx_L1_error));

  /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __Pyx_TraceLine(1096,0,__PYX_ERR(0, 1096, __pyx_L1_error))
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1097
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __Pyx_TraceLine(1097,0,__PYX_ERR(0, 1097, __pyx_L1_error))
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1098
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __Pyx_TraceLine(1098,0,__PYX_ERR(0, 1098, __pyx_L1_error))
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1100
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  __Pyx_TraceLine(1100,0,__PYX_ERR(0, 1100, __pyx_L1_error))
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1101
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __Pyx_TraceLine(1101,0,__PYX_ERR(0, 1101, __pyx_L1_error))
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1103
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_TraceLine(1103,0,__PYX_ERR(0, 1103, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1105
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1105,0,__PYX_ERR(0, 1105, __pyx_L1_error))
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("abs_py_ssize_t", __pyx_f[0], 1111, 1, __PYX_ERR(0, 1111, __pyx_L1_error));

  /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __Pyx_TraceLine(1112,1,__PYX_ERR(0, 1112, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1113
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __Pyx_TraceLine(1113,1,__PYX_ERR(0, 1113, __pyx_L1_error))
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1115
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  __Pyx_TraceLine(1115,1,__PYX_ERR(0, 1115, __pyx_L1_error))
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.abs_py_ssize_t", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("get_best_order", __pyx_f[0], 1118, 1, __PYX_ERR(0, 1118, __pyx_L1_error));

  /* "View.MemoryView":1123
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __Pyx_TraceLine(1123,1,__PYX_ERR(0, 1123, __pyx_L1_error))
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1124
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __Pyx_TraceLine(1124,1,__PYX_ERR(0, 1124, __pyx_L1_error))
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1126
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  __Pyx_TraceLine(1126,1,__PYX_ERR(0, 1126, __pyx_L1_error))
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __Pyx_TraceLine(1127,1,__PYX_ERR(0, 1127, __pyx_L1_error))
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1128
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __Pyx_TraceLine(1128,1,__PYX_ERR(0, 1128, __pyx_L1_error))
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1129
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      __Pyx_TraceLine(1129,1,__PYX_ERR(0, 1129, __pyx_L1_error))
      goto __pyx_L4_break;

      /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1131
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __Pyx_TraceLine(1131,1,__PYX_ERR(0, 1131, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __Pyx_TraceLine(1132,1,__PYX_ERR(0, 1132, __pyx_L1_error))
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1133
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __Pyx_TraceLine(1133,1,__PYX_ERR(0, 1133, __pyx_L1_error))
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1134
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      __Pyx_TraceLine(1134,1,__PYX_ERR(0, 1134, __pyx_L1_error))
      goto __pyx_L7_break;

      /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __Pyx_TraceLine(1136,1,__PYX_ERR(0, 1136, __pyx_L1_error))
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1137
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __Pyx_TraceLine(1137,1,__PYX_ERR(0, 1137, __pyx_L1_error))
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1139
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __Pyx_TraceLine(1139,1,__PYX_ERR(0, 1139, __pyx_L1_error))
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.get_best_order", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1149
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1151
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1156
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1157
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1159
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1160
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1161
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1162
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1164
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1165
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1169
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1170
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  __Pyx_TraceDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("copy_strided_to_strided", __pyx_f[0], 1172, 1, __PYX_ERR(0, 1172, __pyx_L1_error));

  /* "View.MemoryView":1175
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  __Pyx_TraceLine(1175,1,__PYX_ERR(0, 1175, __pyx_L1_error))
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.copy_strided_to_strided", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  __Pyx_TraceDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("slice_get_size", __pyx_f[0], 1179, 1, __PYX_ERR(0, 1179, __pyx_L1_error));

  /* "View.MemoryView":1181
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __Pyx_TraceLine(1181,1,__PYX_ERR(0, 1181, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1183
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __Pyx_TraceLine(1183,1,__PYX_ERR(0, 1183, __pyx_L1_error))
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1184
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __Pyx_TraceLine(1184,1,__PYX_ERR(0, 1184, __pyx_L1_error))
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1186
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __Pyx_TraceLine(1186,1,__PYX_ERR(0, 1186, __pyx_L1_error))
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.slice_get_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("fill_contig_strides_array", __pyx_f[0], 1189, 1, __PYX_ERR(0, 1189, __pyx_L1_error));

  /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __Pyx_TraceLine(1198,1,__PYX_ERR(0, 1198, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1199
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __Pyx_TraceLine(1199,1,__PYX_ERR(0, 1199, __pyx_L1_error))
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1200
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      __Pyx_TraceLine(1200,1,__PYX_ERR(0, 1200, __pyx_L1_error))
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1201
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __Pyx_TraceLine(1201,1,__PYX_ERR(0, 1201, __pyx_L1_error))
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1203
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  __Pyx_TraceLine(1203,1,__PYX_ERR(0, 1203, __pyx_L1_error))
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1204
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      __Pyx_TraceLine(1204,1,__PYX_ERR(0, 1204, __pyx_L1_error))
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1205
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __Pyx_TraceLine(1205,1,__PYX_ERR(0, 1205, __pyx_L1_error))
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1207
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __Pyx_TraceLine(1207,1,__PYX_ERR(0, 1207, __pyx_L1_error))
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.fill_contig_strides_array", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  __Pyx_TraceDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("copy_data_to_temp", __pyx_f[0], 1210, 1, __PYX_ERR(0, 1210, __pyx_L1_error));

  /* "View.MemoryView":1221
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __Pyx_TraceLine(1221,1,__PYX_ERR(0, 1221, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1222
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __Pyx_TraceLine(1222,1,__PYX_ERR(0, 1222, __pyx_L1_error))
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1224
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __Pyx_TraceLine(1224,1,__PYX_ERR(0, 1224, __pyx_L1_error))
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __Pyx_TraceLine(1225,1,__PYX_ERR(0, 1225, __pyx_L1_error))
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1226
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(1226,1,__PYX_ERR(0, 1226, __pyx_L1_error))
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1226, __pyx_L1_error)

    /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1229
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __Pyx_TraceLine(1229,1,__PYX_ERR(0, 1229, __pyx_L1_error))
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1230
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __Pyx_TraceLine(1230,1,__PYX_ERR(0, 1230, __pyx_L1_error))
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1231
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __Pyx_TraceLine(1231,1,__PYX_ERR(0, 1231, __pyx_L1_error))
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1232
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    __Pyx_TraceLine(1232,1,__PYX_ERR(0, 1232, __pyx_L1_error))
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1233
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    __Pyx_TraceLine(1233,1,__PYX_ERR(0, 1233, __pyx_L1_error))
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1235
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  __Pyx_TraceLine(1235,1,__PYX_ERR(0, 1235, __pyx_L1_error))
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1239
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __Pyx_TraceLine(1239,1,__PYX_ERR(0, 1239, __pyx_L1_error))
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __Pyx_TraceLine(1240,1,__PYX_ERR(0, 1240, __pyx_L1_error))
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1241
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      __Pyx_TraceLine(1241,1,__PYX_ERR(0, 1241, __pyx_L1_error))
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __Pyx_TraceLine(1243,1,__PYX_ERR(0, 1243, __pyx_L1_error))
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1244
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    __Pyx_TraceLine(1244,1,__PYX_ERR(0, 1244, __pyx_L1_error))
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1246
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __Pyx_TraceLine(1246,1,__PYX_ERR(0, 1246, __pyx_L1_error))
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1248
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1248,1,__PYX_ERR(0, 1248, __pyx_L1_error))
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);
  __Pyx_TraceCall("_err_extents", __pyx_f[0], 1253, 0, __PYX_ERR(0, 1253, __pyx_L1_error));

  /* "View.MemoryView":1256
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __Pyx_TraceLine(1256,0,__PYX_ERR(0, 1256, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1255
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __Pyx_TraceLine(1255,0,__PYX_ERR(0, 1255, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(0, 1255, __pyx_L1_error)

  /* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_TraceCall("_err_dim", __pyx_f[0], 1259, 0, __PYX_ERR(0, 1259, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1260
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __Pyx_TraceLine(1260,0,__PYX_ERR(0, 1260, __pyx_L1_error))
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 1260, __pyx_L1_error)

  /* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_TraceCall("_err", __pyx_f[0], 1263, 0, __PYX_ERR(0, 1263, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __Pyx_TraceLine(1264,0,__PYX_ERR(0, 1264, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1265
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __Pyx_TraceLine(1265,0,__PYX_ERR(0, 1265, __pyx_L1_error))
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1265, __pyx_L1_error)

    /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1267
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  __Pyx_TraceLine(1267,0,__PYX_ERR(0, 1267, __pyx_L1_error))
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(0, 1267, __pyx_L1_error)
  }

  /* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  __Pyx_TraceDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("memoryview_copy_contents", __pyx_f[0], 1270, 1, __PYX_ERR(0, 1270, __pyx_L1_error));

  /* "View.MemoryView":1278
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __Pyx_TraceLine(1278,1,__PYX_ERR(0, 1278, __pyx_L1_error))
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1279
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __Pyx_TraceLine(1279,1,__PYX_ERR(0, 1279, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1281
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __Pyx_TraceLine(1281,1,__PYX_ERR(0, 1281, __pyx_L1_error))
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1282
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __Pyx_TraceLine(1282,1,__PYX_ERR(0, 1282, __pyx_L1_error))
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1283
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __Pyx_TraceLine(1283,1,__PYX_ERR(0, 1283, __pyx_L1_error))
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __Pyx_TraceLine(1286,1,__PYX_ERR(0, 1286, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __Pyx_TraceLine(1287,1,__PYX_ERR(0, 1287, __pyx_L1_error))
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __Pyx_TraceLine(1288,1,__PYX_ERR(0, 1288, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1289
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __Pyx_TraceLine(1289,1,__PYX_ERR(0, 1289, __pyx_L1_error))
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1291
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __Pyx_TraceLine(1291,1,__PYX_ERR(0, 1291, __pyx_L1_error))
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1293
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __Pyx_TraceLine(1293,1,__PYX_ERR(0, 1293, __pyx_L1_error))
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __Pyx_TraceLine(1294,1,__PYX_ERR(0, 1294, __pyx_L1_error))
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __Pyx_TraceLine(1295,1,__PYX_ERR(0, 1295, __pyx_L1_error))
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1296
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __Pyx_TraceLine(1296,1,__PYX_ERR(0, 1296, __pyx_L1_error))
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1297
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        __Pyx_TraceLine(1297,1,__PYX_ERR(0, 1297, __pyx_L1_error))
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1299
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      __Pyx_TraceLine(1299,1,__PYX_ERR(0, 1299, __pyx_L1_error))
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1299, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __Pyx_TraceLine(1301,1,__PYX_ERR(0, 1301, __pyx_L1_error))
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1302
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __Pyx_TraceLine(1302,1,__PYX_ERR(0, 1302, __pyx_L1_error))
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1302, __pyx_L1_error)

      /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __Pyx_TraceLine(1304,1,__PYX_ERR(0, 1304, __pyx_L1_error))
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __Pyx_TraceLine(1306,1,__PYX_ERR(0, 1306, __pyx_L1_error))
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1307
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __Pyx_TraceLine(1307,1,__PYX_ERR(0, 1307, __pyx_L1_error))
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1309
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __Pyx_TraceLine(1309,1,__PYX_ERR(0, 1309, __pyx_L1_error))
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(0, 1309, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1310
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __Pyx_TraceLine(1310,1,__PYX_ERR(0, 1310, __pyx_L1_error))
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1312,1,__PYX_ERR(0, 1312, __pyx_L1_error))
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __Pyx_TraceLine(1315,1,__PYX_ERR(0, 1315, __pyx_L1_error))
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __Pyx_TraceLine(1316,1,__PYX_ERR(0, 1316, __pyx_L1_error))
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __Pyx_TraceLine(1317,1,__PYX_ERR(0, 1317, __pyx_L1_error))
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __Pyx_TraceLine(1318,1,__PYX_ERR(0, 1318, __pyx_L1_error))
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __Pyx_TraceLine(1320,1,__PYX_ERR(0, 1320, __pyx_L1_error))
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1322
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __Pyx_TraceLine(1322,1,__PYX_ERR(0, 1322, __pyx_L1_error))
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1323
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      __Pyx_TraceLine(1323,1,__PYX_ERR(0, 1323, __pyx_L1_error))
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __Pyx_TraceLine(1324,1,__PYX_ERR(0, 1324, __pyx_L1_error))
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1325
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      __Pyx_TraceLine(1325,1,__PYX_ERR(0, 1325, __pyx_L1_error))
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1326
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __Pyx_TraceLine(1326,1,__PYX_ERR(0, 1326, __pyx_L1_error))
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1328,1,__PYX_ERR(0, 1328, __pyx_L1_error))
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1331
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __Pyx_TraceLine(1331,1,__PYX_ERR(0, 1331, __pyx_L1_error))
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(0, 1331, __pyx_L1_error)

    /* "View.MemoryView":1332
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __Pyx_TraceLine(1332,1,__PYX_ERR(0, 1332, __pyx_L1_error))
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(0, 1332, __pyx_L1_error)

    /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1334
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __Pyx_TraceLine(1334,1,__PYX_ERR(0, 1334, __pyx_L1_error))
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1335
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  __Pyx_TraceLine(1335,1,__PYX_ERR(0, 1335, __pyx_L1_error))
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __Pyx_TraceLine(1336,1,__PYX_ERR(0, 1336, __pyx_L1_error))
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1338
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __Pyx_TraceLine(1338,1,__PYX_ERR(0, 1338, __pyx_L1_error))
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1339
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __Pyx_TraceLine(1339,1,__PYX_ERR(0, 1339, __pyx_L1_error))
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("broadcast_leading", __pyx_f[0], 1342, 1, __PYX_ERR(0, 1342, __pyx_L1_error));

  /* "View.MemoryView":1346
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __Pyx_TraceLine(1346,1,__PYX_ERR(0, 1346, __pyx_L1_error))
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1348
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  __Pyx_TraceLine(1348,1,__PYX_ERR(0, 1348, __pyx_L1_error))
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1349
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    __Pyx_TraceLine(1349,1,__PYX_ERR(0, 1349, __pyx_L1_error))
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1350
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    __Pyx_TraceLine(1350,1,__PYX_ERR(0, 1350, __pyx_L1_error))
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1351
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    __Pyx_TraceLine(1351,1,__PYX_ERR(0, 1351, __pyx_L1_error))
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1353
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __Pyx_TraceLine(1353,1,__PYX_ERR(0, 1353, __pyx_L1_error))
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1354
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    __Pyx_TraceLine(1354,1,__PYX_ERR(0, 1354, __pyx_L1_error))
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1355
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    __Pyx_TraceLine(1355,1,__PYX_ERR(0, 1355, __pyx_L1_error))
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1356
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(1356,1,__PYX_ERR(0, 1356, __pyx_L1_error))
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.broadcast_leading", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("refcount_copying", __pyx_f[0], 1364, 1, __PYX_ERR(0, 1364, __pyx_L1_error));

  /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __Pyx_TraceLine(1368,1,__PYX_ERR(0, 1368, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1369
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __Pyx_TraceLine(1369,1,__PYX_ERR(0, 1369, __pyx_L1_error))
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.refcount_copying", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);
  __Pyx_TraceCall("refcount_objects_in_slice_with_gil", __pyx_f[0], 1373, 0, __PYX_ERR(0, 1373, __pyx_L1_error));

  /* "View.MemoryView":1376
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __Pyx_TraceLine(1376,0,__PYX_ERR(0, 1376, __pyx_L1_error))
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.refcount_objects_in_slice_with_gil", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);
  __Pyx_TraceCall("refcount_objects_in_slice", __pyx_f[0], 1379, 0, __PYX_ERR(0, 1379, __pyx_L1_error));

  /* "View.MemoryView":1383
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __Pyx_TraceLine(1383,0,__PYX_ERR(0, 1383, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __Pyx_TraceLine(1384,0,__PYX_ERR(0, 1384, __pyx_L1_error))
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __Pyx_TraceLine(1385,0,__PYX_ERR(0, 1385, __pyx_L1_error))
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1386
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        __Pyx_TraceLine(1386,0,__PYX_ERR(0, 1386, __pyx_L1_error))
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1388
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      __Pyx_TraceLine(1388,0,__PYX_ERR(0, 1388, __pyx_L1_error))
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1390
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    __Pyx_TraceLine(1390,0,__PYX_ERR(0, 1390, __pyx_L1_error))
    /*else*/ {

      /* "View.MemoryView":1391
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __Pyx_TraceLine(1391,0,__PYX_ERR(0, 1391, __pyx_L1_error))
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1393
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(1393,0,__PYX_ERR(0, 1393, __pyx_L1_error))
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.refcount_objects_in_slice", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {
  __Pyx_TraceDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("slice_assign_scalar", __pyx_f[0], 1399, 1, __PYX_ERR(0, 1399, __pyx_L1_error));

  /* "View.MemoryView":1402
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __Pyx_TraceLine(1402,1,__PYX_ERR(0, 1402, __pyx_L1_error))
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1403
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __Pyx_TraceLine(1403,1,__PYX_ERR(0, 1403, __pyx_L1_error))
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1405
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1405,1,__PYX_ERR(0, 1405, __pyx_L1_error))
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView.slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceCall("_slice_assign_scalar", __pyx_f[0], 1409, 1, __PYX_ERR(0, 1409, __pyx_L1_error));

  /* "View.MemoryView":1413
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __Pyx_TraceLine(1413,1,__PYX_ERR(0, 1413, __pyx_L1_error))
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __Pyx_TraceLine(1414,1,__PYX_ERR(0, 1414, __pyx_L1_error))
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __Pyx_TraceLine(1416,1,__PYX_ERR(0, 1416, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1417
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __Pyx_TraceLine(1417,1,__PYX_ERR(0, 1417, __pyx_L1_error))
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1418
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      __Pyx_TraceLine(1418,1,__PYX_ERR(0, 1418, __pyx_L1_error))
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1419
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __Pyx_TraceLine(1419,1,__PYX_ERR(0, 1419, __pyx_L1_error))
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1421
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  __Pyx_TraceLine(1421,1,__PYX_ERR(0, 1421, __pyx_L1_error))
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1422
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __Pyx_TraceLine(1422,1,__PYX_ERR(0, 1422, __pyx_L1_error))
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1424
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_TraceLine(1424,1,__PYX_ERR(0, 1424, __pyx_L1_error))
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("View.MemoryView._slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(__pyx_codeobj__24)
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);
  __Pyx_TraceCall("__pyx_unpickle_Enum", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */
  __Pyx_TraceLine(4,0,__PYX_ERR(0, 4, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__25, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __Pyx_TraceLine(5,0,__PYX_ERR(0, 5, __pyx_L1_error))
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __Pyx_TraceLine(6,0,__PYX_ERR(0, 6, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __Pyx_TraceLine(7,0,__PYX_ERR(0, 7, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __Pyx_TraceLine(8,0,__PYX_ERR(0, 8, __pyx_L1_error))
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    __Pyx_TraceLine(9,0,__PYX_ERR(0, 9, __pyx_L1_error))
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_TraceLine(10,0,__PYX_ERR(0, 10, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);
  __Pyx_TraceCall("__pyx_unpickle_Enum__set_state", __pyx_f[0], 11, 0, __PYX_ERR(0, 11, __pyx_L1_error));

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  __Pyx_TraceLine(12,0,__PYX_ERR(0, 12, __pyx_L1_error))
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  __Pyx_TraceLine(13,0,__PYX_ERR(0, 13, __pyx_L1_error))
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __Pyx_TraceLine(14,0,__PYX_ERR(0, 14, __pyx_L1_error))
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD;

static PyObject *__pyx_tp_new_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *p;
  PyObject *o = __pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C*)__pyx_vtabptr_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD;
  p->calibration_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD(PyObject *o) {
  struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *p = (struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->calibration_coeffs);
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C)) __pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD);
}

static int __pyx_tp_traverse_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *p = (struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)o;
  e = ((likely(__pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C)) ? ((__pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C->tp_traverse) ? __pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD)); if (e) return e;
  if (p->calibration_coeffs) {
    e = (*v)(((PyObject *)p->calibration_coeffs), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *p = (struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *)o;
  if (likely(__pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C)) { if (__pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C->tp_clear) __pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD);
  tmp = ((PyObject*)p->calibration_coeffs);
  p->calibration_coeffs = ((struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_calibration_coeffs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_1__get__(o);
}

static int __pyx_setprop_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_calibration_coeffs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_3__set__(o, v);
  }
  else {
    return __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_18calibration_coeffs_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD[] = {
  {"_call", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_5_call, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_4_call},
  {"grad", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_7grad, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_6grad},
  {"hessian", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_9hessian, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_8hessian},
  {"dynamics", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_11dynamics, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_10dynamics},
  {"omega", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_13omega, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_12omega},
  {"auxderivs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_15auxderivs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_14auxderivs},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_17__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_19__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD[] = {
  {(char *)"calibration_coeffs", __pyx_getprop_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_calibration_coeffs, __pyx_setprop_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_calibration_coeffs, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Ham_AvgS2precess_simple_cython_PA_AD", /*tp_name*/
  sizeof(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  __pyx_pw_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_3__call__, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD, /*tp_traverse*/
  __pyx_tp_clear_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_array___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryview___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryviewslice___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_Ham_AvgS2precess_simple_cython_PA_AD(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_Ham_AvgS2precess_simple_cython_PA_AD},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "Ham_AvgS2precess_simple_cython_PA_AD",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_n_s_Ham_AvgS2precess_simple_cython_P, __pyx_k_Ham_AvgS2precess_simple_cython_P, sizeof(__pyx_k_Ham_AvgS2precess_simple_cython_P), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_n_u_a6, __pyx_k_a6, sizeof(__pyx_k_a6), 0, 1, 0, 1},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_auxderivs, __pyx_k_auxderivs, sizeof(__pyx_k_auxderivs), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_call, __pyx_k_call, sizeof(__pyx_k_call), 0, 0, 1, 1},
  {&__pyx_n_s_chi1_v, __pyx_k_chi1_v, sizeof(__pyx_k_chi1_v), 0, 0, 1, 1},
  {&__pyx_n_s_chi2_v, __pyx_k_chi2_v, sizeof(__pyx_k_chi2_v), 0, 0, 1, 1},
  {&__pyx_n_s_chiL1, __pyx_k_chiL1, sizeof(__pyx_k_chiL1), 0, 0, 1, 1},
  {&__pyx_n_s_chiL2, __pyx_k_chiL2, sizeof(__pyx_k_chiL2), 0, 0, 1, 1},
  {&__pyx_n_s_chi_1, __pyx_k_chi_1, sizeof(__pyx_k_chi_1), 0, 0, 1, 1},
  {&__pyx_n_s_chi_2, __pyx_k_chi_2, sizeof(__pyx_k_chi_2), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_u_dSO, __pyx_k_dSO, sizeof(__pyx_k_dSO), 0, 1, 0, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_dynamics, __pyx_k_dynamics, sizeof(__pyx_k_dynamics), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_grad, __pyx_k_grad, sizeof(__pyx_k_grad), 0, 0, 1, 1},
  {&__pyx_n_s_hessian, __pyx_k_hessian, sizeof(__pyx_k_hessian), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_m_1, __pyx_k_m_1, sizeof(__pyx_k_m_1), 0, 0, 1, 1},
  {&__pyx_n_s_m_2, __pyx_k_m_2, sizeof(__pyx_k_m_2), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_omega, __pyx_k_omega, sizeof(__pyx_k_omega), 0, 0, 1, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_params, __pyx_k_params, sizeof(__pyx_k_params), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_reshape, __pyx_k_reshape, sizeof(__pyx_k_reshape), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 944, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 152, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 181, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(0, 406, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(0, 615, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 834, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1279
 *         cdef double d2Hdprdpphi = (-(dHeffdpr*dHeffdpphi/H**3)*(M**2/nu) + d2Heffdprdpphi/H) * M*M / nu
 * 
 *         return np.array([d2Hdr2, d2Hdrdphi, d2Hdrdpr, d2Hdrdpphi, d2Hdrdphi, d2Hdphi2, d2Hdphidpr, d2Hdphidpphi, d2Hdrdpr, d2Hdphidpr, d2Hdpr2, d2Hdprdpphi, d2Hdrdpphi, d2Hdphidpphi, d2Hdprdpphi, d2Hdpphi2]).reshape(4, 4)             # <<<<<<<<<<<<<<
 * 
 *     cdef double xi(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):
 */
  __pyx_tuple_ = PyTuple_Pack(2, __pyx_int_4, __pyx_int_4); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(2, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__17 = PyTuple_New(1); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__17, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
  __pyx_slice__20 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__20)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__20);
  __Pyx_GIVEREF(__pyx_slice__20);

  /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_tuple__25 = PyTuple_Pack(3, __pyx_int_184977713, __pyx_int_136983863, __pyx_int_112105877); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__31 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C = __Pyx_ImportType(__pyx_t_1, "pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C", "Hamiltonian_v5PHM_C", sizeof(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_vtabptr_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C = (struct __pyx_vtabstruct_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C*)__Pyx_GetVtable(__pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C->tp_dict); if (unlikely(!__pyx_vtabptr_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_vtabptr_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD = &__pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD;
  __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.__pyx_base = *__pyx_vtabptr_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C;
  __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.__pyx_base._call = (PyObject *(*)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch))__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD__call;
  __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.__pyx_base.grad = (PyObject *(*)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch))__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_grad;
  __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.__pyx_base.hessian = (PyObject *(*)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch))__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_hessian;
  __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.__pyx_base.dynamics = (PyObject *(*)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch))__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_dynamics;
  __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.__pyx_base.omega = (double (*)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch))__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_omega;
  __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.xi = (double (*)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double))__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_xi;
  __pyx_vtable_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.auxderivs = (PyObject *(*)(struct __pyx_obj_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD *, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double, double, double, double, double, double, int __pyx_skip_dispatch))__pyx_f_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_36Ham_AvgS2precess_simple_cython_PA_AD_auxderivs;
  __pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.tp_base = __pyx_ptype_8pyseobnr_3eob_11hamiltonian_19Hamiltonian_v5PHM_C_Hamiltonian_v5PHM_C;
  if (PyType_Ready(&__pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD) < 0) __PYX_ERR(1, 17, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.tp_dictoffset && __pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD.tp_dict, __pyx_vtabptr_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD) < 0) __PYX_ERR(1, 17, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Ham_AvgS2precess_simple_cython_P, (PyObject *)&__pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD) < 0) __PYX_ERR(1, 17, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD) < 0) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_ptype_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD = &__pyx_type_8pyseobnr_3eob_11hamiltonian_36Ham_AvgS2precess_simple_cython_PA_AD_Ham_AvgS2precess_simple_cython_PA_AD;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(0, 280, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(0, 280, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(0, 331, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(0, 331, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(0, 331, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyseobnr.eob.utils.containers"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8pyseobnr_3eob_5utils_10containers_PhysicalParams = __Pyx_ImportType(__pyx_t_1, "pyseobnr.eob.utils.containers", "PhysicalParams", sizeof(struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_PhysicalParams), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8pyseobnr_3eob_5utils_10containers_PhysicalParams) __PYX_ERR(4, 8, __pyx_L1_error)
  __pyx_vtabptr_8pyseobnr_3eob_5utils_10containers_PhysicalParams = (struct __pyx_vtabstruct_8pyseobnr_3eob_5utils_10containers_PhysicalParams*)__Pyx_GetVtable(__pyx_ptype_8pyseobnr_3eob_5utils_10containers_PhysicalParams->tp_dict); if (unlikely(!__pyx_vtabptr_8pyseobnr_3eob_5utils_10containers_PhysicalParams)) __PYX_ERR(4, 8, __pyx_L1_error)
  __pyx_ptype_8pyseobnr_3eob_5utils_10containers_CalibCoeffs = __Pyx_ImportType(__pyx_t_1, "pyseobnr.eob.utils.containers", "CalibCoeffs", sizeof(struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_CalibCoeffs), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8pyseobnr_3eob_5utils_10containers_CalibCoeffs) __PYX_ERR(4, 23, __pyx_L1_error)
  __pyx_ptype_8pyseobnr_3eob_5utils_10containers_Dynamics = __Pyx_ImportType(__pyx_t_1, "pyseobnr.eob.utils.containers", "Dynamics", sizeof(struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_Dynamics), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8pyseobnr_3eob_5utils_10containers_Dynamics) __PYX_ERR(4, 26, __pyx_L1_error)
  __pyx_ptype_8pyseobnr_3eob_5utils_10containers_FluxParams = __Pyx_ImportType(__pyx_t_1, "pyseobnr.eob.utils.containers", "FluxParams", sizeof(struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_FluxParams), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8pyseobnr_3eob_5utils_10containers_FluxParams) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_ptype_8pyseobnr_3eob_5utils_10containers_EOBParams = __Pyx_ImportType(__pyx_t_1, "pyseobnr.eob.utils.containers", "EOBParams", sizeof(struct __pyx_obj_8pyseobnr_3eob_5utils_10containers_EOBParams), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8pyseobnr_3eob_5utils_10containers_EOBParams) __PYX_ERR(4, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initHam_AvgS2precess_simple_cython_PA_AD(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initHam_AvgS2precess_simple_cython_PA_AD(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_Ham_AvgS2precess_simple_cython_PA_AD(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_Ham_AvgS2precess_simple_cython_PA_AD(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_Ham_AvgS2precess_simple_cython_PA_AD(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  __Pyx_TraceDeclarations
  PyObject *__pyx_t_1 = NULL;
  static PyThread_type_lock __pyx_t_2[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'Ham_AvgS2precess_simple_cython_PA_AD' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_Ham_AvgS2precess_simple_cython_PA_AD(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 2, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("Ham_AvgS2precess_simple_cython_PA_AD", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 2, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 2, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 2, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pyseobnr__eob__hamiltonian__Ham_AvgS2precess_simple_cython_PA_AD) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 2, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD")) {
      if (unlikely(PyDict_SetItemString(modules, "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD", __pyx_m) < 0)) __PYX_ERR(1, 2, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(1, 2, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(1, 2, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  #endif
  __Pyx_TraceCall("__Pyx_PyMODINIT_FUNC PyInit_Ham_AvgS2precess_simple_cython_PA_AD(void)", __pyx_f[1], 2, 0, __PYX_ERR(1, 2, __pyx_L1_error));

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":5
 * 
 * cimport cython
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * 
 */
  __Pyx_TraceLine(5,0,__PYX_ERR(1, 5, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":25
 *         return self._call(q,p,chi1_v,chi2_v,m_1,m_2,chi_1,chi_2,chiL1,chiL2)
 * 
 *     cpdef _call(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 *         """
 *         Evaluate the Hamiltonian as well as several potentials.
 */
  __Pyx_TraceLine(25,0,__PYX_ERR(1, 25, __pyx_L1_error))


  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":160
 * 
 * 
 *     cpdef grad(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 *         """
 *         Compute the gradient of the Hamiltonian in polar coordinates.
 */
  __Pyx_TraceLine(160,0,__PYX_ERR(1, 160, __pyx_L1_error))


  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":468
 *         return dHdr, dHdphi, dHdpr, dHdpphi
 * 
 *     cpdef hessian(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */
  __Pyx_TraceLine(468,0,__PYX_ERR(1, 468, __pyx_L1_error))


  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1281
 *         return np.array([d2Hdr2, d2Hdrdphi, d2Hdrdpr, d2Hdrdpphi, d2Hdrdphi, d2Hdphi2, d2Hdphidpr, d2Hdphidpphi, d2Hdrdpr, d2Hdphidpr, d2Hdpr2, d2Hdprdpphi, d2Hdrdpphi, d2Hdphidpphi, d2Hdprdpphi, d2Hdpphi2]).reshape(4, 4)
 * 
 *     cdef double xi(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */
  __Pyx_TraceLine(1281,0,__PYX_ERR(1, 1281, __pyx_L1_error))


  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1348
 *         return xi
 * 
 *     cpdef dynamics(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */
  __Pyx_TraceLine(1348,0,__PYX_ERR(1, 1348, __pyx_L1_error))


  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1657
 *         return dHdr, dHdphi, dHdpr, dHdpphi,H,xi
 * 
 *     cpdef double omega(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */
  __Pyx_TraceLine(1657,0,__PYX_ERR(1, 1657, __pyx_L1_error))


  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":1820
 *         return omega
 * 
 *     cpdef auxderivs(self, double[:]q,double[:]p,double[:]chi1_v,double[:]chi2_v,double m_1,double m_2,double chi_1,double chi_2,double chiL1,double chiL2):             # <<<<<<<<<<<<<<
 * 
 *         """
 */
  __Pyx_TraceLine(1820,0,__PYX_ERR(1, 1820, __pyx_L1_error))


  /* "pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx":2
 * 
 * # cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False,profile=True, linetrace=True             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __Pyx_TraceLine(2,0,__PYX_ERR(1, 2, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */
  __Pyx_TraceLine(734,0,__PYX_ERR(2, 734, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */
  __Pyx_TraceLine(737,0,__PYX_ERR(2, 737, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */
  __Pyx_TraceLine(740,0,__PYX_ERR(2, 740, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */
  __Pyx_TraceLine(743,0,__PYX_ERR(2, 743, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */
  __Pyx_TraceLine(746,0,__PYX_ERR(2, 746, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */
  __Pyx_TraceLine(749,0,__PYX_ERR(2, 749, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */
  __Pyx_TraceLine(928,0,__PYX_ERR(2, 928, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */
  __Pyx_TraceLine(932,0,__PYX_ERR(2, 932, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */
  __Pyx_TraceLine(940,0,__PYX_ERR(2, 940, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */
  __Pyx_TraceLine(946,0,__PYX_ERR(2, 946, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */
  __Pyx_TraceLine(952,0,__PYX_ERR(2, 952, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */
  __Pyx_TraceLine(966,0,__PYX_ERR(2, 966, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */
  __Pyx_TraceLine(981,0,__PYX_ERR(2, 981, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */
  __Pyx_TraceLine(996,0,__PYX_ERR(2, 996, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */
  __Pyx_TraceLine(1006,0,__PYX_ERR(2, 1006, __pyx_L1_error))


  /* "../../../../local/tmp/sossokine/build-env-10if7lw0/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */
  __Pyx_TraceLine(1013,0,__PYX_ERR(2, 1013, __pyx_L1_error))


  /* "View.MemoryView":210
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __Pyx_TraceLine(210,0,__PYX_ERR(0, 210, __pyx_L1_error))
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */
  __Pyx_TraceLine(227,0,__PYX_ERR(0, 227, __pyx_L1_error))


  /* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */
  __Pyx_TraceLine(245,0,__PYX_ERR(0, 245, __pyx_L1_error))


  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __Pyx_TraceLine(287,0,__PYX_ERR(0, 287, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __Pyx_TraceLine(288,0,__PYX_ERR(0, 288, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(289,0,__PYX_ERR(0, 289, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __Pyx_TraceLine(292,0,__PYX_ERR(0, 292, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(293,0,__PYX_ERR(0, 293, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */
  __Pyx_TraceLine(299,0,__PYX_ERR(0, 299, __pyx_L1_error))


  /* "View.MemoryView":317
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __Pyx_TraceLine(317,0,__PYX_ERR(0, 317, __pyx_L1_error))
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":318
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __Pyx_TraceLine(318,0,__PYX_ERR(0, 318, __pyx_L1_error))
  __pyx_t_2[0] = PyThread_allocate_lock();
  __pyx_t_2[1] = PyThread_allocate_lock();
  __pyx_t_2[2] = PyThread_allocate_lock();
  __pyx_t_2[3] = PyThread_allocate_lock();
  __pyx_t_2[4] = PyThread_allocate_lock();
  __pyx_t_2[5] = PyThread_allocate_lock();
  __pyx_t_2[6] = PyThread_allocate_lock();
  __pyx_t_2[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_2, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */
  __Pyx_TraceLine(395,0,__PYX_ERR(0, 395, __pyx_L1_error))


  /* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */
  __Pyx_TraceLine(433,0,__PYX_ERR(0, 433, __pyx_L1_error))


  /* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */
  __Pyx_TraceLine(443,0,__PYX_ERR(0, 443, __pyx_L1_error))


  /* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */
  __Pyx_TraceLine(451,0,__PYX_ERR(0, 451, __pyx_L1_error))


  /* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */
  __Pyx_TraceLine(483,0,__PYX_ERR(0, 483, __pyx_L1_error))


  /* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */
  __Pyx_TraceLine(487,0,__PYX_ERR(0, 487, __pyx_L1_error))


  /* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */
  __Pyx_TraceLine(503,0,__PYX_ERR(0, 503, __pyx_L1_error))


  /* "View.MemoryView":551
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(551,0,__PYX_ERR(0, 551, __pyx_L1_error))
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */
  __Pyx_TraceLine(659,0,__PYX_ERR(0, 659, __pyx_L1_error))


  /* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */
  __Pyx_TraceLine(665,0,__PYX_ERR(0, 665, __pyx_L1_error))


  /* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */
  __Pyx_TraceLine(668,0,__PYX_ERR(0, 668, __pyx_L1_error))


  /* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __Pyx_TraceLine(702,0,__PYX_ERR(0, 702, __pyx_L1_error))


  /* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */
  __Pyx_TraceLine(712,0,__PYX_ERR(0, 712, __pyx_L1_error))


  /* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */
  __Pyx_TraceLine(809,0,__PYX_ERR(0, 809, __pyx_L1_error))


  /* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */
  __Pyx_TraceLine(912,0,__PYX_ERR(0, 912, __pyx_L1_error))


  /* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */
  __Pyx_TraceLine(945,0,__PYX_ERR(0, 945, __pyx_L1_error))


  /* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */
  __Pyx_TraceLine(981,0,__PYX_ERR(0, 981, __pyx_L1_error))


  /* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */
  __Pyx_TraceLine(987,0,__PYX_ERR(0, 987, __pyx_L1_error))


  /* "View.MemoryView":997
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(997,0,__PYX_ERR(0, 997, __pyx_L1_error))
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */
  __Pyx_TraceLine(1001,0,__PYX_ERR(0, 1001, __pyx_L1_error))


  /* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */
  __Pyx_TraceLine(1054,0,__PYX_ERR(0, 1054, __pyx_L1_error))


  /* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */
  __Pyx_TraceLine(1065,0,__PYX_ERR(0, 1065, __pyx_L1_error))


  /* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */
  __Pyx_TraceLine(1082,0,__PYX_ERR(0, 1082, __pyx_L1_error))


  /* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */
  __Pyx_TraceLine(1089,0,__PYX_ERR(0, 1089, __pyx_L1_error))


  /* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */
  __Pyx_TraceLine(1111,0,__PYX_ERR(0, 1111, __pyx_L1_error))


  /* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */
  __Pyx_TraceLine(1118,0,__PYX_ERR(0, 1118, __pyx_L1_error))


  /* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */
  __Pyx_TraceLine(1172,0,__PYX_ERR(0, 1172, __pyx_L1_error))


  /* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */
  __Pyx_TraceLine(1179,0,__PYX_ERR(0, 1179, __pyx_L1_error))


  /* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */
  __Pyx_TraceLine(1189,0,__PYX_ERR(0, 1189, __pyx_L1_error))


  /* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */
  __Pyx_TraceLine(1210,0,__PYX_ERR(0, 1210, __pyx_L1_error))


  /* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */
  __Pyx_TraceLine(1253,0,__PYX_ERR(0, 1253, __pyx_L1_error))


  /* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */
  __Pyx_TraceLine(1259,0,__PYX_ERR(0, 1259, __pyx_L1_error))


  /* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */
  __Pyx_TraceLine(1263,0,__PYX_ERR(0, 1263, __pyx_L1_error))


  /* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */
  __Pyx_TraceLine(1270,0,__PYX_ERR(0, 1270, __pyx_L1_error))


  /* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */
  __Pyx_TraceLine(1342,0,__PYX_ERR(0, 1342, __pyx_L1_error))


  /* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */
  __Pyx_TraceLine(1364,0,__PYX_ERR(0, 1364, __pyx_L1_error))


  /* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */
  __Pyx_TraceLine(1373,0,__PYX_ERR(0, 1373, __pyx_L1_error))


  /* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */
  __Pyx_TraceLine(1379,0,__PYX_ERR(0, 1379, __pyx_L1_error))


  /* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */
  __Pyx_TraceLine(1399,0,__PYX_ERR(0, 1399, __pyx_L1_error))


  /* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */
  __Pyx_TraceLine(1409,0,__PYX_ERR(0, 1409, __pyx_L1_error))


  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __Pyx_TraceLine(1,0,__PYX_ERR(0, 1, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_TraceLine(11,0,__PYX_ERR(0, 11, __pyx_L1_error))

  __Pyx_TraceReturn(Py_None, 0);

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* Profile */
#if CYTHON_PROFILE
static int __Pyx_TraceSetupAndCall(PyCodeObject** code,
                                   PyFrameObject** frame,
                                   PyThreadState* tstate,
                                   const char *funcname,
                                   const char *srcfile,
                                   int firstlineno) {
    PyObject *type, *value, *traceback;
    int retval;
    if (*frame == NULL || !CYTHON_PROFILE_REUSE_FRAME) {
        if (*code == NULL) {
            *code = __Pyx_createFrameCodeObject(funcname, srcfile, firstlineno);
            if (*code == NULL) return 0;
        }
        *frame = PyFrame_New(
            tstate,                          /*PyThreadState *tstate*/
            *code,                           /*PyCodeObject *code*/
            __pyx_d,                  /*PyObject *globals*/
            0                                /*PyObject *locals*/
        );
        if (*frame == NULL) return 0;
        if (CYTHON_TRACE && (*frame)->f_trace == NULL) {
            Py_INCREF(Py_None);
            (*frame)->f_trace = Py_None;
        }
#if PY_VERSION_HEX < 0x030400B1
    } else {
        (*frame)->f_tstate = tstate;
#endif
    }
    __Pyx_PyFrame_SetLineNumber(*frame, firstlineno);
    retval = 1;
    __Pyx_EnterTracing(tstate);
    __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
    #if CYTHON_TRACE
    if (tstate->c_tracefunc)
        retval = tstate->c_tracefunc(tstate->c_traceobj, *frame, PyTrace_CALL, NULL) == 0;
    if (retval && tstate->c_profilefunc)
    #endif
        retval = tstate->c_profilefunc(tstate->c_profileobj, *frame, PyTrace_CALL, NULL) == 0;
    __Pyx_LeaveTracing(tstate);
    if (retval) {
        __Pyx_ErrRestoreInState(tstate, type, value, traceback);
        return __Pyx_IsTracing(tstate, 0, 0) && retval;
    } else {
        Py_XDECREF(type);
        Py_XDECREF(value);
        Py_XDECREF(traceback);
        return -1;
    }
}
static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno) {
    PyCodeObject *py_code = 0;
#if PY_MAJOR_VERSION >= 3
    py_code = PyCode_NewEmpty(srcfile, funcname, firstlineno);
    if (likely(py_code)) {
        py_code->co_flags |= CO_OPTIMIZED | CO_NEWLOCALS;
    }
#else
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    py_funcname = PyString_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    py_srcfile = PyString_FromString(srcfile);
    if (unlikely(!py_srcfile)) goto bad;
    py_code = PyCode_New(
        0,
        0,
        0,
        CO_OPTIMIZED | CO_NEWLOCALS,
        __pyx_empty_bytes,     /*PyObject *code,*/
        __pyx_empty_tuple,     /*PyObject *consts,*/
        __pyx_empty_tuple,     /*PyObject *names,*/
        __pyx_empty_tuple,     /*PyObject *varnames,*/
        __pyx_empty_tuple,     /*PyObject *freevars,*/
        __pyx_empty_tuple,     /*PyObject *cellvars,*/
        py_srcfile,       /*PyObject *filename,*/
        py_funcname,      /*PyObject *name,*/
        firstlineno,
        __pyx_empty_bytes      /*PyObject *lnotab*/
    );
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
#endif
    return py_code;
}
#endif

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#if PY_VERSION_HEX >= 0x030A0000 || defined(HAVE_STDARG_PROTOTYPES)
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


/* MemviewSliceIsContig */
static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* TypeInfoCompare */
  static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
  static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    if (buf->len > 0) {
        for (i = 0; i < ndim; i++) {
            spec = axes_specs[i];
            if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
                goto fail;
            if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
                goto fail;
        }
        if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
            goto fail;
    }
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(double *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj) {
    double value = __pyx_PyFloat_AsDouble(obj);
    if ((value == (double)-1) && PyErr_Occurred())
        return 0;
    *(double *) itemp = value;
    return 1;
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* MemviewSliceCopyTemplate */
  static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
